/*
 * plugin.h: API agreement between ieee1905 plugin manager and plugins
 *
 * Author: Vivek Dutta <vivek.dutta@iopsys.eu>
 *
 * Copyright (C) 2020, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef _PLUGIN_H_
#define _PLUGIN_H_

#include <stdint.h>
#include <linux/types.h>

#ifndef IFNAMELEN
#define IFNAMELEN 16
#endif

enum {
	PLUGIN_FUNC_PROCESS,
	PLUGIN_FUNC_VALIDATE,
	PLUGIN_FUNC_NUM,
};

typedef int (*plugin_init_t)(void *priv);
typedef int (*plugin_exit_t)(void *priv);
typedef int (*plugin_func_t)(void *priv);

struct plugin {
	char name[128];
	uint8_t id[IFNAMELEN];
	plugin_init_t init;
	plugin_exit_t exit;
	plugin_func_t ops[PLUGIN_FUNC_NUM];
	void *handle;
};

/**
 * @brief data exchanged between plugin-manager and map-plugin,
 *
 * Incoming packets
 * intf_name => name of receive interface name
 * mac_ascii => origin mac address in hex string 'xx:xx:xx:xx:xx:xx' format
 * cstream   => cmdu stream
 * profile   => Registered map-module profile
 * flag      => True, if map-controller also registered
 *
 * Outgoing packets
 * intf_name => name of out interface name
 * mac_ascii => destination mac address in hex string 'xx:xx:xx:xx:xx:xx' format
 * cstream   => cmdu stream
 * flag      => True, if message_id needs to be regenerated before sending
 * rel_multicast_flag => True, if the message is a reliable multicast 
 *
 */
struct plugin_process_data {
	char intf_name[16];
	char mac_ascii[18];
	struct cmdu_stream *cstream;
	uint8_t profile;
	bool flag;
	bool rel_multicast_flag;
};

/**
 * @brief The ieee1905 multiap plugin manager(ieee1905 stack) will implement
 * this function. Whenever a multiap module has any CMDU (request/response)
 * to transmit, it will call this function.
 *
 * @param data cmdu related information
 *
 * @return boolean true, if message sent successfully, otherwise false
 */
extern bool cmdu_dispacher(struct plugin_process_data *data);

/**
 * @brief The ieee1905 multiap plugin manager(ieee1905 stack) will implement
 * this function. Whenever a multiap module get registered with map-plugin
 * this api will update the plugin-manager.
 *
 * @param set true, to set the values. false, to clear the values
 * @param role_bit 0x1 for map-agent, 0x2 for map-controller
 * @param profile 0 for map profile-1, 1 for map profile-2
 */
extern void update_module_db(bool set, uint8_t role_bit, uint8_t profile);

#endif
