/*
 * 1905_hle.c: IEEE-1905 hle stack
 *
 * Author: Vivek Dutta <vivek.dutta@iopsys.eu>
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "1905_hle.h"
#include "1905_alme.h"

#include <stdio.h>		// printf
#include <stdlib.h>		// malloc
#include <unistd.h>		// getopt
#include <string.h>		// strtok

#include <arpa/inet.h>		// socket(), AF_INET, htons(), ...

#include <errno.h>		// errno
#include <easy/easy.h>

////////////////////////////////////////////////////////////////////////////////
// Static (auxiliary) private functions, structures and macros
////////////////////////////////////////////////////////////////////////////////

// Return a properly filled structure representing the desired ALME REQUEST
// Some types of ALME requests require arguments. These are taken from the
// arguments the executable was called with.
//
static INT8U *_build_alme_request(int alme_request_type, char *data)
{
	INT8U *ret = NULL;

	switch (alme_request_type) {
	case ALME_TYPE_GET_INTF_LIST_REQUEST:
		{
			struct getIntfListRequestALME *p;

			p = (struct getIntfListRequestALME *)malloc(sizeof(struct getIntfListRequestALME));
			if (NULL == p) {
				ERROR("Could not allocate getIntfListRequestALME structure!");
				return NULL;
			}
			p->alme_type = ALME_TYPE_GET_INTF_LIST_REQUEST;

			ret = (INT8U *) p;
			break;
		}
	case ALME_TYPE_GET_METRIC_REQUEST:
		{
			struct getMetricRequestALME *p;

			INT8U mac_address[6] = {0x00};

			// When an ALME-GET-METRIC.request is solicited, the user can either
			// ask for a specific neighbor (in that case an extra argument is
			// provided) or for all neighbors (in that case no extra argument is
			// provided).
			//
			if (NULL != data) {
				hwaddr_aton(data, mac_address);
			}

			p = (struct getMetricRequestALME *)malloc(sizeof(struct getMetricRequestALME));
			if (NULL == p) {
				ERROR("Could not allocate getMetricRequestALME structure!");
				return NULL;
			}
			p->alme_type = ALME_TYPE_GET_METRIC_REQUEST;
			memcpy(p->interface_address, mac_address, 6);

			ret = (INT8U *) p;
			break;
		}
	case ALME_TYPE_CUSTOM_COMMAND_REQUEST:
		{
			struct customCommandRequestALME *p;

			if (NULL == data) {
				// No extra argument was provided
				//
				ERROR("Invalid arguments for 'ALME-CUSTOM-COMMAND' message");
				return NULL;
			}

			p = (struct customCommandRequestALME *)malloc(sizeof(struct customCommandRequestALME));
			if (NULL == p) {
				ERROR("Could not allocate customCommandRequestALME structure!");
				return NULL;
			}
			p->alme_type = ALME_TYPE_CUSTOM_COMMAND_REQUEST;

			if (data != NULL) {
				if (0 == strcmp(data, "dnd")) {
					p->command = CUSTOM_COMMAND_DUMP_NETWORK_DEVICES;
				} else {
					ERROR("Invalid arguments for 'ALME-CUSTOM-COMMAND' message");
					free(p);
					return NULL;
				}
			}

			ret = (INT8U *) p;
			break;
		}
	default:
		{
			ERROR("ERROR: Unknown ALME message type: %d", alme_request_type);
			break;
		}
	}			// end of switch

	return ret;
}

// Sends an ALME REQUEST message to an AL entity:
//
//   - 'server_ip_and_port' is a string containing the "IP:port" where the AL
//     TCP server is listening (ex: "10.32.1.44:8888")
//
//   - 'alme_request' is a pointer to the ALME REQUEST payload (as generated by
//     "forge_1905_ALME_from_structure()")
//
//   - 'alme_request_len' is the number of bytes of 'alme_request'
//
//   - 'alme_reply' is a pointer to a buffer which is 'alme_reply_len' bytes
//      long where the response from the AL entity (either an ALME RESPONSE or
//      an ALME CONFIRMATION message) will be placed.
//
//   - 'alme_request_len' is the capacity of the 'alme_reply' buffer. It is also
//     used as an output argument that will contain the actual length of the
//     reply.
//
// Note that the caller is responsible for freeing both 'alme_request' and
// 'alme_response' after they are no longer needed (ie. this function does not
// allocate/free memory at all).
//
static int _sendAlmeRequestAndWaitForReply(char *server_ip_and_port, INT8U * alme_request, int alme_request_len, INT8U * alme_reply, int *alme_reply_len)
{
	int sock;

	struct sockaddr_in server;

	ssize_t total_sent;

	ssize_t received;
	ssize_t total_received;

	char *aux;
	char *ip;
	char *port;

	aux = strdup(server_ip_and_port);
	ip = strtok(aux, ":");
	port = strtok(NULL, ":");

	if (NULL == ip || NULL == port) {
		ERROR("Invalid address format. Must follow this template: '<ip_address>:<port_number>'");
		return 0;
	}
	//Create socket
	//
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (-1 == sock) {
		ERROR("socket() failed with errno=%d (%s)", errno, strerror(errno));
		return 0;
	}
	//Connect to remote server
	//
	memset(&server, 0, sizeof(server));
	server.sin_addr.s_addr = inet_addr(ip);
	server.sin_family = AF_INET;
	server.sin_port = htons(atoi(port));

	free(aux);

	if (connect(sock, (struct sockaddr *)&server, sizeof(server)) < 0) {
		ERROR("connect() failed with errno=%d (%s)", errno, strerror(errno));
		return 0;
	}
	// Send the ALME REQUEST message
	//
	INFO("Sending ALME request message (%d byte(s) long)...", alme_request_len);
	total_sent = 0;
	do {
		ssize_t sent;

		sent = send(sock, alme_request, alme_request_len, 0);

		if (-1 == sent) {
			ERROR("send() failed with errno=%d (%s)", errno, strerror(errno));
			return 0;
		}
		INFO("%zd byte(s) sent", sent);

		total_sent += sent;
	} while (total_sent < alme_request_len);

	// Close the socket for writing. This will inform the other end that we
	// are done (ie. send a "FIN" TCP packet)
	//
	INFO("ALME request sent. Closing writing end of the socket descriptor...");
	if (-1 == shutdown(sock, SHUT_WR)) {
		ERROR("shutdown(\"SHUT_WR\") failed with errno=%d (%s)", errno, strerror(errno));
		return 0;
	}
	// Receive a reply from the server
	//
	INFO("Waiting for the ALME reply...");
	total_received = 0;
	while ((received = recv(sock, alme_reply + total_received, *alme_reply_len - total_received, 0)) > 0) {
		// Keep reading until the server closes the connection
		//
		if (-1 == received) {
			ERROR("recv() failed with errno=%d (%s)", errno, strerror(errno));
			return 0;
		}
		INFO("%zd byte(s) received", received);

		total_received += received;

		if (total_received >= *alme_reply_len) {
			// This message is too big.
			//
			ERROR("Error! Received message is too big");
			return 0;
		}
	}

	INFO("ALME reply received (%zd bytes in total). Closing socket...", total_received);
	if (-1 == close(sock)) {
		ERROR("close() failed with errno=%d (%s)", errno, strerror(errno));
		return 0;
	}

	*alme_reply_len = total_received;

	return 1;
}

////////////////////////////////////////////////////////////////////////////////
// External public functions
////////////////////////////////////////////////////////////////////////////////
void set_hle_debug_level(unsigned char level)
{
	set_print_level(level);
}

int execute_hle_cmd(char *al_ip_port, unsigned short alme_request_type, char *data)
{
	INT8U *alme_request_structure = NULL;
	INT8U *alme_request_payload = NULL;
	INT16U alme_request_payload_len = 0;

#ifndef MAX_NETWORK_SEGMENT_SIZE
#define MAX_REPLY_SIZE (100*MAX_NETWORK_SEGMENT_SIZE)
#endif

	INT8U *alme_reply_structure;
	INT8U alme_reply_payload[MAX_REPLY_SIZE];
	int alme_reply_payload_len = MAX_REPLY_SIZE;

	char aux[300 * 1024];

	int i;

	if (NULL == al_ip_port) {
		ERROR("ERROR: You *must* provide an AL address (example: '10.9.123.1:9077')");
		return (1);
	}
	// Build the ALME structure and print it to stdout
	//
	alme_request_structure = _build_alme_request(alme_request_type, data);
	if (NULL == alme_request_structure) {
		ERROR("ERROR: The ALME REQUEST structure could not be build.");
		return (1);
	}
	INFO("Displaying contents of the ALME REQUEST that is going to be sent:");
	visit_1905_ALME_structure(alme_request_structure, print_callback, print_info, "", NULL);

	// From the structure, generate a bit stream
	//
	alme_request_payload = forge_1905_ALME_from_structure(alme_request_structure, &alme_request_payload_len);
	if (NULL == alme_request_payload) {
		ERROR("ERROR: The ALME REQUEST payload could not be build.");
		return (1);
	}
	INFO("Displaying the bit stream associated to this ALME REQUEST structure (%d byte(s) long):", alme_request_payload_len);
	aux[0] = 0;
	for (i = 0; i < alme_request_payload_len; i++) {
		char item[10];

		sprintf(item, "0x%02x ", alme_request_payload[i]);
		strcat(aux, item);
	}
	INFO("%s", aux);
	free_1905_ALME_structure(alme_request_structure);

	// Send that bit stream to the AL entity and wait for a response
	//
	INFO("Sending bit stream to %s (len = %d)...", al_ip_port, alme_request_payload_len);
	if (0 == _sendAlmeRequestAndWaitForReply(al_ip_port, alme_request_payload, alme_request_payload_len, alme_reply_payload, &alme_reply_payload_len)) {
		ERROR("ERROR: AL communication problem");
		return (1);
	}
	free_1905_ALME_packet(alme_request_payload);

	INFO("Displaying bit stream associated to the ALME RESPONSE/CONFIRMATION structure (%d byte(s) long):", alme_reply_payload_len);
	aux[0] = 0;
	for (i = 0; i < alme_reply_payload_len; i++) {
		char item[10];

		sprintf(item, "0x%02x ", alme_reply_payload[i]);
		strcat(aux, item);
	}
	INFO("%s", aux);

	// Convert the response back into a structure and print it to stdout
	//
	alme_reply_structure = parse_1905_ALME_from_packet(alme_reply_payload);
	if (NULL == alme_reply_structure) {
		ERROR("ERROR: Cannot parse ALME RESPONSE/CONFIRMATION");
	}
	visit_1905_ALME_structure(alme_reply_structure, print_callback, print_debug, "", NULL);

	return 0;
}
