/*
 * 1905_cmdus.h: IEEE-1905 AL cmdu related functions
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef _1905_CDMUS_H_
#define _1905_CDMUS_H_
#include <stdint.h>
#include <stdbool.h>
#include <linux/types.h>

#ifndef MAX_NETWORK_SEGMENT_SIZE
#define MAX_NETWORK_SEGMENT_SIZE 1500
#endif

////////////////////////////////////////////////////////////////////////////////
// CMDU message type as detailed in "IEEE Std 1905.1-2013, Table 6-4"
////////////////////////////////////////////////////////////////////////////////
#define CMDU_TYPE_TOPOLOGY_DISCOVERY               0x0000
#define CMDU_TYPE_TOPOLOGY_NOTIFICATION            0x0001
#define CMDU_TYPE_TOPOLOGY_QUERY                   0x0002
#define CMDU_TYPE_TOPOLOGY_RESPONSE                0x0003
#define CMDU_TYPE_VENDOR_SPECIFIC                  0x0004
#define CMDU_TYPE_LINK_METRIC_QUERY                0x0005
#define CMDU_TYPE_LINK_METRIC_RESPONSE             0x0006
#define CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH      0x0007
#define CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE    0x0008
#define CMDU_TYPE_AP_AUTOCONFIGURATION_WSC         0x0009
#define CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW       0x000a
#define CMDU_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION   0x000b
#define CMDU_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION    0x000c
#define CMDU_TYPE_HIGHER_LAYER_QUERY               0x000d
#define CMDU_TYPE_HIGHER_LAYER_RESPONSE            0x000e
#define CMDU_TYPE_INTERFACE_POWER_CHANGE_REQUEST   0x000f
#define CMDU_TYPE_INTERFACE_POWER_CHANGE_RESPONSE  0x0010
#define CMDU_TYPE_GENERIC_PHY_QUERY                0x0011
#define CMDU_TYPE_GENERIC_PHY_RESPONSE             0x0012
#define LAST_1905_CMDU                             CMDU_TYPE_GENERIC_PHY_RESPONSE

////////////////////////////////////////////////////////////////////////////////
// CMDU message version
////////////////////////////////////////////////////////////////////////////////
#define CMDU_MESSAGE_VERSION_1905_1_2013  (0x00)

////////////////////////////////////////////////////////////////////////////////
// CMDU associated structures
////////////////////////////////////////////////////////////////////////////////
struct cmdu_cstruct {
	uint8_t message_version;
	// One of "CMDU_MESSAGE_VERSION_*" values

	uint16_t message_type;
	// Any of the CMDU_TYPE_* types

	uint16_t message_id;
	// Identifies the message

	uint8_t relay_indicator;
	// Set to '1' to indicate that his packet
	// must be relayed by 1905 AL to its
	// neighbors.

	uint8_t **tlv_cstruct_list;
	// NULL-terminated list of pointers to TLV
	// structures.
	// The "end of message" TLV is not included
	// in this list.
};

struct cmdu_stream {
	uint8_t message_version;
	// One of "CMDU_MESSAGE_VERSION_*" values

	uint16_t message_type;
	// Any of the CMDU_TYPE_* types

	uint16_t message_id;
	// Identifies the message

	uint8_t relay_indicator;
	// Set to '1' to indicate that his packet
	// must be relayed by 1905 AL to its
	// neighbors.

	uint8_t nbr_tlv_streams;

	uint8_t **tlv_stream_list;
	// contains nbr_tlv_streams number of binary streams
	// each corresponds to one specific tlv.
	// The "end of message" TLV is not included
	// in this list.
};
////////////////////////////////////////////////////////////////////////////////
// Main API functions
////////////////////////////////////////////////////////////////////////////////

bool check_1905_cmdu_rules(struct cmdu_cstruct *cmdu);
// This function receives a NULL-terminated list of pointers to streams
// containing ETH layer packet data (ie. offset +14 in a raw network packet,
// just after the source MAC address, destination MAC address and ETH type
// fields).
//
// The payload of each stream must contain a 1905 CMDU as detailed in ""IEEE
// Std 1905.1-2013, Section 6.2 (Table 6-3 and beyond)"
//
// The list of received streams must "belong" to one same "message":
//   * The 'message_id' must be the same one in all of them
//   * The 'last_fragment_indicator' flag must be set to '1' in *exactly* one
//     (and only one!) of them.
//
// Note that the streams do not need to be provided in order. The function will
// take care of that.
//
// In the typical case where there is no fragmentation, this function will
// receive a single stream (actually, a list of two pointers: one pointing to
// the stream and another one set to NULL to mark the end of the list) with the
// 'last_fragment_indicator' set to '1', otherwise it will receive a list of
// streams where only the last one has 'last_fragment_indicator' set to '1' and
// all of them share the same 'message_id'.
//
// For each possible CMDU message type, the following list of TLVs is returned:
//
//   CMDU_TYPE_TOPOLOGY_DISCOVERY
//     - 1 TLV_TYPE_AL_MAC_ADDRESS_TYPE
//     - 1 TLV_TYPE_MAC_ADDRESS_TYPE
//
//   CMDU_TYPE_TOPOLOGY_NOTIFICATION
//     - 1 TLV_TYPE_AL_MAC_ADDRESS_TYPE
//
//   CMDU_TYPE_TOPOLOGY_QUERY
//     - <no requirements>
//
//   CMDU_TYPE_TOPOLOGY_RESPONSE
//     - 1 TLV_TYPE_DEVICE_INFORMATION_TYPE
//     - 0 or more TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES
//     - 0 or more TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST
//     - 0 or more TLV_TYPE_NEIGHBOR_DEVICE_LIST
//
//   CMDU_TYPE_VENDOR_SPECIFIC
//     - 1 TLV_TYPE_VENDOR_SPECIFIC (must be the first one received!)
//     - 0 or more TLVs of any type (always after the previous one)
//
//   CMDU_TYPE_LINK_METRIC_QUERY
//     - 1 TLV_TYPE_LINK_METRIC_QUERY
//
//   CMDU_TYPE_LINK_METRIC_RESPONSE
//     - For each valid 1905 neighbor one of the followings options
//       - 1 TLV_TYPE_TRANSMITTER_LINK_METRIC
//       - 1 TLV_TYPE_RECEIVER_LINK_METRIC
//       - 1 TLV_TYPE_TRANSMITTER_LINK_METRIC *and* 1 TLV_TYPE_RECEIVER_LINK_METRIC
//
//   CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH
//     - 1 TLV_TYPE_AL_MAC_ADDRESS_TYPE
//     - 1 TLV_TYPE_SEARCHED_ROLE
//     - 1 TLV_TYPE_AUTOCONFIG_FREQ_BAND
//
//   CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE
//     - 1 TLV_TYPE_SUPPORTED_ROLE
//     - 1 TLV_TYPE_SUPPORTED_FREQ_BAND
//
//   CMDU_TYPE_AP_AUTOCONFIGURATION_WSC
//     - 1 TLV_TYPE_WSC
//
//   CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW
//     - 1 TLV_TYPE_AL_MAC_ADDRESS_TYPE
//     - 1 TLV_TYPE_SUPPORTED_ROLE
//     - 1 TLV_TYPE_SUPPORTED_FREQ_BAND
//
//   CMDU_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION
//     - 1 TLV_TYPE_AL_MAC_ADDRESS_TYPE
//     - 1 TLV_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION
//
//   CMDU_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION
//     - 1 TLV_TYPE_AL_MAC_ADDRESS_TYPE
//     - 1 TLV_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION
//
//   CMDU_AP_CAPABILITY_QUERY
//     < NO TLV >
//
//   CMDU_AP_CAPABILITY_REPORT
//     - 1 AP CAPABILITY TLV
//     - 1 or more AP RADIO BASIC CAPABILITIES TLV
//     - 0 or more AP HT CAPABILITIES TLV
//     - 0 or more AP VHT CAPABILITIES TLV
//     - 0 or more AP HE CAPABILITIES TLV
//
// In addition, *all* of them must *always* end with one TLV_TYPE_END_OF_MESSAGE
// TLV.
//
// If the aggregated payload of all the streams does not contain one of the TLVs
// required for its 'message_type', this function returns an error (a NULL
// pointer).
//
// If the aggregated payload of all the streams contains more TLVs than those
// required for its 'message_type', they are ignored (ie. not contained in the
// returned structure)
//
// The returned structure will then contain:
//   - 'message_version' set to the same value contained in all streams
//   - 'message_type' set to the same value contained in all streams
//   - 'message_id' set to the same value contained in all streams
//   - 'relay_indicator' set to the same value contained in all streams
//   - 'tlv_cstruct_list' is a NULL-terminated list of pointers to the TLVs as
//     explained before (note that TLV_TYPE_END_OF_MESSAGE is *not* included in
//     this list)
//
// If any type of error/inconsistency is found, a NULL pointer is returned
// instead, otherwise remember to free the received structure once you don't
// need it anymore (using the "free_1905_CMDU_structure()" function)
//
//   NOTE: An 'inconsistency' happens when one of the fragment, for example,
//         contains a different 'message_version' from another fragment or if
//         two streams contain fragments with the same 'fragment_id', etc...
//
struct cmdu_cstruct *parse_1905_CMDU_from_packets(uint8_t **packet_streams);

// This is the opposite of "parse_1905_CMDU_from_packets()": it receives a
// pointer to a TLV structure and then returns a list of pointers to fragmented
// streams.
//
// Note that if everything fits under MAX_NETWORK_SEGMENT_SIZE size, only one
// stream will be returned (actually, what will be returned is a list of two
// elements, containing the pointer to the stream and a NULL entry)
//
// Fragments will be returned in order so that only the last one will contain
// the 'last_fragment_indicator' flag set.
//
// 'lens' is also an output argument: it is dynamically allocated by this
// function and contains the number of bytes of each returned stream and a '0'
// element at the end.
//
// The provided 'memory_structure' must have its fields properly filled or else
// this function will return an error (ie. a NULL pointer).
// What does this mean? Easy: for each type of message only the expected TLVs
// can be included in the 'tlv_cstruct_list' list.
// The list of acceptable values for each message type is detailed in the
// "parse_1905_CMDU_from_packets()" function documentation.
//
//   NOTE: Notice how, in "parse_1905_CMDU_from_packets()", unexpected TLVs are
//         discarded, but here they produce an error. This is expected!
//
// In addition, the 'relay_indicator' field will only be respected when
// 'message_type' == CMDU_TYPE_VENDOR_SPECIFIC. In all other cases the value
// that is defined in the standard will be used instead. So don't worry about
// setting it to the correct value (it will be done automatically for you).
//
// One more thing: the provided 'memory_structure' is not freed by this
// function, thus, once this function returns, you will be responsible for
// freeing three things:
//
//   - The 'memory_structure' CMDU structure (you were already responsible for),
//     that you can free with 'free_1905_CMDU_structure()'
//
//   - The 'lens' array, which you can later free with "FREE()"
//
//   - The returned streams and the array where they are contained. Both can
//     be freed with 'free_1905_CMDU_packets()'
//
//
uint8_t **forge_1905_CMDU_from_structure(struct cmdu_cstruct *memory_structure, uint16_t **lens);

////////////////////////////////////////////////////////////////////////////////
// Utility API functions
////////////////////////////////////////////////////////////////////////////////

// Return the 'mid', 'fragment_id' and 'last_fragment_indicator' of the CMDU
// contained in the given 'stream' in the provided output variables.
//
// Return "0" if an error preventing the parsing takes place, "1" otherwise.
//
bool get_cmdu_header_from_packet(uint8_t *stream, uint16_t *mid,
					 uint8_t *fragment_id, uint8_t *last_fragment_indicator);

// This function receives a pointer to a CMDU structure and then traverses it
// and all nested structures, calling "FREE()" on each one of them
//
//
void free_1905_CMDU_structure(struct cmdu_cstruct *memory_structure);
void free_tlv_list_CMDU_structure(struct cmdu_cstruct *memory_structure);
void free_tlv_internal_list_CMDU_structure(struct cmdu_cstruct *memory_structure);

// This function receives a pointer to a list of streams (such as the one
// returned by 'forge_1905_CMDU_from_structure()' and frees all the associated
// structures
//
//
void free_1905_CMDU_packets(uint8_t **packet_streams);

// This function returns '0' if the two given pointers represent CMDU structures
// that contain the same data
//
uint8_t compare_1905_CMDU_structures(struct cmdu_cstruct *memory_structure_1, struct cmdu_cstruct *memory_structure_2);

// Use this function for debug purposes. It turns a CMDU_TYPE_* variable into its
// string representation.
//
// Example: CMDU_TYPE_TOPOLOGY_NOTIFICATION --> "CMDU_TYPE_TOPOLOGY_NOTIFICATION"
//
// Return "Unknown" if the provided type does not exist.
//
char *convert_1905_CMDU_type_to_string(uint16_t cmdu_type);
uint8_t **forge_1905_TLVs_from_CMDU(struct cmdu_cstruct *cmdu, uint16_t **lens, uint8_t *total_tlvs);
uint16_t get_tlv_packet_len(uint8_t *tlv_stream);
uint8_t **prepare_cmdu_packets_from_cmdu_stream(struct cmdu_stream *memory_structure, uint16_t **lens);
uint8_t count_tlv_present_instream(uint8_t *tlv_stream, uint16_t stream_len);
void free_cmdu_cstream(struct cmdu_stream *cstream);
struct cmdu_cstruct *convert_stream_to_cstruct(struct cmdu_stream *cstream);
struct cmdu_stream *convert_cmdu_cstruct_to_stream(struct cmdu_cstruct *cmdu);
struct cmdu_stream *convert_packet_to_tlv_streams(uint8_t *packet, uint16_t plen);

/**
 * @ brief Function to free cmdu_cstruct
 *
 * @param cmdu cmdu c structure
 *
 */
void ieee1905_free_cmdu(struct cmdu_cstruct *cmdu);

/**
 * @ brief Function to count the tlvs present in binary stream, count stops
 * when it get end-of-message TLV, behaviour is uncertain if the tlv binary
 * stream is not having end-of-message TLV at the end of TLV stream.
 *
 * @param tlv_stream tlv binary stream
 *
 * @return count number of tlvs present in the tlv binary stream
 */
uint8_t count_tlv_present_instream_unsafe(uint8_t *tlv_stream);

/**
 * @brief Function to build a cmdu c-structure
 *
 * @param type cmdu message_type as defined in ieee1905 specification
 * @param mid cmdu message_id as defined in ieee1905 specs
 * @param relay cmdu relay indicator as defined in ieee1905 specs
 * @param tlvs Binary streams of tlvs associated with cmdu
 *
 * @return NULL if decoding fails
 * @return cmdu_cstruct if decoding succeeds
 */
struct cmdu_cstruct *ieee1905_build_cmdu(uint16_t type, uint16_t mid,
					 uint8_t relay, uint8_t *tlvs);
#endif
