/*
 * common_datatypes.h: IEEE-1905 handler's common header
 *
 * Author: Vivek Dutta <vivek.dutta@iopsys.eu>
 *
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef COMMON_DEFS_H
#define COMMON_DEFS_H
#include <stdint.h>
#include <stdbool.h>
#include <linux/types.h>

#include <arpa/inet.h>
#include <string.h>
#include <stdio.h>

#include <stdlib.h>
#include <errno.h>

#define NUM_ELEMS(x) (sizeof(x) / sizeof((x)[0]))

#define INT8U uint8_t
#define INT16U uint16_t
#define INT32U uint32_t
#define INT8S int8_t
#define INT16S int16_t
#define INT32S int32_t

// BIT manipulation Macros
#define BIT(m, x)      ((x >> m) & 0x0001)
// NOTE: n must be greater than m
#define BITS(n, m, x)  ((x >> m) & ((1 <<  (n-m+1)) - 1))
#define MODIFY_BIT(m, x, v) \
	do { \
		x = (x & (~(1 << m))); \
		x |= (((v) & 1) << m); \
	} while (0)

// NOTE: n must be greater than m
#define MODIFY_BITS(n, m, x, v) \
	do { \
		(x) = ((x) & (~(((1 << ((n)-(m)+1)) - 1) << (m)))); \
		(x) |= (((v)  & ((1 << ((n)-(m)+1)) - 1)) << (m)); \
	} while (0)

// Logging Macros
#define DEBUG_PRINT 0

#define PRINT_DBG(fmt, args...) \
	do { \
		if (DEBUG_PRINT) \
			printf(fmt"\n", ##args); \
	} while (0)

#define PRINT_ERR(fmt, args...) \
	fprintf(stderr, "[%s:%d]"fmt"\n", __func__, __LINE__, ##args)

// Allocation Macros
#define ERROR_ALLOC_MESSAGE "Insufficient Memory!"
#define MALLOC(p, n) \
	do { \
		(p) = malloc(sizeof(*(p)) * (n)); \
		if (!(p)) { \
			PRINT_ERR(ERROR_ALLOC_MESSAGE); \
			exit(ENOMEM); \
		} \
	} while (0)

#define CALLOC(p, n) \
	do { \
		(p) = calloc((n), sizeof(*(p))); \
		if (!(p)) { \
			PRINT_ERR(ERROR_ALLOC_MESSAGE); \
			exit(ENOMEM); \
		} \
	} while (0)

#define REALLOC(p, n) \
	do { \
		(p) = realloc((p), sizeof(*(p)) * (n)); \
		if (!(p)) { \
			PRINT_ERR(ERROR_ALLOC_MESSAGE); \
			exit(ENOMEM); \
		} \
	} while (0)

#define FREE(p) \
	do { \
		if (p) \
			free(p); \
		p = NULL; \
	} while (0)

#define ETHERTYPE_1905  (0x893a)
#define ETHERTYPE_LLDP  (0x88cc)

// 1905 multicast address ("01:80:C2:00:00:13")
#define MCAST_1905_B0 (0x01)
#define MCAST_1905_B1 (0x80)
#define MCAST_1905_B2 (0xC2)
#define MCAST_1905_B3 (0x00)
#define MCAST_1905_B4 (0x00)
#define MCAST_1905_B5 (0x13)
#define MCAST_1905  {MCAST_1905_B0, MCAST_1905_B1, MCAST_1905_B2, MCAST_1905_B3, MCAST_1905_B4, MCAST_1905_B5}

// LLDP nearest bridge multicast address ("01:80:C2:00:00:0E")
#define MCAST_LLDP_B0  (0x01)
#define MCAST_LLDP_B1  (0x80)
#define MCAST_LLDP_B2  (0xC2)
#define MCAST_LLDP_B3  (0x00)
#define MCAST_LLDP_B4  (0x00)
#define MCAST_LLDP_B5  (0x0E)
#define MCAST_LLDP  {MCAST_LLDP_B0, MCAST_LLDP_B1, MCAST_LLDP_B2, MCAST_LLDP_B3, MCAST_LLDP_B4, MCAST_LLDP_B5}

// Auxiliary functions to:
//
//   A) Extract 1, 2 or 4 bytes from a stream received from the network.
//
//   B) Insert  1, 2 or 4 bytes into a stream which is going to be sent into
//      the network.
//
// These functions do three things:
//
//   1. Avoid unaligned memory accesses (which might cause slowdowns or even
//      exceptions on some architectures)
//
//   2. Convert from network order to host order (and the other way)
//
//   3. Advance the packet pointer as many bytes as those which have just
//      been extracted/inserted.

// Extract/insert 1 byte
static inline void _E1B(uint8_t **packet_ppointer, uint8_t *memory_pointer)
{
	*memory_pointer = **packet_ppointer;
	(*packet_ppointer) += 1;
}

static inline void _I1B(const uint8_t *memory_pointer, uint8_t **packet_ppointer)
{
	**packet_ppointer = *memory_pointer;
	(*packet_ppointer) += 1;
}

// Extract/insert 2 bytes
static inline void _E2B(uint8_t **packet_ppointer, uint16_t *memory_pointer)
{
	uint16_t mem = 0;
	*(((uint8_t *) &mem) + 0) = **packet_ppointer;
	(*packet_ppointer)++;
	*(((uint8_t *) &mem) + 1) = **packet_ppointer;
	(*packet_ppointer)++;
	*memory_pointer = ntohs(mem);
}

static inline void _I2B(const uint16_t *memory_pointer, uint8_t **packet_ppointer)
{
	uint16_t mem = htons(*memory_pointer);
	**packet_ppointer = *(((uint8_t *) &mem) + 0);
	(*packet_ppointer)++;
	**packet_ppointer = *(((uint8_t *) &mem) + 1);
	(*packet_ppointer)++;
}

// Extract/insert 4 bytes
static inline void _E4B(uint8_t **packet_ppointer, uint32_t *memory_pointer)
{
	uint32_t mem = 0;
	*(((uint8_t *) &mem) + 0) = **packet_ppointer;
	(*packet_ppointer)++;
	*(((uint8_t *) &mem) + 1) = **packet_ppointer;
	(*packet_ppointer)++;
	*(((uint8_t *) &mem) + 2) = **packet_ppointer;
	(*packet_ppointer)++;
	*(((uint8_t *) &mem) + 3) = **packet_ppointer;
	(*packet_ppointer)++;
	*memory_pointer = ntohl(mem);
}

static inline void _I4B(const uint32_t *memory_pointer, uint8_t **packet_ppointer)
{
	uint32_t mem = htonl(*memory_pointer);
	**packet_ppointer = *(((uint8_t *) &mem) + 0);
	(*packet_ppointer)++;
	**packet_ppointer = *(((uint8_t *) &mem) + 1);
	(*packet_ppointer)++;
	**packet_ppointer = *(((uint8_t *) &mem) + 2);
	(*packet_ppointer)++;
	**packet_ppointer = *(((uint8_t *) &mem) + 3);
	(*packet_ppointer)++;
}

// Extract/insert N bytes (ignore endianess)
static inline void _EnB(uint8_t **packet_ppointer,
			void *memory_pointer,
			uint32_t n)
{
	memcpy(memory_pointer, *packet_ppointer, n);
	(*packet_ppointer) += n;
}

static inline void _InB(const void *memory_pointer,
			uint8_t **packet_ppointer,
			uint32_t n)
{
	memcpy(*packet_ppointer, memory_pointer, n);
	(*packet_ppointer) += n;
}

#endif
