/*
 * 1905_cmdu.c: 1905 cmdu handling
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#include <stdlib.h>
#include "1905_cmdus.h"
#include "1905_tlvs.h"

#include "common_defs.h"
////////////////////////////////////////////////////////////////////////////////
// Auxiliary, static tables
////////////////////////////////////////////////////////////////////////////////

//   WARNING:
//     If the CMDU message type changes (ie. the definition of  CMD_TYPE_*)
//     the following tables will have to be adapted (as the array index depends
//     on that).
//     Fortunately this should never happen.

// These tables marks, for each CMDU type of message, which TLVs are:
//
//   1. Require to be present zero or more times
//
//   2. required to be present exactly once
//
// The values in these tables were obtained from "IEEE Std 1905.1-2013, Section
// 6.3"
//
//
// TODO:
//     Right now this mechanism only considers either "zero or more" or "exactly
//     one" possibilities... however, in the "1a" update of the standard, there
//     are new types of TLVs that can appear "zero or one" and "one or more"
//     times.
//     For now I'm treating:
//       A) ...the "zero or one" type as "zero or more" (this
//          happens with the "push button generic phy event notification TLV",
//          the "control URL TLV" and the "IPv4/v6 TLVs") and...
//       B) ...the "one or more" type as "exactly one" (the "interface power
//          change information type TLV" and the "interface power change status
//          TLV").
//     Case (B) is not really a problem (in fact, I think "one or more" is an
//     error in the standard for these TLVs... as it should be "exactly one"...
//     maybe this will be corrected in a future update).
//     However, because of case (A), we could end up considering valid CMDUs
//     with, for example, more than one "IPv4 TLVs" (which is clearly an error).
//
//
//
static uint32_t _zeroormore_tlvs_for_cmdu[] = {
	/* CMDU_TYPE_TOPOLOGY_DISCOVERY             */ 0x00000000,
	/* CMDU_TYPE_TOPOLOGY_NOTIFICATION          */ 0x00000000,
	/* CMDU_TYPE_TOPOLOGY_QUERY                 */ 0x00000000,
	/* CMDU_TYPE_TOPOLOGY_RESPONSE              */
	1 << TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES |
		1 << TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST |
		1 << TLV_TYPE_NEIGHBOR_DEVICE_LIST |
		1 << TLV_TYPE_POWER_OFF_INTERFACE |
		1 << TLV_TYPE_L2_NEIGHBOR_DEVICE,
	/* CMDU_TYPE_VENDOR_SPECIFIC                */ 0xffffffff,
	/* CMDU_TYPE_LINK_METRIC_QUERY              */ 0x00000000,
	/* CMDU_TYPE_LINK_METRIC_RESPONSE           */ 1 << TLV_TYPE_TRANSMITTER_LINK_METRIC |
		1 << TLV_TYPE_RECEIVER_LINK_METRIC,
	/* CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH    */ 0x00000000,
	/* CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE  */ 0x00000000,
	/* CMDU_TYPE_AP_AUTOCONFIGURATION_WSC       */ 0x00000000,
	/* CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW     */ 0x00000000,
	/* CMDU_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION */ 1 << TLV_TYPE_GENERIC_PHY_EVENT_NOTIFICATION,
	/* CMDU_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION  */ 0x00000000,
	/* CMDU_TYPE_HIGHER_LAYER_QUERY             */ 0x00000000,
	/* CMDU_TYPE_HIGHER_LAYER_RESPONSE          */ 1 << TLV_TYPE_CONTROL_URL |
		1 << TLV_TYPE_IPV4 |
		1 << TLV_TYPE_IPV6,
	/* CMDU_TYPE_INTERFACE_POWER_CHANGE_REQUEST */ 0x00000000,
	/* CMDU_TYPE_INTERFACE_POWER_CHANGE_RESPONSE */ 0x00000000,
	/* CMDU_TYPE_GENERIC_PHY_QUERY              */ 0x00000000,
	/* CMDU_TYPE_GENERIC_PHY_RESPONSE           */ 0x00000000,
};

static uint32_t _exactlyone_tlvs_for_cmdu[] = {
	/* CMDU_TYPE_TOPOLOGY_DISCOVERY             */ 1 << TLV_TYPE_AL_MAC_ADDRESS_TYPE |
		1 << TLV_TYPE_MAC_ADDRESS_TYPE,
	/* CMDU_TYPE_TOPOLOGY_NOTIFICATION          */ 1 << TLV_TYPE_AL_MAC_ADDRESS_TYPE,
	/* CMDU_TYPE_TOPOLOGY_QUERY                 */ 0x00000000,
	/* CMDU_TYPE_TOPOLOGY_RESPONSE              */ 1 << TLV_TYPE_DEVICE_INFORMATION_TYPE,
	/* CMDU_TYPE_VENDOR_SPECIFIC                */ 0x00000000,
	/* CMDU_TYPE_LINK_METRIC_QUERY              */ 1 << TLV_TYPE_LINK_METRIC_QUERY,
	/* CMDU_TYPE_LINK_METRIC_RESPONSE           */ 0x00000000,
	/* CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH    */ 1 << TLV_TYPE_AL_MAC_ADDRESS_TYPE |
		1 << TLV_TYPE_SEARCHED_ROLE |
		1 << TLV_TYPE_AUTOCONFIG_FREQ_BAND,
	/* CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE  */ 1 << TLV_TYPE_SUPPORTED_ROLE |
		1 << TLV_TYPE_SUPPORTED_FREQ_BAND,
	/* CMDU_TYPE_AP_AUTOCONFIGURATION_WSC       */ 1 << TLV_TYPE_WSC,
	/* CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW     */ 1 << TLV_TYPE_AL_MAC_ADDRESS_TYPE |
		1 << TLV_TYPE_SUPPORTED_ROLE |
		1 << TLV_TYPE_SUPPORTED_FREQ_BAND,
	/* CMDU_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION */ 1 << TLV_TYPE_AL_MAC_ADDRESS_TYPE |
		1 << TLV_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION,
	/* CMDU_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION  */ 1 << TLV_TYPE_AL_MAC_ADDRESS_TYPE |
		1 << TLV_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION,
	/* CMDU_TYPE_HIGHER_LAYER_QUERY             */ 0x00000000,
	/* CMDU_TYPE_HIGHER_LAYER_RESPONSE          */
	    1 << TLV_TYPE_AL_MAC_ADDRESS_TYPE | 1 << TLV_TYPE_1905_PROFILE_VERSION |
		    1 << TLV_TYPE_DEVICE_IDENTIFICATION,
	/* CMDU_TYPE_INTERFACE_POWER_CHANGE_REQUEST */ 1 << TLV_TYPE_INTERFACE_POWER_CHANGE_INFORMATION,
	/* CMDU_TYPE_INTERFACE_POWER_CHANGE_RESPONSE */ 1 << TLV_TYPE_INTERFACE_POWER_CHANGE_STATUS,
	/* CMDU_TYPE_GENERIC_PHY_QUERY              */ 0x00000000,
	/* CMDU_TYPE_GENERIC_PHY_RESPONSE           */ 1 << TLV_TYPE_GENERIC_PHY_DEVICE_INFORMATION,
};

// The following table tells us the value of the 'relay_indicator' flag for
// each type of CMDU message.
//
// The values were obtained from "IEEE Std 1905.1-2013, Table 6-4"
//
// Note that '0xff' is a special value that means: "this CMDU message type can
// have the flag set to either '0' or '1' and its actual value for this
// particular message must be specified in some other way"
//
static uint8_t _relayed_CMDU[] = {
	/* CMDU_TYPE_TOPOLOGY_DISCOVERY             */ 0,
	/* CMDU_TYPE_TOPOLOGY_NOTIFICATION          */ 1,
	/* CMDU_TYPE_TOPOLOGY_QUERY                 */ 0,
	/* CMDU_TYPE_TOPOLOGY_QUERY                 */ 0,
	/* CMDU_TYPE_VENDOR_SPECIFIC                */ 0xff,
	/* CMDU_TYPE_LINK_METRIC_QUERY              */ 0,
	/* CMDU_TYPE_LINK_METRIC_RESPONSE           */ 0,
	/* CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH    */ 1,
	/* CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE  */ 0,
	/* CMDU_TYPE_AP_AUTOCONFIGURATION_WSC       */ 0,
	/* CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW     */ 1,
	/* CMDU_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION */ 1,
	/* CMDU_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION  */ 1,
	/* CMDU_TYPE_HIGHER_LAYER_QUERY             */ 0,
	/* CMDU_TYPE_HIGHER_LAYER_RESPONSE          */ 0,
	/* CMDU_TYPE_INTERFACE_POWER_CHANGE_REQUEST */ 0,
	/* CMDU_TYPE_INTERFACE_POWER_CHANGE_RESPONSE */ 0,
	/* CMDU_TYPE_GENERIC_PHY_QUERY              */ 0,
	/* CMDU_TYPE_GENERIC_PHY_RESPONSE           */ 0,
};

////////////////////////////////////////////////////////////////////////////////
// Auxiliary static functions
////////////////////////////////////////////////////////////////////////////////
uint16_t get_tlv_packet_len(uint8_t *tlv_stream)
{
	uint8_t *p;
	uint8_t type;
	uint16_t len;

	if (tlv_stream == NULL)
		return 0;

	p = tlv_stream;
	_E1B(&p, &type);
	_E2B(&p, &len);

	// tlv packet does not have length of tlv_type(1 Byte) and
	// tlv_len(2 Bytes) included
	return len + 3;
}

uint8_t count_tlv_present_instream(uint8_t *tlv_stream, uint16_t stream_len)
{
	uint8_t *p = tlv_stream;
	uint8_t count =0;
	uint16_t tlv_len = 0, parsed_len = 0;
	uint8_t tlv_type;

	if (tlv_stream == NULL ||
	    stream_len == 0) {
		PRINT_ERR("Stream is empty");
		return 0;
	}

	while(parsed_len < stream_len) {
		_E1B(&p, &tlv_type);
		_E2B(&p, &tlv_len);

		// Assuming we reached the end of cmdu's tlv stream
		if (tlv_type == TLV_TYPE_END_OF_MESSAGE)
			break;

		p = p + tlv_len;
		parsed_len += tlv_len + 3;
		count++;
	}

	return count;
}

void free_cmdu_cstream(struct cmdu_stream *cstream)
{
	uint8_t i;

	for (i = 0; i < cstream->nbr_tlv_streams; ++i) {
		FREE(cstream->tlv_stream_list[i]);
	}

	if (cstream->nbr_tlv_streams)
		FREE(cstream->tlv_stream_list);

	FREE(cstream);
}

struct cmdu_cstruct *convert_stream_to_cstruct(struct cmdu_stream *cstream)
{
	struct cmdu_cstruct *cmdu;
	uint8_t i, tlv_count;

	CALLOC(cmdu, 1);

	cmdu->message_version = cstream->message_version;
	cmdu->message_type = cstream->message_type;
	cmdu->message_id = cstream->message_id;
	cmdu->relay_indicator = cstream->relay_indicator;

	CALLOC(cmdu->tlv_cstruct_list, (cstream->nbr_tlv_streams + 1));

	tlv_count = 0;
	for (i = 0; i < cstream->nbr_tlv_streams; ++i) {
		uint8_t *tlv = parse_1905_TLV_from_packet(cstream->tlv_stream_list[i]);

		if (tlv != NULL)
			cmdu->tlv_cstruct_list[tlv_count++] = tlv;
		else {

			//PRINT_ERR("## Failed to decode the tlv|0x%02x|", *(cstream->tlv_stream_list[i]));
			for (size_t ii = 0; ii < i; ii++)
				free_1905_TLV_structure(cmdu->tlv_cstruct_list[ii], true);

			free(cmdu->tlv_cstruct_list);
			free(cmdu);
			return NULL;
			// might return null from here
		}

	}
	cmdu->tlv_cstruct_list[tlv_count] = NULL;

	return cmdu;
}

struct cmdu_stream *convert_cmdu_cstruct_to_stream(struct cmdu_cstruct *cmdu)
{
	struct cmdu_stream *cstream;
	uint8_t i;

	CALLOC(cstream, 1);

	cstream->message_version = cmdu->message_version;
	cstream->message_type = cmdu->message_type;
	cstream->message_id = cmdu->message_id;
	cstream->relay_indicator = cmdu->relay_indicator;

	i = 0;
	while (cmdu->tlv_cstruct_list[i] != NULL)
		i++;

	cstream->nbr_tlv_streams = i;

	if (cstream->nbr_tlv_streams)
		CALLOC(cstream->tlv_stream_list, i);

	for (i = 0; i < cstream->nbr_tlv_streams; ++i) {
		uint8_t *stream = forge_1905_TLV_from_structure(cmdu->tlv_cstruct_list[i]);

		if (stream != NULL)
			cstream->tlv_stream_list[i] = stream;
		else {
			PRINT_ERR("##### failed to decode the tlv|%x|", *stream);
			// might return null from here
		}

	}

	return cstream;
}

// Each CMDU must follow some rules regarding which TLVs they can contain
// depending on their type.
//
// This is extracted from "IEEE Std 1905.1-2013, Section 6.2":
//
//   1. When generating a CMDU:
//      a) It shall include all of the TLVs that are listed for the message
//      b) It shall not include any other TLV that is not listed for the message
//      c) It may additionally include zero or more vendor specific TLVs
//
//   2. When receiving a CMDU:
//      a) It may process or ignore any vendor specific TLVs
//      b) It shall ignore all TLVs that are not specified for the message
//      c) It shall ignore the entire message if the message does not include
//         all of the TLVs that are listed for this message
//
//  In both cases, this function returns:
//    'false' --> If 'p' did not respect the rules
//    'true' --> If 'p' is a valid cmdu structure
bool check_1905_cmdu_rules(struct cmdu_cstruct *p)
{
	uint8_t i;
	uint8_t counter[TLV_TYPE_LAST];
	uint8_t tlv_type;

	if (p == NULL) {
		// Invalid arguments
		PRINT_DBG("Invalid CMDU structure");
		return false;
	}

	if (p->tlv_cstruct_list == NULL) {
		PRINT_DBG("Invalid tlv list in CMDU structure");
		return false;
	}

	if (p->message_type > LAST_1905_CMDU) {
		PRINT_ERR("Not a valid 1905 CMDU type|0x%04x|", p->message_type);
		return false;
	}

	// First of all, count how many times each type of TLV message appears in
	// the structure. We will use this information later
	for (i = 0; i < TLV_TYPE_LAST; i++)
		counter[i] = 0;

	i = 0;
	while (p->tlv_cstruct_list[i] != NULL) {
		tlv_type = *(p->tlv_cstruct_list[i]);

		if (tlv_type > TLV_TYPE_LAST) {
			PRINT_ERR("tlv_type|0x%02x| present in cmdu|0x%04x|is not a ieee1905 tlv",
				  tlv_type, p->message_type);
			return false;
		}
		counter[tlv_type]++;
		i++;
	}

	// Rules 1.a and 2.c check the same thing : make sure the structure
	// contains, *at least*, the required TLVs
	// If not, return false
	//
	// The required TLVs are those contained in the "_exactlyone_tlvs_for_cmdu"
	// table.
	for (i = 0; i < TLV_TYPE_LAST; i++) {
		if ((counter[i] != 1) &&
		    (_exactlyone_tlvs_for_cmdu[p->message_type] & (1 << i))) {
			PRINT_DBG("Once TLV %s appeared %d times at index |%d|",
				  convert_1905_TLV_type_to_string(i),
				  counter[i], i);
			return false;
		}
	}

	// Rules 1.b and 2.b also both check for the same thing (unexpected TLVs)
	//
	// Unexpected TLVs are those that do not appear in neither the
	// "_exactlyone_tlvs_for_cmdu" nor the "_zeroormore_tlvs_for_cmdu" tables
	//
	for (i = 0; i < TLV_TYPE_LAST; i++) {
		if ((counter[i] != 0) &&
		    (i != TLV_TYPE_VENDOR_SPECIFIC) &&
		    !(_zeroormore_tlvs_for_cmdu[p->message_type] & (1 << i)) &&
		    !(_exactlyone_tlvs_for_cmdu[p->message_type] & (1 << i))
		    ) {
			PRINT_DBG("Extra TLV %s present %d times",
				  convert_1905_TLV_type_to_string(i), counter[i]);
			return false;
		}
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Actual API functions
////////////////////////////////////////////////////////////////////////////////

struct cmdu_cstruct *parse_1905_CMDU_from_packets(uint8_t **packet_streams)
{
	struct cmdu_cstruct *ret;

	uint8_t fragments_nr;
	uint8_t current_fragment;

	uint8_t tlvs_nr;

	uint8_t error;

	if (packet_streams == NULL) {
		// Invalid arguments
		//
		PRINT_DBG("NULL packet_streams");
		return NULL;
	}
	// Find out how many streams/fragments we have received
	//
	fragments_nr = 0;
	while (*(packet_streams + fragments_nr))
		fragments_nr++;

	if (fragments_nr == 0) {
		// No streams supplied!
		//
		PRINT_DBG("No fragments supplied");
		return NULL;
	}
	// Allocate the return structure.
	// Initially it will contain an empty list of TLVs that we will later
	// re-allocate and fill.
	//
	CALLOC(ret, 1);
	CALLOC(ret->tlv_cstruct_list, 1);
	ret->tlv_cstruct_list[0] = NULL;
	tlvs_nr = 0;

	// Next, parse each fragment
	//
	error = 0;
	for (current_fragment = 0; current_fragment < fragments_nr; current_fragment++) {
		uint8_t *p;
		uint8_t i;

		uint8_t message_version;
		uint8_t reserved_field;
		uint16_t message_type;
		uint16_t message_id;
		uint8_t fragment_id;
		uint8_t indicators;

		uint8_t relay_indicator;
		uint8_t last_fragment_indicator;

		uint8_t *parsed;

		// We want to traverse fragments in order, thus lets search for the
		// fragment whose 'fragment_id' matches 'current_fragment' (which will
		// monotonically increase starting at '0')
		//
		for (i = 0; i < fragments_nr; i++) {
			p = *(packet_streams + i);

			// The 'fragment_id' field is the 7th byte (offset 6)
			//
			if (current_fragment == *(p + 6))
				break;

		}
		if (i == fragments_nr) {
			// One of the fragments is missing!
			//
			error = 1;
			break;
		}
		// At this point 'p' points to the stream whose 'fragment_id' is
		// 'current_fragment'

		// Let's parse the header fields
		//
		_E1B(&p, &message_version);
		_E1B(&p, &reserved_field);
		_E2B(&p, &message_type);
		_E2B(&p, &message_id);
		_E1B(&p, &fragment_id);
		_E1B(&p, &indicators);

		last_fragment_indicator = (indicators & 0x80) >> 7;	// MSB and 2nd MSB
		relay_indicator = (indicators & 0x40) >> 6;	// of the
		// 'indicators'
		// field

		if (current_fragment == 0) {
			// This is the first fragment, thus fill the 'common' values.
			// We will later (in later fragments) check that their values always
			// remain the same
			//
			ret->message_version = message_version;
			ret->message_type = message_type;
			ret->message_id = message_id;
			ret->relay_indicator = relay_indicator;
		} else {
			// Check for consistency in all 'common' values
			//
			if ((ret->message_version != message_version) ||
			    (ret->message_type != message_type) ||
			    (ret->message_id != message_id) ||
			    (ret->relay_indicator != relay_indicator)
			    ) {
				// Fragments with different common fields were detected!
				//
				error = 2;
				break;
			}
		}

		// Regarding the 'relay_indicator', depending on the message type, it
		// can only have a valid specific value
		//
		if (_relayed_CMDU[message_type] == 0xff) {
			// Special, case. All values are allowed
		} else {
			// Check if the value for this type of message is valid
			//
			if (_relayed_CMDU[message_type] != relay_indicator) {
				error = 3;
				break;
			}
		}

		// Regarding the 'last_fragment_indicator' flag, the following condition
		// must be met: the last fragement (and only it!) must have it set to
		// '1'
		//
		if ((last_fragment_indicator == 1) && (current_fragment < fragments_nr - 1)) {
			// 'last_fragment_indicator' appeared *before* the last fragment
			//
			error = 4;
			break;
		}
		if ((last_fragment_indicator == 0) && (current_fragment == fragments_nr - 1)) {
			// 'last_fragment_indicator' did not appear in the last fragment
			//
			error = 5;
			break;
		}
		// We can now parse the TLVs. 'p' is pointing to the first one at this
		// moment
		//
		while (p != NULL) {
			parsed = parse_1905_TLV_from_packet(p);
			if (parsed == NULL) {
				PRINT_ERR("parsing of tlv failed");

				error = 6;
				break;
			}

			if (*parsed == TLV_TYPE_END_OF_MESSAGE) {
				// No more TLVs
				//
				free_1905_TLV_structure(parsed, true);
				break;
			}
			// Advance 'p' to the next TLV.
			//
			uint8_t tlv_type;
			uint16_t tlv_len;

			_E1B(&p, &tlv_type);
			_E2B(&p, &tlv_len);

			p += tlv_len;

			// Add this new TLV to the list (the list needs to be re-allocated
			// with more space first)
			//
			tlvs_nr++;
			REALLOC(ret->tlv_cstruct_list, (tlvs_nr + 1));
			ret->tlv_cstruct_list[tlvs_nr - 1] = parsed;
			ret->tlv_cstruct_list[tlvs_nr] = NULL;
		}
		if (error != 0)
			break;
	}

	if (error == 0) {
		// Ok then... we now have our output structure properly filled.
		// However, there is one last battery of checks we must perform:
		//
		//   - CMDU_TYPE_VENDOR_SPECIFIC: The first TLV *must* be of type
		//     TLV_TYPE_VENDOR_SPECIFIC
		//
		//   - All the other message types: Some TLVs (different for each of
		//     them) can only appear once, others can appear zero or more times
		//     and others must be ignored.
		//     The 'check_1905_cmdu_rules()' takes care of this for us.
		//
		PRINT_DBG("CMDU type: %s", convert_1905_CMDU_type_to_string(ret->message_type));

		if (ret->message_type == CMDU_TYPE_VENDOR_SPECIFIC) {
			if (ret->tlv_cstruct_list == NULL ||
			    ret->tlv_cstruct_list[0] == NULL ||
			    TLV_TYPE_VENDOR_SPECIFIC != *(ret->tlv_cstruct_list[0])) {
				error = 7;
			}
		} else {
			if (check_1905_cmdu_rules(ret) == false) {
				PRINT_ERR("Structure is missing some required TLVs");
				PRINT_DBG("List of present TLVs:");

				if (ret->tlv_cstruct_list != NULL) {
					uint8_t i;

					i = 0;
					while (ret->tlv_cstruct_list[i]) {
						PRINT_DBG("  - %s", convert_1905_TLV_type_to_string(*(ret->tlv_cstruct_list[i])));
						i++;
					}
					PRINT_DBG("  - <END>");
				} else {
					PRINT_DBG("  - <NONE>");
				}

				free_1905_CMDU_structure(ret);
				return NULL;
			}
		}
	}
	// Finally! If we get this far without errors we are already done, otherwise
	// free everything and return NULL
	//
	if (error != 0) {
		PRINT_DBG("Parsing error %d", error);
		free_1905_CMDU_structure(ret);
		return NULL;
	}

	return ret;
}

uint8_t **forge_1905_TLVs_from_CMDU(struct cmdu_cstruct *cmdu, uint16_t **lens, uint8_t *total_tlvs)
{
	uint8_t **ret;
	uint16_t *tlv_len_list;
	uint8_t tlv_nr = 0, i;

	if (cmdu == NULL || lens == NULL)
		return NULL;

	if (cmdu->tlv_cstruct_list == NULL)
		return NULL;

	while (cmdu->tlv_cstruct_list[tlv_nr] != NULL)
		tlv_nr++;

	if (tlv_nr == 0) {
		*total_tlvs = 0;
		PRINT_ERR("## No TLV in cmdu(%s) ##",
			convert_1905_CMDU_type_to_string(cmdu->message_type));
		return NULL;
	}

	CALLOC(ret, tlv_nr);
	CALLOC(tlv_len_list, tlv_nr);

	uint8_t forged_tlv_count = 0;

	for (i = 0; i < tlv_nr; i++) {
		uint16_t tlv_stream_size;
		uint8_t *tlv = forge_1905_TLV_from_structure(cmdu->tlv_cstruct_list[i]);
		tlv_stream_size = get_tlv_packet_len(tlv);

		if (tlv) {
			ret[forged_tlv_count] = tlv;
			tlv_len_list[forged_tlv_count] = tlv_stream_size;
			forged_tlv_count++;
		} else {
			tlv_len_list[i] = 0;
		}
	}
	*total_tlvs = forged_tlv_count;
	*lens = tlv_len_list;
	return ret;
}

uint8_t **prepare_cmdu_packets_from_cmdu_stream(struct cmdu_stream *memory_structure, uint16_t **lens)
{
	uint8_t **ret;
	uint16_t *tlv_len_list;

	uint8_t i;
	uint8_t tlv_start;
	uint8_t tlv_stop;
	uint8_t fragments_nr;
	uint32_t max_tlvs_block_size;
	uint8_t error = 0;

	if (NULL == memory_structure || NULL == lens) {
		return NULL;
	}

	// Allocate the return streams.
	// Initially we will just have an empty list (ie. it contains a single
	// element marking the end-of-list: a NULL pointer)
	CALLOC(ret, 1);
	CALLOC(tlv_len_list, 1);

	fragments_nr = 0;

	// Let's create as many streams as needed so that all of them fit in
	// MAX_NETWORK_SEGMENT_SIZE bytes.
	//
	// More specifically, each of the fragments that we are going to generate
	// will have a size equal to the sum of:
	//
	//   - 6 bytes (destination MAC address)
	//   - 6 bytes (origin MAC address)
	//   - 2 bytes (ETH type)
	//   - 1 byte  (CMDU message version)
	//   - 1 byte  (CMDU reserved field)
	//   - 2 bytes (CMDU message type)
	//   - 2 bytes (CMDU message id)
	//   - 1 byte  (CMDU fragment id)
	//   - 1 byte  (CMDU flags/indicators)
	//   - X bytes (size of all TLVs contained in the fragment)
	//   - 3 bytes (TLV_TYPE_END_OF_MESSAGE TLV)
	//
	// In other words, X (the size of all the TLVs that are going to be inside
	// this fragment) can not be greater than MAX_NETWORK_SEGMENT_SIZE - 6 - 6 -
	// 2 - 1 - 1 - 2 - 2 - 1 - 1 - 3 = MAX_NETWORK_SEGMENT_SIZE - 25 bytes.

	// 3 for end of tlv message
	max_tlvs_block_size = MAX_NETWORK_SEGMENT_SIZE - 25 - 3;
	tlv_start = 0;
	tlv_stop = 0;
	do {
		uint8_t *s;

		uint16_t current_X_size;

		uint8_t reserved_field;
		uint8_t fragment_id;
		uint8_t indicators;

		uint8_t no_space;

		current_X_size = 0;
		no_space = 0;

		for (i = tlv_start; i < memory_structure->nbr_tlv_streams; i++) {
			uint16_t tlv_stream_size;

			tlv_stream_size = get_tlv_packet_len(memory_structure->tlv_stream_list[i]);

			if (current_X_size + tlv_stream_size < max_tlvs_block_size) {
				tlv_stop++;
			} else {
				// There is no space for more TLVs
				no_space = 1;
				break;
			}

			current_X_size += tlv_stream_size;
		}

		if (tlv_start == tlv_stop) {
			if (no_space == 1) {
				// One *single* TLV does not fit in a fragment!
				// This is an error... there is no way to split one single TLV into
				// several fragments according to the standard.
				//
				error = 1;
				break;
			}
		}
		// Now that we know how many TLVs are going to be embedded inside this
		// fragment (from 'tlv_start' up to -and not including- 'tlv_stop'),
		// let's build it
		fragments_nr++;

		REALLOC(ret, (fragments_nr + 1));
		CALLOC(ret[fragments_nr - 1], MAX_NETWORK_SEGMENT_SIZE);
		ret[fragments_nr] = NULL;

		REALLOC(tlv_len_list, (fragments_nr + 1));
		tlv_len_list[fragments_nr - 1] = 0;	// To be updated a few lines later
		tlv_len_list[fragments_nr] = 0;

		s = ret[fragments_nr - 1];

		reserved_field = 0;
		fragment_id = fragments_nr - 1;
		indicators = 0;

		// Set 'last_fragment_indicator' flag (bit #7)
		if (tlv_stop == memory_structure->nbr_tlv_streams)
			indicators |= 1 << 7;

		// Set 'relay_indicator' flag (bit #6)
		indicators |= memory_structure->relay_indicator << 6;
		// TODO: add a check to validate the relay bit

		_I1B(&memory_structure->message_version, &s);
		_I1B(&reserved_field, &s);
		_I2B(&memory_structure->message_type, &s);
		_I2B(&memory_structure->message_id, &s);
		_I1B(&fragment_id, &s);
		_I1B(&indicators, &s);

		for (i = tlv_start; i < tlv_stop; i++) {
			uint8_t *tlv_stream;
			uint16_t tlv_stream_size;

			tlv_stream = memory_structure->tlv_stream_list[i];
			tlv_stream_size = get_tlv_packet_len(tlv_stream);

			memcpy(s, tlv_stream, tlv_stream_size);

			s += tlv_stream_size;
		}

		// Don't forget to add the last three octects representing the
		// TLV_TYPE_END_OF_MESSAGE message
		*s = 0x0;
		s++;
		*s = 0x0;
		s++;
		*s = 0x0;
		s++;

		// Update the length return value
		tlv_len_list[fragments_nr - 1] = abs(s - ret[fragments_nr - 1]);

		// And advance the TLV pointer so that, if more fragments are needed,
		// the next one starts where we have stopped.
		tlv_start = tlv_stop;

	} while (tlv_start < memory_structure->nbr_tlv_streams);

	// Finally! If we get this far without errors we are already done, otherwise
	// free everything and return NULL
	if (error != 0) {
		free_1905_CMDU_packets(ret);
		FREE(tlv_len_list);
		return NULL;
	}
	*lens = tlv_len_list;

	return ret;
}

uint8_t **forge_1905_CMDU_from_structure(struct cmdu_cstruct *memory_structure, uint16_t **lens)
{
	uint8_t **ret;
	uint16_t *tlv_lens;

	uint8_t tlv_start;
	uint8_t tlv_stop;

	uint8_t fragments_nr;

	uint32_t max_tlvs_block_size;

	uint8_t error;

	error = 0;

	if (NULL == memory_structure || NULL == lens) {
		PRINT_ERR("Invalid input arguments");
		return NULL;
	}

	if (memory_structure->tlv_cstruct_list == NULL) {
		PRINT_ERR("Invalid tlv list in input arguments");
		return NULL;
	}

	// Before anything else, let's check that the CMDU 'rules' are satisfied:
	if (check_1905_cmdu_rules(memory_structure) == false) {
		PRINT_ERR("cmdu validation failed for cmdu_type|0x%04x|",
			  memory_structure->message_type);
		return NULL;
	}
	// Allocate the return streams.
	// Initially we will just have an empty list (ie. it contains a single
	// element marking the end-of-list: a NULL pointer)
	//
	CALLOC(ret, 1);
	CALLOC(tlv_lens, 1);

	fragments_nr = 0;

	// Let's create as many streams as needed so that all of them fit in
	// MAX_NETWORK_SEGMENT_SIZE bytes.
	//
	// More specifically, each of the fragments that we are going to generate
	// will have a size equal to the sum of:
	//
	//   - 6 bytes (destination MAC address)
	//   - 6 bytes (origin MAC address)
	//   - 2 bytes (ETH type)
	//   - 1 byte  (CMDU message version)
	//   - 1 byte  (CMDU reserved field)
	//   - 2 bytes (CMDU message type)
	//   - 2 bytes (CMDU message id)
	//   - 1 byte  (CMDU fragment id)
	//   - 1 byte  (CMDU flags/indicators)
	//   - X bytes (size of all TLVs contained in the fragment)
	//   - 3 bytes (TLV_TYPE_END_OF_MESSAGE TLV)
	//
	// In other words, X (the size of all the TLVs that are going to be inside
	// this fragmen) can not be greater than MAX_NETWORK_SEGMENT_SIZE - 6 - 6 -
	// 2 - 1 - 1 - 2 - 2 - 1 - 1 - 3 = MAX_NETWORK_SEGMENT_SIZE - 25 bytes.
	//
	max_tlvs_block_size = MAX_NETWORK_SEGMENT_SIZE - 25;
	tlv_start = 0;
	tlv_stop = 0;
	do {
		uint8_t *s;
		uint8_t i;

		uint16_t current_X_size;

		uint8_t reserved_field;
		uint8_t fragment_id;
		uint8_t indicators;

		uint8_t no_space;

		current_X_size = 0;
		no_space = 0;
		while (memory_structure->tlv_cstruct_list[tlv_stop]) {
			uint8_t *p;
			uint8_t *tlv_stream;
			uint16_t tlv_stream_size;

			p = memory_structure->tlv_cstruct_list[tlv_stop];

			tlv_stream = forge_1905_TLV_from_structure(p);
			tlv_stream_size = get_tlv_packet_len(tlv_stream);
			FREE(tlv_stream);

			if (current_X_size + tlv_stream_size < max_tlvs_block_size) {
				tlv_stop++;
			} else {
				// There is no space for more TLVs
				//
				no_space = 1;
				break;
			}

			current_X_size += tlv_stream_size;
		}

		if ((tlv_start == tlv_stop) && (no_space == 1)) {
				error = 1;
				break;
		}
		// Now that we know how many TLVs are going to be embedded inside this
		// fragment (from 'tlv_start' up to -and not including- 'tlv_stop'),
		// let's build it
		//
		fragments_nr++;

		REALLOC(ret, (fragments_nr + 1));
		CALLOC(ret[fragments_nr - 1], MAX_NETWORK_SEGMENT_SIZE);
		ret[fragments_nr] = NULL;

		REALLOC(tlv_lens, (fragments_nr + 1));
		tlv_lens[fragments_nr - 1] = 0;	// To be updated a few lines later
		tlv_lens[fragments_nr] = 0;

		s = ret[fragments_nr - 1];

		reserved_field = 0;
		fragment_id = fragments_nr - 1;
		indicators = 0;

		// Set 'last_fragment_indicator' flag (bit #7)
		//
		if (memory_structure->tlv_cstruct_list[tlv_stop] == NULL)
			indicators |= 1 << 7;

		// Set 'relay_indicator' flag (bit #6)
		//
		if (_relayed_CMDU[memory_structure->message_type] == 0xff) {
			// Special, case. Respect what the caller told us
			//
			indicators |= memory_structure->relay_indicator << 6;
		} else {
			// Use the fixed value for this type of message according to the
			// standard
			//
			indicators |= _relayed_CMDU[memory_structure->message_type] << 6;
		}

		_I1B(&memory_structure->message_version, &s);
		_I1B(&reserved_field, &s);
		_I2B(&memory_structure->message_type, &s);
		_I2B(&memory_structure->message_id, &s);
		_I1B(&fragment_id, &s);
		_I1B(&indicators, &s);

		for (i = tlv_start; i < tlv_stop; i++) {
			uint8_t *tlv_stream;
			uint16_t tlv_stream_size;

			tlv_stream = forge_1905_TLV_from_structure(memory_structure->tlv_cstruct_list[i]);
			tlv_stream_size = get_tlv_packet_len(tlv_stream);

			memcpy(s, tlv_stream, tlv_stream_size);
			FREE(tlv_stream);

			s += tlv_stream_size;
		}

		// Don't forget to add the last three octects representing the
		// TLV_TYPE_END_OF_MESSAGE message
		//
		*s = 0x0;
		s++;
		*s = 0x0;
		s++;
		*s = 0x0;
		s++;

		// Update the length return value
		tlv_lens[fragments_nr - 1] = abs(s - ret[fragments_nr - 1]);

		// And advance the TLV pointer so that, if more fragments are needed,
		// the next one starts where we have stopped.
		//
		tlv_start = tlv_stop;

	} while (memory_structure->tlv_cstruct_list[tlv_start]);

	// Finally! If we get this far without errors we are already done, otherwise
	// free everything and return NULL
	//
	if (error != 0) {
		free_1905_CMDU_packets(ret);
		FREE(tlv_lens);
		return NULL;
	}
	*lens = tlv_lens;

	return ret;
}

bool get_cmdu_header_from_packet(uint8_t *stream, uint16_t *mid, uint8_t *fragment_id, uint8_t *last_fragment_indicator)
{
	uint8_t message_version;
	uint8_t reserved_field;
	uint16_t message_type;
	uint8_t indicators;

	if ((stream == NULL) ||
	    (mid == NULL) ||
	    (fragment_id == NULL) ||
	    (last_fragment_indicator == NULL))
		return false;

	// Let's parse the header fields
	_E1B(&stream, &message_version);
	_E1B(&stream, &reserved_field);
	_E2B(&stream, &message_type);
	_E2B(&stream, mid);
	_E1B(&stream, fragment_id);
	_E1B(&stream, &indicators);

	if (message_version != CMDU_MESSAGE_VERSION_1905_1_2013) {
		PRINT_ERR("Unsupported cmdu Version!");
		return false;
	}

	*last_fragment_indicator = (indicators & 0x80) >> 7;	// MSB and 2nd MSB

	return true;
}

struct cmdu_stream *convert_packet_to_tlv_streams(uint8_t *packet, uint16_t plen)
{
	struct cmdu_stream *cstream;
	uint8_t indicators = 0;
	uint8_t count = 0;
	uint16_t parsed_len = 0;
	uint8_t *p;

	CALLOC(cstream, 1);

	// We want to traverse fragments in order, thus lets search for the
	// fragment whose 'fragment_id' matches 'current_fragment' (which will
	// monotonically increase starting at '0')

	p = packet;
	// Let's parse the header fields
	_E1B(&p, &cstream->message_version);
	p++; // skip the reserved field
	_E2B(&p, &cstream->message_type);
	_E2B(&p, &cstream->message_id);
	p++; // skip fragment id
	_E1B(&p, &indicators);

	plen = plen - (1 + 1 + 2 + 2 + 1 + 1);

	cstream->relay_indicator = (indicators & 0x40) >> 6;
	cstream->nbr_tlv_streams = count_tlv_present_instream(p, plen);

	if (cstream->nbr_tlv_streams) {
		CALLOC(cstream->tlv_stream_list, cstream->nbr_tlv_streams);
		do {
			uint8_t *q = p;
			uint8_t tlv_type;
			uint16_t tlv_len;

			_E1B(&q, &tlv_type);
			_E2B(&q, &tlv_len);

			if (tlv_type == TLV_TYPE_END_OF_MESSAGE)
				break;

			if (parsed_len + tlv_len + 3 > plen)
				break;

			CALLOC(cstream->tlv_stream_list[count], (tlv_len + 3));
			memcpy(cstream->tlv_stream_list[count], p, (tlv_len + 3));

			p = p + tlv_len + 3;
			parsed_len += tlv_len + 3;
			count++;
		} while (p != NULL);
	}

	return cstream;
}

void free_1905_CMDU_structure(struct cmdu_cstruct *memory_structure)
{

	if ((memory_structure != NULL) &&
	    (memory_structure->tlv_cstruct_list != NULL)) {
		uint8_t i;

		i = 0;
		while (memory_structure->tlv_cstruct_list[i]) {
			free_1905_TLV_structure(memory_structure->tlv_cstruct_list[i], true);
			i++;
		}
		FREE(memory_structure->tlv_cstruct_list);
	}

	FREE(memory_structure);
}

void free_tlv_list_CMDU_structure(struct cmdu_cstruct *memory_structure)
{

	if ((memory_structure != NULL) &&
	    (memory_structure->tlv_cstruct_list != NULL)) {
		uint8_t i;

		i = 0;
		while (memory_structure->tlv_cstruct_list[i]) {
			free_1905_TLV_structure(memory_structure->tlv_cstruct_list[i], true);
			i++;
		}
		FREE(memory_structure->tlv_cstruct_list);
	}
}

void free_tlv_internal_list_CMDU_structure(struct cmdu_cstruct *memory_structure)
{
	if ((memory_structure != NULL) &&
	    (memory_structure->tlv_cstruct_list != NULL)) {
		uint8_t i;

		i = 0;
		while (memory_structure->tlv_cstruct_list[i]) {
			free_1905_TLV_structure(memory_structure->tlv_cstruct_list[i], false);
			i++;
		}
		FREE(memory_structure->tlv_cstruct_list);
	}
}

void free_1905_CMDU_packets(uint8_t **packet_streams)
{
	uint8_t i;

	if (packet_streams == NULL)
		return;

	i = 0;
	while (packet_streams[i]) {
		FREE(packet_streams[i]);
		i++;
	}
	FREE(packet_streams);
}

uint8_t compare_1905_CMDU_structures(struct cmdu_cstruct *memory_structure_1, struct cmdu_cstruct *memory_structure_2)
{
	uint8_t i;

	if (memory_structure_1 == NULL || memory_structure_2 == NULL)
		return 1;

	if (memory_structure_1->tlv_cstruct_list == NULL ||
	    memory_structure_2->tlv_cstruct_list == NULL)
		return 1;

	if ((memory_structure_1->message_version != memory_structure_2->message_version) ||
	    (memory_structure_1->message_type != memory_structure_2->message_type) ||
	    (memory_structure_1->message_id != memory_structure_2->message_id) ||
	    (memory_structure_1->relay_indicator != memory_structure_2->relay_indicator))
		return 1;

	i = 0;
	while (memory_structure_1->tlv_cstruct_list[i] != NULL) {
		if (compare_1905_TLV_structures(memory_structure_1->tlv_cstruct_list[i],
						memory_structure_2->tlv_cstruct_list[i]) != 0)
			return 1;

		i++;
	}
	// No more TLVs to compare! Return '0' (structures are equal)
	if (memory_structure_1->tlv_cstruct_list[i] == NULL &&
	    memory_structure_2->tlv_cstruct_list[i] == NULL)
		return 0;

	// This point should never be reached
	return 1;
}

char *convert_1905_CMDU_type_to_string(uint16_t cmdu_type)
{
	switch (cmdu_type) {
	case CMDU_TYPE_TOPOLOGY_DISCOVERY:
		return "CMDU_TYPE_TOPOLOGY_DISCOVERY";
	case CMDU_TYPE_TOPOLOGY_NOTIFICATION:
		return "CMDU_TYPE_TOPOLOGY_NOTIFICATION";
	case CMDU_TYPE_TOPOLOGY_QUERY:
		return "CMDU_TYPE_TOPOLOGY_QUERY";
	case CMDU_TYPE_TOPOLOGY_RESPONSE:
		return "CMDU_TYPE_TOPOLOGY_RESPONSE";
	case CMDU_TYPE_VENDOR_SPECIFIC:
		return "CMDU_TYPE_VENDOR_SPECIFIC";
	case CMDU_TYPE_LINK_METRIC_QUERY:
		return "CMDU_TYPE_LINK_METRIC_QUERY";
	case CMDU_TYPE_LINK_METRIC_RESPONSE:
		return "CMDU_TYPE_LINK_METRIC_RESPONSE";
	case CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH:
		return "CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH";
	case CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE:
		return "CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE";
	case CMDU_TYPE_AP_AUTOCONFIGURATION_WSC:
		return "CMDU_TYPE_AP_AUTOCONFIGURATION_WSC";
	case CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW:
		return "CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW";
	case CMDU_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION:
		return "CMDU_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION";
	case CMDU_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION:
		return "CMDU_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION";
	case CMDU_TYPE_GENERIC_PHY_QUERY:
		return "CMDU_TYPE_GENERIC_PHY_QUERY";
	case CMDU_TYPE_GENERIC_PHY_RESPONSE:
		return "CMDU_TYPE_GENERIC_PHY_RESPONSE";
	case CMDU_TYPE_HIGHER_LAYER_QUERY:
		return "CMDU_TYPE_HIGHER_LAYER_QUERY";
	case CMDU_TYPE_HIGHER_LAYER_RESPONSE:
		return "CMDU_TYPE_HIGHER_LAYER_RESPONSE";
	case CMDU_TYPE_INTERFACE_POWER_CHANGE_REQUEST:
		return "CMDU_TYPE_INTERFACE_POWER_CHANGE_REQUEST";
	case CMDU_TYPE_INTERFACE_POWER_CHANGE_RESPONSE:
		return "CMDU_TYPE_INTERFACE_POWER_CHANGE_RESPONSE";
	default:
		return "UNKNOWN";
	}
	return "";
}

// count includes end-of-message tlv
uint8_t count_tlv_present_instream_unsafe(uint8_t *tlv_stream)
{
	uint8_t *p = tlv_stream;
	uint8_t tlv_type, count = 0;
	uint16_t tlv_len = 0;

	if (tlv_stream == NULL) {
		PRINT_ERR("Stream is empty");
		return 0;
	}

	do {
		_E1B(&p, &tlv_type);
		_E2B(&p, &tlv_len);

		p = p + tlv_len;
		count++;
	} while(tlv_type != TLV_TYPE_END_OF_MESSAGE);

	return count;
}

void ieee1905_free_cmdu(struct cmdu_cstruct *cmdu)
{
	if (cmdu)
		free_1905_CMDU_structure(cmdu);
}

struct cmdu_cstruct *ieee1905_build_cmdu(uint16_t type, uint16_t mid,
					 uint8_t relay, uint8_t *tlvs)
{
	uint8_t i, tlv_count, j;
	struct cmdu_cstruct *cmdu;
	uint8_t *cur_stream, *tlv = NULL;

	cmdu = calloc(1, sizeof(struct cmdu_cstruct));
	if (!cmdu) {
		PRINT_ERR("Insufficient memory!");
		return NULL;
	}

	cmdu->message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	cmdu->message_type = type;
	cmdu->message_id = mid;
	cmdu->relay_indicator = relay;

	tlv_count = count_tlv_present_instream_unsafe(tlvs);
	CALLOC(cmdu->tlv_cstruct_list, (tlv_count + 1));
	if (cmdu->tlv_cstruct_list == NULL) {
		PRINT_ERR("Out of memory!!");
		FREE(cmdu);
		return NULL;
	}

	j = 0;
	cur_stream = tlvs;
	for (i = 0; i < tlv_count - 1; i++) {
		uint16_t len = get_tlv_packet_len(cur_stream);

		tlv = parse_1905_TLV_from_packet(cur_stream);
		if (tlv)
			cmdu->tlv_cstruct_list[j++] =  tlv;

		// Move to the next tlv
		cur_stream = cur_stream + len;
	}

	if (j != tlv_count - 1) {
		PRINT_ERR("Failed to decode one or more tlvs(%d:%d)", j, tlv_count);
		cmdu->tlv_cstruct_list[j] = NULL;
	}
	cmdu->tlv_cstruct_list[tlv_count] = NULL;

	return cmdu;
}
