/*
 * 1905_tlvs.c: 1905 tlv handling
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "1905_tlvs.h"
#include "common_defs.h"

////////////////////////////////////////////////////////////////////////////////
// Actual API functions
////////////////////////////////////////////////////////////////////////////////

uint8_t *parse_1905_TLV_from_packet(uint8_t *packet_stream)
{
	if (packet_stream == NULL) {
		PRINT_ERR("{%s}Null packet stream", __func__);
		return NULL;
	}
	// The first byte of the stream is the "Type" field from the TLV structure.
	// Valid values for this byte are the following ones...
	//
	switch (*packet_stream) {
	case TLV_TYPE_END_OF_MESSAGE:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.1"

			struct endOfMessageTLV *ret;

			uint8_t *p;
			uint16_t len;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			// According to the standard, the length *must* be 0
			//
			if (len != 0) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}

			ret->tlv_type = TLV_TYPE_END_OF_MESSAGE;

			return (uint8_t *) ret;
		}

	case TLV_TYPE_VENDOR_SPECIFIC:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.2"

			struct vendorSpecificTLV *ret;

			uint8_t *p;
			uint16_t len;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			// According to the standard, the length *must* be at least "3"
			//
			if (len < 3) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}

			ret->tlv_type = TLV_TYPE_VENDOR_SPECIFIC;

			_E1B(&p, &ret->vendorOUI[0]);
			_E1B(&p, &ret->vendorOUI[1]);
			_E1B(&p, &ret->vendorOUI[2]);

			ret->m_nr = len - 3;

			if (ret->m_nr) {
				CALLOC(ret->m, ret->m_nr);
				_EnB(&p, ret->m, ret->m_nr);
			} else {
				ret->m = NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.3"

			struct alMacAddressTypeTLV *ret;

			uint8_t *p;
			uint16_t len;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			// According to the standard, the length *must* be 6
			//
			if (len != 6) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}

			ret->tlv_type = TLV_TYPE_AL_MAC_ADDRESS_TYPE;

			_EnB(&p, ret->al_mac_address, 6);

			return (uint8_t *) ret;
		}

	case TLV_TYPE_MAC_ADDRESS_TYPE:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.4"

			struct macAddressTypeTLV *ret;

			uint8_t *p;
			uint16_t len;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			// According to the standard, the length *must* be 6
			//
			if (len != 6) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}

			ret->tlv_type = TLV_TYPE_MAC_ADDRESS_TYPE;

			_EnB(&p, ret->mac_address, 6);

			return (uint8_t *) ret;
		}

	case TLV_TYPE_DEVICE_INFORMATION_TYPE:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.5"

			struct deviceInformationTypeTLV *ret;

			uint8_t *p;
			uint16_t len;
			uint8_t i;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			ret->tlv_type = TLV_TYPE_DEVICE_INFORMATION_TYPE;

			_EnB(&p, ret->al_mac_address, 6);
			_E1B(&p, &ret->local_interfaces_nr);

			CALLOC(ret->local_interfaces, ret->local_interfaces_nr);

			for (i = 0; i < ret->local_interfaces_nr; i++) {
				_EnB(&p, ret->local_interfaces[i].mac_address, 6);
				_E2B(&p, &ret->local_interfaces[i].media_type);
				_E1B(&p, &ret->local_interfaces[i].media_specific_data_size);

				if ((ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11B_2_4_GHZ) ||
				    (ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11G_2_4_GHZ) ||
				    (ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11A_5_GHZ) ||
				    (ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11N_2_4_GHZ) ||
				    (ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11N_5_GHZ) ||
				    (ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AC_5_GHZ) ||
				    (ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AD_60_GHZ) ||
				    (ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AF_GHZ)
				    ) {
					uint8_t aux;

					if (ret->local_interfaces[i].media_specific_data_size != 10) {
						// Malformed packet
						//
						FREE(ret->local_interfaces);
						FREE(ret);
						return NULL;
					}

					_EnB(&p, ret->local_interfaces[i].media_specific_data.ieee80211.network_membership, 6);
					_E1B(&p, &aux);
					ret->local_interfaces[i].media_specific_data.ieee80211.role = aux >> 4;
					_E1B(&p, &ret->local_interfaces[i].media_specific_data.ieee80211.ap_channel_band);
					_E1B(&p, &ret->local_interfaces[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_1);
					_E1B(&p, &ret->local_interfaces[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_2);

				} else if ((ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_1901_WAVELET) ||
					   (ret->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_1901_FFT)
				    ) {
					if (ret->local_interfaces[i].media_specific_data_size != 7) {
						// Malformed packet
						//
						FREE(ret->local_interfaces);
						FREE(ret);
						return NULL;
					}
					_EnB(&p, ret->local_interfaces[i].media_specific_data.ieee1901.network_identifier, 7);
				} else {
					if (ret->local_interfaces[i].media_specific_data_size != 0) {
						// Malformed packet
						//
						FREE(ret->local_interfaces);
						FREE(ret);
						return NULL;
					}
				}
			}

			if (p - (packet_stream + 3) != len) {
				// Malformed packet
				//
				FREE(ret->local_interfaces);
				FREE(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.6"

			struct deviceBridgingCapabilityTLV *ret;

			uint8_t *p;
			uint16_t len;
			uint8_t i, j;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			ret->tlv_type = TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES;

			if (len == 0) {
				FREE(ret);
				return NULL;
			}

			_E1B(&p, &ret->bridging_tuples_nr);

			if (ret->bridging_tuples_nr > 0) {
				CALLOC(ret->bridging_tuples, ret->bridging_tuples_nr);

				for (i = 0; i < ret->bridging_tuples_nr; i++) {
					_E1B(&p, &ret->bridging_tuples[i].bridging_tuple_macs_nr);

					if (ret->bridging_tuples[i].bridging_tuple_macs_nr > 0) {
						CALLOC(ret->bridging_tuples[i].bridging_tuple_macs,
						       ret->bridging_tuples[i].bridging_tuple_macs_nr);

						for (j = 0; j < ret->bridging_tuples[i].bridging_tuple_macs_nr; j++) {
							_EnB(&p, ret->bridging_tuples[i].bridging_tuple_macs[j].mac_address, 6);
						}
					}
				}
			}

			if (p - (packet_stream + 3) != len) {
				// Malformed packet
				//
				for (i = 0; i < ret->bridging_tuples_nr; i++) {
					FREE(ret->bridging_tuples[i].bridging_tuple_macs);
				}
				FREE(ret->bridging_tuples);
				FREE(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.8"

			struct non1905NeighborDeviceListTLV *ret;

			uint8_t *p;
			uint16_t len;
			uint8_t i;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			// According to the standard, the length *must* be "6 + 6*n"
			//
			if (0 != ((len - 6) % 6)) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}
			ret->tlv_type = TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST;

			_EnB(&p, ret->local_mac_address, 6);

			ret->non_1905_neighbors_nr = (len - 6) / 6;

			CALLOC(ret->non_1905_neighbors, ret->non_1905_neighbors_nr);

			for (i = 0; i < ret->non_1905_neighbors_nr; i++) {
				_EnB(&p, ret->non_1905_neighbors[i].mac_address, 6);
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_NEIGHBOR_DEVICE_LIST:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.9"

			struct neighborDeviceListTLV *ret;

			uint8_t *p;
			uint16_t len;
			uint8_t i;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			// According to the standard, the length *must* be "6 + 7*n"
			// "6+1"
			//
			if (0 != ((len - 6) % 7)) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}
			ret->tlv_type = TLV_TYPE_NEIGHBOR_DEVICE_LIST;

			_EnB(&p, ret->local_mac_address, 6);

			ret->neighbors_nr = (len - 6) / 7;

			CALLOC(ret->neighbors, ret->neighbors_nr);

			for (i = 0; i < ret->neighbors_nr; i++) {
				uint8_t aux;

				_EnB(&p, ret->neighbors[i].mac_address, 6);
				_E1B(&p, &aux);

				if (aux & 0x80) {
					ret->neighbors[i].bridge_flag = 1;
				} else {
					ret->neighbors[i].bridge_flag = 0;
				}
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_LINK_METRIC_QUERY:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.10"

			struct linkMetricQueryTLV *ret;

			uint8_t *p;
			uint16_t len;

			uint8_t destination;
			uint8_t link_metrics_type;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			// According to the standard, the length *must* be 8
			//
			if (len != 8) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}

			ret->tlv_type = TLV_TYPE_LINK_METRIC_QUERY;

			_E1B(&p, &destination);
			_EnB(&p, ret->specific_neighbor, 6);

			if (destination >= 2) {
				// Reserved (invalid) value received
				//
				FREE(ret);
				return NULL;
			} else if (destination == 0) {
				ret->destination = LINK_METRIC_QUERY_TLV_ALL_NEIGHBORS;
				memset(ret->specific_neighbor, 0, 6);
			} else if (destination == 1) {
				ret->destination = LINK_METRIC_QUERY_TLV_SPECIFIC_NEIGHBOR;
			} else {
				// This code cannot be reached
				//
				FREE(ret);
				return NULL;
			}

			_E1B(&p, &link_metrics_type);

			if (link_metrics_type >= 3) {
				// Reserved (invalid) value received
				//
				FREE(ret);
				return NULL;
			} else if (link_metrics_type == 0) {
				ret->link_metrics_type = LINK_METRIC_QUERY_TLV_TX_LINK_METRICS_ONLY;
			} else if (link_metrics_type == 1) {
				ret->link_metrics_type = LINK_METRIC_QUERY_TLV_RX_LINK_METRICS_ONLY;
			} else if (link_metrics_type == 2) {
				ret->link_metrics_type = LINK_METRIC_QUERY_TLV_BOTH_TX_AND_RX_LINK_METRICS;
			} else {
				// This code cannot be reached
				//
				FREE(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_TRANSMITTER_LINK_METRIC:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.11"

			struct transmitterLinkMetricTLV *ret;

			uint8_t *p;
			uint16_t len;
			uint8_t i;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			// According to the standard, the length *must* be "12+29*n" where
			// "n" is "1" or greater
			//
			if ((12 + 29 * 1) > len) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}
			if (0 != (len - 12) % 29) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}

			ret->tlv_type = TLV_TYPE_TRANSMITTER_LINK_METRIC;

			_EnB(&p, ret->local_al_address, 6);
			_EnB(&p, ret->neighbor_al_address, 6);

			ret->transmitter_link_metrics_nr = (len - 12) / 29;

			CALLOC(ret->transmitter_link_metrics, ret->transmitter_link_metrics_nr);

			for (i = 0; i < ret->transmitter_link_metrics_nr; i++) {
				_EnB(&p, ret->transmitter_link_metrics[i].local_interface_address, 6);
				_EnB(&p, ret->transmitter_link_metrics[i].neighbor_interface_address, 6);

				_E2B(&p, &ret->transmitter_link_metrics[i].intf_type);
				_E1B(&p, &ret->transmitter_link_metrics[i].bridge_flag);
				_E4B(&p, &ret->transmitter_link_metrics[i].packet_errors);
				_E4B(&p, &ret->transmitter_link_metrics[i].transmitted_packets);
				_E2B(&p, &ret->transmitter_link_metrics[i].mac_throughput_capacity);
				_E2B(&p, &ret->transmitter_link_metrics[i].link_availability);
				_E2B(&p, &ret->transmitter_link_metrics[i].phy_rate);
			}

			if (p - (packet_stream + 3) != len) {
				// Malformed packet
				//
				FREE(ret->transmitter_link_metrics);
				FREE(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_RECEIVER_LINK_METRIC:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.12"

			struct receiverLinkMetricTLV *ret;

			uint8_t *p;
			uint16_t len;
			uint8_t i;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			// According to the standard, the length *must* be "12+23*n" where
			// "n" is "1" or greater
			//
			if ((12 + 23 * 1) > len) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}
			if (0 != (len - 12) % 23) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}

			ret->tlv_type = TLV_TYPE_RECEIVER_LINK_METRIC;

			_EnB(&p, ret->local_al_address, 6);
			_EnB(&p, ret->neighbor_al_address, 6);

			ret->receiver_link_metrics_nr = (len - 12) / 23;

			CALLOC(ret->receiver_link_metrics, ret->receiver_link_metrics_nr);

			for (i = 0; i < ret->receiver_link_metrics_nr; i++) {
				_EnB(&p, ret->receiver_link_metrics[i].local_interface_address, 6);
				_EnB(&p, ret->receiver_link_metrics[i].neighbor_interface_address, 6);

				_E2B(&p, &ret->receiver_link_metrics[i].intf_type);
				_E4B(&p, &ret->receiver_link_metrics[i].packet_errors);
				_E4B(&p, &ret->receiver_link_metrics[i].packets_received);
				_E1B(&p, &ret->receiver_link_metrics[i].rssi);
			}

			if (p - (packet_stream + 3) != len) {
				// Malformed packet
				//
				FREE(ret->receiver_link_metrics);
				FREE(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_LINK_METRIC_RESULT_CODE:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.13"

			struct linkMetricResultCodeTLV *ret;

			uint8_t *p;
			uint16_t len;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			// According to the standard, the length *must* be 1
			//
			if (len != 1) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}

			ret->tlv_type = TLV_TYPE_LINK_METRIC_RESULT_CODE;

			_E1B(&p, &ret->result_code);

			return (uint8_t *) ret;
		}

	case TLV_TYPE_SEARCHED_ROLE:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.14"

			struct searchedRoleTLV *ret;

			uint8_t *p;
			uint16_t len;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			// According to the standard, the length *must* be 1
			//
			if (len != 1) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}

			ret->tlv_type = TLV_TYPE_SEARCHED_ROLE;

			_E1B(&p, &ret->role);

			return (uint8_t *) ret;
		}

	case TLV_TYPE_AUTOCONFIG_FREQ_BAND:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.15"

			struct autoconfigFreqBandTLV *ret;

			uint8_t *p;
			uint16_t len;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			// According to the standard, the length *must* be 1
			//
			if (len != 1) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}

			ret->tlv_type = TLV_TYPE_AUTOCONFIG_FREQ_BAND;

			_E1B(&p, &ret->freq_band);

			return (uint8_t *) ret;
		}

	case TLV_TYPE_SUPPORTED_ROLE:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.16"

			struct supportedRoleTLV *ret;

			uint8_t *p;
			uint16_t len;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			// According to the standard, the length *must* be 1
			//
			if (len != 1) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}

			ret->tlv_type = TLV_TYPE_SUPPORTED_ROLE;

			_E1B(&p, &ret->role);

			return (uint8_t *) ret;
		}

	case TLV_TYPE_SUPPORTED_FREQ_BAND:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.17"

			struct supportedFreqBandTLV *ret;

			uint8_t *p;
			uint16_t len;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			// According to the standard, the length *must* be 1
			//
			if (len != 1) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}

			ret->tlv_type = TLV_TYPE_SUPPORTED_FREQ_BAND;

			_E1B(&p, &ret->freq_band);

			return (uint8_t *) ret;
		}

	case TLV_TYPE_WSC:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.18"

			struct wscTLV *ret;

			uint8_t *p;
			uint16_t len;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			ret->tlv_type = TLV_TYPE_WSC;
			ret->wsc_frame_size = len;

			if (len > 0) {
				CALLOC(ret->wsc_frame, len);
				_EnB(&p, ret->wsc_frame, len);
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.19"

			struct pushButtonEventNotificationTLV *ret;

			uint8_t *p;
			uint16_t len;
			uint8_t i;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			ret->tlv_type = TLV_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION;

			if (len == 0) {
				FREE(ret);
				return NULL;
			}

			_E1B(&p, &ret->media_types_nr);

			CALLOC(ret->media_types, ret->media_types_nr);

			for (i = 0; i < ret->media_types_nr; i++) {
				_E2B(&p, &ret->media_types[i].media_type);
				_E1B(&p, &ret->media_types[i].media_specific_data_size);

				if ((ret->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11B_2_4_GHZ) ||
				    (ret->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11G_2_4_GHZ) ||
				    (ret->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11A_5_GHZ) ||
				    (ret->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11N_2_4_GHZ) ||
				    (ret->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11N_5_GHZ) ||
				    (ret->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11AC_5_GHZ) ||
				    (ret->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11AD_60_GHZ) ||
				    (ret->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11AF_GHZ)
				    ) {
					uint8_t aux;

					if (ret->media_types[i].media_specific_data_size != 10) {
						// Malformed packet
						//
						FREE(ret->media_types);
						FREE(ret);
						return NULL;
					}

					_EnB(&p, ret->media_types[i].media_specific_data.ieee80211.network_membership, 6);
					_E1B(&p, &aux);
					ret->media_types[i].media_specific_data.ieee80211.role = aux >> 4;
					_E1B(&p, &ret->media_types[i].media_specific_data.ieee80211.ap_channel_band);
					_E1B(&p, &ret->media_types[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_1);
					_E1B(&p, &ret->media_types[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_2);

				} else if ((ret->media_types[i].media_type == MEDIA_TYPE_IEEE_1901_WAVELET) ||
					   (ret->media_types[i].media_type == MEDIA_TYPE_IEEE_1901_FFT)
				    ) {
					if (ret->media_types[i].media_specific_data_size != 7) {
						// Malformed packet
						//
						FREE(ret->media_types);
						FREE(ret);
						return NULL;
					}
					_EnB(&p, ret->media_types[i].media_specific_data.ieee1901.network_identifier, 7);
				} else {
					if (ret->media_types[i].media_specific_data_size != 0) {
						// Malformed packet
						//
						FREE(ret->media_types);
						FREE(ret);
						return NULL;
					}
				}
			}

			if (p - (packet_stream + 3) != len) {
				// Malformed packet
				//
				FREE(ret->media_types);
				FREE(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.20"

			struct pushButtonJoinNotificationTLV *ret;

			uint8_t *p;
			uint16_t len;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			// According to the standard, the length *must* be 20
			//
			if (len != 20) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}

			ret->tlv_type = TLV_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION;

			_EnB(&p, ret->al_mac_address, 6);
			_E2B(&p, &ret->message_identifier);
			_EnB(&p, ret->mac_address, 6);
			_EnB(&p, ret->new_mac_address, 6);

			return (uint8_t *) ret;
		}

	case TLV_TYPE_GENERIC_PHY_DEVICE_INFORMATION:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.21"

			struct genericPhyDeviceInformationTypeTLV *ret;

			uint8_t *p;
			uint16_t len;
			uint8_t i;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			ret->tlv_type = TLV_TYPE_GENERIC_PHY_DEVICE_INFORMATION;

			_EnB(&p, ret->al_mac_address, 6);
			_E1B(&p, &ret->local_interfaces_nr);

			if (ret->local_interfaces_nr > 0) {
				CALLOC(ret->local_interfaces, ret->local_interfaces_nr);

				for (i = 0; i < ret->local_interfaces_nr; i++) {
					_EnB(&p, ret->local_interfaces[i].local_interface_address, 6);
					_EnB(&p, ret->local_interfaces[i].generic_phy_common_data.oui, 3);
					_E1B(&p, &ret->local_interfaces[i].generic_phy_common_data.variant_index);
					_EnB(&p, ret->local_interfaces[i].variant_name, 32);
					_E1B(&p, &ret->local_interfaces[i].generic_phy_description_xml_url_len);
					_E1B(&p, &ret->local_interfaces[i].generic_phy_common_data.media_specific_bytes_nr);

					if (ret->local_interfaces[i].generic_phy_description_xml_url_len > 0) {
						CALLOC(ret->local_interfaces[i].generic_phy_description_xml_url, \
							ret->local_interfaces[i].generic_phy_description_xml_url_len);
						_EnB(&p, ret->local_interfaces[i].generic_phy_description_xml_url,
						     ret->local_interfaces[i].generic_phy_description_xml_url_len);
					}

					if (ret->local_interfaces[i].generic_phy_common_data.media_specific_bytes_nr > 0) {
						CALLOC(ret->local_interfaces[i].generic_phy_common_data.media_specific_bytes,
						       ret->local_interfaces[i].generic_phy_common_data.media_specific_bytes_nr);
						_EnB(&p, ret->local_interfaces[i].generic_phy_common_data.media_specific_bytes,
						     ret->local_interfaces[i].generic_phy_common_data.media_specific_bytes_nr);
					}
				}
			}

			if (p - (packet_stream + 3) != len) {
				// Malformed packet
				//
				for (i = 0; i < ret->local_interfaces_nr; i++) {
					if (ret->local_interfaces[i].generic_phy_description_xml_url_len > 0) {
						FREE(ret->local_interfaces[i].generic_phy_description_xml_url);
					}

					if (ret->local_interfaces[i].generic_phy_common_data.media_specific_bytes_nr > 0) {
						FREE(ret->local_interfaces[i].generic_phy_common_data.media_specific_bytes);
					}
				}
				FREE(ret->local_interfaces);
				FREE(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_DEVICE_IDENTIFICATION:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.22"

			struct deviceIdentificationTypeTLV *ret;

			uint8_t *p;
			uint16_t len;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			// According to the standard, the length *must* be 20
			//
			if (len != 192) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}

			ret->tlv_type = TLV_TYPE_DEVICE_IDENTIFICATION;

			_EnB(&p, ret->friendly_name, 64);
			_EnB(&p, ret->manufacturer_name, 64);
			_EnB(&p, ret->manufacturer_model, 64);

			return (uint8_t *) ret;
		}

	case TLV_TYPE_CONTROL_URL:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.23"

			struct controlUrlTypeTLV *ret;

			uint8_t *p;
			uint16_t len;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			ret->tlv_type = TLV_TYPE_CONTROL_URL;

			if (len > 0) {
				CALLOC(ret->url, len);
				_EnB(&p, ret->url, len);
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_IPV4:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.24"

			struct ipv4TypeTLV *ret;

			uint8_t *p;
			uint16_t len;
			uint8_t i, j;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			ret->tlv_type = TLV_TYPE_IPV4;

			if (len == 0) {
				FREE(ret);
				return NULL;
			}

			_E1B(&p, &ret->ipv4_interfaces_nr);

			if (ret->ipv4_interfaces_nr > 0) {
				CALLOC(ret->ipv4_interfaces, ret->ipv4_interfaces_nr);

				for (i = 0; i < ret->ipv4_interfaces_nr; i++) {
					_EnB(&p, ret->ipv4_interfaces[i].mac_address, 6);
					_E1B(&p, &ret->ipv4_interfaces[i].ipv4_nr);

					if (ret->ipv4_interfaces[i].ipv4_nr > 0) {
						CALLOC(ret->ipv4_interfaces[i].ipv4, ret->ipv4_interfaces[i].ipv4_nr);

						for (j = 0; j < ret->ipv4_interfaces[i].ipv4_nr; j++) {
							_E1B(&p, &ret->ipv4_interfaces[i].ipv4[j].type);
							_EnB(&p, ret->ipv4_interfaces[i].ipv4[j].ipv4_address, 4);
							_EnB(&p, ret->ipv4_interfaces[i].ipv4[j].ipv4_dhcp_server, 4);
						}
					}
				}
			}

			if (p - (packet_stream + 3) != len) {
				// Malformed packet
				//
				for (i = 0; i < ret->ipv4_interfaces_nr; i++) {
					if (ret->ipv4_interfaces[i].ipv4_nr > 0) {
						FREE(ret->ipv4_interfaces[i].ipv4);
					}
				}
				FREE(ret->ipv4_interfaces);
				FREE(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_IPV6:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.25"

			struct ipv6TypeTLV *ret;

			uint8_t *p;
			uint16_t len;
			uint8_t i, j;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			ret->tlv_type = TLV_TYPE_IPV6;

			if (len == 0) {
				FREE(ret);
				return NULL;
			}

			_E1B(&p, &ret->ipv6_interfaces_nr);

			if (ret->ipv6_interfaces_nr > 0) {
				CALLOC(ret->ipv6_interfaces, ret->ipv6_interfaces_nr);

				for (i = 0; i < ret->ipv6_interfaces_nr; i++) {
					_EnB(&p, ret->ipv6_interfaces[i].mac_address, 6);
					_EnB(&p, ret->ipv6_interfaces[i].ipv6_link_local_address, 16);
					_E1B(&p, &ret->ipv6_interfaces[i].ipv6_nr);

					if (ret->ipv6_interfaces[i].ipv6_nr > 0) {
						CALLOC(ret->ipv6_interfaces[i].ipv6, ret->ipv6_interfaces[i].ipv6_nr);

						for (j = 0; j < ret->ipv6_interfaces[i].ipv6_nr; j++) {
							_E1B(&p, &ret->ipv6_interfaces[i].ipv6[j].type);
							_EnB(&p, ret->ipv6_interfaces[i].ipv6[j].ipv6_address, 16);
							_EnB(&p, ret->ipv6_interfaces[i].ipv6[j].ipv6_address_origin, 16);
						}
					}
				}
			}

			if (p - (packet_stream + 3) != len) {
				// Malformed packet
				//
				for (i = 0; i < ret->ipv6_interfaces_nr; i++) {
					if (ret->ipv6_interfaces[i].ipv6_nr > 0) {
						FREE(ret->ipv6_interfaces[i].ipv6);
					}
				}
				FREE(ret->ipv6_interfaces);
				FREE(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_GENERIC_PHY_EVENT_NOTIFICATION:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.26"

			struct pushButtonGenericPhyEventNotificationTLV *ret;

			uint8_t *p;
			uint16_t len;
			uint8_t i;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			ret->tlv_type = TLV_TYPE_GENERIC_PHY_EVENT_NOTIFICATION;

			if (len == 0) {
				FREE(ret);
				return NULL;
			}

			_E1B(&p, &ret->local_interfaces_nr);

			if (ret->local_interfaces_nr > 0) {
				CALLOC(ret->local_interfaces, ret->local_interfaces_nr);

				for (i = 0; i < ret->local_interfaces_nr; i++) {
					_EnB(&p, ret->local_interfaces[i].oui, 3);
					_E1B(&p, &ret->local_interfaces[i].variant_index);
					_E1B(&p, &ret->local_interfaces[i].media_specific_bytes_nr);

					if (ret->local_interfaces[i].media_specific_bytes_nr > 0) {
						CALLOC(ret->local_interfaces[i].media_specific_bytes,
						       ret->local_interfaces[i].media_specific_bytes_nr);
						_EnB(&p, ret->local_interfaces[i].media_specific_bytes,
						     ret->local_interfaces[i].media_specific_bytes_nr);
					}
				}
			}

			if (p - (packet_stream + 3) != len) {
				// Malformed packet
				//
				for (i = 0; i < ret->local_interfaces_nr; i++) {
					if (ret->local_interfaces[i].media_specific_bytes_nr > 0) {
						FREE(ret->local_interfaces[i].media_specific_bytes);
					}
				}
				FREE(ret->local_interfaces);
				FREE(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_1905_PROFILE_VERSION:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.27"

			struct x1905ProfileVersionTLV *ret;

			uint8_t *p;
			uint16_t len;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			// According to the standard, the length *must* be 1
			//
			if (len != 1) {
				// Malformed packet
				//
				FREE(ret);
				return NULL;
			}

			ret->tlv_type = TLV_TYPE_1905_PROFILE_VERSION;

			_E1B(&p, &ret->profile);

			return (uint8_t *) ret;
		}

	case TLV_TYPE_POWER_OFF_INTERFACE:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.28"

			struct powerOffInterfaceTLV *ret;

			uint8_t *p;
			uint16_t len;
			uint8_t i;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			ret->tlv_type = TLV_TYPE_POWER_OFF_INTERFACE;

			if (len == 0) {
				FREE(ret);
				return NULL;
			}

			_E1B(&p, &ret->power_off_interfaces_nr);

			if (ret->power_off_interfaces_nr > 0) {
				CALLOC(ret->power_off_interfaces, ret->power_off_interfaces_nr);

				for (i = 0; i < ret->power_off_interfaces_nr; i++) {
					_EnB(&p, ret->power_off_interfaces[i].interface_address, 6);
					_E2B(&p, &ret->power_off_interfaces[i].media_type);
					_EnB(&p, ret->power_off_interfaces[i].generic_phy_common_data.oui, 3);
					_E1B(&p, &ret->power_off_interfaces[i].generic_phy_common_data.variant_index);
					_E1B(&p, &ret->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes_nr);

					if (ret->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes_nr > 0) {
						CALLOC(ret->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes, \
						       ret->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes_nr);
						_EnB(&p, ret->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes,
						     ret->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes_nr);
					}
				}
			}

			if (p - (packet_stream + 3) != len) {
				// Malformed packet
				//
				for (i = 0; i < ret->power_off_interfaces_nr; i++) {
					if (ret->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes_nr > 0) {
						FREE(ret->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes);
					}
				}
				FREE(ret->power_off_interfaces);
				FREE(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_INTERFACE_POWER_CHANGE_INFORMATION:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.29"

			struct interfacePowerChangeInformationTLV *ret;

			uint8_t *p;
			uint16_t len;
			uint8_t i;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			ret->tlv_type = TLV_TYPE_INTERFACE_POWER_CHANGE_INFORMATION;

			if (len == 0) {
				FREE(ret);
				return NULL;
			}

			_E1B(&p, &ret->power_change_interfaces_nr);

			if (ret->power_change_interfaces_nr > 0) {
				CALLOC(ret->power_change_interfaces, ret->power_change_interfaces_nr);

				for (i = 0; i < ret->power_change_interfaces_nr; i++) {
					_EnB(&p, ret->power_change_interfaces[i].interface_address, 6);
					_E1B(&p, &ret->power_change_interfaces[i].requested_power_state);
				}
			}

			if (p - (packet_stream + 3) != len) {
				// Malformed packet
				//
				FREE(ret->power_change_interfaces);
				FREE(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_INTERFACE_POWER_CHANGE_STATUS:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.30"

			struct interfacePowerChangeStatusTLV *ret;

			uint8_t *p;
			uint16_t len;
			uint8_t i;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			ret->tlv_type = TLV_TYPE_INTERFACE_POWER_CHANGE_STATUS;

			if (len == 0) {
				FREE(ret);
				return NULL;
			}

			_E1B(&p, &ret->power_change_interfaces_nr);

			if (ret->power_change_interfaces_nr > 0) {
				CALLOC(ret->power_change_interfaces, ret->power_change_interfaces_nr);

				for (i = 0; i < ret->power_change_interfaces_nr; i++) {
					_EnB(&p, ret->power_change_interfaces[i].interface_address, 6);
					_E1B(&p, &ret->power_change_interfaces[i].result);
				}
			}

			if (p - (packet_stream + 3) != len) {
				// Malformed packet
				//
				if (ret->power_change_interfaces_nr > 0) {
					FREE(ret->power_change_interfaces);
				}
				FREE(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	case TLV_TYPE_L2_NEIGHBOR_DEVICE:
		{
			// This parsing is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.31"

			struct l2NeighborDeviceTLV *ret;

			uint8_t *p;
			uint16_t len;
			uint8_t i, j, k;

			CALLOC(ret, 1);

			p = packet_stream + 1;
			_E2B(&p, &len);

			ret->tlv_type = TLV_TYPE_L2_NEIGHBOR_DEVICE;

			if (len == 0) {
				FREE(ret);
				return NULL;
			}

			_E1B(&p, &ret->local_interfaces_nr);

			if (ret->local_interfaces_nr > 0) {
				CALLOC(ret->local_interfaces, ret->local_interfaces_nr);

				for (i = 0; i < ret->local_interfaces_nr; i++) {
					_EnB(&p, ret->local_interfaces[i].local_mac_address, 6);
					_E2B(&p, &ret->local_interfaces[i].l2_neighbors_nr);

					if (ret->local_interfaces[i].l2_neighbors_nr > 0) {
						CALLOC(ret->local_interfaces[i].l2_neighbors, \
						       ret->local_interfaces[i].l2_neighbors_nr);

						for (j = 0; j < ret->local_interfaces[i].l2_neighbors_nr; j++) {
							_EnB(&p, ret->local_interfaces[i].l2_neighbors[j].l2_neighbor_mac_address, 6);
							_E2B(&p, &ret->local_interfaces[i].l2_neighbors[j].behind_mac_addresses_nr);

							if (ret->local_interfaces[i].l2_neighbors[j].behind_mac_addresses_nr > 0) {
								ret->local_interfaces[i].l2_neighbors[j].behind_mac_addresses =
								    (uint8_t(*)[6]) malloc(sizeof(uint8_t[6]) *
											      ret->local_interfaces[i].l2_neighbors[j].behind_mac_addresses_nr);

								for (k = 0; k < ret->local_interfaces[i].l2_neighbors[j].behind_mac_addresses_nr; k++) {
									_EnB(&p, ret->local_interfaces[i].l2_neighbors[j].behind_mac_addresses[k], 6);
								}
							}
						}
					}
				}
			}

			if (p - (packet_stream + 3) != len) {
				// Malformed packet
				//
				for (i = 0; i < ret->local_interfaces_nr; i++) {
					for (j = 0; j < ret->local_interfaces[i].l2_neighbors_nr; j++) {
						FREE(ret->local_interfaces[i].l2_neighbors[j].behind_mac_addresses);
					}
					FREE(ret->local_interfaces[i].l2_neighbors);
				}
				FREE(ret->local_interfaces);
				FREE(ret);
				return NULL;
			}

			return (uint8_t *) ret;
		}

	default:
		{
			// Ignore
			//
			return NULL;
		}

	}

	// This code cannot be reached
	//
	return NULL;
}

uint8_t *forge_1905_TLV_from_structure(uint8_t *memory_structure)
{
	if (memory_structure == NULL) {
		return NULL;
	}
	// The first byte of any of the valid structures is always the "tlv_type"
	// field.
	//
	switch (*memory_structure) {
	case TLV_TYPE_END_OF_MESSAGE:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.1"

			uint8_t *ret, *p;
			struct endOfMessageTLV *m;

			uint16_t tlv_length;

			m = (struct endOfMessageTLV *)memory_structure;

			tlv_length = 0;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			return ret;
		}

	case TLV_TYPE_VENDOR_SPECIFIC:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.2"

			uint8_t *ret, *p;
			struct vendorSpecificTLV *m;

			uint16_t tlv_length;

			m = (struct vendorSpecificTLV *)memory_structure;

			tlv_length = 3 + m->m_nr;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_I1B(&m->vendorOUI[0], &p);
			_I1B(&m->vendorOUI[1], &p);
			_I1B(&m->vendorOUI[2], &p);
			_InB(m->m, &p, m->m_nr);

			return ret;
		}

	case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.3"

			uint8_t *ret, *p;
			struct alMacAddressTypeTLV *m;

			uint16_t tlv_length;

			m = (struct alMacAddressTypeTLV *)memory_structure;

			tlv_length = 6;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->al_mac_address, &p, 6);

			return ret;
		}

	case TLV_TYPE_MAC_ADDRESS_TYPE:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.4"

			uint8_t *ret, *p;
			struct macAddressTypeTLV *m;

			uint16_t tlv_length;

			m = (struct macAddressTypeTLV *)memory_structure;

			tlv_length = 6;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->mac_address, &p, 6);

			return ret;
		}

	case TLV_TYPE_DEVICE_INFORMATION_TYPE:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.5"

			uint8_t *ret, *p;
			struct deviceInformationTypeTLV *m;

			uint16_t tlv_length;

			uint8_t i;

			m = (struct deviceInformationTypeTLV *)memory_structure;

			tlv_length = 7;	// AL MAC (6 bytes) + number of ifaces (1 bytes)
			for (i = 0; i < m->local_interfaces_nr; i++) {
				tlv_length += 6 + 2 + 1;	// MAC (6 bytes) + media type (2
				// bytes) + number of octets (1 byte)

				tlv_length += m->local_interfaces[i].media_specific_data_size;
			}

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->al_mac_address, &p, 6);
			_I1B(&m->local_interfaces_nr, &p);

			for (i = 0; i < m->local_interfaces_nr; i++) {
				_InB(m->local_interfaces[i].mac_address, &p, 6);
				_I2B(&m->local_interfaces[i].media_type, &p);
				_I1B(&m->local_interfaces[i].media_specific_data_size, &p);

				if ((m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11B_2_4_GHZ) ||
				    (m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11G_2_4_GHZ) ||
				    (m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11A_5_GHZ) ||
				    (m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11N_2_4_GHZ) ||
				    (m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11N_5_GHZ) ||
				    (m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AC_5_GHZ) ||
				    (m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AD_60_GHZ) ||
				    (m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AF_GHZ)
				    ) {
					uint8_t aux;

					if (m->local_interfaces[i].media_specific_data_size != 10) {
						// Malformed structure
						//
						FREE(ret);
						return NULL;
					}

					_InB(m->local_interfaces[i].media_specific_data.ieee80211.network_membership, &p, 6);
					aux = m->local_interfaces[i].media_specific_data.ieee80211.role << 4;
					_I1B(&aux, &p);
					_I1B(&m->local_interfaces[i].media_specific_data.ieee80211.ap_channel_band, &p);
					_I1B(&m->local_interfaces[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_1, &p);
					_I1B(&m->local_interfaces[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_2, &p);

				} else if ((m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_1901_WAVELET) ||
					   (m->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_1901_FFT)
				    ) {
					if (m->local_interfaces[i].media_specific_data_size != 7) {
						// Malformed structure
						//
						FREE(ret);
						return NULL;
					}
					_InB(m->local_interfaces[i].media_specific_data.ieee1901.network_identifier, &p, 7);
				} else {
					if (m->local_interfaces[i].media_specific_data_size != 0) {
						// Malformed structure
						//
						FREE(ret);
						return NULL;
					}
				}
			}

			return ret;
		}

	case TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.6"

			uint8_t *ret, *p;
			struct deviceBridgingCapabilityTLV *m;

			uint16_t tlv_length;

			uint8_t i, j;

			m = (struct deviceBridgingCapabilityTLV *)memory_structure;

			tlv_length = 1;	// number of bridging tuples (1 bytes)
			for (i = 0; i < m->bridging_tuples_nr; i++) {
				tlv_length += 1;	// number of MAC addresses (1 bytes)
				tlv_length += 6 * m->bridging_tuples[i].bridging_tuple_macs_nr;
			}

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_I1B(&m->bridging_tuples_nr, &p);

			for (i = 0; i < m->bridging_tuples_nr; i++) {
				_I1B(&m->bridging_tuples[i].bridging_tuple_macs_nr, &p);

				for (j = 0; j < m->bridging_tuples[i].bridging_tuple_macs_nr; j++) {
					_InB(m->bridging_tuples[i].bridging_tuple_macs[j].mac_address, &p, 6);
				}
			}

			return ret;
		}

	case TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.8"

			uint8_t *ret, *p;
			struct non1905NeighborDeviceListTLV *m;

			uint16_t tlv_length;

			uint8_t i;

			m = (struct non1905NeighborDeviceListTLV *)memory_structure;

			tlv_length = 6 + 6 * m->non_1905_neighbors_nr;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->local_mac_address, &p, 6);

			for (i = 0; i < m->non_1905_neighbors_nr; i++) {
				_InB(m->non_1905_neighbors[i].mac_address, &p, 6);
			}

			return ret;
		}

	case TLV_TYPE_NEIGHBOR_DEVICE_LIST:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.9"

			uint8_t *ret, *p;
			struct neighborDeviceListTLV *m;

			uint16_t tlv_length;

			uint8_t i;

			m = (struct neighborDeviceListTLV *)memory_structure;

			tlv_length = 6 + 7 * m->neighbors_nr;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->local_mac_address, &p, 6);

			for (i = 0; i < m->neighbors_nr; i++) {
				uint8_t aux;

				_InB(m->neighbors[i].mac_address, &p, 6);

				if (m->neighbors[i].bridge_flag == 1) {
					aux = 1 << 7;
					_I1B(&aux, &p);
				} else {
					aux = 0;
					_I1B(&aux, &p);
				}
			}

			return ret;
		}

	case TLV_TYPE_LINK_METRIC_QUERY:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.10"

			uint8_t *ret, *p;
			struct linkMetricQueryTLV *m;

			uint16_t tlv_length;

			m = (struct linkMetricQueryTLV *)memory_structure;

			tlv_length = 8;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_I1B(&m->destination, &p);

			if (m->destination == LINK_METRIC_QUERY_TLV_SPECIFIC_NEIGHBOR) {
				_InB(m->specific_neighbor, &p, 6);
			} else {
				uint8_t empty_address[] = { m->link_metrics_type, 0x00, 0x00, 0x00, 0x00, 0x00 };

				_InB(empty_address, &p, 6);

				// Ugh? Why is the first value set to "m->link_metrics_type"
				// instead of "0x00"? What kind of black magic is this?
				//
				// Well... it turns out there is a reason for this. Take a
				// chair and let me explain.
				//
				// The original 1905 standard document (and also its later "1a"
				// update) describe the "metric query TLV" fields like this:
				//
				//   - Field #1: 1 octet set to "8" (tlv_type)
				//   - Field #2: 1 octet set to "8" (tlv_length)
				//   - Field #3: 1 octet set to "0" or "1" (destination)
				//   - Field #4: 6 octets set to the MAC address of a neighbour
				//               when field #3 is set "1"
				//   - Field #5: 1 octet set to "0", "1", "2" or "3" (link_
				//               _metrics_type)
				//
				// The problem is that we don't know what to put inside field
				// #4 when Field #3 is set to "0" ("all neighbors") instead of
				// "1" ("specific neighbor").
				//
				// A "reasonable" solution would be to set all bytes from field
				// #4 to "0x00". *However*, one could also think that the
				// correct thing to do is to not include the field at all (ie.
				// skip from field #3 to field #5).
				//
				// Now... this is actually insane. Typically protocols have a
				// fixed number of fields (whenever possible) to make it easier
				// for parsers (in fact, this would be the only exception to
				// this rule in the whole 1905 standard). Then... why would
				// someone think that not including field #4 is a good idea?
				//
				// Well... because this is what the "description" of field #3
				// reads on the standard:
				//
				//   "If the value is 0, then the EUI-48 field is not present;
				//    if the value is 1, then the EUI-48 field shall be present"
				//
				// ...and "not present" seems to imply not to include it
				// (although one could argue that it could also mean "set all
				// bytes to zero).
				//
				// I really think the standard means "set to zero" instead of
				// "not including it" (even if the wording seems to imply
				// otherwise). Why? For two reasons:
				//
				//   1. The standard says field #2 must *always* be "8" (and if
				//      field #4 could not be included, this value should be
				//      allowed to also take the value of 6)
				//
				//   2. There is no other place in the whole standard where a
				//      field can be present or not.
				//
				// Despite what I have just said, *some implementations* seem
				// to have taken the other route, and expect field #4 *not* to
				// be present (even if field #2 is set to "8"!!).
				//
				// When we send one "all neighbors" topology query to one of
				// these implementations they will interpret the first byte of
				// field #4 as the contents of field #5.
				//
				// And that's why when querying for all neighbors, because the
				// contents of field #4 don't really matter, we are going to
				// set its first byte to the same value as field #5.
				// This way all implementations, no matter how they decided to
				// interpret the standard, will work :)
			}

			_I1B(&m->link_metrics_type, &p);

			return ret;
		}

	case TLV_TYPE_TRANSMITTER_LINK_METRIC:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.11"

			uint8_t *ret, *p;
			struct transmitterLinkMetricTLV *m;

			uint16_t tlv_length;

			uint8_t i;

			m = (struct transmitterLinkMetricTLV *)memory_structure;

			tlv_length = 12 + 29 * m->transmitter_link_metrics_nr;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->local_al_address, &p, 6);
			_InB(m->neighbor_al_address, &p, 6);

			for (i = 0; i < m->transmitter_link_metrics_nr; i++) {
				_InB(m->transmitter_link_metrics[i].local_interface_address, &p, 6);
				_InB(m->transmitter_link_metrics[i].neighbor_interface_address, &p, 6);
				_I2B(&m->transmitter_link_metrics[i].intf_type, &p);
				_I1B(&m->transmitter_link_metrics[i].bridge_flag, &p);
				_I4B(&m->transmitter_link_metrics[i].packet_errors, &p);
				_I4B(&m->transmitter_link_metrics[i].transmitted_packets, &p);
				_I2B(&m->transmitter_link_metrics[i].mac_throughput_capacity, &p);
				_I2B(&m->transmitter_link_metrics[i].link_availability, &p);
				_I2B(&m->transmitter_link_metrics[i].phy_rate, &p);
			}

			return ret;
		}

	case TLV_TYPE_RECEIVER_LINK_METRIC:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.12"

			uint8_t *ret, *p;
			struct receiverLinkMetricTLV *m;

			uint16_t tlv_length;

			uint8_t i;

			m = (struct receiverLinkMetricTLV *)memory_structure;

			tlv_length = 12 + 23 * m->receiver_link_metrics_nr;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->local_al_address, &p, 6);
			_InB(m->neighbor_al_address, &p, 6);

			for (i = 0; i < m->receiver_link_metrics_nr; i++) {
				_InB(m->receiver_link_metrics[i].local_interface_address, &p, 6);
				_InB(m->receiver_link_metrics[i].neighbor_interface_address, &p, 6);
				_I2B(&m->receiver_link_metrics[i].intf_type, &p);
				_I4B(&m->receiver_link_metrics[i].packet_errors, &p);
				_I4B(&m->receiver_link_metrics[i].packets_received, &p);
				_I1B(&m->receiver_link_metrics[i].rssi, &p);
			}

			return ret;
		}

	case TLV_TYPE_LINK_METRIC_RESULT_CODE:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.13"

			uint8_t *ret, *p;
			struct linkMetricResultCodeTLV *m;

			uint16_t tlv_length;

			m = (struct linkMetricResultCodeTLV *)memory_structure;

			tlv_length = 1;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			if (m->result_code != LINK_METRIC_RESULT_CODE_TLV_INVALID_NEIGHBOR) {
				// Malformed structure
				//
				FREE(ret);
				return NULL;
			}

			_I1B(&m->result_code, &p);

			return ret;
		}

	case TLV_TYPE_SEARCHED_ROLE:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.14"

			uint8_t *ret, *p;
			struct searchedRoleTLV *m;

			uint16_t tlv_length;

			m = (struct searchedRoleTLV *)memory_structure;

			tlv_length = 1;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			if (m->role != IEEE80211_ROLE_REGISTRAR) {
				// Malformed structure
				//
				FREE(ret);
				return NULL;
			}

			_I1B(&m->role, &p);

			return ret;
		}

	case TLV_TYPE_AUTOCONFIG_FREQ_BAND:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.14"

			uint8_t *ret, *p;
			struct autoconfigFreqBandTLV *m;

			uint16_t tlv_length;

			m = (struct autoconfigFreqBandTLV *)memory_structure;

			tlv_length = 1;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			if ((m->freq_band != IEEE80211_FREQUENCY_BAND_2_4_GHZ) &&
			    (m->freq_band != IEEE80211_FREQUENCY_BAND_5_GHZ) && (m->freq_band != IEEE80211_FREQUENCY_BAND_60_GHZ)
			    ) {
				// Malformed structure
				//
				FREE(ret);
				return NULL;
			}

			_I1B(&m->freq_band, &p);

			return ret;
		}

	case TLV_TYPE_SUPPORTED_ROLE:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.16"

			uint8_t *ret, *p;
			struct supportedRoleTLV *m;

			uint16_t tlv_length;

			m = (struct supportedRoleTLV *)memory_structure;

			tlv_length = 1;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			if (m->role != IEEE80211_ROLE_REGISTRAR) {
				// Malformed structure
				//
				FREE(ret);
				return NULL;
			}

			_I1B(&m->role, &p);

			return ret;
		}

	case TLV_TYPE_SUPPORTED_FREQ_BAND:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.17"

			uint8_t *ret, *p;
			struct supportedFreqBandTLV *m;

			uint16_t tlv_length;

			m = (struct supportedFreqBandTLV *)memory_structure;

			tlv_length = 1;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			if ((m->freq_band != IEEE80211_FREQUENCY_BAND_2_4_GHZ) &&
			    (m->freq_band != IEEE80211_FREQUENCY_BAND_5_GHZ) && (m->freq_band != IEEE80211_FREQUENCY_BAND_60_GHZ)
			    ) {
				// Malformed structure
				//
				FREE(ret);
				return NULL;
			}

			_I1B(&m->freq_band, &p);

			return ret;
		}

	case TLV_TYPE_WSC:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.18"

			uint8_t *ret, *p;
			struct wscTLV *m;

			uint16_t tlv_length;

			m = (struct wscTLV *)memory_structure;

			tlv_length = m->wsc_frame_size;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->wsc_frame, &p, m->wsc_frame_size);

			return ret;
		}

	case TLV_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.19"

			uint8_t *ret, *p;
			struct pushButtonEventNotificationTLV *m;

			uint16_t tlv_length;

			uint8_t i;

			m = (struct pushButtonEventNotificationTLV *)memory_structure;

			tlv_length = 1;	// number of media types (1 byte)
			for (i = 0; i < m->media_types_nr; i++) {
				tlv_length += 2 + 1;	//  media type (2 bytes) +
				//  number of octets (1 byte)

				tlv_length += m->media_types[i].media_specific_data_size;
			}

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_I1B(&m->media_types_nr, &p);

			for (i = 0; i < m->media_types_nr; i++) {
				_I2B(&m->media_types[i].media_type, &p);
				_I1B(&m->media_types[i].media_specific_data_size, &p);

				if ((m->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11B_2_4_GHZ) ||
				    (m->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11G_2_4_GHZ) ||
				    (m->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11A_5_GHZ) ||
				    (m->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11N_2_4_GHZ) ||
				    (m->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11N_5_GHZ) ||
				    (m->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11AC_5_GHZ) ||
				    (m->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11AD_60_GHZ) ||
				    (m->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11AF_GHZ)
				    ) {
					uint8_t aux;

					if (m->media_types[i].media_specific_data_size != 10) {
						// Malformed structure
						//
						FREE(ret);
						return NULL;
					}

					_InB(m->media_types[i].media_specific_data.ieee80211.network_membership, &p, 6);
					aux = m->media_types[i].media_specific_data.ieee80211.role << 4;
					_I1B(&aux, &p);
					_I1B(&m->media_types[i].media_specific_data.ieee80211.ap_channel_band, &p);
					_I1B(&m->media_types[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_1, &p);
					_I1B(&m->media_types[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_2, &p);

				} else if ((m->media_types[i].media_type == MEDIA_TYPE_IEEE_1901_WAVELET) ||
					   (m->media_types[i].media_type == MEDIA_TYPE_IEEE_1901_FFT)
				    ) {
					if (m->media_types[i].media_specific_data_size != 7) {
						// Malformed structure
						//
						FREE(ret);
						return NULL;
					}
					_InB(m->media_types[i].media_specific_data.ieee1901.network_identifier, &p, 7);
				} else {
					if (m->media_types[i].media_specific_data_size != 0) {
						// Malformed structure
						//
						FREE(ret);
						return NULL;
					}
				}
			}

			return ret;
		}

	case TLV_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.20"

			uint8_t *ret, *p;
			struct pushButtonJoinNotificationTLV *m;

			uint16_t tlv_length;

			m = (struct pushButtonJoinNotificationTLV *)memory_structure;

			tlv_length = 20;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->al_mac_address, &p, 6);
			_I2B(&m->message_identifier, &p);
			_InB(m->mac_address, &p, 6);
			_InB(m->new_mac_address, &p, 6);

			return ret;
		}

	case TLV_TYPE_GENERIC_PHY_DEVICE_INFORMATION:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.21"

			uint8_t *ret, *p;
			struct genericPhyDeviceInformationTypeTLV *m;

			uint16_t tlv_length;

			uint8_t i;

			m = (struct genericPhyDeviceInformationTypeTLV *)memory_structure;

			tlv_length = 6;	// AL MAC address (6 bytes)
			tlv_length += 1;	// number of local interfaces (1 bytes)
			for (i = 0; i < m->local_interfaces_nr; i++) {
				tlv_length += 6;	// local interface MAC address (6 bytes)
				tlv_length += 3;	// OUI (2 bytes)
				tlv_length += 1;	// variant_index (1 byte)
				tlv_length += 32;	// variant_name (32 bytes)
				tlv_length += 1;	// URL len (1 byte)
				tlv_length += 1;	// media specific bytes number (1 bytes)
				tlv_length += m->local_interfaces[i].generic_phy_description_xml_url_len;
				// URL bytes
				tlv_length += m->local_interfaces[i].generic_phy_common_data.media_specific_bytes_nr;
				// media specific bytes
			}

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->al_mac_address, &p, 6);
			_I1B(&m->local_interfaces_nr, &p);

			for (i = 0; i < m->local_interfaces_nr; i++) {
				_InB(m->local_interfaces[i].local_interface_address, &p, 6);
				_InB(m->local_interfaces[i].generic_phy_common_data.oui, &p, 3);
				_I1B(&m->local_interfaces[i].generic_phy_common_data.variant_index, &p);
				_InB(m->local_interfaces[i].variant_name, &p, 32);
				_I1B(&m->local_interfaces[i].generic_phy_description_xml_url_len, &p);
				_I1B(&m->local_interfaces[i].generic_phy_common_data.media_specific_bytes_nr, &p);
				if (m->local_interfaces[i].generic_phy_description_xml_url_len > 0) {
					_InB(m->local_interfaces[i].generic_phy_description_xml_url, &p,
					     m->local_interfaces[i].generic_phy_description_xml_url_len);
				}
				if (m->local_interfaces[i].generic_phy_common_data.media_specific_bytes_nr > 0) {
					_InB(m->local_interfaces[i].generic_phy_common_data.media_specific_bytes, &p,
					     m->local_interfaces[i].generic_phy_common_data.media_specific_bytes_nr);
				}
			}

			return ret;
		}

	case TLV_TYPE_DEVICE_IDENTIFICATION:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.21"

			uint8_t *ret, *p;
			struct deviceIdentificationTypeTLV *m;

			uint16_t tlv_length;

			m = (struct deviceIdentificationTypeTLV *)memory_structure;

			tlv_length = 192;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->friendly_name, &p, 64);
			_InB(m->manufacturer_name, &p, 64);
			_InB(m->manufacturer_model, &p, 64);

			return ret;
		}

	case TLV_TYPE_CONTROL_URL:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.23"

			uint8_t *ret, *p;
			struct controlUrlTypeTLV *m;

			uint16_t tlv_length;

			m = (struct controlUrlTypeTLV *)memory_structure;

			tlv_length = strlen(m->url) + 1;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_InB(m->url, &p, tlv_length);

			return ret;
		}

	case TLV_TYPE_IPV4:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.24"

			uint8_t *ret, *p;
			struct ipv4TypeTLV *m;

			uint16_t tlv_length;

			uint8_t i, j;

			m = (struct ipv4TypeTLV *)memory_structure;

			tlv_length = 1;	// number of entries (1 bytes)
			for (i = 0; i < m->ipv4_interfaces_nr; i++) {
				tlv_length += 6;	// interface MAC address (6 bytes)
				tlv_length += 1;	// number of IPv4s (1 bytes)
				tlv_length += (1 + 4 + 4) * m->ipv4_interfaces[i].ipv4_nr;
			}

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_I1B(&m->ipv4_interfaces_nr, &p);

			for (i = 0; i < m->ipv4_interfaces_nr; i++) {
				_InB(m->ipv4_interfaces[i].mac_address, &p, 6);
				_I1B(&m->ipv4_interfaces[i].ipv4_nr, &p);

				for (j = 0; j < m->ipv4_interfaces[i].ipv4_nr; j++) {
					_I1B(&m->ipv4_interfaces[i].ipv4[j].type, &p);
					_InB(m->ipv4_interfaces[i].ipv4[j].ipv4_address, &p, 4);
					_InB(m->ipv4_interfaces[i].ipv4[j].ipv4_dhcp_server, &p, 4);
				}
			}

			return ret;
		}

	case TLV_TYPE_IPV6:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.25"

			uint8_t *ret, *p;
			struct ipv6TypeTLV *m;

			uint16_t tlv_length;

			uint8_t i, j;

			m = (struct ipv6TypeTLV *)memory_structure;

			tlv_length = 1;	// number of entries (1 bytes)
			for (i = 0; i < m->ipv6_interfaces_nr; i++) {
				tlv_length += 6;	// interface MAC address (6 bytes)
				tlv_length += 16;	// interface ipv6 local link address (16 bytes)
				tlv_length += 1;	// number of ipv6s (1 bytes)
				tlv_length += (1 + 16 + 16) * m->ipv6_interfaces[i].ipv6_nr;
			}

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_I1B(&m->ipv6_interfaces_nr, &p);

			for (i = 0; i < m->ipv6_interfaces_nr; i++) {
				_InB(m->ipv6_interfaces[i].mac_address, &p, 6);
				_InB(m->ipv6_interfaces[i].ipv6_link_local_address, &p, 16);
				_I1B(&m->ipv6_interfaces[i].ipv6_nr, &p);

				for (j = 0; j < m->ipv6_interfaces[i].ipv6_nr; j++) {
					_I1B(&m->ipv6_interfaces[i].ipv6[j].type, &p);
					_InB(m->ipv6_interfaces[i].ipv6[j].ipv6_address, &p, 16);
					_InB(m->ipv6_interfaces[i].ipv6[j].ipv6_address_origin, &p, 16);
				}
			}

			return ret;
		}

	case TLV_TYPE_GENERIC_PHY_EVENT_NOTIFICATION:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.26"

			uint8_t *ret, *p;
			struct pushButtonGenericPhyEventNotificationTLV *m;

			uint16_t tlv_length;

			uint8_t i;

			m = (struct pushButtonGenericPhyEventNotificationTLV *)memory_structure;

			tlv_length = 1;	// number of local interfaces (1 bytes)
			for (i = 0; i < m->local_interfaces_nr; i++) {
				tlv_length += 3;	// OUI (2 bytes)
				tlv_length += 1;	// variant_index (1 byte)
				tlv_length += 1;	// media specific bytes number (1 bytes)
				tlv_length += m->local_interfaces[i].media_specific_bytes_nr;
				// media specific bytes
			}

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_I1B(&m->local_interfaces_nr, &p);

			for (i = 0; i < m->local_interfaces_nr; i++) {
				_InB(m->local_interfaces[i].oui, &p, 3);
				_I1B(&m->local_interfaces[i].variant_index, &p);
				_I1B(&m->local_interfaces[i].media_specific_bytes_nr, &p);
				if (m->local_interfaces[i].media_specific_bytes_nr > 0) {
					_InB(m->local_interfaces[i].media_specific_bytes, &p, m->local_interfaces[i].media_specific_bytes_nr);
				}
			}

			return ret;
		}

	case TLV_TYPE_1905_PROFILE_VERSION:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.27"

			uint8_t *ret, *p;
			struct x1905ProfileVersionTLV *m;

			uint16_t tlv_length;

			m = (struct x1905ProfileVersionTLV *)memory_structure;

			tlv_length = 1;

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);

			if (m->profile != PROFILE_1905_1 && m->profile != PROFILE_1905_1A) {
				// Malformed structure
				//
				FREE(ret);
				return NULL;
			}

			_I1B(&m->profile, &p);

			return ret;
		}

	case TLV_TYPE_POWER_OFF_INTERFACE:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.28"

			uint8_t *ret, *p;
			struct powerOffInterfaceTLV *m;

			uint16_t tlv_length;

			uint8_t i;

			m = (struct powerOffInterfaceTLV *)memory_structure;

			tlv_length = 1;	// number of power off interfaces (1 bytes)
			for (i = 0; i < m->power_off_interfaces_nr; i++) {
				tlv_length += 6;	// interface MAC address (6 bytes)
				tlv_length += 2;	// media type (2 bytes)
				tlv_length += 3;	// OUI (2 bytes)
				tlv_length += 1;	// variant_index (1 byte)
				tlv_length += 1;	// media specific bytes number (1 bytes)
				tlv_length += m->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes_nr;
				// media specific bytes
			}

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_I1B(&m->power_off_interfaces_nr, &p);

			for (i = 0; i < m->power_off_interfaces_nr; i++) {
				_InB(m->power_off_interfaces[i].interface_address, &p, 6);
				_I2B(&m->power_off_interfaces[i].media_type, &p);
				_InB(m->power_off_interfaces[i].generic_phy_common_data.oui, &p, 3);
				_I1B(&m->power_off_interfaces[i].generic_phy_common_data.variant_index, &p);
				_I1B(&m->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes_nr, &p);
				if (m->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes_nr > 0) {
					_InB(m->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes, &p,
					     m->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes_nr);
				}
			}

			return ret;
		}

	case TLV_TYPE_INTERFACE_POWER_CHANGE_INFORMATION:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.29"

			uint8_t *ret, *p;
			struct interfacePowerChangeInformationTLV *m;

			uint16_t tlv_length;

			uint8_t i;

			m = (struct interfacePowerChangeInformationTLV *)memory_structure;

			tlv_length = 1;	// number of interfaces (1 bytes)
			tlv_length += (6 + 1) * m->power_change_interfaces_nr;


			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_I1B(&m->power_change_interfaces_nr, &p);

			for (i = 0; i < m->power_change_interfaces_nr; i++) {
				_InB(m->power_change_interfaces[i].interface_address, &p, 6);
				_I1B(&m->power_change_interfaces[i].requested_power_state, &p);
			}

			return ret;
		}

	case TLV_TYPE_INTERFACE_POWER_CHANGE_STATUS:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.30"

			uint8_t *ret, *p;
			struct interfacePowerChangeStatusTLV *s;

			uint16_t tlv_length;

			uint8_t i;

			s = (struct interfacePowerChangeStatusTLV *)memory_structure;

			tlv_length = 1;	// number of interfaces (1 bytes)
			tlv_length += (6 + 1) * s->power_change_interfaces_nr;


			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&s->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_I1B(&s->power_change_interfaces_nr, &p);

			for (i = 0; i < s->power_change_interfaces_nr; i++) {
				_InB(s->power_change_interfaces[i].interface_address, &p, 6);
				_I1B(&s->power_change_interfaces[i].result, &p);
			}

			return ret;
		}

	case TLV_TYPE_L2_NEIGHBOR_DEVICE:
		{
			// This forging is done according to the information detailed in
			// "IEEE Std 1905.1-2013 Section 6.4.31"

			uint8_t *ret, *p;
			struct l2NeighborDeviceTLV *m;

			uint16_t tlv_length;

			uint8_t i, j, k;

			m = (struct l2NeighborDeviceTLV *)memory_structure;

			tlv_length = 1;	// number of entries (1 bytes)
			for (i = 0; i < m->local_interfaces_nr; i++) {
				tlv_length += 6;	// interface MAC address (6 bytes)
				tlv_length += 2;	// number of neighbors (2 bytes)

				for (j = 0; j < m->local_interfaces[i].l2_neighbors_nr; j++) {
					tlv_length += 6;	// neighbor MAC address (6 bytes)
					tlv_length += 2;	// number of "behind" MACs (1 bytes)
					tlv_length += 6 * m->local_interfaces[i].l2_neighbors[j].behind_mac_addresses_nr;
				}
			}

			CALLOC(ret, (1 + 2 + tlv_length));
			p = ret;

			_I1B(&m->tlv_type, &p);
			_I2B(&tlv_length, &p);
			_I1B(&m->local_interfaces_nr, &p);

			for (i = 0; i < m->local_interfaces_nr; i++) {
				_InB(m->local_interfaces[i].local_mac_address, &p, 6);
				_I2B(&m->local_interfaces[i].l2_neighbors_nr, &p);

				for (j = 0; j < m->local_interfaces[i].l2_neighbors_nr; j++) {
					_InB(m->local_interfaces[i].l2_neighbors[j].l2_neighbor_mac_address, &p, 6);
					_I2B(&m->local_interfaces[i].l2_neighbors[j].behind_mac_addresses_nr, &p);

					for (k = 0; k < m->local_interfaces[i].l2_neighbors[j].behind_mac_addresses_nr; k++) {
						_InB(m->local_interfaces[i].l2_neighbors[j].behind_mac_addresses[k], &p, 6);
					}
				}
			}

			return ret;
		}
	default:
		{
			// Ignore
			//
			return NULL;
		}

	}

	// This code cannot be reached
	//
	return NULL;
}

void free_1905_TLV_structure(uint8_t *memory_structure, bool all)
{
	if (memory_structure == NULL) {
		return;
	}
	// The first byte of any of the valid structures is always the "tlv_type"
	// field.
	//
	switch (*memory_structure) {
	case TLV_TYPE_END_OF_MESSAGE:
	case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
	case TLV_TYPE_MAC_ADDRESS_TYPE:
	case TLV_TYPE_LINK_METRIC_QUERY:
	case TLV_TYPE_LINK_METRIC_RESULT_CODE:
	case TLV_TYPE_SEARCHED_ROLE:
	case TLV_TYPE_AUTOCONFIG_FREQ_BAND:
	case TLV_TYPE_SUPPORTED_ROLE:
	case TLV_TYPE_SUPPORTED_FREQ_BAND:
	case TLV_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION:
	case TLV_TYPE_DEVICE_IDENTIFICATION:
	case TLV_TYPE_1905_PROFILE_VERSION:
		{
			if (all)
				FREE(memory_structure);

			return;
		}

	case TLV_TYPE_VENDOR_SPECIFIC:
		{
			struct vendorSpecificTLV *m;

			m = (struct vendorSpecificTLV *)memory_structure;

			if (m->m_nr > 0 && NULL != m->m) {
				FREE(m->m);
			}
			if (all)
				FREE(m);

			return;
		}

	case TLV_TYPE_DEVICE_INFORMATION_TYPE:
		{
			struct deviceInformationTypeTLV *m;

			m = (struct deviceInformationTypeTLV *)memory_structure;

			if (m->local_interfaces_nr > 0 && NULL != m->local_interfaces) {
				FREE(m->local_interfaces);
			}
			if (all)
				FREE(m);

			return;
		}

	case TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES:
		{
			struct deviceBridgingCapabilityTLV *m;
			uint8_t i;

			m = (struct deviceBridgingCapabilityTLV *)memory_structure;

			for (i = 0; i < m->bridging_tuples_nr; i++) {
				if (m->bridging_tuples[i].bridging_tuple_macs_nr > 0 && NULL != m->bridging_tuples[i].bridging_tuple_macs) {
					FREE(m->bridging_tuples[i].bridging_tuple_macs);
				}
			}
			if (m->bridging_tuples_nr > 0 && NULL != m->bridging_tuples) {
				FREE(m->bridging_tuples);
			}
			if (all)
				FREE(m);

			return;
		}

	case TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST:
		{
			struct non1905NeighborDeviceListTLV *m;

			m = (struct non1905NeighborDeviceListTLV *)memory_structure;

			if (m->non_1905_neighbors_nr > 0 && NULL != m->non_1905_neighbors) {
				FREE(m->non_1905_neighbors);
			}
			if (all)
				FREE(m);

			return;
		}

	case TLV_TYPE_NEIGHBOR_DEVICE_LIST:
		{
			struct neighborDeviceListTLV *m;

			m = (struct neighborDeviceListTLV *)memory_structure;

			if (m->neighbors_nr > 0 && NULL != m->neighbors) {
				FREE(m->neighbors);
			}
			if (all)
				FREE(m);

			return;
		}

	case TLV_TYPE_TRANSMITTER_LINK_METRIC:
		{
			struct transmitterLinkMetricTLV *m;

			m = (struct transmitterLinkMetricTLV *)memory_structure;

			if (m->transmitter_link_metrics_nr > 0 && NULL != m->transmitter_link_metrics) {
				FREE(m->transmitter_link_metrics);
			}
			if (all)
				FREE(m);

			return;
		}

	case TLV_TYPE_RECEIVER_LINK_METRIC:
		{
			struct receiverLinkMetricTLV *m;

			m = (struct receiverLinkMetricTLV *)memory_structure;

			if (m->receiver_link_metrics_nr > 0 && NULL != m->receiver_link_metrics) {
				FREE(m->receiver_link_metrics);
			}
			if (all)
				FREE(m);

			return;
		}

	case TLV_TYPE_WSC:
		{
			struct wscTLV *m;

			m = (struct wscTLV *)memory_structure;

			if (m->wsc_frame_size > 0 && NULL != m->wsc_frame) {
				FREE(m->wsc_frame);
			}
			if (all)
				FREE(m);

			return;
		}

	case TLV_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION:
		{
			struct pushButtonEventNotificationTLV *m;

			m = (struct pushButtonEventNotificationTLV *)memory_structure;

			if (m->media_types_nr > 0 && NULL != m->media_types) {
				FREE(m->media_types);
			}
			if (all)
				FREE(m);

			return;
		}

	case TLV_TYPE_GENERIC_PHY_DEVICE_INFORMATION:
		{
			struct genericPhyDeviceInformationTypeTLV *m;
			uint8_t i;

			m = (struct genericPhyDeviceInformationTypeTLV *)memory_structure;

			for (i = 0; i < m->local_interfaces_nr; i++) {
				if (m->local_interfaces[i].generic_phy_description_xml_url_len > 0
				    && NULL != m->local_interfaces[i].generic_phy_description_xml_url) {
					FREE(m->local_interfaces[i].generic_phy_description_xml_url);
				}

				if (m->local_interfaces[i].generic_phy_common_data.media_specific_bytes_nr > 0
				    && NULL != m->local_interfaces[i].generic_phy_common_data.media_specific_bytes) {
					FREE(m->local_interfaces[i].generic_phy_common_data.media_specific_bytes);
				}
			}
			if (m->local_interfaces_nr > 0 && NULL != m->local_interfaces) {
				FREE(m->local_interfaces);
			}
			if (all)
				FREE(m);

			return;
		}

	case TLV_TYPE_CONTROL_URL:
		{
			struct controlUrlTypeTLV *m;

			m = (struct controlUrlTypeTLV *)memory_structure;

			if (m->url != NULL) {
				FREE(m->url);
			}
			if (all)
				FREE(m);

			return;
		}

	case TLV_TYPE_IPV4:
		{
			struct ipv4TypeTLV *m;
			uint8_t i;

			m = (struct ipv4TypeTLV *)memory_structure;

			for (i = 0; i < m->ipv4_interfaces_nr; i++) {
				if (m->ipv4_interfaces[i].ipv4_nr > 0 && NULL != m->ipv4_interfaces[i].ipv4) {
					FREE(m->ipv4_interfaces[i].ipv4);
				}
			}
			if (m->ipv4_interfaces_nr > 0 && NULL != m->ipv4_interfaces) {
				FREE(m->ipv4_interfaces);
			}
			if (all)
				FREE(m);

			return;
		}

	case TLV_TYPE_IPV6:
		{
			struct ipv6TypeTLV *m;
			uint8_t i;

			m = (struct ipv6TypeTLV *)memory_structure;

			for (i = 0; i < m->ipv6_interfaces_nr; i++) {
				if (m->ipv6_interfaces[i].ipv6_nr > 0 && NULL != m->ipv6_interfaces[i].ipv6) {
					FREE(m->ipv6_interfaces[i].ipv6);
				}
			}
			if (m->ipv6_interfaces_nr > 0 && NULL != m->ipv6_interfaces) {
				FREE(m->ipv6_interfaces);
			}
			if (all)
				FREE(m);

			return;
		}

	case TLV_TYPE_GENERIC_PHY_EVENT_NOTIFICATION:
		{
			struct pushButtonGenericPhyEventNotificationTLV *m;

			m = (struct pushButtonGenericPhyEventNotificationTLV *)memory_structure;

			if (m->local_interfaces_nr > 0 && NULL != m->local_interfaces) {
				FREE(m->local_interfaces);
			}
			if (all)
				FREE(m);

			return;
		}

	case TLV_TYPE_POWER_OFF_INTERFACE:
		{
			struct powerOffInterfaceTLV *m;
			uint8_t i;

			m = (struct powerOffInterfaceTLV *)memory_structure;

			for (i = 0; i < m->power_off_interfaces_nr; i++) {
				if (m->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes_nr > 0
				    && NULL != m->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes) {
					FREE(m->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes);
				}
			}
			if (m->power_off_interfaces_nr > 0 && NULL != m->power_off_interfaces) {
				FREE(m->power_off_interfaces);
			}
			if (all)
				FREE(m);

			return;
		}

	case TLV_TYPE_INTERFACE_POWER_CHANGE_INFORMATION:
		{
			struct interfacePowerChangeInformationTLV *m;

			m = (struct interfacePowerChangeInformationTLV *)memory_structure;

			if (m->power_change_interfaces_nr > 0 && NULL != m->power_change_interfaces) {
				FREE(m->power_change_interfaces);
			}
			if (all)
				FREE(m);

			return;
		}

	case TLV_TYPE_INTERFACE_POWER_CHANGE_STATUS:
		{
			struct interfacePowerChangeStatusTLV *m;

			m = (struct interfacePowerChangeStatusTLV *)memory_structure;

			if (m->power_change_interfaces_nr > 0 && NULL != m->power_change_interfaces) {
				FREE(m->power_change_interfaces);
			}
			if (all)
				FREE(m);

			return;
		}

	case TLV_TYPE_L2_NEIGHBOR_DEVICE:
		{
			struct l2NeighborDeviceTLV *m;
			uint8_t i, j;

			m = (struct l2NeighborDeviceTLV *)memory_structure;

			for (i = 0; i < m->local_interfaces_nr; i++) {
				if (m->local_interfaces[i].l2_neighbors_nr > 0 && NULL != m->local_interfaces[i].l2_neighbors) {
					for (j = 0; j < m->local_interfaces[i].l2_neighbors_nr; j++) {
						if (m->local_interfaces[i].l2_neighbors[j].behind_mac_addresses_nr > 0
						    && NULL != m->local_interfaces[i].l2_neighbors[j].behind_mac_addresses) {
							FREE(m->local_interfaces[i].l2_neighbors[j].behind_mac_addresses);
						}
					}
					FREE(m->local_interfaces[i].l2_neighbors);
				}
			}
			if (m->local_interfaces_nr > 0 && NULL != m->local_interfaces) {
				FREE(m->local_interfaces);
			}
			if (all)
				FREE(m);

			return;
		}

	default:
		{
			// Ignore
			//
			return;
		}
	}

	// This code cannot be reached
	//
	return;
}

uint8_t compare_1905_TLV_structures(uint8_t *memory_structure_1, uint8_t *memory_structure_2)
{
	if (NULL == memory_structure_1 || NULL == memory_structure_2) {
		return 1;
	}
	// The first byte of any of the valid structures is always the "tlv_type"
	// field.
	//
	if (*memory_structure_1 != *memory_structure_2) {
		return 1;
	}
	switch (*memory_structure_1) {
	case TLV_TYPE_END_OF_MESSAGE:
		{
			// Nothing to compare (this TLV is always empty)
			//
			return 0;
		}

	case TLV_TYPE_VENDOR_SPECIFIC:
		{
			struct vendorSpecificTLV *p1, *p2;

			p1 = (struct vendorSpecificTLV *)memory_structure_1;
			p2 = (struct vendorSpecificTLV *)memory_structure_2;

			if (p1->vendorOUI[0] != p2->vendorOUI[0] ||
			    p1->vendorOUI[1] != p2->vendorOUI[1] ||
			    p1->vendorOUI[2] != p2->vendorOUI[2] ||
			    p1->m_nr != p2->m_nr ||
			    (memcmp(p1->m, p2->m, p1->m_nr) != 0)) {
				return 1;
			} else {
				return 0;
			}
		}

	case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
		{
			struct alMacAddressTypeTLV *p1, *p2;

			p1 = (struct alMacAddressTypeTLV *)memory_structure_1;
			p2 = (struct alMacAddressTypeTLV *)memory_structure_2;

			if ((memcmp(p1->al_mac_address, p2->al_mac_address, 6) != 0)) {
				return 1;
			} else {
				return 0;
			}
		}

	case TLV_TYPE_MAC_ADDRESS_TYPE:
		{
			struct macAddressTypeTLV *p1, *p2;

			p1 = (struct macAddressTypeTLV *)memory_structure_1;
			p2 = (struct macAddressTypeTLV *)memory_structure_2;

			if ((memcmp(p1->mac_address, p2->mac_address, 6) != 0)) {
				return 1;
			} else {
				return 0;
			}
		}

	case TLV_TYPE_DEVICE_INFORMATION_TYPE:
		{
			struct deviceInformationTypeTLV *p1, *p2;
			uint8_t i;

			p1 = (struct deviceInformationTypeTLV *)memory_structure_1;
			p2 = (struct deviceInformationTypeTLV *)memory_structure_2;

			if (memcmp(p1->al_mac_address, p2->al_mac_address, 6) != 0 ||
			    p1->local_interfaces_nr != p2->local_interfaces_nr) {
				return 1;
			}

			if (p1->local_interfaces_nr > 0 &&
			    (NULL == p1->local_interfaces ||
			     NULL == p2->local_interfaces)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->local_interfaces_nr; i++) {
				if (memcmp(p1->local_interfaces[i].mac_address, p2->local_interfaces[i].mac_address, 6) != 0 ||
				    p1->local_interfaces[i].media_type != p2->local_interfaces[i].media_type ||
				    p1->local_interfaces[i].media_specific_data_size != p2->local_interfaces[i].media_specific_data_size) {
					return 1;
				}

				if ((p1->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11B_2_4_GHZ) ||
				    (p1->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11G_2_4_GHZ) ||
				    (p1->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11A_5_GHZ) ||
				    (p1->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11N_2_4_GHZ) ||
				    (p1->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11N_5_GHZ) ||
				    (p1->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AC_5_GHZ) ||
				    (p1->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AD_60_GHZ) ||
				    (p1->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AF_GHZ)
				    ) {
					if (memcmp
					    (p1->local_interfaces[i].media_specific_data.ieee80211.network_membership,
					     p2->local_interfaces[i].media_specific_data.ieee80211.network_membership, 6) != 0
					    || p1->local_interfaces[i].media_specific_data.ieee80211.role !=
					    p2->local_interfaces[i].media_specific_data.ieee80211.role
					    || p1->local_interfaces[i].media_specific_data.ieee80211.ap_channel_band !=
					    p2->local_interfaces[i].media_specific_data.ieee80211.ap_channel_band
					    || p1->local_interfaces[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_1 !=
					    p2->local_interfaces[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_1) {
						return 1;
					}

				} else if ((p1->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_1901_WAVELET) ||
					   (p1->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_1901_FFT)
				    ) {
					if (memcmp
					    (p1->local_interfaces[i].media_specific_data.ieee1901.network_identifier,
					     p2->local_interfaces[i].media_specific_data.ieee1901.network_identifier, 6) != 0) {
						return 1;
					}
				}
			}

			return 0;
		}

	case TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES:
		{
			struct deviceBridgingCapabilityTLV *p1, *p2;
			uint8_t i, j;

			p1 = (struct deviceBridgingCapabilityTLV *)memory_structure_1;
			p2 = (struct deviceBridgingCapabilityTLV *)memory_structure_2;

			if (p1->bridging_tuples_nr != p2->bridging_tuples_nr) {
				return 1;
			}

			if (p1->bridging_tuples_nr > 0 && (NULL == p1->bridging_tuples || NULL == p2->bridging_tuples)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->bridging_tuples_nr; i++) {
				if (p1->bridging_tuples[i].bridging_tuple_macs_nr != p2->bridging_tuples[i].bridging_tuple_macs_nr) {
					return 1;
				}

				for (j = 0; j < p1->bridging_tuples[i].bridging_tuple_macs_nr; j++) {
					if (memcmp
					    (p1->bridging_tuples[i].bridging_tuple_macs[j].mac_address,
					     p2->bridging_tuples[i].bridging_tuple_macs[j].mac_address, 6) != 0) {
						return 1;
					}
				}
			}

			return 0;
		}

	case TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST:
		{
			struct non1905NeighborDeviceListTLV *p1, *p2;
			uint8_t i;

			p1 = (struct non1905NeighborDeviceListTLV *)memory_structure_1;
			p2 = (struct non1905NeighborDeviceListTLV *)memory_structure_2;

			if (memcmp(p1->local_mac_address, p2->local_mac_address, 6) != 0 ||
			    p1->non_1905_neighbors_nr != p2->non_1905_neighbors_nr) {
				return 1;
			}

			if (p1->non_1905_neighbors_nr > 0 && (NULL == p1->non_1905_neighbors || NULL == p2->non_1905_neighbors)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->non_1905_neighbors_nr; i++) {
				if (memcmp(p1->non_1905_neighbors[i].mac_address, p2->non_1905_neighbors[i].mac_address, 6) != 0) {
					return 1;
				}
			}

			return 0;
		}

	case TLV_TYPE_NEIGHBOR_DEVICE_LIST:
		{
			struct neighborDeviceListTLV *p1, *p2;
			uint8_t i;

			p1 = (struct neighborDeviceListTLV *)memory_structure_1;
			p2 = (struct neighborDeviceListTLV *)memory_structure_2;

			if (memcmp(p1->local_mac_address, p2->local_mac_address, 6) != 0 ||
			    p1->neighbors_nr != p2->neighbors_nr) {
				return 1;
			}

			if (p1->neighbors_nr > 0 && (NULL == p1->neighbors || NULL == p2->neighbors)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->neighbors_nr; i++) {
				if (memcmp(p1->neighbors[i].mac_address, p2->neighbors[i].mac_address, 6) != 0 ||
				    p1->neighbors[i].bridge_flag != p2->neighbors[i].bridge_flag) {
					return 1;
				}
			}

			return 0;
		}

	case TLV_TYPE_LINK_METRIC_QUERY:
		{
			struct linkMetricQueryTLV *p1, *p2;

			p1 = (struct linkMetricQueryTLV *)memory_structure_1;
			p2 = (struct linkMetricQueryTLV *)memory_structure_2;

			if (p1->destination != p2->destination ||
			    memcmp(p1->specific_neighbor, p2->specific_neighbor, 6) != 0 ||
			    p1->link_metrics_type != p2->link_metrics_type) {
				return 1;
			} else {
				return 0;
			}
		}

	case TLV_TYPE_TRANSMITTER_LINK_METRIC:
		{
			struct transmitterLinkMetricTLV *p1, *p2;
			uint8_t i;

			p1 = (struct transmitterLinkMetricTLV *)memory_structure_1;
			p2 = (struct transmitterLinkMetricTLV *)memory_structure_2;

			if (memcmp(p1->local_al_address, p2->local_al_address, 6) != 0 ||
			    memcmp(p1->neighbor_al_address, p2->neighbor_al_address, 6) != 0 ||
			    p1->transmitter_link_metrics_nr != p2->transmitter_link_metrics_nr) {
				return 1;
			}

			if (p1->transmitter_link_metrics_nr > 0 && (NULL == p1->transmitter_link_metrics || NULL == p2->transmitter_link_metrics)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->transmitter_link_metrics_nr; i++) {
				if (memcmp(p1->transmitter_link_metrics[i].local_interface_address,
					   p2->transmitter_link_metrics[i].local_interface_address, 6) != 0
				    || memcmp(p1->transmitter_link_metrics[i].neighbor_interface_address,
						   p2->transmitter_link_metrics[i].neighbor_interface_address, 6) != 0
				    || p1->transmitter_link_metrics[i].intf_type != p2->transmitter_link_metrics[i].intf_type
				    || p1->transmitter_link_metrics[i].bridge_flag != p2->transmitter_link_metrics[i].bridge_flag
				    || p1->transmitter_link_metrics[i].packet_errors != p2->transmitter_link_metrics[i].packet_errors
				    || p1->transmitter_link_metrics[i].transmitted_packets != p2->transmitter_link_metrics[i].transmitted_packets
				    || p1->transmitter_link_metrics[i].mac_throughput_capacity != p2->transmitter_link_metrics[i].mac_throughput_capacity
				    || p1->transmitter_link_metrics[i].link_availability != p2->transmitter_link_metrics[i].link_availability
				    || p1->transmitter_link_metrics[i].phy_rate != p2->transmitter_link_metrics[i].phy_rate) {
					return 1;
				}
			}

			return 0;
		}

	case TLV_TYPE_RECEIVER_LINK_METRIC:
		{
			struct receiverLinkMetricTLV *p1, *p2;
			uint8_t i;

			p1 = (struct receiverLinkMetricTLV *)memory_structure_1;
			p2 = (struct receiverLinkMetricTLV *)memory_structure_2;

			if (memcmp(p1->local_al_address, p2->local_al_address, 6) != 0 ||
			    memcmp(p1->neighbor_al_address, p2->neighbor_al_address, 6) != 0 ||
			    p1->receiver_link_metrics_nr != p2->receiver_link_metrics_nr) {
				return 1;
			}

			if (p1->receiver_link_metrics_nr > 0 && (NULL == p1->receiver_link_metrics || NULL == p2->receiver_link_metrics)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->receiver_link_metrics_nr; i++) {
				if (memcmp(p1->receiver_link_metrics[i].local_interface_address,
					   p2->receiver_link_metrics[i].local_interface_address, 6) != 0
				    || memcmp(p1->receiver_link_metrics[i].neighbor_interface_address,
					   p2->receiver_link_metrics[i].neighbor_interface_address, 6) != 0
				    || p1->receiver_link_metrics[i].intf_type != p2->receiver_link_metrics[i].intf_type
				    || p1->receiver_link_metrics[i].packet_errors != p2->receiver_link_metrics[i].packet_errors
				    || p1->receiver_link_metrics[i].packets_received != p2->receiver_link_metrics[i].packets_received
				    || p1->receiver_link_metrics[i].rssi != p2->receiver_link_metrics[i].rssi) {
					return 1;
				}
			}

			return 0;
		}

	case TLV_TYPE_LINK_METRIC_RESULT_CODE:
		{
			struct linkMetricResultCodeTLV *p1, *p2;

			p1 = (struct linkMetricResultCodeTLV *)memory_structure_1;
			p2 = (struct linkMetricResultCodeTLV *)memory_structure_2;

			if (p1->result_code != p2->result_code) {
				return 1;
			} else {
				return 0;
			}
		}

	case TLV_TYPE_SEARCHED_ROLE:
		{
			struct searchedRoleTLV *p1, *p2;

			p1 = (struct searchedRoleTLV *)memory_structure_1;
			p2 = (struct searchedRoleTLV *)memory_structure_2;

			if (p1->role != p2->role) {
				return 1;
			} else {
				return 0;
			}
		}

	case TLV_TYPE_AUTOCONFIG_FREQ_BAND:
		{
			struct autoconfigFreqBandTLV *p1, *p2;

			p1 = (struct autoconfigFreqBandTLV *)memory_structure_1;
			p2 = (struct autoconfigFreqBandTLV *)memory_structure_2;

			if (p1->freq_band != p2->freq_band) {
				return 1;
			} else {
				return 0;
			}
		}

	case TLV_TYPE_SUPPORTED_ROLE:
		{
			struct supportedRoleTLV *p1, *p2;

			p1 = (struct supportedRoleTLV *)memory_structure_1;
			p2 = (struct supportedRoleTLV *)memory_structure_2;

			if (p1->role != p2->role) {
				return 1;
			} else {
				return 0;
			}
		}

	case TLV_TYPE_SUPPORTED_FREQ_BAND:
		{
			struct supportedFreqBandTLV *p1, *p2;

			p1 = (struct supportedFreqBandTLV *)memory_structure_1;
			p2 = (struct supportedFreqBandTLV *)memory_structure_2;

			if (p1->freq_band != p2->freq_band) {
				return 1;
			} else {
				return 0;
			}
		}

	case TLV_TYPE_WSC:
		{
			struct wscTLV *p1, *p2;

			p1 = (struct wscTLV *)memory_structure_1;
			p2 = (struct wscTLV *)memory_structure_2;

			if (p1->wsc_frame_size != p2->wsc_frame_size
			    || memcmp(p1->wsc_frame, p2->wsc_frame, p1->wsc_frame_size) != 0) {
				return 1;
			}

			return 0;
		}

	case TLV_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION:
		{
			struct pushButtonEventNotificationTLV *p1, *p2;
			uint8_t i;

			p1 = (struct pushButtonEventNotificationTLV *)memory_structure_1;
			p2 = (struct pushButtonEventNotificationTLV *)memory_structure_2;

			if (p1->media_types_nr != p2->media_types_nr) {
				return 1;
			}

			if (p1->media_types_nr > 0 && (NULL == p1->media_types || NULL == p2->media_types)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->media_types_nr; i++) {
				if (p1->media_types[i].media_type != p2->media_types[i].media_type ||
				    p1->media_types[i].media_specific_data_size != p2->media_types[i].media_specific_data_size) {
					return 1;
				}

				if ((p1->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11B_2_4_GHZ) ||
				    (p1->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11G_2_4_GHZ) ||
				    (p1->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11A_5_GHZ) ||
				    (p1->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11N_2_4_GHZ) ||
				    (p1->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11N_5_GHZ) ||
				    (p1->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11AC_5_GHZ) ||
				    (p1->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11AD_60_GHZ) ||
				    (p1->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11AF_GHZ)
				    ) {
					if (memcmp(p1->media_types[i].media_specific_data.ieee80211.network_membership,
					     p2->media_types[i].media_specific_data.ieee80211.network_membership, 6) != 0
					    || p1->media_types[i].media_specific_data.ieee80211.role !=
					    p2->media_types[i].media_specific_data.ieee80211.role
					    || p1->media_types[i].media_specific_data.ieee80211.ap_channel_band !=
					    p2->media_types[i].media_specific_data.ieee80211.ap_channel_band
					    || p1->media_types[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_1 !=
					    p2->media_types[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_1) {
						return 1;
					}

				} else if ((p1->media_types[i].media_type == MEDIA_TYPE_IEEE_1901_WAVELET) ||
					   (p1->media_types[i].media_type == MEDIA_TYPE_IEEE_1901_FFT)
				    ) {
					if (memcmp(p1->media_types[i].media_specific_data.ieee1901.network_identifier,
					     p2->media_types[i].media_specific_data.ieee1901.network_identifier, 6) != 0) {
						return 1;
					}
				}
			}

			return 0;
		}

	case TLV_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION:
		{
			struct pushButtonJoinNotificationTLV *p1, *p2;

			p1 = (struct pushButtonJoinNotificationTLV *)memory_structure_1;
			p2 = (struct pushButtonJoinNotificationTLV *)memory_structure_2;

			if (memcmp(p1->al_mac_address, p2->al_mac_address, 6) != 0
			    || p1->message_identifier != p2->message_identifier
			    || memcmp(p1->mac_address, p2->al_mac_address, 6) != 0
			    || memcmp(p1->new_mac_address, p2->al_mac_address, 6) != 0) {
				return 1;
			} else {
				return 0;
			}
		}

	case TLV_TYPE_DEVICE_IDENTIFICATION:
		{
			struct deviceIdentificationTypeTLV *p1, *p2;

			p1 = (struct deviceIdentificationTypeTLV *)memory_structure_1;
			p2 = (struct deviceIdentificationTypeTLV *)memory_structure_2;

			if (memcmp(p1->friendly_name, p2->friendly_name, 64) != 0 ||
			    memcmp(p1->manufacturer_name, p2->manufacturer_name, 64) != 0 ||
			    memcmp(p1->manufacturer_model, p2->manufacturer_model, 64) != 0) {
				return 1;
			} else {
				return 0;
			}
		}

	case TLV_TYPE_CONTROL_URL:
		{
			struct controlUrlTypeTLV *p1, *p2;

			p1 = (struct controlUrlTypeTLV *)memory_structure_1;
			p2 = (struct controlUrlTypeTLV *)memory_structure_2;

			if (memcmp(p1->url, p2->url, strlen(p1->url) + 1) != 0) {
				return 1;
			} else {
				return 0;
			}
		}

	case TLV_TYPE_IPV4:
		{
			struct ipv4TypeTLV *p1, *p2;
			uint8_t i, j;

			p1 = (struct ipv4TypeTLV *)memory_structure_1;
			p2 = (struct ipv4TypeTLV *)memory_structure_2;

			if (p1->ipv4_interfaces_nr != p2->ipv4_interfaces_nr) {
				return 1;
			}

			if (p1->ipv4_interfaces_nr > 0 && (NULL == p1->ipv4_interfaces || NULL == p2->ipv4_interfaces)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->ipv4_interfaces_nr; i++) {
				if (memcmp(p1->ipv4_interfaces[i].mac_address,
					   p2->ipv4_interfaces[i].mac_address, 6) != 0
				    || p1->ipv4_interfaces[i].ipv4_nr != p2->ipv4_interfaces[i].ipv4_nr) {
					return 1;
				}

				for (j = 0; j < p1->ipv4_interfaces[i].ipv4_nr; j++) {
					if (p1->ipv4_interfaces[i].ipv4[j].type != p2->ipv4_interfaces[i].ipv4[j].type ||
					    memcmp(p1->ipv4_interfaces[i].ipv4[j].ipv4_address, p2->ipv4_interfaces[i].ipv4[j].ipv4_address, 4) != 0 ||
					    memcmp(p1->ipv4_interfaces[i].ipv4[j].ipv4_dhcp_server, p2->ipv4_interfaces[i].ipv4[j].ipv4_dhcp_server,
							4) != 0) {
						return 1;
					}
				}
			}

			return 0;
		}

	case TLV_TYPE_IPV6:
		{
			struct ipv6TypeTLV *p1, *p2;
			uint8_t i, j;

			p1 = (struct ipv6TypeTLV *)memory_structure_1;
			p2 = (struct ipv6TypeTLV *)memory_structure_2;

			if (p1->ipv6_interfaces_nr != p2->ipv6_interfaces_nr) {
				return 1;
			}

			if (p1->ipv6_interfaces_nr > 0 && (NULL == p1->ipv6_interfaces || NULL == p2->ipv6_interfaces)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->ipv6_interfaces_nr; i++) {
				if (memcmp(p1->ipv6_interfaces[i].mac_address, p2->ipv6_interfaces[i].mac_address, 6) != 0 ||
				    p1->ipv6_interfaces[i].ipv6_nr != p2->ipv6_interfaces[i].ipv6_nr) {
					return 1;
				}

				for (j = 0; j < p1->ipv6_interfaces[i].ipv6_nr; j++) {
					if (p1->ipv6_interfaces[i].ipv6[j].type != p2->ipv6_interfaces[i].ipv6[j].type ||
					    memcmp(p1->ipv6_interfaces[i].ipv6[j].ipv6_address, p2->ipv6_interfaces[i].ipv6[j].ipv6_address, 16) != 0 ||
					    memcmp(p1->ipv6_interfaces[i].ipv6[j].ipv6_address_origin, p2->ipv6_interfaces[i].ipv6[j].ipv6_address_origin,
							16) != 0) {
						return 1;
					}
				}
			}

			return 0;
		}

	case TLV_TYPE_GENERIC_PHY_DEVICE_INFORMATION:
		{
			struct genericPhyDeviceInformationTypeTLV *p1, *p2;
			uint8_t i;

			p1 = (struct genericPhyDeviceInformationTypeTLV *)memory_structure_1;
			p2 = (struct genericPhyDeviceInformationTypeTLV *)memory_structure_2;

			if (memcmp(p1->al_mac_address, p2->al_mac_address, 6) != 0 || p1->local_interfaces_nr != p2->local_interfaces_nr) {
				return 1;
			}

			if (p1->local_interfaces_nr > 0 && (NULL == p1->local_interfaces || NULL == p2->local_interfaces)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->local_interfaces_nr; i++) {
				if (memcmp(p1->local_interfaces[i].local_interface_address, p2->local_interfaces[i].local_interface_address, 6) != 0 ||
				    memcmp(p1->local_interfaces[i].generic_phy_common_data.oui, p2->local_interfaces[i].generic_phy_common_data.oui,
						3) != 0
				    || p1->local_interfaces[i].generic_phy_common_data.variant_index !=
				    p2->local_interfaces[i].generic_phy_common_data.variant_index
				    || memcmp(p1->local_interfaces[i].variant_name, p2->local_interfaces[i].variant_name, 32) != 0
				    || p1->local_interfaces[i].generic_phy_description_xml_url_len !=
				    p2->local_interfaces[i].generic_phy_description_xml_url_len
				    || p1->local_interfaces[i].generic_phy_common_data.media_specific_bytes_nr !=
				    p2->local_interfaces[i].generic_phy_common_data.media_specific_bytes_nr
				    || memcmp(p1->local_interfaces[i].generic_phy_description_xml_url,
						   p2->local_interfaces[i].generic_phy_description_xml_url,
						   p1->local_interfaces[i].generic_phy_description_xml_url_len) != 0
				    || memcmp(p1->local_interfaces[i].generic_phy_common_data.media_specific_bytes,
						   p2->local_interfaces[i].generic_phy_common_data.media_specific_bytes,
						   p1->local_interfaces[i].generic_phy_common_data.media_specific_bytes_nr) != 0) {
					return 1;
				}
			}

			return 0;
		}

	case TLV_TYPE_GENERIC_PHY_EVENT_NOTIFICATION:
		{
			struct pushButtonGenericPhyEventNotificationTLV *p1, *p2;
			uint8_t i;

			p1 = (struct pushButtonGenericPhyEventNotificationTLV *)memory_structure_1;
			p2 = (struct pushButtonGenericPhyEventNotificationTLV *)memory_structure_2;

			if (p1->local_interfaces_nr != p2->local_interfaces_nr) {
				return 1;
			}

			if (p1->local_interfaces_nr > 0 && (NULL == p1->local_interfaces || NULL == p2->local_interfaces)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->local_interfaces_nr; i++) {
				if (memcmp(p1->local_interfaces[i].oui, p2->local_interfaces[i].oui, 3) != 0 ||
				    p1->local_interfaces[i].variant_index != p2->local_interfaces[i].variant_index ||
				    p1->local_interfaces[i].media_specific_bytes_nr != p2->local_interfaces[i].media_specific_bytes_nr ||
				    memcmp(p1->local_interfaces[i].media_specific_bytes, p2->local_interfaces[i].media_specific_bytes,
						p1->local_interfaces[i].media_specific_bytes_nr) != 0) {
					return 1;
				}
			}

			return 0;
		}

	case TLV_TYPE_1905_PROFILE_VERSION:
		{
			struct x1905ProfileVersionTLV *p1, *p2;

			p1 = (struct x1905ProfileVersionTLV *)memory_structure_1;
			p2 = (struct x1905ProfileVersionTLV *)memory_structure_2;

			if (p1->profile != p2->profile) {
				return 1;
			} else {
				return 0;
			}
		}

	case TLV_TYPE_POWER_OFF_INTERFACE:
		{
			struct powerOffInterfaceTLV *p1, *p2;
			uint8_t i;

			p1 = (struct powerOffInterfaceTLV *)memory_structure_1;
			p2 = (struct powerOffInterfaceTLV *)memory_structure_2;

			if (p1->power_off_interfaces_nr != p2->power_off_interfaces_nr) {
				return 1;
			}

			if (p1->power_off_interfaces_nr > 0 && (NULL == p1->power_off_interfaces || NULL == p2->power_off_interfaces)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->power_off_interfaces_nr; i++) {
				if (memcmp(p1->power_off_interfaces[i].interface_address, p2->power_off_interfaces[i].interface_address, 6) != 0 ||
				    p1->power_off_interfaces[i].media_type != p2->power_off_interfaces[i].media_type ||
				    memcmp(p1->power_off_interfaces[i].generic_phy_common_data.oui,
						p2->power_off_interfaces[i].generic_phy_common_data.oui, 3) != 0
				    || p1->power_off_interfaces[i].generic_phy_common_data.variant_index !=
				    p2->power_off_interfaces[i].generic_phy_common_data.variant_index
				    || p1->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes_nr !=
				    p2->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes_nr
				    || memcmp(p1->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes,
						   p2->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes,
						   p1->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes_nr) != 0) {
					return 1;
				}
			}

			return 0;
		}

	case TLV_TYPE_INTERFACE_POWER_CHANGE_INFORMATION:
		{
			struct interfacePowerChangeInformationTLV *p1, *p2;
			uint8_t i;

			p1 = (struct interfacePowerChangeInformationTLV *)memory_structure_1;
			p2 = (struct interfacePowerChangeInformationTLV *)memory_structure_2;

			if (p1->power_change_interfaces_nr != p2->power_change_interfaces_nr) {
				return 1;
			}

			if (p1->power_change_interfaces_nr > 0 && (NULL == p1->power_change_interfaces || NULL == p2->power_change_interfaces)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->power_change_interfaces_nr; i++) {
				if (memcmp(p1->power_change_interfaces[i].interface_address, p2->power_change_interfaces[i].interface_address, 6) != 0 ||
				    p1->power_change_interfaces[i].requested_power_state != p2->power_change_interfaces[i].requested_power_state) {
					return 1;
				}
			}

			return 0;
		}

	case TLV_TYPE_INTERFACE_POWER_CHANGE_STATUS:
		{
			struct interfacePowerChangeStatusTLV *p1, *p2;
			uint8_t i;

			p1 = (struct interfacePowerChangeStatusTLV *)memory_structure_1;
			p2 = (struct interfacePowerChangeStatusTLV *)memory_structure_2;

			if (p1->power_change_interfaces_nr != p2->power_change_interfaces_nr) {
				return 1;
			}

			if (p1->power_change_interfaces_nr > 0 && (NULL == p1->power_change_interfaces || NULL == p2->power_change_interfaces)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->power_change_interfaces_nr; i++) {
				if (memcmp(p1->power_change_interfaces[i].interface_address, p2->power_change_interfaces[i].interface_address, 6) != 0 ||
				    p1->power_change_interfaces[i].result != p2->power_change_interfaces[i].result) {
					return 1;
				}
			}

			return 0;
		}

	case TLV_TYPE_L2_NEIGHBOR_DEVICE:
		{
			struct l2NeighborDeviceTLV *p1, *p2;
			uint8_t i, j, k;

			p1 = (struct l2NeighborDeviceTLV *)memory_structure_1;
			p2 = (struct l2NeighborDeviceTLV *)memory_structure_2;

			if (p1->local_interfaces_nr != p2->local_interfaces_nr) {
				return 1;
			}

			if (p1->local_interfaces_nr > 0 && (NULL == p1->local_interfaces || NULL == p2->local_interfaces)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->local_interfaces_nr; i++) {
				if (memcmp(p1->local_interfaces[i].local_mac_address, p2->local_interfaces[i].local_mac_address, 6) != 0 ||
				    p1->local_interfaces[i].l2_neighbors_nr != p2->local_interfaces[i].l2_neighbors_nr) {
					return 1;
				}

				if (p1->local_interfaces[i].l2_neighbors_nr > 0
				    && (NULL == p1->local_interfaces[i].l2_neighbors || NULL == p2->local_interfaces[i].l2_neighbors)) {
					// Malformed structure
					//
					return 1;
				}

				for (j = 0; j < p1->local_interfaces[i].l2_neighbors_nr; j++) {
					if (memcmp
					    (p1->local_interfaces[i].l2_neighbors[j].l2_neighbor_mac_address,
					     p2->local_interfaces[i].l2_neighbors[j].l2_neighbor_mac_address, 6) != 0
					    || p1->local_interfaces[i].l2_neighbors[j].behind_mac_addresses_nr !=
					    p2->local_interfaces[i].l2_neighbors[j].behind_mac_addresses_nr) {
						return 1;
					}

					if (p1->local_interfaces[i].l2_neighbors[j].behind_mac_addresses_nr > 0
					    && (NULL == p1->local_interfaces[i].l2_neighbors[j].behind_mac_addresses
						|| NULL == p2->local_interfaces[i].l2_neighbors[j].behind_mac_addresses)) {
						// Malformed structure
						//
						return 1;
					}

					for (k = 0; k < p1->local_interfaces[i].l2_neighbors[j].behind_mac_addresses_nr; k++) {
						if (memcmp
						    (p1->local_interfaces[i].l2_neighbors[j].behind_mac_addresses[k],
						     p2->local_interfaces[i].l2_neighbors[j].behind_mac_addresses[k], 6) != 0) {
							return 1;
						}
					}
				}
			}

			return 0;
		}

	default:
		{
			// Unknown structure type
			//
			return 1;
		}
	}

	// This code cannot be reached
	//
	return 1;
}

char *convert_1905_TLV_type_to_string(uint8_t tlv_type)
{
	switch (tlv_type) {
	case TLV_TYPE_END_OF_MESSAGE:
		return "TLV_TYPE_END_OF_MESSAGE";
	case TLV_TYPE_VENDOR_SPECIFIC:
		return "TLV_TYPE_VENDOR_SPECIFIC";
	case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
		return "TLV_TYPE_AL_MAC_ADDRESS_TYPE";
	case TLV_TYPE_MAC_ADDRESS_TYPE:
		return "TLV_TYPE_MAC_ADDRESS_TYPE";
	case TLV_TYPE_DEVICE_INFORMATION_TYPE:
		return "TLV_TYPE_DEVICE_INFORMATION_TYPE";
	case TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES:
		return "TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES";
	case TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST:
		return "TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST";
	case TLV_TYPE_NEIGHBOR_DEVICE_LIST:
		return "TLV_TYPE_NEIGHBOR_DEVICE_LIST";
	case TLV_TYPE_LINK_METRIC_QUERY:
		return "TLV_TYPE_LINK_METRIC_QUERY";
	case TLV_TYPE_TRANSMITTER_LINK_METRIC:
		return "TLV_TYPE_TRANSMITTER_LINK_METRIC";
	case TLV_TYPE_RECEIVER_LINK_METRIC:
		return "TLV_TYPE_RECEIVER_LINK_METRIC";
	case TLV_TYPE_LINK_METRIC_RESULT_CODE:
		return "TLV_TYPE_LINK_METRIC_RESULT_CODE";
	case TLV_TYPE_SEARCHED_ROLE:
		return "TLV_TYPE_SEARCHED_ROLE";
	case TLV_TYPE_AUTOCONFIG_FREQ_BAND:
		return "TLV_TYPE_AUTOCONFIG_FREQ_BAND";
	case TLV_TYPE_SUPPORTED_ROLE:
		return "TLV_TYPE_SUPPORTED_ROLE";
	case TLV_TYPE_SUPPORTED_FREQ_BAND:
		return "TLV_TYPE_SUPPORTED_FREQ_BAND";
	case TLV_TYPE_WSC:
		return "TLV_TYPE_WSC";
	case TLV_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION:
		return "TLV_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION";
	case TLV_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION:
		return "TLV_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION";
	case TLV_TYPE_GENERIC_PHY_DEVICE_INFORMATION:
		return "TLV_TYPE_GENERIC_PHY_DEVICE_INFORMATION";
	case TLV_TYPE_DEVICE_IDENTIFICATION:
		return "TLV_TYPE_DEVICE_IDENTIFICATION";
	case TLV_TYPE_CONTROL_URL:
		return "TLV_TYPE_CONTROL_URL";
	case TLV_TYPE_IPV4:
		return "TLV_TYPE_IPV4";
	case TLV_TYPE_IPV6:
		return "TLV_TYPE_IPV6";
	case TLV_TYPE_GENERIC_PHY_EVENT_NOTIFICATION:
		return "TLV_TYPE_GENERIC_PHY_EVENT_NOTIFICATION";
	case TLV_TYPE_1905_PROFILE_VERSION:
		return "TLV_TYPE_1905_PROFILE_VERSION";
	case TLV_TYPE_POWER_OFF_INTERFACE:
		return "TLV_TYPE_POWER_OFF_INTERFACE";
	case TLV_TYPE_INTERFACE_POWER_CHANGE_INFORMATION:
		return "TLV_TYPE_INTERFACE_POWER_CHANGE_INFORMATION";
	case TLV_TYPE_INTERFACE_POWER_CHANGE_STATUS:
		return "TLV_TYPE_INTERFACE_POWER_CHANGE_STATUS";
	case TLV_TYPE_L2_NEIGHBOR_DEVICE:
		return "TLV_TYPE_L2_NEIGHBOR_DEVICE";
	default:
		return "Unknown";
	}
	return "";
}

char *convert_band_to_string(uint8_t freq_band)
{
	switch (freq_band) {
	case IEEE80211_FREQUENCY_BAND_2_4_GHZ:
		return "2_4GHZ";
	case IEEE80211_FREQUENCY_BAND_5_GHZ:
		return "5GHZ";
	case IEEE80211_FREQUENCY_BAND_60_GHZ:
		return "60GHZ";
	}
	return NULL;
}

bool get_freq_band_from_media_type(uint16_t media_type, uint8_t *freq_band)
{
	bool ret = false;

	if (freq_band == NULL)
		return false;

	switch (media_type) {
	case MEDIA_TYPE_IEEE_802_11B_2_4_GHZ:
	case MEDIA_TYPE_IEEE_802_11G_2_4_GHZ:
	case MEDIA_TYPE_IEEE_802_11N_2_4_GHZ:
		*freq_band = IEEE80211_FREQUENCY_BAND_2_4_GHZ;
		ret = true;
		break;
	case MEDIA_TYPE_IEEE_802_11A_5_GHZ:
	case MEDIA_TYPE_IEEE_802_11N_5_GHZ:
	case MEDIA_TYPE_IEEE_802_11AC_5_GHZ:
		*freq_band = IEEE80211_FREQUENCY_BAND_5_GHZ;
		ret = true;
		break;
	case MEDIA_TYPE_IEEE_802_11AD_60_GHZ:
		*freq_band = IEEE80211_FREQUENCY_BAND_60_GHZ;
		ret = true;
		break;
	}
	return ret;
}
