/*
 * al_wsc.c: IEEE-1905 AL wsc handling
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <uci.h>
#include <libubus.h>

#include "utils.h"
#include "wsc.h"
#include "platform_crypto.h"

////////////////////////////////////////////////////////////////////////////////
// Private functions and data
////////////////////////////////////////////////////////////////////////////////

// "defines" used to fill the M1 and M2 message fields
//
#define ATTR_VERSION           (0x104a)
#define ATTR_MSG_TYPE          (0x1022)
#define WPS_M1                 (0x04)
#define WPS_M2                 (0x05)
#define ATTR_UUID_E            (0x1047)
#define ATTR_UUID_R            (0x1048)
#define ATTR_MAC_ADDR          (0x1020)
#define ATTR_ENROLLEE_NONCE    (0x101a)
#define ATTR_REGISTRAR_NONCE   (0x1039)
#define ATTR_PUBLIC_KEY        (0x1032)
#define ATTR_AUTH_TYPE_FLAGS   (0x1004)
#define ATTR_ENCR_TYPE_FLAGS   (0x1010)
#define ATTR_CONN_TYPE_FLAGS   (0x100d)
#define WPS_CONN_ESS           (0x01)
#define WPS_CONN_IBSS          (0x02)
#define ATTR_CONFIG_METHODS    (0x1008)
#define WPS_CONFIG_VIRT_PUSHBUTTON (0x0280)
#define WPS_CONFIG_PHY_PUSHBUTTON  (0x0480)
#define ATTR_WPS_STATE         (0x1044)
#define WPS_STATE_NOT_CONFIGURED (1)
#define WPS_STATE_CONFIGURED     (2)
#define ATTR_MANUFACTURER      (0x1021)
#define ATTR_MODEL_NAME        (0x1023)
#define ATTR_MODEL_NUMBER      (0x1024)
#define ATTR_SERIAL_NUMBER     (0x1042)
#define ATTR_PRIMARY_DEV_TYPE  (0x1054)
#define WPS_DEV_COMPUTER                           (1)
#define WPS_DEV_COMPUTER_PC                       (1)
#define WPS_DEV_COMPUTER_SERVER                   (2)
#define WPS_DEV_COMPUTER_MEDIA_CENTER             (3)
#define WPS_DEV_COMPUTER_ULTRA_MOBILE             (4)
#define WPS_DEV_COMPUTER_NOTEBOOK                 (5)
#define WPS_DEV_COMPUTER_DESKTOP                  (6)
#define WPS_DEV_COMPUTER_MID                      (7)
#define WPS_DEV_COMPUTER_NETBOOK                  (8)
#define WPS_DEV_COMPUTER_TABLET                   (9)
#define WPS_DEV_INPUT                              (2)
#define WPS_DEV_INPUT_KEYBOARD                    (1)
#define WPS_DEV_INPUT_MOUSE                       (2)
#define WPS_DEV_INPUT_JOYSTICK                    (3)
#define WPS_DEV_INPUT_TRACKBALL                   (4)
#define WPS_DEV_INPUT_GAMING                      (5)
#define WPS_DEV_INPUT_REMOTE                      (6)
#define WPS_DEV_INPUT_TOUCHSCREEN                 (7)
#define WPS_DEV_INPUT_BIOMETRIC_READER            (8)
#define WPS_DEV_INPUT_BARCODE_READER              (9)
#define WPS_DEV_PRINTER                            (3)
#define WPS_DEV_PRINTER_PRINTER                   (1)
#define WPS_DEV_PRINTER_SCANNER                   (2)
#define WPS_DEV_PRINTER_FAX                       (3)
#define WPS_DEV_PRINTER_COPIER                    (4)
#define WPS_DEV_PRINTER_ALL_IN_ONE                (5)
#define WPS_DEV_CAMERA                             (4)
#define WPS_DEV_CAMERA_DIGITAL_STILL_CAMERA       (1)
#define WPS_DEV_CAMERA_VIDEO                      (2)
#define WPS_DEV_CAMERA_WEB                        (3)
#define WPS_DEV_CAMERA_SECURITY                   (4)
#define WPS_DEV_STORAGE                            (5)
#define WPS_DEV_STORAGE_NAS                       (1)
#define WPS_DEV_NETWORK_INFRA                      (6)
#define WPS_DEV_NETWORK_INFRA_AP                  (1)
#define WPS_DEV_NETWORK_INFRA_ROUTER              (2)
#define WPS_DEV_NETWORK_INFRA_SWITCH              (3)
#define WPS_DEV_NETWORK_INFRA_GATEWAY             (4)
#define WPS_DEV_NETWORK_INFRA_BRIDGE              (5)
#define WPS_DEV_DISPLAY                            (7)
#define WPS_DEV_DISPLAY_TV                        (1)
#define WPS_DEV_DISPLAY_PICTURE_FRAME             (2)
#define WPS_DEV_DISPLAY_PROJECTOR                 (3)
#define WPS_DEV_DISPLAY_MONITOR                   (4)
#define WPS_DEV_MULTIMEDIA                         (8)
#define WPS_DEV_MULTIMEDIA_DAR                    (1)
#define WPS_DEV_MULTIMEDIA_PVR                    (2)
#define WPS_DEV_MULTIMEDIA_MCX                    (3)
#define WPS_DEV_MULTIMEDIA_SET_TOP_BOX            (4)
#define WPS_DEV_MULTIMEDIA_MEDIA_SERVER           (5)
#define WPS_DEV_MULTIMEDIA_PORTABLE_VIDEO_PLAYER  (6)
#define WPS_DEV_GAMING                             (9)
#define WPS_DEV_GAMING_XBOX                       (1)
#define WPS_DEV_GAMING_XBOX360                    (2)
#define WPS_DEV_GAMING_PLAYSTATION                (3)
#define WPS_DEV_GAMING_GAME_CONSOLE               (4)
#define WPS_DEV_GAMING_PORTABLE_DEVICE            (5)
#define WPS_DEV_PHONE                             (10)
#define WPS_DEV_PHONE_WINDOWS_MOBILE              (1)
#define WPS_DEV_PHONE_SINGLE_MODE                 (2)
#define WPS_DEV_PHONE_DUAL_MODE                   (3)
#define WPS_DEV_PHONE_SP_SINGLE_MODE              (4)
#define WPS_DEV_PHONE_SP_DUAL_MODE                (5)
#define WPS_DEV_AUDIO                             (11)
#define WPS_DEV_AUDIO_TUNER_RECV                  (1)
#define WPS_DEV_AUDIO_SPEAKERS                    (2)
#define WPS_DEV_AUDIO_PMP                         (3)
#define WPS_DEV_AUDIO_HEADSET                     (4)
#define WPS_DEV_AUDIO_HEADPHONES                  (5)
#define WPS_DEV_AUDIO_MICROPHONE                  (6)
#define WPS_DEV_AUDIO_HOME_THEATRE                (7)
#define ATTR_DEV_NAME          (0x1011)
#define ATTR_RF_BANDS          (0x103c)
#define WPS_RF_24GHZ           (0x01)
#define WPS_RF_50GHZ           (0x02)
#define WPS_RF_60GHZ           (0x04)
#define ATTR_ASSOC_STATE       (0x1002)
#define WPS_ASSOC_NOT_ASSOC     (0)
#define WPS_ASSOC_CONN_SUCCESS  (1)
#define ATTR_DEV_PASSWORD_ID   (0x1012)
#define DEV_PW_PUSHBUTTON      (0x0004)
#define ATTR_CONFIG_ERROR      (0x1009)
#define WPS_CFG_NO_ERROR       (0)
#define ATTR_OS_VERSION        (0x102d)
#define ATTR_VENDOR_EXTENSION  (0x1049)
#define WFA_VENDOR_ID_1 	   (0x00)
#define WFA_VENDOR_ID_2 	   (0x37)
#define WFA_VENDOR_ID_3 	   (0x2A)
#define IOP_VENDOR_ID_1	  	(0x00)
#define IOP_VENDOR_ID_2	  	(0x22)
#define IOP_VENDOR_ID_3	  	(0x07)
#define WFA_ELEM_VERSION2      (0x00)
#define WPS_VERSION            (0x20)
#define WFA_MAP_EXT_ATTR	   (0x06)
#define WFA_IOPSYS_EXT_ATTR	   (0x32)
#define ATTR_SSID              (0x1045)
#define ATTR_BRIDGE              (0x46)
#define ATTR_PROTO              (0x47)
#define ATTR_VID              (0x48)
#define ATTR_BR_IP              (0x49)
#define ATTR_BK_SSID              (0x4A)
#define ATTR_BK_KEY              (0x4B)
#define ATTR_RF_SUBAND		(0x4C)
#define ATTR_AUTH_TYPE         (0x1003)
#define ATTR_ENCR_TYPE         (0x100f)
#define ATTR_NETWORK_KEY       (0x1027)
#define ATTR_KEY_WRAP_AUTH     (0x101e)
#define ATTR_ENCR_SETTINGS     (0x1018)
#define ATTR_AUTHENTICATOR     (0x1005)



// Keys sizes
#define WPS_AUTHKEY_LEN    32
#define WPS_KEYWRAPKEY_LEN 16
#define WPS_EMSK_LEN       32

// This is the key derivation function used in the WPS standard to obtain a
// final hash that is later used for encryption.
//
// The output is stored in the memory buffer pointed by 'res', which must be
// "SHA256_MAC_LEN" bytes long (ie. 're_len' must always be "SHA256_MAC_LEN",
// even if it is an input argument)
//
static void _wps_key_derivation_function(INT8U *key, INT8U *label_prefix, INT32U label_prefix_len, char *label, INT8U *res, INT32U res_len)
{
	INT8U i_buf[4];
	INT8U key_bits[4];

	INT8U *addr[4];
	INT32U len[4];

	INT32U i, iter;

	INT8U hash[SHA256_MAC_LEN];
	INT8U *opos;

	INT32U left;

	INT8U *p;
	INT32U aux;

	aux = res_len * 8;
	p = key_bits;

	_I4B(&aux, &p);

	addr[0] = i_buf;
	addr[1] = label_prefix;
	addr[2] = (INT8U *) label;
	addr[3] = key_bits;
	len[0] = sizeof(i_buf);
	len[1] = label_prefix_len;
	len[2] = strlen(label);
	len[3] = sizeof(key_bits);

	iter = (res_len + SHA256_MAC_LEN - 1) / SHA256_MAC_LEN;
	opos = res;
	left = res_len;

	for (i = 1; i <= iter; i++) {
		p = i_buf;
		_I4B(&i, &p);

		PLATFORM_HMAC_SHA256(key, SHA256_MAC_LEN, 4, addr, len, hash);

		if (i < iter) {
			memcpy(opos, hash, SHA256_MAC_LEN);
			opos += SHA256_MAC_LEN;
			left -= SHA256_MAC_LEN;
		} else {
			memcpy(opos, hash, left);
		}
	}
}

//////////////////////////////////////// Enrollee functions ////////////////////
INT8U wscBuildM1(char *interface_name, INT8U **m1, INT16U *m1_size, void **key, INT8U *mac, INT8U rf_bands, INT16U auth_types, INT16U encryption_types)
{
	INT8U *buffer;

	struct wsc_key *private_key;

	INT8U *p;

	INT8U aux8;
	INT16U aux16;
	INT32U aux32;

	if (NULL == interface_name || NULL == m1 || NULL == m1_size || NULL == key) {
		fprintf(stderr, "Invalid arguments !\n");
		return 0;
	}

	CALLOC(buffer, 1000);
	p = buffer;

	// VERSION
	{
		aux16 = ATTR_VERSION;
		_I2B(&aux16, &p);
		aux16 = 1;
		_I2B(&aux16, &p);
		aux8 = 0x10;
		_I1B(&aux8, &p);
	}

	// MESSAGE TYPE
	{
		aux16 = ATTR_MSG_TYPE;
		_I2B(&aux16, &p);
		aux16 = 1;
		_I2B(&aux16, &p);
		aux8 = WPS_M1;
		_I1B(&aux8, &p);
	}

	// UUID
	{
		const char *uuid = "0000000000000000";

		aux16 = ATTR_UUID_E;
		_I2B(&aux16, &p);
		aux16 = 16;
		_I2B(&aux16, &p);
		_InB(uuid, &p, 16);
	}

	// MAC ADDRESS
	{
		aux16 = ATTR_MAC_ADDR;
		_I2B(&aux16, &p);
		aux16 = 6;
		_I2B(&aux16, &p);
		_InB(mac, &p, 6);
	}

	// ENROLLEE NONCE
	{
		INT8U enrollee_nonce[16];

		PLATFORM_GET_RANDOM_BYTES(enrollee_nonce, 16);

		aux16 = ATTR_ENROLLEE_NONCE;
		_I2B(&aux16, &p);
		aux16 = 16;
		_I2B(&aux16, &p);
		_InB(enrollee_nonce, &p, 16);
	}

	// PUBLIC KEY
	{
		INT8U *priv, *pub;
		INT16U priv_len = 0, pub_len = 0;

		PLATFORM_GENERATE_DH_KEY_PAIR(&priv, &priv_len, &pub, &pub_len);
		// TODO: ZERO PAD the pub key (doesn't seem to be really needed though)
		fprintf(stderr, "M1 plen|%d| publen|%d|\n", priv_len, pub_len);

		aux16 = ATTR_PUBLIC_KEY;
		_I2B(&aux16, &p);
		aux16 = pub_len;
		_I2B(&aux16, &p);
		_InB(pub, &p, pub_len);
		// The private key is one of the output arguments
		CALLOC(private_key, 1);
		CALLOC(private_key->key, priv_len);
		private_key->key_len = priv_len;
		if (priv_len > 0)
			memcpy(private_key->key, priv, priv_len);


		memcpy(private_key->mac, mac, 6);
		free(pub);
		free(priv);
	}

	// AUTHENTICATION TYPES
	{
		aux16 = ATTR_AUTH_TYPE_FLAGS;
		_I2B(&aux16, &p);
		aux16 = 2;
		_I2B(&aux16, &p);
		_I2B(&auth_types, &p);
	}

	// ENCRYPTION TYPES
	{
//			encryption_types |= WPS_ENCR_NONE;
//			encryption_types |= WPS_ENCR_TKIP;
//			encryption_types |= WPS_ENCR_AES;

		aux16 = ATTR_ENCR_TYPE_FLAGS;
		_I2B(&aux16, &p);
		aux16 = 2;
		_I2B(&aux16, &p);
		_I2B(&encryption_types, &p);
	}

	// CONNECTION TYPES
	{
		// Two possible types: ESS or IBSS. In the 1905 context, enrollees will
		// always want to acts as "ESS" to create an "extended" network where
		// all APs share the same credentials as the registrar.

		aux16 = ATTR_CONN_TYPE_FLAGS;
		_I2B(&aux16, &p);
		aux16 = 1;
		_I2B(&aux16, &p);
		aux8 = WPS_CONN_ESS;
		_I1B(&aux8, &p);
	}

	// CONFIGURATION METHODS
	{
		// In the 1905 context, the configuration methods the AP is willing to
		// offer will always be these two

		aux16 = ATTR_CONFIG_METHODS;
		_I2B(&aux16, &p);
		aux16 = 2;
		_I2B(&aux16, &p);
		aux16 = WPS_CONFIG_PHY_PUSHBUTTON | WPS_CONFIG_VIRT_PUSHBUTTON;
		_I2B(&aux16, &p);
	}

	// WPS STATE
	{
		aux16 = ATTR_WPS_STATE;
		_I2B(&aux16, &p);
		aux16 = 1;
		_I2B(&aux16, &p);
		aux8 = WPS_STATE_NOT_CONFIGURED;
		_I1B(&aux8, &p);
	}

	// MANUFACTURER
	{
		const char *manufacturer_name = "Unknown";

		aux16 = ATTR_MANUFACTURER;
		_I2B(&aux16, &p);
		aux16 = strlen(manufacturer_name);
		_I2B(&aux16, &p);
		_InB(manufacturer_name, &p, strlen(manufacturer_name));
	}

	// MODEL NAME
	{
		const char *model_name = "Unknown";

		aux16 = ATTR_MODEL_NAME;
		_I2B(&aux16, &p);
		aux16 = strlen(model_name);
		_I2B(&aux16, &p);
		_InB(model_name, &p, strlen(model_name));
	}

	// MODEL NUMBER
	{
		const char *model_number = "00000000";

		aux16 = ATTR_MODEL_NUMBER;
		_I2B(&aux16, &p);
		aux16 = strlen(model_number);
		_I2B(&aux16, &p);
		_InB(model_number, &p, strlen(model_number));
	}

	// SERIAL NUMBER
	{
		const char *serial_number = "00000000";

		aux16 = ATTR_SERIAL_NUMBER;
		_I2B(&aux16, &p);
		aux16 = strlen(serial_number);
		_I2B(&aux16, &p);
		_InB(serial_number, &p, strlen(serial_number));
	}

	// PRIMARY DEVICE TYPE
	{
		// In the 1905 context, they node sending a M1 message will always be
		// (at least) a "network router"

		INT8U oui[4] = { 0x00, 0x50, 0xf2, 0x00 };	// Fixed value from the
		// WSC spec

		aux16 = ATTR_PRIMARY_DEV_TYPE;
		_I2B(&aux16, &p);
		aux16 = 8;
		_I2B(&aux16, &p);
		aux16 = WPS_DEV_NETWORK_INFRA;
		_I2B(&aux16, &p);
		_InB(oui, &p, 4);
		aux16 = WPS_DEV_NETWORK_INFRA_ROUTER;
		_I2B(&aux16, &p);
	}

	// DEVICE NAME
	{
		const char *device_name = "Unknown";

		aux16 = ATTR_DEV_NAME;
		_I2B(&aux16, &p);
		aux16 = strlen(device_name);
		_I2B(&aux16, &p);
		_InB(device_name, &p, strlen(device_name));
	}

	// RF BANDS
	{
		// TODO
		// We should here list all supported freq bands (2.4, 5.0, 60 GHz)...
		// however each interfaces is already "pre-configured" to one specific
		// freq band... thus we will always report back a single value (instead
		// of and OR'ed list).
		// This should probably be improved in the future.

		aux16 = ATTR_RF_BANDS;
		_I2B(&aux16, &p);
		aux16 = 1;
		_I2B(&aux16, &p);
		_I1B(&rf_bands, &p);
	}

	// ASSOCIATION STATE
	{
		aux16 = ATTR_ASSOC_STATE;
		_I2B(&aux16, &p);
		aux16 = 2;
		_I2B(&aux16, &p);
		aux16 = WPS_ASSOC_NOT_ASSOC;
		_I2B(&aux16, &p);
	}

	// DEVICE PASSWORD ID
	{
		aux16 = ATTR_DEV_PASSWORD_ID;
		_I2B(&aux16, &p);
		aux16 = 2;
		_I2B(&aux16, &p);
		aux16 = DEV_PW_PUSHBUTTON;
		_I2B(&aux16, &p);
	}

	// CONFIG ERROR
	{
		aux16 = ATTR_CONFIG_ERROR;
		_I2B(&aux16, &p);
		aux16 = 2;
		_I2B(&aux16, &p);
		aux16 = WPS_CFG_NO_ERROR;
		_I2B(&aux16, &p);
	}

	// OS VERSION
	{
		// TODO: Fill with actual properties from the interface

		INT32U os_version = 0x00000001;

		aux16 = ATTR_OS_VERSION;
		_I2B(&aux16, &p);
		aux16 = 4;
		_I2B(&aux16, &p);
		aux32 = 0x80000000 | os_version;
		_I4B(&aux32, &p);
	}

	// VENDOR EXTENSIONS
	{
		aux16 = ATTR_VENDOR_EXTENSION;
		_I2B(&aux16, &p);
		aux16 = 6;
		_I2B(&aux16, &p);
		aux8 = WFA_VENDOR_ID_1;
		_I1B(&aux8, &p);
		aux8 = WFA_VENDOR_ID_2;
		_I1B(&aux8, &p);
		aux8 = WFA_VENDOR_ID_3;
		_I1B(&aux8, &p);
		aux8 = WFA_ELEM_VERSION2;
		_I1B(&aux8, &p);
		aux8 = 1;
		_I1B(&aux8, &p);
		aux8 = WPS_VERSION;
		_I1B(&aux8, &p);
	}

	*m1 = buffer;
	*m1_size = abs(p - buffer);
	*key = private_key;

	return 1;
}

INT8U wscProcessM2(void *key, INT8U *m1, INT16U m1_size, INT8U *m2, INT16U m2_size, struct mdata *out, char *ifname)
{
	INT8U *p;
	struct wsc_key *k;

	// "Useful data" we want to extract from M2
	INT8U mapie = 0;
	INT8U vid = 0;
	INT8U bridge[16] = {0};
	INT8U proto[8] = {0};
	INT8U bk_ssid[33] = {0};
	INT8U bk_key[65] = {0};
	INT32U br_ip = 0x00;
	INT8U ssid[33];
	INT8U ssid_present;
	INT8U bssid[6];
	INT8U bssid_present;
	INT16U auth_type;
	INT8U auth_type_present;
	INT16U encryption_type;
	INT8U encryption_type_present;
	INT8U network_key[65];
	INT8U network_key_present;
	INT8U band = 0;

	// Keys we need to compute to authenticate and decrypt M2
	INT8U authkey[WPS_AUTHKEY_LEN];
	INT8U keywrapkey[WPS_KEYWRAPKEY_LEN];
	INT8U emsk[WPS_EMSK_LEN];

	if (key == NULL || m1 == NULL || m1_size == 0) {
		fprintf(stderr, "%p %p %p\n", key, m1, m1_size);
		fprintf(stderr, "Missing m1 or key\n");
		return 0;
	}

	if (m2 == NULL || m2_size == 0) {
		fprintf(stderr, "Missing m2\n");
		return 0;
	}
	// "Intermediary data" we also need to extract from M2 to obtain the keys
	// that will let us decrypt the "useful data" from M2
	//
	INT8U *m2_nonce = NULL;
	INT8U m2_nonce_present;
	INT8U *m2_pubkey = NULL;
	INT8U m2_pubkey_present;
	INT16U m2_pubkey_len = 0;
	INT8U *m2_encrypted_settings = NULL;
	INT8U m2_encrypted_settings_present;
	INT16U m2_encrypted_settings_len = 0;
	INT8U *m2_authenticator = NULL;
	INT8U m2_authenticator_present;

	// "Intermediary data" we also need to extract from M1 to obtain the keys
	// that will let us decrypt the "useful data" from M2
	//
	INT8U *m1_nonce = NULL;
	INT8U m1_nonce_present;
	/*INT8U  *m1_pubkey; */ INT8U m1_pubkey_present;
	INT16U m1_pubkey_len;

	// "Intermediary data" contained in the "key" argument also needed to obtain
	// the keys that will let us decrypt the "useful data" from M2
	//
	INT8U *m1_privkey;
	INT16U m1_privkey_len;
	INT8U *m1_mac;

	k = (struct wsc_key *)key;
	m1_privkey = k->key;
	m1_privkey_len = k->key_len;
	m1_mac = k->mac;

	// Extract "intermediary data" from M2
	m2_nonce_present = 0;
	m2_pubkey_present = 0;
	m2_encrypted_settings_present = 0;
	m2_authenticator_present = 0;
	p = m2;
	while (abs(p - m2) < m2_size) {
		INT16U attr_type;
		INT16U attr_len;

		_E2B(&p, &attr_type);
		_E2B(&p, &attr_len);

		if (attr_type == ATTR_REGISTRAR_NONCE) {
			if (attr_len != 16) {
				fprintf(stderr, "Incorrect length (%d) for ATTR_REGISTRAR_NONCE\n", attr_len);
				return 0;
			}
			m2_nonce = p;

			p += attr_len;
			m2_nonce_present = 1;
		} else if (attr_type == ATTR_PUBLIC_KEY) {
			m2_pubkey_len = attr_len;
			m2_pubkey = p;

			p += attr_len;
			m2_pubkey_present = 1;
		} else if (attr_type == ATTR_ENCR_SETTINGS) {
			m2_encrypted_settings_len = attr_len;
			m2_encrypted_settings = p;

			p += attr_len;
			m2_encrypted_settings_present = 1;
		} else if (attr_type == ATTR_AUTHENTICATOR) {
			if (attr_len != 8) {
				fprintf(stderr, "Incorrect length (%d) for ATTR_AUTHENTICATOR\n", attr_len);
				return 0;
			}
			m2_authenticator = p;

			p += attr_len;
			m2_authenticator_present = 1;
		} else {
			p += attr_len;
		}
	}
	if (0 == m2_nonce_present || 0 == m2_pubkey_present || 0 == m2_encrypted_settings_present || 0 == m2_authenticator_present) {
		fprintf(stderr, "Missing attributes in the received M2 message\n");
		return 0;
	}
	// Extract "intermediary data" from M1
	m1_nonce_present = 0;
	m1_pubkey_present = 0;
	p = m1;
	while (abs(p - m1) < m1_size) {
		INT16U attr_type;
		INT16U attr_len;

		_E2B(&p, &attr_type);
		_E2B(&p, &attr_len);

		if (attr_type == ATTR_ENROLLEE_NONCE) {
			if (attr_len != 16) {
				fprintf(stderr, "Incorrect length (%d) for ATTR_REGISTRAR_NONCE\n", attr_len);
				return 0;
			}
			m1_nonce = p;

			p += attr_len;
			m1_nonce_present = 1;
		} else if (attr_type == ATTR_PUBLIC_KEY) {
			m1_pubkey_len = attr_len;
			/*m1_pubkey     = p; */

			p += m1_pubkey_len;
			m1_pubkey_present = 1;
		} else if (attr_type == ATTR_RF_BANDS) {
			_E1B(&p, &band);
		} else {
			p += attr_len;
		}
	}

	if (0 == m1_nonce_present || 0 == m1_pubkey_present) {
		fprintf(stderr, "Missing attributes in the received M1 message\n");
		return 0;
	}

	// With all the information we have just extracted from M1 and M2, obtain
	// the authentication/encryption keys.
	{
		INT8U *shared_secret;
		INT16U shared_secret_len;

		INT8U *addr[3];
		INT32U len[3];

		INT8U dhkey[SHA256_MAC_LEN];
		INT8U kdk[SHA256_MAC_LEN];

		INT8U keys[WPS_AUTHKEY_LEN + WPS_KEYWRAPKEY_LEN + WPS_EMSK_LEN];

		// With the enrolle public key (which we received in M1) and our private
		// key (which we generated above), obtain the Diffie Hellman shared
		// secret (when receiving M2, the enrollee will be able to obtain this
		// same shared secret using its private key and ou public key -contained
		// in M2-)
		//
		PLATFORM_COMPUTE_DH_SHARED_SECRET(&shared_secret, &shared_secret_len, m2_pubkey, m2_pubkey_len, m1_privkey, m1_privkey_len);
		// TODO: ZERO PAD the shared_secret (doesn't seem to be really needed
		// though)

		// Next, obtain the SHA-256 digest of this shared secret. We will call
		// this "dhkey"
		//
		addr[0] = shared_secret;
		len[0] = shared_secret_len;

		PLATFORM_SHA256(1, addr, len, dhkey);

		// Next, concatenate three things (the enrolle nonce contained in M1,
		// the enrolle MAC address, and the nonce we just generated before, and
		// calculate its HMAC (hash message authentication code) using "dhkey"
		// as the secret key.
		//
		addr[0] = m1_nonce;
		addr[1] = m1_mac;
		addr[2] = m2_nonce;
		len[0] = 16;
		len[1] = 6;
		len[2] = 16;

		PLATFORM_HMAC_SHA256(dhkey, SHA256_MAC_LEN, 3, addr, len, kdk);

		// Finally, take "kdk" and using a function provided in the "Wi-Fi
		// simple configuration" standard, obtain THREE KEYS that we will use
		// later ("authkey", "keywrapkey" and "emsk")
		//
		_wps_key_derivation_function(kdk, NULL, 0, "Wi-Fi Easy and Secure Key Derivation", keys, sizeof(keys));

		memcpy(authkey, keys, WPS_AUTHKEY_LEN);
		memcpy(keywrapkey, keys + WPS_AUTHKEY_LEN, WPS_KEYWRAPKEY_LEN);
		memcpy(emsk, keys + WPS_AUTHKEY_LEN + WPS_KEYWRAPKEY_LEN, WPS_EMSK_LEN);

		fprintf(stderr, "WPS keys: \n");
		fprintf(stderr, " Registrar pubkey(%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", m2_pubkey_len, m2_pubkey[0], m2_pubkey[1],
		      m2_pubkey[2], m2_pubkey[m2_pubkey_len - 3], m2_pubkey[m2_pubkey_len - 2], m2_pubkey[m2_pubkey_len - 1]);
		fprintf(stderr, " Enrollee privkey(%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", m1_privkey_len, m1_privkey[0], m1_privkey[1],
		      m1_privkey[2], m1_privkey[m1_privkey_len - 3], m1_privkey[m1_privkey_len - 2], m1_privkey[m1_privkey_len - 1]);
		fprintf(stderr, " Shared secret(%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", shared_secret_len, shared_secret[0],
		      shared_secret[1], shared_secret[2], shared_secret[shared_secret_len - 3], shared_secret[shared_secret_len - 2],
		      shared_secret[shared_secret_len - 1]);
		fprintf(stderr, " DH key( 32 bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", dhkey[0], dhkey[1], dhkey[2], dhkey[29],
		      dhkey[30], dhkey[31]);
		fprintf(stderr, " Enrollee nonce( 16 bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", m1_nonce[0], m1_nonce[1], m1_nonce[2],
		      m1_nonce[13], m1_nonce[14], m1_nonce[15]);
		fprintf(stderr, " Registrar nonce( 16 bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", m2_nonce[0], m2_nonce[1], m2_nonce[2],
		      m2_nonce[13], m2_nonce[14], m2_nonce[15]);
		fprintf(stderr, " KDK( 32 bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", kdk[0], kdk[1], kdk[2], kdk[29], kdk[30],
		      kdk[31]);
		fprintf(stderr, " authkey(%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", authkey[0], authkey[1], authkey[2],
		      authkey[WPS_AUTHKEY_LEN - 3], authkey[WPS_AUTHKEY_LEN - 2], authkey[WPS_AUTHKEY_LEN - 1]);
		fprintf(stderr, " keywrapkey(%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", keywrapkey[0], keywrapkey[1], keywrapkey[2],
		      keywrapkey[WPS_KEYWRAPKEY_LEN - 3], keywrapkey[WPS_KEYWRAPKEY_LEN - 2], keywrapkey[WPS_KEYWRAPKEY_LEN - 1]);
		fprintf(stderr, " emsk(%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", emsk[0], emsk[1], emsk[2], emsk[WPS_EMSK_LEN - 3],
		      emsk[WPS_EMSK_LEN - 2], emsk[WPS_EMSK_LEN - 1]);
		FREE(shared_secret);
	}

	// With the just computed key, check the message authentication
	{
		// Concatenate M1 and M2 (excluding the last 12 bytes, where the
		// authenticator attribute is contained) and calculate the HMAC, then
		// check it against the actual authenticator attribute value.
		//
		INT8U hash[SHA256_MAC_LEN];

		INT8U *addr[2];
		INT32U len[2];

		addr[0] = m1;
		addr[1] = m2;
		len[0] = m1_size;
		len[1] = m2_size - 12;

		PLATFORM_HMAC_SHA256(authkey, WPS_AUTHKEY_LEN, 2, addr, len, hash);

		if (memcmp(m2_authenticator, hash, 8) != 0) {
			fprintf(stderr, "Message M2 authentication failed\n");
			return 0;
		}
	}

	// With the just computed keys, decrypt the message and check the keywrap
	{
		INT8U *plain;
		INT32U plain_len;

		INT8U m2_keywrap_present;

		plain = m2_encrypted_settings + AES_BLOCK_SIZE;
		plain_len = m2_encrypted_settings_len - AES_BLOCK_SIZE;

		fprintf(stderr, "AP settings before decryption (%d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n",
		      plain_len, plain[0], plain[1], plain[2],
		      plain[plain_len - 3], plain[plain_len - 2],
		      plain[plain_len - 1]);

		fprintf(stderr, "IV (%d bytes)                           : 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n",
		      AES_BLOCK_SIZE, m2_encrypted_settings[0],
		      m2_encrypted_settings[1], m2_encrypted_settings[2],
		      m2_encrypted_settings[AES_BLOCK_SIZE - 3],
		      m2_encrypted_settings[AES_BLOCK_SIZE - 2],
		      m2_encrypted_settings[AES_BLOCK_SIZE - 1]);

		PLATFORM_AES_DECRYPT(keywrapkey, m2_encrypted_settings, plain, plain_len);
		fprintf(stderr, "AP settings after decryption (%d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n",
		      plain_len, plain[0], plain[1], plain[2],
		      plain[plain_len - 3], plain[plain_len - 2], plain[plain_len - 1]);

		// Remove padding
		plain_len -= plain[plain_len - 1];

		// Parse contents of decrypted settings
		ssid_present = 0;
		bssid_present = 0;
		auth_type_present = 0;
		encryption_type_present = 0;
		network_key_present = 0;
		m2_keywrap_present = 0;
		p = plain;
		while ((INT32U) (abs(p - plain)) < plain_len) {
			INT16U attr_type;
			INT16U attr_len;

			_E2B(&p, &attr_type);
			_E2B(&p, &attr_len);

			if (attr_type == ATTR_SSID) {
				_EnB(&p, ssid, attr_len);
				ssid[attr_len] = 0x00;
				ssid_present = 1;
			} else if (attr_type == ATTR_AUTH_TYPE) {
				_E2B(&p, &auth_type);
				auth_type_present = 1;
			} else if (attr_type == ATTR_ENCR_TYPE) {
				_E2B(&p, &encryption_type);
				encryption_type_present = 1;
			} else if (attr_type == ATTR_NETWORK_KEY) {
				_EnB(&p, network_key, attr_len);
				network_key[attr_len] = 0x00;
				network_key_present = 1;
			} else if (attr_type == ATTR_MAC_ADDR) {
				_EnB(&p, bssid, attr_len);
				bssid_present = 1;
			} else if (attr_type == ATTR_KEY_WRAP_AUTH) {
				// This attribute is always the last one contained in the plain
				// text, thus 4 bytes *before* where "p" is pointing right now
				// is the end of the plain text blob whose HMAC we are going to
				// compute to check the keywrap.
				//
				INT8U *end_of_hmac;
				INT8U hash[SHA256_MAC_LEN];

				INT8U *addr[1];
				INT32U len[1];

				end_of_hmac = p - 4;

				addr[0] = plain;
				len[0] = abs(end_of_hmac - plain);

				PLATFORM_HMAC_SHA256(authkey, WPS_AUTHKEY_LEN, 1, addr, len, hash);

				if (memcmp(p, hash, 8) != 0) {
					fprintf(stderr, "Message M2 keywrap failed\n", attr_len);
					return 0;
				}

				p += attr_len;
				m2_keywrap_present = 1;
			} else if (ATTR_VENDOR_EXTENSION == attr_type) {
				INT8U id[3];
				INT8U *end;
				INT8U subelem;
				INT8U len;

				// May be one or more subelements (Section 12 of WSC spec)
				end = p + attr_len;
				_EnB(&p, id, sizeof(id));
				if (id[0] == WFA_VENDOR_ID_1 &&
						id[1] == WFA_VENDOR_ID_2 &&
						id[2] == WFA_VENDOR_ID_3) {
					while (p < end) {
						_E1B(&p, &subelem);
						_E1B(&p, &len);
						if (subelem == WFA_MAP_EXT_ATTR) {
							// Map extension subelement will be 1 byte
							_E1B(&p, &mapie);
						} else
							p += len;
					}
				} else if (id[0] == IOP_VENDOR_ID_1 &&
						id[1] == IOP_VENDOR_ID_2 &&
						id[2] == IOP_VENDOR_ID_3) {
					while (p < end) {
						_E1B(&p, &subelem);
						_E1B(&p, &len);

						if (subelem == ATTR_BRIDGE)
							_EnB(&p, bridge, len);
						else if (subelem == ATTR_PROTO)
							_EnB(&p, proto, len);
						else if (subelem == ATTR_VID)
							_E1B(&p, &vid);
						else if (subelem == ATTR_BR_IP)
							_EnB(&p, &br_ip, len);
						else if (subelem == ATTR_BK_SSID)
							_EnB(&p, bk_ssid, len);
						else if (subelem == ATTR_BK_KEY)
							_EnB(&p, bk_key, len);
						else
							p += len;
					}
				}
			} else
				p += attr_len;
		}
		if ( (ssid_present &
		     bssid_present &
		     auth_type_present &
		     encryption_type_present &
		     network_key_present &
		     m2_keywrap_present) == 0) {
			fprintf(stderr, "Missing attributes in the configuration settings received in the M2 message\n");
			return 0;
		}
	}

	memcpy(out->output.ssid, ssid, sizeof(ssid));
	memcpy(out->output.network_key, network_key, sizeof(network_key));
	out->output.auth_types = auth_type;
	out->output.encryption_types = encryption_type;
	out->output.mapie = mapie;
	memcpy(out->output.bridge, bridge, sizeof(bridge));
	memcpy(out->output.proto, proto, sizeof(proto));
	out->output.br_ip = br_ip;
	memcpy(out->output.bk_ssid, bk_ssid, sizeof(bk_ssid));
	memcpy(out->output.bk_key, bk_key, sizeof(bk_key));
	out->output.vid = vid;
	memcpy(out->output.bssid, bssid, sizeof(bssid));
	out->output.band = band;

	return 1;
}

//////////////////////////////////////// Registrar functions ///////////////////
INT8U wscBuildM2(INT8U *m1, INT16U m1_size, INT8U **m2, INT16U *m2_size, INT8U rf_bands, INT8U *mac, struct mdata *in)
{
	INT8U *buffer;

	INT8U *p;

	INT8U aux8;
	INT16U var16;
	INT32U aux32;

	INT8U *m1_mac_address = NULL;
	INT8U m1_mac_address_present;
	INT8U *m1_nonce = NULL;
	INT8U m1_nonce_present;
	INT8U *m1_pubkey = NULL;
	INT8U m1_pubkey_present;
	INT16U m1_pubkey_len = 0;
	INT8U m1_rf_band = WPS_RF_24GHZ;

	INT8U *local_privkey;
	INT16U local_privkey_len;

	INT8U authkey[WPS_AUTHKEY_LEN];
	INT8U keywrapkey[WPS_KEYWRAPKEY_LEN];
	INT8U emsk[WPS_EMSK_LEN];

	INT8U registrar_nonce[16];
	INT8U m2_mac[6] = {0};

	INT16U encryption_types;
	INT16U auth_types;

	// We first need to extract the following parameters contained in "M1":
	//
	//   - Mac address
	//   - Nounce
	//   - Public key
	//
	m1_mac_address_present = 0;
	m1_nonce_present = 0;
	m1_pubkey_present = 0;

	p = m1;

	if (mac)
		memcpy(m2_mac, mac, sizeof(m2_mac));

	while (abs(p - m1) < m1_size) {
		INT16U attr_type;
		INT16U attr_len;

		_E2B(&p, &attr_type);
		_E2B(&p, &attr_len);

		if (attr_type == ATTR_MAC_ADDR) {
			if (attr_len != 6) {
				fprintf(stderr, "Incorrect length (%d) for ATTR_MAC_ADDR\n", attr_len);
				return 0;
			}
			m1_mac_address = p;

			p += attr_len;
			m1_mac_address_present = 1;
		} else if (attr_type == ATTR_ENROLLEE_NONCE) {
			if (attr_len != 16) {
				fprintf(stderr, "Incorrect length (%d) for ATTR_ENROLLEE_NONCE\n", attr_len);
				return 0;
			}
			m1_nonce = p;

			p += attr_len;
			m1_nonce_present = 1;
		} else if (attr_type == ATTR_PUBLIC_KEY) {
			m1_pubkey_len = attr_len;
			m1_pubkey = p;

			p += attr_len;
			m1_pubkey_present = 1;
		} else if (attr_type == ATTR_RF_BANDS) {
			_E1B(&p, &m1_rf_band);
		} else {
			p += attr_len;
		}
	}

	CALLOC(buffer, 1000);
	p = buffer;

	// VERSION
	{
		var16 = ATTR_VERSION;
		_I2B(&var16, &p);
		var16 = 1;
		_I2B(&var16, &p);
		aux8 = 0x10;
		_I1B(&aux8, &p);
	}

	// MESSAGE TYPE
	{
		var16 = ATTR_MSG_TYPE;
		_I2B(&var16, &p);
		var16 = 1;
		_I2B(&var16, &p);
		aux8 = WPS_M2;
		_I1B(&aux8, &p);
	}

	// ENROLLEE NONCE
	{
		var16 = ATTR_ENROLLEE_NONCE;
		_I2B(&var16, &p);
		var16 = 16;
		_I2B(&var16, &p);
		_InB(m1_nonce, &p, 16);
	}

	// REGISTRAR NONCE
	{
		PLATFORM_GET_RANDOM_BYTES(registrar_nonce, 16);

		var16 = ATTR_REGISTRAR_NONCE;
		_I2B(&var16, &p);
		var16 = 16;
		_I2B(&var16, &p);
		_InB(registrar_nonce, &p, 16);
	}

	// UUID
	{
		const char *uuid = "0000000000000000";

		var16 = ATTR_UUID_R;
		_I2B(&var16, &p);
		var16 = 16;
		_I2B(&var16, &p);
		_InB(uuid, &p, 16);
	}

	// PUBLIC KEY
	{
		INT8U *priv, *pub;
		INT16U priv_len = 0, pub_len = 0;

		PLATFORM_GENERATE_DH_KEY_PAIR(&priv, &priv_len, &pub, &pub_len);
		fprintf(stderr, "plen|%d|, publen|%d|\n", priv_len, pub_len);
		// TODO: ZERO PAD the pub key (doesn't seem to be really needed though)

		var16 = ATTR_PUBLIC_KEY;
		_I2B(&var16, &p);
		var16 = pub_len;
		_I2B(&var16, &p);
		_InB(pub, &p, pub_len);

		// We will use it later... save it.
		local_privkey = priv;
		local_privkey_len = priv_len;
		free(pub);
	}

	// Key derivation (no bytes are written to the output buffer in the next
	// block of code, we just obtain three cryptographic keys that are needed
	// later
	{
		INT8U *shared_secret;
		INT16U shared_secret_len;

		INT8U *addr[3];
		INT32U len[3];

		INT8U dhkey[SHA256_MAC_LEN];
		INT8U kdk[SHA256_MAC_LEN];

		INT8U keys[WPS_AUTHKEY_LEN + WPS_KEYWRAPKEY_LEN + WPS_EMSK_LEN];

		// With the enrolle public key (which we received in M1) and our private
		// key (which we generated above), obtain the Diffie Hellman shared
		// secret (when receiving M2, the enrollee will be able to obtain this
		// same shared secret using its private key and our public key
		// -contained in M2-)
		//
		PLATFORM_COMPUTE_DH_SHARED_SECRET(&shared_secret, &shared_secret_len, m1_pubkey, m1_pubkey_len, local_privkey, local_privkey_len);
		// TODO: ZERO PAD the shared_secret (doesn't seem to be really needed
		// though)

		// Next, obtain the SHA-256 digest of this shared secret. We will call
		// this "dhkey"
		//
		addr[0] = shared_secret;
		len[0] = shared_secret_len;

		PLATFORM_SHA256(1, addr, len, dhkey);

		// Next, concatenate three things (the enrollee nonce contained in M1,
		// the enrolle MAC address -also contained in M1-, and the nonce we just
		// generated before and calculate its HMAC (hash message authentication
		// code) using "dhkey" as the secret key.
		//
		addr[0] = m1_nonce;
		addr[1] = m1_mac_address;
		addr[2] = registrar_nonce;
		len[0] = 16;
		len[1] = 6;
		len[2] = 16;

		PLATFORM_HMAC_SHA256(dhkey, SHA256_MAC_LEN, 3, addr, len, kdk);

		// Finally, take "kdk" and using a function provided in the "Wi-Fi
		// simple configuration" standard, obtain THREE KEYS that we will use
		// later ("authkey", "keywrapkey" and "emsk")
		//
		_wps_key_derivation_function(kdk, NULL, 0, "Wi-Fi Easy and Secure Key Derivation", keys, sizeof(keys));

		memcpy(authkey, keys, WPS_AUTHKEY_LEN);
		memcpy(keywrapkey, keys + WPS_AUTHKEY_LEN, WPS_KEYWRAPKEY_LEN);
		memcpy(emsk, keys + WPS_AUTHKEY_LEN + WPS_KEYWRAPKEY_LEN, WPS_EMSK_LEN);

		fprintf(stderr, "WPS keys: \n");
		fprintf(stderr, "  Enrollee pubkey   (%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", m1_pubkey_len, m1_pubkey[0], m1_pubkey[1],
		      m1_pubkey[2], m1_pubkey[m1_pubkey_len - 3], m1_pubkey[m1_pubkey_len - 2], m1_pubkey[m1_pubkey_len - 1]);
		fprintf(stderr, "  registrar privkey (%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", local_privkey_len, local_privkey[0],
		      local_privkey[1], local_privkey[2], local_privkey[local_privkey_len - 3], local_privkey[local_privkey_len - 2],
		      local_privkey[local_privkey_len - 1]);
		fprintf(stderr, "  shared secret     (%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", shared_secret_len, shared_secret[0],
		      shared_secret[1], shared_secret[2], shared_secret[shared_secret_len - 3], shared_secret[shared_secret_len - 2],
		      shared_secret[shared_secret_len - 1]);
		fprintf(stderr, "  dh key            ( 32 bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", dhkey[0], dhkey[1], dhkey[2], dhkey[29],
		      dhkey[30], dhkey[31]);
		fprintf(stderr, "  enrollee nonce    ( 16 bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", m1_nonce[0], m1_nonce[1], m1_nonce[2],
		      m1_nonce[13], m1_nonce[14], m1_nonce[15]);
		fprintf(stderr, "  registrar nonce   ( 16 bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", registrar_nonce[0], registrar_nonce[1],
		      registrar_nonce[2], registrar_nonce[13], registrar_nonce[14], registrar_nonce[15]);
		fprintf(stderr, "  kdk               ( 32 bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", kdk[0], kdk[1], kdk[2], kdk[29], kdk[30],
		      kdk[31]);
		fprintf(stderr, "  auth_key           (%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", authkey[0], authkey[1], authkey[2],
		      authkey[WPS_AUTHKEY_LEN - 3], authkey[WPS_AUTHKEY_LEN - 2], authkey[WPS_AUTHKEY_LEN - 1]);
		fprintf(stderr, "  keywrap_key        (%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", keywrapkey[0], keywrapkey[1], keywrapkey[2],
		      keywrapkey[WPS_KEYWRAPKEY_LEN - 3], keywrapkey[WPS_KEYWRAPKEY_LEN - 2], keywrapkey[WPS_KEYWRAPKEY_LEN - 1]);
		fprintf(stderr, "  emsk              (%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", emsk[0], emsk[1], emsk[2], emsk[WPS_EMSK_LEN - 3],
		      emsk[WPS_EMSK_LEN - 2], emsk[WPS_EMSK_LEN - 1]);
		FREE(shared_secret);
	}

	// AUTHENTICATION TYPES
	{
		auth_types = in->input.auth_types;


		var16 = ATTR_AUTH_TYPE_FLAGS;
		_I2B(&var16, &p);
		var16 = 2;
		_I2B(&var16, &p);
		_I2B(&auth_types, &p);
	}

	// ENCRYPTION TYPES
	{
		encryption_types = in->input.encryption_types;

		var16 = ATTR_ENCR_TYPE_FLAGS;
		_I2B(&var16, &p);
		var16 = 2;
		_I2B(&var16, &p);
		_I2B(&encryption_types, &p);
	}

	// CONNECTION TYPES
	{
		// Two possible types: ESS or IBSS. In the 1905 context, registrars
		// will always be "ESS", meaning they are willing to have their
		// credentials cloned by other APs in order to end up with a network
		// which is "roaming-friendly" ("ESS": "extended service set")

		var16 = ATTR_CONN_TYPE_FLAGS;
		_I2B(&var16, &p);
		var16 = 1;
		_I2B(&var16, &p);
		aux8 = WPS_CONN_ESS;
		_I1B(&aux8, &p);
	}

	// CONFIGURATION METHODS
	{
		// In the 1905 context, the configuration methods the AP is willing to
		// offer will always be these two

		var16 = ATTR_CONFIG_METHODS;
		_I2B(&var16, &p);
		var16 = 2;
		_I2B(&var16, &p);
		var16 = WPS_CONFIG_PHY_PUSHBUTTON | WPS_CONFIG_VIRT_PUSHBUTTON;
		_I2B(&var16, &p);
	}

	// MANUFACTURER
	{
		const char *manufacturer_name = "Unknown";

		var16 = ATTR_MANUFACTURER;
		_I2B(&var16, &p);
		var16 = strlen(manufacturer_name);
		_I2B(&var16, &p);
		_InB(manufacturer_name, &p, strlen(manufacturer_name));
	}

	// MODEL NAME
	{
		const char *model_name = "Unknown";

		var16 = ATTR_MODEL_NAME;
		_I2B(&var16, &p);
		var16 = strlen(model_name);
		_I2B(&var16, &p);
		_InB(model_name, &p, strlen(model_name));
	}

	// MODEL NUMBER
	{
		const char *model_number = "00000000";

		var16 = ATTR_MODEL_NUMBER;
		_I2B(&var16, &p);
		var16 = strlen(model_number);
		_I2B(&var16, &p);
		_InB(model_number, &p, strlen(model_number));
	}

	// SERIAL NUMBER
	{
		const char *serial_number = "00000000";

		var16 = ATTR_SERIAL_NUMBER;
		_I2B(&var16, &p);
		var16 = strlen(serial_number);
		_I2B(&var16, &p);
		_InB(serial_number, &p, strlen(serial_number));
	}

	// PRIMARY DEVICE TYPE
	{
		// In the 1905 context, they node sending a M2 message will always be
		// (at least) a "network router"

		INT8U oui[4] = { 0x00, 0x50, 0xf2, 0x00 };	// Fixed value from the
		// WSC spec

		var16 = ATTR_PRIMARY_DEV_TYPE;
		_I2B(&var16, &p);
		var16 = 8;
		_I2B(&var16, &p);
		var16 = WPS_DEV_NETWORK_INFRA;
		_I2B(&var16, &p);
		_InB(oui, &p, 4);
		var16 = WPS_DEV_NETWORK_INFRA_ROUTER;
		_I2B(&var16, &p);
	}

	// DEVICE NAME
	{
		const char *device_name = "Unknown";

		var16 = ATTR_DEV_NAME;
		_I2B(&var16, &p);
		var16 = strlen(device_name);
		_I2B(&var16, &p);
		_InB(device_name, &p, strlen(device_name));
	}

	// RF BANDS
	{
		var16 = ATTR_RF_BANDS;
		_I2B(&var16, &p);
		var16 = 1;
		_I2B(&var16, &p);
		_I1B(&rf_bands, &p);
	}

	// ASSOCIATION STATE
	{
		var16 = ATTR_ASSOC_STATE;
		_I2B(&var16, &p);
		var16 = 2;
		_I2B(&var16, &p);
		var16 = WPS_ASSOC_CONN_SUCCESS;
		_I2B(&var16, &p);
	}

	// CONFIG ERROR
	{
		var16 = ATTR_CONFIG_ERROR;
		_I2B(&var16, &p);
		var16 = 2;
		_I2B(&var16, &p);
		var16 = WPS_CFG_NO_ERROR;
		_I2B(&var16, &p);
	}

	// DEVICE PASSWORD ID
	{
		var16 = ATTR_DEV_PASSWORD_ID;
		_I2B(&var16, &p);
		var16 = 2;
		_I2B(&var16, &p);
		var16 = DEV_PW_PUSHBUTTON;
		_I2B(&var16, &p);
	}

	// OS VERSION
	{
		// TODO: Fill with actual properties from the interface

		INT32U os_version = 0x00000001;

		var16 = ATTR_OS_VERSION;
		_I2B(&var16, &p);
		var16 = 4;
		_I2B(&var16, &p);
		aux32 = 0x80000000 | os_version;
		_I4B(&aux32, &p);
	}

	// VENDOR EXTENSIONS
	{
		var16 = ATTR_VENDOR_EXTENSION;
		_I2B(&var16, &p);
		var16 = 6;
		_I2B(&var16, &p);
		aux8 = WFA_VENDOR_ID_1;
		_I1B(&aux8, &p);
		aux8 = WFA_VENDOR_ID_2;
		_I1B(&aux8, &p);
		aux8 = WFA_VENDOR_ID_3;
		_I1B(&aux8, &p);
		aux8 = WFA_ELEM_VERSION2;
		_I1B(&aux8, &p);
		aux8 = 1;
		_I1B(&aux8, &p);
		aux8 = WPS_VERSION;
		_I1B(&aux8, &p);
	}

	// ENCRYPTED SETTINGS
	{
		// This is what we are going to do next:
		//
		//   1. Fill a tmp buffer ("aux") ith credential attributes (SSID,
		//      network key, etc...)
		//
		//   2. Add an HMAC to this tmp buffer (so that the enrollee, when
		//      receiving this buffer in M2, can be sure no one has tampered
		//      with these attributes)
		//
		//   3. Encryp the message + HMAC with AES (so that no one else
		//      snooping can have a look at these attributes)
		char *ssid;
		char *network_key;

		INT8U plain[312];
		INT8U hash[SHA256_MAC_LEN];
		INT8U *iv_start;
		INT8U *data_start;
		INT8U *r;
		INT8U pad_elements_nr;

		INT8U *addr[1];
		INT32U len[1];

		ssid = (char *) in->input.ssid;
		network_key = (char *) in->input.network_key;

		r = plain;

		// SSID
		var16 = ATTR_SSID;
		_I2B(&var16, &r);
		var16 = strlen(ssid);
		_I2B(&var16, &r);
		_InB(ssid, &r, strlen(ssid));

		// AUTH TYPE
		var16 = ATTR_AUTH_TYPE;
		_I2B(&var16, &r);
		var16 = 2;
		_I2B(&var16, &r);
		var16 = auth_types;
		_I2B(&var16, &r);

		// ENCRYPTION TYPE
		var16 = ATTR_ENCR_TYPE;
		_I2B(&var16, &r);
		var16 = 2;
		_I2B(&var16, &r);
		var16 = encryption_types;
		_I2B(&var16, &r);

		// NETWORK KEY
		var16 = ATTR_NETWORK_KEY;
		_I2B(&var16, &r);
		var16 = strlen(network_key);
		_I2B(&var16, &r);
		_InB(network_key, &r, strlen(network_key));

		// MAC ADDR
		var16 = ATTR_MAC_ADDR;
		_I2B(&var16, &r);
		var16 = 6;
		_I2B(&var16, &r);
		_InB(m2_mac, &r, 6);

		// MULTI-AP VENDOR EXTENSIONS
		var16 = ATTR_VENDOR_EXTENSION;
		_I2B(&var16, &r);
		var16 = 9;
		_I2B(&var16, &r);
		aux8 = WFA_VENDOR_ID_1;
		_I1B(&aux8, &r);
		aux8 = WFA_VENDOR_ID_2;
		_I1B(&aux8, &r);
		aux8 = WFA_VENDOR_ID_3;
		_I1B(&aux8, &r);
		aux8 = WFA_ELEM_VERSION2;
		_I1B(&aux8, &r);
		aux8 = 1;
		_I1B(&aux8, &r);
		aux8 = WPS_VERSION;
		_I1B(&aux8, &r);

		// MULTI-AP EXTENSION SUBELEMENT
		aux8 = WFA_MAP_EXT_ATTR;
		_I1B(&aux8, &r);
		aux8 = 1;
		_I1B(&aux8, &r);
		aux8 = in->input.mapie;
		_I1B(&aux8, &r);

		// IOPSYS VENDOR EXTENSION
		var16 = ATTR_VENDOR_EXTENSION;
		_I2B(&var16, &r);
		// IOP vendor ID + interface len + proto len + VID
		var16 = 3
				+ 2 + strlen(in->input.bridge)
				+ 2 + strlen(in->input.proto)
				+ 3
				+ 6
				+ 2 + strlen(in->input.bk_ssid)
				+ 2 + strlen(in->input.bk_key);
		_I2B(&var16, &r);

		// IOPSYS ID
		aux8 = IOP_VENDOR_ID_1;
		_I1B(&aux8, &r);
		aux8 = IOP_VENDOR_ID_2;
		_I1B(&aux8, &r);
		aux8 = IOP_VENDOR_ID_3;
		_I1B(&aux8, &r);

		// BRIDGE
		aux8 = ATTR_BRIDGE;
		_I1B(&aux8, &r);
		aux8 = strlen(in->input.bridge);
		_I1B(&aux8, &r);
		_InB(in->input.bridge, &r, strlen(in->input.bridge));

		// PROTO
		aux8 = ATTR_PROTO;
		_I1B(&aux8, &r);
		aux8 = strlen(in->input.proto);
		_I1B(&aux8, &r);
		_InB(in->input.proto, &r, strlen(in->input.proto));

		// VID
		aux8 = ATTR_VID;
		_I1B(&aux8, &r);
		aux8 = 1;
		_I1B(&aux8, &r);
		aux8 = in->input.vid;
		_I1B(&aux8, &r);

		// BRIDGE IP ADDRESS
		aux8 = ATTR_BR_IP;
		_I1B(&aux8, &r);
		aux8 = 4;
		_I1B(&aux8, &r);
		_InB(&in->input.br_ip, &r, 4);

		// BACKHAUL SSID
		aux8 = ATTR_BK_SSID;
		_I1B(&aux8, &r);
		aux8 = strlen(in->input.bk_ssid);
		_I1B(&aux8, &r);
		_InB(in->input.bk_ssid, &r, strlen(in->input.bk_ssid));

		// BACKHAUL KEY
		aux8 = ATTR_BK_KEY;
		_I1B(&aux8, &r);
		aux8 = strlen(in->input.bk_key);
		_I1B(&aux8, &r);
		_InB(in->input.bk_key, &r, strlen(in->input.bk_key));

		fprintf(stderr, "AP configuration settings that we are going to send:\n");
		fprintf(stderr, "  - SSID            : %s\n", ssid);
		fprintf(stderr, "  - BSSID           : %02x:%02x:%02x:%02x:%02x:%02x\n",
		     m2_mac[0], m2_mac[1],
		     m2_mac[2], m2_mac[3],
		      m2_mac[4], m2_mac[5]);
		fprintf(stderr, "  - AUTH_TYPE       : 0x%04x\n", auth_types);
		fprintf(stderr, "  - ENCRYPTION_TYPE : 0x%04x\n", encryption_types);
		fprintf(stderr, "  - NETWORK_KEY     : %s\n", network_key);
		fprintf(stderr, "  - MAP_EXTENSION   : 0x%02x\n", in->input.mapie);
		fprintf(stderr, "  - BRIDGE          : %s\n", in->input.bridge);
		fprintf(stderr, "  - PROTO           : %s\n", in->input.proto);
		fprintf(stderr, "  - VID             : 0x%02x\n", in->input.vid);
		fprintf(stderr, "  - BR_IP           : %d.%d.%d.%d\n",
				((in->input.br_ip) & 0xff),
				((in->input.br_ip >> 8) & 0xff),
				((in->input.br_ip >> 16) & 0xff),
				((in->input.br_ip >> 24) & 0xff));
		fprintf(stderr, "  - BK_SSID         : %s\n", in->input.bk_ssid);
		fprintf(stderr, "  - BK_KEY          : %s\n", in->input.bk_key);
		// Obtain the HMAC of the whole plain buffer using "authkey" as the
		// secret key.
		//
		addr[0] = plain;
		len[0] = abs(r - plain);
		PLATFORM_HMAC_SHA256(authkey, WPS_AUTHKEY_LEN, 1, addr, len, hash);

		// ...and add it to the same plain buffer (well, only the first 8 bytes
		// of the hash)
		//
		var16 = ATTR_KEY_WRAP_AUTH;
		_I2B(&var16, &r);
		var16 = 8;
		_I2B(&var16, &r);
		_InB(hash, &r, 8);

		// Finally, encrypt everything with AES and add the resulting blob to
		// the M2 buffer, as an "ATTR_ENCR_SETTINGS" attribute
		//
		//// Pad the length of the message to encrypt to a multiple of
		//// AES_BLOCK_SIZE. The new padded bytes must have their value equal to
		//// the amount of bytes padded (PKCS#5 v2.0 pad)
		////
		pad_elements_nr = AES_BLOCK_SIZE - (abs(r - plain) % AES_BLOCK_SIZE);
		for (aux8 = 0; aux8 < pad_elements_nr; aux8++) {
			_I1B(&pad_elements_nr, &r);
		}
		//// Add the attribute header ("type" and "lenght") to the M2 buffer,
		//// followed by the IV and the data to encrypt.
		////
		var16 = ATTR_ENCR_SETTINGS;
		_I2B(&var16, &p);
		var16 = AES_BLOCK_SIZE + abs(r - plain);
		_I2B(&var16, &p);
		iv_start = p;
		PLATFORM_GET_RANDOM_BYTES(p, AES_BLOCK_SIZE);
		p += AES_BLOCK_SIZE;
		data_start = p;
		_InB(plain, &p, abs(r - plain));
		//// Encrypt the data IN-PLACE. Note that the "ATTR_ENCR_SETTINGS"
		//// attribute containes both the IV and the encrypted data.
		////
		INT32U setting_len = abs(r - plain);

		fprintf(stderr, "AP settings before encryption (%d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", setting_len, data_start[0], data_start[1],
		      data_start[2], data_start[setting_len - 3], data_start[setting_len - 2], data_start[setting_len - 1]);
		fprintf(stderr, "IV (%d bytes)                           : 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", AES_BLOCK_SIZE, iv_start[0], iv_start[1],
		      iv_start[2], iv_start[AES_BLOCK_SIZE - 3], iv_start[AES_BLOCK_SIZE - 2], iv_start[AES_BLOCK_SIZE - 1]);
		PLATFORM_AES_ENCRYPT(keywrapkey, iv_start, data_start, setting_len);
		fprintf(stderr, "AP settings after  encryption (%d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x\n", setting_len, data_start[0], data_start[1],
		      data_start[2], data_start[setting_len - 3], data_start[setting_len - 2], data_start[setting_len - 1]);
	}

	// AUTHENTICATOR
	{
		// This one is easy: concatenate M1 and M2 (everything in the M2 buffer
		// up to this point) and calculate the HMAC, then append it to M2 as
		// a new (and final!) attribute
		//
		INT8U hash[SHA256_MAC_LEN];

		INT8U *addr[2];
		INT32U len[2];

		addr[0] = m1;
		addr[1] = buffer;
		len[0] = m1_size;
		len[1] = abs(p - buffer);

		PLATFORM_HMAC_SHA256(authkey, WPS_AUTHKEY_LEN, 2, addr, len, hash);

		var16 = ATTR_AUTHENTICATOR;
		_I2B(&var16, &p);
		var16 = 8;
		_I2B(&var16, &p);
		_InB(hash, &p, 8);
	}

	*m2 = buffer;
	*m2_size = abs(p - buffer);

	free(local_privkey);
	return 1;
}
