/*
 * al_datamodel.c: IEEE-1905 AL datamodel handling
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "utils.h"
#include "al_datamodel.h"
#include "cmdu_visit.h"
#include "platform_interfaces.h"

#include <easy/easy.h>
#include <linux/if.h>
////////////////////////////////////////////////////////////////////////////////
// Private stuff
////////////////////////////////////////////////////////////////////////////////

static struct _dataModel {
	INT8U network_registrar_nr;
	struct _networkRegistrar {
		INT8U registrar_type;
		INT8U registrar_mac_address[6];
	} *network_registrar;

	INT8U al_mac_address[6];
	INT8U local_interfaces_nr;

	struct _localInterface {
		char name[IFNAMSIZ];
		INT8U mac_address[6];

		INT8U neighbors_nr;

		struct _neighbor {
			INT8U al_mac_address[6];
			INT8U remote_interfaces_nr;

			struct _remoteInterface {
				INT8U mac_address[6];
				INT32U last_topology_discovery_ts;
				INT32U last_bridge_discovery_ts;

			} *remote_interfaces;

		} *neighbors;

	} *local_interfaces;

	INT8U network_devices_nr;

	struct _networkDevice {
		INT32U update_timestamp;

		struct deviceInformationTypeTLV *info;
		INT8U bridges_nr;
		struct deviceBridgingCapabilityTLV **bridges;

		INT8U non1905_neighbors_nr;
		struct non1905NeighborDeviceListTLV **non1905_neighbors;

		INT8U x1905_neighbors_nr;
		struct neighborDeviceListTLV **x1905_neighbors;

		INT8U power_off_nr;
		struct powerOffInterfaceTLV **power_off;

		INT8U l2_neighbors_nr;
		struct l2NeighborDeviceTLV **l2_neighbors;

		struct genericPhyDeviceInformationTypeTLV *generic_phy;

		struct x1905ProfileVersionTLV *profile;

		struct deviceIdentificationTypeTLV *identification;

		struct controlUrlTypeTLV *control_url;

		struct ipv4TypeTLV *ipv4;

		struct ipv6TypeTLV *ipv6;

		INT8U metrics_with_neighbors_nr;
		struct _metricsWithNeighbor {
			INT8U neighbor_al_mac_address[6];

			INT32U tx_metrics_timestamp;
			struct transmitterLinkMetricTLV *tx_metrics;

			INT32U rx_metrics_timestamp;
			struct receiverLinkMetricTLV *rx_metrics;

		} *metrics_with_neighbors;

		INT8U extensions_nr;
		struct vendorSpecificTLV **extensions;

	} *network_devices;
	// This list will always contain at least ONE entry,
	// containing the info of the *local* device.
} sgDataModel;

// Given a 'mac_address', return a pointer to the "struct _localInterface" that
// represents the local interface with that address.
// Returns NONE if such a local interface could not be found.
//
static struct _localInterface *_macAddressToLocalInterfaceStruct(INT8U *mac_address)
{
	INT8U i;

	if (mac_address != NULL) {
		for (i = 0; i < sgDataModel.local_interfaces_nr; i++) {
			if (memcmp(sgDataModel.local_interfaces[i].mac_address, mac_address, 6) == 0) {
				return &sgDataModel.local_interfaces[i];
			}
		}
	}
	// Not found!
	//
	return NULL;
}

// Given a 'name', return a pointer to the "struct _localInterface" that
// represents the local interface with that interface name.
// Returns NONE if such a local interface could not be found.
//
static struct _localInterface *_nameToLocalInterfaceStruct(char *name)
{
	INT8U i;

	if (name != NULL) {
		for (i = 0; i < sgDataModel.local_interfaces_nr; i++) {
			if (memcmp(sgDataModel.local_interfaces[i].name, name, strlen(sgDataModel.local_interfaces[i].name) + 1) == 0) {
				return &sgDataModel.local_interfaces[i];
			}
		}
	}
	// Not found!
	//
	return NULL;
}

// Given an 'al_mac_address', return a pointer to the "struct _neighbor" that
// represents a 1905 neighbor with that 'al_mac_address' visible from the
// provided 'local_interface_name'.
// Returns NONE if such a neighbor could not be found.
//
static struct _neighbor *_alMacAddressToNeighborStruct(char *local_interface_name, INT8U *al_mac_address)
{
	INT8U i;
	struct _localInterface *x;

	x = _nameToLocalInterfaceStruct(local_interface_name);
	// Non existent interface
	if (x == NULL)
		return NULL;

	if (al_mac_address != NULL) {
		for (i = 0; i < x->neighbors_nr; i++) {
			if (memcmp(x->neighbors[i].al_mac_address, al_mac_address, 6) == 0) {
				return &x->neighbors[i];
			}
		}
	}
	// Not found!
	return NULL;
}

// Given a 'mac_address', return a pointer to the "struct _remoteInterface" that
// represents an interface in the provided 1905 neighbor (identified by its
// 'neighbor_al_mac_address' and visible in 'local_interface) that contains that
// address.
// Returns NONE if such a remote interface could not be found.
//
static struct _remoteInterface *_macAddressToRemoteInterfaceStruct(char *local_interface_name, INT8U *neighbor_al_mac_address, INT8U *mac_address)
{
	INT8U i;
	struct _neighbor *x;

	x = _alMacAddressToNeighborStruct(local_interface_name, neighbor_al_mac_address);
	// Non existent neighbor
	if (x == NULL)
		return NULL;

	if (mac_address != NULL) {
		for (i = 0; i < x->remote_interfaces_nr; i++) {
			if (memcmp(x->remote_interfaces[i].mac_address, mac_address, 6) == 0) {
				return &x->remote_interfaces[i];
			}
		}
	}
	// Not found!
	return NULL;
}

// When a new 1905 neighbor is discovered on a local interface, this function
// must be called to update the database.
// Returns '0' if there was a problem (out of memory, etc...), '2' if the
// neighbor had already been inserted, '1' if the new neighbor was succesfully
// inserted.
//
static INT8U _insertNeighbor(char *local_interface_name, INT8U *al_mac_address)
{
	struct _localInterface *x;

	// First, make sure the interface exists
	x = _nameToLocalInterfaceStruct(local_interface_name);
	if (x == NULL) {
		return 0;
	}
	// Next, make sure this neighbor does not already exist
	//
	if (_alMacAddressToNeighborStruct(local_interface_name, al_mac_address) != NULL) {
		// The neighbor exists! We don't need to do anything special.
		//
		return 2;
	}

	if (x->neighbors_nr == 0)
		CALLOC(x->neighbors, 1);
	else
		REALLOC(x->neighbors, (x->neighbors_nr + 1));

	memcpy(x->neighbors[x->neighbors_nr].al_mac_address, al_mac_address, 6);
	x->neighbors[x->neighbors_nr].remote_interfaces_nr = 0;
	x->neighbors[x->neighbors_nr].remote_interfaces = NULL;

	x->neighbors_nr++;

	return 1;
}

// When a new interface of a 1905 neighbor is discovered, this function must be
// called to update the database.
// Returns '0' if there was a problem (out of memory, etc...), '2' if the
// neighbor interface had already been inserted, '1' if the new neighbor
// interface was successfully inserted.
//
static INT8U _insertNeighborInterface(char *local_interface_name, INT8U *neighbor_al_mac_address, INT8U *mac_address)
{
	struct _neighbor *x;

	// First, make sure the interface and neighbor exist
	x = _alMacAddressToNeighborStruct(local_interface_name, neighbor_al_mac_address);
	if (x == NULL) {
		return 0;
	}
	// Next, make sure this neighbor interface does not already exist
	//
	if (_macAddressToRemoteInterfaceStruct(local_interface_name, neighbor_al_mac_address, mac_address) != NULL) {
		// The neighbor exists! We don't need to do anything special.
		//
		return 2;
	}

	if (x->remote_interfaces_nr == 0)
		CALLOC(x->remote_interfaces, 1);
	else
		REALLOC(x->remote_interfaces, (x->remote_interfaces_nr + 1));

	memcpy(x->remote_interfaces[x->remote_interfaces_nr].mac_address, mac_address, 6);
	x->remote_interfaces[x->remote_interfaces_nr].last_topology_discovery_ts = 0;
	x->remote_interfaces[x->remote_interfaces_nr].last_bridge_discovery_ts = 0;

	x->remote_interfaces_nr++;

	return 1;
}

////////////////////////////////////////////////////////////////////////////////
// API functions (only available to the 1905 core itself, ie. files inside the
// 'libieee1905' folder)
////////////////////////////////////////////////////////////////////////////////

void DMinit(void)
{
	sgDataModel.network_registrar_nr = 0;
	sgDataModel.network_registrar = NULL;

	sgDataModel.al_mac_address[0] = 0x00;
	sgDataModel.al_mac_address[1] = 0x00;
	sgDataModel.al_mac_address[2] = 0x00;
	sgDataModel.al_mac_address[3] = 0x00;
	sgDataModel.al_mac_address[4] = 0x00;
	sgDataModel.al_mac_address[5] = 0x00;

	sgDataModel.local_interfaces_nr = 0;
	sgDataModel.local_interfaces = NULL;

	// Regarding the "network_devices" list, we will init it with one element,
	// representing the local node
	sgDataModel.network_devices_nr = 1;
	CALLOC(sgDataModel.network_devices, 1);

	sgDataModel.network_devices[0].update_timestamp = get_time();
	sgDataModel.network_devices[0].info = NULL;
	sgDataModel.network_devices[0].bridges_nr = 0;
	sgDataModel.network_devices[0].bridges = NULL;
	sgDataModel.network_devices[0].non1905_neighbors_nr = 0;
	sgDataModel.network_devices[0].non1905_neighbors = NULL;
	sgDataModel.network_devices[0].x1905_neighbors_nr = 0;
	sgDataModel.network_devices[0].x1905_neighbors = NULL;
	sgDataModel.network_devices[0].power_off_nr = 0;
	sgDataModel.network_devices[0].power_off = NULL;
	sgDataModel.network_devices[0].l2_neighbors_nr = 0;
	sgDataModel.network_devices[0].l2_neighbors = NULL;
	sgDataModel.network_devices[0].generic_phy = NULL;
	sgDataModel.network_devices[0].profile = NULL;
	sgDataModel.network_devices[0].identification = NULL;
	sgDataModel.network_devices[0].control_url = NULL;
	sgDataModel.network_devices[0].ipv4 = NULL;
	sgDataModel.network_devices[0].ipv6 = NULL;
	sgDataModel.network_devices[0].metrics_with_neighbors_nr = 0;
	sgDataModel.network_devices[0].metrics_with_neighbors = NULL;
	sgDataModel.network_devices[0].extensions = NULL;
	sgDataModel.network_devices[0].extensions_nr = 0;

	return;
}

void DM_cleanup(void)
{
	INT8U i, j;

	if (sgDataModel.network_registrar_nr)
		FREE(sgDataModel.network_registrar);

	if (sgDataModel.local_interfaces_nr) {
		if (sgDataModel.local_interfaces->neighbors_nr) {
			if (sgDataModel.local_interfaces->neighbors->remote_interfaces_nr) {
				FREE(sgDataModel.local_interfaces->neighbors->remote_interfaces);
			}
			FREE(sgDataModel.local_interfaces->neighbors);
		}
		FREE(sgDataModel.local_interfaces);
	}

	for (i = 1; i < sgDataModel.network_devices_nr; i++) {
		struct _networkDevice *y;

		y = &sgDataModel.network_devices[i];

		if (y->info != NULL) {
			free_1905_TLV_structure((INT8U *) y->info, true);
			y->info = NULL;
		}

		for (j = 0; j < y->bridges_nr; j++) {
			free_1905_TLV_structure((INT8U *) y->bridges[j], true);
		}
		if (0 != y->bridges_nr && NULL != y->bridges) {
			FREE(y->bridges);
			y->bridges_nr = 0;
			y->bridges = NULL;
		}

		for (j = 0; j < y->non1905_neighbors_nr; j++) {
			free_1905_TLV_structure((INT8U *) y->non1905_neighbors[j], true);
		}
		if (0 != y->non1905_neighbors_nr && NULL != y->non1905_neighbors) {
			FREE(y->non1905_neighbors);
			y->non1905_neighbors_nr = 0;
			y->non1905_neighbors = NULL;
		}

		for (j = 0; j < y->x1905_neighbors_nr; j++) {
			free_1905_TLV_structure((INT8U *) y->x1905_neighbors[j], true);
		}
		if (0 != y->x1905_neighbors_nr && NULL != y->x1905_neighbors) {
			FREE(y->x1905_neighbors);
			y->x1905_neighbors_nr = 0;
			y->x1905_neighbors = NULL;
		}

		for (j = 0; j < y->power_off_nr; j++) {
			free_1905_TLV_structure((INT8U *) y->power_off[j], true);
		}
		if (0 != y->power_off_nr && NULL != y->power_off) {
			FREE(y->power_off);
			y->power_off_nr = 0;
			y->power_off = NULL;
		}

		for (j = 0; j < y->l2_neighbors_nr; j++) {
			free_1905_TLV_structure((INT8U *) y->l2_neighbors[j], true);
		}
		if (0 != y->l2_neighbors_nr && NULL != y->l2_neighbors) {
			FREE(y->l2_neighbors);
			y->l2_neighbors_nr = 0;
			y->l2_neighbors = NULL;
		}

		if (y->generic_phy != NULL) {
			free_1905_TLV_structure((INT8U *) y->generic_phy, true);
			y->generic_phy = NULL;
		}

		if (y->profile != NULL) {
			free_1905_TLV_structure((INT8U *) y->profile, true);
			y->profile = NULL;
		}

		if (y->identification != NULL) {
			free_1905_TLV_structure((INT8U *) y->identification, true);
			y->identification = NULL;
		}

		if (y->control_url != NULL) {
			free_1905_TLV_structure((INT8U *) y->control_url, true);
			y->control_url = NULL;
		}

		if (y->ipv4 != NULL) {
			free_1905_TLV_structure((INT8U *) y->ipv4, true);
			y->ipv4 = NULL;
		}

		if (y->ipv6 != NULL) {
			free_1905_TLV_structure((INT8U *) y->ipv6, true);
			y->ipv6 = NULL;
		}

		for (j = 0; j < y->metrics_with_neighbors_nr; j++) {
			free_1905_TLV_structure((INT8U *) y->metrics_with_neighbors[j].tx_metrics, true);
			free_1905_TLV_structure((INT8U *) y->metrics_with_neighbors[j].rx_metrics, true);
		}
		if (0 != y->metrics_with_neighbors_nr && NULL != y->metrics_with_neighbors) {
			FREE(y->metrics_with_neighbors);
			y->metrics_with_neighbors = NULL;
		}
		if (sgDataModel.network_devices[j].metrics_with_neighbors_nr == 0)
			FREE(sgDataModel.network_devices[j].metrics_with_neighbors);

	}
	FREE(sgDataModel.network_devices);
}

void DMalMacSet(INT8U *al_mac_address)
{
	memcpy(sgDataModel.al_mac_address, al_mac_address, 6);

	return;
}

INT8U *DMalMacGet(void)
{
	return sgDataModel.al_mac_address;
}

bool DMalMacConfigured(void)
{
	if (hwaddr_is_zero(sgDataModel.al_mac_address))
		return false;

	return true;
}

bool DMregistrarInsert(INT8U *reg_mac, INT16U intf_type)
{
	uint8_t registrar_type;
	INT8U reg_nr = sgDataModel.network_registrar_nr;

	switch(intf_type) {
	case MEDIA_TYPE_IEEE_802_11B_2_4_GHZ:
	case MEDIA_TYPE_IEEE_802_11G_2_4_GHZ:
	case MEDIA_TYPE_IEEE_802_11N_2_4_GHZ:
		registrar_type = IEEE80211_FREQUENCY_BAND_2_4_GHZ;
		break;
	case MEDIA_TYPE_IEEE_802_11A_5_GHZ:
	case MEDIA_TYPE_IEEE_802_11N_5_GHZ:
	case MEDIA_TYPE_IEEE_802_11AC_5_GHZ:
		registrar_type = IEEE80211_FREQUENCY_BAND_5_GHZ;
		break;
	case MEDIA_TYPE_IEEE_802_11AD_60_GHZ:
	case MEDIA_TYPE_IEEE_802_11AF_GHZ:
		registrar_type = IEEE80211_FREQUENCY_BAND_60_GHZ;
		break;
	default:
		return false;
	}

	if (reg_nr == 0)
		CALLOC(sgDataModel.network_registrar, 1);
	else
		REALLOC(sgDataModel.network_registrar, (reg_nr+1));

	sgDataModel.network_registrar[reg_nr].registrar_type = registrar_type;
	memcpy(&sgDataModel.network_registrar[reg_nr].registrar_mac_address, reg_mac, 6);
	sgDataModel.network_registrar_nr++;

	return true;
}

INT8U *DMregistrarMacGet(INT8U reg_type)
{
	INT8U i = 0;

	for (i = 0; i < sgDataModel.network_registrar_nr; i++) {
		if (reg_type == sgDataModel.network_registrar[i].registrar_type)
			return sgDataModel.network_registrar[i].registrar_mac_address;
	}
	return NULL;
}

bool DMregistrarMacGetAll(INT8U reg_type, char *reg_buff)
{
	INT8U i = 0;
	size_t len;

	for (i = 0; i < sgDataModel.network_registrar_nr; i++) {
		if (reg_type == sgDataModel.network_registrar[i].registrar_type) {
			char mac[20] = {0};

			hwaddr_ntoa(sgDataModel.network_registrar[i].registrar_mac_address, mac);
			// Below is not type buffer needs to be a csv mac address
			strcat(mac, ",");
			strncat(reg_buff, mac, 1024);
		}
	}
	// Remove last comma
	len = strlen(reg_buff);
	if (len) {
		reg_buff[len-1] = '\0';
		return true;
	}

	return false;
}

INT8U DMinsertInterface(char *name, INT8U *mac_address)
{
	struct _localInterface *x;

	// First, make sure this interface does not already exist
	x = _nameToLocalInterfaceStruct(name);
	if (x != NULL) {
		// The interface exists!
		//
		// Even if it already exists, if the provided 'mac_address' and the
		// already existing entry match, do not return an error.
		//
		if (memcmp(x->mac_address, mac_address, 6) == 0) {
			// Ok
			//
			return 1;
		} else {
			// Interface exists and its MAC address is different from the
			// provided one. Maybe the caller should first "remove" this
			// interface and then try to add the new one.
			//
			return 0;
		}
	}

	if (sgDataModel.local_interfaces_nr == 0)
		CALLOC(sgDataModel.local_interfaces, 1);
	else
		REALLOC(sgDataModel.local_interfaces, (sgDataModel.local_interfaces_nr + 1));

	strcpy(sgDataModel.local_interfaces[sgDataModel.local_interfaces_nr].name, name);
	memcpy(sgDataModel.local_interfaces[sgDataModel.local_interfaces_nr].mac_address, mac_address, 6);
	sgDataModel.local_interfaces[sgDataModel.local_interfaces_nr].neighbors = NULL;
	sgDataModel.local_interfaces[sgDataModel.local_interfaces_nr].neighbors_nr = 0;

	sgDataModel.local_interfaces_nr++;

	return 1;
}

char *DMmacToInterfaceName(INT8U *mac_address)
{
	struct _localInterface *x;

	x = _macAddressToLocalInterfaceStruct(mac_address);

	if (x != NULL) {
		return x->name;
	} else {
		// Not found!
		//
		return NULL;
	}
}

INT8U *DMinterfaceNameToMac(char *interface_name)
{
	INT8U i;

	if (interface_name != NULL) {
		for (i = 0; i < sgDataModel.local_interfaces_nr; i++) {
			if (memcmp(sgDataModel.local_interfaces[i].name, interface_name,
				   strlen(sgDataModel.local_interfaces[i].name) + 1) == 0)
				return sgDataModel.local_interfaces[i].mac_address;
		}
	}
	// Not found!
	return NULL;
}

INT8U(*DMgetListOfInterfaceNeighbors(char *local_interface_name, INT8U *al_mac_addresses_nr))[6]
{
	INT8U i;
	INT8U(*ret)[6];

	struct _localInterface *x;

	// Non existent interface
	x = _nameToLocalInterfaceStruct(local_interface_name);
	if (x == NULL) {
		*al_mac_addresses_nr = 0;
		return NULL;
	}

	if (x->neighbors_nr == 0) {
		*al_mac_addresses_nr = 0;
		return NULL;
	}

	ret = (INT8U(*)[6]) malloc(sizeof(INT8U[6]) * x->neighbors_nr);
	if (ret == NULL) {
		ERROR("Out of memory!");
		exit(ENOMEM);
	}

	for (i = 0; i < x->neighbors_nr; i++) {
		ret[i][0] = x->neighbors[i].al_mac_address[0];
		ret[i][1] = x->neighbors[i].al_mac_address[1];
		ret[i][2] = x->neighbors[i].al_mac_address[2];
		ret[i][3] = x->neighbors[i].al_mac_address[3];
		ret[i][4] = x->neighbors[i].al_mac_address[4];
		ret[i][5] = x->neighbors[i].al_mac_address[5];
	}

	*al_mac_addresses_nr = x->neighbors_nr;
	return ret;
}

INT8U(*DMgetListOfNeighbors(INT8U *al_mac_addresses_nr))[6]
{
	INT8U i, j, k;

	INT8U total;
	INT8U(*ret)[6];

	if (al_mac_addresses_nr == NULL) {
		return NULL;
	}

	total = 0;
	ret = NULL;

	for (i = 0; i < sgDataModel.local_interfaces_nr; i++) {
		for (j = 0; j < sgDataModel.local_interfaces[i].neighbors_nr; j++) {
			// Check for duplicates
			//
			INT8U already_present;

			already_present = 0;
			for (k = 0; k < total; k++) {
				if (memcmp(&ret[k], sgDataModel.local_interfaces[i].neighbors[j].al_mac_address, 6) == 0) {
					already_present = 1;
					break;
				}
			}

			if (already_present == 1) {
				continue;
			}
			// If we get here, this is a new neighbor and we need to add it to
			// the list
			//
			if (ret == NULL) {
				ret = (INT8U(*)[6]) malloc(sizeof(INT8U[6]));
			} else {
				ret = (INT8U(*)[6]) realloc(ret, sizeof(INT8U[6]) * (total + 1));
			}
			if (ret == NULL) {
				ERROR("Out of memory!");
				exit(ENOMEM);
			}
			memcpy(&ret[total], sgDataModel.local_interfaces[i].neighbors[j].al_mac_address, 6);

			total++;
		}
	}

	*al_mac_addresses_nr = total;

	return ret;
}

INT8U(*DMgetListOfLinksWithNeighbor(INT8U *neighbor_al_mac_address, char ***interfaces, INT8U *links_nr))[6]
{
	INT8U i, j, k;
	INT8U total;

	INT8U(*ret)[6];
	char **intfs;

	total = 0;
	ret = NULL;
	intfs = NULL;

	for (i = 0; i < sgDataModel.local_interfaces_nr; i++) {
		for (j = 0; j < sgDataModel.local_interfaces[i].neighbors_nr; j++) {
			// Filter neighbor (we are just interested in
			// 'neighbor_al_mac_address')
			//
			if (memcmp(neighbor_al_mac_address, sgDataModel.local_interfaces[i].neighbors[j].al_mac_address, 6) != 0) {
				continue;
			}

			for (k = 0; k < sgDataModel.local_interfaces[i].neighbors[j].remote_interfaces_nr; k++) {
				// This is a new link between the local AL and the remote AL.
				// Add it.
				if (ret == NULL) {
					ret = (INT8U(*)[6]) malloc(sizeof(INT8U[6]));
					CALLOC(intfs, 1);
				} else {
					ret = (INT8U(*)[6]) realloc(ret, sizeof(INT8U[6]) * (total + 1));
					REALLOC(intfs, (total+1));
				}
				/*
				if (ret) {
					ERROR("Out of memory!");
					exit(ENOMEM);
				}
				*/

				memcpy(&ret[total], sgDataModel.local_interfaces[i].neighbors[j].remote_interfaces[k].mac_address, 6);
				intfs[total] = sgDataModel.local_interfaces[i].name;

				total++;
			}
		}
	}

	*links_nr = total;
	*interfaces = intfs;

	return ret;
}

void DMfreeListOfLinksWithNeighbor(INT8U(*p)[6], char **interfaces, INT8U links_nr)
{
	if (links_nr == 0) {
		return;
	}

	if (interfaces != NULL) {
		FREE(interfaces);
	}

	if (p != NULL) {
		FREE(p);
	}

	return;
}

INT8U DMupdateDiscoveryTimeStamps(INT8U *receiving_interface_addr, INT8U *al_mac_address, INT8U *mac_address, INT8U timestamp_type, INT32U *ellapsed)
{
	char *receiving_interface_name;

	struct _remoteInterface *x;

	INT32U aux1, aux2;
	INT8U insert_result;
	INT8U ret;

	ret = 2;

	if (receiving_interface_addr == NULL) {
		ERROR("Invalid 'receiving_interface_addr'");
		return 0;
	}

	receiving_interface_name = DMmacToInterfaceName(receiving_interface_addr);
	if (receiving_interface_name == NULL) {
		ERROR("receiving_interface_addr (%02x:%02x:%02x:%02x:%02x:%02x) does not match any local interface",
		      receiving_interface_addr[0],
		      receiving_interface_addr[1],
		      receiving_interface_addr[2],
		      receiving_interface_addr[3],
		      receiving_interface_addr[4],
		      receiving_interface_addr[5]);
		return 0;
	}

	insert_result = _insertNeighbor(receiving_interface_name, al_mac_address);
	if (insert_result == 0 ||
	    _insertNeighborInterface(receiving_interface_name, al_mac_address, mac_address) == 0
	    ) {
		ERROR("Could not create new entries in the database");
		return 0;
	}

	// This is the first time we know of this neighbor (ie. the neighbor
	// has been inserted in the data model for the first time)
	if (insert_result == 1)
		ret = 1;

	x = _macAddressToRemoteInterfaceStruct(receiving_interface_name, al_mac_address, mac_address);

	DEBUG("New discovery timestamp udpate:");
	DEBUG("  - local_interface      : %s", receiving_interface_name);
	DEBUG("  - 1905 neighbor AL MAC : %02x:%02x:%02x:%02x:%02x:%02x:",
	      al_mac_address[0], al_mac_address[1],
	      al_mac_address[2], al_mac_address[3],
	      al_mac_address[4], al_mac_address[5]);
	DEBUG("  - remote interface MAC : %02x:%02x:%02x:%02x:%02x:%02x:",
	      mac_address[0], mac_address[1],
	      mac_address[2], mac_address[3],
	      mac_address[4], mac_address[5]);

	aux1 = x->last_topology_discovery_ts;
	aux2 = x->last_bridge_discovery_ts;

	switch (timestamp_type) {
	case TIMESTAMP_TOPOLOGY_DISCOVERY:
		{
			INT32U aux;

			aux = get_time();

			if (ellapsed != NULL) {
				if (ret == 2) {
					*ellapsed = aux - x->last_topology_discovery_ts;
				} else {
					*ellapsed = 0;
				}
			}

			x->last_topology_discovery_ts = aux;
			break;
		}
	case TIMESTAMP_BRIDGE_DISCOVERY:
		{
			INT32U aux;

			aux = get_time();

			if (ellapsed != NULL) {
				if (ret == 2) {
					*ellapsed = aux - x->last_bridge_discovery_ts;
				} else {
					*ellapsed = 0;
				}
			}
			x->last_bridge_discovery_ts = aux;
			break;
		}
	default:
		{
			ERROR("Unknown 'timestamp_type' (%d)", timestamp_type);

			return 0;
		}
	}

	DEBUG("  - topology disc TS     : %d --> %d", aux1, x->last_topology_discovery_ts);
	DEBUG("  - bridge   disc TS     : %d --> %d", aux2, x->last_bridge_discovery_ts);

	return ret;
}

INT8U DMisLinkBridged(char *local_interface_name, INT8U *neighbor_al_mac_address, INT8U *neighbor_mac_address)
{
	struct _remoteInterface *x;
	INT32U aux;

	x = _macAddressToRemoteInterfaceStruct(local_interface_name,
					       neighbor_al_mac_address,
					       neighbor_mac_address);
	if (x == NULL)
		return 2;

	if (x->last_topology_discovery_ts > x->last_bridge_discovery_ts) {
		aux = x->last_topology_discovery_ts - x->last_bridge_discovery_ts;
	} else {
		aux = x->last_bridge_discovery_ts - x->last_topology_discovery_ts;
	}

	// Links is *not* bridged
	if (aux < DISCOVERY_THRESHOLD_MS)
		return 0;
	else
		return 1;
}

INT8U DMisNeighborBridged(char *local_interface_name, INT8U *neighbor_al_mac_address)
{
	struct _neighbor *x;
	INT8U i;

	x = _alMacAddressToNeighborStruct(local_interface_name, neighbor_al_mac_address);
	// Non existent neighbor
	if (x == NULL)
		return 2;

	for (i = 0; i < x->remote_interfaces_nr; i++) {
		if (DMisLinkBridged(local_interface_name, neighbor_al_mac_address, x->remote_interfaces[i].mac_address) == 1) {
			// If at least one link is bridged, then this neighbor is
			// considered to be bridged.
			//
			return 1;
		}
	}

	// All links are not bridged, thus the neighbor is considered to be not
	// bridged also.
	//
	return 0;
}

INT8U DMisInterfaceBridged(char *local_interface_name)
{
	struct _localInterface *x;

	INT8U i;

	x = _nameToLocalInterfaceStruct(local_interface_name);
	if (x == NULL) {
		ERROR("Invalid local interface name");
		return 2;
	}

	for (i = 0; i < x->neighbors_nr; i++) {
		if (DMisNeighborBridged(local_interface_name, x->neighbors[i].al_mac_address) == 1) {
			// If at least one neighbor is bridged, then this interface is
			// considered to be bridged.
			//
			return 1;
		}
	}

	// All neighbors are not bridged, thus the interface is considered to be not
	// bridged also.
	//
	return 0;
}

INT8U *DMmacToAlMac(INT8U *mac_address)
{
	INT8U i, j, k;

	INT8U *al_mac;
	CALLOC(al_mac, 6);

	if (memcmp(sgDataModel.al_mac_address, mac_address, 6) == 0) {
		memcpy(al_mac, sgDataModel.al_mac_address, 6);
		return al_mac;
	}

	for (i = 0; i < sgDataModel.local_interfaces_nr; i++) {
		if (memcmp(sgDataModel.local_interfaces[i].mac_address, mac_address, 6) == 0) {
			memcpy(al_mac, sgDataModel.al_mac_address, 6);
			return al_mac;
		}

		for (j = 0; j < sgDataModel.local_interfaces[i].neighbors_nr; j++) {
			if (memcmp(sgDataModel.local_interfaces[i].neighbors[j].al_mac_address, mac_address, 6) == 0) {
				memcpy(al_mac, sgDataModel.local_interfaces[i].neighbors[j].al_mac_address, 6);
				return al_mac;
			}

			for (k = 0; k < sgDataModel.local_interfaces[i].neighbors[j].remote_interfaces_nr; k++) {
				if (memcmp(sgDataModel.local_interfaces[i].neighbors[j].remote_interfaces[k].mac_address, mac_address, 6) == 0) {
					memcpy(al_mac, sgDataModel.local_interfaces[i].neighbors[j].al_mac_address, 6);
					return al_mac;
				}
			}
		}
	}

	FREE(al_mac);
	return NULL;
}

INT8U DMupdateNetworkDeviceInfo(INT8U *al_mac_address,
				INT8U in_update, struct deviceInformationTypeTLV *info,
				INT8U br_update, struct deviceBridgingCapabilityTLV **bridges, INT8U bridges_nr,
				INT8U no_update, struct non1905NeighborDeviceListTLV **non1905_neighbors, INT8U non1905_neighbors_nr,
				INT8U x1_update, struct neighborDeviceListTLV **x1905_neighbors, INT8U x1905_neighbors_nr,
				INT8U po_update, struct powerOffInterfaceTLV **power_off, INT8U power_off_nr,
				INT8U l2_update, struct l2NeighborDeviceTLV **l2_neighbors, INT8U l2_neighbors_nr,
				INT8U ge_update, struct genericPhyDeviceInformationTypeTLV *generic_phy,
				INT8U pr_update, struct x1905ProfileVersionTLV *profile,
				INT8U id_update, struct deviceIdentificationTypeTLV *identification,
				INT8U co_update, struct controlUrlTypeTLV *control_url,
				INT8U v4_update, struct ipv4TypeTLV *ipv4, INT8U v6_update, struct ipv6TypeTLV *ipv6)
{
	INT8U i, j;

	if ((al_mac_address == NULL) ||
	    (br_update == 1 && (bridges_nr > 0 && NULL == bridges)) ||
	    (no_update == 1 && (non1905_neighbors_nr > 0 && NULL == non1905_neighbors)) ||
	    (x1_update == 1 && (x1905_neighbors_nr > 0 && NULL == x1905_neighbors)) ||
	    (po_update == 1 && (power_off_nr > 0 && NULL == power_off)) || (l2_update == 1 && (l2_neighbors_nr > 0 && NULL == l2_neighbors))
	    ) {
		return 0;
	}
	// First, search for an existing entry with the same AL MAC address
	// Remember that the first entry holds a reference to the *local* node.
	//
	if (memcmp(DMalMacGet(), al_mac_address, 6) == 0) {
		i = 0;
	} else {
		for (i = 1; i < sgDataModel.network_devices_nr; i++) {
			if (sgDataModel.network_devices[i].info != NULL) {
				if (memcmp(sgDataModel.network_devices[i].info->al_mac_address, al_mac_address, 6) == 0) {
					break;
				}
			}
		}
	}

	if (i == sgDataModel.network_devices_nr) {
		// A matching entry was *not* found. Create a new one, but only if this
		// new information contains the "info" TLV (otherwise don't do anything
		// and wait for the "info" TLV to be received in the future)
		//
		if (1 == in_update && NULL != info) {
			if (sgDataModel.network_devices_nr == 0)
				CALLOC(sgDataModel.network_devices, 1);
			else
				REALLOC(sgDataModel.network_devices, (sgDataModel.network_devices_nr + 1));

			sgDataModel.network_devices[sgDataModel.network_devices_nr].update_timestamp = get_time();
			sgDataModel.network_devices[sgDataModel.network_devices_nr].info = 1 == in_update ? info : NULL;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].bridges_nr = 1 == br_update ? bridges_nr : 0;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].bridges = 1 == br_update ? bridges : NULL;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].non1905_neighbors_nr = 1 == no_update ? non1905_neighbors_nr : 0;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].non1905_neighbors = 1 == no_update ? non1905_neighbors : NULL;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].x1905_neighbors_nr = 1 == x1_update ? x1905_neighbors_nr : 0;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].x1905_neighbors = 1 == x1_update ? x1905_neighbors : NULL;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].power_off_nr = 1 == po_update ? power_off_nr : 0;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].power_off = 1 == po_update ? power_off : NULL;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].l2_neighbors_nr = 1 == l2_update ? l2_neighbors_nr : 0;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].l2_neighbors = 1 == l2_update ? l2_neighbors : NULL;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].generic_phy = 1 == ge_update ? generic_phy : NULL;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].profile = 1 == pr_update ? profile : NULL;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].identification = 1 == id_update ? identification : NULL;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].control_url = 1 == co_update ? control_url : NULL;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].ipv4 = 1 == v4_update ? ipv4 : NULL;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].ipv6 = 1 == v6_update ? ipv6 : NULL;

			sgDataModel.network_devices[sgDataModel.network_devices_nr].metrics_with_neighbors_nr = 0;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].metrics_with_neighbors = NULL;

			sgDataModel.network_devices[sgDataModel.network_devices_nr].extensions = NULL;
			sgDataModel.network_devices[sgDataModel.network_devices_nr].extensions_nr = 0;

			sgDataModel.network_devices_nr++;
		}
	} else {
		// A matching entry was found. Update it. But first, free the old TLV
		// structures (but only if a new value was provided!... otherwise retain
		// the old item)
		//
		sgDataModel.network_devices[i].update_timestamp = get_time();

		if (info != NULL) {
			if (sgDataModel.network_devices[i].info != NULL) {
				free_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].info, true);
			}
			sgDataModel.network_devices[i].info = info;
		}

		if (br_update == 1) {
			for (j = 0; j < sgDataModel.network_devices[i].bridges_nr; j++) {
				free_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].bridges[j], true);
			}
			if (sgDataModel.network_devices[i].bridges_nr > 0 && NULL != sgDataModel.network_devices[i].bridges) {
				FREE(sgDataModel.network_devices[i].bridges);
			}
			sgDataModel.network_devices[i].bridges_nr = bridges_nr;
			sgDataModel.network_devices[i].bridges = bridges;
		}

		if (no_update == 1) {
			for (j = 0; j < sgDataModel.network_devices[i].non1905_neighbors_nr; j++) {
				free_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].non1905_neighbors[j], true);
			}
			if (sgDataModel.network_devices[i].non1905_neighbors_nr > 0 && NULL != sgDataModel.network_devices[i].non1905_neighbors) {
				FREE(sgDataModel.network_devices[i].non1905_neighbors);
			}
			sgDataModel.network_devices[i].non1905_neighbors_nr = non1905_neighbors_nr;
			sgDataModel.network_devices[i].non1905_neighbors = non1905_neighbors;
		}

		if (x1_update == 1) {
			for (j = 0; j < sgDataModel.network_devices[i].x1905_neighbors_nr; j++) {
				free_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].x1905_neighbors[j], true);
			}
			if (sgDataModel.network_devices[i].x1905_neighbors_nr > 0 && NULL != sgDataModel.network_devices[i].x1905_neighbors) {
				FREE(sgDataModel.network_devices[i].x1905_neighbors);
			}
			sgDataModel.network_devices[i].x1905_neighbors_nr = x1905_neighbors_nr;
			sgDataModel.network_devices[i].x1905_neighbors = x1905_neighbors;
		}

		if (po_update == 1) {
			for (j = 0; j < sgDataModel.network_devices[i].power_off_nr; j++) {
				free_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].power_off[j], true);
			}
			if (sgDataModel.network_devices[i].power_off_nr > 0 && NULL != sgDataModel.network_devices[i].power_off) {
				FREE(sgDataModel.network_devices[i].power_off);
			}
			sgDataModel.network_devices[i].power_off_nr = power_off_nr;
			sgDataModel.network_devices[i].power_off = power_off;
		}

		if (l2_update == 1) {
			for (j = 0; j < sgDataModel.network_devices[i].l2_neighbors_nr; j++) {
				free_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].l2_neighbors[j], true);
			}
			if (sgDataModel.network_devices[i].l2_neighbors_nr > 0 && NULL != sgDataModel.network_devices[i].l2_neighbors) {
				FREE(sgDataModel.network_devices[i].l2_neighbors);
			}
			sgDataModel.network_devices[i].l2_neighbors_nr = l2_neighbors_nr;
			sgDataModel.network_devices[i].l2_neighbors = l2_neighbors;
		}

		if (ge_update == 1) {
			free_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].generic_phy, true);
			sgDataModel.network_devices[i].generic_phy = generic_phy;
		}

		if (pr_update == 1) {
			free_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].profile, true);
			sgDataModel.network_devices[i].profile = profile;
		}

		if (id_update == 1) {
			free_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].identification, true);
			sgDataModel.network_devices[i].identification = identification;
		}

		if (co_update == 1) {
			free_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].control_url, true);
			sgDataModel.network_devices[i].control_url = control_url;
		}

		if (v4_update == 1) {
			free_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].ipv4, true);
			sgDataModel.network_devices[i].ipv4 = ipv4;
		}

		if (v6_update == 1) {
			free_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].ipv6, true);
			sgDataModel.network_devices[i].ipv6 = ipv6;
		}

	}

	return 1;
}

INT8U DMnetworkDeviceInfoNeedsUpdate(INT8U *al_mac_address)
{
	INT8U i;

	// First, search for an existing entry with the same AL MAC address
	//
	for (i = 0; i < sgDataModel.network_devices_nr; i++) {
		if (sgDataModel.network_devices[i].info != NULL) {
			if (memcmp(sgDataModel.network_devices[i].info->al_mac_address, al_mac_address, 6) == 0) {
				break;
			}
		}
	}

	if (i == sgDataModel.network_devices_nr) {
		// A matching entry was *not* found. Thus a refresh of the information
		// is needed.
		//
		return 1;
	} else {
		// A matching entry was found. Check its timestamp.
		//
		if (get_time() - sgDataModel.network_devices[i].update_timestamp > MAX_AGE * 1000) {
			return 1;
		} else {
			return 0;
		}
	}
}

INT8U DMupdateNetworkDeviceMetrics(INT8U *metrics)
{
	INT8U *FROM_al_mac_address;	// Metrics are reported FROM this AL entity...
	INT8U *TO_al_mac_address;	// ... TO this other one.

	INT8U i, j;

	if (metrics == NULL) {
		ERROR("Invalid 'metrics' argument");
		return 0;
	}
	// Obtain the AL MAC of the devices involved in the metrics report (ie.
	// the "from" and the "to" AL MAC addresses).
	// This information is contained inside the 'metrics' structure itself.
	//
	if (*metrics == TLV_TYPE_TRANSMITTER_LINK_METRIC) {
		struct transmitterLinkMetricTLV *p;

		p = (struct transmitterLinkMetricTLV *)metrics;

		FROM_al_mac_address = p->local_al_address;
		TO_al_mac_address = p->neighbor_al_address;

	} else if (*metrics == TLV_TYPE_RECEIVER_LINK_METRIC) {
		struct receiverLinkMetricTLV *p;

		p = (struct receiverLinkMetricTLV *)metrics;

		FROM_al_mac_address = p->local_al_address;
		TO_al_mac_address = p->neighbor_al_address;
	} else {
		DEBUG("Invalid 'metrics' argument. Type = %d", *metrics);
		return 0;
	}

	// Next, search for an existing entry with the same AL MAC address
	//
	for (i = 0; i < sgDataModel.network_devices_nr; i++) {
		if (sgDataModel.network_devices[i].info == NULL) {
			// We haven't received general info about this device yet.
			// This can happen, for example, when only metrics have been
			// received so far.
			//
			continue;
		}
		if (memcmp(sgDataModel.network_devices[i].info->al_mac_address, FROM_al_mac_address, 6) == 0) {
			break;
		}
	}

	if (i == sgDataModel.network_devices_nr) {
		// A matching entry was *not* found.
		//
		// At this point, even if we could just create a new device entry with
		// everything set to zero (except for the metrics that we have just
		// received), it is probably wiser to simply discard the data.
		//
		// In other words, we should not accept metrics data until the "general
		// info" for this node has been processed.
		//
		DEBUG("Metrics received from an unknown 1905 node (%02x:%02x:%02x:%02x:%02x:%02x). Ignoring data...", FROM_al_mac_address[0],
		      FROM_al_mac_address[1], FROM_al_mac_address[2], FROM_al_mac_address[3], FROM_al_mac_address[4], FROM_al_mac_address[5]);
		return 0;
	}
	// Now that we have found the corresponding neighbor entry (or created a
	// new one) search for a sub-entry that matches the AL MAC of the node the
	// metrics are being reported against.
	//
	for (j = 0; j < sgDataModel.network_devices[i].metrics_with_neighbors_nr; j++) {
		if (memcmp(sgDataModel.network_devices[i].metrics_with_neighbors[j].neighbor_al_mac_address, TO_al_mac_address, 6) == 0) {
			break;
		}
	}

	if (j == sgDataModel.network_devices[i].metrics_with_neighbors_nr) {
		// A matching entry was *not* found. Create a new one
		//
		if (sgDataModel.network_devices[i].metrics_with_neighbors_nr == 0)
			CALLOC(sgDataModel.network_devices[i].metrics_with_neighbors, 1);
		else
			REALLOC(sgDataModel.network_devices[i].metrics_with_neighbors,
				(sgDataModel.network_devices[i].metrics_with_neighbors_nr + 1));

		memcpy(sgDataModel.network_devices[i].metrics_with_neighbors[sgDataModel.network_devices[i].metrics_with_neighbors_nr].neighbor_al_mac_address,
		       TO_al_mac_address, 6);

		if (*metrics == TLV_TYPE_TRANSMITTER_LINK_METRIC) {
			sgDataModel.network_devices[i].metrics_with_neighbors[sgDataModel.network_devices[i].metrics_with_neighbors_nr].tx_metrics_timestamp =
			    get_time();
			sgDataModel.network_devices[i].metrics_with_neighbors[sgDataModel.network_devices[i].metrics_with_neighbors_nr].tx_metrics =
			    (struct transmitterLinkMetricTLV *)metrics;

			sgDataModel.network_devices[i].metrics_with_neighbors[sgDataModel.network_devices[i].metrics_with_neighbors_nr].rx_metrics_timestamp =
			    0;
			sgDataModel.network_devices[i].metrics_with_neighbors[sgDataModel.network_devices[i].metrics_with_neighbors_nr].rx_metrics = NULL;
		} else {
			sgDataModel.network_devices[i].metrics_with_neighbors[sgDataModel.network_devices[i].metrics_with_neighbors_nr].tx_metrics_timestamp =
			    0;
			sgDataModel.network_devices[i].metrics_with_neighbors[sgDataModel.network_devices[i].metrics_with_neighbors_nr].tx_metrics = NULL;

			sgDataModel.network_devices[i].metrics_with_neighbors[sgDataModel.network_devices[i].metrics_with_neighbors_nr].rx_metrics_timestamp =
			    get_time();
			sgDataModel.network_devices[i].metrics_with_neighbors[sgDataModel.network_devices[i].metrics_with_neighbors_nr].rx_metrics =
			    (struct receiverLinkMetricTLV *)metrics;
		}

		sgDataModel.network_devices[i].metrics_with_neighbors_nr++;
	} else {
		// A matching entry was found. Update it. But first, free the old TLV
		// structures.
		//
		if (*metrics == TLV_TYPE_TRANSMITTER_LINK_METRIC) {
			free_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].metrics_with_neighbors[j].tx_metrics, true);

			sgDataModel.network_devices[i].metrics_with_neighbors[j].tx_metrics_timestamp = get_time();
			sgDataModel.network_devices[i].metrics_with_neighbors[j].tx_metrics = (struct transmitterLinkMetricTLV *)metrics;
		} else {
			free_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].metrics_with_neighbors[j].rx_metrics, true);

			sgDataModel.network_devices[i].metrics_with_neighbors[j].rx_metrics_timestamp = get_time();
			sgDataModel.network_devices[i].metrics_with_neighbors[j].rx_metrics = (struct receiverLinkMetricTLV *)metrics;
		}
	}

	return 1;
}

void DMdumpNeighbourSelected(struct blob_buf *bb, char *interface_name)
{
	INT8U i;
	void *array = NULL;

	struct _localInterface *local = NULL;

	if (NULL == bb || NULL == interface_name)
		return;

	local = _nameToLocalInterfaceStruct(interface_name);
	if (local == NULL)
		return;

	blobmsg_add_u32(bb, "neighbors_nr", local->neighbors_nr);
	array = blobmsg_open_array(bb, "neighbors");
	for (i = 0; i < local->neighbors_nr; i++) {
		char mac[18] = { 0x0 };
		hwaddr_ntoa(local->neighbors[i].al_mac_address, mac);
		blobmsg_add_string(bb, NULL, mac);
	}
	blobmsg_close_array(bb, array);

	return;
}

void DMFillNeighbourLinkSelected(struct blob_buf *bb, char *interface_name)
{
	uint8_t i, j, bridge = 0;
	void *table;
	char mac_str[18] = { 0x0 };
	struct _localInterface *local = NULL;
	struct linkMetrics *info_link, linkm;
	struct interfaceInfo *x;
	uint8_t nmac[6];

	memset(&linkm, 0, sizeof(struct linkMetrics));

	if (NULL == bb || NULL == interface_name)
		return;

	local = _nameToLocalInterfaceStruct(interface_name);
	if (local == NULL)
		return;

	info_link = &linkm;
	x = PLATFORM_GET_1905_INTERFACE_INFO(interface_name);
	for (i = 0; i < local->neighbors_nr; i++) {
		for (j = 0; j < local->neighbors[i].remote_interfaces_nr; j++) {
			// Assuming there is only one link between two nodes
			if (PLATFORM_GET_LINK_METRICS(interface_name,
						      local->neighbors[i].remote_interfaces[j].mac_address, info_link)) {
				memcpy(&nmac, local->neighbors[i].remote_interfaces[j].mac_address, 6);
				break;
			}

		}

		table = blobmsg_open_table(bb, NULL);

		hwaddr_ntoa(local->neighbors[i].al_mac_address, mac_str);
		blobmsg_add_string(bb,"ieee1905id", mac_str);

		hwaddr_ntoa(nmac, mac_str);
		blobmsg_add_string(bb,"interfaceid", mac_str);

		blobmsg_add_string(bb, "media_type", get_interface_type_string(x->interface_type));

		bridge = DMisLinkBridged(interface_name, local->neighbors[i].al_mac_address, nmac);
		blobmsg_add_u8(bb, "ieee802dot1bridge", bridge);
		blobmsg_add_u32(bb, "packet_errors_tx", info_link->tx_packet_errors);
		blobmsg_add_u32(bb, "transmitted_packets", info_link->tx_packet_ok);
		blobmsg_add_u16(bb, "mac_throughput_capacity", info_link->tx_max_xput);
		blobmsg_add_u16(bb, "link_availability", info_link->tx_link_availability);
		blobmsg_add_u32(bb, "packet_errors_rx", info_link->rx_packet_errors);
		blobmsg_add_u32(bb, "packets_received", info_link->rx_packet_ok);
		blobmsg_add_u16(bb, "phy_rate", info_link->tx_phy_rate);
		blobmsg_add_u16(bb, "rssi", info_link->rx_rssi);

		blobmsg_close_table(bb, table);
	}
	PLATFORM_FREE_1905_INTERFACE_INFO(x);
}

void DMdumpNeighbour(struct blob_buf *bb)
{
	INT8U(*al_mac_addresses)[6];
	INT8U al_mac_addresses_nr;
	INT8U i;
	void *array = NULL;
	char mac[20] = {0};

	al_mac_addresses = DMgetListOfNeighbors(&al_mac_addresses_nr);
	if (bb == NULL)
		return;

	blobmsg_add_u32(bb, "neighbors_nr", al_mac_addresses_nr);
	array = blobmsg_open_array(bb, "neighbors");
	for (i = 0; i < al_mac_addresses_nr; i++) {
		hwaddr_ntoa(al_mac_addresses[i], mac);
		blobmsg_add_string(bb, NULL, mac);
	}
	blobmsg_close_array(bb, array);

	return;
}

void DMdumpNetworkDevices(PRINT_CALLBACK callback, PRINT_WRITER write_function, struct blob_buf *bb)
{
	// Buffer size to store a prefix string that will be used to show each
	// element of a structure on screen
	//
#define MAX_PREFIX  100

	INT8U i, j;
	void *array = NULL, *table = NULL, *subarray = NULL, *subtable = NULL;

	write_function("  device_nr: %d", sgDataModel.network_devices_nr);

	if (bb) {
		blobmsg_add_u32(bb, "device_nr", sgDataModel.network_devices_nr);
		array = blobmsg_open_array(bb, "devices");
	}
	for (i = 0; i < sgDataModel.network_devices_nr; i++) {
		if (bb) {
			table = blobmsg_open_table(bb, NULL);
		}
		char new_prefix[MAX_PREFIX];

		PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->", i);
		new_prefix[MAX_PREFIX - 1] = 0x0;
		write_function("%supdate timestamp: %d", new_prefix, sgDataModel.network_devices[i].update_timestamp);
		if (bb)
			blobmsg_add_u32(bb, "update_timestamp", sgDataModel.network_devices[i].update_timestamp);

		PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->general_info->", i);
		new_prefix[MAX_PREFIX - 1] = 0x0;
		visit_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].info, callback, write_function, new_prefix, bb);

		PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->bridging_capabilities_nr: %d", i, sgDataModel.network_devices[i].bridges_nr);
		new_prefix[MAX_PREFIX - 1] = 0x0;
		write_function("%s", new_prefix);
		if (bb) {
			blobmsg_add_u32(bb, "bridging_capabilities_nr", sgDataModel.network_devices[i].bridges_nr);
			subarray = blobmsg_open_array(bb, "bridging_capabilities");
		}
		for (j = 0; j < sgDataModel.network_devices[i].bridges_nr; j++) {
			if (bb)
				subtable = blobmsg_open_table(bb, NULL);
			PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->bridging_capabilities[%d]->", i, j);
			new_prefix[MAX_PREFIX - 1] = 0x0;
			visit_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].bridges[j], callback, write_function, new_prefix, bb);
			if (subtable)
				blobmsg_close_table(bb, subtable);
		}
		if (subarray)
			blobmsg_close_array(bb, subarray);

		PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->non_1905_neighbors_nr: %d", i,
				  sgDataModel.network_devices[i].non1905_neighbors_nr);
		new_prefix[MAX_PREFIX - 1] = 0x0;
		write_function("%s", new_prefix);
		if (bb) {
			blobmsg_add_u32(bb, "non_1905_neighbors_nr", sgDataModel.network_devices[i].non1905_neighbors_nr);
			subarray = blobmsg_open_array(bb, "non1905_neighbors");
		}
		for (j = 0; j < sgDataModel.network_devices[i].non1905_neighbors_nr; j++) {
			if (bb)
				subtable = blobmsg_open_table(bb, NULL);
			PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->non_1905_neighbors[%d]->", i, j);
			new_prefix[MAX_PREFIX - 1] = 0x0;
			visit_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].non1905_neighbors[j], callback, write_function, new_prefix, bb);
			if (subtable)
				blobmsg_close_table(bb, subtable);
		}
		if (subarray)
			blobmsg_close_array(bb, subarray);

		PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->x1905_neighbors_nr: %d", i, sgDataModel.network_devices[i].x1905_neighbors_nr);
		new_prefix[MAX_PREFIX - 1] = 0x0;
		write_function("%s", new_prefix);
		if (bb) {
			blobmsg_add_u32(bb, "x1905_neighbors_nr", sgDataModel.network_devices[i].x1905_neighbors_nr);
			subarray = blobmsg_open_array(bb, "1905_neighbors");
		}
		for (j = 0; j < sgDataModel.network_devices[i].x1905_neighbors_nr; j++) {
			if (bb)
				subtable = blobmsg_open_table(bb, NULL);
			PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->x1905_neighbors[%d]->", i, j);
			new_prefix[MAX_PREFIX - 1] = 0x0;
			visit_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].x1905_neighbors[j], callback, write_function, new_prefix, bb);
			if (subtable)
				blobmsg_close_table(bb, subtable);
		}
		if (subarray)
			blobmsg_close_array(bb, subarray);

		PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->power_off_interfaces_nr: %d", i, sgDataModel.network_devices[i].power_off_nr);
		new_prefix[MAX_PREFIX - 1] = 0x0;
		write_function("%s", new_prefix);
		if (bb) {
			blobmsg_add_u32(bb, "power_off_interfaces_nr", sgDataModel.network_devices[i].power_off_nr);
			subarray = blobmsg_open_array(bb, "power_off_interfaces");
		}
		for (j = 0; j < sgDataModel.network_devices[i].power_off_nr; j++) {
			if (bb)
				subtable = blobmsg_open_table(bb, NULL);
			PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->power_off_interfaces[%d]->", i, j);
			new_prefix[MAX_PREFIX - 1] = 0x0;
			visit_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].power_off[j], callback, write_function, new_prefix, bb);
			if (subtable)
				blobmsg_close_table(bb, subtable);
		}
		if (subarray)
			blobmsg_close_array(bb, subarray);

		PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->l2_neighbors_nr: %d", i, sgDataModel.network_devices[i].l2_neighbors_nr);
		new_prefix[MAX_PREFIX - 1] = 0x0;
		write_function("%s", new_prefix);
		if (bb) {
			blobmsg_add_u32(bb, "l2_neighbors_nr", sgDataModel.network_devices[i].l2_neighbors_nr);
			subarray = blobmsg_open_array(bb, "l2_neighbors");
		}
		for (j = 0; j < sgDataModel.network_devices[i].l2_neighbors_nr; j++) {
			if (bb)
				subtable = blobmsg_open_table(bb, NULL);
			PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->l2_neighbors[%d]->", i, j);
			new_prefix[MAX_PREFIX - 1] = 0x0;
			visit_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].l2_neighbors[j], callback, write_function, new_prefix, bb);
			if (subtable)
				blobmsg_close_table(bb, subtable);
		}
		if (subarray)
			blobmsg_close_array(bb, subarray);

		PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->generic_phys->", i);
		new_prefix[MAX_PREFIX - 1] = 0x0;
		visit_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].generic_phy, callback, write_function, new_prefix, bb);

		PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->profile->", i);
		new_prefix[MAX_PREFIX - 1] = 0x0;
		visit_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].profile, callback, write_function, new_prefix, bb);

		PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->identification->", i);
		new_prefix[MAX_PREFIX - 1] = 0x0;
		visit_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].identification, callback, write_function, new_prefix, bb);

		PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->control_url->", i);
		new_prefix[MAX_PREFIX - 1] = 0x0;
		visit_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].control_url, callback, write_function, new_prefix, bb);

		PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->ipv4->", i);
		new_prefix[MAX_PREFIX - 1] = 0x0;
		visit_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].ipv4, callback, write_function, new_prefix, bb);

		PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->ipv6->", i);
		new_prefix[MAX_PREFIX - 1] = 0x0;
		visit_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].ipv6, callback, write_function, new_prefix, bb);

		PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->metrics_nr: %d", i, sgDataModel.network_devices[i].metrics_with_neighbors_nr);
		new_prefix[MAX_PREFIX - 1] = 0x0;
		write_function("%s", new_prefix);
		if (bb) {
			blobmsg_add_u32(bb, "metrics_nr", sgDataModel.network_devices[i].metrics_with_neighbors_nr);
			subarray = blobmsg_open_array(bb, "metrics");
		}
		for (j = 0; j < sgDataModel.network_devices[i].metrics_with_neighbors_nr; j++) {
			if (bb)
				subtable = blobmsg_open_table(bb, NULL);
			PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->metrics[%d]->tx->", i, j);
			new_prefix[MAX_PREFIX - 1] = 0x0;
			if (sgDataModel.network_devices[i].metrics_with_neighbors[j].tx_metrics != NULL) {
				write_function("%slast_updated: %d", new_prefix, sgDataModel.network_devices[i].metrics_with_neighbors[j].tx_metrics_timestamp);
				if (bb)
					blobmsg_add_u32(bb, "tx_last_updated", sgDataModel.network_devices[i].metrics_with_neighbors[j].tx_metrics_timestamp);
				visit_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].metrics_with_neighbors[j].tx_metrics, callback,
							 write_function, new_prefix, bb);
			}
			PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->metrics[%d]->rx->", i, j);
			new_prefix[MAX_PREFIX - 1] = 0x0;
			if (sgDataModel.network_devices[i].metrics_with_neighbors[j].rx_metrics != NULL) {
				write_function("%slast updated: %d", new_prefix, sgDataModel.network_devices[i].metrics_with_neighbors[j].rx_metrics_timestamp);
				if (bb)
					blobmsg_add_u32(bb, "rx_last_updated", sgDataModel.network_devices[i].metrics_with_neighbors[j].rx_metrics_timestamp);
				visit_1905_TLV_structure((INT8U *) sgDataModel.network_devices[i].metrics_with_neighbors[j].rx_metrics, callback,
							 write_function, new_prefix, bb);
			}
			if (subtable)
				blobmsg_close_table(bb, subtable);
		}
		if (subarray)
			blobmsg_close_array(bb, subarray);

		PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "  device[%d]->", index);
		if (table)
			blobmsg_close_table(bb, table);
	}
	if (array)
		blobmsg_close_array(bb, array);

	return;
}

INT8U DMrunGarbageCollector(void)
{
	INT8U i, j, k;
	INT8U removed_entries;
	INT8U original_devices_nr;

	removed_entries = 0;

	// Visit all existing devices, searching for those with a timestamp older
	// than GC_MAX_AGE
	//
	// Note that we skip element "0", which is always the local device. We don't
	// care when it was last updated as it is always updated "on demand", just
	// before someone requests its data (right now the only place where this
	// happens is when using an ALME custom command)
	//
	original_devices_nr = sgDataModel.network_devices_nr;
	for (i = 1; i < sgDataModel.network_devices_nr; i++) {
		INT8U *p = NULL;

		p = DMmacToAlMac(sgDataModel.network_devices[i].info->al_mac_address);
		if ((get_time() - sgDataModel.network_devices[i].update_timestamp > (GC_MAX_AGE * 1000)) ||
		    (sgDataModel.network_devices[i].info != NULL &&
		     p == NULL)
		    ) {
			// Entry too old or with a MAC address no longer registered in the
			// "topology discovery" database. Remove it.
			INT8U al_mac_address[6] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
			struct _networkDevice *x;

			removed_entries++;

			x = &sgDataModel.network_devices[i];

			// First, free all child structures
			//
			if (x->info != NULL) {
				// Save the MAC of the node that is going to be removed for
				// later use
				//
				memcpy(al_mac_address, x->info->al_mac_address, 6);

				DEBUG("Removing old device entry (%02x:%02x:%02x:%02x:%02x:%02x)", x->info->al_mac_address[0], x->info->al_mac_address[1],
				      x->info->al_mac_address[2], x->info->al_mac_address[3], x->info->al_mac_address[4], x->info->al_mac_address[5]);
				free_1905_TLV_structure((INT8U *) x->info, true);
				x->info = NULL;
			} else {
				WARNING("Removing old device entry (Unknown AL MAC)");
			}

			for (j = 0; j < x->bridges_nr; j++) {
				free_1905_TLV_structure((INT8U *) x->bridges[j], true);
			}
			if (0 != x->bridges_nr && NULL != x->bridges) {
				FREE(x->bridges);
				x->bridges_nr = 0;
				x->bridges = NULL;
			}

			for (j = 0; j < x->non1905_neighbors_nr; j++) {
				free_1905_TLV_structure((INT8U *) x->non1905_neighbors[j], true);
			}
			if (0 != x->non1905_neighbors_nr && NULL != x->non1905_neighbors) {
				FREE(x->non1905_neighbors);
				x->non1905_neighbors_nr = 0;
				x->non1905_neighbors = NULL;
			}

			for (j = 0; j < x->x1905_neighbors_nr; j++) {
				free_1905_TLV_structure((INT8U *) x->x1905_neighbors[j], true);
			}
			if (0 != x->x1905_neighbors_nr && NULL != x->x1905_neighbors) {
				FREE(x->x1905_neighbors);
				x->x1905_neighbors_nr = 0;
				x->x1905_neighbors = NULL;
			}

			for (j = 0; j < x->power_off_nr; j++) {
				free_1905_TLV_structure((INT8U *) x->power_off[j], true);
			}
			if (0 != x->power_off_nr && NULL != x->power_off) {
				FREE(x->power_off);
				x->power_off_nr = 0;
				x->power_off = NULL;
			}

			for (j = 0; j < x->l2_neighbors_nr; j++) {
				free_1905_TLV_structure((INT8U *) x->l2_neighbors[j], true);
			}
			if (0 != x->l2_neighbors_nr && NULL != x->l2_neighbors) {
				FREE(x->l2_neighbors);
				x->l2_neighbors_nr = 0;
				x->l2_neighbors = NULL;
			}

			if (x->generic_phy != NULL) {
				free_1905_TLV_structure((INT8U *) x->generic_phy, true);
				x->generic_phy = NULL;
			}

			if (x->profile != NULL) {
				free_1905_TLV_structure((INT8U *) x->profile, true);
				x->profile = NULL;
			}

			if (x->identification != NULL) {
				free_1905_TLV_structure((INT8U *) x->identification, true);
				x->identification = NULL;
			}

			if (x->control_url != NULL) {
				free_1905_TLV_structure((INT8U *) x->control_url, true);
				x->control_url = NULL;
			}

			if (x->ipv4 != NULL) {
				free_1905_TLV_structure((INT8U *) x->ipv4, true);
				x->ipv4 = NULL;
			}

			if (x->ipv6 != NULL) {
				free_1905_TLV_structure((INT8U *) x->ipv6, true);
				x->ipv6 = NULL;
			}

			for (j = 0; j < x->metrics_with_neighbors_nr; j++) {
				free_1905_TLV_structure((INT8U *) x->metrics_with_neighbors[j].tx_metrics, true);
				free_1905_TLV_structure((INT8U *) x->metrics_with_neighbors[j].rx_metrics, true);
			}
			if (0 != x->metrics_with_neighbors_nr && NULL != x->metrics_with_neighbors) {
				FREE(x->metrics_with_neighbors);
				x->metrics_with_neighbors = NULL;
			}
			// Next, remove the _networkDevice entry
			//
			if (i == (sgDataModel.network_devices_nr - 1)) {
				// Last element. It will automatically be removed below (keep
				// reading)
			} else {
				// Place the last element here (we don't care about preserving
				// order)
				//
				sgDataModel.network_devices[i] = sgDataModel.network_devices[sgDataModel.network_devices_nr - 1];
				i--;
			}
			sgDataModel.network_devices_nr--;

			// Next, Remove all references to this node from other node's
			// metrics information entries
			//
			for (j = 0; j < sgDataModel.network_devices_nr; j++) {
				INT8U original_neighbors_nr;

				original_neighbors_nr = sgDataModel.network_devices[j].metrics_with_neighbors_nr;

				for (k = 0; k < sgDataModel.network_devices[j].metrics_with_neighbors_nr; k++) {
					if (0 ==
					    memcmp(al_mac_address, sgDataModel.network_devices[j].metrics_with_neighbors[k].neighbor_al_mac_address, 6)) {
						free_1905_TLV_structure((INT8U *) sgDataModel.network_devices[j].metrics_with_neighbors[k].tx_metrics, true);
						free_1905_TLV_structure((INT8U *) sgDataModel.network_devices[j].metrics_with_neighbors[k].rx_metrics, true);

						// Place last element here (we don't care about
						// preserving order)
						//
						if (k == (sgDataModel.network_devices[j].metrics_with_neighbors_nr - 1)) {
							// Last element. It will automatically be removed
							// below (keep reading)
						} else {
							sgDataModel.network_devices[j].metrics_with_neighbors[k] =
							    sgDataModel.network_devices[j].metrics_with_neighbors[sgDataModel.network_devices[j].
														  metrics_with_neighbors_nr - 1];
							k--;
						}
						sgDataModel.network_devices[j].metrics_with_neighbors_nr--;
					}
				}

				if (original_neighbors_nr != sgDataModel.network_devices[j].metrics_with_neighbors_nr) {
					if (sgDataModel.network_devices[j].metrics_with_neighbors_nr == 0)
						FREE(sgDataModel.network_devices[j].metrics_with_neighbors);
					else
						REALLOC(sgDataModel.network_devices[j].metrics_with_neighbors,
							(sgDataModel.network_devices[j].metrics_with_neighbors_nr));
				}
			}

			// And also from the local interfaces database
			//
			DMremoveALNeighborFromInterface(al_mac_address, "all");
		}

		FREE(p);
	}

	// If at least one element was removed, we need to realloc
	//
	if (original_devices_nr != sgDataModel.network_devices_nr) {
		if (sgDataModel.network_devices_nr == 0)
			FREE(sgDataModel.network_devices);
		else
			REALLOC(sgDataModel.network_devices, (sgDataModel.network_devices_nr));
	}

	return removed_entries;
}

void DMremoveALNeighborFromInterface(INT8U *al_mac_address, char *interface_name)
{
	INT8U i, j;

	for (i = 0; i < sgDataModel.local_interfaces_nr; i++) {
		INT8U original_neighbors_nr;

		if ((memcmp(sgDataModel.local_interfaces[i].name, interface_name, strlen(sgDataModel.local_interfaces[i].name) + 1) != 0) &&
		    (0 != memcmp(interface_name, "all", strlen(interface_name) + 1))
		    ) {
			// Ignore this interface
			//
			continue;
		}

		original_neighbors_nr = sgDataModel.local_interfaces[i].neighbors_nr;

		for (j = 0; j < sgDataModel.local_interfaces[i].neighbors_nr; j++) {
			if (memcmp(al_mac_address, sgDataModel.local_interfaces[i].neighbors[j].al_mac_address, 6) == 0) {
				if (sgDataModel.local_interfaces[i].neighbors[j].remote_interfaces_nr > 0
				    && NULL != sgDataModel.local_interfaces[i].neighbors[j].remote_interfaces) {
					FREE(sgDataModel.local_interfaces[i].neighbors[j].remote_interfaces);

					sgDataModel.local_interfaces[i].neighbors[j].remote_interfaces = NULL;
					sgDataModel.local_interfaces[i].neighbors[j].remote_interfaces_nr = 0;
				}
				// Place last element here (we don't care about preserving
				// order)
				//
				if (j == (sgDataModel.local_interfaces[i].neighbors_nr - 1)) {
					// Last element. It will automatically be removed
					// below (keep reading)
				} else {
					sgDataModel.local_interfaces[i].neighbors[j] =
					    sgDataModel.local_interfaces[i].neighbors[sgDataModel.local_interfaces[i].neighbors_nr - 1];
					j--;
				}
				sgDataModel.local_interfaces[i].neighbors_nr--;
			}
		}

		if (original_neighbors_nr != sgDataModel.local_interfaces[i].neighbors_nr) {
			if (sgDataModel.local_interfaces[i].neighbors_nr == 0)
				FREE(sgDataModel.local_interfaces[i].neighbors);
			else
				REALLOC(sgDataModel.local_interfaces[i].neighbors,
					(sgDataModel.local_interfaces[i].neighbors_nr));
		}
	}
}

struct vendorSpecificTLV ***DMextensionsGet(INT8U *al_mac_address, INT8U **nr)
{
	INT8U i;
	struct vendorSpecificTLV ***extensions;

	// Find device
	if ((al_mac_address == NULL) || (nr == NULL)) {
		ERROR("Invalid argument");
		return NULL;
	}
	// Search for an existing entry with the same AL MAC address
	for (i = 0; i < sgDataModel.network_devices_nr; i++) {
		if (sgDataModel.network_devices[i].info == NULL) {
			// We haven't received general info about this device yet.
			//
			continue;
		}
		if (memcmp(sgDataModel.network_devices[i].info->al_mac_address, al_mac_address, 6) == 0) {
			break;
		}
	}

	if (i == sgDataModel.network_devices_nr) {
		// A matching entry was *not* found.
		DEBUG("Extension received from an unknown 1905 node (%02x:%02x:%02x:%02x:%02x:%02x). Ignoring data...",
		      al_mac_address[0], al_mac_address[1],
		      al_mac_address[2], al_mac_address[3],
		      al_mac_address[4], al_mac_address[5]);
		extensions = NULL;
	} else {
		// Point to the datamodel extensions section
		//
		extensions = &sgDataModel.network_devices[i].extensions;
		*nr = &sgDataModel.network_devices[i].extensions_nr;
	}

	return extensions;
}

/**
 * @brief Function to check if the mac address is of local interfaces or of
 * local al mac address.
 *
 * @param mac 6 byte long mac address
 *
 * @return bool true if the mac address is of local node, false otherwise
 *
 */
bool is_local_mac(INT8U *mac)
{
	INT8U *almac = NULL;

	if (mac == NULL)
		return true;

	if (hwaddr_is_zero(mac))
		return true;

	if (DMmacToInterfaceName(mac) != NULL)
		return true;

	almac = DMalMacGet();
	if (memcmp(mac, almac, 6) == 0)
		return true;

	return false;
}

/**
 * @brief Function to check if the mac address is of ieee1905 multicast
 * address (01:80:C2:00:00:13).
 *
 * @param mac 6 byte long mac address
 *
 * @return bool true if the mac address is of ieee1905 multicast group,
 * false otherwise
 *
 */
bool is_ieee1905_mcast_mac(INT8U *mac)
{
	if (memcmp(mac, "\x01\x80\xc2\x00\x00\x13", 6) == 0)
		return true;

	return false;
}
