/*
 * al_entity.c: IEEE-1905 AL processing
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

// In the comments below, every time a reference is made (ex: "See Section 6.4"
// or "See Table 6-11") we are talking about the contents of the following
// document:
//
//   "IEEE Std 1905.1-2013"

#include "utils.h"
#include "utils_uci.h"

#include "cmdu_visit.h"
#include "1905_alme.h"
#include "lldp_tlvs.h"
#include "lldp_payload.h"

#include "al_entity.h"
#include "al_datamodel.h"
#include "al_send.h"
#include "al_recv.h"
#include "plugin_mngr.h"
#include "al_wsc.h"

#include "platform_interfaces.h"
#include "platform_interfaces_ubus.h"
#include "platform_os.h"

#include <libubox/list.h>

#define FOREVER (1)
#define TIMER_TOKEN_DISCOVERY          (1)
#define TIMER_TOKEN_GARBAGE_COLLECTOR  (2)
#define MAX_REASSAMBLE_AGE (10 * 1000)  // 10 sec
#define TOTAL_FRAGMENTS (300)  // queue depth
#define MAX_FRAGMENTS_PER_SOURCE (60)
#define MAX_FRAGMENTS_PER_CMDU (5)

static bool sg_stop_flag = false;

static void dump_cmdu_fragments(void) __attribute__((unused));
struct tid_list
{
	pthread_t tid;
	struct list_head list;
};

static struct tid_list sg_pid_list;

static bool gs_ap_search_needed = true;
static uint8_t gs_trigger_wsc_renew = 0;

struct cmdu_fragment
{
	uint8_t src_mac[6];
	uint16_t mid;
	uint8_t fid;
	uint32_t age;
	struct cmdu_stream *cstream;
	struct list_head list;
};

static struct cmdu_fragment *gs_cmdu_fragment_head;

void init_pid_list()
{
	INIT_LIST_HEAD(&sg_pid_list.list);
}

int add_tid_to_list(pthread_t tid)
{
	struct tid_list *entry;

	entry = (struct tid_list *) malloc(sizeof(struct tid_list));
	if(!entry)
		return 1;

	entry->tid = tid;

	list_add_tail(&entry->list, &sg_pid_list.list);

	return 0;
}

void join_all_tid_list()
{
	struct tid_list *entry;

        list_for_each_entry(entry, &sg_pid_list.list, list) {
		pthread_join(entry->tid, NULL);
	}
}

void clean_tid_list()
{
	struct tid_list *entry, *tmp;

        list_for_each_entry_safe(entry, tmp, &sg_pid_list.list, list) {
		pthread_cancel(entry->tid);
		list_del(&entry->list);
		FREE(entry);
        }
}

static bool is_interface_available(char *ifname)
{
	bool ret = true;
	struct interfaceInfo *x;

	x = PLATFORM_GET_1905_INTERFACE_INFO(ifname);
	if (x == NULL) {
		WARNING("Could not retrieve info of interface %s", ifname);
		return false;
	}

	if ((x->is_secured == 0) ||
	    ((x->power_state != INTERFACE_POWER_STATE_ON) &&
	     (x->power_state != INTERFACE_POWER_STATE_SAVE))) {
		ret = false;
	}

	PLATFORM_FREE_1905_INTERFACE_INFO(x);
	return ret;
}

static void cmdu_frag_init()
{
	CALLOC(gs_cmdu_fragment_head, 1);

	INIT_LIST_HEAD(&gs_cmdu_fragment_head->list);
}

static void dump_cmdu_fragments(void)
{
	struct list_head *listptr;
	struct cmdu_fragment *entry;

	WARNING("#### Dumping fragments #####");
        list_for_each(listptr, &gs_cmdu_fragment_head->list) {
                entry = list_entry(gs_cmdu_fragment_head->list.next,
				   struct cmdu_fragment, list);
		WARNING("mac (%02x:%02x:%02x:%02x:%02x:%02x)",
		     entry->src_mac[0], entry->src_mac[1],
		     entry->src_mac[2], entry->src_mac[3],
		     entry->src_mac[4], entry->src_mac[5]);
		WARNING("mid(%d), fid(%d), age(%d), cmdu_type(%x)",
		     entry->mid, entry->fid, entry->age,
		     entry->cstream->message_type);
        }
	WARNING("#### Dumping fragments done #####");
}

static void cleanup_cmdu_fragments(void)
{
	struct list_head *listptr;
	struct cmdu_fragment *entry;

	if (gs_cmdu_fragment_head == NULL)
		return;

        list_for_each(listptr, &gs_cmdu_fragment_head->list) {
                entry = list_entry(gs_cmdu_fragment_head->list.next,
				   struct cmdu_fragment, list);

		if (list_empty(&entry->list)) {
			break;
		}

		free_cmdu_cstream(entry->cstream);
		list_del(&entry->list);
		FREE(entry);
        }
	FREE(gs_cmdu_fragment_head);
}
////////////////////////////////////////////////////////////////////////////////
// Private functions and data
////////////////////////////////////////////////////////////////////////////////
static struct cmdu_stream *reassemble_fragmented_packets(uint8_t *packet_buffer, uint16_t len)
{

	uint8_t dst_addr[6];
	uint8_t src_addr[6];
	uint16_t ether_type;

	uint16_t mid;
	uint8_t fragment_id;
	uint8_t last_fragment = 0;

	uint8_t *p;

	p = packet_buffer;

	_EnB(&p, dst_addr, 6);
	_EnB(&p, src_addr, 6);
	_E2B(&p, &ether_type);

	// eth_header(6+6+2) cmdu_header(1+1+2+2+1+1+3)
	if (len < 25) {
		ERROR("Fragment len is less than cmdu fragment |%d|", len);
		return NULL;
	}

	if (ether_type != ETHERTYPE_1905) {
		ERROR("Unsupported ethertype |0x%02x|", ether_type);
		return NULL;
	}

	// strip eth header
	len -= (6 + 6 + 2);

	if (get_cmdu_header_from_packet(p, &mid, &fragment_id, &last_fragment) == false) {
		ERROR("Could not retrieve CMDU header from bit stream");
		return NULL;
	}

	// This is a complete cmdu process this further
	if (fragment_id == 0 &&
	    last_fragment == 1) {
		 return convert_packet_to_tlv_streams(p, len);
	} else {
		// Iterate through the list to find if the packet already present in db
		struct list_head *listptr;
		struct cmdu_fragment *entry, *found_entry = NULL;
		struct cmdu_stream *cstream;
		struct cmdu_stream *temp_stream;
		uint32_t cur_tick, time_elapsed;
		uint8_t src_frag_count = 0;
		uint16_t total_frag_count = 0;

		INFO("Reassembly mid(%d), fid(%d), last|%d|",
		      mid, fragment_id, last_fragment);

		cur_tick = get_time();

		// Filters
		list_for_each(listptr, &gs_cmdu_fragment_head->list) {
			entry = list_entry(gs_cmdu_fragment_head->list.next,
					   struct cmdu_fragment, list);

			if (list_empty(&entry->list)) {
				break;
			}
			time_elapsed = cur_tick - entry->age;
			DEBUG("# loop entry mid(%d:%d), fid(%d), age(%d)",
			      entry->mid, mid, entry->fid, time_elapsed);

			// check age and if exceeds, delete this entry
			if (time_elapsed > MAX_REASSAMBLE_AGE) {
				INFO("Deleting this entry age(%d) expires", time_elapsed);
				INFO("mac (%02x:%02x:%02x:%02x:%02x:%02x)",
					entry->src_mac[0], entry->src_mac[1],
					entry->src_mac[2], entry->src_mac[3],
					entry->src_mac[4], entry->src_mac[5]);
				list_del(&entry->list);
				temp_stream = entry->cstream;
				free_cmdu_cstream(temp_stream);
				continue;
			}

			total_frag_count += entry->fid;
			if (memcmp(&entry->src_mac, &src_addr, 6) == 0) {
				src_frag_count += entry->fid;
			}

			if (entry->mid == mid &&
			    memcmp(&entry->src_mac, &src_addr, 6) == 0) {
				// found the matching entry
				found_entry = entry;
			}
		}

		if (total_frag_count > TOTAL_FRAGMENTS) {
			WARNING("Can't accomodate more fragments");
			return NULL;
		}

		if (src_frag_count > MAX_FRAGMENTS_PER_SOURCE) {
			WARNING("SRC(%02:%02:%02:%02:%02:%02) sending too much fragmented data",
			      src_addr[0], src_addr[1], src_addr[2],
			      src_addr[3], src_addr[4], src_addr[5]);
			return NULL;
		}

		if (found_entry) {
			uint8_t total_tlv_count, entry_tlv_count;
			// ignore the fragments greater than max
			// the cmdu will be aged out latter on
			if (fragment_id > MAX_FRAGMENTS_PER_CMDU) {
				WARNING("Fragment exceeds max fragment per cmdu");
				return NULL;
			}

			// Only process the next fid in sequence, otherwise
			// discard the packet
			if (found_entry->fid + 1 != fragment_id) {
				WARNING("(%d:%d) not next fid",
				      found_entry->fid, fragment_id);
				return NULL;
			}

			temp_stream = convert_packet_to_tlv_streams(p, len);
			if (temp_stream == NULL) {
				WARNING("decoding failed for current frag");
				return NULL;
			}

			// Update the existing entry
			found_entry->fid = fragment_id;

			cstream = found_entry->cstream;
			entry_tlv_count = cstream->nbr_tlv_streams;
			total_tlv_count = temp_stream->nbr_tlv_streams + cstream->nbr_tlv_streams;

			if (cstream->nbr_tlv_streams)
				REALLOC(cstream->tlv_stream_list, total_tlv_count);
			else
				CALLOC(cstream->tlv_stream_list, total_tlv_count);

			cstream->nbr_tlv_streams = total_tlv_count;

			for (uint8_t i = 0; i < temp_stream->nbr_tlv_streams; i++) {
				cstream->tlv_stream_list[entry_tlv_count + i] = temp_stream->tlv_stream_list[i];
			}

			// free temp stream
			free(temp_stream->tlv_stream_list);
			free(temp_stream);

			if (last_fragment == 1) {
				INFO("NBR tlv(%d) last(%d)",
				      cstream->nbr_tlv_streams, last_fragment);
				list_del(&found_entry->list);
				//dump_cmdu_fragments();
				return cstream;
			}
		} else {
			// this should not happen
			if (last_fragment == 1) {
				WARNING("Last frag before first");
				return NULL;
			}
			// This is not the fist segment
			if (fragment_id != 0) {
				WARNING("fid not zero");
				return NULL;
			}

			// Add a new entry
			CALLOC(entry, 1);

			entry->fid = fragment_id;
			entry->mid = mid;
			memcpy(entry->src_mac, src_addr, 6);
			entry->age = get_time();

			cstream = convert_packet_to_tlv_streams(p, len);
			if (cstream == NULL) {
				WARNING("conversion fails fid:0");
				// if conversion fails then no need to store this in db
				FREE(entry);
				return NULL;
			}
			entry->cstream = cstream;
			list_add_tail(&entry->list, &gs_cmdu_fragment_head->list);
			INFO("New entry added mid(%d), fid(%d) in list", mid, fragment_id);
			//dump_cmdu_fragments();
		}

	}

	return NULL;
}

// Returns '1' if the packet has already been processed in the past and thus,
// should be discarded (to avoid network storms). '0' otherwise.
//
// According to what is explained in "Sections 7.5, 7.6 and 7.7" if a
// defragmented packet whose "AL MAC address TLV" and "message id" match one
// that has already been received in the past, then it should be discarded.
//
// I *personally* think the standard is "slightly" wrong here because *not* all
// CMDUs contain an "AL MAC address TLV".
// We could use the ethernet source address instead, however this would only
// work for those messages that are *not* relayed (one same duplicated relayed
// message can arrive at our local node with two different ethernet source
// addresses).
// Fortunately for us, all relayed CMDUs *do* contain an "AL MAC address TLV",
// thus this is what we are going to do:
//
//   1. If the CMDU is a relayed one, check against the "AL MAC" contained in
//      the "AL MAC address TLV"
//
//   2. If the CMDU is *not* a relayed one, check against the ethernet source
//      address
//
// This function keeps track of the latest MAX_DUPLICATES_LOG_ENTRIES tuples
// of ("mac_address", "message_id") and:
//
//   1. If the provided tuple matches an already existing one, this function
//      returns '1'
//
//   2. Otherwise, the entry is added (discarding, if needed, the oldest entry)
//      and this function returns '0'
//
INT8U _checkDuplicates(INT8U *src_mac_address, struct cmdu_stream *cstream)
{
#define MAX_DUPLICATES_LOG_ENTRIES 10

	static INT8U mac_addresses[MAX_DUPLICATES_LOG_ENTRIES][6];
	static INT16U message_ids[MAX_DUPLICATES_LOG_ENTRIES];

	static INT8U start;
	static INT8U total;

	INT8U mac_address[6];

	INT8U i;

	uint8_t *remote_al_mac = DMmacToAlMac(src_mac_address);
	// For relayed CMDUs, use the AL MAC, otherwise use the ethernet src MAC.
	if (remote_al_mac == NULL) {
		memcpy(mac_address, src_mac_address, 6);
	} else {
		memcpy(mac_address, remote_al_mac, 6);
		FREE(remote_al_mac);
	}

	if (cstream->relay_indicator == 1) {
		for (uint8_t i = 0; i < cstream->nbr_tlv_streams; i++) {
			if (*cstream->tlv_stream_list[i] == TLV_TYPE_AL_MAC_ADDRESS_TYPE) {
				struct alMacAddressTypeTLV *t =
					(struct alMacAddressTypeTLV *) cstream->tlv_stream_list[i];

				memcpy(mac_address, t->al_mac_address, 6);
				break;
			}
		}
	}
	// Also, discard relayed CMDUs whose AL MAC is our own (that means someone
	// is retrasnmitting us back a message we originally created)
	if (cstream->relay_indicator == 1) {
		if (memcmp(mac_address, DMalMacGet(), 6) == 0) {
			return 1;
		}
	}

	if (cstream->message_type == CMDU_TYPE_VENDOR_SPECIFIC)
		return 0;

	// Find if the ("mac_address", "message_id") tuple is already present in the
	// database
	for (i = 0; i < total; i++) {
		INT8U index;

		index = (start + i) % MAX_DUPLICATES_LOG_ENTRIES;

		if (0 == memcmp(mac_addresses[index], mac_address, 6)) {

			if (message_ids[index] == cstream->message_id) {
				// The entry already exists!
				return 1;
			} else {
				// Override this mac entry as this al has replied with different mid
				memcpy(mac_addresses[index], mac_addresses[start], 6);
				start --;
				total --;
				start = start % MAX_DUPLICATES_LOG_ENTRIES;
			}
		}
	}

	// This is a new entry, insert it into the cache and return "0"
	if (total < MAX_DUPLICATES_LOG_ENTRIES) {
		// There is space for new entries
		INT8U index;

		index = (start + total) % MAX_DUPLICATES_LOG_ENTRIES;

		memcpy(mac_addresses[index], mac_address, 6);
		message_ids[index] = cstream->message_id;

		total++;
	} else {
		// We need to replace the oldest entry
		memcpy(mac_addresses[start], mac_address, 6);
		message_ids[start] = cstream->message_id;

		start++;

		start = start % MAX_DUPLICATES_LOG_ENTRIES;
	}

	return 0;
}

// According to "Section 7.6", if a received packet has the "relayed multicast"
// bit set, after processing, we must forward it on all authenticated 1905
// interfaces (except on the one where it was received).
//
// This function checks if the provided 'c' structure has that "relayed
// multicast" flag set and, if so, retransmits it on all local interfaces
// (except for the one whose MAC address matches 'receiving_interface_addr') to
// 'destination_mac_addr' and the same "message id" (MID) as the one contained
// in the originally received 'c' structure.
//
void _checkForwarding(INT8U *receiving_interface_addr, INT8U *destination_mac_addr, struct cmdu_stream *cstream)
{
	INT8U i;

	if (cstream->relay_indicator) {
		char **ifs_names;
		INT8U ifs_nr;

		DEBUG("Relay multicast flag set. Forwarding...");

		ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);
		for (i = 0; i < ifs_nr; i++) {
			INT8U authenticated;
			INT8U power_state;

			struct interfaceInfo *x;

			x = PLATFORM_GET_1905_INTERFACE_INFO(ifs_names[i]);
			if (x == NULL) {
				WARNING("Could not retrieve info of interface %s", ifs_names[i]);
				authenticated = 0;
				power_state = INTERFACE_POWER_STATE_OFF;
				continue;
			}

			authenticated = x->is_secured;
			power_state = x->power_state;

			if ((authenticated == 0) ||
			    ((power_state != INTERFACE_POWER_STATE_ON) &&
			     (power_state != INTERFACE_POWER_STATE_SAVE)) ||
			    (memcmp(x->mac_address, receiving_interface_addr, 6) == 0)
			    ) {
				// Do not forward the message on this interface
				PLATFORM_FREE_1905_INTERFACE_INFO(x);
				continue;
			}
			PLATFORM_FREE_1905_INTERFACE_INFO(x);

			// Retransmit message
			INFO("--> 0x%04x (forwarding from %s to %s)", cstream->message_type,
			     DMmacToInterfaceName(receiving_interface_addr), ifs_names[i]);

			if (fragment_send_raw_packets(ifs_names[i], destination_mac_addr, cstream) == false) {
				WARNING("Could not retransmit 1905 message on interface %s", x->name);
			}
		}
		PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);
	}

	return;
}

// This function sends an "AP-autoconfig search" message on all authenticated
// interfaces BUT ONLY if there is at least one unconfigured AP interface on
// this node.
//
// A function has been created for this because the same code is executed from
// three different places:
//
//   - When a new interface becomes authenticated
//
//   - When the *local* push button is pressed and there is at least one
//     interface which does not support this configuration mechanism (ex:
//     ethernet)
//
//   - After a local unconfigured AP interface becomes configured (this is
//     needed in the unlikely situation where there are more than one
//     unconfigured APs in the same node)
//
static void _triggerAPSearchProcess(bool skip_checks)
{
	INT8U i;
	char **ifs_names;
	INT8U ifs_nr;

	if (false == skip_checks) {
		if (false == gs_ap_search_needed)
			return;

		if (get_unconfigured_ap_nbr() == 0) {
			gs_ap_search_needed = false;
			return;
		}
	}

	if (skip_checks == true) {
		ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);
	} else {
		ifs_names = get_list_of_unconfigured_ap(&ifs_nr);
	}

	for (i = 0; i < ifs_nr; i++) {
		INT8U unconfigured_ap_band = 0;
		struct interfaceInfo *x = NULL;

		x = PLATFORM_GET_1905_INTERFACE_INFO(ifs_names[i]);
		if (x == NULL) {
			WARNING("Could not retrieve info of interface %s", ifs_names[i]);
			continue;
		}

		INT8U authenticated = x->is_secured;
		INT8U power_state = x->power_state;

		// non wireless interfaces will be skipped from this check
		if (false == get_freq_band_from_media_type(x->interface_type, &unconfigured_ap_band)) {
			PLATFORM_FREE_1905_INTERFACE_INFO(x);
			continue;
		}

		// this might be redundant
		if (false == skip_checks) {
			if (x->interface_type_data.ieee80211.role != IEEE80211_ROLE_AP) {
				PLATFORM_FREE_1905_INTERFACE_INFO(x);
				continue;
			}
		}

		if ((authenticated == 1) &&
		    ((power_state == INTERFACE_POWER_STATE_ON) ||
		     (power_state == INTERFACE_POWER_STATE_SAVE))) {
			if (send1905APAutoconfigurationSearchPacket(unconfigured_ap_band) == 0) {
				DEBUG("Could not send 1905 AP-autoconfiguration search message");
			}
		}

		PLATFORM_FREE_1905_INTERFACE_INFO(x);
	}

	if (skip_checks)
		PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);
	else
		FREE(ifs_names);
}

static bool db_init()
{
	uint8_t i;
	char **interfaces_names;
	INT8U interfaces_nr;

	// Obtain the list of interfaces that the AL entity is going to manage
	interfaces_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&interfaces_nr);
	if (interfaces_names == NULL) {
		ERROR("No interfaces detected");
		return false;
	}

	for (i = 0; i < interfaces_nr; i++) {
		struct interfaceInfo *x;

		x = PLATFORM_GET_1905_INTERFACE_INFO(interfaces_names[i]);
		if (x == NULL) {
			ERROR("Could not retrieve interface info for %s", interfaces_names[i]);
			continue;
		}
		DMinsertInterface(x->name, x->mac_address);
		PLATFORM_FREE_1905_INTERFACE_INFO(x);
	}
	PLATFORM_FREE_LIST_OF_1905_INTERFACES(interfaces_names, interfaces_nr);

	return true;
}

static bool registrar_init(bool enabled)
{
	bool ret;
	char **interfaces_names;
	INT8U interfaces_nr, i;
	char ssid[32], key[63], enc[20];
	char ssid2[32], key2[63], enc2[20];
	struct interfaceInfo *x;
	uint8_t intf_type;;

	if (enabled == false) {
		INFO("Registrar is not enabled");
		return true;
	}

	// Obtain the list of interfaces that the AL entity is going to manage
	DEBUG("Retrieving list of interfaces visible to the 1905 AL entity...");
	interfaces_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&interfaces_nr);
	if (interfaces_names == NULL) {
		ERROR("No interfaces detected");
		return false;
	}

	for (i = 0; i < interfaces_nr; i++) {
		uint8_t intf_ap_band = IEEE80211_FREQUENCY_BAND_2_4_GHZ;
		intf_type = get_interface_type(interfaces_names[i]);

		// registrar init is only applicable for wireless interfaces
		if (intf_type != INTF_TYPE_WIFI)
			continue;

		x = PLATFORM_GET_1905_INTERFACE_INFO(interfaces_names[i]);
		if (x == NULL) {
			ERROR("Could not retrieve interface info for %s", interfaces_names[i]);
			continue;
		}

		// If this interface is the designated 1905 network registrar
		// interface, save its MAC address to the database
		INFO("Registrar interface name(%s)", interfaces_names[i]);
		DMregistrarInsert(x->mac_address, x->interface_type);
		platform_update_bssid(interfaces_names[i]);

		if (false == get_freq_band_from_media_type(x->interface_type, &intf_ap_band)) {
			PLATFORM_FREE_1905_INTERFACE_INFO(x);
			continue;
		}

		PLATFORM_FREE_1905_INTERFACE_INFO(x);

		ret = uci_get_wireless_interface_option(UCI_WIRELESS, interfaces_names[i], "ssid", ssid, sizeof(ssid));
		ret &= uci_get_wireless_interface_option(UCI_WIRELESS, interfaces_names[i], "key", key, sizeof(key));
		ret &= uci_get_wireless_interface_option(UCI_WIRELESS, interfaces_names[i], "encryption", enc, sizeof(enc));

		if (ret == false) {
			ERROR("Failed to get info from wireless uci for intf(%s)", interfaces_names[i]);
			return false;
		}

		ret = uci_check_wifi_iface(UCI_IEEE1905, interfaces_names[i]);
		if (ret == false) {
			ret = uci_add_wireless_iface_sec(UCI_IEEE1905, interfaces_names[i]);
			if (ret == false) {
				return false;
			}

			uci_set_wireless_interface_option(UCI_IEEE1905, interfaces_names[i], "ssid", ssid);
			uci_set_wireless_interface_option(UCI_IEEE1905, interfaces_names[i], "key", key);
			uci_set_wireless_interface_option(UCI_IEEE1905, interfaces_names[i], "encryption", enc);
			MODIFY_BIT(intf_ap_band, gs_trigger_wsc_renew, 1);

		} else {
			ret = uci_get_wireless_interface_option(UCI_IEEE1905, interfaces_names[i], "ssid", ssid2, sizeof(ssid));
			ret &= uci_get_wireless_interface_option(UCI_IEEE1905, interfaces_names[i], "key", key2, sizeof(key));
			ret &= uci_get_wireless_interface_option(UCI_IEEE1905, interfaces_names[i], "encryption", enc2, sizeof(enc));

			if (ret == false) {
				ERROR("Failed to get wifi-iface for intf(%s) from ieee1905", interfaces_names[i]);
				return false;
			}

			if (strcmp(ssid, ssid2) == 0 &&
			    strcmp(key, key2) == 0 &&
			    strcmp(enc, enc2) == 0) {
				INFO("WSC renew is not required");
			} else {
				MODIFY_BIT(intf_ap_band, gs_trigger_wsc_renew, 1);
				uci_set_wireless_interface_option(UCI_IEEE1905, interfaces_names[i], "ssid", ssid);
				uci_set_wireless_interface_option(UCI_IEEE1905, interfaces_names[i], "key", key);
				uci_set_wireless_interface_option(UCI_IEEE1905, interfaces_names[i], "encryption", enc);
			}
		}
	}

	if (gs_trigger_wsc_renew != 0) {
		INFO("## WSC renew is required (%d)##", gs_trigger_wsc_renew);
	}

	PLATFORM_FREE_LIST_OF_1905_INTERFACES(interfaces_names, interfaces_nr);
	return true;
}

bool register_egress_policy()
{
	char **interfaces_names;
	INT8U interfaces_nr, i;

	// Obtain the list of interfaces that the AL entity is going to manage
	interfaces_names = uci_get_list_of_1905_interfaces(&interfaces_nr);
	if (interfaces_names == NULL || interfaces_nr == 0) {
		ERROR("No interfaces detected");
		return false;
	}

	// We are interested in processing 1905 packets that arrive on any of the
	// 1905 interfaces.
	// For this we are going to tell the platform code that we want to receive
	// a message on the just created queue every time a new 1905 packet arrives
	// on any of those interfaces.
	DEBUG("Registering packet arrival event for each interface...");
	for (i = 0; i < interfaces_nr; i++) {
		struct event1905Packet aux;

		memset(&aux, 0, sizeof(struct event1905Packet));
		strncpy(aux.interface_name, interfaces_names[i], IFNAMSIZ);
		memcpy(aux.interface_mac_address, DMinterfaceNameToMac(aux.interface_name), 6);
		memcpy(aux.al_mac_address, DMalMacGet(), 6);

		if (PLATFORM_REGISTER_QUEUE_EVENT(PLATFORM_QUEUE_EVENT_NEW_1905_PACKET, &aux) == 0) {
			ERROR("Could not register callback for 1905 packets in interface %s", interfaces_names[i]);
			return false;
		}
	}
	uci_free_list_of_1905_interfaces(interfaces_names, interfaces_nr);
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Public functions
////////////////////////////////////////////////////////////////////////////////
INT8U start1905AL(bool registrar_enabled)
{
	INT8U queue_message[MAX_PCAP_QUEUE_SIZE];
	INT8U i;

	// Initialize platform-specific code
	if (init_time() == 0) {
		ERROR("Failed to initialize platform");
		return AL_ERROR_OS;
	}

	// initialise cmdu fragmentation handler
	cmdu_frag_init();

	if (!db_init()) {
		ERROR("db init failed");
		return AL_ERROR_OS;
	}

	// initialise registrar
	if(!registrar_init(registrar_enabled)) {
		ERROR("Registrar init failed");
		return AL_ERROR_OS;
	}

	// Initialize m1 and key
	wsc_init();

	if(!register_egress_policy()) {
		ERROR("pcap policy registration failed");
		return AL_ERROR_OS;
	}

	// Create a queue that will later be used by the platform code to notify us
	// when certain types of "events" take place
	DEBUG("Creating events queue...");
	if (false == PLATFORM_CREATE_QUEUE()) {
		ERROR("Could not create events queue");
		return AL_ERROR_OS;
	}

	// We are also interested in processing a 60 seconds timeout event (so that
	// we can send new discovery messages into the network)
	//
	DEBUG("Registering DISCOVERY time out event (periodic)...");
	{
		struct eventTimeOut aux;

		aux.timeout_ms = 60000;	// 60 seconds
		aux.token = TIMER_TOKEN_DISCOVERY;

		if (PLATFORM_REGISTER_QUEUE_EVENT(PLATFORM_QUEUE_EVENT_TIMEOUT_PERIODIC, &aux) == 0) {
			ERROR("Could not register timer callback");
			return AL_ERROR_OS;
		}
	}

	// ...and a slighlty higher timeout to "clean" the database from nodes that
	// have left the network without notice
	//
	DEBUG("Registering GARBAGE COLLECTOR time out event (periodic)...");
	{
		struct eventTimeOut aux;

		aux.timeout_ms = 70000;	// 70 seconds
		aux.token = TIMER_TOKEN_GARBAGE_COLLECTOR;

		if (PLATFORM_REGISTER_QUEUE_EVENT(PLATFORM_QUEUE_EVENT_TIMEOUT_PERIODIC, &aux) == 0) {
			ERROR("Could not register timer callback");
			return AL_ERROR_OS;
		}
	}

	// As soon as we enter the queue message processing loop we want to start
	// the discovery process as if a "DISCOVERY timeout" event had just
	// happened.
	// In other words, we want the first "DISCOVERY timeout" event to take place
	// at t=0 and then every 60 seconds.
	// In order to "force" this first event at t=0 we use a new timer event,
	// but this time this is a one time (ie. non-periodic) timer which will
	// time out in just one second from now.
	//
	DEBUG("Registering a one time forced DISCOVERY event...");
	{
		struct eventTimeOut aux;

		aux.timeout_ms = 1;
		aux.token = TIMER_TOKEN_DISCOVERY;

		if (PLATFORM_REGISTER_QUEUE_EVENT(PLATFORM_QUEUE_EVENT_TIMEOUT, &aux) == 0) {
			ERROR("Could not register timer callback");
			return AL_ERROR_OS;
		}
	}

	// ...and the "new authenticated link" event, needed to produce the "push
	// button join notification" message.
	//
	DEBUG("Registering the NEW AUTHENTICATED LINK event...");
	if (PLATFORM_REGISTER_QUEUE_EVENT(PLATFORM_QUEUE_EVENT_AUTHENTICATED_LINK, NULL) == 0) {
		ERROR("Could not register 'authenticated link' event");
		return AL_ERROR_OS;
	}
	// Prepare the message queue
	DEBUG("Entering read-process loop...");
	while (FOREVER) {
		memset(&queue_message, 0, MAX_PCAP_QUEUE_SIZE);
		INT8U *p;
		INT8U message_type;
		INT16U message_len;

		if (sg_stop_flag) {
			DEBUG("Stoping all threads");
			al_cleanup();
		}

		DEBUG("Waiting for new queue message...");
		if (PLATFORM_READ_QUEUE(queue_message) == 0) {
			DEBUG("Something went wrong while trying to retrieve a new message from the queue. Ignoring...");
			continue;
		}
		// The first byte of 'queue_message' tells us the type of message that
		// we have just received
		p = &queue_message[0];
		_E1B(&p, &message_type);
		_E2B(&p, &message_len);

		if (message_len >= MAX_PCAP_QUEUE_SIZE) {
			ERROR("Received stream is bigger than queue len, skip ...");
			continue;
		}

		switch (message_type) {
		case PLATFORM_QUEUE_EVENT_NEW_1905_PACKET:
			{
				INT8U *q;
				INT8U dst_addr[6];
				INT8U src_addr[6];
				INT16U ether_type;

				INT8U receiving_interface_addr[6];
				char *receiving_interface_name;

				// The first six bytes of the message payload contain the MAC
				// address of the interface where the packet was received
				//
				_EnB(&p, receiving_interface_addr, 6);

				receiving_interface_name = DMmacToInterfaceName(receiving_interface_addr);
				if (receiving_interface_name == NULL) {
					ERROR("A packet was receiving on MAC %02x:%02x:%02x:%02x:%02x:%02x, which does not match any local interface",
					      receiving_interface_addr[0], receiving_interface_addr[1], receiving_interface_addr[2],
					      receiving_interface_addr[3], receiving_interface_addr[4], receiving_interface_addr[5]);
					continue;
				}

				if (is_interface_available(receiving_interface_name) == false)
					continue;
				q = p;

				// The next bytes are the actual packet payload (ie. the
				// ethernet payload)
				//
				_EnB(&q, dst_addr, 6);
				_EnB(&q, src_addr, 6);
				_E2B(&q, &ether_type);

				DEBUG("New queue message arrived: packet captured on interface %s", receiving_interface_name);
				DEBUG("    Dst address: %02x:%02x:%02x:%02x:%02x:%02x", dst_addr[0], dst_addr[1], dst_addr[2], dst_addr[3], dst_addr[4],
				      dst_addr[5]);
				DEBUG("    Src address: %02x:%02x:%02x:%02x:%02x:%02x", src_addr[0], src_addr[1], src_addr[2], src_addr[3], src_addr[4],
				      src_addr[5]);
				DEBUG("    Ether type : 0x%04x", ether_type);

				switch (ether_type) {
				case ETHERTYPE_LLDP:
					{
						struct PAYLOAD *payload;

						DEBUG("LLDP message received.");

						payload = parse_lldp_PAYLOAD_from_packet(q);

						if (payload == NULL) {
							WARNING("Invalid bridge discovery message. Ignoring...");
						} else {
							DEBUG("LLDP message contents:");
							visit_lldp_PAYLOAD_structure(payload, print_callback, print_debug, "", NULL);

							processLlpdPayload(payload, receiving_interface_addr);

							free_lldp_PAYLOAD_structure(payload);
						}

						break;
					}

				case ETHERTYPE_1905:
					{
						struct cmdu_cstruct *c;
						struct cmdu_stream *cstream;

						DEBUG("CMDU message received. Reassembling...");

						cstream = reassemble_fragmented_packets(p, message_len);

						// if cstream and c are null then it means
						// this was just a fragment part of a big CMDU.
						// The data has been internally cached, waiting for
						// the rest of pieces.
						if (cstream == NULL) {
							break;
						}

						if (_checkDuplicates(src_addr, cstream) == 1) {
							INFO("Received duplicate CMDU on %s (mid = %d). Discarding...",
							      receiving_interface_name, cstream->message_id);
							free_cmdu_cstream(cstream);
							break;
						}

						// cstream = convert_packet_to_tlv_streams(stream, message_len - 14);
						c = convert_stream_to_cstruct(cstream);

						plugin_process(cstream, receiving_interface_addr,
							       src_addr);


						if (c == NULL) {
							DEBUG("CMDU(%x) might be map specific...skip..",
							      cstream->message_type);
							free_cmdu_cstream(cstream);
							break;
						}


						if (check_1905_cmdu_rules(c)) {
							DEBUG("CMDU message contents:");
							visit_1905_CMDU_structure(c, print_callback, print_debug, "", NULL);

							INT8U res;
							DEBUG("Process 1905 CMDU:");
							res = process1905Cmdu(c, receiving_interface_addr, src_addr);

							if (res == PROCESS_CMDU_OK_TRIGGER_AP_SEARCH) {
								_triggerAPSearchProcess(false);
							}

						}
						// It might be necessary to retransmit this
						// message on the rest of interfaces (depending
						// on the "relayed multicast" flag
						_checkForwarding(receiving_interface_addr, dst_addr, cstream);

						free_1905_CMDU_structure(c);
						free_cmdu_cstream(cstream);
						break;
					}

				default:
					{
						WARNING("Unknown ethertype 0x%04x!! Ignoring...", ether_type);
						break;
					}
				}

				break;
			}
#ifdef ALME_OVER_TCP
		case PLATFORM_QUEUE_EVENT_NEW_ALME_MESSAGE:
			{
				// ALME messages contain:
				//
				//   1- one byte with the "client id" (which must be used when
				//      later calling "PLATFORM_SEND_ALME_REPLY()")
				//
				//   2- the bit stream representation of an ALME TLV.
				//
				// We just need to convert it into a struct and process it:
				//
				INT8U alme_client_id;
				INT8U *alme_tlv;

				_E1B(&p, &alme_client_id);

				DEBUG("New queue message arrived: ALME message (client ID = %d).", alme_client_id);

				alme_tlv = parse_1905_ALME_from_packet(p);
				if (alme_tlv == NULL) {
					WARNING("Invalid ALME message. Ignoring...");
				}

				DEBUG("ALME message contents:");
				visit_1905_ALME_structure((INT8U *) alme_tlv, print_callback, print_debug, "", NULL);

				process1905Alme(alme_tlv, alme_client_id);

				free_1905_ALME_structure(alme_tlv);

				break;
			}
#endif
		case PLATFORM_QUEUE_EVENT_TIMEOUT:
		case PLATFORM_QUEUE_EVENT_TIMEOUT_PERIODIC:
			{
				INT32U timer_id;

				// The message payload of this type of messages only contains
				// four bytes with the "timer ID" that expired.
				//
				_E4B(&p, &timer_id);

				DEBUG("New queue message arrived: timer 0x%08x expired", timer_id);

				switch (timer_id) {
				case TIMER_TOKEN_DISCOVERY:
					{
						INT16U mid;

						char **ifs_names;
						INT8U ifs_nr;

						// According to "Section 8.2.1.1" and "Section 8.2.1.2"
						// we now have to send a "Topology discovery message"
						// followed by a "802.1 bridge discovery message" but,
						// according to the rules in "Section 7.2", only on each
						// and every of the *authenticated* 1905 interfaces
						// that are in the state of "PWR_ON" or "PWR_SAVE"
						//
						ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);
						mid = getNextMid();
						for (i = 0; i < ifs_nr; i++) {
							if (is_interface_available(ifs_names[i]) == false)
								continue;

							// Topology discovery message
							if (send1905TopologyDiscoveryPacket(ifs_names[i], mid) == 0) {
								WARNING("Could not send 1905 topology discovery message");
							}
							// 802.1 bridge discovery message
							if (sendLLDPBridgeDiscoveryPacket(ifs_names[i]) == 0) {
								WARNING("Could not send LLDP bridge discovery message");
							}

						}
						PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);
						// Trigger Auto configuration search
						_triggerAPSearchProcess(true);

						// Trigger Auto configuration renew
						if (gs_trigger_wsc_renew) {
							if (BIT(IEEE80211_FREQUENCY_BAND_2_4_GHZ, gs_trigger_wsc_renew))
								send1905APAutoconfigurationRenewPacket(IEEE80211_FREQUENCY_BAND_2_4_GHZ);

							if (BIT(IEEE80211_FREQUENCY_BAND_5_GHZ, gs_trigger_wsc_renew))
								send1905APAutoconfigurationRenewPacket(IEEE80211_FREQUENCY_BAND_5_GHZ);

							if (BIT(IEEE80211_FREQUENCY_BAND_60_GHZ, gs_trigger_wsc_renew))
								send1905APAutoconfigurationRenewPacket(IEEE80211_FREQUENCY_BAND_60_GHZ);

							gs_trigger_wsc_renew = 0;
						}
						break;
					}

				case TIMER_TOKEN_GARBAGE_COLLECTOR:
					{
						DEBUG("Running garbage collector...");

						if (DMrunGarbageCollector() > 0) {
							INT16U mid;

							char **ifs_names;
							INT8U ifs_nr;

							DEBUG("Some elements were removed. Sending a topology change notification...");

							// According to "Section 8.2.2.3" and "Section
							// 7.2", we now have to send a "Topology
							// Notification message) on all *authenticated*
							// interfaces that are in the state of "PWR_ON" or
							// "PWR_SAVE"
							//
							ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);
							mid = getNextMid();
							for (i = 0; i < ifs_nr; i++) {
								if (is_interface_available(ifs_names[i]) == false)
									continue;

								// Topology notification message
								if (send1905TopologyNotificationPacket(ifs_names[i], mid) == 0) {
									WARNING("Could not send 1905 topology discovery message");
								}
							}
							PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);

						}
						break;
					}

				default:
					{
						WARNING("Unknown timer ID!! Ignoring...");
						break;
					}
				}

				break;
			}

		case PLATFORM_QUEUE_EVENT_PUSH_BUTTON:
			{
				INT16U mid;

				char **ifs_names;
				INT8U ifs_nr;

				INT8U *no_push_button;
				INT8U at_least_one_unsupported_interface;

				DEBUG("New queue message arrived: push button event");

				// According to "Section 9.2.2.1", we must first make sure that
				// none of the interfaces is in the middle of a previous "push
				// button" configuration sequence.
				//
				ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);
				mid = getNextMid();

				for (i = 0; i < ifs_nr; i++) {
					struct interfaceInfo *x;

					x = PLATFORM_GET_1905_INTERFACE_INFO(ifs_names[i]);
					if (x == NULL) {
						WARNING("Could not retrieve info of interface %s", ifs_names[i]);
						break;
					} else {
						if (x->push_button_on_going == 1) {
							INFO("Interface %s is in the middle of a previous 'push button' configuration sequence. Ignoring new event...", ifs_names[i]);

							PLATFORM_FREE_1905_INTERFACE_INFO(x);
							break;
						}
						PLATFORM_FREE_1905_INTERFACE_INFO(x);
					}

				}
				if (i < ifs_nr) {
					// Don't do anything
					//
					break;
				}
				// If we get here, none of the interfaces is in the middle of a
				// "push button" configuration process, thus we can initialize
				// the "push button event" on all of our interfaces that support
				// it.
				//
				// Let's see which interfaces support it and keep track of
				// those who don't by setting the corresponding byte in array
				// "no_push_button" to '1'
				//
				CALLOC(no_push_button, ifs_nr);

				for (i = 0; i < ifs_nr; i++) {
					struct interfaceInfo *x;

					x = PLATFORM_GET_1905_INTERFACE_INFO(ifs_names[i]);
					if (x == NULL) {
						WARNING("Could not retrieve info of interface %s", ifs_names[i]);

						no_push_button[i] = 1;
						break;
					} else {
						if (x->power_state == POWER_STATE_PWR_OFF) {
							// Ignore interfaces that are switched off
							//
							DEBUG("Skipping interface %s because it is powered off", ifs_names[i]);
							no_push_button[i] = 1;
						} else if (x->push_button_on_going == 2) {
							// This interface does not support the "push button"
							// configuration process
							//
							DEBUG("Skipping interface %s because it does not support the push button configuration mechanism",
							      ifs_names[i]);
							no_push_button[i] = 2;

							// NOTE: "2" will be used as a special marker to
							//       later trigger the AP search process (see
							//       below)
						} else if ((MEDIA_TYPE_IEEE_802_11B_2_4_GHZ == x->interface_type ||
							    MEDIA_TYPE_IEEE_802_11G_2_4_GHZ == x->interface_type ||
							    MEDIA_TYPE_IEEE_802_11A_5_GHZ == x->interface_type ||
							    MEDIA_TYPE_IEEE_802_11N_2_4_GHZ == x->interface_type ||
							    MEDIA_TYPE_IEEE_802_11N_5_GHZ == x->interface_type ||
							    MEDIA_TYPE_IEEE_802_11AC_5_GHZ == x->interface_type ||
							    MEDIA_TYPE_IEEE_802_11AD_60_GHZ == x->interface_type ||
							    MEDIA_TYPE_IEEE_802_11AF_GHZ == x->interface_type) &&
							   IEEE80211_ROLE_AP != x->interface_type_data.ieee80211.role &&
							   (0x0 != x->interface_type_data.ieee80211.bssid[0] ||
							    0x0 != x->interface_type_data.ieee80211.bssid[1] ||
							    0x0 != x->interface_type_data.ieee80211.bssid[2] ||
							    0x0 != x->interface_type_data.ieee80211.bssid[3] ||
							    0x0 != x->interface_type_data.ieee80211.bssid[4] || 0x0 != x->interface_type_data.ieee80211.bssid[5]
							   )
						    ) {
							// According to "Section 9.2.2.1", an 802.11 STA
							// which is already paired with an AP must *not*
							// start the "push button" configuration process.
							//
							DEBUG("Skipping interface %s because it is a wifi STA already associated to an AP", ifs_names[i]);
							no_push_button[i] = 1;
						} else {
							no_push_button[i] = 0;
						}

						PLATFORM_FREE_1905_INTERFACE_INFO(x);
					}
				}

				// We now have the list of interfaces that need to start their
				// "push button" configuration process. Let's do it:
				//
				at_least_one_unsupported_interface = 0;
				for (i = 0; i < ifs_nr; i++) {
					if (no_push_button[i] == 0) {
						INFO("Starting push button configuration process on interface %s", ifs_names[i]);
						PLATFORM_START_PUSH_BUTTON_CONFIGURATION(ifs_names[i], DMalMacGet(), mid);
					}
					if (no_push_button[i] == 2) {
						at_least_one_unsupported_interface = 1;
					}
				}
				if (at_least_one_unsupported_interface == 1) {
					// The reason for doing this is the next one:
					//
					// Imagine one device with two interfaces: an unconfigured
					// AP wifi interface and an ethernet interface.
					//
					// If we press the button we *need*  to send the
					// "AP search" CMDU... however, because the ethernet interface
					// never starts the "push button configuration" process
					// (because it is not supported!), the interface can never
					// "become authenticated" and trigger the AP search
					// process.
					//
					// That's why we do it here, manually
					//
					_triggerAPSearchProcess(false);
				}
				// Finally, send the notification message (so that the rest of
				// 1905 nodes is aware of this situation) but only on already
				// authenticated interfaces.
				//
				for (i = 0; i < ifs_nr; i++) {

					if (is_interface_available(ifs_names[i]) == false)
						continue;

					// Push button notification message
					if (send1905PushButtonEventNotificationPacket(ifs_names[i], mid, ifs_names, no_push_button, ifs_nr) == 0) {
						WARNING("Could not send 1905 push button event notification message");
					}
				}

				PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);

				break;
			}

		case PLATFORM_QUEUE_EVENT_AUTHENTICATED_LINK:
			{
				// Two different things need to be done when a new interface is
				// authenticated:
				//
				//   1. According to "Section 9.2.2.3", a "push button join
				//      notification" message must be generated and sent.
				//
				//   2. According to "Section 10.1", the "AP-autoconfiguration"
				//      process is triggered.

				INT16U mid;

				INT8U local_mac_addr[6];
				INT8U new_mac_addr[6];
				INT8U original_al_mac_addr[6];
				INT16U original_mid;

				char **ifs_names;
				INT8U ifs_nr;

				// The first six bytes of the message payload contain the MAC
				// address of the interface where the "push button"
				// configuration process succeeded.
				//
				_EnB(&p, local_mac_addr, 6);

				// The next six bytes contain the MAC address of the interface
				// successfully authenticated at the other end.
				//
				_EnB(&p, new_mac_addr, 6);

				// The next six bytes contain the original AL MAC address that
				// started everything.
				//
				_EnB(&p, original_al_mac_addr, 6);

				// Finally, the last two bytes contains the MID of that original
				// message
				//
				_E2B(&p, &original_mid);

				DEBUG("New queue message arrived: authenticated link");
				DEBUG("    Local interface:        %02x:%02x:%02x:%02x:%02x:%02x", local_mac_addr[0], local_mac_addr[1], local_mac_addr[2],
				      local_mac_addr[3], local_mac_addr[4], local_mac_addr[5]);
				DEBUG("    New (remote) interface: %02x:%02x:%02x:%02x:%02x:%02x", new_mac_addr[0], new_mac_addr[1], new_mac_addr[2],
				      new_mac_addr[3], new_mac_addr[4], new_mac_addr[5]);
				DEBUG("    Original AL MAC       : %02x:%02x:%02x:%02x:%02x:%02x", original_al_mac_addr[0], original_al_mac_addr[1],
				      original_al_mac_addr[2], original_al_mac_addr[3], original_al_mac_addr[4], original_al_mac_addr[5]);
				DEBUG("    Original MID          : %d", original_mid);

				ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);

				// If "new_mac_addr" is NULL, this means the interface was
				// "authenticated" as a whole (not at "link level").
				// This happens for ethernet interfaces.
				// In these cases we must *not* send the "push button join
				// notification" message (note, however, that the "AP-
				// autoconfiguration" process does need to be triggered, which
				// is done later)
				//
				if (new_mac_addr[0] == 0x00 &&
				    new_mac_addr[1] == 0x00 &&
				    new_mac_addr[2] == 0x00 && new_mac_addr[3] == 0x00 && new_mac_addr[4] == 0x00 && new_mac_addr[5] == 0x00) {
					DEBUG("NULL new (remote) interface. No 'push button join notification' will be sent.");
				} else {
					// Send the "push button join notification" message on all
					// authenticated interfaces (except for the one just
					// authenticated)
					//
					mid = getNextMid();
					for (i = 0; i < ifs_nr; i++) {

						if (is_interface_available(ifs_names[i]) == false)
							continue;

						send1905PushButtonJoinNotificationPacket(ifs_names[i], mid,
											 original_al_mac_addr,
											 original_mid,
											 local_mac_addr,
											 new_mac_addr);
					}
				}

				PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);

				// Finally, trigger the "AP-autoconfiguration" process
				_triggerAPSearchProcess(false);

				break;
			}

		case PLATFORM_QUEUE_EVENT_TOPOLOGY_CHANGE_NOTIFICATION:
			{
				INT16U mid;

				char **ifs_names;
				INT8U ifs_nr;

				DEBUG("New queue message arrived: topology change notification event");

				// TODO:
				//   1. Find which L2 neighbors are no longer available
				//   2. Set their timestamp to 0
				//   3. Call DMrunGarbageCollector() to remove them from the
				//      database
				//
				// Until this is done, nodes will only be removed from the
				// database when the "TIMER_TOKEN_GARBAGE_COLLECTOR" timer
				// expires.

				// According to "Section 8.2.2.3" and "Section 7.2", we now
				// have to send a "Topology Notification" message) on all
				// *authenticated* interfaces that are in the state of "PWR_ON"
				// or "PWR_SAVE"
				ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);
				mid = getNextMid();
				for (i = 0; i < ifs_nr; i++) {

					if (is_interface_available(ifs_names[i]) == false)
						continue;

					// Topology notification message
					if (send1905TopologyNotificationPacket(ifs_names[i], mid) == 0) {
						WARNING("Could not send 1905 topology discovery message");
					}
				}
				PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);

				break;
			}
		case PLATFORM_QUEUE_EVENT_STOP:
			{
				al_cleanup();
				break;
			}

		default:
			{
				WARNING("Unknown queue message type (%d)", message_type);

				break;
			}
		}
	}

	return 0;
}

void al_cleanup()
{
	INFO("Cleanup called");
	clean_tid_list();
	DEBUG("Cleaning datamodel");
	DM_cleanup();
	DEBUG("Cleaning AL queues");
	PLATFORM_CLEAN_QUEUE();
	DEBUG("Cleaning cmdu fragments");
	cleanup_cmdu_fragments();
	DEBUG("Cleaning WSC data");
	wsc_clean();
	DEBUG("Cleaning interface metadata");
	free_al_intf_metadata();
	DEBUG("Cleaning interface names");
	clean_interface_data();
	DEBUG("Cleaning interface mutex");
	PLATFORM_INTERFACE_MUTEX_DESTROY();
	DEBUG("Remove plugin");
	removePlugin();
	DEBUG("Removing pid file");
	unlink(PID_FILE);
	INFO("Cleanup done now exiting");
	exit(0);
}

void set_stop_flag()
{
	sg_stop_flag = true;
}
