/*
 * al_recv.c: IEEE-1905 AL packet processing
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "utils.h"

#include "al_recv.h"
#include "al_datamodel.h"
#include "al_send.h"
#include "al_wsc.h"
#include "al_ubus.h"

#include "1905_tlvs.h"
#include "1905_cmdus.h"
#include "lldp_tlvs.h"
#include "lldp_payload.h"

#include "platform_interfaces.h"
#include "platform_interfaces_ubus.h"
#include "al_ubus.h"
#include <easy/easy.h>

static bool _is_pbc_interface(struct interfaceInfo *x)
{
	switch (x->interface_type) {
	case MEDIA_TYPE_IEEE_802_11B_2_4_GHZ:
	case MEDIA_TYPE_IEEE_802_11G_2_4_GHZ:
	case MEDIA_TYPE_IEEE_802_11N_2_4_GHZ:
	case MEDIA_TYPE_IEEE_802_11A_5_GHZ:
	case MEDIA_TYPE_IEEE_802_11N_5_GHZ:
	case MEDIA_TYPE_IEEE_802_11AC_5_GHZ:
	case MEDIA_TYPE_IEEE_802_11AD_60_GHZ:
		return true;
	}
	return false;
}
////////////////////////////////////////////////////////////////////////////////
// Public functions (exported only to files in this same folder)
////////////////////////////////////////////////////////////////////////////////

INT8U process1905Cmdu(struct cmdu_cstruct *c, INT8U *receiving_interface_addr, INT8U *src_addr)
{
	if (c == NULL) {
		return PROCESS_CMDU_KO;
	}

	char *recv_intf = DMmacToInterfaceName(receiving_interface_addr);

	if (recv_intf == NULL) {
		ERROR("not able to retrieve src interface name from src mac");
		return PROCESS_CMDU_KO;
	}

	BroadcastCMDU(c, recv_intf, src_addr);

	switch (c->message_type) {
	case CMDU_TYPE_TOPOLOGY_DISCOVERY:
		{
			// When a "topology discovery" is received we must update our
			// internal database (that keeps track of which AL MACs and
			// interface MACs are seen on each interface) and send a "topology
			// query" message asking for more details.

			INT8U *p;
			INT8U i;

			INT8U al_mac_address[6] = {0};
			INT8U mac_address[6] = {0};

			INT8U first_discovery;
			INT32U ellapsed;

			INFO("<-- CMDU_TYPE_TOPOLOGY_DISCOVERY (%s)", recv_intf);

			// We need to update the data model structure, which keeps track
			// of local interfaces, neighbors, and neighbors' interfaces, and
			// what type of discovery messages ("topology discovery" and/or
			// "bridge discovery") have been received on each link.

			if (c->tlv_cstruct_list == NULL) {
				ERROR("Malformed structure.");
				break;
			}
			// First, extract the AL MAC and MAC addresses of the interface
			// which transmitted this "topology discovery" message
			//
			i = 0;
			while (NULL != (p = c->tlv_cstruct_list[i])) {
				switch (*p) {
				case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
					{
						struct alMacAddressTypeTLV *t = (struct alMacAddressTypeTLV *)p;

						memcpy(al_mac_address, t->al_mac_address, 6);

						break;
					}
				case TLV_TYPE_MAC_ADDRESS_TYPE:
					{
						struct macAddressTypeTLV *t = (struct macAddressTypeTLV *)p;

						memcpy(mac_address, t->mac_address, 6);

						break;
					}
				default:
					{
						WARNING("Unexpected TLV (%d) type inside CMDU", *p);
						break;
					}
				}
				i++;
			}

			// Make sure that both the AL MAC and MAC addresses were contained
			// in the CMDU
			//
			if (hwaddr_is_zero(al_mac_address) || hwaddr_is_zero(mac_address)) {
				WARNING("More TLVs were expected inside this CMDU");
				return PROCESS_CMDU_KO;
			}

			DEBUG("AL MAC address = %02x:%02x:%02x:%02x:%02x:%02x",
			      al_mac_address[0], al_mac_address[1],
			      al_mac_address[2], al_mac_address[3],
			      al_mac_address[4], al_mac_address[5]);

			DEBUG("MAC    address = %02x:%02x:%02x:%02x:%02x:%02x",
			      mac_address[0], mac_address[1],
			      mac_address[2], mac_address[3],
			      mac_address[4], mac_address[5]);

			// Next, update the data model
			first_discovery = DMupdateDiscoveryTimeStamps(receiving_interface_addr,
								      al_mac_address,
								      mac_address,
								      TIMESTAMP_TOPOLOGY_DISCOVERY,
								      &ellapsed);
			if (first_discovery == 1) {
#ifdef SPEED_UP_DISCOVERY
				// If the data model did not contain an entry for this neighbor,
				// "manually" (ie. "out of cycle") send a "Topology Discovery"
				// message on the receiving interface.
				// This will speed up the network discovery process, so that
				// the new node does not have to wait until our "60 seconds"
				// timer expires for him to "discover" us
				//
				DEBUG("Is this a new node? Re-scheduling a Topology Discovery so that he 'discovers' us");

				if (send1905TopologyDiscoveryPacket(recv_intf, getNextMid()) == 0) {
					WARNING("Could not send 1905 topology discovery message");
				}
#endif
			}
			// Finally, query the advertising neighbor for (much) more detailed
			// information (but only if we haven't recently queried it!)
			// This will make the other end send us a
			// CMDU_TYPE_TOPOLOGY_RESPONSE message, which we will later
			// process.
			//
			INT8U needs_update_ret = DMnetworkDeviceInfoNeedsUpdate(al_mac_address);
			if (needs_update_ret == 0 ||	// Recently received a Topology Response or....
			    (first_discovery == 2 &&
			     ellapsed < 5000)	// ...recently (<5 seconds) received a Topology Discovery
			    ) {
				// The first condition prevents us from re-asking (ie.
				// re-sending "Topology Queries") to one same node (we already
				// knew of) faster than once every minute.
				//
				// The second condition prevents us from flooding new nodes
				// (from which we haven't received a "Topology Response" yet)
				// with "Topology Queries" faster than once every 5 seconds)
				//
				break;
			}

			if (send1905TopologyQueryPacket(recv_intf, getNextMid(), al_mac_address) == 0) {
				WARNING("Could not send 'topology query' message");
			}

			break;
		}
	case CMDU_TYPE_TOPOLOGY_NOTIFICATION:
		{
			// When a "topology notification" is received we must send a new
			// "topology query" to the sender.
			// The "sender" AL MAC address is contained in the unique TLV
			// embedded in the just received "topology notification" CMDU.

			INT8U *p;
			INT8U i;

			INT8U al_mac_address[6] = {0};

			INFO("<-- CMDU_TYPE_TOPOLOGY_NOTIFICATION (%s)", recv_intf);

			if (c->tlv_cstruct_list == NULL) {
				ERROR("Malformed structure.");
				break;
			}
			// Extract the AL MAC addresses of the interface which transmitted
			// this "topology notification" message
			//
			i = 0;
			while (NULL != (p = c->tlv_cstruct_list[i])) {
				switch (*p) {
				case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
					{
						struct alMacAddressTypeTLV *t = (struct alMacAddressTypeTLV *)p;

						memcpy(al_mac_address, t->al_mac_address, 6);

						break;
					}
				default:
					{
						WARNING("Unexpected TLV (%d) type inside CMDU", *p);
						break;
					}
				}
				i++;
			}

			// Make sure that both the AL MAC and MAC addresses were contained
			// in the CMDU
			if (hwaddr_is_zero(al_mac_address)) {
				WARNING("More TLVs were expected inside this CMDU");
				return PROCESS_CMDU_KO;
			}

			DEBUG("AL MAC address = %02x:%02x:%02x:%02x:%02x:%02x", al_mac_address[0], al_mac_address[1], al_mac_address[2], al_mac_address[3],
			      al_mac_address[4], al_mac_address[5]);

#ifdef SPEED_UP_DISCOVERY
			// We will send a topology discovery back. Why is this useful?
			// Well... imagine a node that has just entered the secure network.
			// The first thing this node will do is sending a
			// "topology notification" which, when received by us, will trigger
			// a "topology query".
			// However, unless we send a "topology discovery" back, the new node
			// will not query us for a while (until we actually send our
			// periodic "topology discovery").
			//
			DEBUG("Is this a new node? Re-scheduling a Topology Discovery so that he 'discovers' us");

			if (send1905TopologyDiscoveryPacket(recv_intf, getNextMid()) == 0) {
				WARNING("Could not send 1905 topology discovery message");
			}
#endif
			// Finally, query the informing node.
			// Note that we don't have to check (as we did in the "topology
			// discovery" case) if we recently updated the data model or not.
			// This is because a "topology notification" *always* implies
			// network changes and thus the device must always be (re)-queried.
			//
			if (send1905TopologyQueryPacket(recv_intf, getNextMid(), al_mac_address) == 0) {
				WARNING("Could not send 'topology query' message");
			}

			break;
		}
	case CMDU_TYPE_TOPOLOGY_QUERY:
		{
			// When a "topology query" is received we must obtain a series of
			// information from the platform and then package and send it back
			// in a "topology response" message.

			INT8U *dst_mac;
			INT8U *p;

			INFO("<-- CMDU_TYPE_TOPOLOGY_QUERY (%s)", recv_intf);

			// We must send the response to the AL MAC of the node who sent the
			// query, however, this AL MAC is *not* contained in the query.
			// The only thing we can do at this point is try to search our AL
			// neighbors data base for a matching MAC.
			//
			p = DMmacToAlMac(src_addr);

			if (p == NULL) {
				// The standard says we should always send to the AL MAC
				// address, however, in these cases, instead of just dropping
				// the packet, sending the response to the 'src' address from
				// the TOPOLOGY QUERY seems the right thing to do.
				//
				dst_mac = src_addr;
				WARNING("Unknown destination AL MAC. Using the 'src' MAC from the TOPOLOGY QUERY (%02x:%02x:%02x:%02x:%02x:%02x)",
					dst_mac[0], dst_mac[1],
					dst_mac[2], dst_mac[3],
					dst_mac[4], dst_mac[5]);
			} else {
				dst_mac = p;
			}

			if (send1905TopologyResponsePacket(recv_intf, c->message_id, dst_mac) == 0) {
				WARNING("Could not send 'topology query' message");
			}

			FREE(p);

			break;
		}
	case CMDU_TYPE_TOPOLOGY_RESPONSE:
		{
			// When a "topology response" is received we must update our
			// internal database (that keeps track of which 1905 devices are
			// present in the network)

			INT8U *p;
			INT8U i;

			struct deviceInformationTypeTLV *info = NULL;
			struct deviceBridgingCapabilityTLV **x = NULL;
			struct non1905NeighborDeviceListTLV **y = NULL;
			struct neighborDeviceListTLV **z = NULL;
			struct powerOffInterfaceTLV **q = NULL;
			struct l2NeighborDeviceTLV **r = NULL;

			INT8U bridges_nr;
			INT8U non1905_neighbors_nr;
			INT8U x1905_neighbors_nr;
			INT8U power_off_nr;
			INT8U l2_neighbors_nr;

			INT8U xi, yi, zi, qi, ri;

			INFO("<-- CMDU_TYPE_TOPOLOGY_RESPONSE (%s)", recv_intf);

			if (c->tlv_cstruct_list == NULL) {
				ERROR("Malformed structure.");
				break;
			}
			// First, extract the device info TLV and "count" how many bridging
			// capability TLVs, non-1905 neighbors TLVs and 1905 neighbors TLVs
			// there are
			//
			bridges_nr = 0;
			non1905_neighbors_nr = 0;
			x1905_neighbors_nr = 0;
			power_off_nr = 0;
			l2_neighbors_nr = 0;
			i = 0;
			while (NULL != (p = c->tlv_cstruct_list[i])) {
				switch (*p) {
				case TLV_TYPE_DEVICE_INFORMATION_TYPE:
					{
						info = (struct deviceInformationTypeTLV *)p;
						break;
					}
				case TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES:
					{
						bridges_nr++;
						break;
					}
				case TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST:
					{
						non1905_neighbors_nr++;
						break;
					}
				case TLV_TYPE_NEIGHBOR_DEVICE_LIST:
					{
						x1905_neighbors_nr++;
						break;
					}
				case TLV_TYPE_POWER_OFF_INTERFACE:
					{
						power_off_nr++;
						break;
					}
				case TLV_TYPE_L2_NEIGHBOR_DEVICE:
					{
						l2_neighbors_nr++;
						break;
					}
				case TLV_TYPE_VENDOR_SPECIFIC:
					{
						// According to the standard, zero or more Vendor
						// Specific TLVs may be present.
						//
						break;
					}
				default:
					{
						WARNING("Unexpected TLV (%d) type inside CMDU", *p);
						break;
					}
				}
				i++;
			}

			// Next, now that we know how many TLVs of each type there are,
			// create an array of pointers big enough to contain them and fill
			// it.
			//
			if (bridges_nr > 0) {
				CALLOC(x, bridges_nr);
			}
			if (non1905_neighbors_nr > 0) {
				CALLOC(y, non1905_neighbors_nr);
			}
			if (x1905_neighbors_nr > 0) {
				CALLOC(z, x1905_neighbors_nr);
			}
			if (power_off_nr > 0) {
				CALLOC(q, power_off_nr);
			}
			if (l2_neighbors_nr > 0) {
				CALLOC(r, l2_neighbors_nr);
			}

			xi = 0;
			yi = 0;
			zi = 0;
			qi = 0;
			ri = 0;
			i = 0;
			while (NULL != (p = c->tlv_cstruct_list[i])) {
				switch (*p) {
				case TLV_TYPE_DEVICE_INFORMATION_TYPE:
					{
						break;
					}
				case TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES:
					{
						x[xi++] = (struct deviceBridgingCapabilityTLV *)p;
						break;
					}
				case TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST:
					{
						y[yi++] = (struct non1905NeighborDeviceListTLV *)p;
						break;
					}
				case TLV_TYPE_NEIGHBOR_DEVICE_LIST:
					{
						z[zi++] = (struct neighborDeviceListTLV *)p;
						break;
					}
				case TLV_TYPE_POWER_OFF_INTERFACE:
					{
						q[qi++] = (struct powerOffInterfaceTLV *)p;
						break;
					}
				case TLV_TYPE_L2_NEIGHBOR_DEVICE:
					{
						r[ri++] = (struct l2NeighborDeviceTLV *)p;
						break;
					}
				case TLV_TYPE_VENDOR_SPECIFIC:
					{
						// According to the standard, zero or more Vendor
						// Specific TLVs may be present.
						//
						free_1905_TLV_structure(p, true);
						break;
					}
				default:
					{
						// We are not interested in other TLVs. Free them
						//
						free_1905_TLV_structure(p, true);
						break;
					}
				}
				i++;
			}

			// The CMDU structure is not needed anymore, but we cannot just let
			// the caller call "free_1905_CMDU_structure()", because it would
			// also free the TLVs references that we need (ie. those saved in
			// the "x", "y", "z", "q" and "r" pointer arrays).
			// The "fix" is easy: just set "c->tlv_cstruct_list" to NULL so that
			// when the caller calls "free_1905_CMDU_structure()", this function
			// ignores (ie. does not free) the "tlv_cstruct_list" list.
			//
			// This will work because and memory won't be lost because:
			//
			//   1. Those TLVs contained in "c->tlv_cstruct_list" that we are not
			//      keeping track off have already been freed (see the "default"
			//      case in the previous "switch" structure).
			//
			//   2. The rest of them will be freed when the data model entry
			//      is replaced/deleted.
			//
			//   3. Setting "C->tlv_cstruct_list" to NULL will cause
			//      "free_1905_CMDU_structure()" to ignore this list.
			//
			FREE(c->tlv_cstruct_list);
			c->tlv_cstruct_list = NULL;

			// Next, update the database. This will take care of duplicate
			// entries (and free TLVs if needed)
			//
			DEBUG("Updating network devices database...");
			DMupdateNetworkDeviceInfo(info->al_mac_address,
						  1, info,
						  1, x, bridges_nr,
						  1, y, non1905_neighbors_nr,
						  1, z, x1905_neighbors_nr,
						  1, q, power_off_nr, 1, r, l2_neighbors_nr, 0, NULL, 0, NULL, 0, NULL, 0, NULL, 0, NULL, 0, NULL);

			// Show all network devices (ie. print them through the logging
			// system)
			//
			DMdumpNetworkDevices(print_callback, print_debug, NULL);

			// And finally, send other queries to the device so that we can
			// keep updating the database once the responses are received
			//
			if (send1905MetricsQueryPacket(recv_intf, getNextMid(), info->al_mac_address) == 0) {
				WARNING("Could not send 'metrics query' message");
			}
			if (send1905HighLayerQueryPacket(recv_intf, getNextMid(), info->al_mac_address) == 0) {
				WARNING("Could not send 'high layer query' message");
			}
			for (i = 0; i < info->local_interfaces_nr; i++) {
				if (info->local_interfaces[i].media_type == MEDIA_TYPE_UNKNOWN) {
					// There is *at least* one generic inteface in the response,
					// thus query for more information
					//
					if (0 ==
					    send1905GenericPhyQueryPacket(recv_intf, getNextMid(), info->al_mac_address)) {
						WARNING("Could not send 'generic phy query' message");
					}
					break;
				}
			}

			// There is one extra thing that needs to be done: send topology
			// query to neighbor's neighbors.
			//
			// This is not strictly necessary for 1905 to work. In fact, as I
			// think the protocol was designed, every node should only be aware
			// of its *direct* neighbors; and it is the HLE responsability to
			// query each node and build the network topology map.
			//
			// However, the 1905 datamodel standard document, interestingly
			// (and, I think, erroneously) includes information from all the
			// nodes (even those that are not direct neighbors).
			//
			// Here we are going to retrieve that information but, because this
			// requires much more memory in the AL node, we will only do this
			// if the user actually expressed his desire to do so when starting
			// the AL entity.
			//
#ifdef IEEE1905_MAP_COMPLETE_NETWORK
			for (i = 0; i < zi; i++) {
				INT8U j;

				// For each neighbor's neighbor on that interface
				for (j = 0; j < z[i]->neighbors_nr; j++) {
					INT8U ii, jj;

					// Discard the current node (obviously)
					//
					if (memcmp(DMalMacGet(), z[i]->neighbors[j].mac_address, 6) == 0) {
						continue;
					}
					// Discard nodes I have just asked for
					//
					for (ii = 0; ii < i; ii++) {
						for (jj = 0; jj < z[ii]->neighbors_nr; jj++) {
							if (memcmp(z[ii]->neighbors[jj].mac_address, z[i]->neighbors[j].mac_address, 6) == 0) {
								continue;
							}
						}
					}

					// Discard neighbors whose information was updated
					// recently (ie. no need to flood the network)
					//
					if (DMnetworkDeviceInfoNeedsUpdate(z[i]->neighbors[j].mac_address) == 0) {
						continue;
					}

					if (0 ==
					    send1905TopologyQueryPacket(recv_intf, getNextMid(),
									z[i]->neighbors[j].mac_address)) {
						WARNING("Could not send 'topology query' message");
					}
				}
			}
#endif

			break;
		}
	case CMDU_TYPE_VENDOR_SPECIFIC:
		{
			INFO("<-- CMDU_TYPE_VENDOR_SPECIFIC (%s)", recv_intf);

			// TODO: Implement vendor specific hooks. Maybe, for now, we should
			// simply call a new "PLATFORM_VENDOR_SPECIFIC_CALLBACK()" function

			break;
		}
	case CMDU_TYPE_LINK_METRIC_QUERY:
		{
			INT8U *p;
			INT8U i;

			INT8U *dst_mac;

			struct linkMetricQueryTLV *t;

			INFO("<-- CMDU_TYPE_LINK_METRIC_QUERY (%s)", recv_intf);

			if (c->tlv_cstruct_list == NULL) {
				ERROR("Malformed structure.");
				break;
			}
			// First, search for the "struct linkMetricQueryTLV"
			//
			i = 0;
			t = NULL;
			while (NULL != (p = c->tlv_cstruct_list[i])) {
				switch (*p) {
				case TLV_TYPE_LINK_METRIC_QUERY:
					{
						t = (struct linkMetricQueryTLV *)p;
						break;
					}
				case TLV_TYPE_VENDOR_SPECIFIC:
					{
						// According to the standard, zero or more Vendor
						// Specific TLVs may be present.
						//
						break;
					}
				default:
					{
						WARNING("Unexpected TLV (%d) type inside CMDU", *p);
						break;
					}
				}
				i++;
			}

			if (t == NULL) {
				WARNING("More TLVs were expected inside this CMDU");
				return PROCESS_CMDU_KO;
			}

			if (t->destination == LINK_METRIC_QUERY_TLV_ALL_NEIGHBORS) {
				DEBUG("Destination = all neighbors");
			} else if (t->destination == LINK_METRIC_QUERY_TLV_SPECIFIC_NEIGHBOR) {
				DEBUG("Destination = specific neighbor (%02x:%02x:%02x:%02x:%02x:%02x)", t->specific_neighbor[0], t->specific_neighbor[1],
				      t->specific_neighbor[2], t->specific_neighbor[3], t->specific_neighbor[4], t->specific_neighbor[5]);
			} else {
				WARNING("Unexpected 'destination' (%d)", t->destination);
				return PROCESS_CMDU_KO;
			}

			if (t->link_metrics_type == LINK_METRIC_QUERY_TLV_TX_LINK_METRICS_ONLY) {
				DEBUG("Type        = Tx metrics only");
			} else if (t->link_metrics_type == LINK_METRIC_QUERY_TLV_RX_LINK_METRICS_ONLY) {
				DEBUG("Type        = Rx metrics only");
			}
			if (t->link_metrics_type == LINK_METRIC_QUERY_TLV_BOTH_TX_AND_RX_LINK_METRICS) {
				DEBUG("Type        = Tx and Rx metrics");
			} else {
				WARNING("Unexpected 'type' (%d)", t->link_metrics_type);
				return PROCESS_CMDU_KO;
			}

			// And finally, send a "metrics response" to the requesting neighbor

			// We must send the response to the AL MAC of the node who sent the
			// query, however, this AL MAC is *not* contained in the query.
			// The only thing we can do at this point is try to search our AL
			// neighbors data base for a matching MAC.
			//
			p = DMmacToAlMac(src_addr);

			if (p == NULL) {
				// The standard says we should always send to the AL MAC
				// address, however, in these cases, instead of just dropping
				// the packet, sending the response to the 'src' address from
				// the METRICS QUERY seems the right thing to do.
				//
				dst_mac = src_addr;
				WARNING("Unknown destination AL MAC. Using the 'src' MAC from the METRICS QUERY (%02x:%02x:%02x:%02x:%02x:%02x)",
					dst_mac[0], dst_mac[1],
					dst_mac[2], dst_mac[3],
					dst_mac[4], dst_mac[5]);
			} else {
				dst_mac = p;
			}

			if (0 ==
			    send1905MetricsResponsePacket(recv_intf, c->message_id, dst_mac, t->destination,
							  t->specific_neighbor, t->link_metrics_type)) {
				WARNING("Could not send 'metrics response' message");
			}

			FREE(p);
			break;
		}
	case CMDU_TYPE_LINK_METRIC_RESPONSE:
		{
			// When a "metrics response" is received we must update our
			// internal database (that keeps track of which 1905 devices are
			// present in the network)

			INT8U *p;
			INT8U i;

			INFO("<-- CMDU_TYPE_LINK_METRIC_RESPONSE (%s)", recv_intf);

			if (c->tlv_cstruct_list == NULL) {
				ERROR("Malformed structure.");
				break;
			}
			// Call "DMupdateNetworkDeviceMetrics()" for each TLV
			//
			DEBUG("Updating network devices database...");

			i = 0;
			while (NULL != (p = c->tlv_cstruct_list[i])) {
				switch (*p) {
				case TLV_TYPE_TRANSMITTER_LINK_METRIC:
				case TLV_TYPE_RECEIVER_LINK_METRIC:
					{
						DMupdateNetworkDeviceMetrics(p);
						break;
					}
				case TLV_TYPE_VENDOR_SPECIFIC:
					{
						// According to the standard, zero or more Vendor
						// Specific TLVs may be present.
						//
						free_1905_TLV_structure(p, true);
						break;
					}
				default:
					{
						WARNING("Unexpected TLV (%d) type inside CMDU", *p);

						free_1905_TLV_structure(p, true);
						break;
					}
				}
				i++;
			}

			// References to the TLVs cannot be freed by the caller (see the
			// comment in "case CMDU_TYPE_TOPOLOGY_RESPONSE:" to understand the
			// following two lines).
			//
			FREE(c->tlv_cstruct_list);
			c->tlv_cstruct_list = NULL;

			// Show all network devices (ie. print them through the logging
			// system)
			//
			DMdumpNetworkDevices(print_callback, print_debug, NULL);

			break;
		}
	case CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH:
		{
			// When a "AP-autoconfig search" is received then, *only* if one
			// of our interfaces is the network AP registrar, an "AP-autoconfig
			// response" message must be sent.
			// Otherwise, the message is ignored.

			INT8U *p;
			INT8U i;

			char **ifs_names;
			INT8U ifs_nr;

			INT8U searched_role_is_present;
			INT8U searched_role = IEEE80211_ROLE_AP;

			INT8U freq_band_is_present;
			INT8U freq_band = IEEE80211_FREQUENCY_BAND_2_4_GHZ;

			INT8U al_mac_address[6] = {0};

			searched_role_is_present = 0;
			freq_band_is_present = 0;

			ERROR("<-- CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH (%s)", recv_intf);

			if (c->tlv_cstruct_list == NULL) {
				ERROR("Malformed structure.");
				break;
			}
			// First, parse the incomming packet to find out three things:
			// - The AL MAC of the node searching for AP-autoconfiguration
			//   parameters.
			// - The "searched role" contained in the "searched role TLV" (must
			//   be "REGISTRAR")
			// - The "freq band" contained in the "autoconfig freq band TLV"
			//   (must match the one of our local registrar interface)
			//
			i = 0;
			while (NULL != (p = c->tlv_cstruct_list[i])) {
				switch (*p) {
				case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
					{
						struct alMacAddressTypeTLV *t = (struct alMacAddressTypeTLV *)p;

						memcpy(al_mac_address, t->al_mac_address, 6);

						break;
					}
				case TLV_TYPE_SEARCHED_ROLE:
					{
						struct searchedRoleTLV *t = (struct searchedRoleTLV *)p;

						searched_role_is_present = 1;
						searched_role = t->role;

						break;
					}
				case TLV_TYPE_AUTOCONFIG_FREQ_BAND:
					{
						struct autoconfigFreqBandTLV *t = (struct autoconfigFreqBandTLV *)p;

						freq_band_is_present = 1;
						freq_band = t->freq_band;

						break;
					}
				default:
					{
						WARNING("Unexpected TLV (%d) type inside CMDU", *p);
						break;
					}
				}
				i++;
			}

			// Make sure that all needed parameters were present in the message
			//
			if (hwaddr_is_zero(al_mac_address) || 0 == searched_role_is_present || 0 == freq_band_is_present) {
				WARNING("More TLVs were expected inside this CMDU");
				return PROCESS_CMDU_KO;
			}

			if (searched_role != IEEE80211_ROLE_AP) {
				WARNING("Unexpected 'searched role'");
				return PROCESS_CMDU_KO;
			}

			// Check our local interfaces, looking for one acting as the
			// registrar.
			// If one is found, send the response.
			ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);
			for (i = 0; i < ifs_nr; i++) {
				struct interfaceInfo *x;
				bool send_autoconfig_response = false;

				x = PLATFORM_GET_1905_INTERFACE_INFO(ifs_names[i]);
				if (x == NULL) {
					WARNING("Could not retrieve info of interface %s", ifs_names[i]);
					continue;
				}

				if (get_interface_type(ifs_names[i]) != INTF_TYPE_WIFI) {
					// Only wifi interfaces can be registrar
					PLATFORM_FREE_1905_INTERFACE_INFO(x);
					continue;
				}

				ERROR("Checking for registrar |%s| band |0x%x|", ifs_names[i], freq_band);
				INT8U *reg_mac = NULL;

				if (x->interface_type_data.ieee80211.role == IEEE80211_ROLE_AP) {
					switch (x->interface_type) {
					case MEDIA_TYPE_IEEE_802_11B_2_4_GHZ:
					case MEDIA_TYPE_IEEE_802_11G_2_4_GHZ:
					case MEDIA_TYPE_IEEE_802_11N_2_4_GHZ:
						{
							if (freq_band == IEEE80211_FREQUENCY_BAND_2_4_GHZ)
								reg_mac = DMregistrarMacGet(IEEE80211_FREQUENCY_BAND_2_4_GHZ);
							break;
						}
					case MEDIA_TYPE_IEEE_802_11A_5_GHZ:
					case MEDIA_TYPE_IEEE_802_11N_5_GHZ:
					case MEDIA_TYPE_IEEE_802_11AC_5_GHZ:
						{
							if (freq_band == IEEE80211_FREQUENCY_BAND_5_GHZ)
								reg_mac = DMregistrarMacGet(IEEE80211_FREQUENCY_BAND_5_GHZ);
							break;
						}
					case MEDIA_TYPE_IEEE_802_11AD_60_GHZ:
						{
							if (freq_band == IEEE80211_FREQUENCY_BAND_60_GHZ)
								reg_mac = DMregistrarMacGet(IEEE80211_FREQUENCY_BAND_60_GHZ);
							break;

						}
					}
				}
				if (reg_mac && (memcmp(&x->mac_address, reg_mac, 6) == 0))
					send_autoconfig_response = true;

				if (send_autoconfig_response) {
					ERROR("Interface %s is AP registrar of same band. Sending response...", ifs_names[i]);

					if (0 == send1905APAutoconfigurationResponsePacket(recv_intf,
											   c->message_id,
											   al_mac_address,
											   freq_band)) {
						WARNING("Could not send 'AP autoconfiguration response' message");
					}
				}
				PLATFORM_FREE_1905_INTERFACE_INFO(x);

			}

			PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);

			break;
		}
	case CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE:
		{
			// When a "AP-autoconfig response" is received then we have to
			// search for the first interface which is an unconfigured AP with
			// the same freq band as the one contained in the message and send
			// a AP-autoconfig WSC-M1

			INT8U *p;
			INT8U i;

			char **ifs_names;
			INT8U ifs_nr;

			INT8U supported_role_is_present;
			INT8U supported_role = IEEE80211_ROLE_AP;

			INT8U supported_freq_band_is_present;
			INT8U supported_freq_band = IEEE80211_FREQUENCY_BAND_2_4_GHZ;

			ERROR("<-- CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE (%s)", recv_intf);

			if (c->tlv_cstruct_list == NULL) {
				ERROR("Malformed structure.");
				break;
			}
			// First, parse the incomming packet to find out two things:
			//   parameters.
			// - The "supported role" contained in the "supported role TLV"
			//   (must be "REGISTRAR")
			// - The "supported freq band" contained in the "supported freq
			// band TLV" (must match the one of our local unconfigured
			// interface)
			i = 0;
			while (NULL != (p = c->tlv_cstruct_list[i])) {
				switch (*p) {
				case TLV_TYPE_SUPPORTED_ROLE:
					{
						struct supportedRoleTLV *t = (struct supportedRoleTLV *)p;

						supported_role_is_present = 1;
						supported_role = t->role;

						break;
					}
				case TLV_TYPE_SUPPORTED_FREQ_BAND:
					{
						struct supportedFreqBandTLV *t = (struct supportedFreqBandTLV *)p;

						supported_freq_band_is_present = 1;
						supported_freq_band = t->freq_band;

						break;
					}
				default:
					{
						WARNING("Unexpected TLV (%d) type inside CMDU", *p);
						break;
					}
				}
				i++;
			}

			// We must send the WSC TLV to the AL MAC of the node who
			// sent the response, however, this AL MAC is *not*
			// contained in the response. The only thing we can do at
			// this point is try to search our AL neighbors data base
			// for a matching MAC.
			//
			INT8U *dst_mac;
			INT8U *p_al;

			p_al = DMmacToAlMac(src_addr);

			if (p_al == NULL) {
				// The standard says we should always send to the AL
				// MAC address, however, in these cases, instead of
				// just dropping the packet, sending the response to
				// the 'src' address from the AUTOCONFIGURATION RESPONSE
				// seems the right thing to do.
				dst_mac = src_addr;
				WARNING("Unknown destination AL MAC. Using the 'src' MAC from the ACS RESPONSE (%02x:%02x:%02x:%02x:%02x:%02x)",
					dst_mac[0], dst_mac[1],
					dst_mac[2], dst_mac[3],
					dst_mac[4], dst_mac[5]);
			} else {
				dst_mac = p_al;
			}

			// Make sure that all needed parameters were present in the message
			if (0 == supported_role_is_present || 0 == supported_freq_band_is_present) {
				WARNING("More TLVs were expected inside this CMDU");
				return PROCESS_CMDU_KO;
			}

			if (supported_role != IEEE80211_ROLE_REGISTRAR) {
				WARNING("Unexpected 'searched role' |%d|", supported_role);
				return PROCESS_CMDU_KO;
			}
			if (!((supported_freq_band == IEEE80211_FREQUENCY_BAND_2_4_GHZ) ||
			      (supported_freq_band == IEEE80211_FREQUENCY_BAND_5_GHZ) ||
			      (supported_freq_band == IEEE80211_FREQUENCY_BAND_60_GHZ))) {
				WARNING("Unexpected freq_band |%d|", supported_freq_band);
				return PROCESS_CMDU_KO;
			}

			// If got Auto Config response means, there is another
			// registrar present in the network.
			// Notify the upper layers about this
			if (DMregistrarMacGet(supported_freq_band) != NULL) {
				ERROR("## Another registrar in network ##");
				notify_another_registrar(recv_intf, dst_mac, supported_freq_band);
				return PROCESS_CMDU_KO;
			}
			// Check our local interfaces, looking for an unconfigured AP
			// interface that matches the freq band.
			// If one is found, send the response.
			ifs_names = get_list_of_unconfigured_ap(&ifs_nr);
			for (i = 0; i < ifs_nr; i++) {
				struct interfaceInfo *x;

				x = PLATFORM_GET_1905_INTERFACE_INFO(ifs_names[i]);
				if (x == NULL) {
					WARNING("Could not retrieve info of interface %s", ifs_names[i]);
					continue;
				}

				if (x->interface_type_data.ieee80211.role != IEEE80211_ROLE_AP) {
					ERROR("Interface (%s) is not an AP", ifs_names[i]);
					PLATFORM_FREE_1905_INTERFACE_INFO(x);
					continue;
				}

				if (supported_freq_band != get_wifi_band(x)) {
					DEBUG("Freq band of intf(%s) unsupported, band|%d|", ifs_names[i], supported_freq_band);
					continue;
				}

				INT8U *m1;
				INT16U m1_size;
				void *key;

				INFO("Unconfigured AP(%s) on same freq band|%d|. Sending WSC-M1...", ifs_names[i], supported_freq_band);

				// Obtain WSC-M1 and send the WSC TLV
				if (wscBuildM1(ifs_names[i], &m1, &m1_size, &key)) {
					send1905APAutoconfigurationWSCPacket(recv_intf,
									     getNextMid(),
									     dst_mac, m1,
									     m1_size);
					PLATFORM_FREE_1905_INTERFACE_INFO(x);
					break;
				}
				PLATFORM_FREE_1905_INTERFACE_INFO(x);
			}
			update_registrar_in_network(supported_freq_band, dst_mac);
			FREE(p_al);
			FREE(ifs_names);

			break;
		}
	case CMDU_TYPE_AP_AUTOCONFIGURATION_WSC:
		{
			INT8U *p;
			INT8U i;

			INT8U *wsc_frame;
			INT16U wsc_frame_size;

			INT8U wsc_type;

			// When a "AP-autoconfig WSC" is received we first have to find out
			// if the contained message is M1 or M2.
			// If it is M1, send an M2 response.
			// If it is M2, apply the received configuration.

			ERROR("<-- CMDU_TYPE_AP_AUTOCONFIGURATION_WSC (%s)", recv_intf);

			if (c->tlv_cstruct_list == NULL) {
				ERROR("Malformed structure.");
				break;
			}

			wsc_frame = NULL;
			wsc_frame_size = 0;
			i = 0;
			while (NULL != (p = c->tlv_cstruct_list[i])) {
				switch (*p) {
				case TLV_TYPE_WSC:
					{
						struct wscTLV *t = (struct wscTLV *)p;

						wsc_frame = t->wsc_frame;
						wsc_frame_size = t->wsc_frame_size;

						break;
					}
				default:
					{
						WARNING("Unexpected TLV (%d) type inside CMDU", *p);
						break;
					}
				}
				i++;
			}

			// Make sure there was a WSC TLV in the message
			if (wsc_frame == NULL) {
				WARNING("More TLVs were expected inside this CMDU");
				return PROCESS_CMDU_KO;
			}

			wsc_type = wscGetType(wsc_frame, wsc_frame_size);

			if (wsc_type == WSC_TYPE_M2) {
				bool m2_processed_done = false;
				// Process it and apply the configuration to the corresponding
				// interface.
				// Don't know the which key to use to decrypt, try all keys
				// one by one
				char **ifs_names = NULL;
				INT8U ifs_nr = 0;

				ifs_names = get_list_of_unconfigured_ap(&ifs_nr);
				INFO("Got M2 checking un-configured iface nr(%d)", ifs_nr);

				if (ifs_nr == 0)
					return PROCESS_CMDU_KO;

				for (size_t i = 0; i < ifs_nr; i++) {
					INT8U *m1;
					INT16U m1_size;
					void *key;

					if (get_wsc_m1_key(ifs_names[i], &m1, &m1_size, &key))
						if (wscProcessM2(key, m1, m1_size, wsc_frame, wsc_frame_size)) {
							m2_processed_done = true;
							break;
						}
				}
				FREE(ifs_names);
				if (m2_processed_done)
					return PROCESS_CMDU_OK_TRIGGER_AP_SEARCH;

			} else if (wsc_type == WSC_TYPE_M1) {
				// We hadn't previously sent an M1 (ie. we are the registrar),
				// thus the contents of the just received message must be M1.
				//
				// Process it and send an M2 response.
				//
				INT8U *m2;
				INT16U m2_size;

				INT8U *dst_mac;
				INT8U *p;

				INFO("Got M1 sending M2");
				// We must send M2 to the AL MAC of the node who sent M1,
				// however, this AL MAC is *not* contained in M1.
				// The only thing we can do at this point is try to search our
				// AL neighbors data base for a matching MAC.
				p = DMmacToAlMac(src_addr);

				if (p == NULL) {
					// The standard says we should always send to the AL MAC
					// address, however, in these cases, instead of just
					// dropping the packet, sending M2 to the 'src' address
					// from the M1 seems the right thing to do.
					//
					dst_mac = src_addr;
					WARNING("Unknown destination AL MAC. Using the 'src' MAC from M1 (%02x:%02x:%02x:%02x:%02x:%02x)",
						dst_mac[0], dst_mac[1],
						dst_mac[2], dst_mac[3],
						dst_mac[4], dst_mac[5]);
				} else {
					dst_mac = p;
				}

				if (wscBuildM2(wsc_frame, wsc_frame_size, &m2,
					       &m2_size)) {
					send1905APAutoconfigurationWSCPacket(recv_intf,
									     getNextMid(),
									     dst_mac,
									     m2,
									     m2_size);
					wscFreeM2(m2, m2_size);
				}
				FREE(p);

			} else {
				WARNING("Unknown type of WSC message!");
			}

			break;
		}
	case CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW:
		{
			// When a "AP-autoconfig renew" is
			uint8_t i, supported_role, ifs_nr;
			uint8_t supported_role_is_present ,supported_freq_band_is_present;
			INT8U al_mac_address[6];
			INT8U *m1, *p;
			INT16U m1_size;
			char **ifs_names;
			void *key;


			memset(&al_mac_address, 0, 6);
			INFO("<-- CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW (%s)", recv_intf);

			if (c->tlv_cstruct_list == NULL) {
				ERROR("Malformed structure.");
				break;
			}

			supported_role_is_present = 0;
			supported_freq_band_is_present = 0;
			supported_role = IEEE80211_ROLE_REGISTRAR;
			i = 0;
			while (NULL != (p = c->tlv_cstruct_list[i])) {
				switch (*p) {
				case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
					{
						struct alMacAddressTypeTLV *t = (struct alMacAddressTypeTLV *)p;
						memcpy(al_mac_address, t->al_mac_address, 6);
						break;
					}
				case TLV_TYPE_SUPPORTED_ROLE:
					{
						struct supportedRoleTLV *t = (struct supportedRoleTLV *)p;
						supported_role_is_present = 1;
						supported_role = t->role;
						break;
					}
				case TLV_TYPE_SUPPORTED_FREQ_BAND:
					{
						supported_freq_band_is_present = 1;
						break;
					}
				default:
					{
						WARNING("Unexpected TLV (%d) type inside CMDU", *p);
						break;
					}
				}
				i++;
			}

			// Make sure that all needed parameters were present in the message
			if (hwaddr_is_zero(al_mac_address)
			    || supported_role_is_present == 0
			    || supported_freq_band_is_present == 0) {
				WARNING("More TLVs were expected inside this CMDU");
				return PROCESS_CMDU_KO;
			}

			if (supported_role != IEEE80211_ROLE_REGISTRAR) {
				WARNING("Unexpected 'supported role' |%d|", supported_role);
				return PROCESS_CMDU_KO;
			}

			ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);

			for (i = 0; i < ifs_nr; i++) {

				// Obtain WSC-M1 and send the WSC TLV
				struct interfaceInfo *x;
				uint8_t ap_freq_band;

				x = PLATFORM_GET_1905_INTERFACE_INFO(ifs_names[i]);
				if (x == NULL) {
					WARNING("Could not retrieve info of interface %s", ifs_names[i]);
					continue;
				}

				if (false == get_freq_band_from_media_type(x->interface_type, &ap_freq_band)) {
					// non wireless interface continue
					PLATFORM_FREE_1905_INTERFACE_INFO(x);
					continue;
				}

				platform_clear_bssid(ifs_names[i]);
				PLATFORM_FREE_1905_INTERFACE_INFO(x);
			}

			for (i = 0; i < ifs_nr; i++) {

				// Obtain WSC-M1 and send the WSC TLV
				struct interfaceInfo *x;
				uint8_t ap_freq_band;

				x = PLATFORM_GET_1905_INTERFACE_INFO(ifs_names[i]);
				if (x == NULL) {
					WARNING("Could not retrieve info of interface %s", ifs_names[i]);
					continue;
				}

				if (false == get_freq_band_from_media_type(x->interface_type, &ap_freq_band)) {
					// non wireless interface continue
					PLATFORM_FREE_1905_INTERFACE_INFO(x);
					continue;
				}

				if (wscBuildM1(ifs_names[i], &m1, &m1_size, &key)) {
					send1905APAutoconfigurationWSCPacket(recv_intf,
									     getNextMid(),
									     al_mac_address, m1,
									     m1_size);
					PLATFORM_FREE_1905_INTERFACE_INFO(x);
					break;
				}
				PLATFORM_FREE_1905_INTERFACE_INFO(x);
			}
			PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);
			break;
		}
	case CMDU_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION:
		{
			// According to "Section 9.2.2.2", when a "push button event
			// notification" is received we have to:
			//
			//   1. Transition *all* interfaces to POWER_STATE_PWR_ON
			//
			//   2. Start the "push button" configuration process in all those
			//     interfaces that:
			//       2.1 Are not 802.11
			//       2.2 Are 802.11 APs, configured as "registrars", but only if
			//           the received message did not contain 802.11 media type
			//           information.

			INT8U *p;
			INT8U i;

			char **ifs_names;
			INT8U ifs_nr;

			INT8U wifi_data_is_present;

			INT8U al_mac_address[6] = {0};

			wifi_data_is_present = 0;

			INFO("<-- CMDU_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION (%s)", recv_intf);

			if (c->tlv_cstruct_list == NULL) {
				ERROR("Malformed structure.");
				break;
			}
			// First, parse the incomming packet to find out if the 'push
			// button' event TLV contains 802.11 data.
			//
			i = 0;
			while (NULL != (p = c->tlv_cstruct_list[i])) {
				switch (*p) {
				case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
					{
						struct alMacAddressTypeTLV *t = (struct alMacAddressTypeTLV *)p;

						memcpy(al_mac_address, t->al_mac_address, 6);

						break;
					}
				case TLV_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION:
					{
						INT8U j;
						struct pushButtonEventNotificationTLV *t = (struct pushButtonEventNotificationTLV *)p;

						for (j = 0; j < t->media_types_nr; j++) {
							if (MEDIA_TYPE_IEEE_802_11B_2_4_GHZ == t->media_types[j].media_type ||
							    MEDIA_TYPE_IEEE_802_11G_2_4_GHZ == t->media_types[j].media_type ||
							    MEDIA_TYPE_IEEE_802_11A_5_GHZ == t->media_types[j].media_type ||
							    MEDIA_TYPE_IEEE_802_11N_2_4_GHZ == t->media_types[j].media_type ||
							    MEDIA_TYPE_IEEE_802_11N_5_GHZ == t->media_types[j].media_type ||
							    MEDIA_TYPE_IEEE_802_11AC_5_GHZ == t->media_types[j].media_type ||
							    MEDIA_TYPE_IEEE_802_11AD_60_GHZ == t->media_types[j].media_type ||
							    MEDIA_TYPE_IEEE_802_11AF_GHZ == t->media_types[j].media_type) {
								wifi_data_is_present = 1;
								break;
							}
						}

						break;
					}
				default:
					{
						WARNING("Unexpected TLV (%d) type inside CMDU", *p);
						break;
					}
				}
				i++;
			}

			if (hwaddr_is_zero(al_mac_address)) {
				WARNING("More TLVs were expected inside this CMDU");
				return PROCESS_CMDU_KO;
			}
			if (wifi_data_is_present == 0) {
				INFO("wifi data missing in 'pb event notification'. Skipping...");
				return PROCESS_CMDU_OK;
			}
			// Next, switch on all interfaces
			ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);

			DEBUG("Transitioning all local interfaces to POWER_ON");

#ifndef DO_NOT_ACCEPT_UNAUTHENTICATED_COMMANDS
			for (i = 0; i < ifs_nr; i++) {
				PLATFORM_SET_INTERFACE_POWER_MODE(ifs_names[i], INTERFACE_POWER_STATE_ON);
			}
#endif
			// Finally, for those non wifi interfaces (or a wifi interface whose
			// MAC address matches the network registrar MAC address), start
			// the "push button" configuration process.
			DEBUG("Starting 'push button' configuration process on all compatible interfaces");
			for (i = 0; i < ifs_nr; i++) {
				bool start_push_button_config = true;
				struct interfaceInfo *x;

				x = PLATFORM_GET_1905_INTERFACE_INFO(ifs_names[i]);
				if (x == NULL) {
					WARNING("Could not retrieve info of interface %s", ifs_names[i]);
					continue;
				}

				if (x->push_button_on_going == 2) {
					INFO("Interface (%s) doesn't support pbc|%d|", ifs_names[i], x->push_button_on_going);
					PLATFORM_FREE_1905_INTERFACE_INFO(x);
					continue;
				}
				if (false == _is_pbc_interface(x)) {
					INFO("Non wireless interface(%s) or pbc |%d| not supported", ifs_names[i], x->push_button_on_going);
					PLATFORM_FREE_1905_INTERFACE_INFO(x);
					continue;
				}

				INT8U *reg_mac = NULL;

				switch (x->interface_type) {
				case MEDIA_TYPE_IEEE_802_11B_2_4_GHZ:
				case MEDIA_TYPE_IEEE_802_11G_2_4_GHZ:
				case MEDIA_TYPE_IEEE_802_11N_2_4_GHZ:
					{
						reg_mac = DMregistrarMacGet(IEEE80211_FREQUENCY_BAND_2_4_GHZ);
						break;
					}
				case MEDIA_TYPE_IEEE_802_11A_5_GHZ:
				case MEDIA_TYPE_IEEE_802_11N_5_GHZ:
				case MEDIA_TYPE_IEEE_802_11AC_5_GHZ:
					{
						reg_mac = DMregistrarMacGet(IEEE80211_FREQUENCY_BAND_5_GHZ);
						break;
					}
				case MEDIA_TYPE_IEEE_802_11AD_60_GHZ:
				case MEDIA_TYPE_IEEE_802_11AF_GHZ:
					{
						reg_mac = DMregistrarMacGet(IEEE80211_FREQUENCY_BAND_60_GHZ);
						break;
					}
				}
				if (reg_mac && (memcmp(&x->mac_address, reg_mac, 6) == 0))
					start_push_button_config = true;

				PLATFORM_FREE_1905_INTERFACE_INFO(x);

				if (start_push_button_config) {
					INFO("Starting push button configuration process on interface %s", ifs_names[i]);
					if (PLATFORM_START_PUSH_BUTTON_CONFIGURATION(ifs_names[i], al_mac_address, c->message_id) == 0) {
						WARNING("Could not start 'push button' configuration process on interface", ifs_names[i]);
					}
				}
			}

			PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);

			break;
		}
	case CMDU_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION:
		{
			INFO("<-- CMDU_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION (%s)", recv_intf);
			// This is already broadcasted using ubus event ieee1905.event
			break;
		}
	case CMDU_TYPE_GENERIC_PHY_QUERY:
		{
			// When a "generic phy query" is received we must reply with the
			// list of local "generic" interfaces inside a "generic phy
			// response" CMDU.
			// Not that even if we don't have any "generic" interface (ie. its
			// 'media type' is "MEDIA_TYPE_UNKNOWN") the response will be sent
			// (containing a TLV that says there are "zero" generic interfaces)

			INT8U *dst_mac;
			INT8U *p;

			INFO("<-- CMDU_TYPE_GENERIC_PHY_QUERY (%s)", recv_intf);

			// We must send the response to the AL MAC of the node who sent the
			// query, however, this AL MAC is *not* contained in the query.
			// The only thing we can do at this point is try to search our AL
			// neighbors data base for a matching MAC.
			//
			p = DMmacToAlMac(src_addr);

			if (p == NULL) {
				// The standard says we should always send to the AL MAC
				// address, however, in these cases, instead of just dropping
				// the packet, sending the response to the 'src' address from
				// the TOPOLOGY QUERY seems the right thing to do.
				//
				dst_mac = src_addr;
				WARNING("Unknown destination AL MAC. Using the 'src' MAC from the GENERIC PHY QUERY (%02x:%02x:%02x:%02x:%02x:%02x)",
					dst_mac[0], dst_mac[1],
					dst_mac[2], dst_mac[3],
					dst_mac[4], dst_mac[5]);
			} else {
				dst_mac = p;
			}

			if (send1905GenericPhyResponsePacket(recv_intf, c->message_id, dst_mac) == 0) {
				WARNING("Could not send 'topology query' message");
			}

			FREE(p);
			break;
		}
	case CMDU_TYPE_GENERIC_PHY_RESPONSE:
		{
			// When a "generic phy response" is received we must update our
			// internal database (that keeps track of which 1905 devices are
			// present in the network)

			struct genericPhyDeviceInformationTypeTLV *t;

			INT8U *p;
			INT8U i;

			INFO("<-- CMDU_TYPE_GENERIC_PHY_RESPONSE (%s)", recv_intf);

			if (c->tlv_cstruct_list == NULL) {
				ERROR("Malformed structure.");
				break;
			}
			// Call "DMupdateGenericPhyInfo()" for the "generic phy
			// device information type TLV"  contained in this CMDU.
			//
			i = 0;
			t = NULL;
			while (NULL != (p = c->tlv_cstruct_list[i])) {
				switch (*p) {
				case TLV_TYPE_GENERIC_PHY_DEVICE_INFORMATION:
					{
						t = (struct genericPhyDeviceInformationTypeTLV *)p;
						break;
					}
				case TLV_TYPE_VENDOR_SPECIFIC:
					{
						// According to the standard, zero or more Vendor
						// Specific TLVs may be present.
						//
						free_1905_TLV_structure(p, true);
						break;
					}
				default:
					{
						WARNING("Unexpected TLV (%d) type inside CMDU", *p);

						free_1905_TLV_structure(p, true);
						break;
					}
				}
				i++;
			}

			if (t == NULL) {
				WARNING("More TLVs were expected inside this CMDU");
				return PROCESS_CMDU_KO;
			}

			DEBUG("Updating network devices database...");
			DMupdateNetworkDeviceInfo(t->al_mac_address,
						  0, NULL,
						  0, NULL, 0,
						  0, NULL, 0, 0, NULL, 0, 0, NULL, 0, 0, NULL, 0, 1, t, 0, NULL, 0, NULL, 0, NULL, 0, NULL, 0, NULL);

			// References to the TLVs cannot be freed by the caller (see the
			// comment in "case CMDU_TYPE_TOPOLOGY_RESPONSE:" to understand the
			// following two lines).
			//
			FREE(c->tlv_cstruct_list);
			c->tlv_cstruct_list = NULL;

			// Show all network devices (ie. print them through the logging
			// system)
			//
			DMdumpNetworkDevices(print_callback, print_debug, NULL);

			break;
		}
	case CMDU_TYPE_HIGHER_LAYER_QUERY:
		{
			// When a "high layer query" is received we must reply with the
			// list of items inside a "high layer response" CMDU.

			INT8U *dst_mac;
			INT8U *p;

			INFO("<-- CMDU_TYPE_HIGHER_LAYER_QUERY (%s)", recv_intf);

			// We must send the response to the AL MAC of the node who sent the
			// query, however, this AL MAC is *not* contained in the query.
			// The only thing we can do at this point is try to search our AL
			// neighbors data base for a matching MAC.
			//
			p = DMmacToAlMac(src_addr);

			if (p == NULL) {
				// The standard says we should always send to the AL MAC
				// address, however, in these cases, instead of just dropping
				// the packet, sending the response to the 'src' address from
				// the TOPOLOGY QUERY seems the right thing to do.
				//
				dst_mac = src_addr;
				WARNING("Unknown destination AL MAC. Using the 'src' MAC from the HIGH LAYER QUERY (%02x:%02x:%02x:%02x:%02x:%02x)",
					dst_mac[0], dst_mac[1],
					dst_mac[2], dst_mac[3],
					dst_mac[4], dst_mac[5]);
			} else {
				dst_mac = p;
			}

			if (send1905HighLayerResponsePacket(recv_intf, c->message_id, dst_mac) == 0) {
				WARNING("Could not send 'high layer response' message");
			}

			FREE(p);
			break;
		}
	case CMDU_TYPE_HIGHER_LAYER_RESPONSE:
		{
			// When a "high layer response" is received we must update our
			// internal database (that keeps track of which 1905 devices are
			// present in the network)

			struct x1905ProfileVersionTLV *profile = NULL;
			struct deviceIdentificationTypeTLV *identification = NULL;
			struct controlUrlTypeTLV *control_url = NULL;
			struct ipv4TypeTLV *ipv4 = NULL;
			struct ipv6TypeTLV *ipv6 = NULL;

			INT8U al_mac_address[6];
			INT8U al_mac_address_is_present;

			INT8U *p;
			INT8U i;

			INFO("<-- CMDU_TYPE_HIGHER_LAYER_RESPONSE (%s)", recv_intf);

			if (c->tlv_cstruct_list == NULL) {
				ERROR("Malformed structure.");
				break;
			}
			// Call "DMupdateGenericPhyInfo()" with each of the TLVs contained
			// in this CMDU
			//
			DEBUG("Updating network devices database...");

			i = 0;
			al_mac_address_is_present = 0;
			while (NULL != (p = c->tlv_cstruct_list[i])) {
				switch (*p) {
				case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
					{
						struct alMacAddressTypeTLV *t = (struct alMacAddressTypeTLV *)p;

						memcpy(al_mac_address, t->al_mac_address, 6);

						al_mac_address_is_present = 1;

						free_1905_TLV_structure(p, true);
						break;
					}
				case TLV_TYPE_1905_PROFILE_VERSION:
					{
						profile = (struct x1905ProfileVersionTLV *)p;
						break;
					}
				case TLV_TYPE_DEVICE_IDENTIFICATION:
					{
						identification = (struct deviceIdentificationTypeTLV *)p;
						break;
					}
				case TLV_TYPE_CONTROL_URL:
					{
						control_url = (struct controlUrlTypeTLV *)p;
						break;
					}
				case TLV_TYPE_IPV4:
					{
						ipv4 = (struct ipv4TypeTLV *)p;
						break;
					}
				case TLV_TYPE_IPV6:
					{
						ipv6 = (struct ipv6TypeTLV *)p;
						break;
					}
				default:
					{
						WARNING("Unexpected TLV (%d) type inside CMDU", *p);

						free_1905_TLV_structure(p, true);
						break;
					}
				}
				i++;
			}

			if (al_mac_address_is_present == 0) {
				WARNING("More TLVs were expected inside this CMDU");
				return PROCESS_CMDU_KO;
			}
			// Next, update the database. This will take care of duplicate
			// entries (and free the TLV if needed)
			//
			DEBUG("Updating network devices database...");
			DMupdateNetworkDeviceInfo(al_mac_address,
						  0, NULL,
						  0, NULL, 0,
						  0, NULL, 0,
						  0, NULL, 0, 0, NULL, 0, 0, NULL, 0, 0, NULL, 1, profile, 1, identification, 1, control_url, 1, ipv4, 1, ipv6);

			// References to the TLVs cannot be freed by the caller (see the
			// comment in "case CMDU_TYPE_TOPOLOGY_RESPONSE:" to understand the
			// following two lines).
			//
			FREE(c->tlv_cstruct_list);
			c->tlv_cstruct_list = NULL;

			// Show all network devices (ie. print them through the logging
			// system)
			//
			DMdumpNetworkDevices(print_callback, print_debug, NULL);

			break;
		}
	case CMDU_TYPE_INTERFACE_POWER_CHANGE_REQUEST:
		{
			// When an "interface power change" request is received we need to
			// set the local interfaces to the requested power modes and reply
			// back with the result of these operations

			struct interfacePowerChangeInformationTLV *t;

			INT8U *p;
			INT8U i;

			INFO("<-- CMDU_TYPE_INTERFACE_POWER_CHANGE_REQUEST (%s)", recv_intf);

			if (c->tlv_cstruct_list == NULL) {
				ERROR("Malformed structure.");
				break;
			}
			// Search for the "interface power change information type" TLV
			//
			i = 0;
			t = NULL;
			while (NULL != (p = c->tlv_cstruct_list[i])) {
				switch (*p) {
				case TLV_TYPE_INTERFACE_POWER_CHANGE_INFORMATION:
					{
						t = (struct interfacePowerChangeInformationTLV *)p;
						break;
					}
				default:
					{
						WARNING("Unexpected TLV (%d) type inside CMDU", *p);

						break;
					}
				}
				i++;
			}

			if (t == NULL) {
				WARNING("More TLVs were expected inside this CMDU");
				return PROCESS_CMDU_KO;
			}

			for (i = 0; i < t->power_change_interfaces_nr; i++) {
				INT8U r;
				INT8U results;

#ifndef DO_NOT_ACCEPT_UNAUTHENTICATED_COMMANDS
				r = PLATFORM_SET_INTERFACE_POWER_MODE(DMmacToInterfaceName(t->power_change_interfaces[i].interface_address),
								      t->power_change_interfaces[i].requested_power_state);
#else
				r = INTERFACE_POWER_RESULT_KO;
#endif

				switch (r) {
				case INTERFACE_POWER_RESULT_EXPECTED:
					{
						results = POWER_STATE_RESULT_COMPLETED;
						break;
					}
				case INTERFACE_POWER_RESULT_NO_CHANGE:
					{
						results = POWER_STATE_RESULT_NO_CHANGE;
						break;
					}
				case INTERFACE_POWER_RESULT_ALTERNATIVE:
					{
						results = POWER_STATE_RESULT_ALTERNATIVE_CHANGE;
						break;
					}
				case INTERFACE_POWER_RESULT_KO:
					{
						WARNING("  Could not set power mode on interface %s",
							DMmacToInterfaceName(t->power_change_interfaces[i].interface_address));
						results = POWER_STATE_RESULT_NO_CHANGE;
						break;
					}
				default:
					{
						WARNING("  Unknown power mode return value: %d", r);
						results = POWER_STATE_RESULT_NO_CHANGE;
						break;
					}
				}

				DEBUG("  Setting interface #%d %s (%02x:%02x:%02x:%02x:%02x:%02x) to %s --> %s", i,
				      DMmacToInterfaceName(t->power_change_interfaces[i].interface_address),
				      t->power_change_interfaces[i].interface_address[0], t->power_change_interfaces[i].interface_address[1],
				      t->power_change_interfaces[i].interface_address[2], t->power_change_interfaces[i].interface_address[3],
				      t->power_change_interfaces[i].interface_address[4], t->power_change_interfaces[i].interface_address[5],
				      t->power_change_interfaces[i].requested_power_state ==
				      POWER_STATE_REQUEST_OFF ? "POWER OFF" : t->power_change_interfaces[i].requested_power_state ==
				      POWER_STATE_REQUEST_ON ? "POWER ON" : t->power_change_interfaces[i].requested_power_state ==
				      POWER_STATE_REQUEST_SAVE ? "POWER SAVE" : "Unknown",
				      results == POWER_STATE_RESULT_COMPLETED ? "Completed" : results == POWER_STATE_RESULT_NO_CHANGE ? "No change" : results ==
				      POWER_STATE_RESULT_ALTERNATIVE_CHANGE ? "Alternative change" : "Unknown");
			}

			break;
		}
	case CMDU_TYPE_INTERFACE_POWER_CHANGE_RESPONSE:
		{
			// When an "interface power change" response is received we don't
			// need to do anything special. Simply log the event.

			struct interfacePowerChangeStatusTLV *t;

			INT8U *p;
			INT8U i;

			INFO("<-- CMDU_TYPE_INTERFACE_POWER_CHANGE_RESPONSE (%s)", recv_intf);

			if (c->tlv_cstruct_list == NULL) {
				ERROR("Malformed structure.");
				break;
			}
			// Search for the "interface power change status" TLV
			//
			i = 0;
			t = NULL;
			while (NULL != (p = c->tlv_cstruct_list[i])) {
				switch (*p) {
				case TLV_TYPE_INTERFACE_POWER_CHANGE_STATUS:
					{
						t = (struct interfacePowerChangeStatusTLV *)p;
						break;
					}
				default:
					{
						WARNING("Unexpected TLV (%d) type inside CMDU", *p);

						break;
					}
				}
				i++;
			}

			if (t == NULL) {
				WARNING("More TLVs were expected inside this CMDU");
				return PROCESS_CMDU_KO;
			}

			for (i = 0; i < t->power_change_interfaces_nr; i++) {
				DEBUG("  Interface #%d %s (%02x:%02x:%02x:%02x:%02x:%02x) --> %s", i,
				      DMmacToInterfaceName(t->power_change_interfaces[i].interface_address),
				      t->power_change_interfaces[i].interface_address[0], t->power_change_interfaces[i].interface_address[1],
				      t->power_change_interfaces[i].interface_address[2], t->power_change_interfaces[i].interface_address[3],
				      t->power_change_interfaces[i].interface_address[4], t->power_change_interfaces[i].interface_address[5],
				      t->power_change_interfaces[i].result ==
				      POWER_STATE_RESULT_COMPLETED ? "Completed" : t->power_change_interfaces[i].result ==
				      POWER_STATE_RESULT_NO_CHANGE ? "No change" : t->power_change_interfaces[i].result ==
				      POWER_STATE_RESULT_ALTERNATIVE_CHANGE ? "Alternative change" : "Unknown");
			}

			break;
		}

	default:
		{
			break;
		}
	}

	return PROCESS_CMDU_OK;
}

INT8U processLlpdPayload(struct PAYLOAD *payload, INT8U *receiving_interface_addr)
{
	INT8U *p;
	INT8U i;

	INT8U al_mac_address[6] = {0};
	INT8U mac_address[6] = {0};

	if (payload == NULL) {
		return 0;
	}

	INFO("<-- LLDP BRIDGE DISCOVERY (%s)", DMmacToInterfaceName(receiving_interface_addr));

	// We need to update the data model structure, which keeps track
	// of local interfaces, neighbors, and neighbors' interfaces, and
	// what type of discovery messages ("topology discovery" and/or
	// "bridge discovery") have been received on each link.

	// First, extract the AL MAC and MAC addresses of the interface
	// which transmitted this bridge discovery message
	//
	i = 0;
	while (NULL != (p = payload->tlv_cstruct_list[i])) {
		switch (*p) {
		case TLV_TYPE_CHASSIS_ID:
			{
				struct chassisIdTLV *t = (struct chassisIdTLV *)p;

				if (t->chassis_id_subtype == CHASSIS_ID_TLV_SUBTYPE_MAC_ADDRESS) {
					memcpy(al_mac_address, t->chassis_id, 6);
				}

				break;
			}
		case TLV_TYPE_MAC_ADDRESS_TYPE:
			{
				struct portIdTLV *t = (struct portIdTLV *)p;

				if (t->port_id_subtype == PORT_ID_TLV_SUBTYPE_MAC_ADDRESS) {
					memcpy(mac_address, t->port_id, 6);
				}

				break;
			}
		case TLV_TYPE_TIME_TO_LIVE:
			{
				break;
			}
		default:
			{
				DEBUG("Ignoring TLV type %d", *p);
				break;
			}
		}
		i++;
	}

	// Make sure that both the AL MAC and MAC addresses were contained
	// in the CMDU
	//
	if (hwaddr_is_zero(al_mac_address) || hwaddr_is_zero(mac_address)) {
		WARNING("More TLVs were expected inside this LLDP message");
		return 0;
	}

	DEBUG("AL MAC address = %02x:%02x:%02x:%02x:%02x:%02x", al_mac_address[0], al_mac_address[1], al_mac_address[2], al_mac_address[3], al_mac_address[4],
	      al_mac_address[5]);
	DEBUG("MAC    address = %02x:%02x:%02x:%02x:%02x:%02x", mac_address[0], mac_address[1], mac_address[2], mac_address[3], mac_address[4], mac_address[5]);

	// Finally, update the data model
	//
	if (DMupdateDiscoveryTimeStamps(receiving_interface_addr, al_mac_address, mac_address, TIMESTAMP_BRIDGE_DISCOVERY, NULL) == 0) {
		WARNING("Problems updating data model with topology response TLVs");
		return 0;
	}

	return 1;
}
