/*
 * al_send.c: IEEE-1905 AL packet send handling
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "utils.h"
#include <libubus.h>

#include <stdarg.h>		// va_list
  // NOTE: This is part of the C standard, thus *all* platforms should have it
  // available... and that's why this include can exist in this "platform
  // independent" file

#include "al_send.h"
#include "al_datamodel.h"

#include "cmdu_visit.h"
#include "lldp_tlvs.h"
#include "lldp_payload.h"

#include "platform_os.h"
#include "platform_interfaces.h"

#include "get_tlvs.h"
#include "al_ubus.h"

////////////////////////////////////////////////////////////////////////////////
// Private functions and data
////////////////////////////////////////////////////////////////////////////////

//******************************************************************************
//******* Functions to fill (and free) TLVs with local data ********************
//******************************************************************************
//
// Note that not *all* types of TLVs have a corresponding function in this
// section. Only those that either:
//
//   a) Are called from more than one place.
//   b) In order to be filled, the local device/node needs to be queried.
//
// Acording to these rules, some of the TLVs that do *not* have a corresponding
// function in this section are, for example, all "power change" related TLVs,
// LLDP TLVS, etc... These will be manually "filled" and "freed" in the specific
// "send*()" function that makes use of them.


// Modify the provided pointers so that they now point to a list of pointers to
// "non1905NeighborDeviceListTLV" and "neighborDeviceListTLV" structures filled
// with all the pertaining information retrieved from the local device.
//
// Example: this is how you would use this function:
//
//   struct non1905NeighborDeviceListTLV **a;   INT8U a_nr;
//   struct neighborDeviceListTLV        **b;   INT8U b_nr;
//
//   obtainLocalNeighborsTLV(&a, &a_nr, &b, &b_nr);
//
//   // a[0] -------> ptr to the first "non1905NeighborDeviceListTLV" structure
//   // a[1] -------> ptr to the second "non1905NeighborDeviceListTLV" structure
//   // ...
//   // a[a_nr-1] --> ptr to the last "non1905NeighborDeviceListTLV" structure
//
//   // b[0] -------> ptr to the first "neighborDeviceListTLV" structure
//   // b[1] -------> ptr to the second "neighborDeviceListTLV" structure
//   // ...
//   // b[b_nr-1] --> ptr to the last "neighborDeviceListTLV" structure
//
void obtainLocalNeighborsTLV(struct non1905NeighborDeviceListTLV ***non_1905_neighbors,
			     INT8U *non_1905_neighbors_nr,
			     struct neighborDeviceListTLV ***neighbors,
			     INT8U *neighbors_nr)
{
	char **interfaces_names;
	INT8U interfaces_names_nr;
	INT8U i, j, k;

	*non_1905_neighbors = NULL;
	*neighbors = NULL;

	*non_1905_neighbors_nr = 0;
	*neighbors_nr = 0;

	INT8U(*al_mac_addresses)[6];
	INT8U al_mac_addresses_nr;

	interfaces_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&interfaces_names_nr);

	for (i = 0; i < interfaces_names_nr; i++) {
		struct interfaceInfo *x;

		struct non1905NeighborDeviceListTLV *no;
		struct neighborDeviceListTLV *yes;

		x = PLATFORM_GET_1905_INTERFACE_INFO(interfaces_names[i]);
		if (x == NULL) {
			WARNING("Could not retrieve neighbors of interface %s", interfaces_names[i]);
			continue;
		}

		al_mac_addresses = DMgetListOfInterfaceNeighbors(interfaces_names[i], &al_mac_addresses_nr);

		CALLOC(no, 1);
		CALLOC(yes, 1);

		no->tlv_type = TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST;
		no->local_mac_address[0] = x->mac_address[0];
		no->local_mac_address[1] = x->mac_address[1];
		no->local_mac_address[2] = x->mac_address[2];
		no->local_mac_address[3] = x->mac_address[3];
		no->local_mac_address[4] = x->mac_address[4];
		no->local_mac_address[5] = x->mac_address[5];
		no->non_1905_neighbors_nr = 0;
		no->non_1905_neighbors = NULL;

		yes->tlv_type = TLV_TYPE_NEIGHBOR_DEVICE_LIST;
		yes->local_mac_address[0] = x->mac_address[0];
		yes->local_mac_address[1] = x->mac_address[1];
		yes->local_mac_address[2] = x->mac_address[2];
		yes->local_mac_address[3] = x->mac_address[3];
		yes->local_mac_address[4] = x->mac_address[4];
		yes->local_mac_address[5] = x->mac_address[5];
		yes->neighbors_nr = 0;
		yes->neighbors = NULL;

		// Decide if each neighbor is a 1905 or a non-1905 neighbor
		//
		if (x->neighbor_mac_addresses_nr != INTERFACE_NEIGHBORS_UNKNOWN) {
			INT8U *al_mac_address_has_been_reported = NULL;

			// Keep track of all the AL MACs that the interface reports he is
			// seeing.
			if (al_mac_addresses_nr != 0) {
				// Originally, none of the neighbors in the data model has been
				// reported...
				CALLOC(al_mac_address_has_been_reported, al_mac_addresses_nr);
			}

			for (j = 0; j < x->neighbor_mac_addresses_nr; j++) {
				INT8U *al_mac;
				INT8U k;

				al_mac = DMmacToAlMac(x->neighbor_mac_addresses[j]);

				if (al_mac == NULL) {
					// Non-1905 neighbor

					INT8U already_added;

					// Make sure it has not already been added
					//
					already_added = 0;
					for (k = 0; k < no->non_1905_neighbors_nr; k++) {
						if (memcmp(x->neighbor_mac_addresses[j], no->non_1905_neighbors[k].mac_address, 6) == 0) {
							already_added = 1;
							break;
						}
					}

					if (already_added == 0) {
						// This is a new neighbor
						//
						if (no->non_1905_neighbors_nr == 0) {
							CALLOC(no->non_1905_neighbors, 1);
						} else {
							REALLOC(no->non_1905_neighbors, (no->non_1905_neighbors_nr + 1));
						}

						no->non_1905_neighbors[no->non_1905_neighbors_nr].mac_address[0] = x->neighbor_mac_addresses[j][0];
						no->non_1905_neighbors[no->non_1905_neighbors_nr].mac_address[1] = x->neighbor_mac_addresses[j][1];
						no->non_1905_neighbors[no->non_1905_neighbors_nr].mac_address[2] = x->neighbor_mac_addresses[j][2];
						no->non_1905_neighbors[no->non_1905_neighbors_nr].mac_address[3] = x->neighbor_mac_addresses[j][3];
						no->non_1905_neighbors[no->non_1905_neighbors_nr].mac_address[4] = x->neighbor_mac_addresses[j][4];
						no->non_1905_neighbors[no->non_1905_neighbors_nr].mac_address[5] = x->neighbor_mac_addresses[j][5];

						no->non_1905_neighbors_nr++;
					}
				} else {
					// 1905 neighbor

					INT8U already_added;

					// Mark this AL MAC as reported
					//
					for (k = 0; k < al_mac_addresses_nr; k++) {
						if (memcmp(al_mac, al_mac_addresses[k], 6) == 0) {
							al_mac_address_has_been_reported[k] = 1;
							break;
						}
					}

					// Make sure it has not already been added
					//
					already_added = 0;
					for (k = 0; k < yes->neighbors_nr; k++) {
						if (memcmp(al_mac, yes->neighbors[k].mac_address, 6) == 0) {
							already_added = 1;
							break;
						}
					}

					if (already_added == 0) {
						// This is a new neighbor
						//
						if (yes->neighbors_nr == 0) {
							CALLOC(yes->neighbors, 1);
						} else {
							REALLOC(yes->neighbors, (yes->neighbors_nr + 1));
						}

						yes->neighbors[yes->neighbors_nr].mac_address[0] = al_mac[0];
						yes->neighbors[yes->neighbors_nr].mac_address[1] = al_mac[1];
						yes->neighbors[yes->neighbors_nr].mac_address[2] = al_mac[2];
						yes->neighbors[yes->neighbors_nr].mac_address[3] = al_mac[3];
						yes->neighbors[yes->neighbors_nr].mac_address[4] = al_mac[4];
						yes->neighbors[yes->neighbors_nr].mac_address[5] = al_mac[5];
						yes->neighbors[yes->neighbors_nr].bridge_flag = DMisNeighborBridged(interfaces_names[i], al_mac);

						yes->neighbors_nr++;
					}

					FREE(al_mac);
				}
			}
			PLATFORM_FREE_1905_INTERFACE_INFO(x);

			// Update the datamodel so that those neighbours whose MAC addresses
			// have not been reported are removed.
			// This will speed up the "removal" of nodes.
			//
			for (j = 0; j < al_mac_addresses_nr; j++) {
				if (al_mac_address_has_been_reported[j] == 0) {
					DMremoveALNeighborFromInterface(al_mac_addresses[j], interfaces_names[i]);
					DMrunGarbageCollector();
				}
			}
			if (al_mac_addresses_nr > 0 && NULL != al_mac_address_has_been_reported) {
				FREE(al_mac_address_has_been_reported);
			}
		} else {
			// The interface reports that it has no way of knowing which MAC
			// neighbors are connected to it.
			// In these cases, *at least* the already known 1905 neighbors
			// (which were discovered by us -not the platform- thanks to the
			// topology discovery process) should be returned.

			for (j = 0; j < al_mac_addresses_nr; j++) {
				INT8U already_added;

				// Make sure it has not already been added
				//
				already_added = 0;
				for (k = 0; k < yes->neighbors_nr; k++) {
					if (memcmp(al_mac_addresses[j], yes->neighbors[k].mac_address, 6) == 0) {
						already_added = 1;
						break;
					}
				}

				if (already_added == 0) {
					// This is a new neighbor
					//
					if (yes->neighbors_nr == 0) {
						CALLOC(yes->neighbors, 1);
					} else {
						REALLOC(yes->neighbors, (yes->neighbors_nr + 1));
					}

					yes->neighbors[yes->neighbors_nr].mac_address[0] = al_mac_addresses[j][0];
					yes->neighbors[yes->neighbors_nr].mac_address[1] = al_mac_addresses[j][1];
					yes->neighbors[yes->neighbors_nr].mac_address[2] = al_mac_addresses[j][2];
					yes->neighbors[yes->neighbors_nr].mac_address[3] = al_mac_addresses[j][3];
					yes->neighbors[yes->neighbors_nr].mac_address[4] = al_mac_addresses[j][4];
					yes->neighbors[yes->neighbors_nr].mac_address[5] = al_mac_addresses[j][5];
					yes->neighbors[yes->neighbors_nr].bridge_flag = DMisNeighborBridged(interfaces_names[i], al_mac_addresses[j]);

					yes->neighbors_nr++;

				}
			}
		}
		FREE(al_mac_addresses);

		// At this point we have, for this particular interface, all the
		// non 1905 neighbors in "no" and all 1905 neighbors in "yes".
		// We just need to add "no" and "yes" to the "non_1905_neighbors"
		// and "neighbors" lists and proceed to the next interface.
		//
		if (no->non_1905_neighbors_nr > 0) {
			// Add this to the list of non-1905 neighbor TLVs
			//
			if (*non_1905_neighbors_nr == 0) {
				CALLOC(*non_1905_neighbors, 1);
			} else {
				REALLOC(*non_1905_neighbors, (*non_1905_neighbors_nr + 1));
			}

			(*non_1905_neighbors)[*non_1905_neighbors_nr] = no;
			(*non_1905_neighbors_nr)++;
		} else {
			FREE(no);
		}

		if (yes->neighbors_nr > 0) {
			// Add this to the list of non-1905 neighbor TLVs
			//
			if (*neighbors_nr == 0) {
				CALLOC(*neighbors, 1);
			} else {
				REALLOC(*neighbors, (*neighbors_nr + 1));
			}

			(*neighbors)[*neighbors_nr] = yes;
			(*neighbors_nr)++;
		} else {
			FREE(yes);
		}
	}

	PLATFORM_FREE_LIST_OF_1905_INTERFACES(interfaces_names, interfaces_names_nr);
}

// Free the contents of the pointers filled by a previous call to
// "obtainLocalNeighborsTLV()".
// This function is called with the same arguments as
// "obtainLocalNeighborsTLV()"
//
void freeLocalNeighborsTLV(struct non1905NeighborDeviceListTLV ***non_1905_neighbors,
			   INT8U *non_1905_neighbors_nr,
			   struct neighborDeviceListTLV ***neighbors,
			   INT8U *neighbors_nr)
{
	INT8U i;

	if (*non_1905_neighbors_nr > 0) {
		for (i = 0; i < *non_1905_neighbors_nr; i++) {
			free_1905_TLV_structure((INT8U *)(*non_1905_neighbors)[i], true);
		}
		FREE(*non_1905_neighbors);
	}
	if (*neighbors_nr > 0) {
		for (i = 0; i < *neighbors_nr; i++) {
			free_1905_TLV_structure((INT8U *)(*neighbors)[i], true);
		}
		FREE(*neighbors);
	}
}

// Return a list of Tx metrics TLV and/or a list of Rx metrics TLV involving
// the local node and the neighbor whose AL MAC address matches
// 'specific_neighbor'.
//
// 'destination' can be either 'LINK_METRIC_QUERY_TLV_ALL_NEIGHBORS' (in which
// case argument 'specific_neighbor' is ignored) or
// 'LINK_METRIC_QUERY_TLV_SPECIFIC_NEIGHBOR' (in which case 'specific_neighbor'
// is the AL MAC of the 1905 node at the other end of the link whose metrics
// are being reported.
//
// 'metrics_type' can be 'LINK_METRIC_QUERY_TLV_TX_LINK_METRICS_ONLY',
// 'LINK_METRIC_QUERY_TLV_RX_LINK_METRICS_ONLY' or
// 'LINK_METRIC_QUERY_TLV_BOTH_TX_AND_RX_LINK_METRICS'.
//
// 'tx' is an output argument which will be set to NULL if 'metrics_type' is
// set to 'LINK_METRIC_QUERY_TLV_RX_LINK_METRICS_ONLY' or point to a list of
// 1 or more (depending on the value of 'destination') pointers to Tx metrics
// TLVs.
//
// 'rx' is an output argument which will be set to NULL if 'metrics_type' is
// set to 'LINK_METRIC_QUERY_TLV_TX_LINK_METRICS_ONLY' or point to a list of
// 1 or more (depending on the value of 'destination') pointers to Rx metrics
// TLVs.
//
// 'nr' is an output argument set to the number of elements if rx and/or tx.
//
// If there is a problem (example: a specific neighbor was not found), this
// function returns '0', otherwise it returns '1'.
//
void obtainLocalMetricsTLVs(INT8U destination, INT8U *specific_neighbor, INT8U metrics_type,
			    struct transmitterLinkMetricTLV ***tx,
			    struct receiverLinkMetricTLV ***rx, INT8U *nr)
{
	INT8U(*al_mac_addresses)[6];
	INT8U al_mac_addresses_nr;

	struct transmitterLinkMetricTLV **tx_tlvs;
	struct receiverLinkMetricTLV **rx_tlvs;

	INT8U total_tlvs;
	INT8U i, j;

	al_mac_addresses = DMgetListOfNeighbors(&al_mac_addresses_nr);

	// We will need either 1 or 'al_mac_addresses_nr' Rx and/or Tx TLVs,
	// depending on the value of the 'destination' argument (ie. one Rx and/or
	// Tx TLV for each neighbor whose metrics we are going to report)
	//
	rx_tlvs = NULL;
	tx_tlvs = NULL;
	if (al_mac_addresses_nr > 0) {
		if (destination == LINK_METRIC_QUERY_TLV_ALL_NEIGHBORS) {
			if (LINK_METRIC_QUERY_TLV_TX_LINK_METRICS_ONLY == metrics_type
			    || LINK_METRIC_QUERY_TLV_BOTH_TX_AND_RX_LINK_METRICS == metrics_type) {
				CALLOC(tx_tlvs, al_mac_addresses_nr);
			}
			if (LINK_METRIC_QUERY_TLV_RX_LINK_METRICS_ONLY == metrics_type
			    || LINK_METRIC_QUERY_TLV_BOTH_TX_AND_RX_LINK_METRICS == metrics_type) {
				CALLOC(rx_tlvs, al_mac_addresses_nr);
			}
		} else {
			if (LINK_METRIC_QUERY_TLV_TX_LINK_METRICS_ONLY == metrics_type
			    || LINK_METRIC_QUERY_TLV_BOTH_TX_AND_RX_LINK_METRICS == metrics_type) {
				CALLOC(tx_tlvs, 1);
			}
			if (LINK_METRIC_QUERY_TLV_RX_LINK_METRICS_ONLY == metrics_type
			    || LINK_METRIC_QUERY_TLV_BOTH_TX_AND_RX_LINK_METRICS == metrics_type) {
				CALLOC(rx_tlvs, 1);
			}
		}
	}
	// Next, for each neighbor, fill the corresponding TLV structure (Rx, Tx or
	// both) that contains the information regarding all possible links that
	// "join" our local node with that neighbor.
	//
	total_tlvs = 0;
	for (i = 0; i < al_mac_addresses_nr; i++) {
		INT8U(*remote_macs)[6];
		char **local_interfaces;
		INT8U links_nr;

		// Check if we are really interested in obtaining metrics information
		// regarding this particular neighbor
		//
		if (LINK_METRIC_QUERY_TLV_SPECIFIC_NEIGHBOR == destination &&
		    0 != memcmp(al_mac_addresses[i], specific_neighbor, 6)
		    ) {
			// Not interested
			//
			continue;
		}
		// Obtain the list of "links" that connect our AL node with this
		// specific neighbor.
		//
		remote_macs = DMgetListOfLinksWithNeighbor(al_mac_addresses[i], &local_interfaces, &links_nr);

		if (links_nr > 0) {

			// If there are 1 or more links between the local node and the
			// neighbor, first fill the TLV "header"
			//
			if (tx_tlvs != NULL) {
				CALLOC(tx_tlvs[total_tlvs], 1);

				tx_tlvs[total_tlvs]->tlv_type = TLV_TYPE_TRANSMITTER_LINK_METRIC;
				memcpy(tx_tlvs[total_tlvs]->local_al_address, DMalMacGet(), 6);
				memcpy(tx_tlvs[total_tlvs]->neighbor_al_address, al_mac_addresses[i], 6);
				tx_tlvs[total_tlvs]->transmitter_link_metrics_nr = links_nr;
				CALLOC(tx_tlvs[total_tlvs]->transmitter_link_metrics, links_nr);
			}
			if (rx_tlvs != NULL) {
				CALLOC(rx_tlvs[total_tlvs], 1);

				rx_tlvs[total_tlvs]->tlv_type = TLV_TYPE_RECEIVER_LINK_METRIC;
				memcpy(rx_tlvs[total_tlvs]->local_al_address, DMalMacGet(), 6);
				memcpy(rx_tlvs[total_tlvs]->neighbor_al_address, al_mac_addresses[i], 6);
				rx_tlvs[total_tlvs]->receiver_link_metrics_nr = links_nr;
				CALLOC(rx_tlvs[total_tlvs]->receiver_link_metrics, links_nr);
			}
			// ...and then, for each link, fill the specific link information:
			//
			for (j = 0; j < links_nr; j++) {
				struct interfaceInfo *f;
				struct linkMetrics *l, linkm;

				memset(&linkm, 0, sizeof(struct linkMetrics));
				l = &linkm;
				if (PLATFORM_GET_LINK_METRICS(local_interfaces[j], remote_macs[j], l) == false) {
					continue;
				}

				f = PLATFORM_GET_1905_INTERFACE_INFO(local_interfaces[j]);

				if (tx_tlvs != NULL) {
					memcpy(tx_tlvs[total_tlvs]->transmitter_link_metrics[j].local_interface_address,
					       DMinterfaceNameToMac(local_interfaces[j]), 6);
					memcpy(tx_tlvs[total_tlvs]->transmitter_link_metrics[j].neighbor_interface_address, remote_macs[j], 6);

					if (f == NULL) {
						tx_tlvs[total_tlvs]->transmitter_link_metrics[j].intf_type = MEDIA_TYPE_UNKNOWN;
					} else {
						tx_tlvs[total_tlvs]->transmitter_link_metrics[j].intf_type = f->interface_type;
					}
					tx_tlvs[total_tlvs]->transmitter_link_metrics[j].bridge_flag =
					    DMisLinkBridged(local_interfaces[j], al_mac_addresses[i], remote_macs[j]);

					tx_tlvs[total_tlvs]->transmitter_link_metrics[j].packet_errors = l->tx_packet_errors;
					tx_tlvs[total_tlvs]->transmitter_link_metrics[j].transmitted_packets = l->tx_packet_ok;
					tx_tlvs[total_tlvs]->transmitter_link_metrics[j].mac_throughput_capacity = l->tx_max_xput;
					tx_tlvs[total_tlvs]->transmitter_link_metrics[j].link_availability = l->tx_link_availability;
					tx_tlvs[total_tlvs]->transmitter_link_metrics[j].phy_rate = l->tx_phy_rate;
				}

				if (rx_tlvs != NULL) {
					memcpy(rx_tlvs[total_tlvs]->receiver_link_metrics[j].local_interface_address, DMinterfaceNameToMac(local_interfaces[j]),
					       6);
					memcpy(rx_tlvs[total_tlvs]->receiver_link_metrics[j].neighbor_interface_address, remote_macs[j], 6);

					if (f == NULL) {
						rx_tlvs[total_tlvs]->receiver_link_metrics[j].intf_type = MEDIA_TYPE_UNKNOWN;
					} else {
						rx_tlvs[total_tlvs]->receiver_link_metrics[j].intf_type = f->interface_type;
					}

					rx_tlvs[total_tlvs]->receiver_link_metrics[j].packet_errors = l->rx_packet_errors;
					rx_tlvs[total_tlvs]->receiver_link_metrics[j].packets_received = l->rx_packet_ok;
					rx_tlvs[total_tlvs]->receiver_link_metrics[j].rssi = l->rx_rssi;
				}

				if (f != NULL) {
					PLATFORM_FREE_1905_INTERFACE_INFO(f);
				}
			}
			total_tlvs++;
		}

		DMfreeListOfLinksWithNeighbor(remote_macs, local_interfaces, links_nr);
	}

	FREE(al_mac_addresses);

	if (LINK_METRIC_QUERY_TLV_SPECIFIC_NEIGHBOR == destination && total_tlvs == 0) {
		// Specific neighbor not found
		//
		*nr = 0;
		*tx = NULL;
		*rx = NULL;
	} else {
		*nr = total_tlvs;
		*tx = tx_tlvs;
		*rx = rx_tlvs;
	}
}

// Free the contents of the pointers filled by a previous call to
// "obtainLocalMetricsTLVs()".
// This function is called with the same three last arguments as
// "obtainLocalMetricsTLVs()"
//
void freeLocalMetricsTLVs(struct transmitterLinkMetricTLV ***tx_tlvs,
			  struct receiverLinkMetricTLV ***rx_tlvs,
			  INT8U *total_tlvs)
{
	INT8U i;

	if (NULL != tx_tlvs && NULL != *tx_tlvs) {
		for (i = 0; i < *total_tlvs; i++) {
			free_1905_TLV_structure((INT8U *)(*tx_tlvs)[i], true);
		}
		FREE(*tx_tlvs);
	}
	if (NULL != rx_tlvs && NULL != *rx_tlvs) {
		for (i = 0; i < *total_tlvs; i++) {
			free_1905_TLV_structure((INT8U *)(*rx_tlvs)[i], true);
		}
		FREE(*rx_tlvs);
	}
}

// Given two pointers to a preallocated "ipv4TypeTLV" and "ipv6TypeTLV"
// structures, fill them with all the pertaining information retrieved from the
// local device.
//
void obtainLocalIpsTLVs(struct ipv4TypeTLV *ipv4, struct ipv6TypeTLV *ipv6)
{
	char **ifs_names;
	INT8U ifs_nr;

	INT8U i, j;

	ipv4->tlv_type = TLV_TYPE_IPV4;
	ipv4->ipv4_interfaces_nr = 0;
	ipv6->tlv_type = TLV_TYPE_IPV6;
	ipv6->ipv6_interfaces_nr = 0;

	ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);

	for (i = 0; i < ifs_nr; i++) {
		struct interfaceInfo *y;

		y = PLATFORM_GET_1905_INTERFACE_INFO(ifs_names[i]);
		if (y == NULL) {
			WARNING("Could not retrieve info of interface %s", ifs_names[i]);
			continue;
		}

		if (y->ipv4_nr != 0) {
			if (ipv4->ipv4_interfaces_nr == 0) {
				CALLOC(ipv4->ipv4_interfaces, 1);
			} else {
				CALLOC(ipv4->ipv4_interfaces, (ipv4->ipv4_interfaces_nr + 1));
			}

			memcpy(ipv4->ipv4_interfaces[ipv4->ipv4_interfaces_nr].mac_address, y->mac_address, 6);
			ipv4->ipv4_interfaces[ipv4->ipv4_interfaces_nr].ipv4_nr = y->ipv4_nr;
			CALLOC(ipv4->ipv4_interfaces[ipv4->ipv4_interfaces_nr].ipv4, y->ipv4_nr);
			for (j = 0; j < y->ipv4_nr; j++) {
				switch (y->ipv4[j].type) {
				case IPV4_UNKNOWN:
					{
						ipv4->ipv4_interfaces[ipv4->ipv4_interfaces_nr].ipv4[j].type = IPV4_TYPE_UNKNOWN;
						break;
					}
				case IPV4_DHCP:
					{
						ipv4->ipv4_interfaces[ipv4->ipv4_interfaces_nr].ipv4[j].type = IPV4_TYPE_DHCP;
						break;
					}
				case IPV4_STATIC:
					{
						ipv4->ipv4_interfaces[ipv4->ipv4_interfaces_nr].ipv4[j].type = IPV4_TYPE_STATIC;
						break;
					}
				case IPV4_AUTOIP:
					{
						ipv4->ipv4_interfaces[ipv4->ipv4_interfaces_nr].ipv4[j].type = IPV4_TYPE_AUTOIP;
						break;
					}
				default:
					{
						WARNING("Invalid IPv4 type %d", y->ipv4[j].type);
						break;
					}
				}
				memcpy(ipv4->ipv4_interfaces[ipv4->ipv4_interfaces_nr].ipv4[j].ipv4_address, y->ipv4[j].address, 4);
				memcpy(ipv4->ipv4_interfaces[ipv4->ipv4_interfaces_nr].ipv4[j].ipv4_dhcp_server, y->ipv4[j].dhcp_server, 4);
			}

			ipv4->ipv4_interfaces_nr++;
		}

		if (y->ipv6_nr != 0) {
			if (ipv6->ipv6_interfaces_nr == 0) {
				CALLOC(ipv6->ipv6_interfaces, 1);
			} else {
				REALLOC(ipv6->ipv6_interfaces, (ipv6->ipv6_interfaces_nr + 1));
			}

			memcpy(ipv6->ipv6_interfaces[ipv6->ipv6_interfaces_nr].mac_address, y->mac_address, 6);
			ipv6->ipv6_interfaces[ipv6->ipv6_interfaces_nr].ipv6_nr = y->ipv6_nr;
			CALLOC(ipv6->ipv6_interfaces[ipv6->ipv6_interfaces_nr].ipv6, y->ipv6_nr);
			for (j = 0; j < y->ipv6_nr; j++) {
				switch (y->ipv6[j].type) {
				case IPV6_UNKNOWN:
					{
						ipv6->ipv6_interfaces[ipv6->ipv6_interfaces_nr].ipv6[j].type = IPV6_TYPE_UNKNOWN;
						break;
					}
				case IPV6_DHCP:
					{
						ipv6->ipv6_interfaces[ipv6->ipv6_interfaces_nr].ipv6[j].type = IPV6_TYPE_DHCP;
						break;
					}
				case IPV6_STATIC:
					{
						ipv6->ipv6_interfaces[ipv6->ipv6_interfaces_nr].ipv6[j].type = IPV6_TYPE_STATIC;
						break;
					}
				case IPV6_SLAAC:
					{
						ipv6->ipv6_interfaces[ipv6->ipv6_interfaces_nr].ipv6[j].type = IPV6_TYPE_SLAAC;
						break;
					}
				default:
					{
						WARNING("Invalid IPv6 type %d", y->ipv6[j].type);
						break;
					}
				}
				memcpy(ipv6->ipv6_interfaces[ipv6->ipv6_interfaces_nr].ipv6[j].ipv6_address, y->ipv6[j].address, 16);
				memcpy(ipv6->ipv6_interfaces[ipv6->ipv6_interfaces_nr].ipv6[j].ipv6_address_origin, y->ipv6[j].origin, 16);
			}

			ipv6->ipv6_interfaces_nr++;
		}

		PLATFORM_FREE_1905_INTERFACE_INFO(y);
	}

	PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);
}

//******************************************************************************
//******* Local device data dump ***********************************************
//******************************************************************************
//
// This function is used to update the entry of the database associated to the
// local node.
//
// Let me explain this in more detail: the database contains information of
// all nodes (the local and remote ones):
//
//   - For remote nodes, every time a response CMDU is received, the TLVs
//     contained in that CMDU are added to the entry of the database associated
//     to that remote node (or updated, if they already existed)
//
//   - For the local node, however, we must "manually" force an "update" so that
//     the database entry associated to the local node contains updated
//     information. *This* is exactly what this function does.
//
// When should we call this function? Well... we are only interested in updating
// this local entry when someone is going to look at it which, as of today,
// only happens when a special ("custom") ALME is received
// ("CUSTOM_COMMAND_DUMP_NETWORK_DEVICES") and, as a result, we must send the
// local information as part of the response.
//
void updateLocalDeviceData(void)
{
	struct deviceInformationTypeTLV *info;
	struct deviceBridgingCapabilityTLV **bridges;
	struct non1905NeighborDeviceListTLV **non1905_neighbors;
	INT8U non1905_neighbors_nr;
	struct neighborDeviceListTLV **x1905_neighbors;
	INT8U x1905_neighbors_nr;
	struct powerOffInterfaceTLV **power_off;
	struct l2NeighborDeviceTLV **l2_neighbors;
	struct genericPhyDeviceInformationTypeTLV *generic_phy;
	struct x1905ProfileVersionTLV *profile;
	struct deviceIdentificationTypeTLV *identification;
	struct controlUrlTypeTLV *control_url;
	struct ipv4TypeTLV *ipv4;
	struct ipv6TypeTLV *ipv6;

	struct transmitterLinkMetricTLV **tx_tlvs;
	struct receiverLinkMetricTLV **rx_tlvs;
	INT8U total_metrics_tlvs;
	INT8U i;

	// We need to allocate these structures in the heap (instead of simply
	// declaring variables in the stack) because they are going to be "saved"
	// in the database when calling "DMupdate*()"
	CALLOC(info, 1);
	CALLOC(bridges, 1);
	CALLOC(bridges[0], 1);
	CALLOC(power_off, 1);
	CALLOC(power_off[0], 1);
	CALLOC(l2_neighbors, 1);
	CALLOC(l2_neighbors[0], 1);
	CALLOC(generic_phy, 1);
	CALLOC(profile, 1);
	CALLOC(identification, 1);
	CALLOC(control_url, 1);
	CALLOC(ipv4, 1);
	CALLOC(ipv6, 1);

	get_deviceInformationTypeTLV(info);
	get_deviceBridgingCapabilityTLV(bridges[0]);
	obtainLocalNeighborsTLV(&non1905_neighbors, &non1905_neighbors_nr, &x1905_neighbors, &x1905_neighbors_nr);
	get_powerOffInterfaceTLV(power_off[0]);
	get_l2NeighborDeviceTLV(l2_neighbors[0]);
	get_genericPhyDeviceInformationTypeTLV(generic_phy);
	get_x1905ProfileVersionTLV(profile);
	get_deviceIdentificationTypeTLV(identification);
	get_controlUrlTypeTLV(control_url);
	obtainLocalIpsTLVs(ipv4, ipv6);

	obtainLocalMetricsTLVs(LINK_METRIC_QUERY_TLV_ALL_NEIGHBORS,
			       NULL, LINK_METRIC_QUERY_TLV_BOTH_TX_AND_RX_LINK_METRICS, &tx_tlvs, &rx_tlvs, &total_metrics_tlvs);

	// The following function will take care of "freeing" the allocated memory
	// if needed
	//
	DMupdateNetworkDeviceInfo(info->al_mac_address,
				  1, info,
				  1, bridges, 1,
				  1, non1905_neighbors, non1905_neighbors_nr,
				  1, x1905_neighbors, x1905_neighbors_nr,
				  1, power_off, 1, 1, l2_neighbors, 1, 1, generic_phy, 1, profile, 1, identification, 1, control_url, 1, ipv4, 1, ipv6);

	// The next function, however, takes care only of the pointers to metrics
	// information, and not of the memory used to hold the list of pointers
	// themselves...
	//
	for (i = 0; i < total_metrics_tlvs; i++) {
		DMupdateNetworkDeviceMetrics((INT8U *) tx_tlvs[i]);
		DMupdateNetworkDeviceMetrics((INT8U *) rx_tlvs[i]);
	}

	// ... and that's why we need to do this (the "0" is so that pointers
	// themselves are not freed, as they are now responsibility of the
	// database)
	//
	total_metrics_tlvs = 0;
	freeLocalMetricsTLVs(&tx_tlvs, &rx_tlvs, &total_metrics_tlvs);
}

////////////////////////////////////////////////////////////////////////////////
// Public functions (exported only to files in this same folder)
////////////////////////////////////////////////////////////////////////////////

bool fragment_send_raw_packets(const char *interface_name, INT8U *dst_mac_address, struct cmdu_stream *cmdu_stream)
{
	INT8U **streams;
	INT16U *streams_lens;
	INT8U total_streams, x;
	bool ret = true;

	streams = prepare_cmdu_packets_from_cmdu_stream(cmdu_stream, &streams_lens);
	if (streams == NULL) {
		// Could not forge the packet. Error?
		WARNING("prepare_cmdu_packets_from_cmdu_stream() failed!");
		return false;
	}

	total_streams = 0;
	while (streams[total_streams]) {
		total_streams++;
	}

	if (total_streams == 0) {
		// Could not forge the packet. Error?
		WARNING("prepare_cmdu_packets_from_cmdu_stream() returned 0 streams!");

		free_1905_CMDU_packets(streams);
		FREE(streams_lens);
		return false;
	}

	x = 0;
	while (streams[x]) {
		if (PLATFORM_SEND_RAW_PACKET(interface_name, dst_mac_address,
					     DMalMacGet(),
					     ETHERTYPE_1905,
					     streams[x],
					     streams_lens[x]) == 0) {
			ERROR("Packet could not be sent!");
			ret = false;
			break;
		}

		x++;
	}

	free_1905_CMDU_packets(streams);
	FREE(streams_lens);

	return ret;
}

INT8U send1905RawPacket(char *interface_name, INT16U mid, INT8U *dst_mac_address, struct cmdu_cstruct *cmdu)
{
	INT8U **streams;
	INT16U *streams_lens;

	INT8U total_streams, x;

	// Loop-back the messages
	if (is_local_mac(dst_mac_address)) {
		BroadcastCMDU(cmdu, NULL, NULL);
		return 1;
	}

	DEBUG("Contents of CMDU to send:");
	visit_1905_CMDU_structure(cmdu, print_callback, print_debug, "", NULL);

	streams = forge_1905_CMDU_from_structure(cmdu, &streams_lens);
	if (streams == NULL) {
		// Could not forge the packet. Error?
		//
		WARNING("forge_1905_CMDU_from_structure() failed!");
		return 0;
	}

	total_streams = 0;
	while (streams[total_streams]) {
		total_streams++;
	}

	if (total_streams == 0) {
		// Could not forge the packet. Error?
		//
		WARNING("forge_1905_CMDU_from_structure() returned 0 streams!");

		free_1905_CMDU_packets(streams);
		FREE(streams_lens);
		return 0;
	}

	x = 0;
	while (streams[x]) {
		DEBUG("Sending 1905 message on interface %s, MID %d, fragment %d/%d", interface_name, mid, x + 1, total_streams);
		if (PLATFORM_SEND_RAW_PACKET(interface_name, dst_mac_address, DMalMacGet(), ETHERTYPE_1905, streams[x], streams_lens[x]) == 0) {
			ERROR("Packet could not be sent!");
		}

		x++;
	}

	free_1905_CMDU_packets(streams);
	FREE(streams_lens);

	return 1;
}

INT8U send1905TopologyDiscoveryPacket(char *interface_name, INT16U mid)
{
	// The "topology discovery" message is a CMDU with two TLVs:
	//   - One AL MAC address type TLV
	//   - One MAC address type TLV

	INT8U mcast_address[] = MCAST_1905;

	struct cmdu_cstruct discovery_message;
	struct alMacAddressTypeTLV al_mac_addr_tlv;
	struct macAddressTypeTLV mac_addr_tlv;

	INFO("--> CMDU_TYPE_TOPOLOGY_DISCOVERY (%s)", interface_name);

	// Fill the AL MAC address type TLV
	if (false == get_alMacAddressTypeTLV(&al_mac_addr_tlv))
		return 0;

	if (false == get_macAddressTypeTLV(&mac_addr_tlv, interface_name))
		return 0;

	// Build the CMDU
	discovery_message.message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	discovery_message.message_type = CMDU_TYPE_TOPOLOGY_DISCOVERY;
	discovery_message.message_id = mid;
	discovery_message.relay_indicator = 0;
	CALLOC(discovery_message.tlv_cstruct_list, 3);
	discovery_message.tlv_cstruct_list[0] = (INT8U *) &al_mac_addr_tlv;
	discovery_message.tlv_cstruct_list[1] = (INT8U *) &mac_addr_tlv;
	discovery_message.tlv_cstruct_list[2] = NULL;

	// Send the packet
	if (send1905RawPacket(interface_name, mid, mcast_address, &discovery_message) == 0) {
		ERROR("Could not send the 1905 packet");
		free_1905_TLV_structure((INT8U *) &al_mac_addr_tlv, false);
		FREE(discovery_message.tlv_cstruct_list);
		return 0;
	}
	// Free memory
	free_tlv_internal_list_CMDU_structure(&discovery_message);
	return 1;
}

INT8U send1905TopologyQueryPacket(char *interface_name, INT16U mid, INT8U *destination_al_mac_address)
{
	// The "topology query" message is a CMDU with no TLVs
	INT8U ret = 1;

	struct cmdu_cstruct query_message;

	INFO("--> CMDU_TYPE_TOPOLOGY_QUERY (%s)", interface_name);
	DEBUG("Sending to %02x:%02x:%02x:%02x:%02x:%02x",
	      destination_al_mac_address[0], destination_al_mac_address[1], destination_al_mac_address[2],
	      destination_al_mac_address[3], destination_al_mac_address[4], destination_al_mac_address[5]);

	// Build the CMDU
	query_message.message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	query_message.message_type = CMDU_TYPE_TOPOLOGY_QUERY;
	query_message.message_id = mid;
	query_message.relay_indicator = 0;
	CALLOC(query_message.tlv_cstruct_list, 1);
	query_message.tlv_cstruct_list[0] = NULL;

	// Send the packet
	if (send1905RawPacket(interface_name, mid, destination_al_mac_address, &query_message) == 0) {
		WARNING("Could not send packet");
		ret = 0;
	}

	// Free memory
	free_tlv_internal_list_CMDU_structure(&query_message);

	return ret;
}

// returns number of tlv contained by cmdu
INT8U prepare1905TopologyResponsePacket(struct cmdu_cstruct *response_message, INT16U mid)
{
	INT8U non_1905_neighbors_nr = 0;
	INT8U neighbors_nr = 0;
	struct non1905NeighborDeviceListTLV **non_1905_neighbors;
	struct neighborDeviceListTLV **neighbors;
	struct deviceInformationTypeTLV *device_info;
	struct deviceBridgingCapabilityTLV *bridge_info;
	struct powerOffInterfaceTLV *power_off;
	struct l2NeighborDeviceTLV *l2_neighbors;
	bool device_info_present, bridge_info_present, power_off_present, l2_neighbors_present;
	INT8U total_tlvs = 0;
	INT8U tlv_index = 0;

	device_info_present = bridge_info_present = power_off_present = l2_neighbors_present = false;

	if (response_message == NULL)
		return 0;

	// Fill all the needed TLVs
	CALLOC(device_info, 1);
	CALLOC(bridge_info, 1);
	CALLOC(power_off, 1);
	CALLOC(l2_neighbors, 1);

	if (get_deviceInformationTypeTLV(device_info)) {
		device_info_present = true;
		total_tlvs++;
	}

	if (get_deviceBridgingCapabilityTLV(bridge_info)) {
		bridge_info_present = true;
		total_tlvs++;
	} else {
		FREE(bridge_info);
	}

	if (get_powerOffInterfaceTLV(power_off)) {
		power_off_present = true;
		total_tlvs++;
	} else {
		FREE(power_off);
	}

	if (get_l2NeighborDeviceTLV(l2_neighbors)) {
		l2_neighbors_present = true;
		total_tlvs++;
	} else {
		FREE(l2_neighbors);
	}

	obtainLocalNeighborsTLV(&non_1905_neighbors, &non_1905_neighbors_nr, &neighbors, &neighbors_nr);
	total_tlvs += neighbors_nr;	// Non-1905 neighbor device list TLVs
	total_tlvs += non_1905_neighbors_nr;	// 1905 Neighbor device list TLVs

	// Build the CMDU
	response_message->message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	response_message->message_type = CMDU_TYPE_TOPOLOGY_RESPONSE;
	response_message->message_id = mid;
	response_message->relay_indicator = 0;
	CALLOC(response_message->tlv_cstruct_list, (total_tlvs + 1));

	if (device_info_present)
		response_message->tlv_cstruct_list[tlv_index++] = (INT8U *) device_info;

	if (bridge_info_present)
		response_message->tlv_cstruct_list[tlv_index++] = (INT8U *) bridge_info;

	for (INT8U j = 0; j < non_1905_neighbors_nr; j++) {
		response_message->tlv_cstruct_list[tlv_index++] = (INT8U *) non_1905_neighbors[j];
	}

	for (INT8U j = 0; j < neighbors_nr; j++) {
		response_message->tlv_cstruct_list[tlv_index++] = (INT8U *) neighbors[j];
	}

	if (power_off_present)
		response_message->tlv_cstruct_list[tlv_index++] = (INT8U *) power_off;

	if (l2_neighbors_present)
		response_message->tlv_cstruct_list[tlv_index++] = (INT8U *) l2_neighbors;

	response_message->tlv_cstruct_list[tlv_index] = NULL;

	return total_tlvs;
}

INT8U send1905TopologyResponsePacket(char *interface_name, INT16U mid, INT8U *destination_al_mac_address)
{
	// The "topology response" message is a CMDU with the following TLVs:
	//   - One device information type TLV
	//   - Zero or one device bridging capability TLVs
	//   - Zero or more non-1905 neighbor device list TLVs
	//   - Zero or more 1905 neighbor device list TLVs
	//   - Zero or more power off interface TLVs
	//   - Zero or more L2 neighbor device TLVs
	//
	//   NOTE: The "non-1905 neighbor" and the "L2 neighbor" device TLVs are
	//   kind of overlaping... but this is what the standard says.
	//
	//   NOTE: Regarding the "device bridging capability", "power off interface"
	//   and "L2 neighbor device" TLVs, the standard says "zero or more" but
	//   it should be "zero or one", as one single TLV of these types can carry
	//   many entries.
	//   That's why in this implementation we are just sending zero or one (no
	//   more!) TLVs of these type. However, in reception (see
	//   "process1905Cmdu()") we will be ready to receive more.

	INT8U ret = 0;
	struct cmdu_cstruct *cmdu;

	CALLOC(cmdu, 1);

	if (prepare1905TopologyResponsePacket(cmdu, mid)) {
		INFO("--> CMDU_TYPE_TOPOLOGY_RESPONSE (%s)", interface_name);
		DEBUG("Sending to %02x:%02x:%02x:%02x:%02x:%02x",
		      destination_al_mac_address[0], destination_al_mac_address[1], destination_al_mac_address[2],
		      destination_al_mac_address[3], destination_al_mac_address[4], destination_al_mac_address[5]);
		if (send1905RawPacket(interface_name, mid, destination_al_mac_address, cmdu) == 0) {
			WARNING("Could not send packet");
			ret = 0;
		}
		ret = 1;
	}
	free_1905_CMDU_structure(cmdu);

	return ret;
}

INT8U send1905TopologyNotificationPacket(char *interface_name, INT16U mid)
{
	// The "topology discovery" message is a CMDU with one TLVs:
	//   - One AL MAC address type TLV

	INT8U ret = 1;

	INT8U mcast_address[] = MCAST_1905;

	struct cmdu_cstruct discovery_message;
	struct alMacAddressTypeTLV al_mac_addr_tlv;

	INFO("--> CMDU_TYPE_TOPOLOGY_NOTIFICATION (%s)", interface_name);

	// Fill all the needed TLVs
	get_alMacAddressTypeTLV(&al_mac_addr_tlv);

	// Build the CMDU
	discovery_message.message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	discovery_message.message_type = CMDU_TYPE_TOPOLOGY_NOTIFICATION;
	discovery_message.message_id = mid;
	discovery_message.relay_indicator = 0;
	CALLOC(discovery_message.tlv_cstruct_list, 2);
	discovery_message.tlv_cstruct_list[0] = (INT8U *) &al_mac_addr_tlv;
	discovery_message.tlv_cstruct_list[1] = NULL;

	// Send the packet
	if (send1905RawPacket(interface_name, mid, mcast_address, &discovery_message) == 0) {
		ERROR("Could not send the 1905 packet");
		ret = 0;
	}

	// Free all allocated (and no longer needed) memory
	free_tlv_internal_list_CMDU_structure(&discovery_message);

	return ret;
}

INT8U send1905MetricsQueryPacket(char *interface_name, INT16U mid, INT8U *destination_al_mac_address)
{
	// The "metrics query" message is a CMDU with one TLV:
	//   - One link metric query TLV

	INT8U ret = 1;

	struct cmdu_cstruct query_message;
	struct linkMetricQueryTLV metric_query_tlv;

	INFO("--> CMDU_TYPE_LINK_METRIC_QUERY (%s)", interface_name);
	DEBUG("Sending to %02x:%02x:%02x:%02x:%02x:%02x",
	      destination_al_mac_address[0], destination_al_mac_address[1], destination_al_mac_address[2],
	      destination_al_mac_address[3], destination_al_mac_address[4], destination_al_mac_address[5]);

	// Fill all the needed TLVs
	get_linkMetricQueryTLV(&metric_query_tlv);

	// Build the CMDU
	query_message.message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	query_message.message_type = CMDU_TYPE_LINK_METRIC_QUERY;
	query_message.message_id = mid;
	query_message.relay_indicator = 0;
	CALLOC(query_message.tlv_cstruct_list, 2);
	query_message.tlv_cstruct_list[0] = (INT8U *) &metric_query_tlv;
	query_message.tlv_cstruct_list[1] = NULL;

	if (send1905RawPacket(interface_name, mid, destination_al_mac_address, &query_message) == 0) {
		WARNING("Could not send packet");
		ret = 0;
	}

	// Free all allocated (and no longer needed) memory
	free_tlv_internal_list_CMDU_structure(&query_message);

	return ret;
}

INT8U send1905MetricsResponsePacket(char *interface_name, INT16U mid, INT8U *destination_al_mac_address,
				    INT8U destination, INT8U *specific_neighbor,
				    INT8U metrics_type)
{
	// The "metrics response" message can be either:
	//
	//   A) A CMDU containing either:
	//      - One Tx link metrics
	//      - One Rx link metrics
	//      - One Rx and one Tx link metrics
	//      ... containing info regarding the link between the current node
	//      and the AL entity whose AL MAC is 'specific_neighbor'
	//
	//   B) A CMDU made by concatenating many CMDUs of "type A" (one for each of
	//      its 1905 neighbors).
	//
	// Case (A) happens when
	//
	//   'destination' == LINK_METRIC_QUERY_TLV_SPECIFIC_NEIGHBOR
	//
	// ...while case (B) takes place when
	//
	//   'destination' == LINK_METRIC_QUERY_TLV_ALL_NEIGHBORS
	//
	INT8U ret;

	struct cmdu_cstruct *response_message;
	struct transmitterLinkMetricTLV **tx_tlvs;
	struct receiverLinkMetricTLV **rx_tlvs;

	INT8U total_tlvs;
	INT8U i, j;

	CALLOC(response_message, 1);
	INFO("--> CMDU_TYPE_LINK_METRIC_RESPONSE (%s)", interface_name);
	DEBUG("Sending to %02x:%02x:%02x:%02x:%02x:%02x",
	      destination_al_mac_address[0], destination_al_mac_address[1], destination_al_mac_address[2],
	      destination_al_mac_address[3], destination_al_mac_address[4], destination_al_mac_address[5]);

	// Fill all the needed TLVs
	//
	obtainLocalMetricsTLVs(destination, specific_neighbor, metrics_type, &tx_tlvs, &rx_tlvs, &total_tlvs);

	// Build the CMDU
	//
	response_message->message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	response_message->message_type = CMDU_TYPE_LINK_METRIC_RESPONSE;
	response_message->message_id = mid;
	response_message->relay_indicator = 0;

	if (NULL == tx_tlvs || NULL == rx_tlvs) {
		CALLOC(response_message->tlv_cstruct_list, (total_tlvs + 1));
	} else {
		CALLOC(response_message->tlv_cstruct_list, ((2 * total_tlvs) + 1));
	}

	j = 0;
	if (tx_tlvs != NULL) {
		for (i = 0; i < total_tlvs; i++) {
			response_message->tlv_cstruct_list[j++] = (INT8U *) (tx_tlvs[i]);
		}
	}
	if (rx_tlvs != NULL) {
		for (i = 0; i < total_tlvs; i++) {
			response_message->tlv_cstruct_list[j++] = (INT8U *) (rx_tlvs[i]);
		}
	}

	response_message->tlv_cstruct_list[j++] = NULL;

	// Send the packet
	//
	if (send1905RawPacket(interface_name, mid, destination_al_mac_address, response_message) == 0) {
		ERROR("Could not allocate memory for TLV buffer");
		ret = 0;
	} else {
		ret = 1;
	}

	// Free all allocated (and no longer needed) memory
	free_1905_CMDU_structure(response_message);
	FREE(rx_tlvs);
	FREE(tx_tlvs);

	return ret;
}

INT8U send1905PushButtonEventNotificationPacket(char *interface_name, INT16U mid, char **all_interfaces_names, INT8U *push_button_mask, INT8U nr)
{
	// The "push button event notification" message is a CMDU with two TLVs:
	//   - One AL MAC address type TLV
	//   - One push button event notification TLV
	//   - Zero or one push button generic phy event notification

	INT8U ret;

	INT8U mcast_address[] = MCAST_1905;

	INT8U media_types_nr;
	INT8U generic_media_types_nr;
	INT8U i, j;

	struct cmdu_cstruct notification_message;
	struct alMacAddressTypeTLV al_mac_addr_tlv;
	struct pushButtonEventNotificationTLV pb_event_tlv;
	struct pushButtonGenericPhyEventNotificationTLV pbg_event_tlv;

	INFO("--> CMDU_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION (%s)", interface_name);

	// Fill the AL MAC address type TLV
	get_alMacAddressTypeTLV(&al_mac_addr_tlv);

	// Fill the push button event notification TLV
	//
	media_types_nr = 0;
	generic_media_types_nr = 0;
	for (i = 0; i < nr; i++) {
		if (push_button_mask[i] == 0) {
			media_types_nr++;
		}
	}

	pb_event_tlv.tlv_type = TLV_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION;
	pb_event_tlv.media_types_nr = media_types_nr;

	if (media_types_nr > 0) {
		CALLOC(pb_event_tlv.media_types, media_types_nr);
	} else {
		pb_event_tlv.media_types = NULL;
	}

	j = 0;
	for (i = 0; i < nr; i++) {
		if (push_button_mask[i] == 0) {
			struct interfaceInfo *x;

			x = PLATFORM_GET_1905_INTERFACE_INFO(all_interfaces_names[i]);
			if (x == NULL) {
				WARNING("Could not retrieve info of interface %s", all_interfaces_names[i]);

				pb_event_tlv.media_types[j].media_type = MEDIA_TYPE_UNKNOWN;
				pb_event_tlv.media_types[j].media_specific_data_size = 0;
			} else {
				pb_event_tlv.media_types[j].media_type = x->interface_type;

				// Fill the rest of media specific fields
				//
				switch (pb_event_tlv.media_types[j].media_type) {
				case MEDIA_TYPE_IEEE_802_3U_FAST_ETHERNET:
				case MEDIA_TYPE_IEEE_802_3AB_GIGABIT_ETHERNET:
				case MEDIA_TYPE_MOCA_V1_1:
					{
						// These interfaces don't require extra data
						//
						pb_event_tlv.media_types[j].media_specific_data_size = 0;

						break;
					}
				case MEDIA_TYPE_IEEE_802_11B_2_4_GHZ:
				case MEDIA_TYPE_IEEE_802_11G_2_4_GHZ:
				case MEDIA_TYPE_IEEE_802_11A_5_GHZ:
				case MEDIA_TYPE_IEEE_802_11N_2_4_GHZ:
				case MEDIA_TYPE_IEEE_802_11N_5_GHZ:
				case MEDIA_TYPE_IEEE_802_11AC_5_GHZ:
				case MEDIA_TYPE_IEEE_802_11AD_60_GHZ:
				case MEDIA_TYPE_IEEE_802_11AF_GHZ:
					{
						pb_event_tlv.media_types[j].media_specific_data_size = 10;
						pb_event_tlv.media_types[j].media_specific_data.ieee80211.network_membership[0] =
						    x->interface_type_data.ieee80211.bssid[0];
						pb_event_tlv.media_types[j].media_specific_data.ieee80211.network_membership[1] =
						    x->interface_type_data.ieee80211.bssid[1];
						pb_event_tlv.media_types[j].media_specific_data.ieee80211.network_membership[2] =
						    x->interface_type_data.ieee80211.bssid[2];
						pb_event_tlv.media_types[j].media_specific_data.ieee80211.network_membership[3] =
						    x->interface_type_data.ieee80211.bssid[3];
						pb_event_tlv.media_types[j].media_specific_data.ieee80211.network_membership[4] =
						    x->interface_type_data.ieee80211.bssid[4];
						pb_event_tlv.media_types[j].media_specific_data.ieee80211.network_membership[5] =
						    x->interface_type_data.ieee80211.bssid[5];
						pb_event_tlv.media_types[j].media_specific_data.ieee80211.role = x->interface_type_data.ieee80211.role;
						pb_event_tlv.media_types[j].media_specific_data.ieee80211.ap_channel_band =
						    x->interface_type_data.ieee80211.ap_channel_band;
						pb_event_tlv.media_types[j].media_specific_data.ieee80211.ap_channel_center_frequency_index_1 =
						    x->interface_type_data.ieee80211.ap_channel_center_frequency_index_1;
						pb_event_tlv.media_types[j].media_specific_data.ieee80211.ap_channel_center_frequency_index_2 =
						    x->interface_type_data.ieee80211.ap_channel_center_frequency_index_2;

						break;
					}
				case MEDIA_TYPE_IEEE_1901_WAVELET:
				case MEDIA_TYPE_IEEE_1901_FFT:
					{
						pb_event_tlv.media_types[j].media_specific_data_size = 7;
						pb_event_tlv.media_types[j].media_specific_data.ieee1901.network_identifier[0] =
						    x->interface_type_data.ieee1901.network_identifier[0];
						pb_event_tlv.media_types[j].media_specific_data.ieee1901.network_identifier[1] =
						    x->interface_type_data.ieee1901.network_identifier[1];
						pb_event_tlv.media_types[j].media_specific_data.ieee1901.network_identifier[2] =
						    x->interface_type_data.ieee1901.network_identifier[2];
						pb_event_tlv.media_types[j].media_specific_data.ieee1901.network_identifier[3] =
						    x->interface_type_data.ieee1901.network_identifier[3];
						pb_event_tlv.media_types[j].media_specific_data.ieee1901.network_identifier[4] =
						    x->interface_type_data.ieee1901.network_identifier[4];
						pb_event_tlv.media_types[j].media_specific_data.ieee1901.network_identifier[5] =
						    x->interface_type_data.ieee1901.network_identifier[5];
						pb_event_tlv.media_types[j].media_specific_data.ieee1901.network_identifier[6] =
						    x->interface_type_data.ieee1901.network_identifier[6];

						break;
					}
				case MEDIA_TYPE_UNKNOWN:
					{
						// Do not include extra data here. It will be included
						// in the accompanying "push button generic phy
						// notification TVL"
						//
						generic_media_types_nr++;
						pb_event_tlv.media_types[j].media_specific_data_size = 0;

						break;
					}
				}
			}
			j++;

			if (x != NULL) {
				PLATFORM_FREE_1905_INTERFACE_INFO(x);
			}
		}
	}

	// Fill the push button generic event notification TLV
	//
	pbg_event_tlv.tlv_type = TLV_TYPE_GENERIC_PHY_EVENT_NOTIFICATION;
	pbg_event_tlv.local_interfaces_nr = generic_media_types_nr;

	if (generic_media_types_nr == 0) {
		pbg_event_tlv.local_interfaces = NULL;
	} else {
		CALLOC(pbg_event_tlv.local_interfaces, generic_media_types_nr);

		j = 0;
		for (i = 0; i < nr; i++) {
			if (push_button_mask[i] == 0) {
				struct interfaceInfo *x;

				x = PLATFORM_GET_1905_INTERFACE_INFO(all_interfaces_names[i]);
				if (x == NULL) {
					WARNING("Could not retrieve info of interface %s", all_interfaces_names[i]);

					continue;
				}

				if (x->interface_type == MEDIA_TYPE_UNKNOWN) {
					// We only care about "generic" interfaces

					INT16U len;
					INT8U *m;

					pbg_event_tlv.local_interfaces[j].oui[0] = x->interface_type_data.other.oui[0];
					pbg_event_tlv.local_interfaces[j].oui[1] = x->interface_type_data.other.oui[1];
					pbg_event_tlv.local_interfaces[j].oui[2] = x->interface_type_data.other.oui[2];
					pbg_event_tlv.local_interfaces[j].variant_index = x->interface_type_data.other.variant_index;

					m = forge_media_specific_blob(&x->interface_type_data.other, &len);

					if (m != NULL) {
						pbg_event_tlv.local_interfaces[j].media_specific_bytes_nr = len;
						pbg_event_tlv.local_interfaces[j].media_specific_bytes = m;
					} else {
						// Ignore media specific data
						//
						pbg_event_tlv.local_interfaces[j].media_specific_bytes_nr = 0;
						pbg_event_tlv.local_interfaces[j].media_specific_bytes = NULL;
					}

					j++;
				}
				PLATFORM_FREE_1905_INTERFACE_INFO(x);
			}
		}
	}

	// Build the CMDU
	//
	notification_message.message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	notification_message.message_type = CMDU_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION;
	notification_message.message_id = mid;
	notification_message.relay_indicator = 1;

	if (generic_media_types_nr != 0) {
		CALLOC(notification_message.tlv_cstruct_list, 4);
		notification_message.tlv_cstruct_list[2] = (INT8U *) &pbg_event_tlv;
		notification_message.tlv_cstruct_list[3] = NULL;
	} else {
		CALLOC(notification_message.tlv_cstruct_list, 3);
		notification_message.tlv_cstruct_list[2] = NULL;
	}

	notification_message.tlv_cstruct_list[0] = (INT8U *) &al_mac_addr_tlv;
	notification_message.tlv_cstruct_list[1] = (INT8U *) &pb_event_tlv;

	// Send the packet
	//
	if (send1905RawPacket(interface_name, mid, mcast_address, &notification_message) == 0) {
		ERROR("Could not send the 1905 packet");
		ret = 0;
	} else {
		ret = 1;
	}

	// Free memory
	free_tlv_internal_list_CMDU_structure(&notification_message);

	return ret;
}

INT8U send1905PushButtonJoinNotificationPacket(char *interface_name, INT16U mid, INT8U *original_al_mac_address, INT16U original_mid,
					       INT8U *local_mac_address, INT8U *new_mac_address)
{
	// The "push button join notification" message is a CMDU with two TLVs:
	//   - One AL MAC address type TLV
	//   - One push button join notification TLV

	INT8U ret = 1;

	INT8U al_mac_address[6];
	INT8U mcast_address[] = MCAST_1905;

	struct cmdu_cstruct notification_message;
	struct alMacAddressTypeTLV al_mac_addr_tlv;
	struct pushButtonJoinNotificationTLV pb_join_tlv;

	INFO("--> CMDU_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION (%s)", interface_name);

	memcpy(al_mac_address, DMalMacGet(), 6);

	// Fill the AL MAC address type TLV
	get_alMacAddressTypeTLV(&al_mac_addr_tlv);

	// Fill the push button join notification TLV
	get_pushButtonJoinNotificationTLV(&pb_join_tlv, original_al_mac_address, original_mid,
					  local_mac_address, new_mac_address);

	// Build the CMDU
	notification_message.message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	notification_message.message_type = CMDU_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION;
	notification_message.message_id = mid;
	notification_message.relay_indicator = 1;
	CALLOC(notification_message.tlv_cstruct_list, 3);
	notification_message.tlv_cstruct_list[0] = (INT8U *) &al_mac_addr_tlv;
	notification_message.tlv_cstruct_list[1] = (INT8U *) &pb_join_tlv;
	notification_message.tlv_cstruct_list[2] = NULL;

	// Send the packet
	if (send1905RawPacket(interface_name, mid, mcast_address, &notification_message) == 0) {
		ERROR("Could not send the 1905 packet");
		ret = 0;
	}

	// Free memory
	free_tlv_internal_list_CMDU_structure(&notification_message);

	return ret;
}

// returns number of tlv contained by cmdu
INT8U prepare1905APAutoconfigurationSearchPacket(struct cmdu_cstruct *search_message, INT16U mid, INT8U freq_band)
{
	// The "AP-autoconfiguration search" message is a CMDU with three TLVs:
	//   - One AL MAC address type TLV
	//   - One searched role TLV
	//   - One autoconfig freq band TLV

	INT8U tlv_count = 0, tlv_index = 0;
	struct alMacAddressTypeTLV *al_mac_addr_tlv;
	struct searchedRoleTLV *searched_role_tlv;
	struct autoconfigFreqBandTLV *ac_freq_band_tlv;
	bool al_mac_tlv_present, searched_role_tlv_present, autoconfig_tlv_present;

	al_mac_tlv_present = searched_role_tlv_present = autoconfig_tlv_present = false;

	CALLOC(al_mac_addr_tlv, 1);
	CALLOC(searched_role_tlv, 1);
	CALLOC(ac_freq_band_tlv, 1);

	// Fill the AL MAC address type TLV
	if (get_alMacAddressTypeTLV(al_mac_addr_tlv)) {
		al_mac_tlv_present = true;
		tlv_count++;
	}

	// Fill the searched role TLV
	if (get_searchedRoleTLV(searched_role_tlv)) {
		searched_role_tlv_present = true;
		tlv_count++;
	}

	// Fill the autoconfig freq band TLV
	if (get_autoconfigFreqBandTLV(ac_freq_band_tlv, freq_band)) {
		autoconfig_tlv_present = true;
		tlv_count++;
	}

	// Build the CMDU
	search_message->message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	search_message->message_type = CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH;
	search_message->message_id = mid;
	search_message->relay_indicator = 1;
	CALLOC(search_message->tlv_cstruct_list, (tlv_count + 1));
	if (al_mac_tlv_present)
		search_message->tlv_cstruct_list[tlv_index++] = (INT8U *) al_mac_addr_tlv;

	if (searched_role_tlv_present)
		search_message->tlv_cstruct_list[tlv_index++] = (INT8U *) searched_role_tlv;

	if (autoconfig_tlv_present)
		search_message->tlv_cstruct_list[tlv_index++] = (INT8U *) ac_freq_band_tlv;

	search_message->tlv_cstruct_list[tlv_index] = NULL;

	return tlv_count;
}

INT8U send1905APAutoconfigurationSearchPacket(INT8U freq_band)
{
	// The "AP-autoconfiguration search" message is a CMDU with three TLVs:
	//   - One AL MAC address type TLV
	//   - One searched role TLV
	//   - One autoconfig freq band TLV

	INT8U ret = 0;
	char **ifs_names;
	uint8_t ifs_nr = 0, i;
	uint16_t mid;
	struct cmdu_cstruct *search_message;
	INT8U mcast_address[] = MCAST_1905;

	CALLOC(search_message, 1);

	mid = getNextMid();
	// Build the CMDU
	if (prepare1905APAutoconfigurationSearchPacket(search_message, mid, freq_band) == 0) {
		return 0;
	}

	ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);

	for (i = 0; i < ifs_nr; i++) {
		INFO("--> CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH (%s)", ifs_names[i]);
		if (send1905RawPacket(ifs_names[i], mid, mcast_address, search_message) == 0) {
			INFO("Could not send packet on (%s) interface", ifs_names[i]);
			ret = 0;
			break;
		}
	}

	PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);
	free_1905_CMDU_structure(search_message);
	return ret;
}

INT8U send1905APAutoconfigurationResponsePacket(char *interface_name, INT16U mid,
						INT8U *destination_al_mac_address, INT8U freq_band)
{
	// The "AP-autoconfiguration response" message is a CMDU with two TLVs:
	//   - One supported role TLV
	//   - One supported freq band TLV

	INT8U ret;

	struct cmdu_cstruct response_message;
	struct supportedRoleTLV supported_role_tlv;
	struct supportedFreqBandTLV supported_freq_band_tlv;

	INFO("--> CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE (%s)", interface_name);

	// Fill the supported role TLV
	get_supportedRoleTLV(&supported_role_tlv);

	// Fill the supported freq band TLV
	get_supportedFreqBandTLV(&supported_freq_band_tlv, freq_band);

	// Build the CMDU
	//
	response_message.message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	response_message.message_type = CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE;
	response_message.message_id = mid;
	response_message.relay_indicator = 0;
	CALLOC(response_message.tlv_cstruct_list, 3);
	response_message.tlv_cstruct_list[0] = (INT8U *) &supported_role_tlv;
	response_message.tlv_cstruct_list[1] = (INT8U *) &supported_freq_band_tlv;
	response_message.tlv_cstruct_list[2] = NULL;

	if (send1905RawPacket(interface_name, mid, destination_al_mac_address, &response_message) == 0) {
		WARNING("Could not send packet");
		ret = 0;
	} else {
		ret = 1;
	}

	// Free memory
	free_tlv_internal_list_CMDU_structure(&response_message);

	return ret;
}

INT8U send1905APAutoconfigurationRenewPacket(uint8_t freq_band)
{
	// The "AP-autoconfiguration Renew" message is a CMDU with three TLVs:
	//     - 1 TLV_TYPE_AL_MAC_ADDRESS_TYPE
	//     - 1 TLV_TYPE_SUPPORTED_ROLE
	//     - 1 TLV_TYPE_SUPPORTED_FREQ_BAND

	INT8U ret = 0;
	uint16_t mid;
	struct cmdu_cstruct renew_cmdu;
	struct alMacAddressTypeTLV al_mac_addr_tlv;
	struct supportedRoleTLV supported_role_tlv;
	struct supportedFreqBandTLV supported_freq_band_tlv;
	INT8U mcast_address[] = MCAST_1905;
	char **ifs_names;
	uint8_t ifs_nr, i;


	if (DMregistrarMacGet(freq_band) == NULL)
		return false;

	// Build the CMDU
	if (get_alMacAddressTypeTLV(&al_mac_addr_tlv) == false) {
		ERROR("failed to get al_mac_addr_tlv");
		return false;
	}

	// Fill the supported role TLV
	if (get_supportedRoleTLV(&supported_role_tlv) == false) {
		ERROR("Failed to get supported role tlv");
		return false;
	}

	// Fill the supported freq band TLV
	if (get_supportedFreqBandTLV(&supported_freq_band_tlv, freq_band) == false) {
		ERROR("Failed to get supported freq band(%d)", freq_band);
		return false;
	}

	mid = getNextMid();

	renew_cmdu.message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	renew_cmdu.message_type = CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW;
	renew_cmdu.message_id = mid;
	renew_cmdu.relay_indicator = 1;

	CALLOC(renew_cmdu.tlv_cstruct_list, 4);
	renew_cmdu.tlv_cstruct_list[0] = (uint8_t *) &al_mac_addr_tlv;
	renew_cmdu.tlv_cstruct_list[1] = (uint8_t *) &supported_role_tlv;
	renew_cmdu.tlv_cstruct_list[2] = (uint8_t *) &supported_freq_band_tlv;
	renew_cmdu.tlv_cstruct_list[3] = NULL;

	// Free memory
	ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);

	for (i = 0; i < ifs_nr; i++) {
		INFO("--> CMDU_TYPE_AP_AUTOCONFIGURATION_RENEW (%s:%d)", ifs_names[i], freq_band);
		if (send1905RawPacket(ifs_names[i], mid, mcast_address, &renew_cmdu) == 0) {
			INFO("Could not send packet on (%s) interface", ifs_names[i]);
			ret = 0;
			break;
		}
	}

	PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);
	free_tlv_internal_list_CMDU_structure(&renew_cmdu);

	return ret;
}

INT8U send1905APAutoconfigurationWSCPacket(char *interface_name, INT16U mid,
					   INT8U *destination_al_mac_address,
					   INT8U *wsc_frame, INT16U wsc_frame_size)
{
	// The "AP-autoconfiguration WSC" message is a CMDU with just one TLVs:
	//   - One WSC TLV

	INT8U ret;

	struct cmdu_cstruct data_message;
	struct wscTLV wsc_tlv;

	INFO("--> CMDU_TYPE_AP_AUTOCONFIGURATION_WSC (%s)", interface_name);

	// Fill the WSC TLV
	get_wscTLV(&wsc_tlv, wsc_frame, wsc_frame_size);

	// Build the CMDU
	data_message.message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	data_message.message_type = CMDU_TYPE_AP_AUTOCONFIGURATION_WSC;
	data_message.message_id = mid;
	data_message.relay_indicator = 0;
	CALLOC(data_message.tlv_cstruct_list, 2);
	data_message.tlv_cstruct_list[0] = (INT8U *) &wsc_tlv;
	data_message.tlv_cstruct_list[1] = NULL;

	if (send1905RawPacket(interface_name, mid, destination_al_mac_address, &data_message) == 0) {
		WARNING("Could not send packet");
		ret = 0;
	} else {
		ret = 1;
	}

	// Free memory
	free_tlv_internal_list_CMDU_structure(&data_message);

	return ret;
}

INT8U send1905GenericPhyQueryPacket(char *interface_name, INT16U mid, INT8U *destination_al_mac_address)
{
	// The "generic phy query" message is a CMDU with no TLVs

	INT8U ret = 1;

	struct cmdu_cstruct query_message;

	INFO("--> CMDU_TYPE_GENERIC_PHY_QUERY (%s)", interface_name);
	DEBUG("Sending to %02x:%02x:%02x:%02x:%02x:%02x",
	      destination_al_mac_address[0], destination_al_mac_address[1], destination_al_mac_address[2],
	      destination_al_mac_address[3], destination_al_mac_address[4], destination_al_mac_address[5]);

	// Build the CMDU
	query_message.message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	query_message.message_type = CMDU_TYPE_GENERIC_PHY_QUERY;
	query_message.message_id = mid;
	query_message.relay_indicator = 0;
	CALLOC(query_message.tlv_cstruct_list, 1);
	query_message.tlv_cstruct_list[0] = NULL;

	// Send packet
	if (send1905RawPacket(interface_name, mid, destination_al_mac_address, &query_message) == 0) {
		WARNING("Could not send packet");
		ret = 0;
	}

	// Free memory
	free_tlv_internal_list_CMDU_structure(&query_message);

	return ret;
}

INT8U send1905GenericPhyResponsePacket(char *interface_name, INT16U mid, INT8U *destination_al_mac_address)
{
	// The "generic phy response" message is a CMDU with the following TLVs:
	//   - One generic phy device information type TLV

	INT8U ret = 1;

	struct cmdu_cstruct response_message;
	struct genericPhyDeviceInformationTypeTLV generic_phy;

	INFO("--> CMDU_TYPE_GENERIC_PHY_RESPONSE (%s)", interface_name);
	DEBUG("Sending to %02x:%02x:%02x:%02x:%02x:%02x",
	      destination_al_mac_address[0], destination_al_mac_address[1], destination_al_mac_address[2],
	      destination_al_mac_address[3], destination_al_mac_address[4], destination_al_mac_address[5]);

	// Fill all the needed TLVs
	get_genericPhyDeviceInformationTypeTLV(&generic_phy);

	// Build the CMDU
	response_message.message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	response_message.message_type = CMDU_TYPE_GENERIC_PHY_RESPONSE;
	response_message.message_id = mid;
	response_message.relay_indicator = 0;
	CALLOC(response_message.tlv_cstruct_list, 2);
	response_message.tlv_cstruct_list[0] = (INT8U *) &generic_phy;
	response_message.tlv_cstruct_list[1] = NULL;

	// Send the packet
	if (send1905RawPacket(interface_name, mid, destination_al_mac_address, &response_message) == 0) {
		WARNING("Could not send packet");
		ret = 0;
	}

	// Free all allocated (and no longer needed) memory
	free_tlv_internal_list_CMDU_structure(&response_message);

	return ret;
}

INT8U send1905HighLayerQueryPacket(char *interface_name, INT16U mid, INT8U *destination_al_mac_address)
{
	// The "high level phy query" message is a CMDU with no TLVs

	INT8U ret = 1;

	struct cmdu_cstruct query_message;

	INFO("--> CMDU_TYPE_HIGHER_LAYER_QUERY (%s)", interface_name);
	DEBUG("Sending to %02x:%02x:%02x:%02x:%02x:%02x",
	      destination_al_mac_address[0], destination_al_mac_address[1], destination_al_mac_address[2],
	      destination_al_mac_address[3], destination_al_mac_address[4], destination_al_mac_address[5]);

	// Build the CMDU
	//
	query_message.message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	query_message.message_type = CMDU_TYPE_HIGHER_LAYER_QUERY;
	query_message.message_id = mid;
	query_message.relay_indicator = 0;
	CALLOC(query_message.tlv_cstruct_list, 1);
	query_message.tlv_cstruct_list[0] = NULL;

	// Send the packet
	if (send1905RawPacket(interface_name, mid, destination_al_mac_address, &query_message) == 0) {
		WARNING("Could not send packet");
		ret = 0;
	}

	// Free memory
	free_tlv_internal_list_CMDU_structure(&query_message);

	return ret;
}

INT8U send1905HighLayerResponsePacket(char *interface_name, INT16U mid, INT8U *destination_al_mac_address)
{
	// The "high layer response" message is a CMDU with the following TLVs:
	//   - One AL MAC address type TLV
	//   - One 1905 profile version TLV
	//   - One device identification type TLV
	//   - Zero or one control URL type TLV
	//   - Zero or one IPv4 type TLV
	//   - Zero or one IPv6 type TLV

	INT8U total_tlvs;
	INT8U i;

	struct cmdu_cstruct response_message;
	struct alMacAddressTypeTLV al_mac_addr_tlv;
	struct x1905ProfileVersionTLV profile_tlv;
	struct deviceIdentificationTypeTLV identification_tlv;
	struct controlUrlTypeTLV control_tlv;
	struct ipv4TypeTLV ipv4_tlv;
	struct ipv6TypeTLV ipv6_tlv;

	INFO("--> CMDU_TYPE_HIGHER_LAYER_RESPONSE (%s)", interface_name);

	// Fill all the needed TLVs
	get_alMacAddressTypeTLV(&al_mac_addr_tlv);
	get_x1905ProfileVersionTLV(&profile_tlv);
	get_deviceIdentificationTypeTLV(&identification_tlv);
	get_controlUrlTypeTLV(&control_tlv);
	obtainLocalIpsTLVs(&ipv4_tlv, &ipv6_tlv);

	// Build the CMDU
	//
	total_tlvs = 3;		// AL MAC, profile and identification

	if (control_tlv.url != NULL) {
		total_tlvs++;
	}
#ifndef SEND_EMPTY_TLVS
	if (ipv4_tlv.ipv4_interfaces_nr != 0)
#endif
	{
		total_tlvs++;
	}
#ifndef SEND_EMPTY_TLVS
	if (ipv6_tlv.ipv6_interfaces_nr != 0)
#endif
	{
		total_tlvs++;
	}

	response_message.message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	response_message.message_type = CMDU_TYPE_HIGHER_LAYER_RESPONSE;
	response_message.message_id = mid;
	response_message.relay_indicator = 0;
	CALLOC(response_message.tlv_cstruct_list, (total_tlvs + 1));
	response_message.tlv_cstruct_list[0] = (INT8U *) &al_mac_addr_tlv;
	response_message.tlv_cstruct_list[1] = (INT8U *) &profile_tlv;
	response_message.tlv_cstruct_list[2] = (INT8U *) &identification_tlv;

	i = 3;
	if (control_tlv.url != NULL) {
		response_message.tlv_cstruct_list[i++] = (INT8U *) &control_tlv;
	}
#ifndef SEND_EMPTY_TLVS
	if (ipv4_tlv.ipv4_interfaces_nr != 0)
#endif
	{
		response_message.tlv_cstruct_list[i++] = (INT8U *) &ipv4_tlv;
	}
#ifndef SEND_EMPTY_TLVS
	if (ipv6_tlv.ipv6_interfaces_nr != 0)
#endif
	{
		response_message.tlv_cstruct_list[i++] = (INT8U *) &ipv6_tlv;
	}

	response_message.tlv_cstruct_list[i++] = NULL;

	// Send the packet
	//
	if (send1905RawPacket(interface_name, mid, destination_al_mac_address, &response_message) == 0) {
		ERROR("Could not send packet");
		FREE(response_message.tlv_cstruct_list);
		return 0;
	}

	// Free all allocated (and no longer needed) memory
	free_tlv_internal_list_CMDU_structure(&response_message);

	return 1;
}

INT8U send1905InterfacePowerChangeRequestPacket(char *interface_name, INT16U mid,
						INT8U *destination_al_mac_address, INT8U(*remote_interfaces)[6],
						INT8U *new_states, INT8U nr)
{
	// NOTE: Right now this function is *not* being used from anywhere. The
	//       reason is that the standard does not say under which circumstances
	//       this packet should be generated.
	//       There should probably exist an HLE primitive that triggers this
	//       event, but there isn't.
	//       I'll just let this function here implemented for the future.

	// The "high layer response" message is a CMDU with the following TLVs:
	//   - One or more interface power change information type TLVs
	//
	// However, it doesn't really make sense to send more than one (after all,
	// one single TLV can contain information regarding as many remote
	// interfaces as desired). The original wording is probably an error in the
	// standard and it should read like this:
	//   - One interface power change information type TLV
	//
	// So... here we are only going to send *one* TLV containing all the remote
	// interfaces requested new states

	INT8U ret;

	struct cmdu_cstruct request_message;
	struct interfacePowerChangeInformationTLV power_change;

	INT8U i;

	if (nr == 0) {
		return 1;
	}

	INFO("--> CMDU_TYPE_INTERFACE_POWER_CHANGE_REQUEST (%s)", interface_name);
	DEBUG("Sending to %02x:%02x:%02x:%02x:%02x:%02x",
	      destination_al_mac_address[0], destination_al_mac_address[1], destination_al_mac_address[2],
	      destination_al_mac_address[3], destination_al_mac_address[4], destination_al_mac_address[5]);

	// Fill the interface power change information type TLV
	//
	power_change.tlv_type = TLV_TYPE_INTERFACE_POWER_CHANGE_INFORMATION;
	power_change.power_change_interfaces_nr = nr;
	CALLOC(power_change.power_change_interfaces, nr);

	for (i = 0; i < nr; i++) {
		memcpy(power_change.power_change_interfaces[i].interface_address, remote_interfaces[i], 6);
		power_change.power_change_interfaces[i].requested_power_state = new_states[i];
	}

	// Build the CMDU
	request_message.message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	request_message.message_type = CMDU_TYPE_INTERFACE_POWER_CHANGE_REQUEST;
	request_message.message_id = mid;
	request_message.relay_indicator = 0;
	CALLOC(request_message.tlv_cstruct_list, 2);
	request_message.tlv_cstruct_list[0] = (INT8U *) &power_change;
	request_message.tlv_cstruct_list[1] = NULL;

	// Send the packet
	if (send1905RawPacket(interface_name, mid, destination_al_mac_address, &request_message) == 0) {
		WARNING("Could not send packet");
		ret = 0;
	} else {
		ret = 1;
	}

	// Free no longer needed memory
	free_tlv_internal_list_CMDU_structure(&request_message);

	return ret;
}

INT8U send1905InterfacePowerChangeResponsePacket(char *interface_name, INT16U mid, INT8U *destination_al_mac_address, INT8U(*local_interfaces)[6],
						 INT8U *results, INT8U nr)
{
	// The "high layer response" message is a CMDU with the following TLVs:
	//   - One or more interface power change status TLVs
	//
	// However, it doesn't really make sense to send more than one (after all,
	// one single TLV can contain information regarding as many remote
	// interfaces as desired). The original wording is probably an error in the
	// standard and it should read like this:
	//   - One interface power change status TLV
	//
	// So... here we are only going to send *one* TLV containing all the local
	// interfaces requested status

	INT8U ret;

	struct cmdu_cstruct response_message;
	struct interfacePowerChangeStatusTLV power_change;

	INT8U i;

	if (nr == 0) {
		return 1;
	}

	INFO("--> CMDU_TYPE_INTERFACE_POWER_CHANGE_response (%s)", interface_name);
	DEBUG("Sending to %02x:%02x:%02x:%02x:%02x:%02x",
	      destination_al_mac_address[0], destination_al_mac_address[1], destination_al_mac_address[2],
	      destination_al_mac_address[3], destination_al_mac_address[4], destination_al_mac_address[5]);

	// Fill the interface power change information type TLV
	power_change.tlv_type = TLV_TYPE_INTERFACE_POWER_CHANGE_INFORMATION;
	power_change.power_change_interfaces_nr = nr;
	CALLOC(power_change.power_change_interfaces, nr);

	for (i = 0; i < nr; i++) {
		memcpy(power_change.power_change_interfaces[i].interface_address, local_interfaces[i], 6);
		power_change.power_change_interfaces[i].result = results[i];
	}

	// Build the CMDU
	response_message.message_version = CMDU_MESSAGE_VERSION_1905_1_2013;
	response_message.message_type = CMDU_TYPE_INTERFACE_POWER_CHANGE_RESPONSE;
	response_message.message_id = mid;
	response_message.relay_indicator = 0;
	CALLOC(response_message.tlv_cstruct_list, 2);
	response_message.tlv_cstruct_list[0] = (INT8U *) &power_change;
	response_message.tlv_cstruct_list[1] = NULL;

	// Send the packet
	//
	if (send1905RawPacket(interface_name, mid, destination_al_mac_address, &response_message) == 0) {
		WARNING("Could not send packet");
		ret = 0;
	} else {
		ret = 1;
	}

	// Free no longer needed memory
	free_tlv_internal_list_CMDU_structure(&response_message);

	return ret;
}

INT8U sendLLDPBridgeDiscoveryPacket(char *interface_name)
{
	INT8U al_mac_address[6];
	INT8U interface_mac_address[6];

	struct chassisIdTLV chassis_id_tlv;
	struct portIdTLV port_id_tlv;
	struct timeToLiveTypeTLV time_to_live_tlv;

	struct PAYLOAD payload;

	INT8U *stream;
	INT16U stream_len;

	INFO("--> LLDP BRIDGE DISCOVERY (%s)", interface_name);

	memcpy(al_mac_address, DMalMacGet(), 6);
	memcpy(interface_mac_address, DMinterfaceNameToMac(interface_name), 6);

	// Fill the chassis ID TLV
	//
	chassis_id_tlv.tlv_type = TLV_TYPE_CHASSIS_ID;
	chassis_id_tlv.chassis_id_subtype = CHASSIS_ID_TLV_SUBTYPE_MAC_ADDRESS;
	chassis_id_tlv.chassis_id[0] = al_mac_address[0];
	chassis_id_tlv.chassis_id[1] = al_mac_address[1];
	chassis_id_tlv.chassis_id[2] = al_mac_address[2];
	chassis_id_tlv.chassis_id[3] = al_mac_address[3];
	chassis_id_tlv.chassis_id[4] = al_mac_address[4];
	chassis_id_tlv.chassis_id[5] = al_mac_address[5];

	// Fill the port ID TLV
	//
	port_id_tlv.tlv_type = TLV_TYPE_PORT_ID;
	port_id_tlv.port_id_subtype = PORT_ID_TLV_SUBTYPE_MAC_ADDRESS;
	port_id_tlv.port_id[0] = interface_mac_address[0];
	port_id_tlv.port_id[1] = interface_mac_address[1];
	port_id_tlv.port_id[2] = interface_mac_address[2];
	port_id_tlv.port_id[3] = interface_mac_address[3];
	port_id_tlv.port_id[4] = interface_mac_address[4];
	port_id_tlv.port_id[5] = interface_mac_address[5];

	// Fill the time to live TLV
	//
	time_to_live_tlv.tlv_type = TLV_TYPE_TIME_TO_LIVE;
	time_to_live_tlv.ttl = TIME_TO_LIVE_TLV_1905_DEFAULT_VALUE;

	// Forge the LLDP payload containing all these TLVs
	//
	payload.tlv_cstruct_list[0] = (INT8U *) &chassis_id_tlv;
	payload.tlv_cstruct_list[1] = (INT8U *) &port_id_tlv;
	payload.tlv_cstruct_list[2] = (INT8U *) &time_to_live_tlv;
	payload.tlv_cstruct_list[3] = NULL;

	stream = forge_lldp_PAYLOAD_from_structure(&payload, &stream_len);

	// Finally, send the packet!
	//
	{
		INT8U mcast_address[] = MCAST_LLDP;

		DEBUG("Sending LLDP bridge discovery message on interface %s", interface_name);
		if (PLATFORM_SEND_RAW_PACKET(interface_name, mcast_address, interface_mac_address, ETHERTYPE_LLDP, stream, stream_len) == 0) {
			ERROR("Packet could not be sent!");
		}
	}

	// Free memory
	free_lldp_PAYLOAD_packet(stream);

	return 1;
}
