/*
 * 1905_alme.c: IEEE-1905 ALME msg handling
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "utils.h"
#include "utils_json.h"

#include "1905_alme.h"
#include "cmdu_visit.h"
#include "platform_interfaces.h"
#include "al_datamodel.h"
#include "al_send.h"

#include <easy/easy.h>

#include <libubox/blobmsg_json.h>

static struct blob_buf *s_alme_blob;
static INT8U gs_alme_client_id;
////////////////////////////////////////////////////////////////////////////////
// Custom (non-standarized) packet structure for standard ALME primitives:
////////////////////////////////////////////////////////////////////////////////
//
// ALME-GET-INTF-LIST.request
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0x01
//
//
// ALME-GET-INTF-LIST.response
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0x02
//  byte # 1: interface_descriptors_nr
//
//  byte # 2: interface_address[0]                                         |
//  byte # 3: interface_address[1]                                         |
//  byte # 4: interface_address[2]                                         |
//  byte # 5: interface_address[3]                                         |
//  byte # 6: interface_address[4]                                         | repeat
//  byte # 7: interface_address[5]                                         | "interface_descriptors_nr"
//  byte # 8: interface_type MSB                                           | times
//  byte # 9: interface_type LSB                                           |
//  byte #10: bridge_flag                                                  |
//  byte #11: vendor_specific_info_nr                                      |
//                                                                         |
//  byte #12: ie_type MSB                  |                               |
//  byte #13: ie_type LSB                  |                               |
//  byte #14: length_field MSB             |                               |
//  byte #15: length_field LSB             |                               |
//  byte #16: oui[0]                       | repeat                        |
//  byte #17: oui[1]                       | "vendor_specific_info_nr"     |
//  byte #18: oui[2]                       | times                         |
//  byte #19: vendor_si[0]                 |                               |
//  byte #20: vendor_si[1]                 |                               |
//  ...                                    |                               |
//  byte #N : vendor_si[length_field-1]    |                               |
//
//
// ALME-SET-INTF-PWR-STATE.request
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0x03
//  byte # 1: interface_address[0]
//  byte # 2: interface_address[1]
//  byte # 3: interface_address[2]
//  byte # 4: interface_address[3]
//  byte # 5: interface_address[4]
//  byte # 6: interface_address[5]
//  byte # 7: power_state
//
//
// ALME-SET-INTF-PWR-STATE.confirm
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0x04
//  byte # 1: interface_address[0]
//  byte # 2: interface_address[1]
//  byte # 3: interface_address[2]
//  byte # 4: interface_address[3]
//  byte # 5: interface_address[4]
//  byte # 6: interface_address[5]
//  byte # 7: reason_code
//
//
// ALME-GET-INTF-PWR-STATE.request
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0x05
//  byte # 1: interface_address[0]
//  byte # 2: interface_address[1]
//  byte # 3: interface_address[2]
//  byte # 4: interface_address[3]
//  byte # 5: interface_address[4]
//  byte # 6: interface_address[5]
//
//
// ALME-GET-INTF-PWR-STATE.response
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0x06
//  byte # 1: interface_address[0]
//  byte # 2: interface_address[1]
//  byte # 3: interface_address[2]
//  byte # 4: interface_address[3]
//  byte # 5: interface_address[4]
//  byte # 6: interface_address[5]
//  byte # 7: reason_code
//
// ALME-SET-FWD-RULE.request
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0x07
//  byte # 1: mac_da[0]
//  byte # 2: mac_da[1]
//  byte # 3: mac_da[2]
//  byte # 4: mac_da[3]
//  byte # 5: mac_da[4]
//  byte # 6: mac_da[5]
//  byte # 7: mac_da_flag
//  byte # 8: mac_sa[0]
//  byte # 9: mac_sa[1]
//  byte #10: mac_sa[2]
//  byte #11: mac_sa[3]
//  byte #12: mac_sa[4]
//  byte #13: mac_sa[5]
//  byte #14: mac_sa_flag
//  byte #15: ether_type MSB
//  byte #16: ether_type LSB
//  byte #17: ether_type_flag
//  byte #18: 0x00 | vid 4 MSbits   -> "vid" is
//  byte #19: vid LSB                  12 bits long
//  byte #20: vid_flag
//  byte #21: 0x00 | pcp 3 LSBits   -> "pcp" is 3 bits long
//  byte #22: pcp_flag
//  byte #23: addresses_nr
//
//  byte #24: addresses[0][0]    |
//  byte #25: addresses[0][1]    | repeat
//  byte #26: addresses[0][2]    | "addresses_nr"
//  byte #27: addresses[0][3]    | times
//  byte #28: addresses[0][4]    | (with addresses[1],
//  byte #29: addresses[0][5]    | addresses[2], etc...)
//
//
// ALME-SET-FWD-RULE.confirm
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0x08
//  byte # 1: rule_id MSB
//  byte # 2: rule_id LSB
//  byte # 3: reason_code
//
//
// ALME-GET-FWD-RULE.request
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0x09
//
//
// ALME-GET-FWD-RULE.response
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0x10
//  byte # 1: rules_nr
//
//  byte # 2: mac_da[0]                                                    |
//  byte # 3: mac_da[1]                                                    |
//  byte # 4: mac_da[2]                                                    |
//  byte # 5: mac_da[3]                                                    |
//  byte # 6: mac_da[4]                                                    |
//  byte # 7: mac_da[5]                                                    |
//  byte # 8: mac_da_flag                                                  |
//  byte # 9: mac_sa[0]                                                    |
//  byte #10: mac_sa[1]                                                    |
//  byte #11: mac_sa[2]                                                    |
//  byte #12: mac_sa[3]                                                    | repeat
//  byte #13: mac_sa[4]                                                    | "rules_nr"
//  byte #14: mac_sa[5]                                                    | times
//  byte #15: mac_sa_flag                                                  |
//  byte #16: ether_type MSB                                               |
//  byte #17: ether_type LSB                                               |
//  byte #18: ether_type_flag                                              |
//  byte #19: 0x00 | vid 4 MSbits   -> "vid" is                            |
//  byte #20: vid LSB                  12 bits long                        |
//  byte #21: vid_flag                                                     |
//  byte #22: 0x00 | pcp 3 LSBits   -> "pcp" is 3 bits long                |
//  byte #23: pcp_flag                                                     |
//  byte #24: addresses_nr                                                 |
//                                                                         |
//  byte #25: addresses[0][0]    |                                         |
//  byte #26: addresses[0][1]    | repeat                                  |
//  byte #27: addresses[0][2]    | "addresses_nr"                          |
//  byte #28: addresses[0][3]    | times                                   |
//  byte #29: addresses[0][4]    | (with addresses[1],                     |
//  byte #30: addresses[0][5]    | addresses[2], etc...)                   |
//                                                                         |
//  byte #N  : last_matched MSB                                            |
//  byte #N+1: last_matched LSB                                            |
//
//
// ALME-MODIFY-FWD-RULE.request
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0x0a
//  byte # 1: rule_id MSB
//  byte # 2: rule_id LSB
//  byte # 3: addresses_nr
//
//  byte #25: addresses[0][0]    |
//  byte #26: addresses[0][1]    | repeat
//  byte #27: addresses[0][2]    | "addresses_nr"
//  byte #28: addresses[0][3]    | times
//  byte #29: addresses[0][4]    | (with addresses[1],
//  byte #30: addresses[0][5]    | addresses[2], etc...)
//
//
// ALME-MODIFY-FWD-RULE.confirm
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0x0b
//  byte # 1: rule_id MSB
//  byte # 2: rule_id LSB
//  byte # 3: reason_code
//
//
// ALME-REMOVE-FWD-RULE.request
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0x0c
//  byte # 1: rule_id MSB
//  byte # 2: rule_id LSB
//
//
// ALME-REMOVE-FWD-RULE.confirm
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0x0d
//  byte # 1: rule_id MSB
//  byte # 2: rule_id LSB
//  byte # 3: reason_code
//
//
// ALME-GET-METRIC.request
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0x0e
//
//  byte # 1: interface_address[0]
//  byte # 2: interface_address[1]
//  byte # 3: interface_address[2]
//  byte # 4: interface_address[3]
//  byte # 5: interface_address[4]
//  byte # 6: interface_address[5]
//
//
// ALME-GET-METRIC.response
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0x0f
//  byte # 1: metrics_nr
//
//  byte # 2: neighbor_dev_address[0]                                      |
//  byte # 3: neighbor_dev_address[1]                                      | repeat
//  byte # 4: neighbor_dev_address[2]                                      | "metrics_nr"
//  byte # 5: neighbor_dev_address[3]                                      | times
//  byte # 6: neighbor_dev_address[4]                                      | (one time for
//  byte # 7: neighbor_dev_address[5]                                      | each local interface
//  byte # 8: local_intf_address[0]                                        | connected to a
//  byte # 9: local_intf_address[1]                                        | remote interface
//  byte #10: local_intf_address[2]                                        | of the neighbor
//  byte #11: local_intf_address[3]                                        | node)
//  byte #12: local_intf_address[4]                                        |
//  byte #13: local_intf_address[5]                                        |
//  byte #14: bridge_flag                                                  |
//                                                                         |
//  byte #15: tlv_type   = 0x9 (transmitter link metrics)                  |
//  byte #16: tlv_length = 12 + 29*1                                       |
//  byte #17: tlv_value[0]         | Only contains one metrics element:    |
//  ...                            | the one involving                     |
//  byte #57: tlv_value[12+29-1]   | "local_intf_address"                  |
//                                                                         |
//  byte #58: tlv_type   = 0x10 (receiver link metrics)                    |
//  byte #59: tlv_length = 12 + 23*1                                       |
//  byte #60: tlv_value[0]         | Only contains one metrics element:    |
//  ...                            | the one involving                     |
//  byte #94: tlv_value[12+23-1]   | "local_intf_address"                  |
//
//  NOTES:
//    * The contents of bytes #17 to #57 are defined in "IEEE Std 1905.1-2013
//      Table 6-17" with "n=1" (ie. only one connected interface -the one
//      that matches "local_intf_address"- is reported)
//    * The contents of bytes #60 to #94 are defined in "IEEE Std 1905.1-2013
//      Table 6-19" with "n=1" (ie. only one connected interface -the one
//      that matches "local_intf_address"- is reported)

////////////////////////////////////////////////////////////////////////////////
// Private (non-standarized) packet structure for custom (not present in the
// standard) ALME primitives:
////////////////////////////////////////////////////////////////////////////////
//
// NOTE: We are using "reserved" 'alme_type' values. We might have to remove
// these new "custom" ALMEs if the standard is ever updated to make use of these
// types.
//
// ALME-CUSTOM-COMMAND.request
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0xf0
//  byte # 1: command
//
// ALME-CUSTOM-COMMAND.response
// ----------------------------------------------------------------------------
//  byte # 0: alme_type = 0xf0
//  byte # 1: length MSB
//  byte # 2: length LSB
//  byte # 3: data[0]                         |
//  ...                                       | Custom response payload
//  byte # (length + 3 - 1) : data[length-1]  |

////////////////////////////////////////////////////////////////////////////////
// Actual API functions
////////////////////////////////////////////////////////////////////////////////

INT8U *parse_1905_ALME_from_packet(INT8U *packet_stream)
{
	if (packet_stream == NULL) {
		return NULL;
	}
	// The first byte of the stream is the "Type" field from the ALME structure.
	// Valid values for this byte are the following ones...
	//
	switch (*packet_stream) {
	case ALME_TYPE_GET_INTF_LIST_REQUEST:
		{
			struct getIntfListRequestALME *ret;

			INT8U *p;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);

			return (INT8U *) ret;
		}

	case ALME_TYPE_GET_INTF_LIST_RESPONSE:
		{
			struct getIntfListResponseALME *ret;

			INT8U *p;
			INT8U i, j;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);
			_E1B(&p, &ret->interface_descriptors_nr);

			if (ret->interface_descriptors_nr > 0) {
				CALLOC(ret->interface_descriptors, ret->interface_descriptors_nr);

				for (i = 0; i < ret->interface_descriptors_nr; i++) {
					_EnB(&p, ret->interface_descriptors[i].interface_address, 6);
					_E2B(&p, &ret->interface_descriptors[i].interface_type);
					_E1B(&p, &ret->interface_descriptors[i].bridge_flag);
					_E1B(&p, &ret->interface_descriptors[i].vendor_specific_info_nr);

					if (ret->interface_descriptors[i].vendor_specific_info_nr > 0) {
						CALLOC(ret->interface_descriptors[i].vendor_specific_info, \
						       ret->interface_descriptors[i].vendor_specific_info_nr);

						for (j = 0; j < ret->interface_descriptors[i].vendor_specific_info_nr; j++) {
							_E2B(&p, &ret->interface_descriptors[i].vendor_specific_info[j].ie_type);
							_E2B(&p, &ret->interface_descriptors[i].vendor_specific_info[j].length_field);
							_EnB(&p, ret->interface_descriptors[i].vendor_specific_info[j].oui, 3);

							if (ret->interface_descriptors[i].vendor_specific_info[j].length_field > 3) {
								CALLOC(ret->interface_descriptors[i].vendor_specific_info[j].vendor_si,\
								       (ret->interface_descriptors[i].vendor_specific_info[j].length_field - 3));

								_EnB(&p, ret->interface_descriptors[i].vendor_specific_info[j].vendor_si,
								     ret->interface_descriptors[i].vendor_specific_info[j].length_field - 3);
							} else {
								ret->interface_descriptors[i].vendor_specific_info[j].vendor_si = NULL;
							}
						}
					}
				}
			}

			return (INT8U *) ret;
		}

	case ALME_TYPE_SET_INTF_PWR_STATE_REQUEST:
		{
			struct setIntfPwrStateRequestALME *ret;

			INT8U *p;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);
			_EnB(&p, ret->interface_address, 6);
			_E1B(&p, &ret->power_state);

			return (INT8U *) ret;
		}

	case ALME_TYPE_SET_INTF_PWR_STATE_CONFIRM:
		{
			struct setIntfPwrStateConfirmALME *ret;

			INT8U *p;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);
			_EnB(&p, ret->interface_address, 6);
			_E1B(&p, &ret->reason_code);

			return (INT8U *) ret;
		}

	case ALME_TYPE_GET_INTF_PWR_STATE_REQUEST:
		{
			struct getIntfPwrStateRequestALME *ret;

			INT8U *p;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);
			_EnB(&p, ret->interface_address, 6);

			return (INT8U *) ret;
		}

	case ALME_TYPE_GET_INTF_PWR_STATE_RESPONSE:
		{
			struct getIntfPwrStateResponseALME *ret;

			INT8U *p;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);
			_EnB(&p, ret->interface_address, 6);
			_E1B(&p, &ret->power_state);

			return (INT8U *) ret;
		}

	case ALME_TYPE_SET_FWD_RULE_REQUEST:
		{
			struct setFwdRuleRequestALME *ret;

			INT8U *p;
			INT8U i;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);
			_EnB(&p, ret->classification_set.mac_da, 6);
			_E1B(&p, &ret->classification_set.mac_da_flag);
			_EnB(&p, ret->classification_set.mac_sa, 6);
			_E1B(&p, &ret->classification_set.mac_sa_flag);
			_E2B(&p, &ret->classification_set.ether_type);
			_E1B(&p, &ret->classification_set.ether_type_flag);
			_E2B(&p, &ret->classification_set.vid);
			_E1B(&p, &ret->classification_set.vid_flag);
			_E1B(&p, &ret->classification_set.pcp);
			_E1B(&p, &ret->classification_set.pcp_flag);
			_E1B(&p, &ret->addresses_nr);

			if (ret->addresses_nr > 0) {
				ret->addresses = (INT8U(*)[6]) malloc(sizeof(INT8U[6]) * ret->addresses_nr);
				if (ret->addresses == NULL) {
					ERROR("Out of memory!");
					exit(ENOMEM);
				}

				for (i = 0; i < ret->addresses_nr; i++) {
					_EnB(&p, ret->addresses[i], 6);
				}
			}

			return (INT8U *) ret;
		}

	case ALME_TYPE_SET_FWD_RULE_CONFIRM:
		{
			struct setFwdRuleConfirmALME *ret;

			INT8U *p;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);
			_E2B(&p, &ret->rule_id);
			_E1B(&p, &ret->reason_code);

			return (INT8U *) ret;
		}

	case ALME_TYPE_GET_FWD_RULES_REQUEST:
		{
			struct getFwdRulesRequestALME *ret;

			INT8U *p;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);

			return (INT8U *) ret;
		}

	case ALME_TYPE_GET_FWD_RULES_RESPONSE:
		{
			struct getFwdRulesResponseALME *ret;

			INT8U *p;
			INT8U i, j;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);
			_E1B(&p, &ret->rules_nr);

			if (ret->rules_nr > 0) {
				CALLOC(ret->rules, ret->rules_nr);

				for (i = 0; i < ret->rules_nr; i++) {
					_EnB(&p, ret->rules[i].classification_set.mac_da, 6);
					_E1B(&p, &ret->rules[i].classification_set.mac_da_flag);
					_EnB(&p, ret->rules[i].classification_set.mac_sa, 6);
					_E1B(&p, &ret->rules[i].classification_set.mac_sa_flag);
					_E2B(&p, &ret->rules[i].classification_set.ether_type);
					_E1B(&p, &ret->rules[i].classification_set.ether_type_flag);
					_E2B(&p, &ret->rules[i].classification_set.vid);
					_E1B(&p, &ret->rules[i].classification_set.vid_flag);
					_E1B(&p, &ret->rules[i].classification_set.pcp);
					_E1B(&p, &ret->rules[i].classification_set.pcp_flag);
					_E1B(&p, &ret->rules[i].addresses_nr);

					if (ret->rules[i].addresses_nr > 0) {
						ret->rules[i].addresses = (INT8U(*)[6]) malloc(sizeof(INT8U[6]) * ret->rules[i].addresses_nr);
						if (ret->rules[i].addresses == NULL) {
							ERROR("Out of memory!");
							exit(ENOMEM);
						}

						for (j = 0; j < ret->rules[i].addresses_nr; j++) {
							_EnB(&p, ret->rules[i].addresses[j], 6);
						}
					}

					_E2B(&p, &ret->rules[i].last_matched);
				}
			}

			return (INT8U *) ret;
		}

	case ALME_TYPE_MODIFY_FWD_RULE_REQUEST:
		{
			struct modifyFwdRuleRequestALME *ret;

			INT8U *p;
			INT8U i;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);
			_E2B(&p, &ret->rule_id);
			_E1B(&p, &ret->addresses_nr);

			if (ret->addresses_nr) {
				ret->addresses = (INT8U(*)[6]) malloc(sizeof(INT8U[6]) * ret->addresses_nr);
				if (ret->addresses == NULL) {
					ERROR("Out of memory!");
					exit(ENOMEM);
				}

				for (i = 0; i < ret->addresses_nr; i++)
					_EnB(&p, ret->addresses[i], 6);
			}

			return (INT8U *) ret;
		}

	case ALME_TYPE_MODIFY_FWD_RULE_CONFIRM:
		{
			struct modifyFwdRuleConfirmALME *ret;

			INT8U *p;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);
			_E2B(&p, &ret->rule_id);
			_E1B(&p, &ret->reason_code);

			return (INT8U *) ret;
		}

	case ALME_TYPE_REMOVE_FWD_RULE_REQUEST:
		{
			struct removeFwdRuleRequestALME *ret;

			INT8U *p;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);
			_E2B(&p, &ret->rule_id);

			return (INT8U *) ret;
		}

	case ALME_TYPE_REMOVE_FWD_RULE_CONFIRM:
		{
			struct removeFwdRuleConfirmALME *ret;

			INT8U *p;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);
			_E2B(&p, &ret->rule_id);
			_E1B(&p, &ret->reason_code);

			return (INT8U *) ret;
		}

	case ALME_TYPE_GET_METRIC_REQUEST:
		{
			struct getMetricRequestALME *ret;

			INT8U *p;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);
			_EnB(&p, ret->interface_address, 6);

			return (INT8U *) ret;
		}

	case ALME_TYPE_GET_METRIC_RESPONSE:
		{
			struct getMetricResponseALME *ret;

			INT8U *p;
			INT8U i;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);
			_E1B(&p, &ret->metrics_nr);

			if (ret->metrics_nr > 0) {
				CALLOC(ret->metrics, ret->metrics_nr);

				for (i = 0; i < ret->metrics_nr; i++) {
					struct transmitterLinkMetricTLV *tx;
					struct receiverLinkMetricTLV *rx;

					INT8U aux1;
					INT16U aux2;

					_EnB(&p, ret->metrics[i].neighbor_dev_address, 6);
					_EnB(&p, ret->metrics[i].local_intf_address, 6);
					_E1B(&p, &ret->metrics[i].bridge_flag);

					tx = (struct transmitterLinkMetricTLV *)parse_1905_TLV_from_packet(p);

					if (NULL == tx ||
					    memcmp(tx->neighbor_al_address, ret->metrics[i].neighbor_dev_address, 6) ||
					    1 != tx->transmitter_link_metrics_nr ||
					    NULL == tx->transmitter_link_metrics ||
					    memcmp(tx->transmitter_link_metrics[0].local_interface_address, ret->metrics[i].local_intf_address, 6)
					    ) {
						// Parsing error
						//
						if (tx != NULL) {
							free_1905_TLV_structure((INT8U *) tx, true);
						}
						FREE(ret->metrics);
						FREE(ret);
						return NULL;
					}

					_E1B(&p, &aux1);
					_E2B(&p, &aux2);
					p += aux2;

					rx = (struct receiverLinkMetricTLV *)parse_1905_TLV_from_packet(p);

					if (NULL == rx ||
					    memcmp(rx->neighbor_al_address, ret->metrics[i].neighbor_dev_address, 6) ||
					    1 != rx->receiver_link_metrics_nr ||
					    NULL == rx->receiver_link_metrics ||
					    memcmp(rx->receiver_link_metrics[0].local_interface_address, ret->metrics[i].local_intf_address, 6)
					    ) {
						// Parsing error
						//
						if (rx != NULL) {
							free_1905_TLV_structure((INT8U *) rx, true);
						}
						FREE(ret->metrics);
						FREE(ret);
						return NULL;
					}

					_E1B(&p, &aux1);
					_E2B(&p, &aux2);
					p += aux2;

					ret->metrics[i].tx_metric = tx;
					ret->metrics[i].rx_metric = rx;
				}
			}

			_E1B(&p, &ret->reason_code);

			return (INT8U *) ret;
		}

	case ALME_TYPE_CUSTOM_COMMAND_REQUEST:
		{
			struct customCommandRequestALME *ret;

			INT8U *p;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);
			_E1B(&p, &ret->command);

			return (INT8U *) ret;
		}

	case ALME_TYPE_CUSTOM_COMMAND_RESPONSE:
		{
			struct customCommandResponseALME *ret;

			INT8U *p;

			p = packet_stream;

			CALLOC(ret, 1);

			_E1B(&p, &ret->alme_type);
			_E2B(&p, &ret->bytes_nr);

			if (ret->bytes_nr > 0) {
				CALLOC(ret->bytes, ret->bytes_nr);
				_EnB(&p, ret->bytes, ret->bytes_nr);
			}

			return (INT8U *) ret;
		}

	default:
		{
			// Ignore
			//
			return NULL;
		}

	}

	// This code cannot be reached
	//
	return NULL;
}

INT8U *forge_1905_ALME_from_structure(INT8U *memory_structure, INT16U *len)
{
	if (memory_structure == NULL) {
		return NULL;
	}
	// The first byte of any of the valid structures is always the "tlv_type"
	// field.
	//
	switch (*memory_structure) {
	case ALME_TYPE_GET_INTF_LIST_REQUEST:
		{
			INT8U *ret, *p;
			struct getIntfListRequestALME *m;

			m = (struct getIntfListRequestALME *)memory_structure;

			*len = 1;	// alme_type

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);

			return ret;
		}

	case ALME_TYPE_GET_INTF_LIST_RESPONSE:
		{
			INT8U *ret, *p;
			struct getIntfListResponseALME *m;

			INT8U i, j;

			m = (struct getIntfListResponseALME *)memory_structure;

			*len = 2;	// alme_type + interface_descriptors_nr
			for (i = 0; i < m->interface_descriptors_nr; i++) {
				*len += 6;	// interface_address
				*len += 2;	// interface_type
				*len += 1;	// bridge_flag
				*len += 1;	// vendor_specific_info_nr

				for (j = 0; j < m->interface_descriptors[i].vendor_specific_info_nr; j++) {
					*len += 7;	//ie_type + length_field + oui
					*len += m->interface_descriptors[i].vendor_specific_info[j].length_field - 3;	// vendor_si
				}
			}

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);
			_I1B(&m->interface_descriptors_nr, &p);

			for (i = 0; i < m->interface_descriptors_nr; i++) {
				_InB(m->interface_descriptors[i].interface_address, &p, 6);
				_I2B(&m->interface_descriptors[i].interface_type, &p);
				_I1B(&m->interface_descriptors[i].bridge_flag, &p);
				_I1B(&m->interface_descriptors[i].vendor_specific_info_nr, &p);

				for (j = 0; j < m->interface_descriptors[i].vendor_specific_info_nr; j++) {
					_I2B(&m->interface_descriptors[i].vendor_specific_info[j].ie_type, &p);
					_I2B(&m->interface_descriptors[i].vendor_specific_info[j].length_field, &p);
					_InB(m->interface_descriptors[i].vendor_specific_info[j].oui, &p, 3);

					if (m->interface_descriptors[i].vendor_specific_info[j].length_field > 3) {
						_InB(m->interface_descriptors[i].vendor_specific_info[j].vendor_si, &p,
						     m->interface_descriptors[i].vendor_specific_info[j].length_field - 3);
					}
				}
			}

			return ret;
		}

	case ALME_TYPE_SET_INTF_PWR_STATE_REQUEST:
		{
			INT8U *ret, *p;
			struct setIntfPwrStateRequestALME *m;

			m = (struct setIntfPwrStateRequestALME *)memory_structure;

			*len = 8;	// alme_type + interface_address + power_state

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);
			_InB(m->interface_address, &p, 6);
			_I1B(&m->power_state, &p);

			return ret;
		}

	case ALME_TYPE_SET_INTF_PWR_STATE_CONFIRM:
		{
			INT8U *ret, *p;
			struct setIntfPwrStateConfirmALME *m;

			m = (struct setIntfPwrStateConfirmALME *)memory_structure;

			*len = 8;	// alme_type + interface_address + reason_code

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);
			_InB(m->interface_address, &p, 6);
			_I1B(&m->reason_code, &p);

			return ret;
		}

	case ALME_TYPE_GET_INTF_PWR_STATE_REQUEST:
		{
			INT8U *ret, *p;
			struct getIntfPwrStateRequestALME *m;

			m = (struct getIntfPwrStateRequestALME *)memory_structure;

			*len = 7;	// alme_type + interface_address

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);
			_InB(m->interface_address, &p, 6);

			return ret;
		}

	case ALME_TYPE_GET_INTF_PWR_STATE_RESPONSE:
		{
			INT8U *ret, *p;
			struct getIntfPwrStateResponseALME *m;

			m = (struct getIntfPwrStateResponseALME *)memory_structure;

			*len = 8;	// alme_type + interface_address + power_state

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);
			_InB(m->interface_address, &p, 6);
			_I1B(&m->power_state, &p);

			return ret;
		}

	case ALME_TYPE_SET_FWD_RULE_REQUEST:
		{
			INT8U *ret, *p;
			struct setFwdRuleRequestALME *m;

			INT8U i;

			m = (struct setFwdRuleRequestALME *)memory_structure;

			*len = 24;	// alme_type + classification_set + addresses_nr
			*len += 6 * m->addresses_nr;	// addresses

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);
			_InB(m->classification_set.mac_da, &p, 6);
			_I1B(&m->classification_set.mac_da_flag, &p);
			_InB(m->classification_set.mac_sa, &p, 6);
			_I1B(&m->classification_set.mac_sa_flag, &p);
			_I2B(&m->classification_set.ether_type, &p);
			_I1B(&m->classification_set.ether_type_flag, &p);
			_I2B(&m->classification_set.vid, &p);
			_I1B(&m->classification_set.vid_flag, &p);
			_I1B(&m->classification_set.pcp, &p);
			_I1B(&m->classification_set.pcp_flag, &p);
			_I1B(&m->addresses_nr, &p);

			for (i = 0; i < m->addresses_nr; i++) {
				_InB(m->addresses[i], &p, 6);
			}

			return ret;
		}

	case ALME_TYPE_SET_FWD_RULE_CONFIRM:
		{
			INT8U *ret, *p;
			struct setFwdRuleConfirmALME *m;

			m = (struct setFwdRuleConfirmALME *)memory_structure;

			*len = 4;	// alme_type + rule_id + reason_code

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);
			_I2B(&m->rule_id, &p);
			_I1B(&m->reason_code, &p);

			return ret;
		}

	case ALME_TYPE_GET_FWD_RULES_REQUEST:
		{
			INT8U *ret, *p;
			struct getFwdRulesRequestALME *m;

			m = (struct getFwdRulesRequestALME *)memory_structure;

			*len = 1;	// alme_type

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);

			return ret;
		}

	case ALME_TYPE_GET_FWD_RULES_RESPONSE:
		{
			INT8U *ret, *p;
			struct getFwdRulesResponseALME *m;

			INT8U i, j;

			m = (struct getFwdRulesResponseALME *)memory_structure;

			*len = 2;	// alme_type + rules_nr
			for (i = 0; i < m->rules_nr; i++) {
				*len += 23;	// classification_set + addresses_nr
				*len += 6 * m->rules[i].addresses_nr;	// addresses
				*len += 2;	// last_matched
			}

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);
			_I1B(&m->rules_nr, &p);

			for (i = 0; i < m->rules_nr; i++) {
				_InB(m->rules[i].classification_set.mac_da, &p, 6);
				_I1B(&m->rules[i].classification_set.mac_da_flag, &p);
				_InB(m->rules[i].classification_set.mac_sa, &p, 6);
				_I1B(&m->rules[i].classification_set.mac_sa_flag, &p);
				_I2B(&m->rules[i].classification_set.ether_type, &p);
				_I1B(&m->rules[i].classification_set.ether_type_flag, &p);
				_I2B(&m->rules[i].classification_set.vid, &p);
				_I1B(&m->rules[i].classification_set.vid_flag, &p);
				_I1B(&m->rules[i].classification_set.pcp, &p);
				_I1B(&m->rules[i].classification_set.pcp_flag, &p);
				_I1B(&m->rules[i].addresses_nr, &p);

				for (j = 0; j < m->rules[i].addresses_nr; j++) {
					_InB(m->rules[i].addresses[j], &p, 6);
				}

				_I2B(&m->rules[i].last_matched, &p);
			}

			return ret;
		}

	case ALME_TYPE_MODIFY_FWD_RULE_REQUEST:
		{
			INT8U *ret, *p;
			struct modifyFwdRuleRequestALME *m;

			INT8U i;

			m = (struct modifyFwdRuleRequestALME *)memory_structure;

			*len = 4;	// alme_type + rule_id + addresses_nr
			*len += 6 * m->addresses_nr;	// addresses

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);
			_I2B(&m->rule_id, &p);
			_I1B(&m->addresses_nr, &p);

			for (i = 0; i < m->addresses_nr; i++) {
				_InB(m->addresses[i], &p, 6);
			}

			return ret;
		}

	case ALME_TYPE_MODIFY_FWD_RULE_CONFIRM:
		{
			INT8U *ret, *p;
			struct modifyFwdRuleConfirmALME *m;

			m = (struct modifyFwdRuleConfirmALME *)memory_structure;

			*len = 4;	// alme_type + rule_id + reason_code

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);
			_I2B(&m->rule_id, &p);
			_I1B(&m->reason_code, &p);

			return ret;
		}

	case ALME_TYPE_REMOVE_FWD_RULE_REQUEST:
		{
			INT8U *ret, *p;
			struct removeFwdRuleRequestALME *m;

			m = (struct removeFwdRuleRequestALME *)memory_structure;

			*len = 3;	// alme_type + rule_id

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);
			_I2B(&m->rule_id, &p);

			return ret;
		}

	case ALME_TYPE_REMOVE_FWD_RULE_CONFIRM:
		{
			INT8U *ret, *p;
			struct removeFwdRuleConfirmALME *m;

			m = (struct removeFwdRuleConfirmALME *)memory_structure;

			*len = 4;	// alme_type + rule_id + reason_code

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);
			_I2B(&m->rule_id, &p);
			_I1B(&m->reason_code, &p);

			return ret;
		}

	case ALME_TYPE_GET_METRIC_REQUEST:
		{
			INT8U *ret, *p;
			struct getMetricRequestALME *m;

			m = (struct getMetricRequestALME *)memory_structure;

			*len = 7;	// alme_type + interface_address

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);
			_InB(m->interface_address, &p, 6);

			return ret;
		}

	case ALME_TYPE_GET_METRIC_RESPONSE:
		{
			INT8U *metric_stream;
			INT8U *ret, *p;
			struct getMetricResponseALME *m;

			INT8U i;

			m = (struct getMetricResponseALME *)memory_structure;

			*len = 2;	// alme_type + metrics_nr
			for (i = 0; i < m->metrics_nr; i++) {

				*len += 6;	// neighbor_dev_address
				*len += 6;	// local_intf_address
				*len += 1;	// bridge_flag

				metric_stream = forge_1905_TLV_from_structure((INT8U *) m->metrics[i].tx_metric);
				if (NULL == metric_stream) {
					// Forging error
					//
					*len = 0;
					return NULL;
				}

				*len += get_tlv_packet_len(metric_stream);
				free_1905_TLV_packet(metric_stream);

				metric_stream = forge_1905_TLV_from_structure((INT8U *) m->metrics[i].rx_metric);
				if (NULL == metric_stream) {
					// Forging error
					//
					*len = 0;
					return NULL;
				}

				*len += get_tlv_packet_len(metric_stream);
				free_1905_TLV_packet(metric_stream);
			}
			*len += 1;	// reason_code

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);
			_I1B(&m->metrics_nr, &p);

			for (i = 0; i < m->metrics_nr; i++) {
				INT16U metric_stream_len;

				struct transmitterLinkMetricTLV *tx;
				struct receiverLinkMetricTLV *rx;

				_InB(m->metrics[i].neighbor_dev_address, &p, 6);
				_InB(m->metrics[i].local_intf_address, &p, 6);
				_I1B(&m->metrics[i].bridge_flag, &p);

				tx = m->metrics[i].tx_metric;
				rx = m->metrics[i].rx_metric;

				if (NULL == tx ||
				    memcmp(tx->neighbor_al_address, m->metrics[i].neighbor_dev_address, 6) ||
				    NULL == tx->transmitter_link_metrics ||
				    memcmp(tx->transmitter_link_metrics[0].local_interface_address, m->metrics[i].local_intf_address, 6) ||
				    NULL == rx ||
				    memcmp(rx->neighbor_al_address, m->metrics[i].neighbor_dev_address, 6) ||
				    NULL == rx->receiver_link_metrics ||
				    memcmp(rx->receiver_link_metrics[0].local_interface_address, m->metrics[i].local_intf_address, 6)
				    ) {
					// Malformed packet
					//
					*len = 0;
					FREE(ret);
					return NULL;
				}

				metric_stream = forge_1905_TLV_from_structure((INT8U *) tx);
				metric_stream_len = get_tlv_packet_len(metric_stream);
				if (NULL == metric_stream || 0 == metric_stream_len) {
					// Forging error
					*len = 0;
					FREE(ret);
					return NULL;
				}
				_InB(metric_stream, &p, metric_stream_len);
				free_1905_TLV_packet(metric_stream);

				metric_stream = forge_1905_TLV_from_structure((INT8U *) rx);
				metric_stream_len = get_tlv_packet_len(metric_stream);
				if (NULL == metric_stream || 0 == metric_stream_len) {
					// Forging error
					*len = 0;
					FREE(ret);
					return NULL;
				}
				_InB(metric_stream, &p, metric_stream_len);
				free_1905_TLV_packet(metric_stream);
			}

			_I1B(&m->reason_code, &p);

			return ret;
		}

	case ALME_TYPE_CUSTOM_COMMAND_REQUEST:
		{
			INT8U *ret, *p;
			struct customCommandRequestALME *m;

			m = (struct customCommandRequestALME *)memory_structure;

			*len = 2;	// alme_type + command

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);
			_I1B(&m->command, &p);

			return ret;
		}

	case ALME_TYPE_CUSTOM_COMMAND_RESPONSE:
		{
			INT8U *ret, *p;
			struct customCommandResponseALME *m;

			m = (struct customCommandResponseALME *)memory_structure;

			*len = 3;	// alme_type + length
			*len += m->bytes_nr;

			CALLOC(ret, *len);
			p = ret;

			_I1B(&m->alme_type, &p);
			_I2B(&m->bytes_nr, &p);

			if (m->bytes_nr > 0) {
				_InB(m->bytes, &p, m->bytes_nr);
			}

			return ret;
		}

	default:
		{
			// Ignore
			//
			return NULL;
		}

	}

	// This code cannot be reached
	//
	return NULL;
}

void free_1905_ALME_structure(INT8U *memory_structure)
{
	if (memory_structure == NULL) {
		return;
	}
	// The first byte of any of the valid structures is always the "alme_type"
	// field.
	//
	switch (*memory_structure) {
	case ALME_TYPE_GET_INTF_LIST_REQUEST:
	case ALME_TYPE_SET_INTF_PWR_STATE_REQUEST:
	case ALME_TYPE_SET_INTF_PWR_STATE_CONFIRM:
	case ALME_TYPE_GET_INTF_PWR_STATE_REQUEST:
	case ALME_TYPE_GET_INTF_PWR_STATE_RESPONSE:
	case ALME_TYPE_SET_FWD_RULE_CONFIRM:
	case ALME_TYPE_GET_FWD_RULES_REQUEST:
	case ALME_TYPE_MODIFY_FWD_RULE_CONFIRM:
	case ALME_TYPE_REMOVE_FWD_RULE_REQUEST:
	case ALME_TYPE_REMOVE_FWD_RULE_CONFIRM:
	case ALME_TYPE_GET_METRIC_REQUEST:
	case ALME_TYPE_CUSTOM_COMMAND_REQUEST:
		{
			FREE(memory_structure);

			return;
		}

	case ALME_TYPE_GET_INTF_LIST_RESPONSE:
		{
			struct getIntfListResponseALME *m;
			INT8U i, j;

			m = (struct getIntfListResponseALME *)memory_structure;

			if (m->interface_descriptors_nr > 0 && NULL != m->interface_descriptors) {
				for (i = 0; i < m->interface_descriptors_nr; i++) {
					if (m->interface_descriptors[i].vendor_specific_info_nr > 0 && NULL != m->interface_descriptors[i].vendor_specific_info) {
						for (j = 0; j < m->interface_descriptors[i].vendor_specific_info_nr; j++) {
							if (m->interface_descriptors[i].vendor_specific_info[j].vendor_si != NULL) {
								FREE(m->interface_descriptors[i].vendor_specific_info[j].vendor_si);
							}
						}
						FREE(m->interface_descriptors[i].vendor_specific_info);
					}
				}
				FREE(m->interface_descriptors);
			}
			FREE(m);

			return;
		}

	case ALME_TYPE_SET_FWD_RULE_REQUEST:
		{
			struct setFwdRuleRequestALME *m;

			m = (struct setFwdRuleRequestALME *)memory_structure;

			if (m->addresses_nr > 0 && NULL != m->addresses) {
				FREE(m->addresses);
			}
			FREE(m);

			return;
		}

	case ALME_TYPE_GET_FWD_RULES_RESPONSE:
		{
			struct getFwdRulesResponseALME *m;
			INT8U i;

			m = (struct getFwdRulesResponseALME *)memory_structure;

			if (m->rules_nr > 0 && NULL != m->rules) {
				for (i = 0; i < m->rules_nr; i++) {
					if (m->rules[i].addresses) {
						FREE(m->rules[i].addresses);
					}
				}
				FREE(m->rules);
			}
			FREE(m);

			return;
		}

	case ALME_TYPE_MODIFY_FWD_RULE_REQUEST:
		{
			struct modifyFwdRuleRequestALME *m;

			m = (struct modifyFwdRuleRequestALME *)memory_structure;

			if (m->addresses_nr > 0 && NULL != m->addresses) {
				FREE(m->addresses);
			}
			FREE(m);

			return;
		}

	case ALME_TYPE_GET_METRIC_RESPONSE:
		{
			struct getMetricResponseALME *m;
			INT8U i;

			m = (struct getMetricResponseALME *)memory_structure;

			if (m->metrics_nr > 0 && NULL != m->metrics) {
				for (i = 0; i < m->metrics_nr; i++) {
					free_1905_TLV_structure((INT8U *) m->metrics[i].tx_metric, true);
					free_1905_TLV_structure((INT8U *) m->metrics[i].rx_metric, true);
				}
				FREE(m->metrics);
			}
			FREE(m);

			return;
		}

	case ALME_TYPE_CUSTOM_COMMAND_RESPONSE:
		{
			struct customCommandResponseALME *m;

			m = (struct customCommandResponseALME *)memory_structure;

			if (m->bytes_nr > 0 && NULL != m->bytes) {
				FREE(m->bytes);
			}
			FREE(m);

			return;
		}

	default:
		{
			// Ignore
			//
			return;
		}
	}

	// This code cannot be reached
	//
	return;
}

INT8U compare_1905_ALME_structures(INT8U *memory_structure_1, INT8U *memory_structure_2)
{
	if (NULL == memory_structure_1 || NULL == memory_structure_2) {
		return 1;
	}
	// The first byte of any of the valid structures is always the "tlv_type"
	// field.
	//
	if (*memory_structure_1 != *memory_structure_2) {
		return 1;
	}
	switch (*memory_structure_1) {
	case ALME_TYPE_GET_INTF_LIST_REQUEST:
		{
			// Nothing to compare (this ALME primitive is always empty)
			//
			return 0;
		}

	case ALME_TYPE_GET_INTF_LIST_RESPONSE:
		{
			struct getIntfListResponseALME *p1, *p2;
			INT8U i, j;

			p1 = (struct getIntfListResponseALME *)memory_structure_1;
			p2 = (struct getIntfListResponseALME *)memory_structure_2;

			if (p1->interface_descriptors_nr != p2->interface_descriptors_nr) {
				return 1;
			}

			if (p1->interface_descriptors_nr > 0 && (NULL == p1->interface_descriptors || NULL == p2->interface_descriptors)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->interface_descriptors_nr; i++) {
				if (memcmp(p1->interface_descriptors[i].interface_address, p2->interface_descriptors[i].interface_address, 6) ||
				    p1->interface_descriptors[i].interface_type != p2->interface_descriptors[i].interface_type ||
				    p1->interface_descriptors[i].bridge_flag != p2->interface_descriptors[i].bridge_flag ||
				    p1->interface_descriptors[i].vendor_specific_info_nr != p2->interface_descriptors[i].vendor_specific_info_nr) {
					return 1;
				}

				if (p1->interface_descriptors[i].vendor_specific_info_nr > 0
				    && (NULL == p1->interface_descriptors[i].vendor_specific_info
					|| NULL == p2->interface_descriptors[i].vendor_specific_info)) {
					// Malformed structure
					//
					return 1;
				}

				for (j = 0; j < p1->interface_descriptors[i].vendor_specific_info_nr; j++) {
					if (p1->interface_descriptors[i].vendor_specific_info[j].ie_type !=
					    p2->interface_descriptors[i].vendor_specific_info[j].ie_type
					    || p1->interface_descriptors[i].vendor_specific_info[j].length_field !=
					    p2->interface_descriptors[i].vendor_specific_info[j].length_field
					    || memcmp(p1->interface_descriptors[i].vendor_specific_info[j].oui,
							   p2->interface_descriptors[i].vendor_specific_info[j].oui, 3)
					    || memcmp(p1->interface_descriptors[i].vendor_specific_info[j].vendor_si,
							   p2->interface_descriptors[i].vendor_specific_info[j].vendor_si,
							   p1->interface_descriptors[i].vendor_specific_info[j].length_field - 3)
					    ) {
						return 1;
					}

				}
			}

			return 0;
		}

	case ALME_TYPE_SET_INTF_PWR_STATE_REQUEST:
		{
			struct setIntfPwrStateRequestALME *p1, *p2;

			p1 = (struct setIntfPwrStateRequestALME *)memory_structure_1;
			p2 = (struct setIntfPwrStateRequestALME *)memory_structure_2;

			if (memcmp(p1->interface_address, p2->interface_address, 6) || p1->power_state != p2->power_state) {
				return 1;
			}

			return 0;
		}

	case ALME_TYPE_SET_INTF_PWR_STATE_CONFIRM:
		{
			struct setIntfPwrStateConfirmALME *p1, *p2;

			p1 = (struct setIntfPwrStateConfirmALME *)memory_structure_1;
			p2 = (struct setIntfPwrStateConfirmALME *)memory_structure_2;

			if (memcmp(p1->interface_address, p2->interface_address, 6) || p1->reason_code != p2->reason_code) {
				return 1;
			}

			return 0;
		}

	case ALME_TYPE_GET_INTF_PWR_STATE_REQUEST:
		{
			struct getIntfPwrStateRequestALME *p1, *p2;

			p1 = (struct getIntfPwrStateRequestALME *)memory_structure_1;
			p2 = (struct getIntfPwrStateRequestALME *)memory_structure_2;

			if (memcmp(p1->interface_address, p2->interface_address, 6)) {
				return 1;
			}

			return 0;
		}

	case ALME_TYPE_GET_INTF_PWR_STATE_RESPONSE:
		{
			struct getIntfPwrStateResponseALME *p1, *p2;

			p1 = (struct getIntfPwrStateResponseALME *)memory_structure_1;
			p2 = (struct getIntfPwrStateResponseALME *)memory_structure_2;

			if (memcmp(p1->interface_address, p2->interface_address, 6) || p1->power_state != p2->power_state) {
				return 1;
			}

			return 0;
		}

	case ALME_TYPE_SET_FWD_RULE_REQUEST:
		{
			struct setFwdRuleRequestALME *p1, *p2;
			INT8U i;

			p1 = (struct setFwdRuleRequestALME *)memory_structure_1;
			p2 = (struct setFwdRuleRequestALME *)memory_structure_2;

			if (memcmp(p1->classification_set.mac_da, p2->classification_set.mac_da, 6) ||
			    p1->classification_set.mac_da_flag != p2->classification_set.mac_da_flag ||
			    memcmp(p1->classification_set.mac_sa, p2->classification_set.mac_sa, 6) ||
			    p1->classification_set.mac_sa_flag != p2->classification_set.mac_sa_flag ||
			    p1->classification_set.ether_type != p2->classification_set.ether_type ||
			    p1->classification_set.ether_type_flag != p2->classification_set.ether_type_flag ||
			    p1->classification_set.vid != p2->classification_set.vid ||
			    p1->classification_set.vid_flag != p2->classification_set.vid_flag ||
			    p1->classification_set.pcp != p2->classification_set.pcp ||
			    p1->classification_set.pcp_flag != p2->classification_set.pcp_flag || p1->addresses_nr != p2->addresses_nr) {
				return 1;
			}

			if (p1->addresses_nr > 0 && (NULL == p1->addresses || NULL == p2->addresses)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->addresses_nr; i++) {
				if (memcmp(p1->addresses[i], p2->addresses[i], 6)) {
					return 1;
				}
			}

			return 0;
		}

	case ALME_TYPE_SET_FWD_RULE_CONFIRM:
		{
			struct setFwdRuleConfirmALME *p1, *p2;

			p1 = (struct setFwdRuleConfirmALME *)memory_structure_1;
			p2 = (struct setFwdRuleConfirmALME *)memory_structure_2;

			if (p1->rule_id != p2->rule_id || p1->reason_code != p2->reason_code) {
				return 1;
			}

			return 0;
		}

	case ALME_TYPE_GET_FWD_RULES_REQUEST:
		{
			// Nothing to compare (this ALME primitive is always empty)
			//
			return 0;
		}

	case ALME_TYPE_GET_FWD_RULES_RESPONSE:
		{
			struct getFwdRulesResponseALME *p1, *p2;
			INT8U i, j;

			p1 = (struct getFwdRulesResponseALME *)memory_structure_1;
			p2 = (struct getFwdRulesResponseALME *)memory_structure_2;

			if (p1->rules_nr != p2->rules_nr) {
				return 1;
			}

			if (p1->rules_nr > 0 && (NULL == p1->rules || NULL == p2->rules)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->rules_nr; i++) {
				if (memcmp(p1->rules[i].classification_set.mac_da, p2->rules[i].classification_set.mac_da, 6) ||
				    p1->rules[i].classification_set.mac_da_flag != p2->rules[i].classification_set.mac_da_flag ||
				    memcmp(p1->rules[i].classification_set.mac_sa, p2->rules[i].classification_set.mac_sa, 6) ||
				    p1->rules[i].classification_set.mac_sa_flag != p2->rules[i].classification_set.mac_sa_flag ||
				    p1->rules[i].classification_set.ether_type != p2->rules[i].classification_set.ether_type ||
				    p1->rules[i].classification_set.ether_type_flag != p2->rules[i].classification_set.ether_type_flag ||
				    p1->rules[i].classification_set.vid != p2->rules[i].classification_set.vid ||
				    p1->rules[i].classification_set.vid_flag != p2->rules[i].classification_set.vid_flag ||
				    p1->rules[i].classification_set.pcp != p2->rules[i].classification_set.pcp ||
				    p1->rules[i].classification_set.pcp_flag != p2->rules[i].classification_set.pcp_flag ||
				    p1->rules[i].addresses_nr != p2->rules[i].addresses_nr || p1->rules[i].last_matched != p2->rules[i].last_matched) {
					return 1;
				}

				if (p1->rules[i].addresses_nr > 0 && (NULL == p1->rules[i].addresses || NULL == p2->rules[i].addresses)) {
					// Malformed structure
					//
					return 1;
				}

				for (j = 0; j < p1->rules[i].addresses_nr; j++) {
					if (memcmp(p1->rules[i].addresses[j], p2->rules[i].addresses[j], 6)) {
						return 1;
					}
				}
			}

			return 0;
		}

	case ALME_TYPE_MODIFY_FWD_RULE_REQUEST:
		{
			struct modifyFwdRuleRequestALME *p1, *p2;
			INT8U i;

			p1 = (struct modifyFwdRuleRequestALME *)memory_structure_1;
			p2 = (struct modifyFwdRuleRequestALME *)memory_structure_2;

			if (p1->rule_id != p2->rule_id || p1->addresses_nr != p2->addresses_nr) {
				return 1;
			}

			if (p1->addresses_nr > 0 && (NULL == p1->addresses || NULL == p2->addresses)) {
				// Malformed structure
				//
				return 1;
			}

			for (i = 0; i < p1->addresses_nr; i++) {
				if (memcmp(p1->addresses[i], p2->addresses[i], 6)) {
					return 1;
				}
			}

			return 0;
		}

	case ALME_TYPE_MODIFY_FWD_RULE_CONFIRM:
		{
			struct modifyFwdRuleConfirmALME *p1, *p2;

			p1 = (struct modifyFwdRuleConfirmALME *)memory_structure_1;
			p2 = (struct modifyFwdRuleConfirmALME *)memory_structure_2;

			if (p1->rule_id != p2->rule_id || p1->reason_code != p2->reason_code) {
				return 1;
			}

			return 0;
		}

	case ALME_TYPE_REMOVE_FWD_RULE_REQUEST:
		{
			struct removeFwdRuleRequestALME *p1, *p2;

			p1 = (struct removeFwdRuleRequestALME *)memory_structure_1;
			p2 = (struct removeFwdRuleRequestALME *)memory_structure_2;

			if (p1->rule_id != p2->rule_id) {
				return 1;
			}

			return 0;
		}

	case ALME_TYPE_REMOVE_FWD_RULE_CONFIRM:
		{
			struct removeFwdRuleConfirmALME *p1, *p2;

			p1 = (struct removeFwdRuleConfirmALME *)memory_structure_1;
			p2 = (struct removeFwdRuleConfirmALME *)memory_structure_2;

			if (p1->rule_id != p2->rule_id || p1->reason_code != p2->reason_code) {
				return 1;
			}

			return 0;
		}

	case ALME_TYPE_GET_METRIC_REQUEST:
		{
			struct getMetricRequestALME *p1, *p2;

			p1 = (struct getMetricRequestALME *)memory_structure_1;
			p2 = (struct getMetricRequestALME *)memory_structure_2;

			if (memcmp(p1->interface_address, p2->interface_address, 6)) {
				return 1;
			}

			return 0;
		}

	case ALME_TYPE_GET_METRIC_RESPONSE:
		{
			struct getMetricResponseALME *p1, *p2;
			INT8U i;

			p1 = (struct getMetricResponseALME *)memory_structure_1;
			p2 = (struct getMetricResponseALME *)memory_structure_2;

			if (p1->metrics_nr != p2->metrics_nr ||
			    p1->reason_code != p2->reason_code)
				return 1;

			if (p1->metrics_nr > 0 &&
			    (p1->metrics == NULL ||
			     p2->metrics == NULL))
				return 1;

			for (i = 0; i < p1->metrics_nr; i++) {
				if (memcmp(p1->metrics[i].neighbor_dev_address, p2->metrics[i].neighbor_dev_address, 6) ||
				    memcmp(p1->metrics[i].local_intf_address, p2->metrics[i].local_intf_address, 6) ||
				    p1->metrics[i].bridge_flag != p2->metrics[i].bridge_flag ||
				    compare_1905_TLV_structures((INT8U *) p1->metrics[i].tx_metric, (INT8U *) p2->metrics[i].tx_metric) ||
				    compare_1905_TLV_structures((INT8U *) p1->metrics[i].rx_metric, (INT8U *) p2->metrics[i].rx_metric)

				    ) {
					return 1;
				}
			}

			return 0;
		}

	case ALME_TYPE_CUSTOM_COMMAND_REQUEST:
		{
			struct customCommandRequestALME *p1, *p2;

			p1 = (struct customCommandRequestALME *)memory_structure_1;
			p2 = (struct customCommandRequestALME *)memory_structure_2;

			if (p1->command != p2->command) {
				return 1;
			}

			return 0;
		}

	case ALME_TYPE_CUSTOM_COMMAND_RESPONSE:
		{
			struct customCommandResponseALME *p1, *p2;

			p1 = (struct customCommandResponseALME *)memory_structure_1;
			p2 = (struct customCommandResponseALME *)memory_structure_2;

			if (p1->bytes_nr != p2->bytes_nr ||
			    memcmp(p1->bytes, p2->bytes, p1->bytes_nr))
				return 1;

			return 0;
		}

	default:
		{
			// Unknown structure type
			return 1;
		}
	}

	// This code cannot be reached
	//
	return 1;
}

void visit_1905_ALME_structure(INT8U *memory_structure, PRINT_CALLBACK callback, PRINT_WRITER write_function, const char *prefix, struct blob_buf *bb)
{
	// Buffer size to store a prefix string that will be used to show each
	// element of a structure on screen
	void *array = NULL, *table = NULL, *subarray = NULL, *subtable = NULL;
#define MAX_PREFIX  100

	if (memory_structure == NULL) {
		return;
	}
	// The first byte of any of the valid structures is always the "tlv_type"
	// field.
	//
	switch (*memory_structure) {
	case ALME_TYPE_GET_INTF_LIST_REQUEST:
		{
			// There is nothing to visit. This TLV is always empty
			//
			return;
		}

	case ALME_TYPE_GET_INTF_LIST_RESPONSE:
		{
			struct getIntfListResponseALME *p;
			INT8U i, j;

			p = (struct getIntfListResponseALME *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->interface_descriptors_nr), "interface_descriptors_nr", "%d",
				 &p->interface_descriptors_nr);

			if (bb)
				array = blobmsg_open_array(bb, "interface_descriptors");
			for (i = 0; i < p->interface_descriptors_nr; i++) {
				if (bb)
					table = blobmsg_open_table(bb, NULL);
				char new_prefix[MAX_PREFIX];

				PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "%sinterface_descriptors[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(p->interface_descriptors[i].interface_address), "interface_address", "0x%02x",
					 p->interface_descriptors[i].interface_address);
				callback(bb, write_function, new_prefix, sizeof(p->interface_descriptors[i].interface_type), "media_type", "%d",
					 &p->interface_descriptors[i].interface_type);
				callback(bb, write_function, new_prefix, sizeof(p->interface_descriptors[i].bridge_flag), "bridge_flag", "%d",
					 &p->interface_descriptors[i].bridge_flag);
				callback(bb, write_function, new_prefix, sizeof(p->interface_descriptors[i].vendor_specific_info_nr), "vendor_specific_info_nr",
					 "%d", &p->interface_descriptors[i].vendor_specific_info_nr);

				if (bb)
					subarray = blobmsg_open_array(bb, "vendor_specific");
				for (j = 0; j < p->interface_descriptors[i].vendor_specific_info_nr; j++) {
					if (bb)
						subtable = blobmsg_open_table(bb, NULL);
					PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "%sinterface_descriptors[%d]->vendor_specific_info[%d]->", prefix, i, j);
					new_prefix[MAX_PREFIX - 1] = 0x0;

					callback(bb, write_function, new_prefix, sizeof(p->interface_descriptors[i].vendor_specific_info[j].ie_type), "ie_type",
						 "%d", &p->interface_descriptors[i].vendor_specific_info[j].ie_type);
					callback(bb, write_function, new_prefix, sizeof(p->interface_descriptors[i].vendor_specific_info[j].length_field),
						 "length_field", "%d", &p->interface_descriptors[i].vendor_specific_info[j].length_field);
					callback(bb, write_function, new_prefix, sizeof(p->interface_descriptors[i].vendor_specific_info[j].oui), "oui",
						 "0x%02x", p->interface_descriptors[i].vendor_specific_info[j].oui);
					callback(bb, write_function, new_prefix, p->interface_descriptors[i].vendor_specific_info[j].length_field - 3,
						 "vendor_si", "0x%02x", p->interface_descriptors[i].vendor_specific_info[j].vendor_si);
					if (subtable)
						blobmsg_close_table(bb, subtable);
				}
				if (subarray)
					blobmsg_close_array(bb, subarray);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case ALME_TYPE_SET_INTF_PWR_STATE_REQUEST:
		{
			struct setIntfPwrStateRequestALME *p;

			p = (struct setIntfPwrStateRequestALME *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->interface_address), "interface_address", "0x%02x", p->interface_address);
			callback(bb, write_function, prefix, sizeof(p->power_state), "power_state", "%d", &p->power_state);

			return;
		}

	case ALME_TYPE_SET_INTF_PWR_STATE_CONFIRM:
		{
			struct setIntfPwrStateConfirmALME *p;

			p = (struct setIntfPwrStateConfirmALME *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->interface_address), "interface_address", "0x%02x", p->interface_address);
			callback(bb, write_function, prefix, sizeof(p->reason_code), "reason_code", "%d", &p->reason_code);

			return;
		}

	case ALME_TYPE_GET_INTF_PWR_STATE_REQUEST:
		{
			struct getIntfPwrStateRequestALME *p;

			p = (struct getIntfPwrStateRequestALME *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->interface_address), "interface_address", "0x%02x", p->interface_address);

			return;
		}

	case ALME_TYPE_GET_INTF_PWR_STATE_RESPONSE:
		{
			struct getIntfPwrStateResponseALME *p;

			p = (struct getIntfPwrStateResponseALME *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->interface_address), "interface_address", "0x%02x", p->interface_address);
			callback(bb, write_function, prefix, sizeof(p->power_state), "power_state", "%d", &p->power_state);

			return;
		}

	case ALME_TYPE_SET_FWD_RULE_REQUEST:
		{
			struct setFwdRuleRequestALME *p;
			INT8U i;

			p = (struct setFwdRuleRequestALME *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->classification_set.mac_da), "classification_set.mac_da", "0x%02x",
				 p->classification_set.mac_da);
			callback(bb, write_function, prefix, sizeof(p->classification_set.mac_da_flag), "classification_set.mac_da_flag", "%d",
				 &p->classification_set.mac_da_flag);
			callback(bb, write_function, prefix, sizeof(p->classification_set.mac_sa), "classification_set.mac_sa", "0x%02x",
				 p->classification_set.mac_sa);
			callback(bb, write_function, prefix, sizeof(p->classification_set.mac_sa_flag), "classification_set.mac_sa_flag", "%d",
				 &p->classification_set.mac_sa_flag);
			callback(bb, write_function, prefix, sizeof(p->classification_set.ether_type), "classification_set.ether_type", "%d",
				 &p->classification_set.ether_type);
			callback(bb, write_function, prefix, sizeof(p->classification_set.ether_type_flag), "classification_set.ether_type_flag", "%d",
				 &p->classification_set.ether_type_flag);
			callback(bb, write_function, prefix, sizeof(p->classification_set.vid), "classification_set.vid", "%d", &p->classification_set.vid);
			callback(bb, write_function, prefix, sizeof(p->classification_set.vid_flag), "classification_set.vid_flag", "%d",
				 &p->classification_set.vid_flag);
			callback(bb, write_function, prefix, sizeof(p->classification_set.pcp), "classification_set.pcp", "%d", &p->classification_set.pcp);
			callback(bb, write_function, prefix, sizeof(p->classification_set.pcp_flag), "classification_set.pcp_flag", "%d",
				 &p->classification_set.pcp_flag);
			callback(bb, write_function, prefix, sizeof(p->addresses_nr), "addresses_nr", "%d", &p->addresses_nr);

			if (bb)
				array = blobmsg_open_array(bb, "addresses");
			for (i = 0; i < p->addresses_nr; i++) {
				if (bb)
					table = blobmsg_open_table(bb, NULL);
				char new_prefix[MAX_PREFIX];

				PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "%saddresses[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(p->addresses[i]), "", "0x%02x", p->addresses[i]);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case ALME_TYPE_SET_FWD_RULE_CONFIRM:
		{
			struct setFwdRuleConfirmALME *p;

			p = (struct setFwdRuleConfirmALME *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->rule_id), "rule_id", "%d", &p->reason_code);
			callback(bb, write_function, prefix, sizeof(p->reason_code), "reason_code", "%d", &p->reason_code);

			return;
		}

	case ALME_TYPE_GET_FWD_RULES_REQUEST:
		{
			// There is nothing to visit. This TLV is always empty
			//
			return;
		}

	case ALME_TYPE_GET_FWD_RULES_RESPONSE:
		{
			struct getFwdRulesResponseALME *p;
			INT8U i, j;

			p = (struct getFwdRulesResponseALME *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->rules_nr), "rules_nr", "%d", &p->rules_nr);

			if (bb)
				array = blobmsg_open_array(bb, "rules");
			for (i = 0; i < p->rules_nr; i++) {
				if (bb)
					blobmsg_open_table(bb, NULL);

				char new_prefix[MAX_PREFIX];

				PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "%srules[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(p->rules[i].addresses_nr), "addresses_nr", "%d", &p->rules[i].addresses_nr);

				if (bb)
					subarray = blobmsg_open_array(bb, "addresses");
				for (j = 0; j < p->rules[i].addresses_nr; j++) {
					if (bb)
						subtable = blobmsg_open_table(bb, NULL);
					PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "%srules[%d]->addresses[%d]", prefix, i, j);
					new_prefix[MAX_PREFIX - 1] = 0x0;

					callback(bb, write_function, new_prefix, sizeof(p->rules[i].addresses[j]), "", "0x%02x", p->rules[i].addresses[j]);
					if (subtable)
						blobmsg_close_table(bb, subtable);
				}
				if (subarray)
					blobmsg_close_array(bb, subarray);

				callback(bb, write_function, new_prefix, sizeof(p->rules[i].last_matched), "last_matched", "%d", &p->rules[i].last_matched);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case ALME_TYPE_MODIFY_FWD_RULE_REQUEST:
		{
			struct modifyFwdRuleRequestALME *fw_rule;
			INT8U i;

			fw_rule = (struct modifyFwdRuleRequestALME *)memory_structure;

			callback(bb, write_function, prefix, sizeof(fw_rule->rule_id), "rule_id", "%d", &fw_rule->rule_id);
			callback(bb, write_function, prefix, sizeof(fw_rule->addresses_nr), "addresses_nr", "%d", &fw_rule->addresses_nr);

			if (bb)
				array = blobmsg_open_array(bb, "addresses");
			for (i = 0; i < fw_rule->addresses_nr; i++) {
				if (bb)
					table = blobmsg_open_table(bb, NULL);
				char new_prefix[MAX_PREFIX];

				PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "%saddresses[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(fw_rule->addresses[i]), "", "0x%02x", fw_rule->addresses[i]);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case ALME_TYPE_MODIFY_FWD_RULE_CONFIRM:
		{
			struct modifyFwdRuleConfirmALME *p;

			p = (struct modifyFwdRuleConfirmALME *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->rule_id), "rule_id", "%d", &p->rule_id);
			callback(bb, write_function, prefix, sizeof(p->reason_code), "reason_code", "%d", &p->reason_code);

			return;
		}

	case ALME_TYPE_REMOVE_FWD_RULE_REQUEST:
		{
			struct removeFwdRuleRequestALME *p;

			p = (struct removeFwdRuleRequestALME *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->rule_id), "rule_id", "%d", &p->rule_id);

			return;
		}

	case ALME_TYPE_REMOVE_FWD_RULE_CONFIRM:
		{
			struct removeFwdRuleConfirmALME *p;

			p = (struct removeFwdRuleConfirmALME *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->rule_id), "rule_id", "%d", &p->rule_id);
			callback(bb, write_function, prefix, sizeof(p->reason_code), "reason_code", "%d", &p->reason_code);

			return;
		}

	case ALME_TYPE_GET_METRIC_REQUEST:
		{
			struct getMetricRequestALME *p;

			p = (struct getMetricRequestALME *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->interface_address), "rule_id", "0x%02x", p->interface_address);

			return;
		}

	case ALME_TYPE_GET_METRIC_RESPONSE:
		{
			struct getMetricResponseALME *p;
			INT8U i;

			p = (struct getMetricResponseALME *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->metrics_nr), "metrics_nr", "%d", &p->metrics_nr);

			if (bb)
				array = blobmsg_open_array(bb, "metrics");
			for (i = 0; i < p->metrics_nr; i++) {
				if (bb)
					table = blobmsg_open_table(bb, NULL);

				char new_prefix[MAX_PREFIX];

				PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "%smetrics[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(p->metrics[i].neighbor_dev_address), "neighbor_dev_address", "0x%02x",
					 p->metrics[i].neighbor_dev_address);
				callback(bb, write_function, new_prefix, sizeof(p->metrics[i].local_intf_address), "local_intf_address", "0x%02x",
					 p->metrics[i].local_intf_address);
				callback(bb, write_function, new_prefix, sizeof(p->metrics[i].bridge_flag), "bridge_flag", "%d", &p->metrics[i].bridge_flag);

				PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "%smetrics[%d]->tx_metric->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				visit_1905_TLV_structure((INT8U *) p->metrics[i].tx_metric, callback, write_function, new_prefix, bb);

				PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "%smetrics[%d]->rx_metric->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				visit_1905_TLV_structure((INT8U *) p->metrics[i].rx_metric, callback, write_function, new_prefix, bb);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case ALME_TYPE_CUSTOM_COMMAND_REQUEST:
		{
			struct customCommandRequestALME *p;

			p = (struct customCommandRequestALME *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->command), "command", "%d", &p->command);

			return;
		}

	case ALME_TYPE_CUSTOM_COMMAND_RESPONSE:
		{
			struct customCommandResponseALME *p;

			p = (struct customCommandResponseALME *)memory_structure;
			if (bb) {
				if (p->bytes) {
					json_object *jobj = json_tokener_parse(p->bytes);

					if (jobj) {
						blobmsg_add_object(bb, jobj);
					}
					json_safe_put(jobj);
				}
			} else {
				callback(bb, write_function, prefix, sizeof(p->bytes_nr), "bytes_nr", "%d", &p->bytes_nr);
				callback(bb, write_function, prefix, p->bytes_nr, "bytes", "%s", p->bytes);
			}

			return;
		}

	default:
		{
			// Ignore
			//
			return;
		}
	}

	// This code cannot be reached
	//
	return;
}

char *convert_1905_ALME_type_to_string(INT8U alme_type)
{
	switch (alme_type) {
	case ALME_TYPE_GET_INTF_LIST_REQUEST:
		return "ALME-GET-INTF-LIST.request";
	case ALME_TYPE_GET_INTF_LIST_RESPONSE:
		return "ALME-GET-INTF-LIST.response";
	case ALME_TYPE_SET_INTF_PWR_STATE_REQUEST:
		return "ALME-SET-INTF-PWR-STATE.request";
	case ALME_TYPE_SET_INTF_PWR_STATE_CONFIRM:
		return "ALME-SET-INTF-PWR-STATE.confirm";
	case ALME_TYPE_GET_INTF_PWR_STATE_REQUEST:
		return "ALME-GET-INTF-PWR-STATE.request";
	case ALME_TYPE_GET_INTF_PWR_STATE_RESPONSE:
		return "ALME-GET-INTF-PWR-STATE.response";
	case ALME_TYPE_SET_FWD_RULE_REQUEST:
		return "ALME-SET-FWD-RULE.request";
	case ALME_TYPE_SET_FWD_RULE_CONFIRM:
		return "ALME-SET-FWD-RULE.confirm";
	case ALME_TYPE_GET_FWD_RULES_REQUEST:
		return "ALME-GET-FWD-RULES.request";
	case ALME_TYPE_GET_FWD_RULES_RESPONSE:
		return "ALME-GET-FWD-RULES.response";
	case ALME_TYPE_MODIFY_FWD_RULE_REQUEST:
		return "ALME-MODIFY-FWD-RULE.request";
	case ALME_TYPE_MODIFY_FWD_RULE_CONFIRM:
		return "ALME-MODIFY-FWD-RULE.confirm";
	case ALME_TYPE_REMOVE_FWD_RULE_REQUEST:
		return "ALME-REMOVE-FWD-RULE.request";
	case ALME_TYPE_REMOVE_FWD_RULE_CONFIRM:
		return "ALME-REMOVE-FWD-RULE.confirm";
	case ALME_TYPE_GET_METRIC_REQUEST:
		return "ALME-GET-METRIC.request";
	case ALME_TYPE_GET_METRIC_RESPONSE:
		return "ALME-GET-METRIC.response";
	case ALME_TYPE_CUSTOM_COMMAND_REQUEST:
		return "ALME_TYPE_CUSTOM_COMMAND_REQUEST";
	default:
		return "Unknown";
	}
}

INT8U get_1905_ALME_type_from_string(char *type)
{
	if (type == NULL)
		return ALME_TYPE_INVALID;

	if (0 == strcmp(type, "ALME-GET-INTF-LIST.request"))
		return ALME_TYPE_GET_INTF_LIST_REQUEST;

	if (0 == strcmp(type, "ALME-GET-INTF-LIST.response"))
		return ALME_TYPE_GET_INTF_LIST_RESPONSE;

	if (0 == strcmp(type, "ALME-SET-INTF-PWR-STATE.request"))
		return ALME_TYPE_SET_INTF_PWR_STATE_REQUEST;

	if (0 == strcmp(type, "ALME-SET-INTF-PWR-STATE.confirm"))
		return ALME_TYPE_SET_INTF_PWR_STATE_CONFIRM;

	if (0 == strcmp(type, "ALME-GET-INTF-PWR-STATE.request"))
		return ALME_TYPE_GET_INTF_PWR_STATE_REQUEST;

	if (0 == strcmp(type, "ALME-GET-INTF-PWR-STATE.response"))
		return ALME_TYPE_GET_INTF_PWR_STATE_RESPONSE;

	if (0 == strcmp(type, "ALME-SET-FWD-RULE.request"))
		return ALME_TYPE_SET_FWD_RULE_REQUEST;

	if (0 == strcmp(type, "ALME-SET-FWD-RULE.confirm"))
		return ALME_TYPE_SET_FWD_RULE_CONFIRM;

	if (0 == strcmp(type, "ALME-GET-FWD-RULES.request"))
		return ALME_TYPE_GET_FWD_RULES_REQUEST;

	if (0 == strcmp(type, "ALME-GET-FWD-RULES.response"))
		return ALME_TYPE_GET_FWD_RULES_RESPONSE;

	if (0 == strcmp(type, "ALME-MODIFY-FWD-RULE.request"))
		return ALME_TYPE_MODIFY_FWD_RULE_REQUEST;

	if (0 == strcmp(type, "ALME-MODIFY-FWD-RULE.confirm"))
		return ALME_TYPE_MODIFY_FWD_RULE_CONFIRM;

	if (0 == strcmp(type, "ALME-REMOVE-FWD-RULE.request"))
		return ALME_TYPE_REMOVE_FWD_RULE_REQUEST;

	if (0 == strcmp(type, "ALME-REMOVE-FWD-RULE.confirm"))
		return ALME_TYPE_REMOVE_FWD_RULE_CONFIRM;

	if (0 == strcmp(type, "ALME-GET-METRIC.request"))
		return ALME_TYPE_GET_METRIC_REQUEST;

	if (0 == strcmp(type, "ALME-GET-METRIC.response"))
		return ALME_TYPE_GET_METRIC_RESPONSE;

	if (0 == strcmp(type, "ALME_TYPE_CUSTOM_COMMAND_REQUEST"))
		return ALME_TYPE_CUSTOM_COMMAND_REQUEST;

	return ALME_TYPE_INVALID;
}

INT8U *build_alme_request(char *alme_cmd, char *data)
{
	INT8U *ret = NULL;
	INT8U alme_request_type = ALME_TYPE_INVALID;

	if (alme_cmd == NULL)
		return NULL;

	alme_request_type = get_1905_ALME_type_from_string(alme_cmd);

	switch (alme_request_type) {
	case ALME_TYPE_GET_INTF_LIST_REQUEST:
		{
			struct getIntfListRequestALME *p;

			p = (struct getIntfListRequestALME *)malloc(sizeof(struct getIntfListRequestALME));
			if (p == NULL) {
				ERROR("Could not allocate getIntfListRequestALME structure!");
				return NULL;
			}
			p->alme_type = ALME_TYPE_GET_INTF_LIST_REQUEST;

			ret = (INT8U *) p;
			break;
		}
	case ALME_TYPE_GET_METRIC_REQUEST:
		{
			struct getMetricRequestALME *p;

			INT8U mac_address[6] = {0x00};

			// When an ALME-GET-METRIC.request is solicited, the user can either
			// ask for a specific neighbor (in that case an extra argument is
			// provided) or for all neighbors (in that case no extra argument is
			// provided).
			//
			if (data != NULL) {
				hwaddr_aton(data, mac_address);
			}

			p = (struct getMetricRequestALME *)malloc(sizeof(struct getMetricRequestALME));
			if (p == NULL) {
				ERROR("Could not allocate getMetricRequestALME structure!");
				return NULL;
			}
			p->alme_type = ALME_TYPE_GET_METRIC_REQUEST;
			memcpy(p->interface_address, mac_address, 6);

			ret = (INT8U *) p;
			break;
		}
	case ALME_TYPE_CUSTOM_COMMAND_REQUEST:
		{
			struct customCommandRequestALME *p;

			if (data == NULL) {
				// No extra argument was provided
				//
				ERROR("Invalid arguments for 'ALME-CUSTOM-COMMAND' message");
				return NULL;
			}

			p = (struct customCommandRequestALME *)malloc(sizeof(struct customCommandRequestALME));
			if (p == NULL) {
				ERROR("Could not allocate customCommandRequestALME structure!");
				return NULL;
			}
			p->alme_type = ALME_TYPE_CUSTOM_COMMAND_REQUEST;

			if (data != NULL) {
				if (0 == strcmp(data, "dnd")) {
					p->command = CUSTOM_COMMAND_DUMP_NETWORK_DEVICES;
				} else {
					ERROR("Invalid arguments for 'ALME-CUSTOM-COMMAND' message");
					free(p);
					return NULL;
				}
			}

			ret = (INT8U *) p;
			break;
		}
	default:
		{
			ERROR("ERROR: Unknown ALME message type: %d", alme_request_type);
			break;
		}
	}			// end of switch

	return ret;
}

void set_alme_blob(struct blob_buf *bb)
{
	s_alme_blob = bb;
}

bool sendALMEwrapper(INT8U *alme)
{
	bool ret = false;
#ifdef ALME_OVER_TCP
	ret = send1905RawALME(alme);
#else
	if (s_alme_blob) {
		visit_1905_ALME_structure(alme, print_callback_blob, print_null, "", s_alme_blob);
		ret = true;
	}
	// reset alme_blob
	s_alme_blob = NULL;
#endif
	return ret;
}

void setALMEClientID(INT8U client_id)
{
	gs_alme_client_id = client_id;
}

// This function is needed to present Tx and Rx TLVs in the way they are
// expected when contained inside an ALME-GET-METRIC.response reply.
//
// Let's explain this in more detail.
//
// Tx and Rx TLVs are "designed" to contain (each one of them) all possible
// links between two AL entities.
//
// In other words, if an AL has 3 neighbors, then 3 Rx (and 3 Tx) TLVs is all
// that is needed to contain all the information we might ever want.
//
// Example:
//
//   In this scenario, "AL 1" just has one neighbor ("AL 2") and thus only one
//   TLV is needed (even if there are *two* possible links)
//
//          eth0 ---------- eth3
//                    |
//     AL 1            -----eth2  AL 2
//
//          eth1 ---------- eth0
//
//   In particular, this unique TLV would be structured like this:
//
//     TLV:
//       - metrics AL1/eth0 <--> AL2/eth3
//       - metrics AL1/eth0 <--> AL2/eth2
//       - metrics AL1/eth1 <--> AL2/eth0
//
// However, when replying to an "ALME-GET-METRIC.response" message, for some
// reason, each Tx/Rx TLV in the list can only contain information for *one*
// local interface.
//
// In other words, in the example above, *even* if just one TLV has the
// capacity of providing information for both interfaces, we would have to
// create *two* TLVs instead, structured like this:
//
//   TLV1:
//     - metrics AL1/eth0 <--> AL2/eth3
//     - metrics AL1/eth0 <--> AL2/eth2
//
//   TLV2:
//     - metrics AL1/eth1 <--> AL2/eth0
//
// This is *obviously* an error in the standard (it causes more memory usage
// and repeated member structures that are not necessary)... but we have to live
// with it.
//
// Anyway... function "obtainLocalMetricsTLVs()" returns a list of pointers to
// Tx/Rx TLVs where each element contains all the links between two neighbors.
// This function ("_reStructureMetricsTLVs()") takes that output and then
// returns a bigger list where each of the TLVs now only contains information
// regarding one local interface.
//
// The "restructured" returned pointers must be freed with a call to
// "freeLocalMetricsTLVs()" when they are no longer needed
//
// Note: this function should always return "1". If it ever returns "0" it means
// there is a design error and the function implementation should be reviewed
//
static INT8U _reStructureMetricsTLVs(struct transmitterLinkMetricTLV ***tx, struct receiverLinkMetricTLV ***rx, INT8U *nr)
{
	struct transmitterLinkMetricTLV **tx_tlvs;
	struct receiverLinkMetricTLV **rx_tlvs;

	struct transmitterLinkMetricTLV **new_tx_tlvs;
	struct receiverLinkMetricTLV **new_rx_tlvs;

	INT8U total_tlvs;
	INT8U new_total_tlvs_rx;
	INT8U new_total_tlvs_tx;

	char **interfaces_names;
	INT8U interfaces_names_nr;

	interfaces_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&interfaces_names_nr);

	INT8U i, j, k;

	tx_tlvs = *tx;
	rx_tlvs = *rx;
	total_tlvs = *nr;

	new_tx_tlvs = NULL;
	new_rx_tlvs = NULL;
	new_total_tlvs_rx = 0;
	new_total_tlvs_tx = 0;

	// For each neighbor
	//
	for (i = 0; i < total_tlvs; i++) {
		// Each "old" TLV (representing a neighbor) will "expand" into as many
		// "new" TLVs as local interfaces can be used to reach that neighbor.
		//
		if (tx_tlvs != NULL) {
			// For each local interface
			//
			for (j = 0; j < interfaces_names_nr; j++) {
				// ...find all TLV metrics associated to this local interface
				//
				for (k = 0; k < tx_tlvs[i]->transmitter_link_metrics_nr; k++) {
					if (0 ==
					    memcmp(DMinterfaceNameToMac(interfaces_names[j]),
							tx_tlvs[i]->transmitter_link_metrics[k].local_interface_address, 6)) {
						// ...and add them
						//
						if (new_tx_tlvs == NULL) {
							// ...as NEW TLVs, if this is the first time
							//
							CALLOC(new_tx_tlvs, 1);
							CALLOC(new_tx_tlvs[0], 1);

							new_tx_tlvs[0]->tlv_type = tx_tlvs[i]->tlv_type;
							memcpy(new_tx_tlvs[0]->local_al_address, tx_tlvs[i]->local_al_address, 6);
							memcpy(new_tx_tlvs[0]->neighbor_al_address, tx_tlvs[i]->neighbor_al_address, 6);
							new_tx_tlvs[0]->transmitter_link_metrics_nr = 1;
							CALLOC(new_tx_tlvs[0]->transmitter_link_metrics, 1);
							memcpy(new_tx_tlvs[0]->transmitter_link_metrics, &(tx_tlvs[i]->transmitter_link_metrics[k]),
							       sizeof(struct _transmitterLinkMetricEntries));

							new_total_tlvs_tx = 1;
						} else {
							// ...or as either NEW TLVs or part of a previously
							// created TLV which is also associated to this same
							// local interface.
							//
							if (0 ==
							    memcmp(new_tx_tlvs[new_total_tlvs_tx - 1]->transmitter_link_metrics[0].local_interface_address,
									tx_tlvs[i]->transmitter_link_metrics[k].local_interface_address, 6)
							    && 0 == memcmp(new_tx_tlvs[new_total_tlvs_tx - 1]->neighbor_al_address,
										tx_tlvs[i]->neighbor_al_address, 6)
							    ) {
								// Part of a previously created one. Append the
								// metrics info.
								REALLOC(new_tx_tlvs[new_total_tlvs_tx - 1]->transmitter_link_metrics,
									(new_tx_tlvs[new_total_tlvs_tx - 1]->transmitter_link_metrics_nr + 1));
								memcpy(&new_tx_tlvs[new_total_tlvs_tx - 1]->
								       transmitter_link_metrics[new_tx_tlvs[new_total_tlvs_tx - 1]->
												transmitter_link_metrics_nr],
								       &(tx_tlvs[i]->transmitter_link_metrics[k]),
								       sizeof(struct _transmitterLinkMetricEntries));
								new_tx_tlvs[new_total_tlvs_tx - 1]->transmitter_link_metrics_nr++;
							} else {
								// New interface. Create new TLV.
								REALLOC(new_tx_tlvs, (new_total_tlvs_tx + 1));
								CALLOC(new_tx_tlvs[new_total_tlvs_tx], 1);

								new_tx_tlvs[new_total_tlvs_tx]->tlv_type = tx_tlvs[i]->tlv_type;
								memcpy(new_tx_tlvs[new_total_tlvs_tx]->local_al_address, tx_tlvs[i]->local_al_address, 6);
								memcpy(new_tx_tlvs[new_total_tlvs_tx]->neighbor_al_address, tx_tlvs[i]->neighbor_al_address, 6);
								new_tx_tlvs[new_total_tlvs_tx]->transmitter_link_metrics_nr = 1;
								CALLOC(new_tx_tlvs[new_total_tlvs_tx]->transmitter_link_metrics, 1);
								memcpy(new_tx_tlvs[new_total_tlvs_tx]->transmitter_link_metrics,
								       &tx_tlvs[i]->transmitter_link_metrics[k], sizeof(struct _transmitterLinkMetricEntries));

								new_total_tlvs_tx++;
							}
						}
					}
				}
			}
		}
		// Repeat THE SAME for the "rx_tlvs" (this is "semi" copy & paste code,
		// because there are differences in the way structures and members are
		// called)
		//
		if (rx_tlvs != NULL) {
			// For each local interface
			//
			for (j = 0; j < interfaces_names_nr; j++) {
				// ...find all TLV metrics associated to this local interface
				//
				for (k = 0; k < rx_tlvs[i]->receiver_link_metrics_nr; k++) {
					if (0 ==
					    memcmp(DMinterfaceNameToMac(interfaces_names[j]), rx_tlvs[i]->receiver_link_metrics[k].local_interface_address,
							6)) {
						// ...and add them
						//
						if (new_rx_tlvs == NULL) {
							// ...as NEW TLVs, if this is the first time
							CALLOC(new_rx_tlvs, 1);
							CALLOC(new_rx_tlvs[0], 1);

							new_rx_tlvs[0]->tlv_type = rx_tlvs[i]->tlv_type;
							memcpy(new_rx_tlvs[0]->local_al_address, rx_tlvs[i]->local_al_address, 6);
							memcpy(new_rx_tlvs[0]->neighbor_al_address, rx_tlvs[i]->neighbor_al_address, 6);
							new_rx_tlvs[0]->receiver_link_metrics_nr = 1;
							CALLOC(new_rx_tlvs[0]->receiver_link_metrics, 1);
							memcpy(new_rx_tlvs[0]->receiver_link_metrics, &rx_tlvs[i]->receiver_link_metrics[k],
							       sizeof(struct _receiverLinkMetricEntries));

							new_total_tlvs_rx = 1;
						} else {
							// ...or as either NEW TLVs or part of a previously
							// created TLV which is also associated to this same
							// local interface.
							//
							if (0 ==
							    memcmp(new_rx_tlvs[new_total_tlvs_rx - 1]->receiver_link_metrics[0].local_interface_address,
									rx_tlvs[i]->receiver_link_metrics[k].local_interface_address, 6)
							    && 0 == memcmp(new_rx_tlvs[new_total_tlvs_rx - 1]->neighbor_al_address,
										rx_tlvs[i]->neighbor_al_address, 6)
							    ) {
								// Part of a previously created one. Append the
								// metrics info.
								REALLOC(new_rx_tlvs[new_total_tlvs_rx - 1]->receiver_link_metrics,\
									(new_rx_tlvs[new_total_tlvs_rx - 1]->receiver_link_metrics_nr + 1));
								memcpy(&new_rx_tlvs[new_total_tlvs_rx - 1]->
								       receiver_link_metrics[new_rx_tlvs[new_total_tlvs_rx - 1]->receiver_link_metrics_nr],
								       (&rx_tlvs[i]->receiver_link_metrics[k]), sizeof(struct _receiverLinkMetricEntries));
								new_rx_tlvs[new_total_tlvs_rx - 1]->receiver_link_metrics_nr++;
							} else {
								// New interface. Create new TLV.
								REALLOC(new_rx_tlvs, (new_total_tlvs_rx + 1));
								CALLOC(new_rx_tlvs[new_total_tlvs_rx], 1);

								new_rx_tlvs[new_total_tlvs_rx]->tlv_type = rx_tlvs[i]->tlv_type;
								memcpy(new_rx_tlvs[new_total_tlvs_rx]->local_al_address, rx_tlvs[i]->local_al_address, 6);
								memcpy(new_rx_tlvs[new_total_tlvs_rx]->neighbor_al_address, rx_tlvs[i]->neighbor_al_address, 6);
								new_rx_tlvs[new_total_tlvs_rx]->receiver_link_metrics_nr = 1;
								CALLOC(new_rx_tlvs[new_total_tlvs_rx]->receiver_link_metrics, 1);
								memcpy(new_rx_tlvs[new_total_tlvs_rx]->receiver_link_metrics,
								       &rx_tlvs[i]->receiver_link_metrics[k], sizeof(struct _receiverLinkMetricEntries));

								new_total_tlvs_rx++;
							}
						}
					}
				}
			}
		}
	}

	PLATFORM_FREE_LIST_OF_1905_INTERFACES(interfaces_names, interfaces_names_nr);

	// Free the old structures
	//
	freeLocalMetricsTLVs(&tx_tlvs, &rx_tlvs, &total_tlvs);

	if (new_total_tlvs_rx != new_total_tlvs_tx) {
		// Something went terribly wrong. This should NEVER happen
		//
		ERROR("_reStructureMetricsTLVs contains a design error. Review it!");
		return 0;
	}
	// And return the new ones
	//
	*tx = new_tx_tlvs;
	*rx = new_rx_tlvs;
	*nr = new_total_tlvs_tx;

	return 1;
}

INT8U send1905InterfaceListResponseALME(void)
{
	INT8U ret;

	struct getIntfListResponseALME *out;

	char **ifs_names;
	INT8U ifs_nr;

	INFO("--> ALME_TYPE_GET_INTF_LIST_RESPONSE");

	// Fill the requested ALME response
	CALLOC(out, 1);
	out->alme_type = ALME_TYPE_GET_INTF_LIST_RESPONSE;

	ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);
	if (ifs_nr == 0) {
		out->interface_descriptors_nr = 0;
		out->interface_descriptors = NULL;
	} else {
		INT8U i;

		out->interface_descriptors_nr = ifs_nr;
		CALLOC(out->interface_descriptors, ifs_nr);

		for (i = 0; i < out->interface_descriptors_nr; i++) {
			struct interfaceInfo *x;

			x = PLATFORM_GET_1905_INTERFACE_INFO(ifs_names[i]);
			if (x == NULL) {
				WARNING("Could not retrieve info of interface %s", ifs_names[i]);

				out->interface_descriptors[i].interface_address[0] = 0x00;
				out->interface_descriptors[i].interface_address[1] = 0x00;
				out->interface_descriptors[i].interface_address[2] = 0x00;
				out->interface_descriptors[i].interface_address[3] = 0x00;
				out->interface_descriptors[i].interface_address[4] = 0x00;
				out->interface_descriptors[i].interface_address[5] = 0x00;
				out->interface_descriptors[i].interface_type = MEDIA_TYPE_UNKNOWN;
				out->interface_descriptors[i].bridge_flag = 0;
				out->interface_descriptors[i].vendor_specific_info_nr = 0;
				out->interface_descriptors[i].vendor_specific_info = NULL;
			} else {
				out->interface_descriptors[i].interface_address[0] = x->mac_address[0];
				out->interface_descriptors[i].interface_address[1] = x->mac_address[1];
				out->interface_descriptors[i].interface_address[2] = x->mac_address[2];
				out->interface_descriptors[i].interface_address[3] = x->mac_address[3];
				out->interface_descriptors[i].interface_address[4] = x->mac_address[4];
				out->interface_descriptors[i].interface_address[5] = x->mac_address[5];
				out->interface_descriptors[i].interface_type = x->interface_type;

				out->interface_descriptors[i].bridge_flag = DMisInterfaceBridged(ifs_names[i]);

				if (x->vendor_specific_elements_nr == 0) {
					out->interface_descriptors[i].vendor_specific_info_nr = 0;
					out->interface_descriptors[i].vendor_specific_info = NULL;
				} else {
					INT8U j;

					out->interface_descriptors[i].vendor_specific_info_nr = x->vendor_specific_elements_nr;
					CALLOC(out->interface_descriptors[i].vendor_specific_info, \
					       x->vendor_specific_elements_nr);

					for (j = 0; j < out->interface_descriptors[i].vendor_specific_info_nr; j++) {
						out->interface_descriptors[i].vendor_specific_info[j].ie_type = 1;
						out->interface_descriptors[i].vendor_specific_info[j].length_field =
						    x->vendor_specific_elements[j].vendor_data_len + 3;
						out->interface_descriptors[i].vendor_specific_info[j].oui[0] = x->vendor_specific_elements[j].oui[0];
						out->interface_descriptors[i].vendor_specific_info[j].oui[1] = x->vendor_specific_elements[j].oui[1];
						out->interface_descriptors[i].vendor_specific_info[j].oui[2] = x->vendor_specific_elements[j].oui[2];

						CALLOC(out->interface_descriptors[i].vendor_specific_info[j].vendor_si, \
						       x->vendor_specific_elements[j].vendor_data_len);
						memcpy(out->interface_descriptors[i].vendor_specific_info[j].vendor_si,
						       x->vendor_specific_elements[j].vendor_data, x->vendor_specific_elements[j].vendor_data_len);
					}
				}
			}

			if (x != NULL) {
				PLATFORM_FREE_1905_INTERFACE_INFO(x);
			}
		}
	}

	PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);

	// Send the packet
	if (sendALMEwrapper((INT8U *) out) == 0) {
		ERROR("Could not send the 1905 ALME reply");
		ret = 1;
	} else {
		ret = 0;
	}

	// Free memory
	//
	free_1905_ALME_structure((INT8U *) out);

	return ret;
}

INT8U send1905MetricsResponseALME(INT8U *mac_address)
{
	INT8U ret;

	struct getMetricResponseALME *out;

	struct transmitterLinkMetricTLV **tx_tlvs;
	struct receiverLinkMetricTLV **rx_tlvs;

	INT8U total_tlvs;
	INT8U res;

	INT8U i;

	INFO("--> ALME_TYPE_GET_METRIC_RESPONSE");

	// Fill the requested ALME response
	//
	if (mac_address == NULL) {
		obtainLocalMetricsTLVs(LINK_METRIC_QUERY_TLV_ALL_NEIGHBORS, NULL,
				       LINK_METRIC_QUERY_TLV_BOTH_TX_AND_RX_LINK_METRICS, &tx_tlvs, &rx_tlvs, &total_tlvs);
	} else {
		obtainLocalMetricsTLVs(LINK_METRIC_QUERY_TLV_SPECIFIC_NEIGHBOR, mac_address,
				       LINK_METRIC_QUERY_TLV_BOTH_TX_AND_RX_LINK_METRICS, &tx_tlvs, &rx_tlvs, &total_tlvs);
	}

	// Reorder Tx/Rx TLVs in the way they are expected inside an ALME metrics
	// response (which is different from what you have in a "regular" TLV for
	// some strange reason, maybe a "bug" in the standard)
	//
	res = _reStructureMetricsTLVs(&tx_tlvs, &rx_tlvs, &total_tlvs);

	CALLOC(out, 1);
	out->alme_type = ALME_TYPE_GET_METRIC_RESPONSE;

	if (0 == total_tlvs || 0 == res) {
		out->reason_code = REASON_CODE_UNMATCHED_NEIGHBOR_MAC_ADDRESS;
		out->metrics_nr = 0;
		out->metrics = NULL;
	} else {
		out->reason_code = REASON_CODE_SUCCESS;
		out->metrics_nr = total_tlvs;

		if (total_tlvs > 0) {
			CALLOC(out->metrics, total_tlvs);
		} else {
			out->metrics = NULL;
		}

		for (i = 0; i < total_tlvs; i++) {
			memcpy(out->metrics[i].neighbor_dev_address, tx_tlvs[i]->neighbor_al_address, 6);
			memcpy(out->metrics[i].local_intf_address, tx_tlvs[i]->transmitter_link_metrics[0].local_interface_address, 6);
			out->metrics[i].bridge_flag =
			    DMisLinkBridged(out->metrics[i].local_intf_address, out->metrics[i].neighbor_dev_address,
					    tx_tlvs[i]->transmitter_link_metrics[0].neighbor_interface_address);
			out->metrics[i].tx_metric = tx_tlvs[i];
			out->metrics[i].rx_metric = rx_tlvs[i];
		}
	}

	// Send the packet
	//
	if (sendALMEwrapper((INT8U *) out) == 0) {
		ERROR("Could not send the 1905 ALME reply");
		ret = 1;
	} else {
		ret = 0;
	}

	// Before free'ing the "out" structure, set Tx and Rx pointer to NULL so
	// that "free_1905_ALME_structure()" ignores them.
	// They will be freed later, in "freeLocalMetricsTLVs()"
	//
	out->metrics_nr = 0;
	for (i = 0; i < total_tlvs; i++) {
		out->metrics[i].tx_metric = NULL;
		out->metrics[i].rx_metric = NULL;
	}
	FREE(out->metrics);
	out->metrics = NULL;
	free_1905_ALME_structure((INT8U *) out);

	freeLocalMetricsTLVs(&tx_tlvs, &rx_tlvs, &total_tlvs);

	return ret;
}

INT8U send1905CustomCommandResponseALME(INT8U command)
{
	INT8U ret;

	struct customCommandResponseALME *out;

	INFO("--> ALME_TYPE_CUSTOM_COMMAND_RESPONSE");

	// Fill the requested ALME response
	CALLOC(out, 1);
	out->alme_type = ALME_TYPE_CUSTOM_COMMAND_RESPONSE;

	switch (command) {
	case CUSTOM_COMMAND_DUMP_NETWORK_DEVICES:
		{
			// Update the information regarding the local node
			//
			updateLocalDeviceData();

			// Dump the database (which contains information from the local and
			// remote nodes) into a text buffer and send that as a reponse
			//

			struct blob_buf bb;

			memset(&bb, 0, sizeof(bb));
			blob_buf_init(&bb, 0);

			DMdumpNetworkDevices(print_callback_blob, print_null, &bb);

			out->bytes = blobmsg_format_json(bb.head, true);
			out->bytes_nr = strlen(out->bytes);

			blob_buf_free(&bb);
			break;
		}
	}

	// Send the packet
	//
	if (sendALMEwrapper((INT8U *) out) == 0) {
		ERROR("Could not send the 1905 ALME reply");
		ret = 1;
	} else {
		ret = 0;
	}

	// Free memory not needed anymore
	//
	if (command == CUSTOM_COMMAND_DUMP_NETWORK_DEVICES) {
		// Here we will free the memory buffer *and* set the pointer in the
		// "out" structure to NULL, so that later "free_1905_ALME_structure()"
		// doesn't try to free it again
		//
		out->bytes_nr = 0;
		out->bytes = NULL;
	}
	free_1905_ALME_structure((INT8U *) out);

	return ret;
}

#ifdef ALME_OVER_TCP
INT8U send1905RawALME(INT8U *alme)
{
	INT8U *packet_out;
	INT16U packet_out_len;

	DEBUG("Contents of ALME reply to send:");
	visit_1905_ALME_structure((INT8U *) alme, print_callback, print_debug, "", NULL);

	// Use the getIntfListResponseALME structure to forge the packet
	// bit stream
	//
	packet_out = forge_1905_ALME_from_structure((INT8U *) alme, &packet_out_len);
	if (packet_out == NULL) {
		WARNING("forge_1905_ALME_from_structure() failed.");

		PLATFORM_SEND_ALME_REPLY(gs_alme_client_id, NULL, 0);
		return 0;
	}
	// Send the ALME reply back
	//
	PLATFORM_SEND_ALME_REPLY(gs_alme_client_id, packet_out, packet_out_len);

	free_1905_ALME_packet(packet_out);

	return 1;
}
#endif
INT8U process1905Alme(INT8U *alme_tlv)
{
	if (alme_tlv == NULL) {
		return 0;
	}
	// The first byte of the 'alme_tlv' structure always contains its type
	//
	switch (*alme_tlv) {
	case ALME_TYPE_GET_INTF_LIST_REQUEST:
		{
			// Obtain the list of local interfaces, retrieve detailed info for
			// each of them, build a response, and send it back.
			//
			INFO("<-- ALME_TYPE_GET_INTF_LIST_REQUEST");

			send1905InterfaceListResponseALME();

			break;
		}
	case ALME_TYPE_SET_INTF_PWR_STATE_REQUEST:
		{
			INFO("<-- ALME_TYPE_SET_INTF_PWR_STATE_REQUEST");
			break;
		}
	case ALME_TYPE_GET_INTF_PWR_STATE_REQUEST:
		{
			INFO("<-- ALME_TYPE_GET_INTF_PWR_STATE_REQUEST");
			break;
		}
	case ALME_TYPE_SET_FWD_RULE_REQUEST:
		{
			INFO("<-- ALME_TYPE_SET_FWD_RULE_REQUEST");
			break;
		}
	case ALME_TYPE_GET_FWD_RULES_REQUEST:
		{
			INFO("<-- ALME_TYPE_GET_FWD_RULES_REQUEST");
			break;
		}
	case ALME_TYPE_MODIFY_FWD_RULE_REQUEST:
		{
			INFO("<-- ALME_TYPE_MODIFY_FWD_RULE_REQUEST");
			break;
		}
	case ALME_TYPE_REMOVE_FWD_RULE_REQUEST:
		{
			INFO("<-- ALME_TYPE_MODIFY_FWD_RULE_CONFIRM");
			break;
		}
	case ALME_TYPE_GET_METRIC_REQUEST:
		{
			// Obtain the requested metrics, build a response, and send it back.
			//
			struct getMetricRequestALME *p;

			INFO("<-- ALME_TYPE_GET_METRIC_REQUEST");

			p = (struct getMetricRequestALME *)alme_tlv;

			if (hwaddr_is_zero(p->interface_address)) {
				// Request metrics against all neighbors
				send1905MetricsResponseALME(NULL);
			} else {
				// Request metrics against one specific neighbor
				send1905MetricsResponseALME(p->interface_address);
			}

			break;
		}
	case ALME_TYPE_CUSTOM_COMMAND_REQUEST:
		{
			struct customCommandRequestALME *p;

			INFO("<-- ALME_TYPE_CUSTOM_COMMAND_REQUEST");

			p = (struct customCommandRequestALME *)alme_tlv;

			send1905CustomCommandResponseALME(p->command);

			break;
		}
	case ALME_TYPE_GET_INTF_LIST_RESPONSE:
	case ALME_TYPE_SET_INTF_PWR_STATE_CONFIRM:
	case ALME_TYPE_GET_INTF_PWR_STATE_RESPONSE:
	case ALME_TYPE_SET_FWD_RULE_CONFIRM:
	case ALME_TYPE_GET_FWD_RULES_RESPONSE:
	case ALME_TYPE_MODIFY_FWD_RULE_CONFIRM:
	case ALME_TYPE_REMOVE_FWD_RULE_CONFIRM:
	case ALME_TYPE_GET_METRIC_RESPONSE:
	case ALME_TYPE_CUSTOM_COMMAND_RESPONSE:
		{
			// These messages should never be receiving by an AL entity. It is
			// the AL entity the one who generates them and then sends them to
			// the HLE.
			//
			WARNING("ALME RESPONSE/CONFIRM message received (type = %d). Ignoring...", *alme_tlv);
			break;
		}

	default:
		{
			WARNING("Unknown ALME message received (type = %d). Ignoring...", *alme_tlv);
			break;
		}
	}

	return 1;
}
