/*
 * get_tlvs.c: Get TLV related functions
 *
 * Author: Vivek Dutta <vivek.dutta@iopsys.eu>
 *
 * Copyright (C) 2020, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "get_tlvs.h"
#include "utils.h"
#include "platform_os.h"
#include "platform_interfaces.h"
#include "al_datamodel.h"

// End of message TLV associated structures ("Section 6.4.1")
bool get_endOfMessageTLV(struct endOfMessageTLV *tlv)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct endOfMessageTLV));
	tlv->tlv_type = TLV_TYPE_END_OF_MESSAGE;

	return true;
}

// Vendor specific TLV associated structures ("Section 6.4.2")
bool get_vendorSpecificTLV(struct vendorSpecificTLV *tlv)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct vendorSpecificTLV));

	tlv->tlv_type = TLV_TYPE_VENDOR_SPECIFIC;

	WARNING("dummy values");
	return true;
}

// AL MAC address type TLV associated structures ("Section 6.4.3")
bool get_alMacAddressTypeTLV(struct alMacAddressTypeTLV *tlv)
{
	INT8U al_mac_address[6] = {0};

	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct alMacAddressTypeTLV));

	memcpy(al_mac_address, DMalMacGet(), 6);

	tlv->tlv_type = TLV_TYPE_AL_MAC_ADDRESS_TYPE;
	tlv->al_mac_address[0] = al_mac_address[0];
	tlv->al_mac_address[1] = al_mac_address[1];
	tlv->al_mac_address[2] = al_mac_address[2];
	tlv->al_mac_address[3] = al_mac_address[3];
	tlv->al_mac_address[4] = al_mac_address[4];
	tlv->al_mac_address[5] = al_mac_address[5];

	return true;
}

// MAC address type TLV associated structures ("Section 6.4.4")
bool get_macAddressTypeTLV(struct macAddressTypeTLV *tlv, char *interface_name)
{
	if (tlv == NULL)
		return false;

	if (interface_name == NULL)
		return false;

	memset(tlv, 0, sizeof(struct macAddressTypeTLV));
	tlv->tlv_type = TLV_TYPE_MAC_ADDRESS_TYPE;
	memcpy(&tlv->mac_address, DMinterfaceNameToMac(interface_name), 6);

	return true;
}

// Device information type TLV associated structures ("Section 6.4.5")
bool get_deviceInformationTypeTLV(struct deviceInformationTypeTLV *tlv)
{
	INT8U al_mac_address[6] = {0};

	char **interfaces_names;
	INT8U interfaces_names_nr;
	INT8U i;

	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct deviceInformationTypeTLV));
	tlv->tlv_type = TLV_TYPE_DEVICE_INFORMATION_TYPE;

	memcpy(al_mac_address, DMalMacGet(), 6);

	tlv->al_mac_address[0] = al_mac_address[0];
	tlv->al_mac_address[1] = al_mac_address[1];
	tlv->al_mac_address[2] = al_mac_address[2];
	tlv->al_mac_address[3] = al_mac_address[3];
	tlv->al_mac_address[4] = al_mac_address[4];
	tlv->al_mac_address[5] = al_mac_address[5];
	tlv->local_interfaces_nr = 0;
	tlv->local_interfaces = NULL;

	interfaces_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&interfaces_names_nr);

	for (i = 0; i < interfaces_names_nr; i++) {
		DEBUG("interface name: %s\n", interfaces_names[i]);
	}

	if (interfaces_names_nr == 1 && if_isbridge(interfaces_names[0])) {
		char ifnames[32][16] = {0};
		int n = 32;
                int ret;

		ret = br_get_iflist(interfaces_names[0], &n, ifnames);
		FREE(interfaces_names);

		CALLOC(interfaces_names, n);
		interfaces_names_nr = n;

		for (i = 0; i < n; ++i) {
			interfaces_names[i] = strdup(ifnames[i]);
		}
	}

	for (i = 0; i < interfaces_names_nr; i++) {
		DEBUG("interface name: %s\n", interfaces_names[i]);
	}

	// Add all interfaces that are *not* in "POWER OFF" mode
	//
	for (i = 0; i < interfaces_names_nr; i++) {
		struct interfaceInfo *x;

		x = PLATFORM_GET_1905_INTERFACE_INFO(interfaces_names[i]);
		if (x == NULL) {
			// Error retrieving information for this interface.
			// Ignore it.
			continue;
		}

		//if (x->power_state == INTERFACE_POWER_STATE_OFF) {
		//	fprintf(stderr, "CONTINUING LOOP DUE TO STATE OFF\n");
		//	// Ignore interfaces that are in "POWER OFF" mode (they will
		//	// be included in the "power off" TLV, later, on this same
		//	// CMDU)
		//	//
		//	PLATFORM_FREE_1905_INTERFACE_INFO(x);
		//	continue;
		//}

		if (tlv->local_interfaces_nr == 0) {
			CALLOC(tlv->local_interfaces, 1);
		} else {
			REALLOC(tlv->local_interfaces, (tlv->local_interfaces_nr + 1));
		}

		tlv->local_interfaces[tlv->local_interfaces_nr].mac_address[0] = x->mac_address[0];
		tlv->local_interfaces[tlv->local_interfaces_nr].mac_address[1] = x->mac_address[1];
		tlv->local_interfaces[tlv->local_interfaces_nr].mac_address[2] = x->mac_address[2];
		tlv->local_interfaces[tlv->local_interfaces_nr].mac_address[3] = x->mac_address[3];
		tlv->local_interfaces[tlv->local_interfaces_nr].mac_address[4] = x->mac_address[4];
		tlv->local_interfaces[tlv->local_interfaces_nr].mac_address[5] = x->mac_address[5];
		tlv->local_interfaces[tlv->local_interfaces_nr].media_type = x->interface_type;
		switch (x->interface_type) {
		case MEDIA_TYPE_IEEE_802_11B_2_4_GHZ:
		case MEDIA_TYPE_IEEE_802_11G_2_4_GHZ:
		case MEDIA_TYPE_IEEE_802_11A_5_GHZ:
		case MEDIA_TYPE_IEEE_802_11N_2_4_GHZ:
		case MEDIA_TYPE_IEEE_802_11N_5_GHZ:
		case MEDIA_TYPE_IEEE_802_11AC_5_GHZ:
		case MEDIA_TYPE_IEEE_802_11AD_60_GHZ:
		case MEDIA_TYPE_IEEE_802_11AF_GHZ:
			{
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data_size = 10;
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.ieee80211.network_membership[0] =
				    x->interface_type_data.ieee80211.bssid[0];
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.ieee80211.network_membership[1] =
				    x->interface_type_data.ieee80211.bssid[1];
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.ieee80211.network_membership[2] =
				    x->interface_type_data.ieee80211.bssid[2];
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.ieee80211.network_membership[3] =
				    x->interface_type_data.ieee80211.bssid[3];
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.ieee80211.network_membership[4] =
				    x->interface_type_data.ieee80211.bssid[4];
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.ieee80211.network_membership[5] =
				    x->interface_type_data.ieee80211.bssid[5];
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.ieee80211.role =
				    x->interface_type_data.ieee80211.role;
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.ieee80211.ap_channel_band =
				    x->interface_type_data.ieee80211.ap_channel_band;
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.ieee80211.
				    ap_channel_center_frequency_index_1 = x->interface_type_data.ieee80211.ap_channel_center_frequency_index_1;
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.ieee80211.
				    ap_channel_center_frequency_index_2 = x->interface_type_data.ieee80211.ap_channel_center_frequency_index_2;
				break;
			}
		case MEDIA_TYPE_IEEE_1901_FFT:
			{
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data_size = 7;
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.ieee1901.network_identifier[0] =
				    x->interface_type_data.ieee1901.network_identifier[0];
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.ieee1901.network_identifier[1] =
				    x->interface_type_data.ieee1901.network_identifier[1];
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.ieee1901.network_identifier[2] =
				    x->interface_type_data.ieee1901.network_identifier[2];
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.ieee1901.network_identifier[3] =
				    x->interface_type_data.ieee1901.network_identifier[3];
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.ieee1901.network_identifier[4] =
				    x->interface_type_data.ieee1901.network_identifier[4];
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.ieee1901.network_identifier[5] =
				    x->interface_type_data.ieee1901.network_identifier[5];
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.ieee1901.network_identifier[6] =
				    x->interface_type_data.ieee1901.network_identifier[6];
				break;
			}
		default:
			{
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data_size = 0;
				tlv->local_interfaces[tlv->local_interfaces_nr].media_specific_data.dummy = 0;
				break;
			}
		}
		tlv->local_interfaces_nr++;

		PLATFORM_FREE_1905_INTERFACE_INFO(x);
	}

	PLATFORM_FREE_LIST_OF_1905_INTERFACES(interfaces_names, interfaces_names_nr);
	return true;
}

// Device bridging capability TLV associated structures ("Section 6.4.6")
bool get_deviceBridgingCapabilityTLV(struct deviceBridgingCapabilityTLV *tlv)
{
	struct bridge *br = NULL;
	INT8U br_nr;
	INT8U i, j;
	bool ret = true;

	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct deviceBridgingCapabilityTLV));
	tlv->tlv_type = TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES;

	br = PLATFORM_GET_LIST_OF_BRIDGES(&br_nr);
	if ((br == NULL) || 0 == br_nr) {
		// No bridge info
		tlv->bridging_tuples_nr = 0;
		tlv->bridging_tuples = NULL;
#ifdef SEND_EMPTY_TLVS
		ret = true;
#else
		ret = false;
#endif
	} else {
		tlv->bridging_tuples_nr = br_nr;
		CALLOC(tlv->bridging_tuples, br_nr);

		for (i = 0; i < br_nr; i++) {
			tlv->bridging_tuples[i].bridging_tuple_macs_nr = br[i].bridged_interfaces_nr;

			if (br[i].bridged_interfaces_nr == 0) {
				tlv->bridging_tuples[i].bridging_tuple_macs = NULL;
			} else {
				CALLOC(tlv->bridging_tuples[i].bridging_tuple_macs, br[i].bridged_interfaces_nr);

				for (j = 0; j < tlv->bridging_tuples[i].bridging_tuple_macs_nr; j++) {
					INT8U mac_address[6];

					memcpy(mac_address, DMinterfaceNameToMac(br[i].bridged_interfaces[j]), 6);
					tlv->bridging_tuples[i].bridging_tuple_macs[j].mac_address[0] = mac_address[0];
					tlv->bridging_tuples[i].bridging_tuple_macs[j].mac_address[1] = mac_address[1];
					tlv->bridging_tuples[i].bridging_tuple_macs[j].mac_address[2] = mac_address[2];
					tlv->bridging_tuples[i].bridging_tuple_macs[j].mac_address[3] = mac_address[3];
					tlv->bridging_tuples[i].bridging_tuple_macs[j].mac_address[4] = mac_address[4];
					tlv->bridging_tuples[i].bridging_tuple_macs[j].mac_address[5] = mac_address[5];
				}
			}
		}
		PLATFORM_FREE_LIST_OF_BRIDGES(br, br_nr);
	}
	return ret;
}

// Non-1905 neighbor device list TLV associated structures ("Section 6.4.8")
bool get_non1905NeighborDeviceListTLV(struct non1905NeighborDeviceListTLV *tlv, char *interface_name)
{
	struct interfaceInfo *x = NULL;
	INT8U j;

	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct non1905NeighborDeviceListTLV));
	tlv->tlv_type = TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST;

	x = PLATFORM_GET_1905_INTERFACE_INFO(interface_name);
	if (x == NULL) {
		WARNING("Could not retrieve neighbors of interface %s", interface_name);
		return false;
	}

	memcpy(&tlv->local_mac_address, &x->mac_address, 6);
	tlv->non_1905_neighbors_nr = 0;
	tlv->non_1905_neighbors = NULL;

	// Decide if each neighbor is a 1905 or a non-1905 neighbor
	if (x->neighbor_mac_addresses_nr != INTERFACE_NEIGHBORS_UNKNOWN) {

		for (j = 0; j < x->neighbor_mac_addresses_nr; j++) {
			INT8U *al_mac;
			INT8U k;

			al_mac = DMmacToAlMac(x->neighbor_mac_addresses[j]);

			if (al_mac == NULL) {
				// Non-1905 neighbor
				INT8U already_added;

				// Make sure it has not already been added
				already_added = 0;
				for (k = 0; k < tlv->non_1905_neighbors_nr; k++) {
					if (memcmp(x->neighbor_mac_addresses[j],
						   tlv->non_1905_neighbors[k].mac_address, 6) == 0) {
						already_added = 1;
						break;
					}
				}

				if (already_added == 0) {
					// This is a new neighbor
					if (tlv->non_1905_neighbors_nr == 0)
						CALLOC(tlv->non_1905_neighbors, 1);
					else
						REALLOC(tlv->non_1905_neighbors, (tlv->non_1905_neighbors_nr + 1));

					tlv->non_1905_neighbors[tlv->non_1905_neighbors_nr].mac_address[0] =
						x->neighbor_mac_addresses[j][0];
					tlv->non_1905_neighbors[tlv->non_1905_neighbors_nr].mac_address[1] =
						x->neighbor_mac_addresses[j][1];
					tlv->non_1905_neighbors[tlv->non_1905_neighbors_nr].mac_address[2] =
						x->neighbor_mac_addresses[j][2];
					tlv->non_1905_neighbors[tlv->non_1905_neighbors_nr].mac_address[3] =
						x->neighbor_mac_addresses[j][3];
					tlv->non_1905_neighbors[tlv->non_1905_neighbors_nr].mac_address[4] =
						x->neighbor_mac_addresses[j][4];
					tlv->non_1905_neighbors[tlv->non_1905_neighbors_nr].mac_address[5] =
						x->neighbor_mac_addresses[j][5];

					tlv->non_1905_neighbors_nr++;
				}
			}
			FREE(al_mac);
		}
		PLATFORM_FREE_1905_INTERFACE_INFO(x);

	}
	return true;
}

// Neighbor device TLV associated structures ("Section 6.4.9")
bool get_neighborDeviceListTLV(struct neighborDeviceListTLV *tlv, char *interface_name)
{
	struct interfaceInfo *x = NULL;
	INT8U j, k;

	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct neighborDeviceListTLV));
	tlv->tlv_type = TLV_TYPE_NEIGHBOR_DEVICE_LIST;

	x = PLATFORM_GET_1905_INTERFACE_INFO(interface_name);
	if (x == NULL) {
		WARNING("Could not retrieve neighbors of interface %s", interface_name);
		return false;
	}

	memcpy(&tlv->local_mac_address, &x->mac_address, 6);
	tlv->neighbors_nr = 0;
	tlv->neighbors = NULL;

	// Decide if each neighbor is a 1905 or a non-1905 neighbor
	if (x->neighbor_mac_addresses_nr != INTERFACE_NEIGHBORS_UNKNOWN) {

		for (j = 0; j < x->neighbor_mac_addresses_nr; j++) {
			INT8U *al_mac;

			al_mac = DMmacToAlMac(x->neighbor_mac_addresses[j]);

			if (al_mac != NULL) {
				// 1905 neighbor

				INT8U already_added;

				// Make sure it has not already been added
				already_added = 0;
				for (k = 0; k < tlv->neighbors_nr; k++) {
					if (memcmp(al_mac, tlv->neighbors[k].mac_address, 6) == 0) {
						already_added = 1;
						break;
					}
				}

				if (already_added == 0) {
					// This is a new neighbor
					if (tlv->neighbors_nr == 0)
						CALLOC(tlv->neighbors, 1);
					else
						REALLOC(tlv->neighbors, (tlv->neighbors_nr + 1));

					tlv->neighbors[tlv->neighbors_nr].mac_address[0] = al_mac[0];
					tlv->neighbors[tlv->neighbors_nr].mac_address[1] = al_mac[1];
					tlv->neighbors[tlv->neighbors_nr].mac_address[2] = al_mac[2];
					tlv->neighbors[tlv->neighbors_nr].mac_address[3] = al_mac[3];
					tlv->neighbors[tlv->neighbors_nr].mac_address[4] = al_mac[4];
					tlv->neighbors[tlv->neighbors_nr].mac_address[5] = al_mac[5];
					tlv->neighbors[tlv->neighbors_nr].bridge_flag = DMisNeighborBridged(interface_name, al_mac);

					tlv->neighbors_nr++;
				}

				FREE(al_mac);
			}
		}
		PLATFORM_FREE_1905_INTERFACE_INFO(x);

	} else {
		// The interface reports that it has no way of knowing which MAC
		// neighbors are connected to it.
		// In these cases, *at least* the already known 1905 neighbors
		// (which were discovered by us -not the platform- thanks to the
		// topology discovery process) should be returned.
		INT8U(*al_mac_addresses)[6];
		INT8U al_mac_addresses_nr = 0;

		al_mac_addresses = DMgetListOfNeighbors(&al_mac_addresses_nr);

		for (j = 0; j < al_mac_addresses_nr; j++) {
			INT8U already_added;

			// Make sure it has not already been added
			already_added = 0;
			for (k = 0; k < tlv->neighbors_nr; k++) {
				if (memcmp(al_mac_addresses[j], tlv->neighbors[k].mac_address, 6) == 0) {
					already_added = 1;
					break;
				}
			}

			if (already_added == 0) {
				// This is a new neighbor
				if (tlv->neighbors_nr == 0)
					CALLOC(tlv->neighbors, 1);
				else
					REALLOC(tlv->neighbors, (tlv->neighbors_nr + 1));

				tlv->neighbors[tlv->neighbors_nr].mac_address[0] = al_mac_addresses[j][0];
				tlv->neighbors[tlv->neighbors_nr].mac_address[1] = al_mac_addresses[j][1];
				tlv->neighbors[tlv->neighbors_nr].mac_address[2] = al_mac_addresses[j][2];
				tlv->neighbors[tlv->neighbors_nr].mac_address[3] = al_mac_addresses[j][3];
				tlv->neighbors[tlv->neighbors_nr].mac_address[4] = al_mac_addresses[j][4];
				tlv->neighbors[tlv->neighbors_nr].mac_address[5] = al_mac_addresses[j][5];
				tlv->neighbors[tlv->neighbors_nr].bridge_flag = DMisNeighborBridged(interface_name, al_mac_addresses[j]);

				tlv->neighbors_nr++;

			}
		}
		FREE(al_mac_addresses);
	}
	return true;
}

// Link metric query TLV associated structures ("Section 6.4.10")
bool get_linkMetricQueryTLV(struct linkMetricQueryTLV *tlv)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct linkMetricQueryTLV));
	tlv->tlv_type = TLV_TYPE_LINK_METRIC_QUERY;
	tlv->destination = LINK_METRIC_QUERY_TLV_ALL_NEIGHBORS;
	memset(&tlv->specific_neighbor, 0, 6);
	tlv->link_metrics_type = LINK_METRIC_QUERY_TLV_BOTH_TX_AND_RX_LINK_METRICS;

	return true;
}

bool get_secific_linkMetricQueryTLV(struct linkMetricQueryTLV *tlv, INT8U *specific_neighbor, INT8U metrix_type)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct linkMetricQueryTLV));
	tlv->tlv_type = TLV_TYPE_LINK_METRIC_QUERY;

	tlv->destination = LINK_METRIC_QUERY_TLV_SPECIFIC_NEIGHBOR;
	memcpy(&tlv->specific_neighbor, specific_neighbor, 6);
	tlv->link_metrics_type = metrix_type;

	return true;
}

// Transmitter link metric TLV associated structures ("Section 6.4.11")
bool get_transmitterLinkMetricTLV(struct transmitterLinkMetricTLV *tlv)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct transmitterLinkMetricTLV));
	tlv->tlv_type = TLV_TYPE_TRANSMITTER_LINK_METRIC;

	WARNING("dummy values");
	return true;
}

// Receiver link metric TLV associated structures ("Section 6.4.12")
bool get_receiverLinkMetricTLV(struct receiverLinkMetricTLV *tlv)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct receiverLinkMetricTLV));
	tlv->tlv_type = TLV_TYPE_RECEIVER_LINK_METRIC;

	WARNING("dummy values");
	return true;
}

// Link metric result code TLV associated structures ("Section 6.4.13")
bool get_linkMetricResultCodeTLV(struct linkMetricResultCodeTLV *tlv)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct linkMetricResultCodeTLV));
	tlv->tlv_type = TLV_TYPE_LINK_METRIC_RESULT_CODE;

	WARNING("dummy values");
	return true;
}

// Searched role TLV associated structures ("Section 6.4.14")
bool get_searchedRoleTLV(struct searchedRoleTLV *tlv)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct searchedRoleTLV));
	tlv->tlv_type = TLV_TYPE_SEARCHED_ROLE;
	tlv->role = IEEE80211_ROLE_REGISTRAR;  // Search registrar

	return true;
}

// Autoconfig frequency band TLV associated structures ("Section 6.4.15")
bool get_autoconfigFreqBandTLV(struct autoconfigFreqBandTLV *tlv, INT8U freq_band)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct autoconfigFreqBandTLV));
	tlv->tlv_type = TLV_TYPE_AUTOCONFIG_FREQ_BAND;
	tlv->freq_band = freq_band;

	return true;
}

// Supported role TLV associated structures ("Section 6.4.16")
bool get_supportedRoleTLV(struct supportedRoleTLV *tlv)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct supportedRoleTLV));
	tlv->tlv_type = TLV_TYPE_SUPPORTED_ROLE;
	tlv->role = IEEE80211_ROLE_REGISTRAR; // Registrar

	return true;
}

// Supported frequency band TLV associated structures ("Section 6.4.17")
bool get_supportedFreqBandTLV(struct supportedFreqBandTLV *tlv, INT8U freq_band)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct supportedFreqBandTLV));
	tlv->tlv_type = TLV_TYPE_SUPPORTED_FREQ_BAND;
	tlv->freq_band = freq_band;

	return true;
}

// Supported frequency band TLV associated structures ("Section 6.4.18")
bool get_wscTLV(struct wscTLV *tlv, INT8U *wsc_frame, INT16U wsc_frame_size)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct wscTLV));
	tlv->tlv_type = TLV_TYPE_WSC;
	tlv->wsc_frame_size = wsc_frame_size;
	CALLOC(tlv->wsc_frame, wsc_frame_size);
	memcpy(tlv->wsc_frame, wsc_frame, wsc_frame_size);

	return true;
}

// Push button event notification TLV associated structures ("Section 6.4.19")
bool get_pushButtonEventNotificationTLV(struct pushButtonEventNotificationTLV *tlv)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct pushButtonEventNotificationTLV));
	tlv->tlv_type = TLV_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION;

	WARNING("dummy values");
	return true;
}

// Push button join notification TLV associated structures ("Section 6.4.20")
bool get_pushButtonJoinNotificationTLV(struct pushButtonJoinNotificationTLV *tlv, INT8U *original_al_mac_address,
				       INT16U original_mid, INT8U *local_mac_address, INT8U *new_mac_address)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct pushButtonJoinNotificationTLV));
	tlv->tlv_type = TLV_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION;
	tlv->al_mac_address[0] = original_al_mac_address[0];
	tlv->al_mac_address[1] = original_al_mac_address[1];
	tlv->al_mac_address[2] = original_al_mac_address[2];
	tlv->al_mac_address[3] = original_al_mac_address[3];
	tlv->al_mac_address[4] = original_al_mac_address[4];
	tlv->al_mac_address[5] = original_al_mac_address[5];
	tlv->message_identifier = original_mid;
	tlv->mac_address[0] = local_mac_address[0];
	tlv->mac_address[1] = local_mac_address[1];
	tlv->mac_address[2] = local_mac_address[2];
	tlv->mac_address[3] = local_mac_address[3];
	tlv->mac_address[4] = local_mac_address[4];
	tlv->mac_address[5] = local_mac_address[5];
	tlv->new_mac_address[0] = new_mac_address[0];
	tlv->new_mac_address[1] = new_mac_address[1];
	tlv->new_mac_address[2] = new_mac_address[2];
	tlv->new_mac_address[3] = new_mac_address[3];
	tlv->new_mac_address[4] = new_mac_address[4];
	tlv->new_mac_address[5] = new_mac_address[5];

	return true;
}

// Generic PHY device information TLV associated structures ("Section 6.4.21")
bool get_genericPhyDeviceInformationTypeTLV(struct genericPhyDeviceInformationTypeTLV *tlv)
{
	INT8U al_mac_address[6] = {0};
	INT8U i;

	char **interfaces_names;
	INT8U interfaces_names_nr;

	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct genericPhyDeviceInformationTypeTLV));
	tlv->tlv_type = TLV_TYPE_GENERIC_PHY_DEVICE_INFORMATION;

	memcpy(al_mac_address, DMalMacGet(), 6);

	tlv->al_mac_address[0] = al_mac_address[0];
	tlv->al_mac_address[1] = al_mac_address[1];
	tlv->al_mac_address[2] = al_mac_address[2];
	tlv->al_mac_address[3] = al_mac_address[3];
	tlv->al_mac_address[4] = al_mac_address[4];
	tlv->al_mac_address[5] = al_mac_address[5];

	interfaces_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&interfaces_names_nr);

	tlv->local_interfaces_nr = 0;
	tlv->local_interfaces = NULL;

	for (i = 0; i < interfaces_names_nr; i++) {
		struct interfaceInfo *x;

		x = PLATFORM_GET_1905_INTERFACE_INFO(interfaces_names[i]);
		if (x == NULL) {
			// Error retrieving information for this interface.
			// Ignore it.
			WARNING("Could not retrieve info of interface %s", interfaces_names[i]);
			continue;
		}

		if (x->interface_type == MEDIA_TYPE_UNKNOWN) {
			// We are only interested in "generic" interfaces

			INT16U len;
			INT8U *m;

			if (tlv->local_interfaces_nr == 0)
				CALLOC(tlv->local_interfaces, 1);
			else
				REALLOC(tlv->local_interfaces, (tlv->local_interfaces_nr + 1));

			memcpy(tlv->local_interfaces[tlv->local_interfaces_nr].local_interface_address, x->mac_address, 6);
			tlv->local_interfaces[tlv->local_interfaces_nr].generic_phy_common_data.oui[0] = x->interface_type_data.other.oui[0];
			tlv->local_interfaces[tlv->local_interfaces_nr].generic_phy_common_data.oui[1] = x->interface_type_data.other.oui[1];
			tlv->local_interfaces[tlv->local_interfaces_nr].generic_phy_common_data.oui[2] = x->interface_type_data.other.oui[2];
			tlv->local_interfaces[tlv->local_interfaces_nr].generic_phy_common_data.variant_index =
			    x->interface_type_data.other.variant_index;

			m = forge_media_specific_blob(&x->interface_type_data.other, &len);
			if (m != NULL) {
				tlv->local_interfaces[tlv->local_interfaces_nr].generic_phy_common_data.media_specific_bytes_nr = len;
				tlv->local_interfaces[tlv->local_interfaces_nr].generic_phy_common_data.media_specific_bytes = m;
			} else {
				tlv->local_interfaces[tlv->local_interfaces_nr].generic_phy_common_data.media_specific_bytes_nr = 0;
				tlv->local_interfaces[tlv->local_interfaces_nr].generic_phy_common_data.media_specific_bytes = NULL;
			}

			memcpy(tlv->local_interfaces[tlv->local_interfaces_nr].variant_name, x->interface_type_data.other.variant_name,
			       strlen(x->interface_type_data.other.variant_name) + 1);
			tlv->local_interfaces[tlv->local_interfaces_nr].generic_phy_description_xml_url_len =
			    strlen(x->interface_type_data.other.generic_phy_description_xml_url) + 1;

			CALLOC(tlv->local_interfaces[tlv->local_interfaces_nr].generic_phy_description_xml_url, \
			       tlv->local_interfaces[tlv->local_interfaces_nr].generic_phy_description_xml_url_len);
			memcpy(tlv->local_interfaces[tlv->local_interfaces_nr].generic_phy_description_xml_url,
			       x->interface_type_data.other.generic_phy_description_xml_url,
			       tlv->local_interfaces[tlv->local_interfaces_nr].generic_phy_description_xml_url_len);

			tlv->local_interfaces_nr++;
		}
		PLATFORM_FREE_1905_INTERFACE_INFO(x);
	}

	PLATFORM_FREE_LIST_OF_1905_INTERFACES(interfaces_names, interfaces_names_nr);
	return true;
}

// Device identification type TLV associated structures ("Section 6.4.22")
bool get_deviceIdentificationTypeTLV(struct deviceIdentificationTypeTLV *tlv)
{
	struct deviceInfo *x = NULL;

	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct deviceIdentificationTypeTLV));
	tlv->tlv_type = TLV_TYPE_DEVICE_IDENTIFICATION;

	x = PLATFORM_GET_DEVICE_INFO();
	if (x == NULL) {
		ERROR("Could not retrieve device info");
		return false;
	}

	memcpy(tlv->friendly_name, x->friendly_name, strlen(x->friendly_name) + 1);
	memcpy(tlv->manufacturer_name, x->manufacturer_name, strlen(x->manufacturer_name) + 1);
	memcpy(tlv->manufacturer_model, x->manufacturer_model, strlen(x->manufacturer_model) + 1);
	return true;
}

// Control URL type TLV associated structures ("Section 6.4.23")
bool get_controlUrlTypeTLV(struct controlUrlTypeTLV *tlv)
{
	struct deviceInfo *x = NULL;

	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct controlUrlTypeTLV));
	tlv->tlv_type = TLV_TYPE_CONTROL_URL;
	tlv->url = NULL;

	x = PLATFORM_GET_DEVICE_INFO();
	if (x == NULL) {
		ERROR("Could not retrieve device info");
		return false;
	}

	if (x->control_url != NULL)
		tlv->url = strdup(x->control_url);

	return true;
}

// IPv4 type TLV associated structures ("Section 6.4.24")
bool get_ipv4TypeTLV(struct ipv4TypeTLV *tlv)
{
	char **ifs_names;
	INT8U ifs_nr;

	INT8U i, j;

	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct ipv4TypeTLV));
	tlv->tlv_type = TLV_TYPE_IPV4;

	tlv->ipv4_interfaces_nr = 0;

	ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);

	for (i = 0; i < ifs_nr; i++) {
		struct interfaceInfo *y;

		y = PLATFORM_GET_1905_INTERFACE_INFO(ifs_names[i]);
		if (y == NULL) {
			WARNING("Could not retrieve info of interface %s", ifs_names[i]);
			continue;
		}

		if (y->ipv4_nr != 0) {
			if (tlv->ipv4_interfaces_nr == 0)
				CALLOC(tlv->ipv4_interfaces, 1);
			else
				REALLOC(tlv->ipv4_interfaces, (tlv->ipv4_interfaces_nr + 1));

			memcpy(tlv->ipv4_interfaces[tlv->ipv4_interfaces_nr].mac_address, y->mac_address, 6);
			tlv->ipv4_interfaces[tlv->ipv4_interfaces_nr].ipv4_nr = y->ipv4_nr;
			CALLOC(tlv->ipv4_interfaces[tlv->ipv4_interfaces_nr].ipv4, y->ipv4_nr);
			for (j = 0; j < y->ipv4_nr; j++) {
				switch (y->ipv4[j].type) {
				case IPV4_UNKNOWN:
					{
						tlv->ipv4_interfaces[tlv->ipv4_interfaces_nr].ipv4[j].type = IPV4_TYPE_UNKNOWN;
						break;
					}
				case IPV4_DHCP:
					{
						tlv->ipv4_interfaces[tlv->ipv4_interfaces_nr].ipv4[j].type = IPV4_TYPE_DHCP;
						break;
					}
				case IPV4_STATIC:
					{
						tlv->ipv4_interfaces[tlv->ipv4_interfaces_nr].ipv4[j].type = IPV4_TYPE_STATIC;
						break;
					}
				case IPV4_AUTOIP:
					{
						tlv->ipv4_interfaces[tlv->ipv4_interfaces_nr].ipv4[j].type = IPV4_TYPE_AUTOIP;
						break;
					}
				default:
					{
						WARNING("Invalid IPv4 type %d", y->ipv4[j].type);
						break;
					}
				}
				memcpy(tlv->ipv4_interfaces[tlv->ipv4_interfaces_nr].ipv4[j].ipv4_address, y->ipv4[j].address, 4);
				memcpy(tlv->ipv4_interfaces[tlv->ipv4_interfaces_nr].ipv4[j].ipv4_dhcp_server, y->ipv4[j].dhcp_server, 4);
			}

			tlv->ipv4_interfaces_nr++;
		}

		PLATFORM_FREE_1905_INTERFACE_INFO(y);
	}

	PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);
	return true;
}

// IPv6 type TLV associated structures ("Section 6.4.25")
bool get_ipv6TypeTLV(struct ipv6TypeTLV *tlv)
{
	char **ifs_names;
	INT8U ifs_nr;

	INT8U i, j;

	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct ipv6TypeTLV));
	tlv->tlv_type = TLV_TYPE_IPV6;

	tlv->ipv6_interfaces_nr = 0;

	ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);

	for (i = 0; i < ifs_nr; i++) {
		struct interfaceInfo *y;

		y = PLATFORM_GET_1905_INTERFACE_INFO(ifs_names[i]);
		if (y == NULL) {
			WARNING("Could not retrieve info of interface %s", ifs_names[i]);
			continue;
		}

		if (y->ipv6_nr != 0) {
			if (tlv->ipv6_interfaces_nr == 0)
				CALLOC(tlv->ipv6_interfaces, 1);
			else
				REALLOC(tlv->ipv6_interfaces, (tlv->ipv6_interfaces_nr + 1));

			memcpy(tlv->ipv6_interfaces[tlv->ipv6_interfaces_nr].mac_address, y->mac_address, 6);
			tlv->ipv6_interfaces[tlv->ipv6_interfaces_nr].ipv6_nr = y->ipv6_nr;
			CALLOC(tlv->ipv6_interfaces[tlv->ipv6_interfaces_nr].ipv6, y->ipv6_nr);
			for (j = 0; j < y->ipv6_nr; j++) {
				switch (y->ipv6[j].type) {
				case IPV6_UNKNOWN:
					{
						tlv->ipv6_interfaces[tlv->ipv6_interfaces_nr].ipv6[j].type = IPV6_TYPE_UNKNOWN;
						break;
					}
				case IPV6_DHCP:
					{
						tlv->ipv6_interfaces[tlv->ipv6_interfaces_nr].ipv6[j].type = IPV6_TYPE_DHCP;
						break;
					}
				case IPV6_STATIC:
					{
						tlv->ipv6_interfaces[tlv->ipv6_interfaces_nr].ipv6[j].type = IPV6_TYPE_STATIC;
						break;
					}
				case IPV6_SLAAC:
					{
						tlv->ipv6_interfaces[tlv->ipv6_interfaces_nr].ipv6[j].type = IPV6_TYPE_SLAAC;
						break;
					}
				default:
					{
						WARNING("Invalid IPv6 type %d", y->ipv6[j].type);
						break;
					}
				}
				memcpy(tlv->ipv6_interfaces[tlv->ipv6_interfaces_nr].ipv6[j].ipv6_address, y->ipv6[j].address, 16);
				memcpy(tlv->ipv6_interfaces[tlv->ipv6_interfaces_nr].ipv6[j].ipv6_address_origin, y->ipv6[j].origin, 16);
			}

			tlv->ipv6_interfaces_nr++;
		}

		PLATFORM_FREE_1905_INTERFACE_INFO(y);
	}

	PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);
	return true;
}

// Push button generic PHY event notification TLV structures ("Section 6.4.26")
bool get_pushButtonGenericPhyEventNotificationTLV(struct pushButtonGenericPhyEventNotificationTLV *tlv)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct pushButtonGenericPhyEventNotificationTLV));
	tlv->tlv_type = TLV_TYPE_GENERIC_PHY_EVENT_NOTIFICATION;

	WARNING("dummy values");
	return true;
}

// Profile version TLV associated structures ("Section 6.4.27")
bool get_x1905ProfileVersionTLV(struct x1905ProfileVersionTLV *tlv)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct x1905ProfileVersionTLV));
	tlv->tlv_type = TLV_TYPE_1905_PROFILE_VERSION;
	tlv->profile = PROFILE_1905_1A;

	return true;
}

// Power off interface TLV associated structures ("Section 6.4.28")
bool get_powerOffInterfaceTLV(struct powerOffInterfaceTLV *tlv)
{
	char **interfaces_names;
	INT8U interfaces_names_nr;
	INT8U i;

	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct powerOffInterfaceTLV));
	tlv->tlv_type = TLV_TYPE_POWER_OFF_INTERFACE;

	tlv->power_off_interfaces_nr = 0;
	tlv->power_off_interfaces = NULL;

	interfaces_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&interfaces_names_nr);

	// Search for interfaces in "POWER OFF" mode
	//
	for (i = 0; i < interfaces_names_nr; i++) {
		struct interfaceInfo *x;

		x = PLATFORM_GET_1905_INTERFACE_INFO(interfaces_names[i]);
		if (x == NULL) {
			// Error retrieving information for this interface.
			// Ignore it.
			//
			continue;
		}

		if (x->power_state != INTERFACE_POWER_STATE_OFF) {
			// Ignore interfaces that are not in "POWER OFF" mode
			//
			PLATFORM_FREE_1905_INTERFACE_INFO(x);
			continue;
		}

		if (tlv->power_off_interfaces_nr == 0)
			CALLOC(tlv->power_off_interfaces, 1);
		else
			REALLOC(tlv->power_off_interfaces, (tlv->power_off_interfaces_nr + 1));

		memcpy(tlv->power_off_interfaces[tlv->power_off_interfaces_nr].interface_address, x->mac_address, 6);

		tlv->power_off_interfaces[tlv->power_off_interfaces_nr].media_type = x->interface_type;

		// Only when the media type is "MEDIA_TYPE_UNKNOWN", fill the
		// rest of fields
		//
		if (tlv->power_off_interfaces[tlv->power_off_interfaces_nr].media_type != MEDIA_TYPE_UNKNOWN) {
			// Set everything to "zero"
			//
			tlv->power_off_interfaces[tlv->power_off_interfaces_nr].generic_phy_common_data.oui[0] = 0x00;
			tlv->power_off_interfaces[tlv->power_off_interfaces_nr].generic_phy_common_data.oui[1] = 0x00;
			tlv->power_off_interfaces[tlv->power_off_interfaces_nr].generic_phy_common_data.oui[2] = 0x00;
			tlv->power_off_interfaces[tlv->power_off_interfaces_nr].generic_phy_common_data.variant_index = 0;
			tlv->power_off_interfaces[tlv->power_off_interfaces_nr].generic_phy_common_data.media_specific_bytes_nr = 0;
			tlv->power_off_interfaces[tlv->power_off_interfaces_nr].generic_phy_common_data.media_specific_bytes = NULL;
		} else {
			INT16U len;
			INT8U *m;

			tlv->power_off_interfaces[tlv->power_off_interfaces_nr].generic_phy_common_data.oui[0] =
			    x->interface_type_data.other.oui[0];
			tlv->power_off_interfaces[tlv->power_off_interfaces_nr].generic_phy_common_data.oui[1] =
			    x->interface_type_data.other.oui[1];
			tlv->power_off_interfaces[tlv->power_off_interfaces_nr].generic_phy_common_data.oui[2] =
			    x->interface_type_data.other.oui[2];
			tlv->power_off_interfaces[tlv->power_off_interfaces_nr].generic_phy_common_data.variant_index =
			    x->interface_type_data.other.variant_index;

			m = forge_media_specific_blob(&x->interface_type_data.other, &len);

			if (m != NULL) {
				tlv->power_off_interfaces[tlv->power_off_interfaces_nr].generic_phy_common_data.media_specific_bytes_nr = len;
				tlv->power_off_interfaces[tlv->power_off_interfaces_nr].generic_phy_common_data.media_specific_bytes = m;
			} else {
				// Ignore media specific data
				//
				tlv->power_off_interfaces[tlv->power_off_interfaces_nr].generic_phy_common_data.media_specific_bytes_nr = 0;
				tlv->power_off_interfaces[tlv->power_off_interfaces_nr].generic_phy_common_data.media_specific_bytes = NULL;
			}
		}
		tlv->power_off_interfaces_nr++;

		PLATFORM_FREE_1905_INTERFACE_INFO(x);
	}

	PLATFORM_FREE_LIST_OF_1905_INTERFACES(interfaces_names, interfaces_names_nr);
	if (tlv->power_off_interfaces_nr == 0) {
#ifdef SEND_EMPTY_TLVS
		return true;
#else
		return false;
#endif
	}
	return true;
}

// Interface power change information TLV structures ("Section 6.4.29")
bool get_interfacePowerChangeInformationTLV(struct interfacePowerChangeInformationTLV *tlv)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct interfacePowerChangeInformationTLV));
	tlv->tlv_type = TLV_TYPE_INTERFACE_POWER_CHANGE_INFORMATION;

	WARNING("dummy values");
	return true;
}

// Interface power change status TLV associated structures ("Section 6.4.29")
bool get_interfacePowerChangeStatusTLV(struct interfacePowerChangeStatusTLV *tlv)
{
	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct interfacePowerChangeStatusTLV));
	tlv->tlv_type = TLV_TYPE_INTERFACE_POWER_CHANGE_STATUS;

	WARNING("dummy values");
	return true;
}

// L2 neighbor device TLV associated structures ("Section 6.4.31")
bool get_l2NeighborDeviceTLV(struct l2NeighborDeviceTLV *tlv)
{
	char **interfaces_names;
	INT8U interfaces_names_nr;
	INT8U i, j;

	if (tlv == NULL)
		return false;

	memset(tlv, 0, sizeof(struct l2NeighborDeviceTLV));

	tlv->tlv_type = TLV_TYPE_L2_NEIGHBOR_DEVICE;
	tlv->local_interfaces_nr = 0;
	tlv->local_interfaces = NULL;

	interfaces_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&interfaces_names_nr);

	// Search for interfaces in "POWER OFF" mode
	//
	for (i = 0; i < interfaces_names_nr; i++) {
		struct interfaceInfo *x;

		x = PLATFORM_GET_1905_INTERFACE_INFO(interfaces_names[i]);
		if (x == NULL)
			continue;

		if (x->neighbor_mac_addresses_nr == 0 ||
		    x->neighbor_mac_addresses_nr == INTERFACE_NEIGHBORS_UNKNOWN) {
			// Ignore interfaces that do not have (or cannot report) L2
			// neighbors
			PLATFORM_FREE_1905_INTERFACE_INFO(x);
			continue;
		}

		if (tlv->local_interfaces_nr == 0)
			CALLOC(tlv->local_interfaces, 1);
		else
			REALLOC(tlv->local_interfaces, (tlv->local_interfaces_nr + 1));

		memcpy(tlv->local_interfaces[tlv->local_interfaces_nr].local_mac_address, x->mac_address, 6);

		tlv->local_interfaces[tlv->local_interfaces_nr].l2_neighbors_nr = x->neighbor_mac_addresses_nr;
		CALLOC(tlv->local_interfaces[tlv->local_interfaces_nr].l2_neighbors,
		       x->neighbor_mac_addresses_nr);

		for (j = 0; j < x->neighbor_mac_addresses_nr; j++) {
			memcpy(tlv->local_interfaces[tlv->local_interfaces_nr].l2_neighbors[j].l2_neighbor_mac_address,
			       x->neighbor_mac_addresses[j], 6);

			// TODO: Modify "struct interfaceInfo" in "platform_interfaces.h"
			// to provide "behind MACs" information.
			// But first... find out WHAT "behind MACs" really means!!
			//
			tlv->local_interfaces[tlv->local_interfaces_nr].l2_neighbors[j].behind_mac_addresses_nr = 0;
			tlv->local_interfaces[tlv->local_interfaces_nr].l2_neighbors[j].behind_mac_addresses = NULL;
		}

		tlv->local_interfaces_nr++;

		PLATFORM_FREE_1905_INTERFACE_INFO(x);
	}

	PLATFORM_FREE_LIST_OF_1905_INTERFACES(interfaces_names, interfaces_names_nr);

	if (tlv->local_interfaces_nr == 0) {
#ifdef SEND_EMPTY_TLVS
		return true;
#else
		return false;
#endif
	}

	return true;
}

