/*
 * main.c: IEEE-1905 main
 *
 * Author: Yashvardhan <y.yashvardhan@iopsys.eu>
 * Author: Vivek Dutta <vivek.dutta@iopsys.eu>
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include "utils.h"
#include "utils_json.h"
#include "1905_ubus.h"
#include "al_ubus.h"
#include "al_datamodel.h"
#include "plugin_mngr.h"
#include "alme_server.h"
#include "platform_interfaces.h"
#include "platform_interfaces_ubus.h"
#include "platform_os.h"
#include "al_entity.h"
#include "utils_uci.h"

#include <stdio.h>		// printf
#include <syslog.h>
#include <unistd.h>		// getopt
#include <stdlib.h>		// exit
#include <string.h>		// strtok
#include <pthread.h>		// pthread_*
#include <json-c/json.h>	// json_*
#include <easy/easy.h>
#include <signal.h>

static bool gs_registrar_enabled;

#define PLUGIN_LIB "libwifimap2"
#define PLUGIN_PATH "/usr/lib/ieee1905"

static void DeviceInfoInit()
{
	char friendly_name[32], mname[32], mmodel[32], control_url[64];

	uci_get_option_db("device", "deviceinfo", "FriendlyName", friendly_name, sizeof(friendly_name));
	uci_get_option_db("device", "deviceinfo", "Manufacturer", mname, sizeof(mname));
	uci_get_option_db("device", "deviceinfo", "ModelName", mmodel, sizeof(mmodel));

	snprintf(control_url, sizeof(control_url), "http://192.168.1.1");

	update_device_info(friendly_name, mname, mmodel, control_url);
}

static void al_sighandler(int sig)
{
	uint8_t msg[3] = {0x10, 0, 0};
	INFO("Received signal(%d)", sig);
	set_stop_flag();
	sendMessageToAlQueue(msg, 3);
}

void configure_al_uci()
{
	char temp[128];
	uint8_t val;


	DMinit();

	if (uci_get_option("ieee1905", "ieee1905", "enabled", temp, 128)) {
		if(get_boolean_val(temp) == false) {
			fprintf(stderr, "AL-Agent is not enabled");
			al_cleanup();
		}
	} else {
		fprintf(stderr, "AL-Agent is not enabled");
		al_cleanup();
	}

	if (uci_get_option("ieee1905", "ieee1905", "debug", temp, 128)) {
		if (uci_get_option("ieee1905", "ieee1905", "debug_level", temp, 128)) {
			val = atoi(temp);
			set_print_level(val);
		}
	} else {
		set_print_level(3);
	}

	if (uci_get_option("ieee1905", "ieee1905", "macaddress", temp, 20)) {
		uint8_t almac[6];
		if ((strncasecmp(temp, "auto", 4) != 0)) {
			hwaddr_aton(temp, almac);
			DMalMacSet(almac);
		}
	}

	if (uci_get_option("ieee1905", "ieee1905", "registrar", temp, 128)) {
		gs_registrar_enabled = get_boolean_val(temp);
	} else {
		gs_registrar_enabled = false;
	}

	if (uci_get_option("ieee1905", "ieee1905", "cmdu_event", temp, 10)) {
		val = get_boolean_val(temp);
		configure_cmdu_event((bool)val);
	} else {
		configure_cmdu_event(false);
	}

	if (uci_get_option("ieee1905", "ieee1905", "map_plugin", temp, 10)) {
		val = get_boolean_val(temp);
		if (val) {
			INFO("Installing multiAP plugin");
			installPlugin(PLUGIN_PATH, PLUGIN_LIB);
		}
	}

	if (PLATFORM_INTERFACE_MUTEX_INIT() == false) {
		ERROR("Platform interface mutex init failed");
		al_cleanup();
	}

	uci_configure_al_iface();
	if (DMalMacConfigured() == false) {
		uint8_t mac[6];
		if (get_mac_addr("br-lan", mac)) {
			WARNING("Configure br-lan mac as almac");
			DMalMacSet(mac);
		} else if (get_mac_addr("br-wan", mac)) {
			WARNING("Configure br-wan mac as almac");
			DMalMacSet(mac);
		} else {
			ERROR("AL-mac not configured, exiting....");
			al_cleanup();
		}
	}
}

void pid_init()
{
	FILE *fd;
	pid_t pid;

	fd = fopen(PID_FILE, "w");
	if (fd == NULL) {
		fprintf(stderr, "PID failed to write\n");
		exit(1);
	}

	pid = getpid();

	fprintf(fd, "%ld", (long int)pid);
	fclose(fd);
}


int al_init_start(void)
{
	pthread_t tid;

	init_pid_list();
	INFO("Update basic info from uci");
	configure_al_uci();
	updateIntfBasicInfo();
	INFO("Update device info");
	DeviceInfoInit();

	// Start Ubus thread here
	INFO("Creating ubus thread");
	if (pthread_create(&tid, NULL, create_ubus_thread, NULL) != 0) {
		ERROR("Error starting ubus thread");
		al_cleanup();
	}

	add_tid_to_list(tid);
#ifdef ALME_OVER_TCP
	join_alme_server();
#endif
	INFO("Starting AL-Agent");
	if (start1905AL(gs_registrar_enabled) != 0) {
		ERROR("Failed to start AL-Agent");
		al_cleanup();
	}

	join_all_tid_list();

	return 0;
}

static void signal_init()
{
	set_sighandler(SIGINT, al_sighandler);
	set_sighandler(SIGTERM, al_sighandler);
}

int main(void)
{
	openlog("ieee1905", LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);

	pid_init();
	signal_init();

	al_init_start();

	al_cleanup();
	return 0;
}
