/*
 * platform_interfaces.c: IEEE-1905 interface related utilities
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "utils.h"
#include "utils_uci.h"
#include "platform_interfaces.h"
#include "platform_interfaces_ubus.h"
#include "platform_os.h"
#include "1905_tlvs.h"
#include "al_ubus.h"

#include <stdio.h>		// printf(), popen()
#include <stdlib.h>		// malloc(), ssize_t
#include <stdarg.h>		// va_*
#include <string.h>		// strdup()
#include <errno.h>		// errno
#include <unistd.h>		// sleep()

#include <sys/ioctl.h>		// ioctl(), SIOCGIFINDEX
#include <net/if.h>		// struct ifreq, IFNAMSIZ
#include <arpa/inet.h>		// socket, AF_INTER, htons(), ...
#include <linux/if_packet.h>	// sockaddr_ll
#include <netinet/ether.h>	// ETH_P_ALL, ETH_A_LEN
#include <unistd.h>		// close()
#include <pthread.h>		// pthread_create(), mutex functions
#include <easy/easy.h>
#include "wifi.h"
////////////////////////////////////////////////////////////////////////////////
// Private data and functions
////////////////////////////////////////////////////////////////////////////////

// Global variables to store the list of names of the interfaces and their
// status.
//
static int sg_interfaces_nr;
static char **sg_interfaces_list;
extern void get_bridge_info(struct bridge_info *binfo);
// The interfaces status variables can be accessed/modified from different
// threads:
//   - The "main" AL thread (where "start1905AL()" runs)
//   - The "push button" configuration process thread
//
// Thus, their access must be protected with a mutex
//
static pthread_mutex_t sg_interface_mutex = PTHREAD_MUTEX_INITIALIZER;

// This function returns '0' if there was a problem,  otherwise it returns an
// ID identifying the interface type.
//

// These are the data structures and functions in charge of handling the press
// of physical/virtual buttons on the platform
//
struct _pushButtonThreadData {
	char *interface_name;
	INT8U al_mac_address[6];
	INT16U mid;
};

static void *_pushButtonConfigurationThread(void *p_arg)
{
	// This function is executed when the "push button" configuration mechanism
	// is started on an interface.
	//
	// It will wait until the process either:
	//
	//   A) Fails. In this case the interface will remain on its previous state
	//     (either "secure" or "not secure")
	//
	//   B) Succeeds. In this case the interface status will be set to
	//      "authenticated", no matter what its previous state was,  and a new
	//      "PLATFORM_QUEUE_EVENT_AUTHENTICATED_LINK" message will be posted to
	//      the AL queue.
	//      This message contains both the just autenticated local interface MAC
	//      address and the MAC of the interface at the other end of this same
	//      link.
	//
	// Behaviour depending in the interface type:
	//
	//   - WIFI: (TODO)
	//       The WPS procedure starts and if it succeeds a message is posted to
	//       the AL queue containing the MAC address of the wifi node at the
	//       other end of the now encrypted link.

	struct _pushButtonThreadData *aux;
	struct interfaceInfo *x;
	uint8_t *p;

	bool push_button_result = false;
	unsigned char new_mac[6];
	INT16U interface_type;
	INT8U local_interface_mac_address[6];

	aux = (struct _pushButtonThreadData *)p_arg;

	DEBUG("[PLATFORM] *Push button configuration thread* Starting on interface %s", aux->interface_name);

	x = PLATFORM_GET_1905_INTERFACE_INFO(aux->interface_name);
	if (x == NULL) {
		ERROR("[PLATFORM] *Push button configuration thread* Error retrieving interface %s information");
		return NULL;
	}

	interface_type = x->interface_type;
	memcpy(local_interface_mac_address, x->mac_address, 6);

	PLATFORM_FREE_1905_INTERFACE_INFO(x);

	new_mac[0] = 0x00;
	new_mac[1] = 0x00;
	new_mac[2] = 0x00;
	new_mac[3] = 0x00;
	new_mac[4] = 0x00;
	new_mac[5] = 0x00;

	// Start the "push button process"

	{
		// *********************************************************************
		// ********************** REGULAR INTERFACE ****************************
		// *********************************************************************

		switch (interface_type) {
		case MEDIA_TYPE_IEEE_802_3U_FAST_ETHERNET:
		case MEDIA_TYPE_IEEE_802_3AB_GIGABIT_ETHERNET:
			{
				// Ethernet interfaces do not support the 'push button' configuration mechanism.
				//
				break;
			}
		case MEDIA_TYPE_IEEE_802_11B_2_4_GHZ:
		case MEDIA_TYPE_IEEE_802_11G_2_4_GHZ:
		case MEDIA_TYPE_IEEE_802_11A_5_GHZ:
		case MEDIA_TYPE_IEEE_802_11N_2_4_GHZ:
		case MEDIA_TYPE_IEEE_802_11N_5_GHZ:
		case MEDIA_TYPE_IEEE_802_11AC_5_GHZ:
		case MEDIA_TYPE_IEEE_802_11AD_60_GHZ:
		case MEDIA_TYPE_IEEE_802_11AF_GHZ:
			{
				int status = start_wifi_pbc(aux->interface_name);

				INFO("#### wifi wps status |%d|###", status);
				break;
			}
		case MEDIA_TYPE_IEEE_1901_WAVELET:
		case MEDIA_TYPE_IEEE_1901_FFT:
			{
				// TODO
				ERROR("NOT Supported!!!");
				break;
			}
		case MEDIA_TYPE_MOCA_V1_1:
			{
				// TODO
				ERROR("NOT Supported!!!");
				break;
			}
		case MEDIA_TYPE_UNKNOWN:
			{
				ERROR("[PLATFORM] *Push button configuration thread* Unknown interface type");
				break;
			}
		default:
			{
				ERROR("[PLATFORM] *Push button configuration thread* Impossible interface type %d",
				      interface_type);
				break;
			}
		}
	}

	// Now wait until the process finishes
	if (get_wifi_wps_status(aux->interface_name)) {
		// The "push button process" has finished!
		push_button_result = true;
		get_pbc_join_mac(aux->interface_name, new_mac);
	}

	if (false == push_button_result) {
		INFO("[PBT] Timeout or error on interface %s. Stopping...", aux->interface_name);
	} else {
		INFO("[PBT] Success! New device (%02x:%02x:%02x:%02x:%02x:%02x) on interface %s. Stopping...",
		      new_mac[0], new_mac[1], new_mac[2], new_mac[3], new_mac[4], new_mac[5], aux->interface_name);

		// Post a "PLATFORM_QUEUE_EVENT_AUTHENTICATED_LINK" message
		{
			INT8U message[3 + 20];
			INT16U mlen = 20;
			INT16U mid = aux->mid;

			message[0] = PLATFORM_QUEUE_EVENT_AUTHENTICATED_LINK;
			p = &message[1];
			_I2B(&mlen, &p);

			_InB(&local_interface_mac_address[0], &p, 6);
			_InB(&new_mac[0], &p, 6);
			_InB(&aux->al_mac_address[0], &p, 6);
			_I2B(&mid, &p);

			DEBUG("[PBT] Sending 23 bytes to queue (0x%02x, 0x%02x, 0x%02x, ...)", message[0], message[1],
			      message[2]);

			if (sendMessageToAlQueue(message, 3 + 20) == 0) {
				ERROR("[PBT] Error sending message to queue from _pushButtonThread()");
				free(p_arg);
				return NULL;
			}
		}
	}

	free(p_arg);
	return NULL;
}

bool PLATFORM_INTERFACE_MUTEX_INIT(void)
{
	if (pthread_mutex_init(&sg_interface_mutex, NULL) != 0) {
		ERROR("Interface mutex init failed");
		return false;
	}
	return true;
}

void PLATFORM_INTERFACE_MUTEX_DESTROY(void)
{
	pthread_mutex_destroy(&sg_interface_mutex);
}

void addInterface(const char *long_interface_name)
{
	pthread_mutex_lock(&sg_interface_mutex);
	if (sg_interfaces_nr == 0) {
		CALLOC(sg_interfaces_list, 1);
	} else {
		REALLOC(sg_interfaces_list, (sg_interfaces_nr + 1));
	}

	sg_interfaces_list[sg_interfaces_nr] = strdup(long_interface_name);

	sg_interfaces_nr++;
	pthread_mutex_unlock(&sg_interface_mutex);

	return;
}

void clean_interface_data()
{
	uint8_t i;

	pthread_mutex_lock(&sg_interface_mutex);

	for (i = 0; i < sg_interfaces_nr; i++) {
		FREE(sg_interfaces_list[i]);
	}
	sg_interfaces_nr = 0;
	FREE(sg_interfaces_list);

	pthread_mutex_unlock(&sg_interface_mutex);
}
////////////////////////////////////////////////////////////////////////////////
// Platform API: Interface related functions to be used by platform-independent
// files (functions declarations are  found in "../interfaces/utils.h)
////////////////////////////////////////////////////////////////////////////////
bool get_mac_addr(const char *if_name, INT8U *mac)
{
	bool ret = true;
	if (if_name == NULL || mac == NULL)
		return false;

	// Get mac-address using ioctl
	struct ifreq s;
	int fd;

	strcpy(s.ifr_name, if_name);
	fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
	if (ioctl(fd, SIOCGIFHWADDR, &s) != 0) {
		ERROR("Could not obtain MAC address of interface %s", if_name);
		close(fd);
		ret = false;
	}
	memcpy(mac, s.ifr_addr.sa_data, 6);
	close(fd);

	return ret;
}

int get_ifindex(const char *interface_name)
{
	int sock, ret = -1;
	struct ifreq ifr;

	sock = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
	if (-1 == sock) {
		ERROR("error opening socket(%s) errno=%d (%s)",
		      interface_name, errno, strerror(errno));
		return -1;
	}

	strncpy(ifr.ifr_name, interface_name, IFNAMSIZ);
	if (ioctl(sock, SIOCGIFINDEX, &ifr) == -1) {
		ERROR("ioctl (%s',SIOCGIFINDEX) ret errno=%d (%s) on RAW socket",
		      interface_name, errno, strerror(errno));
		close(sock);
		return -1;
	}

	ret = ifr.ifr_ifindex;
	close(sock);

	return ret;
}

char **PLATFORM_GET_LIST_OF_1905_INTERFACES(INT8U *nr)
{
	INT8U i;

	pthread_mutex_lock(&sg_interface_mutex);
	char **list;

	CALLOC(list, sg_interfaces_nr);
	*nr = sg_interfaces_nr;
	for (i = 0; i < sg_interfaces_nr; ++i) {
		list[i] = strdup(sg_interfaces_list[i]);
	}
	pthread_mutex_unlock(&sg_interface_mutex);
	return list;
}

void PLATFORM_FREE_LIST_OF_1905_INTERFACES(char **x, INT8U nr)
{
	INT8U i;

	for (i = 0; i < nr; ++i) {
		FREE(x[i]);
	}

	FREE(x);
	return;
}

void fill_dummy_interface_info(struct interfaceInfo *m)
{
	m->mac_address[0] = 0x00;
	m->mac_address[1] = 0x00;
	m->mac_address[2] = 0x00;
	m->mac_address[3] = 0x00;
	m->mac_address[4] = 0x00;
	m->mac_address[5] = 0x00;

	memcpy(m->manufacturer_name, "Unknown", strlen("Unknown") + 1);
	memcpy(m->model_name, "Unknown", strlen("Unknown") + 1);
	memcpy(m->model_number, "00000000", strlen("00000000") + 1);
	memcpy(m->serial_number, "00000000", strlen("00000000") + 1);
	memcpy(m->device_name, "Unknown", strlen("Unknown") + 1);
	memcpy(m->uuid, "0000000000000000", strlen("0000000000000000") + 1);

	m->interface_type = MEDIA_TYPE_UNKNOWN;
	m->interface_type_data.other.oui[0] = 0x00;
	m->interface_type_data.other.oui[1] = 0x00;
	m->interface_type_data.other.oui[2] = 0x00;
	m->interface_type_data.other.generic_phy_description_xml_url = NULL;
	m->interface_type_data.other.variant_index = 0;
	m->interface_type_data.other.variant_name = 0;
	m->interface_type_data.other.media_specific.unsupported.bytes_nr = 0;
	m->interface_type_data.other.media_specific.unsupported.bytes = NULL;

	m->is_secured = 0;
	m->push_button_on_going = 2;	// "2" means "unsupported"
	m->push_button_new_mac_address[0] = 0x00;
	m->push_button_new_mac_address[1] = 0x00;
	m->push_button_new_mac_address[2] = 0x00;
	m->push_button_new_mac_address[3] = 0x00;
	m->push_button_new_mac_address[4] = 0x00;
	m->push_button_new_mac_address[5] = 0x00;

	m->power_state = INTERFACE_POWER_STATE_OFF;
	m->neighbor_mac_addresses_nr = INTERFACE_NEIGHBORS_UNKNOWN;
	m->neighbor_mac_addresses = NULL;

	m->ipv4_nr = 0;
	m->ipv4 = NULL;
	m->ipv6_nr = 0;
	m->ipv6 = NULL;

	m->vendor_specific_elements_nr = 0;
	m->vendor_specific_elements = NULL;
}

int iface_get_hwaddr(char *ifname, uint8_t *mac)
{
	int fd;
	struct ifreq ifr;

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (!socket)
		return -1;

	ifr.ifr_addr.sa_family = AF_INET;
	strncpy(ifr.ifr_name , ifname , IFNAMSIZ-1);

	if (ioctl(fd, SIOCGIFHWADDR, &ifr)) {
		close(fd);
		return -1;
	}

	close(fd);

	memcpy(mac, ifr.ifr_hwaddr.sa_data, 6);

	DEBUG("Mac : %.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n" , mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);

	return 0;
}

struct interfaceInfo *PLATFORM_GET_1905_INTERFACE_INFO(char *interface_name)
{
	struct interfaceInfo *m;
	bool found_mac = true;
	INT8U i;

	DEBUG("[PLATFORM] Retrieving info for interface %s", interface_name);

	m = (struct interfaceInfo *)calloc(1, sizeof(struct interfaceInfo));
	if (m == NULL) {
		ERROR("[PLATFORM] Not enough memory for the 'interface' structure");
		return NULL;
	}

	m->name = strdup(interface_name);
	fill_dummy_interface_info(m);

	// Give "sane" values in case any of the following parameters can not be filled later
	if (get_interface_info(interface_name, m))
		found_mac = false;

	DEBUG("[PLATFORM]   mac_address                 : %02x:%02x:%02x:%02x:%02x:%02x",
	      m->mac_address[0], m->mac_address[1], m->mac_address[2],
	      m->mac_address[3], m->mac_address[4], m->mac_address[5]);
	DEBUG("[PLATFORM]   manufacturer_name           : %s", m->manufacturer_name);
	DEBUG("[PLATFORM]   model_name                  : %s", m->model_name);
	DEBUG("[PLATFORM]   model_number                : %s", m->model_number);
	DEBUG("[PLATFORM]   serial_number               : %s", m->serial_number);
	DEBUG("[PLATFORM]   device_name                 : %s", m->device_name);
	DEBUG("[PLATFORM]   uuid                        : %s", m->uuid);
	DEBUG("[PLATFORM]   interface_type              : %d", m->interface_type);
	if (MEDIA_TYPE_IEEE_802_11B_2_4_GHZ == m->interface_type ||
	    MEDIA_TYPE_IEEE_802_11G_2_4_GHZ == m->interface_type ||
	    MEDIA_TYPE_IEEE_802_11A_5_GHZ == m->interface_type ||
	    MEDIA_TYPE_IEEE_802_11N_2_4_GHZ == m->interface_type ||
	    MEDIA_TYPE_IEEE_802_11N_5_GHZ == m->interface_type ||
	    MEDIA_TYPE_IEEE_802_11AC_5_GHZ == m->interface_type ||
	    MEDIA_TYPE_IEEE_802_11AD_60_GHZ == m->interface_type ||
	    MEDIA_TYPE_IEEE_802_11AF_GHZ == m->interface_type) {
		DEBUG("[PLATFORM]     ieee80211 data");
		DEBUG("[PLATFORM]       bssid                       : %02x:%02x:%02x:%02x:%02x:%02x",
		      m->interface_type_data.ieee80211.bssid[0], m->interface_type_data.ieee80211.bssid[1],
		      m->interface_type_data.ieee80211.bssid[2], m->interface_type_data.ieee80211.bssid[3],
		      m->interface_type_data.ieee80211.bssid[4], m->interface_type_data.ieee80211.bssid[5]);
		DEBUG("[PLATFORM]       ssid                        : %s", m->interface_type_data.ieee80211.ssid);
		DEBUG("[PLATFORM]       role                        : %d", m->interface_type_data.ieee80211.role);
		DEBUG("[PLATFORM]       ap_channel_band             : 0x%02x",
		      m->interface_type_data.ieee80211.ap_channel_band);
		DEBUG("[PLATFORM]       ap_channel_center_f1        : 0x%02x",
		      m->interface_type_data.ieee80211.ap_channel_center_frequency_index_1);
		DEBUG("[PLATFORM]       ap_channel_center_f2        : 0x%02x",
		      m->interface_type_data.ieee80211.ap_channel_center_frequency_index_2);
		DEBUG("[PLATFORM]       authentication_mode         : 0x%04x",
		      m->interface_type_data.ieee80211.authentication_mode);
		DEBUG("[PLATFORM]       encryption_mode             : 0x%04x",
		      m->interface_type_data.ieee80211.encryption_mode);
		DEBUG("[PLATFORM]       network_key                 : %s",
		      m->interface_type_data.ieee80211.network_key);
	} else if (MEDIA_TYPE_IEEE_1901_WAVELET == m->interface_type || MEDIA_TYPE_IEEE_1901_FFT == m->interface_type) {
		DEBUG("[PLATFORM]     ieee1901 data");
		DEBUG("[PLATFORM]       network_identifier          : %02x:%02x:%02x:%02x:%02x:%02x:%02x",
		      m->interface_type_data.ieee1901.network_identifier[0],
		      m->interface_type_data.ieee1901.network_identifier[1],
		      m->interface_type_data.ieee1901.network_identifier[2],
		      m->interface_type_data.ieee1901.network_identifier[3],
		      m->interface_type_data.ieee1901.network_identifier[4],
		      m->interface_type_data.ieee1901.network_identifier[5],
		      m->interface_type_data.ieee1901.network_identifier[6]);
	} else if (m->interface_type == MEDIA_TYPE_UNKNOWN) {
		INT16U len;
		INT8U *data;

		fill_dummy_interface_info(m);

		DEBUG("[PLATFORM]     generic interface data");
		DEBUG("[PLATFORM]       OUI                           : %02x:%02x:%02x",
		      m->interface_type_data.other.oui[0],
		      m->interface_type_data.other.oui[1],
		      m->interface_type_data.other.oui[2]);
		DEBUG("[PLATFORM]       URL description               : %s",
		      NULL == m->interface_type_data.other.generic_phy_description_xml_url ?
		      "<none>" : m->interface_type_data.other.generic_phy_description_xml_url);
		DEBUG("[PLATFORM]       variant index                 : %d",
		      m->interface_type_data.other.variant_index);
		DEBUG("[PLATFORM]       variant name                  : %s",
		      NULL == m->interface_type_data.other.variant_name ?
		      "<none>" : m->interface_type_data.other.variant_name);
		data = forge_media_specific_blob(&m->interface_type_data.other, &len);
		if (data != NULL) {
			if (len > 4) {
				DEBUG("[PLATFORM] media specific data (%d bytes) : 0x%02x:0x%02x:0x%02x:0x%02x:0x%02x.",
				      len, data[0], data[1], data[2], data[3], data[4]);
			} else {
				DEBUG("[PLATFORM]       media specific data (%d bytes)", len);
			}
			free_media_specific_blob(data);
		}
	}
	DEBUG("[PLATFORM]   is_secure                   : %d", m->is_secured);
	DEBUG("[PLATFORM]   push_button_on_going        : %d", m->push_button_on_going);
	DEBUG("[PLATFORM]   push_button_new_mac_address : %02x:%02x:%02x:%02x:%02x:%02x",
	      m->push_button_new_mac_address[0], m->push_button_new_mac_address[1],
	      m->push_button_new_mac_address[2], m->push_button_new_mac_address[3],
	      m->push_button_new_mac_address[4], m->push_button_new_mac_address[5]);
	DEBUG("[PLATFORM]   power_state                 : %d", m->power_state);
	DEBUG("[PLATFORM]   neighbor_mac_addresses_nr   : %d", m->neighbor_mac_addresses_nr);
	if (m->neighbor_mac_addresses_nr != INTERFACE_NEIGHBORS_UNKNOWN) {
		for (i = 0; i < m->neighbor_mac_addresses_nr; i++) {
			DEBUG("[PLATFORM]     - neighbor #%d                : %02x:%02x:%02x:%02x:%02x:%02x",
			      i, m->neighbor_mac_addresses[i][0],
			      m->neighbor_mac_addresses[i][1],
			      m->neighbor_mac_addresses[i][2],
			      m->neighbor_mac_addresses[i][3],
			      m->neighbor_mac_addresses[i][4],
			      m->neighbor_mac_addresses[i][5]);
		}
	}
	DEBUG("[PLATFORM]   IPs                         : %d", m->ipv4_nr + m->ipv6_nr);
	for (i = 0; i < m->ipv4_nr; i++) {
		DEBUG("[PLATFORM]     - IPv4 #%d : %d.%d.%d.%d (type = %s, dhcpserver = %d.%d.%d.%d)",
		      i, m->ipv4[i].address[0],
		      m->ipv4[i].address[1], m->ipv4[i].address[2], m->ipv4[i].address[3],
		      m->ipv4[i].type == IPV4_UNKNOWN ? "unknown" : m->ipv4[i].type == IPV4_DHCP ? "dhcp" :
		      m->ipv4[i].type == IPV4_STATIC ? "static" : m->ipv4[i].type == IPV4_AUTOIP ? "auto" :
		      "error", m->ipv4[i].dhcp_server[0], m->ipv4[i].dhcp_server[1],
		      m->ipv4[i].dhcp_server[2], m->ipv4[i].dhcp_server[3]);
	}
	for (i = 0; i < m->ipv6_nr; i++) {
		DEBUG
		    ("[PLATFORM]     - IPv6 #%d : %02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x (type = %s, origin = %02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x)",
		     i, m->ipv6[i].address[0], m->ipv6[i].address[1], m->ipv6[i].address[2], m->ipv6[i].address[3], m->ipv6[i].address[4],
		     m->ipv6[i].address[5], m->ipv6[i].address[6], m->ipv6[i].address[7], m->ipv6[i].address[8], m->ipv6[i].address[9], m->ipv6[i].address[10],
		     m->ipv6[i].address[11], m->ipv6[i].address[12], m->ipv6[i].address[13], m->ipv6[i].address[14], m->ipv6[i].address[15],
		     m->ipv6[i].type == IPV6_UNKNOWN ? "unknown" : m->ipv6[i].type == IPV6_DHCP ? "dhcp" : m->ipv6[i].type ==
		     IPV6_STATIC ? "static" : m->ipv6[i].type == IPV6_SLAAC ? "slaac" : "error", m->ipv6[i].origin[0], m->ipv6[i].origin[1],
		     m->ipv6[i].origin[2], m->ipv6[i].origin[3], m->ipv6[i].origin[4], m->ipv6[i].origin[5], m->ipv6[i].origin[6], m->ipv6[i].origin[7],
		     m->ipv6[i].origin[8], m->ipv6[i].origin[9], m->ipv6[i].origin[10], m->ipv6[i].origin[11], m->ipv6[i].origin[12], m->ipv6[i].origin[13],
		     m->ipv6[i].origin[14], m->ipv6[i].origin[15]);
	}
	DEBUG("[PLATFORM]   vendor_specific_elements_nr : %d", m->vendor_specific_elements_nr);
	for (i = 0; i < m->vendor_specific_elements_nr; i++) {
		char buffer[256] = {0};
		size_t vendor_data_len = m->vendor_specific_elements[i].vendor_data_len;

		DEBUG("[PLATFORM]     - vendor %d");
		DEBUG("[PLATFORM]         OUI                       : %02x:%02x:%02x", m->vendor_specific_elements[i].oui[0],
		      m->vendor_specific_elements[i].oui[1], m->vendor_specific_elements[i].oui[2]);
		DEBUG("[PLATFORM]         vendor_data_len           : %d", vendor_data_len);
		btostr(m->vendor_specific_elements[i].vendor_data, ((vendor_data_len < 128)?vendor_data_len:128), buffer);
		DEBUG("[PLATFORM]         vendor_data               : %s", buffer);
	}

	if (!found_mac)
		iface_get_hwaddr(interface_name, m->mac_address);

	return m;
}

void PLATFORM_FREE_1905_INTERFACE_INFO(struct interfaceInfo *x)
{
	INT8U i;

	free(x->name);

	if (x->interface_type == MEDIA_TYPE_UNKNOWN) {
		if (!(x->interface_type_data.other.oui[0] == 0x00 &&
		      x->interface_type_data.other.oui[1] == 0x19 &&
		      x->interface_type_data.other.oui[2] == 0xA7 &&
		      0 == strcmp("http://handle.itu.int/11.1002/3000/1706", x->interface_type_data.other.generic_phy_description_xml_url) &&
		      (x->interface_type_data.other.variant_index == 1 || x->interface_type_data.other.variant_index == 2
		       || x->interface_type_data.other.variant_index == 3 || x->interface_type_data.other.variant_index == 4)
		    )
		    ) {
			if (0 != x->interface_type_data.other.media_specific.unsupported.bytes_nr
			    && NULL != x->interface_type_data.other.media_specific.unsupported.bytes) {
				free(x->interface_type_data.other.media_specific.unsupported.bytes);
			}
		}

		if (x->interface_type_data.other.generic_phy_description_xml_url != NULL) {
			free(x->interface_type_data.other.generic_phy_description_xml_url);
		}
		if (x->interface_type_data.other.variant_name != NULL) {
			free(x->interface_type_data.other.variant_name);
		}
	}

	if (x->neighbor_mac_addresses_nr > 0 && INTERFACE_NEIGHBORS_UNKNOWN != x->neighbor_mac_addresses_nr && NULL != x->neighbor_mac_addresses) {
		free(x->neighbor_mac_addresses);
	}

	if (x->ipv4_nr > 0 && NULL != x->ipv4) {
		free(x->ipv4);
	}

	if (x->ipv6_nr > 0 && NULL != x->ipv6) {
		free(x->ipv6);
	}

	if (x->vendor_specific_elements_nr > 0) {
		for (i = 0; i < x->vendor_specific_elements_nr; i++) {
			if (x->vendor_specific_elements[i].vendor_data_len > 0 && NULL != x->vendor_specific_elements[i].vendor_data) {
				free(x->vendor_specific_elements[i].vendor_data);
			}
		}
		free(x->vendor_specific_elements);
	}

	free(x);
}

bool PLATFORM_GET_LINK_METRICS(char *local_interface_name,
			       INT8U *neighbor_interface_address,
			       struct linkMetrics *linkm)
{
	int status;
	uint8_t intf_type;

	if (linkm == NULL)
		return false;

	status = get_mac_addr(local_interface_name, linkm->local_interface_address);
	if (status == (int) false) {
		return false;
	}

	// Copy the remote interface MAC address
	memcpy(linkm->neighbor_interface_address, neighbor_interface_address, 6);

	intf_type = get_interface_type(local_interface_name);

	//Here we get the interface type if it is not ethernet then it is wifi
	if (intf_type == INTF_TYPE_WIFI) {
		struct wifi_sta sta_info;
		memset(&sta_info, 0, sizeof(struct wifi_sta));
		status = wifi_get_sta_info(local_interface_name, neighbor_interface_address, &sta_info);
		if (status != 0) {
			return false;
		}

		linkm->tx_packet_ok = sta_info.stats.tx_pkts;
		linkm->tx_packet_errors = sta_info.stats.tx_err_pkts;
		linkm->tx_max_xput = (uint16_t) sta_info.maxrate;
		linkm->tx_phy_rate = (uint16_t) sta_info.tx_rate.rate;
		linkm->tx_link_availability = sta_info.airtime;
		linkm->rx_packet_ok = sta_info.stats.rx_pkts;
		linkm->rx_packet_errors = sta_info.stats.rx_fail_pkts;
		linkm->rx_rssi = sta_info.rssi_avg;
		linkm->measures_window = sta_info.conn_time;

	} else if (intf_type == INTF_TYPE_ETHERNET) {
		// Obtain how much time the process collecting stats has been running
		//
		//     TODO: This should be set to the amount of seconds ellapsed since
		//     the interface was brought up. However I could not find an easy
		//     way to obtain this information in Linux.
		//     For now, we will simply set this to the amount of seconds since
		//     the system was started, which is typically correct on most
		//     cases.
		//

		linkm->measures_window = get_time() / 1000;
		// Obtain the amount of (correct and incorrect) packets transmitted
		// to 'neighbor_interface_address' in the last
		// 'linkm->measures_window' seconds.
		//
		//   TODO: In Linux there is no easy way to obtain this
		//   information. We will just report the amount of *total* packets
		//   transmitted from our local interface, no matter the
		//   destination.
		//   This information will only be correct when the local interface
		//   is connected to one single remote interface... however we
		//   better report this than nothing at all.
		//
		// This is done by reading the contents of files
		//
		//   "/sys/class/net/<interface_name>/statistics/tx_packets" and
		//   "/sys/class/net/<interface_name>/statistics/tx_errors"
		//
		linkm->tx_packet_ok = (INT32U) read_interface_param32(local_interface_name, "statistics/tx_packets");
		linkm->tx_packet_errors = (INT32U) read_interface_param32(local_interface_name, "statistics/tx_errors");

		// Obtain the estimatid max MAC xput and PHY rate when transmitting
		// data from "A" to "B".
		//
		//   TODO: The same considerations as in the previous parameters
		//   apply here.
		//
		// This is done by reading the contents of file
		//
		//   "/sys/class/net/<interface_name>/speed
		//
		// NOTE: I'll set both parameters to the same value. Is there a
		// better way to do this?
		//
		linkm->tx_max_xput = (INT16U) read_interface_param32(local_interface_name, "speed");
		linkm->tx_phy_rate = (INT16U) read_interface_param32(local_interface_name, "speed");

		// Obtain the estimated average percentage of time that the link is
		// available for transmission.
		//
		//   TODO: I'll just say "100% of the time" for now.
		//
		linkm->tx_link_availability = 100;

		// Obtain the amount of (correct and incorrect) packets received from
		// 'neighbor_interface_address' in the last 'linkm->measures_window'
		// seconds.
		//
		//   TODO: In Linux there is no easy way to obtain this information. We
		//   will just report the amount of *total* packets received on our
		//   local interface, no matter the origin.
		//   This information will only be correct when the local interface is
		//   connected to one single remote interface... however we better
		//   report this than nothing at all.
		//
		// This is done by reading the contents of files
		//
		//   "/sys/class/net/<interface_name>/statistics/rx_packets" and
		//   "/sys/class/net/<interface_name>/statistics/rx_errors"
		//
		linkm->rx_packet_ok = (INT32U) read_interface_param32(local_interface_name, "statistics/rx_packets");
		linkm->rx_packet_errors = (INT32U) read_interface_param32(local_interface_name, "statistics/rx_errors");

		// Obtain the estimated RX RSSI
		//   TODO: ???
		linkm->rx_rssi = 0;
	} else {
		// Generic interface type not supported
		return false;
	}

	return true;
}

struct bridge *PLATFORM_GET_LIST_OF_BRIDGES(INT8U *nr)
{
	struct bridge *br = NULL;
	struct bridge_info br_info;

	*nr = 0;
	memset(&br_info, 0, sizeof(struct bridge_info));

	get_bridge_info(&br_info);

	if (br_info.br_nr) {
		*nr = br_info.br_nr;
		br = br_info.br;
	}

	return br;
}

void PLATFORM_FREE_LIST_OF_BRIDGES(struct bridge *x, INT8U nr)
{
	uint8_t i, j;

	if (nr == 0 || x == NULL)
		return;

	for (i = 0; i < nr; i++) {
		for (j = 0; j < x[i].bridged_interfaces_nr; j++) {
			free(x[i].bridged_interfaces[j]);
		}
		free(x[i].name);
	}

	free(x);
}

INT8U PLATFORM_SEND_RAW_PACKET(const char *interface_name, INT8U *dst_mac, INT8U *src_mac, INT16U eth_type, INT8U *payload, INT16U payload_len)
{
	int i, first_time;
	int sock;

	char aux1[200];
	char aux2[10];

	int ifindex;
	struct sockaddr_ll socket_address;

	INT8U buffer[MAX_NETWORK_SEGMENT_SIZE];
	struct ether_header *eh;

	// Print packet (used for debug purposes)
	DEBUG("[PLATFORM] Preparing to send RAW packet:");
	DEBUG("[PLATFORM]   - Interface name = %s", interface_name);
	DEBUG("[PLATFORM]   - DST  MAC       = %02x:%02x:%02x:%02x:%02x:%02x",
	      dst_mac[0], dst_mac[1], dst_mac[2],
	      dst_mac[3], dst_mac[4], dst_mac[5]);

	DEBUG("[PLATFORM]   - SRC  MAC       = %02x:%02x:%02x:%02x:%02x:%02x",
	      src_mac[0], src_mac[1], src_mac[2],
	      src_mac[3], src_mac[4], src_mac[5]);

	DEBUG("[PLATFORM]   - Ether type     = 0x%04x", eth_type);
	DEBUG("[PLATFORM]   - Payload length = %d", payload_len);

	aux1[0] = 0x0;
	aux2[0] = 0x0;
	first_time = 1;
	for (i = 0; i < payload_len; i++) {
		snprintf(aux2, 6, "0x%02x ", payload[i]);
		strncat(aux1, aux2, 200 - strlen(aux1) - 1);

		if (0 != i && 0 == (i + 1) % 8) {
			if (first_time == 1) {
				DEBUG("[PLATFORM]   - Payload        = %s", aux1);
				first_time = 0;
			} else {
				DEBUG("[PLATFORM]                      %s", aux1);
			}
			aux1[0] = 0x0;
		}
	}
	if (first_time == 1) {
		DEBUG("[PLATFORM]   - Payload        = %s", aux1);
	} else {
		DEBUG("[PLATFORM]                      %s", aux1);
	}

	// Open RAW socket
	DEBUG("[PLATFORM] Opening RAW socket");
	// Retrieve ethernet interface index
	ifindex = get_ifindex(interface_name);
	if (-1 == ifindex) {
		ERROR("failed to get interface index |%s|", interface_name);
		return 0;
	}
	DEBUG("[PLATFORM] Successfully got interface index %d", ifindex);

	sock = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
	if (-1 == sock) {
		ERROR("error opening socket(%s) errno=%d (%s)",
		      interface_name, errno, strerror(errno));
		return 0;
	}

	// Empy buffer
	memset(buffer, 0, MAX_NETWORK_SEGMENT_SIZE);

	// Fill ethernet header
	//
	eh = (struct ether_header *)buffer;
	eh->ether_dhost[0] = dst_mac[0];
	eh->ether_dhost[1] = dst_mac[1];
	eh->ether_dhost[2] = dst_mac[2];
	eh->ether_dhost[3] = dst_mac[3];
	eh->ether_dhost[4] = dst_mac[4];
	eh->ether_dhost[5] = dst_mac[5];
	eh->ether_shost[0] = src_mac[0];
	eh->ether_shost[1] = src_mac[1];
	eh->ether_shost[2] = src_mac[2];
	eh->ether_shost[3] = src_mac[3];
	eh->ether_shost[4] = src_mac[4];
	eh->ether_shost[5] = src_mac[5];
	eh->ether_type = htons(eth_type);

	// Fill buffer
	memcpy(buffer + sizeof(*eh), payload, payload_len);

	// Prepare sockaddr_ll
	memset(&socket_address, 0, sizeof(socket_address));
	socket_address.sll_ifindex = ifindex;
	socket_address.sll_halen = ETH_ALEN;
	socket_address.sll_addr[0] = dst_mac[0];
	socket_address.sll_addr[1] = dst_mac[1];
	socket_address.sll_addr[2] = dst_mac[2];
	socket_address.sll_addr[3] = dst_mac[3];
	socket_address.sll_addr[4] = dst_mac[4];
	socket_address.sll_addr[5] = dst_mac[5];
	socket_address.sll_addr[6] = 0x00;
	socket_address.sll_addr[7] = 0x00;


	DEBUG("[PLATFORM] Sending data to RAW socket");
	// 60 is the minimum ethernet frame length
	if (-1 == sendto(sock, buffer, sizeof(*eh) + payload_len >= 60 ? sizeof(*eh) + payload_len : 60,
			 0, (struct sockaddr *)&socket_address, sizeof(socket_address))
	    ) {
		ERROR("[PLATFORM] sendto('%s') returned with errno=%d (%s)",
		      interface_name, errno, strerror(errno));
		close(sock);
		return 0;
	}
	DEBUG("[PLATFORM] Data sent!");

	close(sock);
	return 1;
}

INT8U PLATFORM_START_PUSH_BUTTON_CONFIGURATION(char *interface_name, INT8U *al_mac_address, INT16U mid)
{
	pthread_t thread;
	struct _pushButtonThreadData *p;
	struct interfaceInfo *x;

	// Make sure the interface:
	//   - is not already in the middle of a "push button" configuration
	//     process and...
	//   - ...it has support for the "push button" configuration mechanism.
	//
	x = PLATFORM_GET_1905_INTERFACE_INFO(interface_name);
	if (x == NULL) {
		return 0;
	}
	if (x->push_button_on_going != 0) {
		PLATFORM_FREE_1905_INTERFACE_INFO(x);
		return 1;
	}
	PLATFORM_FREE_1905_INTERFACE_INFO(x);

	p = (struct _pushButtonThreadData *)malloc(sizeof(struct _pushButtonThreadData));
	if (p == NULL) {
		// Out of memory
		//
		return 0;
	}

	p->interface_name = strdup(interface_name);
	p->al_mac_address[0] = al_mac_address[0];
	p->al_mac_address[1] = al_mac_address[1];
	p->al_mac_address[2] = al_mac_address[2];
	p->al_mac_address[3] = al_mac_address[3];
	p->al_mac_address[4] = al_mac_address[4];
	p->al_mac_address[5] = al_mac_address[5];
	p->mid = mid;

	if (pthread_create(&thread, NULL, _pushButtonConfigurationThread, (void *)p) != 0) {
		ERROR("Failed to create push button config thread!!!");
		return 0;
	}

	return 1;
}

INT8U PLATFORM_SET_INTERFACE_POWER_MODE(char *interface_name, INT8U power_mode)
{
	// TODO
	INFO("Not implemented |%s|, |%x|", interface_name, power_mode);
	switch (power_mode) {
	case INTERFACE_POWER_STATE_ON:
		{
			DEBUG("[PLATFORM] %s --> POWER ON", interface_name);
			break;
		}
	case INTERFACE_POWER_STATE_OFF:
		{
			DEBUG("[PLATFORM] %s --> POWER OFF", interface_name);
			break;
		}
	case INTERFACE_POWER_STATE_SAVE:
		{
			DEBUG("[PLATFORM] %s --> POWER SAVE", interface_name);
			break;
		}
	default:
		{
			WARNING("[PLATFORM] Unknown power mode for interface %s (%d)", interface_name, power_mode);
			return 0;
		}
	}

	return INTERFACE_POWER_RESULT_EXPECTED;
}

INT8U PLATFORM_CONFIGURE_80211_AP(char *interface_name, INT8U *ssid, INT8U *bssid,
				  INT16U auth_type, INT16U encryption_type, INT8U *network_key)
{
	bool ret;
        char auth_type_str[20] = {0};

	INFO("Applying WSC configuration (%s): ", interface_name);
	INFO("  - SSID            : %s", ssid);
	INFO("  - BSSID           : %02x:%02x:%02x:%02x:%02x:%02x",
	     bssid[0], bssid[1], bssid[2], bssid[3], bssid[4], bssid[5]);
	INFO("  - AUTH_TYPE       : 0x%04x", auth_type);
	INFO("  - ENCRYPTION_TYPE : 0x%04x", encryption_type);
	INFO("  - NETWORK_KEY     : %s", network_key);

	if (false == get_encryption_value(auth_type, encryption_type, auth_type_str, 20)) {
		ERROR("Unsupported encryption or cipher received!!");
		return 0;
	}


	ret = uci_check_wifi_iface(UCI_IEEE1905, interface_name);
	if (ret == false) {
		ret = uci_add_wireless_iface_sec(UCI_IEEE1905, interface_name);
		if (ret == false)
			return 0;
	}

	// Set uci in ieee1905
	uci_set_wireless_interface_option(UCI_IEEE1905, interface_name, "ssid", (char *) ssid);
	uci_set_wireless_interface_option(UCI_IEEE1905, interface_name, "key", (char *) network_key);
	uci_set_wireless_interface_option(UCI_IEEE1905, interface_name, "encryption", auth_type_str);

	// Set uci in wireless
	uci_set_wireless_interface_option(UCI_WIRELESS, interface_name, "ssid", (char *) ssid);
	uci_set_wireless_interface_option(UCI_WIRELESS, interface_name, "key", (char *) network_key);
	uci_set_wireless_interface_option(UCI_WIRELESS, interface_name, "encryption", auth_type_str);

	uci_reload_services();
	return 1;
}

INT8U *forge_media_specific_blob(struct genericInterfaceType *m, INT16U *len)
{
#define ITU_T_GHN_XML "http://handle.itu.int/11.1002/3000/1706"

	INT8U *ret;

	ret = NULL;

	if (m->generic_phy_description_xml_url == NULL) {
		return NULL;
	}

	if (memcmp(m->generic_phy_description_xml_url, ITU_T_GHN_XML, strlen(ITU_T_GHN_XML) + 1) == 0) {
		// This XML file defines the *same* media specific data format for all
		// interfaces that meet the following requirements:
		//
		//   - OUI     = 00:19:A7
		//   - Variant = 0, 1, 2, 3 or 4 (it also defines 10 and 11, but we will
		//               ignore these)

		if (m->oui[0] != 0x00 || m->oui[1] != 0x19 || m->oui[2] != 0xa7) {
			// Unknown OUI
		} else if (m->variant_index != 0 && m->variant_index != 1 && m->variant_index != 2 && m->variant_index != 3 && m->variant_index != 4) {
			// Unknown variant
		} else {
			// The 1905 media specific field is made up of FIVE bytes:
			//
			//   0x01, 0x00, 0x02, dni[0], dni[1]
			//
			// (see ITU-T G.9979 Tables 8.2 and 8.3)
			//
			*len = 5;
			CALLOC(ret, *len);
			ret[0] = 0x01;
			ret[1] = 0x00;
			ret[2] = 0x02;
			ret[3] = m->media_specific.ituGhn.dni[0];
			ret[4] = m->media_specific.ituGhn.dni[1];
		}
	}

	if (ret == NULL) {
		// If we get to this point and "ret" is still "NULL", that means that the
		// "XML"/"OUI"/"variant_index" combination has not been recognized, and thus
		// we will simply return the contents of the "unsupported" structure.
		//
		*len = m->media_specific.unsupported.bytes_nr;
		CALLOC(ret, *len);

		memcpy(ret, m->media_specific.unsupported.bytes, *len);
	}

	return ret;
}

const char *get_interface_power_string(uint8_t power_type)
{

	switch(power_type) {
	case INTERFACE_POWER_STATE_ON:
		return "On";
	case INTERFACE_POWER_STATE_SAVE:
		return "Power_Save";
	case INTERFACE_POWER_STATE_OFF:
		return "Off";
	}

	return "Unsupported";
}

const char *get_interface_type_string(INT16U if_type)
{

	switch (if_type) {
	case MEDIA_TYPE_IEEE_802_3U_FAST_ETHERNET:
		return "IEEE 802.3u";
	case MEDIA_TYPE_IEEE_802_3AB_GIGABIT_ETHERNET:
		return "IEEE 802.3ab";
	case MEDIA_TYPE_IEEE_802_11B_2_4_GHZ:
		return "IEEE 802.11b";
	case MEDIA_TYPE_IEEE_802_11G_2_4_GHZ:
		return "IEEE 802.11g";
	case MEDIA_TYPE_IEEE_802_11A_5_GHZ:
		return "IEEE 802.11a";
	case MEDIA_TYPE_IEEE_802_11N_2_4_GHZ:
		return "IEEE 802.11n 2.4";
	case MEDIA_TYPE_IEEE_802_11N_5_GHZ:
		return "IEEE 802.11n 5.0";
	case MEDIA_TYPE_IEEE_802_11AC_5_GHZ:
		return "IEEE 802.11ac";
	case MEDIA_TYPE_IEEE_802_11AD_60_GHZ:
		return "IEEE 802.11ad";
	case MEDIA_TYPE_IEEE_802_11AF_GHZ:
		return "IEEE 802.11af";
	case MEDIA_TYPE_IEEE_1901_WAVELET:
		return "IEEE 1901 Wavelet";
	case MEDIA_TYPE_IEEE_1901_FFT:
		return "IEEE 1901 FFT";
	case MEDIA_TYPE_MOCA_V1_1:
		return "MoCAv1.1";
	case MEDIA_TYPE_UNKNOWN:
		return "Generic PHY";
	}
	return "UNKNOWN";
}

bool get_encryption_value(INT16U auth_type, INT16U encryption_type, char *encrypt_val, size_t elen)
{
	if (encrypt_val == NULL)
                return false;

	if ((auth_type & WPS_AUTH_WPAPSK) && (auth_type & WPS_AUTH_WPA2PSK))
		strncat(encrypt_val, "psk-mixed", elen);
	else if ((auth_type & WPS_AUTH_WPA) && (auth_type & WPS_AUTH_WPA2))
                strncat(encrypt_val, "wpa-mixed", elen);
	else if (auth_type & WPS_AUTH_WPAPSK)
		strncat(encrypt_val, "psk", elen);
        else if (auth_type & WPS_AUTH_WPA2PSK)
		strncat(encrypt_val, "psk2", elen);
	else if (auth_type & WPS_AUTH_WPA)
		strncat(encrypt_val, "wpa", elen);
	else if (auth_type & WPS_AUTH_WPA2)
		strncat(encrypt_val, "wpa2", elen);
	else
		return false;

	//Check for the encryption type
	if((encryption_type & WPS_ENCR_TKIP) && (encryption_type & WPS_ENCR_AES))
		strncat(encrypt_val, "+tkip+aes", elen);
	else if(encryption_type & WPS_ENCR_TKIP)
		strncat(encrypt_val, "+tkip", elen);
	else if(encryption_type & WPS_ENCR_AES)
		strncat(encrypt_val, "+aes", elen);

	return true;
}
