/*
 * platform_interfaces_ubus.c: IEEE-1905 interface info over ubus
 *
 * Author: Vivek Dutta <vivek.dutta@iopsys.eu>
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "utils.h"
#include "utils_uci.h"
#include "platform_interfaces.h"	// struct interfaceInfo
#include "platform_interfaces_ubus.h"
#include "1905_tlvs.h"
#include "al_ubus.h"

#include <stdlib.h>		// malloc()
#include <unistd.h>
#include <errno.h>
#include <ifaddrs.h>
#include <net/if.h>
#include "easy/easy.h"
#include "wifi.h"


////////////////////////////////////////////////////////////////////////////////
// Private data and functions
////////////////////////////////////////////////////////////////////////////////
static INT8S get_metadata_index(char *ifname);
typedef struct {
	char name[IF_NAMESIZE];
	INT8U intf_type;
	INT8U bssid[6];
	char network_key[64];
	INT8U mac_address[6];
	INT8U new_pbc_join_mac[6];
} interface_basic_info_t;

static interface_basic_info_t *gp_interface_info;
static INT8U gtotal_defined_interfaces;


#ifndef IF_OPER_UNKNOWN
enum {
	IF_OPER_UNKNOWN,
	IF_OPER_NOTPRESENT,
	IF_OPER_DOWN,
	IF_OPER_LOWERLAYERDOWN,
	IF_OPER_TESTING,
	IF_OPER_DORMANT,
	IF_OPER_UP,
};
#endif /* IF_OPER_UNKNOWN */

// 2 - 2.4 G, 5 - 5G band
INT8U get_wifi_band(struct interfaceInfo *x)
{
	if (x->interface_type_data.ieee80211.ap_channel_center_frequency_index_1 > 14)
		return IEEE80211_FREQUENCY_BAND_5_GHZ;

	return IEEE80211_FREQUENCY_BAND_2_4_GHZ;
}

char **get_list_of_unconfigured_ap(INT8U *ap_nr)
{
	INT8U i;
	INT8U count = 0;
	char **ap_list;

	CALLOC(ap_list, gtotal_defined_interfaces);

	if (ap_nr == NULL)
		return NULL;

	for (i = 0; i < gtotal_defined_interfaces; i++) {
		if (hwaddr_is_zero(gp_interface_info[i].bssid) &&
		    (gp_interface_info[i].intf_type == INTF_TYPE_WIFI)) {
			ap_list[count] = gp_interface_info[i].name;
			count++;
		}
	}
	*ap_nr = count;
	return ap_list;
}

INT8U get_unconfigured_ap_nbr(void)
{
	INT8U number = 0;
	char **ap_list = get_list_of_unconfigured_ap(&number);

	FREE(ap_list);

	return number;
}

bool update_pbc_join_mac(char *ifname, INT8U *mac)
{
	INT8S index = get_metadata_index(ifname);

	memcpy(&gp_interface_info[index].new_pbc_join_mac, mac, 6);
	return true;
}

bool get_pbc_join_mac(char *ifname, INT8U *mac)
{
	INT8S index = get_metadata_index(ifname);

	memcpy(mac, &gp_interface_info[index].new_pbc_join_mac, 6);
	memset(&gp_interface_info[index].new_pbc_join_mac, 0, 6);
	return true;
}
int start_wifi_pbc(char *ifname)
{
	struct wps_param wps = {
		.role = WPS_REGISTRAR,
		.method = WPS_METHOD_PBC
	};
	return wifi_start_wps(ifname, wps);
}

// ret = true(wps succsess), false(wps timeout/fail)
bool get_wifi_wps_status(char *interface_name)
{
	bool ret = false;
	enum wps_status wps_code = 0;

	wifi_get_wps_status(interface_name, &wps_code);
	while (wps_code == WPS_STATUS_PROCESSING) {
		wps_code = 0;
		wifi_get_wps_status(interface_name, &wps_code);
		sleep(5);
	}

	if (wps_code == WPS_STATUS_SUCCESS)
		ret = true;

	return ret;
}
#define NEIGH_CMD "ip neigh show dev %s|awk '{print $3}'|sort|uniq"

// Returns an INT32U obtained by reading the first line of file
// "/sys/class/net/<interface_name>/<parameter_name>"
//
INT32S read_interface_param32(char *interface_name, char *parameter_name)
{
	INT32S ret;
	FILE *fp;
	char sys_path[100];

	snprintf(sys_path, sizeof(sys_path), "/sys/class/net/%s/%s", interface_name, parameter_name);

	ret = 0;
	fp = fopen(sys_path, "r");
	if (fp != NULL) {
		char aux[30];

		if (fgets(aux, sizeof(aux), fp) != NULL)
			ret = atoi(aux);

		fclose(fp);
	}

	return ret;
}

char *read_interface_param(char *interface_name, char *parameter_name)
{
	char *ret = NULL;
	FILE *fp;
	char sys_path[100];

	snprintf(sys_path, sizeof(sys_path), "/sys/class/net/%s/%s", interface_name, parameter_name);

	fp = fopen(sys_path, "r");
	if (fp != NULL) {
		char aux[30];

		if (fgets(aux, sizeof(aux), fp) != NULL)
			ret = strdup(aux);

		fclose(fp);
	}

	return ret;
}

INT8U (*get_neigh(char *ifname, INT8U *neigh_nr))[6]
{
	char cmd[256] = {0};
	FILE *fp = NULL;
	char *line = NULL;
	size_t len = 0;
	INT8U (*ret)[6] = NULL;
	INT8U count = 0;

	if (ifname == NULL)
		return NULL;

	snprintf(cmd, 256, NEIGH_CMD, ifname);

	fp = popen(cmd, "r");
	if (fp == NULL) {
		ERROR("unable to execute command");
		return NULL;
	}

	while (getline(&line, &len, fp) != -1) {
		if (len < 17) {
			FREE(line);
			len = 0;
			continue;
		}

		ret = (INT8U(*)[6]) realloc(ret, sizeof(INT8U[6]) * (count + 1));

		hwaddr_aton(line, ret[count]);
		count++;
		FREE(line);
		len = 0;
	}
	FREE(line);
	pclose(fp);

	*neigh_nr = count;

	return ret;
}

INT8U (*get_neigh_wifi(char *ifname, INT8U *neigh_nr))[6]
{
	INT8U (*ret)[6] = NULL;
	unsigned char stas[768] = {0};	/* max 128 per interface */
	int count = 128;
	int i = 0;

	*neigh_nr = 0;

	if (ifname == NULL)
		return NULL;

	if (wifi_get_assoclist(ifname, stas, &count) != 0)
		return NULL;

	ret = (INT8U(*)[6]) malloc(sizeof(INT8U[6]) * count);
	if (ret == NULL) {
		ERROR("Out of memory!");
		exit(ENOMEM);
	}

	for (i = 0; i < count; i++) {
		unsigned char *sta = &stas[i*6];

		ret[i][0] = sta[0];
		ret[i][1] = sta[1];
		ret[i][2] = sta[2];
		ret[i][3] = sta[3];
		ret[i][4] = sta[4];
		ret[i][5] = sta[5];
	}

	*neigh_nr = count;

	return ret;
}

bool fill_wifi_sta_info(char *interface_name, struct interfaceInfo *m)
{
	m->interface_type_data.ieee80211.role = IEEE80211_ROLE_AP;
	struct wifi_bss bss_info;

	memset(&bss_info, 0, sizeof(struct wifi_bss));

	wifi_sta_get_ap_info(interface_name, &bss_info);

	// Retrieve SSID information
	memcpy(m->interface_type_data.ieee80211.ssid, bss_info.ssid, 33);

	// Retrieve bssid
	memcpy(m->interface_type_data.ieee80211.bssid, bss_info.bssid, 6);

	m->interface_type = MEDIA_TYPE_IEEE_802_11N_2_4_GHZ;
	if (bss_info.band == BAND_5) {
		switch (bss_info.oper_std) {
		case 2: // a
			m->interface_type = MEDIA_TYPE_IEEE_802_11A_5_GHZ;
			break;
		case 3: // n
			m->interface_type = MEDIA_TYPE_IEEE_802_11N_5_GHZ;
			break;
		case 4: //ac
			m->interface_type = MEDIA_TYPE_IEEE_802_11AC_5_GHZ;
			break;
		}
	} else {
		switch (bss_info.oper_std) {
		case 0: // b
			m->interface_type = MEDIA_TYPE_IEEE_802_11B_2_4_GHZ;
			break;
		case 1: // g
			m->interface_type = MEDIA_TYPE_IEEE_802_11G_2_4_GHZ;
			break;
		case 3: // n
			m->interface_type = MEDIA_TYPE_IEEE_802_11N_2_4_GHZ;
			break;
		}
	}
	// Get sec related params
	m->is_secured                                        = 1;

	m->interface_type_data.ieee80211.encryption_mode = IEEE80211_AUTH_MODE_OPEN;

	if (bss_info.enc & WIFI_CIPHER_AES)
		m->interface_type_data.ieee80211.encryption_mode = IEEE80211_ENCRYPTION_MODE_AES;

	if (bss_info.enc & WIFI_CIPHER_TKIP)
		m->interface_type_data.ieee80211.encryption_mode = IEEE80211_ENCRYPTION_MODE_TKIP;

	m->interface_type_data.ieee80211.ap_channel_band = bss_info.curr_bw;
	m->interface_type_data.ieee80211.ap_channel_center_frequency_index_1 = bss_info.channel;
	// TODO: This need to change for 80+80
	m->interface_type_data.ieee80211.ap_channel_center_frequency_index_2 = bss_info.channel;

	//Retrieve list of connected devices
	m->neighbor_mac_addresses    = NULL;
	m->push_button_on_going        = 2; // "2" means "not supported"

	m->power_state = INTERFACE_POWER_STATE_ON;
	m->ipv4_nr                     = 0;
	m->ipv4                        = NULL;
	m->ipv6_nr                     = 0;
	m->ipv6                        = NULL;
	m->vendor_specific_elements_nr = 0;
	m->vendor_specific_elements    = NULL;
	return true;
}

static bool _fill_interface_type(char *interface_name, struct wifi_ap *ap_info, struct interfaceInfo *m)
{
	struct wifi_radio radio;
	INT32U channel = 0;
	enum wifi_bw bw;
	INT16U itype = MEDIA_TYPE_UNKNOWN;

	memset(&radio, 0, sizeof(struct wifi_radio));
	const char *radio_name = get_ap_radio_name(interface_name);

	wifi_get_channel(radio_name, &channel, &bw);
	if (channel && channel > 14) {
		if (ap_info->bss.oper_std & WIFI_A)
			itype = MEDIA_TYPE_IEEE_802_11A_5_GHZ;

		if (ap_info->bss.oper_std & WIFI_N)
			itype = MEDIA_TYPE_IEEE_802_11N_5_GHZ;

		if (ap_info->bss.oper_std & WIFI_AC)
			itype = MEDIA_TYPE_IEEE_802_11AC_5_GHZ;
	} else {
		if (ap_info->bss.oper_std & WIFI_B)
			itype = MEDIA_TYPE_IEEE_802_11B_2_4_GHZ;

		if (ap_info->bss.oper_std & WIFI_G)
			itype = MEDIA_TYPE_IEEE_802_11G_2_4_GHZ;

		if (ap_info->bss.oper_std & WIFI_N)
			itype = MEDIA_TYPE_IEEE_802_11N_2_4_GHZ;
	}
	wifi_get_bandwidth(radio_name, &bw);
	m->interface_type = itype;
	m->interface_type_data.ieee80211.ap_channel_band = bw;
	m->interface_type_data.ieee80211.ap_channel_center_frequency_index_1 = channel;
	// TODO: This need to change for 80+80
	m->interface_type_data.ieee80211.ap_channel_center_frequency_index_2 = channel;

	return true;
}

bool fill_wifi_dummy_info(char *interface_name, struct interfaceInfo *m)
{
	DEBUG("Filling info for |%s|", interface_name);

	m->interface_type_data.ieee80211.role = IEEE80211_ROLE_AP;

	// Retrieve SSID information
	char ssid[64] = {0};

	strncpy(m->interface_type_data.ieee80211.ssid, ssid, 64);

	// Not working band is always reported as 2_4
	m->interface_type = MEDIA_TYPE_IEEE_802_11N_2_4_GHZ;

	// encryption
	m->is_secured = 1;
	m->interface_type_data.ieee80211.encryption_mode     = IEEE80211_ENCRYPTION_MODE_TKIP |
		IEEE80211_ENCRYPTION_MODE_AES;

	m->interface_type_data.ieee80211.authentication_mode = IEEE80211_AUTH_MODE_WPA2PSK;
	// bssid should be retrived from db as 00:00:00:00:00:00 needed for unconfigured devices
	// this will get updated on receiving M2
	INT8S index = get_metadata_index(interface_name);

	strcpy(m->interface_type_data.ieee80211.network_key, gp_interface_info[index].network_key);
	memcpy(&m->interface_type_data.ieee80211.bssid, &gp_interface_info[index].bssid, 6);

	// Retrieve list of connected devices
	m->neighbor_mac_addresses_nr = 0;
	m->neighbor_mac_addresses    = NULL;

	// wps status
	m->push_button_on_going = 0; // "2" means "not supported"

	// Power state of interface
	m->power_state = INTERFACE_POWER_STATE_ON;

	m->ipv4_nr                     = 0;
	m->ipv4                        = NULL;
	m->ipv6_nr                     = 0;
	m->ipv6                        = NULL;
	m->vendor_specific_elements_nr = 0;
	m->vendor_specific_elements    = NULL;

	return true;
}

bool fill_wifi_ap_info(char *interface_name, struct wifi_ap *ap_info, struct interfaceInfo *m)
{
	DEBUG("Filling info for |%s|", interface_name);

	m->interface_type_data.ieee80211.role = IEEE80211_ROLE_AP;

	// Retrieve SSID information
	char ssid[64] = {0};

	wifi_get_ssid(interface_name, ssid);
	strncpy(m->interface_type_data.ieee80211.ssid, ssid, 64);

	// Not working band is always reported as 2_4
	_fill_interface_type(interface_name, ap_info, m);

	// encryption
	m->is_secured = 1;
	m->interface_type_data.ieee80211.encryption_mode = IEEE80211_ENCRYPTION_MODE_NONE;

	if (ap_info->bss.rsn.pair_ciphers == WIFI_CIPHER_AES)
		m->interface_type_data.ieee80211.encryption_mode |= IEEE80211_ENCRYPTION_MODE_AES;

	if (ap_info->bss.rsn.pair_ciphers == WIFI_CIPHER_TKIP)
		m->interface_type_data.ieee80211.encryption_mode |= IEEE80211_ENCRYPTION_MODE_TKIP;

	m->interface_type_data.ieee80211.authentication_mode = 0;

	if (ap_info->bss.security & (1 << WIFI_SECURITY_WPA))
		m->interface_type_data.ieee80211.authentication_mode |= IEEE80211_AUTH_MODE_WPA;
	else if (ap_info->bss.security & (1 << WIFI_SECURITY_WPAPSK))
		m->interface_type_data.ieee80211.authentication_mode |= IEEE80211_AUTH_MODE_WPAPSK;
	else if (ap_info->bss.security & (1 << WIFI_SECURITY_WPA2))
		m->interface_type_data.ieee80211.authentication_mode |= IEEE80211_AUTH_MODE_WPA2;
	else if (ap_info->bss.security & (1 << WIFI_SECURITY_WPA2PSK))
		m->interface_type_data.ieee80211.authentication_mode |= IEEE80211_AUTH_MODE_WPA2PSK;

	/*
	ERROR("## ap_info(%s) bss rsn akms (0x%x) ##", interface_name, ap_info->bss.rsn.akms);
	if (ap_info->bss.rsn.akms & WIFI_AKM_DOT1X)
		m->interface_type_data.ieee80211.authentication_mode |= IEEE80211_AUTH_MODE_WPA;
	else if (ap_info->bss.rsn.akms & WIFI_AKM_PSK)
		m->interface_type_data.ieee80211.authentication_mode |= IEEE80211_AUTH_MODE_WPAPSK;
	else if (ap_info->bss.rsn.akms & WIFI_AKM_FT_DOT1X)
		m->interface_type_data.ieee80211.authentication_mode |= IEEE80211_AUTH_MODE_WPA2;
	else if (ap_info->bss.rsn.akms & WIFI_AKM_FT_PSK)
		m->interface_type_data.ieee80211.authentication_mode |= IEEE80211_AUTH_MODE_WPA2PSK;

	*/
	if (m->interface_type_data.ieee80211.authentication_mode == 0) {
		m->is_secured = 0;
	}

	// bssid should be retrived from db as 00:00:00:00:00:00 needed for unconfigured devices
	// this will get updated on receiving M2
	INT8S index = get_metadata_index(interface_name);

	strcpy(m->interface_type_data.ieee80211.network_key, gp_interface_info[index].network_key);
	memcpy(&m->interface_type_data.ieee80211.bssid, &gp_interface_info[index].bssid, 6);

	// Retrieve list of connected devices
	m->neighbor_mac_addresses    = get_neigh_wifi(interface_name, &m->neighbor_mac_addresses_nr);

	// wps status
	m->push_button_on_going = 0; // "2" means "not supported"
	enum wps_status wps_code = 0;

	wifi_get_wps_status(interface_name, &wps_code);
	if (wps_code == WPS_STATUS_PROCESSING)
		m->push_button_on_going        = 1;

	// Power state of interface
	m->power_state = INTERFACE_POWER_STATE_ON;

	/* TODO: This code is not working
	 * ifopstatus_t opstatus;
	 * wifi_get_ifoperstatus(interface_name, &opstatus);
	 * if(IF_OPER_UP != opstatus)
	 *	m->power_state = INTERFACE_POWER_STATE_OFF;
	 */

	m->ipv4_nr                     = 0;
	m->ipv4                        = NULL;
	m->ipv6_nr                     = 0;
	m->ipv6                        = NULL;
	m->vendor_specific_elements_nr = 0;
	m->vendor_specific_elements    = NULL;

	return true;
}

bool platform_update_bssid(char *ifname)
{
	struct wifi_ap ap_details;

	memset(&ap_details, 0, sizeof(struct wifi_ap));

	if (wifi_ap_info(ifname, &ap_details) == 0) {
		INT8S index = get_metadata_index(ifname);

		memcpy(&gp_interface_info[index].bssid, &ap_details.bss.bssid, 6);
		return true;
	}
	return false;
}

bool platform_clear_bssid(char *ifname)
{
	INT8S index = get_metadata_index(ifname);

	uci_del_wireless_interface_sec(UCI_IEEE1905, ifname);

	if (-1 == index)
		return false;

	memset(&gp_interface_info[index].bssid, 0, 6);

	return true;
}

bool fill_wifi_interface_info(char *interface_name, struct interfaceInfo *m)
{
	struct wifi_ap ap_details;
	struct wifi_sta sta;

	memset(&ap_details, 0, sizeof(struct wifi_ap));
	memset(&sta, 0, sizeof(struct wifi_sta));

	if (wifi_ap_info(interface_name, &ap_details) == 0)
		return fill_wifi_ap_info(interface_name, &ap_details, m);
	else if (wifi_sta_info(interface_name, &sta) == 0)
		return fill_wifi_sta_info(interface_name, m);

	fill_wifi_dummy_info(interface_name, m);
	DEBUG("Error in fetching AP/STA info, filling dummy info");
	return false;
}

void copy_ip(struct sockaddr_in *addr, struct sockaddr_in *dhcp, struct _ipv4 *ip)
{
	INT8U *p = NULL;
	INT8U *p1 = NULL;
	INT8U i = 0;

	p = (INT8U *)&(addr->sin_addr.s_addr);
	p1 = (INT8U *)&(dhcp->sin_addr.s_addr);
	for (i = 0; i < 4; i++) {
		ip->address[i] = *p;
		ip->dhcp_server[i] = *p1;
		p++; p1++;
	}
}

struct _ipv4 *get_ipv4(char *ifname, INT8U *ipv4_nr)
{
	INT8U count = 0;
	struct _ipv4 *ipv4_data = NULL;
	struct ifaddrs *ifaddr, *ifa_it;

	if (getifaddrs(&ifaddr) == -1) {
		ERROR("getifaddrs failed");
		return NULL;
	}

	for (ifa_it = ifaddr; ifa_it != NULL; ifa_it = ifa_it->ifa_next) {
		if (strncmp(ifname, ifa_it->ifa_name, strlen(ifname)) != 0)
			continue;

		if (ifa_it->ifa_addr->sa_family == AF_INET) {
			ipv4_data = (struct _ipv4 *)realloc(ipv4_data, sizeof(struct _ipv4)*(count + 1));
			struct sockaddr_in *addr;
			struct sockaddr_in dhcp;

			addr = (struct sockaddr_in *)(ifa_it->ifa_addr);
			inet_aton("127.1.1.1", &(dhcp.sin_addr));
			ipv4_data[count].type = 1;
			copy_ip(addr, &dhcp, &ipv4_data[count]);
			count++;
		}
	}

	*ipv4_nr = count;
	freeifaddrs(ifaddr);

	return ipv4_data;
}

bool fill_eth_interface_info(char *interface_name, struct interfaceInfo *m)
{
	INT32S speed = 0;

	DEBUG("Filling info for |%s|", interface_name);

	speed = read_interface_param32(interface_name, "speed");
	if (speed == 100)
		m->interface_type = MEDIA_TYPE_IEEE_802_3U_FAST_ETHERNET;
	else
		m->interface_type = MEDIA_TYPE_IEEE_802_3AB_GIGABIT_ETHERNET;

	m->is_secured                                        = 1;

	//Retrieve list of connected devices
	m->neighbor_mac_addresses_nr = INTERFACE_NEIGHBORS_UNKNOWN;
	m->neighbor_mac_addresses = get_neigh(interface_name, &m->neighbor_mac_addresses_nr);

	m->push_button_on_going        = 2; // "2" means "not supported"
	char *state = read_interface_param(interface_name, "operstate");

	if (state && strncasecmp(state, "up", 2) == 0)
		m->power_state                 = INTERFACE_POWER_STATE_ON;
	else
		m->power_state                 = INTERFACE_POWER_STATE_OFF;

	if (state)
		FREE(state);

	m->ipv4_nr                     = 0;
	m->ipv4				= get_ipv4(interface_name, &m->ipv4_nr);
	//m->ipv4				= NULL;

	m->ipv6_nr			= 0;
	m->ipv6				= NULL;
	m->vendor_specific_elements_nr = 0;
	m->vendor_specific_elements    = NULL;

	return 1;
}

INT8U get_interface_type(char *interface_name)
{
	INT8U ret = INTF_TYPE_UNKNOWN;

	if (!(gp_interface_info && gtotal_defined_interfaces)) {
		ERROR("Al-interface empty or not defined");
		return ret;
	}
	if (interface_name == NULL) {
		ERROR("Interface name is null");
		return ret;
	}
	size_t iflen = strlen(interface_name);

	for (size_t i = 0; i < gtotal_defined_interfaces; ++i) {
		if (strncmp(interface_name, gp_interface_info[i].name, iflen) == 0) {
			ret = gp_interface_info[i].intf_type;
			break;
		}
	}
	return ret;
}

void updateIntfBasicInfo()
{
	uint8_t i;
	char temp[64] = {0};
	char key[64] = {0};
	char **ifs_names;
	INT8U ifs_nr;

	ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);

	if (ifs_nr) {
		gtotal_defined_interfaces = ifs_nr;
		CALLOC(gp_interface_info, gtotal_defined_interfaces);
	}

	for (i = 0; i < ifs_nr; i++) {
		char *if_name = ifs_names[i];
		bool ret;

		strncpy(gp_interface_info[i].name, if_name, IF_NAMESIZE);

		if (check_wireless_ifname(if_name)) {
			gp_interface_info[i].intf_type = INTF_TYPE_WIFI;
			ret = uci_get_wireless_interface_option("wireless", if_name, "key", temp, sizeof(temp));
			if(ret) {
				strncat(gp_interface_info[i].network_key, temp, 64);
				ret = uci_get_wireless_interface_option("ieee1905", if_name, "key", key, sizeof(key));
				if (ret) {
					if (strncmp(temp, key, 64) == 0) {
						platform_update_bssid(if_name);
					}
				}
			}
		} else {
			gp_interface_info[i].intf_type = INTF_TYPE_ETHERNET;
		}

		get_mac_addr(if_name, gp_interface_info[i].mac_address);
	}
	PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);
}

static INT8S get_metadata_index(char *ifname)
{
	if (!(gp_interface_info && gtotal_defined_interfaces)) {
		ERROR("Al-interface empty or not defined");
		return -1;
	}
	if (!ifname) {
		ERROR("Interface name is null");
		return -1;
	}
	size_t iflen = strlen(ifname);

	for (size_t i = 0; i < gtotal_defined_interfaces; ++i) {
		if (strncmp(ifname, gp_interface_info[i].name, iflen) == 0)
			return i;

	}
	INFO("Interface metadata not found");
	return -1;
}

void free_al_intf_metadata(void)
{
	FREE(gp_interface_info);
}

int get_interface_info(char *interface_name, struct interfaceInfo *m)
{
	INT8S index = get_metadata_index(interface_name);

	if (-1 == index)
		return -1;

	memcpy(m->mac_address, gp_interface_info[index].mac_address, 6);

	switch (gp_interface_info[index].intf_type) {
	case INTF_TYPE_ETHERNET:
		{
			fill_eth_interface_info(interface_name, m);
			break;
		}
	case INTF_TYPE_WIFI:
		{
			fill_wifi_interface_info(interface_name, m);
			break;
		}
	default:
		{
			break;
		}
	}

	return 0;
}
