/*
 * platform_os.c: Generic platform wrappers
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "utils.h"
#include "utils_json.h"
#include "platform_os.h"
#include "platform_interfaces_ubus.h"
#include "al_entity.h"

#include <stdio.h>
#include <stdlib.h>		// free(), malloc(), ...
#include <string.h>		// memcpy(), memcmp(), ...
#include <pthread.h>		// threads and mutex functions
#include <errno.h>		// errno
#include <poll.h>		// poll()
#include <sys/inotify.h>	// inotify_*()
#include <unistd.h>		// read(), sleep()
#include <signal.h>
#include <easy/easy.h>
#include <net/if.h>

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <linux/if_packet.h>
#include <net/ethernet.h>

#include <libubox/blobmsg_json.h>
#include <libubox/list.h>
////////////////////////////////////////////////////////////////////////////////
// Private functions, structures and macros
////////////////////////////////////////////////////////////////////////////////

// *********** IPC stuff *******************************************************
#define IPC_QUEUE_DEPTH (100)
struct q_msg
{
	uint16_t msg_len;
	uint8_t *data;
	struct list_head list;
};

static struct q_msg *gs_qmsg_head;
static pthread_mutex_t gs_ipc_mutex;
static pthread_cond_t gs_ipc_cond = PTHREAD_COND_INITIALIZER;
static uint8_t gs_qdepth;

// *********** Packet capture stuff ********************************************

static pthread_mutex_t pcap_filters_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t pcap_filters_cond = PTHREAD_COND_INITIALIZER;
static int pcap_filters_flag;

static void dbg_dump(const char *h, unsigned char *buf, int len)
{
	int i;

	if (h && h[0])
		fprintf(stderr, "Dump: %s\n", h);

	for (i = 0; i < len; i++) {
		if (i && i % 8 == 0)
			fprintf(stderr, "\n");

		fprintf(stderr, "%02x ", buf[i]);
	}
	fprintf(stderr, "\n");
}

int lldp_rx_loop(char *ifname)
{
	int sk;
	int reuse = 1;
	struct sockaddr_ll sa;
	struct packet_mreq mr;
	unsigned char buffer[MAX_NETWORK_SEGMENT_SIZE] = { 0 };
	ssize_t rlen = 0;
	struct ifreq ifr;
	int sockfd;
	int ifindex;
	uint8_t iface_macaddr[ETH_ALEN] = { 0 };

	if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		return -1;
	}

	strncpy(ifr.ifr_name, ifname, 15);
	ifr.ifr_name[15] = '\0';
	if (ioctl(sockfd, SIOCGIFINDEX, &ifr) == -1) {
		close(sockfd);
		return -1;
	}

	ifindex = ifr.ifr_ifindex;
	printf("%s(): ifindex = %d\n", __func__, ifindex);

	ifr.ifr_addr.sa_family = AF_INET;
	if (ioctl(sockfd, SIOCGIFHWADDR, &ifr) == -1) {
		close(sockfd);
		return -1;
	}
	memcpy(iface_macaddr, ifr.ifr_hwaddr.sa_data, ETH_ALEN);
	close(sockfd);


	sk = socket(AF_PACKET, SOCK_RAW, htons(ETHERTYPE_LLDP));
	if (sk < 0) {
		fprintf(stderr, "socket err\n");
		return -1;
	}

	if (setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse)) != 0) {
		fprintf(stderr, "reuseaddr err\n");
		goto out;
	}

	sa.sll_family = PF_PACKET;
	sa.sll_protocol = htons(ETHERTYPE_LLDP);
	sa.sll_halen = ETH_ALEN;
	sa.sll_ifindex = ifindex;
	if ((bind(sk, (struct sockaddr *)&sa, sizeof(struct sockaddr_ll))) != 0) {
		fprintf(stderr, "bind() error\n");
		close(sk);
		return -1;
	}

	/* subscribe to lldp multicast */
	memset(&mr,0,sizeof(mr));
	mr.mr_ifindex = sa.sll_ifindex;
	mr.mr_type = PACKET_MR_MULTICAST;
	mr.mr_alen = ETH_ALEN;
	memcpy(mr.mr_address, "\x01\x80\xC2\x00\x00\x0E", ETH_ALEN);
	if (setsockopt(sk, SOL_PACKET, PACKET_ADD_MEMBERSHIP, &mr, sizeof(mr)) < 0) {
		fprintf(stderr, "%s: setsockopt error (%s)\n", __func__,
			strerror(errno));
		close(sk);
		return -1;
	}

	pthread_mutex_lock(&pcap_filters_mutex);
	pcap_filters_flag--;
	if (pcap_filters_flag == 0)
		pthread_cond_signal(&pcap_filters_cond);
	pthread_mutex_unlock(&pcap_filters_mutex);

	/* receive and process in a loop forever */
	do {
		uint8_t message[MAX_PCAP_QUEUE_SIZE];
		uint16_t message_len, mlen;
		uint8_t *p;

		memset(&message, 0, MAX_PCAP_QUEUE_SIZE);
		DEBUG("%s: before receive\n", __func__);
		rlen = recvfrom(sk, &message[9], sizeof(message) - 9, 0, NULL, NULL);
		if (rlen == -1) {
			fprintf(stderr, "recv err\n");
			goto out;
		}
		//dbg_dump("", &message[9], rlen);

		message_len = rlen + 9;
		mlen = (uint16_t)(6 + rlen);

		message[0] = PLATFORM_QUEUE_EVENT_NEW_1905_PACKET;
		p = &message[1];
		_I2B(&mlen, &p);
		_InB(&iface_macaddr, &p, 6);

		DEBUG("Sending %d bytes to queue (0x%02x, 0x%02x, 0x%02x, ...)",
		      message_len, message[0], message[1], message[2]);


		if (sendMessageToAlQueue(message, message_len) == 0) {
			ERROR("Error sending message to AL-Queue");
			break;
		}
	} while (1);

out:
	close(sk);
	return 0;
}

static void *_lldpLoopThread(void *p)
{
	char *ifname = (char *)p;


	if (p == NULL) {
		pthread_mutex_lock(&pcap_filters_mutex);
		pcap_filters_flag--;
		if (pcap_filters_flag == 0)
			pthread_cond_signal(&pcap_filters_cond);
		pthread_mutex_unlock(&pcap_filters_mutex);

		return NULL;
	}

	lldp_rx_loop(ifname);

	ERROR("[PLATFORM] *lldp thread* Exiting thread (interface %s)", ifname);
	return NULL;
}

int ieee1905_rx_loop(char *ifname)
{
	int sk;
	int reuse = 1;
	struct sockaddr_ll sa;
	struct packet_mreq mr;
	unsigned char buffer[MAX_NETWORK_SEGMENT_SIZE] = { 0 };
	ssize_t rlen = 0;
	struct ifreq ifr;
	int sockfd;
	int ifindex;
	uint8_t iface_macaddr[ETH_ALEN] = { 0 };

	if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		return -1;
	}

	strncpy(ifr.ifr_name, ifname, 15);
	ifr.ifr_name[15] = '\0';
	if (ioctl(sockfd, SIOCGIFINDEX, &ifr) == -1) {
		close(sockfd);
		return -1;
	}

	ifindex = ifr.ifr_ifindex;
	printf("%s(): ifindex = %d\n", __func__, ifindex);

	ifr.ifr_addr.sa_family = AF_INET;
	if (ioctl(sockfd, SIOCGIFHWADDR, &ifr) == -1) {
		close(sockfd);
		fprintf(stderr, "get hwaddr err\n");
		return -1;
	}
	memcpy(iface_macaddr, ifr.ifr_hwaddr.sa_data, ETH_ALEN);
	close(sockfd);


	sk = socket(AF_PACKET, SOCK_RAW, htons(ETHERTYPE_1905));
	if (sk < 0) {
		fprintf(stderr, "socket err\n");
		return -1;
	}

	if (setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse)) != 0) {
		fprintf(stderr, "reuseaddr err\n");
		goto out;
	}

	sa.sll_family = PF_PACKET;
	sa.sll_protocol = htons(ETHERTYPE_1905);
	sa.sll_halen = ETH_ALEN;
	sa.sll_ifindex = ifindex;
	if ((bind(sk, (struct sockaddr *)&sa, sizeof(struct sockaddr_ll))) != 0) {
		fprintf(stderr, "bind() error\n");
		close(sk);
		return -1;
	}

	/* subscribe to 1905 multicast */
	memset(&mr,0,sizeof(mr));
	mr.mr_ifindex = sa.sll_ifindex;
	mr.mr_type = PACKET_MR_MULTICAST;
	mr.mr_alen = ETH_ALEN;
	memcpy(mr.mr_address, "\x01\x80\xC2\x00\x00\x13", ETH_ALEN);
	if (setsockopt(sk, SOL_PACKET, PACKET_ADD_MEMBERSHIP, &mr, sizeof(mr)) < 0) {
		fprintf(stderr, "%s: setsockopt error (%s)\n", __func__,
			strerror(errno));
		close(sk);
		return -1;
	}

	pthread_mutex_lock(&pcap_filters_mutex);
	pcap_filters_flag--;
	if (pcap_filters_flag == 0)
		pthread_cond_signal(&pcap_filters_cond);
	pthread_mutex_unlock(&pcap_filters_mutex);

	/* receive and process in a loop forever */
	do {
		uint8_t message[MAX_PCAP_QUEUE_SIZE];
		uint16_t message_len, mlen;
		uint8_t *p;

		memset(&message, 0, MAX_PCAP_QUEUE_SIZE);
		DEBUG("%s: before receive\n", __func__);
		rlen = recvfrom(sk, &message[9], sizeof(message) - 9, 0, NULL, NULL);
		if (rlen == -1) {
			fprintf(stderr, "recv err\n");
			goto out;
		}
		//dbg_dump("", &message[9], rlen);

		message_len = rlen + 9;
		mlen = (uint16_t)(6 + rlen);

		message[0] = PLATFORM_QUEUE_EVENT_NEW_1905_PACKET;
		p = &message[1];
		_I2B(&mlen, &p);
		_InB(&iface_macaddr, &p, 6);

		DEBUG("Sending %d bytes to queue (0x%02x, 0x%02x, 0x%02x, ...)",
		      message_len, message[0], message[1], message[2]);


		if (sendMessageToAlQueue(message, message_len) == 0) {
			ERROR("Error sending message to AL-Queue");
			break;
		}
	} while (1);

out:
	close(sk);
	return 0;
}

static void *_i1905LoopThread(void *p)
{
	char *ifname = (char *)p;


	if (p == NULL) {
		pthread_mutex_lock(&pcap_filters_mutex);
		pcap_filters_flag--;
		if (pcap_filters_flag == 0)
			pthread_cond_signal(&pcap_filters_cond);
		pthread_mutex_unlock(&pcap_filters_mutex);

		return NULL;
	}

	ieee1905_rx_loop(ifname);

	ERROR("[PLATFORM] *1905 thread* Exiting thread (interface %s)", ifname);
	return NULL;
}

// *********** Timers stuff ****************************************************

// We use the POSIX timers API to implement PLATFORM timers
// It works like this:
//
//   - When the PLATFORM API user calls "PLATFORM_REGISTER_QUEUE_EVENT()" with
//     'PLATFORM_QUEUE_EVENT_TIMEOUT*', a new POSIX timer is created.
//
//   - When the timer expires, the POSIX API creates a thread for us and makes
//     it run function '_timerHandler()'
//
//   - '_timerHandler()' simply deletes (or reprograms, depending on the type
//     of timer) the timer and sends a message to a queue so that the user can
//     later be aware of the timer expiration with a call to
//     "PLATFORM_QUEUE_READ()"

struct _timerHandlerThreadData {
	INT32U token;
	INT8U periodic;
	timer_t timer_id;
};

static void _timerHandler(union sigval s)
{
	struct _timerHandlerThreadData *aux;

	INT8U message[3 + 4];
	INT16U packet_len, mlen;
	INT32U token;
	uint8_t *p;

	aux = (struct _timerHandlerThreadData *)s.sival_ptr;

	// In order to build the message that will be inserted into the queue, we
	// need to follow the "message format" defines in the documentation of
	// function 'PLATFORM_REGISTER_QUEUE_EVENT()'
	//
	packet_len = 4 + 3;
	mlen = 4;
	token = aux->token;

	if (aux->periodic == 1)
		message[0] = PLATFORM_QUEUE_EVENT_TIMEOUT_PERIODIC;
	else
		message[0] = PLATFORM_QUEUE_EVENT_TIMEOUT;

	p = &message[1];
	_I2B(&mlen, &p);
	_I4B(&token, &p);

	DEBUG("[PLATFORM] *Timer handler* Sending %d bytes to queue (%02x, %02x, %02x, ...)", packet_len, message[0], message[1], message[2]);

	if (sendMessageToAlQueue(message, packet_len) == 0) {
		ERROR("Error sending message to AL-Queue");
	}

	if (aux->periodic == 1) {
		// Periodic timer are automatically re-armed. We don't need to do
		// anything
	} else {
		// Delete the asociater timer
		//
		timer_delete(aux->timer_id);

		// Free 'struct _timerHandlerThreadData', as we don't need it any more
		//
		free(aux);
	}

	return;
}

void sendPushButtonEvent(char *buff)
{
	json_object *jobj = NULL;
	char *temp;

	jobj = json_tokener_parse((const char *)buff);
	if (jobj == NULL)
		return;

	temp = _json_object_get_string(jobj, "action");
	if (BOOL_STRCMP("released", temp)) {
		INT8U message[3] = { PLATFORM_QUEUE_EVENT_PUSH_BUTTON, 0x0, 0x0 };

		DEBUG("*Push button event* Sending 3 bytes to queue (0x%02x, 0x%02x, 0x%02x)", message[0], message[1], message[2]);

		if (sendMessageToAlQueue(message, 3) == 0) {
			ERROR("Error sending message to AL-Queue");
		}
	}
	FREE(temp);
	json_safe_put(jobj);
}

void sendTopologyChangeEvent(char *buff)
{
	json_object *jobj;
	char *mac_str, *vif, *action;
	INT8U mac[6] = {0, 0, 0, 0, 0, 0};
	INT8U message[3] = {PLATFORM_QUEUE_EVENT_TOPOLOGY_CHANGE_NOTIFICATION, 0x0, 0x0};

	jobj = json_tokener_parse((const char *) buff);
	if (jobj == NULL) {
		ERROR("Failed to decode topology change(%s)", buff);
		return;
	}

	mac_str = _json_object_get_string(jobj, "macaddr");
	vif = _json_object_get_string(jobj, "vif");
	action = _json_object_get_string(jobj, "action");
	if (mac_str == NULL || vif == NULL || action == NULL) {
		ERROR("required info missing (%s) in topology change", buff);
		json_safe_put(jobj);
		FREE(mac_str);
		FREE(vif);
		FREE(action);
		return;
	}

	hwaddr_aton(mac_str, mac);
	DEBUG("sta(%s), vif(%s), action(%s)", mac_str, vif, action);
	if (0 == strcmp("associated", action)) {
		update_pbc_join_mac(vif, mac);
	}

	DEBUG("*Topology change monitor event* Sending 3 bytes to queue (0x%02x, 0x%02x, 0x%02x)", message[0], message[1], message[2]);

	if (sendMessageToAlQueue(message, 3) == 0) {
		ERROR("*Topology change monitor event* Error sending message to queue");
	}

	FREE(mac_str);
	FREE(vif);
	FREE(action);
	json_safe_put(jobj);
}

INT8U sendMessageToAlQueue(INT8U *message, INT16U message_len)
{
	struct q_msg *msg;

	if (message == NULL || message_len == 0) {
		ERROR("[PLATFORM] Invalid message");
		return 1;
	}

	if (pthread_mutex_lock(&gs_ipc_mutex) != 0) {
		ERROR("Failed to lock the mutex, exiting");
		return 0;
	}

	// check if queue is available
	if (gs_qmsg_head == NULL) {
		ERROR("Queue head not available exiting");
		return 0;
	}

	if (gs_qdepth == IPC_QUEUE_DEPTH) {
		WARNING("## IPC queue full, ignoring packets ##");
		pthread_mutex_unlock(&gs_ipc_mutex);
		return 1;
	}

	msg = calloc(1, sizeof(*msg));
	if (msg == NULL) {
		ERROR("Out of memory !!");
		pthread_mutex_unlock(&gs_ipc_mutex);
		return 0;
	}
	msg->msg_len = message_len;
	msg->data = malloc(sizeof(uint8_t) * message_len);
	if (msg->data == NULL) {
		ERROR("Out of memory !");
		free(msg);
		pthread_mutex_unlock(&gs_ipc_mutex);
		return 0;
	}
	memcpy(msg->data, message, message_len);

	INIT_LIST_HEAD(&msg->list);
	list_add_tail(&msg->list, &gs_qmsg_head->list);
	gs_qdepth++;

	pthread_cond_signal(&gs_ipc_cond);
	pthread_mutex_unlock(&gs_ipc_mutex);

	return 1;
}

////////////////////////////////////////////////////////////////////////////////
// Platform API: Device information functions to be used by platform-independent
// files (functions declarations are  found in "../interfaces/platform_os.h)
////////////////////////////////////////////////////////////////////////////////

static struct deviceInfo sg_devInfo;

void update_device_info(char *fname, char *mname, char *mmodel, char *url)
{
	INFO("Adding fname(%s), mname(%s)", fname, mname);
	INFO("Adding mmodel(%s), url(%s)", mmodel, url);
	strncpy(sg_devInfo.friendly_name, fname, 64);
	strncpy(sg_devInfo.manufacturer_name, mname, 64);
	strncpy(sg_devInfo.manufacturer_model, mmodel, 64);
	strncpy(sg_devInfo.control_url, url, 64);
}

struct deviceInfo *PLATFORM_GET_DEVICE_INFO(void)
{
	return &sg_devInfo;
}

////////////////////////////////////////////////////////////////////////////////
// Platform API: IPC related functions to be used by platform-independent
// files (functions declarations are  found in "../interfaces/platform_os.h)
////////////////////////////////////////////////////////////////////////////////
void PLATFORM_CLEAN_QUEUE(void)
{
	struct q_msg *entry, *node;

	if (gs_qmsg_head == NULL)
		return;

	pthread_mutex_lock(&gs_ipc_mutex);
        list_for_each_entry_safe(entry, node, &gs_qmsg_head->list, list) {
		free(entry->data);
		list_del(&entry->list);
		free(entry);
        }
	FREE(gs_qmsg_head);
	pthread_mutex_unlock(&gs_ipc_mutex);
	pthread_mutex_destroy(&gs_ipc_mutex);
}

bool PLATFORM_CREATE_QUEUE(void)
{
	CALLOC(gs_qmsg_head, 1);
	INIT_LIST_HEAD(&gs_qmsg_head->list);
	pthread_mutex_init(&gs_ipc_mutex, NULL);
	return true;
}

INT8U PLATFORM_REGISTER_QUEUE_EVENT(INT8U event_type, void *data)
{
	switch (event_type) {
	case PLATFORM_QUEUE_EVENT_NEW_1905_PACKET:
		{
			struct event1905Packet *p1;
			pthread_t i1905_thread;
			pthread_t lldp_thread;
			int num_thrs = 2;

			if (data == NULL)
				return 0;

			p1 = (struct event1905Packet *)data;

			pthread_mutex_lock(&pcap_filters_mutex);
			pcap_filters_flag = num_thrs;
			pthread_mutex_unlock(&pcap_filters_mutex);

			if (pthread_create(&i1905_thread, NULL, _i1905LoopThread, (void *)p1->interface_name) != 0) {
				fprintf(stderr, "Failed to create sniffer i1905 thread(%s)!!!", p1->interface_name);
				return 0;
			}

			add_tid_to_list(i1905_thread);

			if (pthread_create(&lldp_thread, NULL, _lldpLoopThread, (void *)p1->interface_name) != 0) {
				fprintf(stderr, "Failed to create lldp thread(%s)!!!", p1->interface_name);
				return 0;
			}

			add_tid_to_list(lldp_thread);

			pthread_mutex_lock(&pcap_filters_mutex);
			while (pcap_filters_flag > 0) {
				pthread_cond_wait(&pcap_filters_cond, &pcap_filters_mutex);
			}
			pthread_mutex_unlock(&pcap_filters_mutex);
			fprintf(stderr, "Main thread resumed...\n");

			break;
		}

	case PLATFORM_QUEUE_EVENT_TIMEOUT:
	case PLATFORM_QUEUE_EVENT_TIMEOUT_PERIODIC:
		{
			struct eventTimeOut *p1;
			struct _timerHandlerThreadData *p2;

			struct sigevent se;
			struct itimerspec its;
			timer_t timer_id;

			p1 = (struct eventTimeOut *)data;

			if (p1->token > MAX_TIMER_TOKEN) {
				// Invalid arguments
				//
				return 0;
			}

			p2 = (struct _timerHandlerThreadData *)malloc(sizeof(struct _timerHandlerThreadData));
			if (p2 == NULL) {
				// Out of memory
				//
				return 0;
			}

			p2->token = p1->token;
			p2->periodic = PLATFORM_QUEUE_EVENT_TIMEOUT_PERIODIC == event_type ? 1 : 0;

			// Next, create the timer. Note that it will be automatically
			// destroyed (by us) in the callback function
			//
			memset(&se, 0, sizeof(se));
			se.sigev_notify = SIGEV_THREAD;
			se.sigev_notify_function = _timerHandler;
			se.sigev_value.sival_ptr = (void *)p2;

			// TODO: delete timers on exit
			if (-1 == timer_create(CLOCK_REALTIME, &se, &timer_id)) {
				// Failed to create a new timer
				//
				free(p2);
				return 0;
			}
			p2->timer_id = timer_id;

			// Finally, arm/start the timer
			//
			its.it_value.tv_sec = p1->timeout_ms / 1000;
			its.it_value.tv_nsec = (p1->timeout_ms % 1000) * 1000000;
			its.it_interval.tv_sec = PLATFORM_QUEUE_EVENT_TIMEOUT_PERIODIC == event_type ? its.it_value.tv_sec : 0;
			its.it_interval.tv_nsec = PLATFORM_QUEUE_EVENT_TIMEOUT_PERIODIC == event_type ? its.it_value.tv_nsec : 0;

			if (timer_settime(timer_id, 0, &its, NULL) != 0) {
				// Problems arming the timer
				//
				free(p2);
				timer_delete(timer_id);
				return 0;
			}

			break;
		}

	case PLATFORM_QUEUE_EVENT_AUTHENTICATED_LINK:
		{
			// The AL entity is telling us that it is capable of processing
			// "authenticated link" events.
			//
			// We don't really need to do anything here. The interface specific
			// thread will be created when the AL entity calls the
			// "PLATFORM_START_PUSH_BUTTON_CONFIGURATION()" function.

			break;
		}

	default:
		{
			// Unknown event type!!
			//
			return 0;
		}
	}

	return 1;
}

INT8U PLATFORM_READ_QUEUE(INT8U *message_buffer)
{
	ssize_t len;
	INT16U payload_len;
	uint8_t *p;
	struct q_msg *msg;


	pthread_mutex_lock(&gs_ipc_mutex);
	while (list_empty(&gs_qmsg_head->list))
		pthread_cond_wait(&gs_ipc_cond, &gs_ipc_mutex);

	msg = list_first_entry(&gs_qmsg_head->list, struct q_msg, list);
	if (!msg)
		return 0;

	len = msg->msg_len;
	memcpy(message_buffer, msg->data, len);
	free(msg->data);
	list_del(&msg->list);
	free(msg);

	if (gs_qdepth)
		gs_qdepth--;

	pthread_mutex_unlock(&gs_ipc_mutex);

	// All messages are TLVs where the second and third bytes indicate the
	// total length of the payload. This value *must* match "len-3"
	if (len < 3) {
		ERROR("[PLATFORM] msg_len less than 3 bytes (minimum TLV size)");
		return 0;
	} else {
		p = &message_buffer[1];

		DEBUG("Receiving %d bytes from queue (%02x, %02x, %02x, ...)",
		      len, message_buffer[0], message_buffer[1], message_buffer[2]);

		_E2B(&p, &payload_len);

		if (payload_len != len - 3) {
			ERROR("[PLATFORM] msg_len %d bytes, but the TLV is %d bytes", len, payload_len + 3);
			return 0;
		}
	}

	return 1;
}
