/*
 * plugin_mngr.c: Plugin manager to support map-plugin
 *
 * Author: Vivek Dutta <vivek.dutta@iopsys.eu>
 *
 * Copyright (C) 2020, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "plugin_mngr.h"
#include "1905_cmdus.h"
#include "1905_ubus.h"
#include "utils_json.h"
#include "platform_interfaces.h"
#include "al_datamodel.h"
#include "al_send.h"
#include "1905_tlvs.h"

#include <libubox/blobmsg_json.h>
#include <easy/easy.h>
#include <dlfcn.h>

struct module_db {
	enum map_profile profile;
	INT8U role_bit;
};

static struct module_db sg_registered_module_data = {MAP_PROFILE_1, MAP_ROLE_AGENT};

// Public functions
void update_module_db(bool set, INT8U role_bit, INT8U profile)
{
	if (set) {
		INFO("Set role|%d|, profile|%d|", role_bit, profile);
		sg_registered_module_data.profile = profile;
		sg_registered_module_data.role_bit |= role_bit;
	} else {
		sg_registered_module_data.profile = MAP_PROFILE_1;
		sg_registered_module_data.role_bit ^= role_bit;
	}
}

INT8U get_map_registered_mode(void)
{
	return sg_registered_module_data.role_bit;
}

INT8U get_map_profile_type(void)
{
	return sg_registered_module_data.profile;
}

static struct plugin *s_map_plugin;
////////////////////////////////////////////////////////////////////////////////
// Public functions (CMDU Rx/Tx callback processing).
////////////////////////////////////////////////////////////////////////////////
bool installPlugin(char *path, char *libname)
{
	void *handle;
	char libpath[256] = { 0 };
	bool ret = true;

	INFO("Installing multiAP plugin(%s/%s)", path, libname);
	snprintf(libpath, 256, "%s/%s.so", path, libname);
	handle = dlopen(libpath, RTLD_GLOBAL | RTLD_NOW);
	if (handle == NULL) {
		ERROR("MultiAP dl open failed |%s|", dlerror());
		handle = NULL;
		return false;
	}

	s_map_plugin = dlsym(handle, libname);
	if (!s_map_plugin) {
		WARNING("Failed to load initialize_map sym(%s)", dlerror());
		ret = false;
	}

	s_map_plugin->handle = handle;

	INFO("Calling init of (%s)", s_map_plugin->name);
	return ret;
}

void removePlugin(void)
{
	if (s_map_plugin == NULL)
		return;

	dlclose(s_map_plugin->handle);
}

int plugin_cleanup(struct ubus_context *ctx)
{
	if (s_map_plugin == NULL)
		return 0;

	return s_map_plugin->exit(ctx);
}

static int plugin_send_blob(struct plugin_process_data *pdata)
{
	uint8_t i;
	struct cmdu_stream *c_stream;
	char *tlv_str_buff, *tlvs;
	uint16_t tlv_len = 0;
	size_t buff_len;
	struct blob_buf bb;

	c_stream = pdata->cstream;
	if (c_stream == NULL) {
		PRINT_ERR("## Invalid input cmdu stream ##");
		return false;
	}

	buff_len = 0;
	for (i = 0; i < c_stream->nbr_tlv_streams; ++i) {
		tlv_len = get_tlv_packet_len(c_stream->tlv_stream_list[i]);
		buff_len = buff_len + tlv_len;
	}

	tlvs = calloc((1 + 6 + buff_len * 2), sizeof(char));
	if (!tlvs) {
		PRINT_ERR("Insufficient memory!");
		return false;
	}

	for (i = 0; i < c_stream->nbr_tlv_streams; ++i) {
		tlv_len = get_tlv_packet_len(c_stream->tlv_stream_list[i]);
		tlv_str_buff = calloc((tlv_len * 2 + 1), sizeof(char));

		btostr(c_stream->tlv_stream_list[i], tlv_len, tlv_str_buff);
		strcat(tlvs, tlv_str_buff);
		free(tlv_str_buff);
	}
	// append end-of-message tlv
	strcat(tlvs, "000000");

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_u32(&bb, "type", c_stream->message_type);
	blobmsg_add_u32(&bb, "mid", c_stream->message_id);
	blobmsg_add_string(&bb, "ingress", pdata->intf_name);
	blobmsg_add_string(&bb, "source", pdata->mac_ascii);
	blobmsg_add_string(&bb, "tlvs", tlvs);
	free(tlvs);

	return send_blob_to_ubus_plugin_mq(&bb);
}

bool plugin_process(struct cmdu_stream *c_stream, INT8U *receiving_interface_addr, INT8U *origin)
{
	struct plugin_process_data pdata;
	char *recv_intf, mac_str[18] = { };

	if (! s_map_plugin)
		return false;

	if (c_stream == NULL) {
		ERROR("## Invalid input param ##");
		return false;
	}

	// Only broadcast messages if agent or controller is registered
	if (sg_registered_module_data.role_bit == 0) {
		INFO("Agent/Controller not registered with the plugin");
		return false;
	}

	if (receiving_interface_addr)
		recv_intf = DMmacToInterfaceName(receiving_interface_addr);
	else
		recv_intf = "lo";

	strncpy(pdata.intf_name, recv_intf, IFNAMELEN);

	if (origin)
		hwaddr_ntoa(origin, mac_str);
	else
		strcpy(mac_str, "00:00:00:00:00:00");

	strncpy(pdata.mac_ascii, mac_str, 18);

	pdata.cstream = c_stream;
	pdata.profile = sg_registered_module_data.profile;
	pdata.flag = sg_registered_module_data.role_bit & MAP_ROLE_CONTROLLER;

	//if (s_map_plugin->ops[PLUGIN_FUNC_VALIDATE](&pdata))
		plugin_send_blob(&pdata);

	return true;
}

/* in below ubus call, data is handled in below function
 * ubus call 1905.multiap send '{"type":2,"out_intf":"br-lan","destination":"02:22:07:9f:b7:ff"}'
 * '{"type":7,"out_intf":"br-lan","destination":"02:22:07:9f:b7:ff","data":"010006022207765c570d0001000e000100}'
 * ubus call 1905.multiap send_array
 * '{"type":7,"out_intf":"br-lan","destination":"02:22:07:9f:b7:ff","tlv_array":["010006022207765c57","0d000100","0e000100"]}'
 */
bool cmdu_dispacher(struct plugin_process_data *pdata)
{
	INT8U dest[6] = {0};
	struct cmdu_stream *c_stream;
	char **ifs_names;
        INT8U ifs_nr;
	INT8U(*al_mac_addresses)[6];
        INT8U al_mac_addresses_nr;
	INT8U *almac;

	if (s_map_plugin == NULL) {
		INFO("Plugin not available");
		return false;
	}

	hwaddr_aton(pdata->mac_ascii, dest);

	almac = DMalMacGet();

	c_stream = (struct cmdu_stream *)pdata->cstream;
	if (pdata->flag)
		c_stream->message_id = getNextMid();

	// Loop-back the cmdu if the destination address is of local AL or is of
	// one of the local registered interface or empty
	if (is_local_mac(dest))
		return plugin_process(pdata->cstream, almac, almac);

	// If cmdu is destined for ieee1905 mcast, also loop-back the cmdu
	if (is_ieee1905_mcast_mac(dest))
		plugin_process(pdata->cstream, almac, almac);

	// Send the packet accross
	fragment_send_raw_packets(pdata->intf_name, dest, pdata->cstream);

	//If the packet is a reliable multicast packet then the packet
	//along with multicast also needs to be unicast
        if(pdata->rel_multicast_flag == true) {
		ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);
                for (uint8_t i = 0; i < ifs_nr; i++) {

			//Check here that the interface is info is preasent
			struct interfaceInfo *x;
			x = PLATFORM_GET_1905_INTERFACE_INFO(ifs_names[i]);
	                if (x == NULL) {
				WARNING("Could not retrieve neighbors of interface %s", ifs_names[i]);
				continue;
			}
			if (x->power_state == 0x02)
				continue;
                        //get all the neighbor mac address to the interface names
			al_mac_addresses = DMgetListOfInterfaceNeighbors(ifs_names[i], &al_mac_addresses_nr);
			for (uint8_t j = 0; j < al_mac_addresses_nr; j++) {
				//reset the relay flag
				c_stream->relay_indicator = 0;
				fragment_send_raw_packets(ifs_names[i], al_mac_addresses[j], c_stream);
			}
			FREE(al_mac_addresses);
		}
		PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);
	}
	INFO("## Map mid of sent msg |%d|", c_stream->message_id);

	return true;
}

int plugin_map_init(struct ubus_context *ctx)
{
	if (s_map_plugin == NULL)
		return 0;

	return s_map_plugin->init(ctx);
}

