/*
 * 1905_ubus.c: IEEE-1905 ubus handler
 *
 * Author: Vivek Dutta <vivek.dutta@iopsys.eu>
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include "1905_ubus.h"
#include "common.h"
#include "utils.h"
#include "1905_cmdus.h"
#include "platform_interfaces.h"
#include "al_ubus.h"
#include "1905_alme.h"
#include "al_datamodel.h"
#include "al_send.h"
#include "cmdu_visit.h"
#include "plugin_mngr.h"
#include "utils_json.h"
#include "utils_uci.h"
#include "common_defs.h"
#include "platform_interfaces.h"
#include "al_entity.h"

#include <string.h>
#include <pthread.h>

#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubus.h>
#include <easy/easy.h>

#if !defined(ALME_OVER_UBUS) && !defined(ALME_OVER_TCP)
#define ALME_OVER_UBUS
#elif defined(ALME_OVER_UBUS) && defined(ALME_OVER_TCP)
#error  "Please select one 'ALME_OVER_UBUS' or 'ALME_OVER_TCP' at a time"
#endif

#define IEEE1905_AL_OBJ "ieee1905.al."

struct ubus_thread_data {
	struct ubus_context *ctx;
	struct uloop_timeout timeout;
	struct ubus_event_handler wps;
	struct ubus_event_handler notification;
	uint8_t num_obj;
	struct ubus_object **obj_list;
	struct uloop_fd cmdu_fd;
	struct uloop_fd cmdu_plugin_fd;
	int cmdu_mq_socket[2];
	int cmdu_plugin_mq_socket[2];
};

// Global Variables
static struct ubus_thread_data sg_u;

#define cmdu_rx_socket sg_u.cmdu_mq_socket[0]
#define cmdu_tx_socket sg_u.cmdu_mq_socket[1]
#define cmdu_plugin_rx_socket sg_u.cmdu_plugin_mq_socket[0]
#define cmdu_plugin_tx_socket sg_u.cmdu_plugin_mq_socket[1]
#define ULOOP_TIMEOUT (1000)
#define INTERFACE_NAME_LEN 128

static void send_kill_signal(void)
{
	pid_t pid;

	pid = getpid();
	ERROR("Error occurred in ubus thread Sending SIGINT to self(%x) for cleanup", pid);
	kill(pid, SIGINT);
}

enum {
	SEND_CMDU_TYPE,		/* cmdu type */
	SEND_OUT_INTF,
	SEND_CMDU_DEST,
	SEND_CMDU_DATA,		/* tlv stream */
	_SEND_POLICY_MAX,
};

static const struct blobmsg_policy send_query_policy[_SEND_POLICY_MAX] = {
	[SEND_CMDU_TYPE] = { .name = "type", .type = BLOBMSG_TYPE_INT32 },
	[SEND_OUT_INTF] = { .name = "egress", .type = BLOBMSG_TYPE_STRING },
	[SEND_CMDU_DEST] = { .name = "dst_macaddr", .type = BLOBMSG_TYPE_STRING },
	[SEND_CMDU_DATA] = { .name = "data", .type = BLOBMSG_TYPE_STRING },
};

enum {
	_MAC_ADDR,
	__MAX_MAC_POLICY
};

static const struct blobmsg_policy mac_addr_policy[__MAX_MAC_POLICY] = {
	[_MAC_ADDR] = {.name = "dst_macaddr", .type = BLOBMSG_TYPE_STRING},
};

static int get_selected_neighbours(struct ubus_context *actx, struct ubus_object *obj __attribute__ ((unused)),
				   struct ubus_request_data *req, const char *method __attribute__ ((unused)),
				   struct blob_attr *msg __attribute__ ((unused)))
{
	char interface_name[128] = { '\0' };
	const char *pname;
	uint8_t len = strlen(IEEE1905_AL_OBJ);

	if (strlen(obj->name) <= len) {
		ERROR("Failed to get interface from |%s|", method);
		return UBUS_STATUS_UNKNOWN_ERROR;
	}
	pname = &obj->name[len];

	// strip last . from pname
	strncpy(interface_name, pname, 128);

	DEBUG("Requested Interface |%s|", interface_name);

	struct blob_buf bb;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	DMdumpNeighbourSelected(&bb, interface_name);

	ubus_send_reply(actx, req, bb.head);
	blob_buf_free(&bb);
	return 0;
}

static int send_query(struct ubus_context *actx, struct ubus_object *obj __attribute__((unused)),
		      struct ubus_request_data *req, const char *method __attribute__ ((unused)),
		      struct blob_attr *msg)
{
	struct blob_attr *tb[_SEND_POLICY_MAX] = { NULL };
	INT8U dmac[6];
	uint32_t cmdu_type;
	char *intf_name = NULL, *p;
	int ret = 0;
	uint16_t mid = 0;
	struct blob_buf bb;

	memset(&bb, 0, sizeof(struct blob_buf));
	memset(&dmac, 0, 6);

	ret = blobmsg_parse(send_query_policy, _SEND_POLICY_MAX, tb, blob_data(msg), blob_len(msg));
	if (ret != 0) {
		INFO("parsing failed (%s)", ubus_strerror(ret));
		return UBUS_STATUS_INVALID_ARGUMENT;
	}

	if (!tb[SEND_CMDU_TYPE]) {
		INFO("## cmdu type missing ##");
		return UBUS_STATUS_INVALID_ARGUMENT;
	}

	if (tb[SEND_CMDU_DEST]) {
	    p = blobmsg_get_string(tb[SEND_CMDU_DEST]);
	    hwaddr_aton(p, dmac);
	}

	cmdu_type = blobmsg_get_u32(tb[SEND_CMDU_TYPE]);

	if (tb[SEND_OUT_INTF])
		intf_name = blobmsg_get_string(tb[SEND_OUT_INTF]);

	blob_buf_init(&bb, 0);
	if (is_local_mac(dmac)) {
		mid = getNextMid();
		blobmsg_add_u32(&bb, "mid", mid);
		blobmsg_add_u8(&bb, "status", true);
		ubus_send_reply(actx, req, bb.head);
		blob_buf_free(&bb);

		switch (cmdu_type) {
		case CMDU_TYPE_TOPOLOGY_QUERY:
			send1905TopologyQueryPacket(intf_name, mid, dmac);
			break;
		case CMDU_TYPE_LINK_METRIC_QUERY:
			send1905MetricsResponsePacket(intf_name, mid, dmac,
				    LINK_METRIC_QUERY_TLV_ALL_NEIGHBORS, NULL,
				    LINK_METRIC_QUERY_TLV_BOTH_TX_AND_RX_LINK_METRICS);
			break;
		case CMDU_TYPE_HIGHER_LAYER_QUERY:
			send1905HighLayerResponsePacket(intf_name, mid, dmac);
			break;
		case CMDU_TYPE_GENERIC_PHY_QUERY:
			send1905GenericPhyResponsePacket(intf_name, mid, dmac);
			break;
		default:
			WARNING("Cmdu type (0x%02x) not supported!!", cmdu_type);
		}

	} else {
		if (intf_name == NULL) {
			INFO("## intf_name missing ##");
			return UBUS_STATUS_INVALID_ARGUMENT;
		}

		mid = getNextMid();
		blobmsg_add_u32(&bb, "mid", mid);
		switch (cmdu_type) {
		case CMDU_TYPE_TOPOLOGY_QUERY:
			ret = send1905TopologyQueryPacket(intf_name, mid, dmac);
			break;
		case CMDU_TYPE_LINK_METRIC_QUERY:
			ret = send1905MetricsQueryPacket(intf_name, mid, dmac);
			break;
		case CMDU_TYPE_HIGHER_LAYER_QUERY:
			ret = send1905HighLayerQueryPacket(intf_name, mid, dmac);
			break;
		case CMDU_TYPE_GENERIC_PHY_QUERY:
			ret = send1905GenericPhyQueryPacket(intf_name, mid, dmac);
			break;
		default:
			WARNING("Cmdu type (0x%02x) not supported!!", cmdu_type);
		}

		if (ret == 1)
			blobmsg_add_u8(&bb, "status", true);
		else
			blobmsg_add_u8(&bb, "status", false);

		ubus_send_reply(actx, req, bb.head);
		blob_buf_free(&bb);
	}

	return 0;
}

static int get_remote_topology(struct ubus_context *actx, struct ubus_object *obj,
			       struct ubus_request_data *req, const char *method __attribute__ ((unused)), struct blob_attr *msg)
{
	struct blob_attr *tb[__MAX_MAC_POLICY] = { NULL };
	char interface_name[128] = { '\0' };
	INT8U dmac[6];
	INT16U mid = 0;
	uint8_t len = strlen(IEEE1905_AL_OBJ);
	const char *pname;
	char *p;

	if (strlen(obj->name) > len) {
		pname = &obj->name[len];
	} else {
		pname = &obj->name[0];
	}

	// strip last . from pname
	strncpy(interface_name, pname, 128);

	memset(&dmac, 0, 6);
	if (blobmsg_parse(mac_addr_policy, __MAX_MAC_POLICY, tb, blob_data(msg), blob_len(msg)) == 0) {
		if (tb[_MAC_ADDR]) {
			p = blobmsg_get_string(tb[_MAC_ADDR]);
			hwaddr_aton(p, dmac);
		}
	}

	struct blob_buf bb;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	mid = getNextMid();
	blobmsg_add_u32(&bb, "mid", mid);

	// Check if the dmac is local al mac
	if (is_local_mac(dmac)) {
		blobmsg_add_u8(&bb, "status", true);
		ubus_send_reply(actx, req, bb.head);
		struct cmdu_cstruct res;

		if (prepare1905TopologyResponsePacket(&res, mid)) {
			BroadcastCMDU(&res, NULL, NULL);
			free_tlv_list_CMDU_structure(&res);
		}

	} else {
		uint8_t *remote_al = DMmacToAlMac(dmac);
		if (remote_al == NULL) {
			WARNING("Either destination mac not present or its non-1905 device");
			blobmsg_add_u8(&bb, "status", false);
		} else {
			if (DMinterfaceNameToMac(interface_name) == NULL) {
				uint8_t interface_num = 0;
				char **interfaces = PLATFORM_GET_LIST_OF_1905_INTERFACES(&interface_num);
				for (uint8_t i = 0; i < interface_num; i++) {
					send1905TopologyQueryPacket(interfaces[i], mid, remote_al);
				}
				PLATFORM_FREE_LIST_OF_1905_INTERFACES(interfaces, interface_num);
				blobmsg_add_u8(&bb, "status", true);
			} else {
				if (send1905TopologyQueryPacket(interface_name, mid, remote_al) == 0) {
					blobmsg_add_u8(&bb, "status", false);
					WARNING("Could not send 'topology query' message");
				} else {
					blobmsg_add_u8(&bb, "status", true);
				}
			}
		}
		ubus_send_reply(actx, req, bb.head);
		FREE(remote_al);
	}
	blob_buf_free(&bb);

	return 0;
}

static int ieee1905_interface_info(struct ubus_context *actx, struct ubus_object *obj,
				   struct ubus_request_data *req, const char *method __attribute__ ((unused)),
				   struct blob_attr *msg __attribute__ ((unused)))
{
	char interface[128] = { '\0' };
	const char *pname;
	uint8_t len = strlen(IEEE1905_AL_OBJ);

	if (strlen(obj->name) <= len) {
		ERROR("Failed to get interface from |%s|", method);
		return UBUS_STATUS_UNKNOWN_ERROR;
	}
	pname = &obj->name[len];

	// strip last . from pname
	strncpy(interface, pname, 128);

	struct blob_buf bb;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	DEBUG("Filling interface|%s| info", interface);
	fill_interface_info(interface, &bb);
	ubus_send_reply(actx, req, bb.head);
	blob_buf_free(&bb);
	return 0;
}

static int ieee1905_all_interface_info(struct ubus_context *actx, struct ubus_object *obj __attribute__ ((unused)),
				struct ubus_request_data *req, const char *method __attribute__ ((unused)),
				struct blob_attr *msg __attribute__ ((unused)))
{

	struct blob_buf bb;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	fill_interface_info_all(&bb);

	ubus_send_reply(actx, req, bb.head);
	blob_buf_free(&bb);
	return 0;
}

static int ieee1905_link_info(struct ubus_context *actx, struct ubus_object *obj,
                                   struct ubus_request_data *req, const char *method __attribute__ ((unused)),
                                   struct blob_attr *msg __attribute__ ((unused)))
{
        char interface[128] = { '\0' };
	const char *pname;
	void *array;
	uint8_t len = strlen(IEEE1905_AL_OBJ);

	if (strlen(obj->name) <= len) {
		ERROR("Failed to get interface from |%s|", method);
		return UBUS_STATUS_UNKNOWN_ERROR;
	}
	pname = &obj->name[len];

	strncpy(interface, pname, 128);

        struct blob_buf bb;

        memset(&bb, 0, sizeof(struct blob_buf));
        blob_buf_init(&bb, 0);

        DEBUG("Filling interface|%s| info", interface);

	array = blobmsg_open_array(&bb, "links");
	DMFillNeighbourLinkSelected(&bb, interface);
	blobmsg_close_array(&bb, array);
        //fill_metric_info(interface, &bb);
        ubus_send_reply(actx, req, bb.head);
        blob_buf_free(&bb);
        return 0;
}


enum {
	ALME_CMD,
	ALME_DATA,
	__ALME_MAX
};

static const struct blobmsg_policy ieee1905_alme_policy[__ALME_MAX] = {
	[ALME_CMD] = {.name = "cmd", .type = BLOBMSG_TYPE_STRING},
	[ALME_DATA] = {.name = "data", .type = BLOBMSG_TYPE_STRING},
};

static int ieee1905_alme_handle(struct ubus_context *actx, struct ubus_object *obj __attribute__ ((unused)),
			 struct ubus_request_data *req, const char *method __attribute__ ((unused)), struct blob_attr *msg)
{
	struct blob_attr *tb[__ALME_MAX] = { NULL };
	char *alme_cmd = NULL;
	char *data = NULL;
	INT8U *alme_tlv = NULL;
	INT8U i = 0;
	void *array = NULL;

	struct blob_buf bb;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

#ifdef ALME_OVER_TCP
	blobmsg_add_u8(&bb, "status", false);
	blobmsg_add_string(&&b, "reason", "ALME supported over TCP, recompile with alme ubus support");
#else
	if (blobmsg_parse(ieee1905_alme_policy, __ALME_MAX, tb, blob_data(msg), blob_len(msg))) {
		array = blobmsg_open_array(&bb, "cmd(s)");
		for (i = ALME_TYPE_GET_INTF_LIST_REQUEST; i < ALME_TYPE_GET_METRIC_RESPONSE; i += 2) {
			blobmsg_add_string(&bb, NULL, convert_1905_ALME_type_to_string(i));
		}
		blobmsg_add_string(&bb, NULL, convert_1905_ALME_type_to_string(ALME_TYPE_CUSTOM_COMMAND_REQUEST));
		blobmsg_close_array(&bb, array);
		ubus_send_reply(actx, req, bb.head);
		blob_buf_free(&bb);
		return 0;
	}
	set_alme_blob(&bb);

	if (tb[ALME_CMD])
		alme_cmd = (char *)blobmsg_get_string(tb[ALME_CMD]);
	else
		return UBUS_STATUS_INVALID_ARGUMENT;

	if (tb[ALME_DATA])
		data = (char *)blobmsg_get_string(tb[ALME_DATA]);

	alme_tlv = build_alme_request(alme_cmd, data);

	if (alme_tlv)
		process1905Alme(alme_tlv);
#endif
	ubus_send_reply(actx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

static int get_local_interfaces(struct ubus_context *actx, struct ubus_object *obj __attribute__ ((unused)),
				struct ubus_request_data *req, const char *method __attribute__ ((unused)),
				struct blob_attr *msg __attribute__ ((unused)))
{

	char **interfaces = NULL;
	INT8U interface_num = 0;
	INT8U i = 0;
	void *array = NULL;

	interfaces = uci_get_list_of_1905_interfaces(&interface_num);
	struct blob_buf bb;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_u32(&bb, "interface_nr", interface_num);
	array = blobmsg_open_array(&bb, "names");
	for (i = 0; i < interface_num; ++i) {
		blobmsg_add_string(&bb, NULL, interfaces[i]);
	}
	blobmsg_close_array(&bb, array);
	ubus_send_reply(actx, req, bb.head);
	blob_buf_free(&bb);
	uci_free_list_of_1905_interfaces(interfaces, interface_num);
	return 0;
}

static int get_local_neighbours(struct ubus_context *actx, struct ubus_object *obj __attribute__ ((unused)),
				struct ubus_request_data *req, const char *method __attribute__ ((unused)),
				struct blob_attr *msg __attribute__ ((unused)))
{

	struct blob_buf bb;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	DMdumpNeighbour(&bb);

	ubus_send_reply(actx, req, bb.head);
	blob_buf_free(&bb);
	return 0;
}

static struct ubus_method ieee1905_common_methods[] = {
	UBUS_METHOD_NOARG("info", ieee1905_all_interface_info),
	UBUS_METHOD_NOARG("interfaces", get_local_interfaces),
	UBUS_METHOD("query_topology", get_remote_topology, mac_addr_policy),
	UBUS_METHOD("send_query", send_query, send_query_policy),
	UBUS_METHOD_NOARG("neighbors", get_local_neighbours),
	UBUS_METHOD("alme", ieee1905_alme_handle, ieee1905_alme_policy),
};

static struct ubus_method ieee1905_methods[] = {
	UBUS_METHOD_NOARG("info", ieee1905_interface_info),
	UBUS_METHOD_NOARG("neighbors", get_selected_neighbours),
	UBUS_METHOD_NOARG("link_info", ieee1905_link_info),
	UBUS_METHOD("query_topology", get_remote_topology, mac_addr_policy),
};

static struct ubus_object_type ieee1905_common = UBUS_OBJECT_TYPE(IEEE1905, ieee1905_common_methods);
static struct ubus_object_type ieee1905_cmdu = UBUS_OBJECT_TYPE(IEEE1905, ieee1905_methods);


static struct ubus_object ieee1905_common_obj = {
	.name = IEEE1905,
	.type = &ieee1905_common,
	.methods = ieee1905_common_methods,
	.n_methods = ARRAY_SIZE(ieee1905_common_methods),
};

static void event_handler(__attribute__((unused)) struct ubus_context *ctx,
		   __attribute__((unused)) struct ubus_event_handler *ev,
		   const char *type, struct blob_attr *msg)
{
	// Checks thread cancellation
	pthread_testcancel();
	ieee1905_event_handler(type, msg);
}

static void register_ubus_events(void)
{
	sg_u.wps.cb = event_handler;
	ubus_register_event_handler(sg_u.ctx, &sg_u.wps, WPS_BUTTON_EVENT);

	sg_u.notification.cb = event_handler;
	ubus_register_event_handler(sg_u.ctx, &sg_u.notification, NOTIFICATION_CAHNGE_EVENT);
}

static void uloop_periodic_run(struct uloop_timeout *t)
{
	struct blob_buf bb;
	int bytes_read;

	DEBUG("-: UBUS_HEARTBEAT(%d) :-", ULOOP_TIMEOUT);

	// Publish any pending enqued events
	bytes_read = recv(cmdu_rx_socket, &bb, sizeof(struct blob_buf), 0);
	if (bytes_read != -1 && bytes_read == sizeof(struct blob_buf)) {
		ubus_send_event(sg_u.ctx, AL_EVENT_NAME, bb.head);
		blob_buf_free(&bb);
	}

	bytes_read = recv(cmdu_plugin_rx_socket, &bb, sizeof(struct blob_buf), 0);
	if (bytes_read != -1 && bytes_read == sizeof(struct blob_buf)) {
		ubus_send_event(sg_u.ctx, "ieee1905.cmdu", bb.head);
		blob_buf_free(&bb);
	}
	// Checks thread cancellation
	pthread_testcancel();
	uloop_timeout_cancel(t);
	sg_u.timeout.cb = uloop_periodic_run;
	uloop_timeout_set(&sg_u.timeout, ULOOP_TIMEOUT);
}

static void read_cmdu_callback(struct uloop_fd *u, unsigned int events __attribute__((unused)))
{
	struct blob_buf bb;
	int bytes_read;

	// Checks thread cancellation
	pthread_testcancel();

	bytes_read = recv(cmdu_rx_socket, &bb, sizeof(struct blob_buf), 0);
	if (bytes_read != sizeof(struct blob_buf)) {
		return;
	}

	ubus_send_event(sg_u.ctx, AL_EVENT_NAME, bb.head);
	blob_buf_free(&bb);
}

static void read_cmdu_plugin_callback(struct uloop_fd *u, unsigned int events __attribute__((unused)))
{
	struct blob_buf bb;
	int bytes_read;

	// Checks thread cancellation
	pthread_testcancel();

	bytes_read = recv(cmdu_plugin_rx_socket, &bb, sizeof(struct blob_buf), 0);
	if (bytes_read != sizeof(struct blob_buf)) {
		return;
	}

	ubus_send_event(sg_u.ctx, "ieee1905.cmdu", bb.head);
	blob_buf_free(&bb);
}

static int init_ubus_mq()
{
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, sg_u.cmdu_mq_socket) == -1) {
		ERROR("Failed to create ubus mq socketpair");
		return 1;
	}

	if (socketpair(AF_UNIX, SOCK_STREAM, 0, sg_u.cmdu_plugin_mq_socket) == -1) {
		ERROR("Failed to create ubus mq socketpair");
		return 1;
	}

	return 0;
}


static int ieee1905_ubus_add_interfaces(void)
{
	uint8_t num_intf, index, j;
	char **interfaces;
	char buff[INTERFACE_NAME_LEN];

	interfaces = uci_get_list_of_1905_interfaces(&num_intf);

	if (interfaces == NULL || num_intf == 0) {
		INFO("al-iface not configured/enabled in uci");
		return 1;
	}

	sg_u.obj_list = (struct ubus_object **)calloc(num_intf, sizeof(struct ubus_object *));
	if (sg_u.obj_list == NULL) {
		ERROR("Failed to allocate memory for ubus object");
		return 1;
	}

	for (index = 0; index < num_intf; ++index) {
		sg_u.obj_list[index] = (struct ubus_object *) calloc(1, sizeof(struct ubus_object));
		if (sg_u.obj_list[index] == NULL) {
			ERROR("Out of memory !!!");

			for (j = index; j != 0; j--)
				FREE(sg_u.obj_list[j]);

			FREE(sg_u.obj_list);
			uci_free_list_of_1905_interfaces(interfaces, num_intf);
			return 1;
		}
	}

	for (index = 0; index < num_intf; ++index) {
		char *obj_name;

		strcpy(buff, IEEE1905_AL_OBJ);
		strcat(buff, interfaces[index]);
		obj_name = strdup(buff);

		sg_u.obj_list[index]->name = obj_name;
		sg_u.obj_list[index]->type = &ieee1905_cmdu;
		sg_u.obj_list[index]->methods = ieee1905_methods;
		sg_u.obj_list[index]->n_methods = ARRAY_SIZE(ieee1905_methods);

		if (ubus_add_object(sg_u.ctx, sg_u.obj_list[index]) != UBUS_STATUS_OK) {
			ERROR("Failed to register ieee1905 object");
			FREE(obj_name);
			FREE(sg_u.obj_list[index]);
			return 1;
		}
		sg_u.num_obj = index;
	}

	sg_u.num_obj = num_intf;
	uci_free_list_of_1905_interfaces(interfaces, num_intf);
	return 0;
}

static void _remove_ubus_objects(void)
{
	ubus_unregister_event_handler(sg_u.ctx, &sg_u.wps);
	ubus_unregister_event_handler(sg_u.ctx, &sg_u.notification);
	ubus_remove_object(sg_u.ctx, &ieee1905_common_obj);
	for (INT8U index = 0; index < sg_u.num_obj; ++index) {
		ubus_remove_object(sg_u.ctx, sg_u.obj_list[index]);
		free((char *)sg_u.obj_list[index]->name);
		FREE(sg_u.obj_list[index]);
	}
	FREE(sg_u.obj_list);
}

static void ubus_thread_cleanup(__attribute__((unused)) void *arg)
{
	DEBUG("ubus thread cleanup called");
	// Cleanup starts from here
	if (sg_u.ctx) {
		_remove_ubus_objects();
		plugin_cleanup(sg_u.ctx);
		ubus_free(sg_u.ctx);
	}
	uloop_done();
	DEBUG("ubus thread cleanup done");
}

static int ieee1905_ubus_init(void)
{

	if (init_ubus_mq())
		return 1;

	sg_u.ctx = ubus_connect(NULL);
	if (!sg_u.ctx) {
		ERROR("Not able to create ubus connection");
		return 1;
	}

	ubus_add_uloop(sg_u.ctx);

	// Add common functionality with init
	INFO("Registering for ubus events");
	register_ubus_events();

	// Add heartbeat timer
	sg_u.timeout.cb = uloop_periodic_run;
	uloop_timeout_set(&sg_u.timeout, ULOOP_TIMEOUT);

	INFO("Registering for ubus objects");
	if (ubus_add_object(sg_u.ctx, &ieee1905_common_obj) != UBUS_STATUS_OK) {
		printf("Failed to register ieee1905 object");
		return 1;
	}

	// Add per interface object
	if(ieee1905_ubus_add_interfaces())
		return 1;

	// Add objects from plugin as well
	plugin_map_init(sg_u.ctx);

	INFO("Adding ubus mq in uloop");
	sg_u.cmdu_fd.cb = read_cmdu_callback;
	sg_u.cmdu_fd.fd = cmdu_rx_socket;
	uloop_fd_add(&sg_u.cmdu_fd, ULOOP_READ | ULOOP_EDGE_TRIGGER);

	sg_u.cmdu_plugin_fd.cb = read_cmdu_plugin_callback;
	sg_u.cmdu_plugin_fd.fd = cmdu_plugin_rx_socket;
	uloop_fd_add(&sg_u.cmdu_plugin_fd, ULOOP_READ | ULOOP_EDGE_TRIGGER);

	return 0;
}

void *create_ubus_thread(void *arg __attribute__((unused)))
{
	uloop_init();

	if (ieee1905_ubus_init()) {
		INFO("Failed to init ubus thread");
		al_cleanup();
		return NULL;
	}

	// Register cleanup method on thread cancel
	pthread_cleanup_push(ubus_thread_cleanup, NULL);
	uloop_run();
	pthread_cleanup_pop(1);

	// If ubus thread terminates, it will send SIGINT to main process
	// to start the cleanup
	ERROR("send kill signal to self for cleanup");
	send_kill_signal();

	return NULL;
}

/*
 * ownership of blob buf transffered to ubus thread if send is successful.
 */
int send_blob_to_ubus_mq(struct blob_buf *bb)
{
	int ret = 0, sent_bytes;

	sent_bytes = send(cmdu_tx_socket, bb, sizeof(struct blob_buf), 0);
	if (sent_bytes != sizeof(struct blob_buf)) {
		ERROR("error sending blob data(%d) to ubus queue", sent_bytes);
		blob_buf_free(bb);
		ret = 1;
	}

	return ret;
}

/*
 * ownership of blob buf transffered to map-plugin if send is successful.
 */
int send_blob_to_ubus_plugin_mq(struct blob_buf *bb)
{
	int ret = 0, sent_bytes;

	sent_bytes = send(cmdu_plugin_tx_socket, bb, sizeof(struct blob_buf), 0);
	if (sent_bytes != sizeof(struct blob_buf)) {
		ERROR("error sending blob data(%d) to ubus queue", sent_bytes);
		blob_buf_free(bb);
		ret = 1;
	}

	return ret;
}
