/*
 * al_recv.c: IEEE-1905 AL packet processing
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "al_ubus.h"
#include "al_send.h"
#include "al_datamodel.h"
#include "1905_cmdus.h"
#include "1905_ubus.h"
#include "platform_interfaces.h"
#include "platform_os.h"
#include "cmdu_visit.h"
#include "utils_uci.h"
#include "utils_json.h"

#include <arpa/inet.h>
#include <easy/easy.h>

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>

#define IPV4_ARR_LEN	16
#define IPV6_ARR_LEN	40

static bool sg_cmdu_event_enable;
static char gUbusRegistrar2dot4[1024];
static char gUbusRegistrar5[1024];
static char gUbusRegistrar60[1024];


void notify_another_registrar(char *recv_intf, INT8U *dst_mac, INT8U freq_band)
{
	struct ubus_context *sctx = ubus_connect(NULL);

	if (!sctx) {
		ERROR("Failed to connect with ubus");
		return;
	}

	struct blob_buf bb;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "error", "Another registrar detected!");
	blobmsg_add_string(&bb, "recv_intf", recv_intf);
	char mac[18] = {0};

	hwaddr_ntoa(dst_mac, mac);
	blobmsg_add_string(&bb, "src_mac", mac);
	blobmsg_add_string(&bb, "band", convert_band_to_string(freq_band));

	ubus_send_event(sctx, AL_EVENT_NAME, bb.head);

	blob_buf_free(&bb);
	ubus_free(sctx);

}

void ieee1905_event_handler(const char *type, struct blob_attr *msg)
{
	char *str = NULL;

	if ((msg == NULL) || (type == NULL))
		return;

	str = (char *)blobmsg_format_json_indent(msg, true, -1);

	if (str == NULL) {
		ERROR("No payload in ubus event");
		return;
	}

	if (BOOL_STRCMP(WPS_BUTTON_EVENT, type)) {
		sendPushButtonEvent(str);
	}
	if (BOOL_STRCMP(NOTIFICATION_CAHNGE_EVENT, type)) {
		sendTopologyChangeEvent(str);
	}
	FREE(str);
}

void configure_cmdu_event(bool action)
{
	sg_cmdu_event_enable = action;
}

bool fill_interface_info(char *interface_name, struct blob_buf *bb)
{
	struct interfaceInfo *info;

	info = PLATFORM_GET_1905_INTERFACE_INFO(interface_name);
	if (info == NULL) {
		ERROR("Error in getting interface information");
		return false;
	}

	blobmsg_add_string(bb, "name", info->name);

	char mac[18] = { '\0' };

	hwaddr_ntoa(info->mac_address, mac);

	blobmsg_add_string(bb, "mac", mac);
	blobmsg_add_string(bb, "interface_type", get_interface_type_string(info->interface_type));
	//Note: Need to add interfacetype data here, skipping for now
	blobmsg_add_u8(bb, "secured", info->is_secured);
	blobmsg_add_u32(bb, "pushbutton_status", info->push_button_on_going);
	blobmsg_add_string(bb, "power_state", get_interface_power_string(info->power_state));
	// Note: Fill ipv4 neighbour details here
	blobmsg_add_u32(bb, "neighbour_nr", info->neighbor_mac_addresses_nr);

	blobmsg_add_u32(bb, "vendor_nr", info->vendor_specific_elements_nr);

	PLATFORM_FREE_1905_INTERFACE_INFO(info);

	return true;
}

static bool get_registrar_freq_band(char *reg_freq_band)
{
	bool found;

	if (reg_freq_band == NULL)
		return false;

	found = false;
	if (DMregistrarMacGet(IEEE80211_FREQUENCY_BAND_2_4_GHZ) != NULL) {
		strncpy(reg_freq_band, "802.11 2.4 GHz", 1024);
		found = true;
	}

	if (DMregistrarMacGet(IEEE80211_FREQUENCY_BAND_5_GHZ) != NULL) {
		if (found)
			strncat(reg_freq_band, ",", 1024);
		strncat(reg_freq_band, "802.11 5 GHz", 1024);
		found = true;
	}

	if (DMregistrarMacGet(IEEE80211_FREQUENCY_BAND_60_GHZ) != NULL) {
		if (found)
			strncat(reg_freq_band, ",", 1024);
		strncat(reg_freq_band, "802.11 60 GHz", 1024);
	}

	return true;
}

bool fill_interface_info_all(struct blob_buf *bb)
{
	char **ifs_names;
	INT8U ifs_nr;
	char reg_freq_band[1024] = {0};
	char al_mac_str[18] = {0};
	void *array;
	uint8_t *almac;

	ifs_names = uci_get_list_of_1905_interfaces(&ifs_nr);
	blobmsg_add_string(bb, "version", "1905.1a");

	almac = DMalMacGet();
	hwaddr_ntoa(almac, al_mac_str);
	blobmsg_add_string(bb, "ieee1905id", al_mac_str);

	if (ifs_nr) {
		blobmsg_add_string(bb, "status", "Enabled");
	} else {
		blobmsg_add_string(bb, "status", "Disabled");
	}

	blobmsg_add_u32(bb, "last_change", (get_time()/1000));

	get_registrar_freq_band(reg_freq_band);
	blobmsg_add_string(bb, "registrar_freq_band", reg_freq_band);

	blobmsg_add_string(bb, "registrar2dot4", gUbusRegistrar2dot4);

        blobmsg_add_string(bb, "registrar5", gUbusRegistrar5);

        blobmsg_add_string(bb, "registrar60", gUbusRegistrar60);

	blobmsg_add_u32(bb, "interface_num", ifs_nr);

	array = blobmsg_open_array(bb, "interfaces");
	for (uint8_t i = 0; i < ifs_nr; i++) {
		DEBUG("Filling interface|%s| info", ifs_names[i]);
		blobmsg_add_string(bb, NULL, ifs_names[i]);
	}
	blobmsg_close_array(bb, array);
	uci_free_list_of_1905_interfaces(ifs_names, ifs_nr);
	return true;
}

static bool _validate_cmdu_broadcast(struct cmdu_cstruct *c)
{
	if (c == NULL)
		return false;

	if (false == sg_cmdu_event_enable)
		return false;

	switch (c->message_type) {
	case CMDU_TYPE_TOPOLOGY_DISCOVERY:
	case CMDU_TYPE_TOPOLOGY_NOTIFICATION:
	case CMDU_TYPE_TOPOLOGY_RESPONSE:
	case CMDU_TYPE_AP_AUTOCONFIGURATION_SEARCH:
	case CMDU_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION:
	case CMDU_TYPE_AP_AUTOCONFIGURATION_RESPONSE:
	case CMDU_TYPE_LINK_METRIC_RESPONSE:
	case CMDU_TYPE_GENERIC_PHY_RESPONSE:
	case CMDU_TYPE_INTERFACE_POWER_CHANGE_RESPONSE:
	case CMDU_TYPE_HIGHER_LAYER_RESPONSE:
		return true;
	}
	return false;
}

void BroadcastCMDU(struct cmdu_cstruct *cmdu, char *ingress, uint8_t *source)
{
	INT8U **streams = NULL;
	INT16U *streams_lens = NULL;
	INT8U total_streams = 0, i;

	if (false == _validate_cmdu_broadcast(cmdu))
		return;

	struct blob_buf bb;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	streams = forge_1905_TLVs_from_CMDU(cmdu, &streams_lens, &total_streams);

	blobmsg_add_u32(&bb, "type", cmdu->message_type);
	blobmsg_add_u32(&bb, "mid", cmdu->message_id);

	if (ingress) {
		blobmsg_add_string(&bb, "ingress", ingress);
	}

	if (source) {
		char smac[18] = {0};
		hwaddr_ntoa(source, smac);
		blobmsg_add_string(&bb, "source", smac);
	}
	// count tlv buffer size
	uint32_t tlv_buffer = 0;
	for (i = 0; i < total_streams; i++) {
		tlv_buffer += streams_lens[i];
	}

	char *buff = NULL, *bp;

	CALLOC(buff, (1 + tlv_buffer * 2 + 6 + 1));
	bp = buff;
	for (i = 0; i < total_streams; ++i) {
		bp = buff + strlen(buff);
		btostr(streams[i], streams_lens[i], bp);
		FREE(streams[i]);
	}

	// append end-of-message tlv_string
	bp = buff + strlen(buff);
	sprintf(bp, "%06x", 0);

	blobmsg_add_string(&bb, "tlvs", buff);
	FREE(buff);
	FREE(streams);
	FREE(streams_lens);

	send_blob_to_ubus_mq(&bb);
}

//Adding the registrar mertics based on the AP-autoconfiguration response messages
void update_registrar_in_network(INT8U supported_freq_band, INT8U *al_mac)
{
	char macStr[18] = {0};
	char *reg, *p;

	if (al_mac == NULL)
		return;

	//Convert the destination mac address to string
	hwaddr_ntoa(al_mac, macStr);
	//Check the supported freq band
	if(supported_freq_band == IEEE80211_FREQUENCY_BAND_2_4_GHZ) {
		reg = gUbusRegistrar2dot4;
	}
	else if(supported_freq_band == IEEE80211_FREQUENCY_BAND_5_GHZ) {
		reg = gUbusRegistrar5;
	}
	else if(supported_freq_band == IEEE80211_FREQUENCY_BAND_60_GHZ) {
		reg = gUbusRegistrar60;
	}
	else{
		reg = NULL;
	}

	if (reg == NULL)
		return;

	p = strcasestr(reg, macStr);
	int len = strlen(reg);
	// add if not present
	if (p == NULL && (len < (1023 - 18))) {
		if (len) {
			strncat(reg, ",", 1);
		}
		strncat(reg, macStr, 18);
		reg[1023] = '\0';
	}
}

bool check_wireless_ifname(const char *ifname)
{
	uint32_t id;
	struct ubus_context *ctx = ubus_connect(NULL);
	char object[128] = {0};

	if (ifname == NULL)
		return false;

	snprintf(object, 128, "wifi.ap.%s", ifname);

	if (!ctx) {
		ERROR("ubus_connect failed");
		return false;
	}

	if (ubus_lookup_id(ctx, object, &id)) {
		ubus_free(ctx);
		return false;
	}

	ubus_free(ctx);
	return true;
}

int ubus_call(const char *object, const char *method, struct blob_buf *data, UBUS_CB callback, void *cb_arg)
{
	uint32_t id;
	struct ubus_context *ctx = ubus_connect(NULL);

	if (!ctx) {
		ERROR("ubus_connect failed");
		return UBUS_STATUS_UNKNOWN_ERROR;
	}

	if (ubus_lookup_id(ctx, object, &id)) {
		ERROR("(%s) not present", object);
		ubus_free(ctx);
		return UBUS_STATUS_UNKNOWN_ERROR;
	}

	// Invoke Ubus to get data from uspd
	if (ubus_invoke(ctx, id, method, data->head, callback, cb_arg, 1000)) {
		ERROR("ubus call failed");
		ubus_free(ctx);
		return UBUS_STATUS_UNKNOWN_ERROR;
	}

	ubus_free(ctx);
	return UBUS_STATUS_OK;
}

bool uci_reload_services(void)
{
	struct blob_buf bb;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	INFO("## Reloading uci config");
	if (ubus_call("uci", "reload_config", &bb, NULL, NULL) == UBUS_STATUS_OK)
		return true;

	return false;
}

static void add_bridge_cb(struct ubus_request *req,
			  int mtype __attribute__((unused)),
			  struct blob_attr *msg)
{
	char *str;
	const char *br_name;
	json_object *jbr;

	if (!msg)
		return;

	br_name = (const char *) req->priv;

	if (br_name == NULL)
		return;

	str = blobmsg_format_json_indent(msg, true, -1);
	if (str == NULL)
		return;

	json_object *jroot = json_tokener_parse(str);
	FREE(str);

	if (jroot == NULL)
		return;

	if (!json_object_object_get_ex(jroot, br_name, &jbr)) {
		ERROR("Parsing of bridge info failed");
		json_safe_put(jroot);
		return;
	}

	char *type = _json_object_get_string(jbr, "type");
	if (strncmp(type, "bridge", 6) == 0) {

		FREE(type);
		json_object *jmem;
		if (!json_object_object_get_ex(jbr, "bridge-members", &jmem)) {
			json_safe_put(jroot);
			ERROR("bridge members not found");
			return;
		}

		INT8U intf_nbr = (INT8U) json_object_array_length(jmem);
		for (size_t i = 0; i < intf_nbr; ++i) {
			json_object *elem = json_object_array_get_idx(jmem, i);

			if (elem == NULL)
				continue;

			const char *name = json_object_get_string(elem);
			if (check_wireless_ifname(name)) {
				addInterface(name);
			}
		}

		if (DMalMacConfigured() == false) {
			char *mac_str = _json_object_get_string(jbr, "macaddr");
			INFO("al mac not configured configure(%s)", mac_str);
			uint8_t mac[6];
			hwaddr_aton(mac_str, mac);
			DMalMacSet(mac);
			FREE(mac_str);
		}

		// Finally add the bridge interface as eth interface
		addInterface(br_name);
	} else {
		FREE(type);
		ERROR("(%s) is not a bridge interface", br_name);
	}
	json_safe_put(jroot);
}

void add_interfaces_from_bridge(const char *br_name)
{
	struct blob_buf bb;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	ubus_call("network.device", "status", &bb, add_bridge_cb, (void *)br_name);
}


static bool does_bridge_has_registered_interface(char *brname, json_object *jbr)
{
	uint8_t interface_num = 0, i, j, intf_nbr;
	char **interfaces;
	bool ret = false;
	const char *name;
	json_object *jmem;

	interfaces = PLATFORM_GET_LIST_OF_1905_INTERFACES(&interface_num);
	for (i = 0; i < interface_num; i++) {
		if (strcmp(interfaces[i], brname) == 0) {
			ret = true;
			break;
		}

		if (json_object_object_get_ex(jbr, "bridge-members", &jmem) == 0)
			break;

		intf_nbr = (uint8_t) json_object_array_length(jmem);
		for (j = 0; j < intf_nbr; ++j) {
			json_object *elem = json_object_array_get_idx(jmem, j);

			if (elem == NULL)
				continue;

			name = json_object_get_string(elem);
			if (strcmp(interfaces[i], name) == 0) {
				ret = true;
				break;
			}
		}
		if (ret)
			break;
	}

	PLATFORM_FREE_LIST_OF_1905_INTERFACES(interfaces, interface_num);
	return ret;
}
static bool does_bridge_has_this_interface(json_object *jbr, const char *ifname)
{
	uint8_t i, intf_nbr;
	const char *name;
	bool ret = false;
	json_object *jmem;

	if (json_object_object_get_ex(jbr, "bridge-members", &jmem) == 0)
		return false;

	intf_nbr = (uint8_t) json_object_array_length(jmem);
	for (i = 0; i < intf_nbr; ++i) {
		json_object *elem = json_object_array_get_idx(jmem, i);

		if (elem == NULL)
			continue;

		name = json_object_get_string(elem);
		if (strcmp(ifname, name) == 0) {
			ret = true;
			break;
		}
	}

	return ret;
}
static void add_bridge_intf(struct bridge_info *brinfo, char *brname, char *ifname)
{
	uint8_t i, index, intf_index;
	bool found;

	// Check if bridge is added
	found = false;
	index = 0;
	for (i = 0; i < brinfo->br_nr; i++) {
		if (strcmp(brinfo->br[i].name, brname) == 0) {
			index = i;
			found = true;
			break;
		}
	}

	// if not found add the bridge in the list
	if (!found) {
		index = brinfo->br_nr;
		brinfo->br[index].name = strdup(brname);
		brinfo->br_nr++;
	}

	// check overflow condition
	intf_index = brinfo->br[index].bridged_interfaces_nr;
	if (intf_index + 1 == MAX_BR_INTERFACES)
		return;

	// check if interface present for that bridge
	for (i = 0; i < intf_index; i++) {
		if (strcmp(brinfo->br[index].bridged_interfaces[i], ifname) == 0)
			return;
	}

	// Add interface in bridge
	brinfo->br[index].bridged_interfaces[intf_index] = strdup(ifname);
	brinfo->br[index].bridged_interfaces_nr++;
}

static void bridge_cb(struct ubus_request *req,
			  int mtype __attribute__((unused)),
			  struct blob_attr *msg)
{
	char *str, *type;
	json_object *jroot;
	struct bridge_info *binfo = req->priv;
	uint8_t interface_num = 0, i, count;
	char **interfaces;

	if (!msg || !binfo)
		return;

	str = blobmsg_format_json_indent(msg, true, -1);
	if (str == NULL)
		return;

	jroot = json_tokener_parse(str);
	free(str);

	if (!jroot) {
		ERROR("Unable to parse the output");
		return;
	}

	count = 0;
	json_object_object_foreach(jroot, key, jval) {
		// skip non-bridge interfaces
		type = _json_object_get_string(jval, "type");
		if (strcmp(type, "bridge")) {
			free(type);
			continue;
		}

		if (does_bridge_has_registered_interface(key, jval))
			count++;

		free(type);
	}


	binfo->br_nr = 0;
	if (count) {
		binfo->br = (struct bridge *) calloc(count, sizeof(struct bridge));
		if (!binfo->br)
			return;
	} else {
		binfo->br = NULL;
		json_safe_put(jroot);
		return;
	}

	interfaces = PLATFORM_GET_LIST_OF_1905_INTERFACES(&interface_num);
	json_object_object_foreach(jroot, key2, jval2) {
		// skip non-bridge interfaces
		type = _json_object_get_string(jval2, "type");
		if (strcmp(type, "bridge")) {
			free(type);
			continue;
		}

		for (i = 0; i < interface_num; ++i) {
			if (strcmp(key2, interfaces[i]) == 0) {
				add_bridge_intf(binfo, key2, interfaces[i]);
				continue;
			}

			if (does_bridge_has_this_interface(jval2, interfaces[i]))
				add_bridge_intf(binfo, key2, interfaces[i]);

		}
		free(type);
	}
	PLATFORM_FREE_LIST_OF_1905_INTERFACES(interfaces, interface_num);
	json_safe_put(jroot);
}

void get_bridge_info(struct bridge_info *binfo)
{
	struct blob_buf bb;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	ubus_call("network.device", "status", &bb, bridge_cb, binfo);
	blob_buf_free(&bb);
}

