/*
 * cmdu_visit.c: Converts CMDU to human readable format
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "utils.h"

#include "cmdu_visit.h"

////////////////////////////////////////////////////////////////////////////////
// Actual API functions
////////////////////////////////////////////////////////////////////////////////

void visit_1905_CMDU_structure(struct cmdu_cstruct *memory_structure, PRINT_CALLBACK callback, PRINT_WRITER write_function, const char *prefix, struct blob_buf *bb)
{
	// Buffer size to store a prefix string that will be used to show each
	// element of a structure on screen
	//
#define MAX_PREFIX  100

	INT8U i;
	void *array, *table;

	if (memory_structure == NULL) {
		return;
	}

	callback(bb, write_function, prefix, sizeof(memory_structure->message_version), "message_version", "%d", &memory_structure->message_version);
	callback(bb, write_function, prefix, sizeof(memory_structure->message_type), "message_type", "%d", &memory_structure->message_type);
	callback(bb, write_function, prefix, sizeof(memory_structure->message_id), "message_id", "%d", &memory_structure->message_id);
	callback(bb, write_function, prefix, sizeof(memory_structure->relay_indicator), "relay_indicator", "%d", &memory_structure->relay_indicator);

	if (memory_structure->tlv_cstruct_list == NULL) {
		return;
	}

	i = 0;
	if (bb)
		array = blobmsg_open_array(bb, "TLVs");
	while (memory_structure->tlv_cstruct_list[i] != NULL) {
		if (bb)
			table = blobmsg_open_table(bb, NULL);
		// In order to make it easier for the callback() function to present
		// useful information, append the type of the TLV to the prefix
		char new_prefix[MAX_PREFIX];

		snprintf(new_prefix, MAX_PREFIX - 1, "%sTLV(%s)->", prefix,
			 convert_1905_TLV_type_to_string(*(memory_structure->tlv_cstruct_list[i])));
		new_prefix[MAX_PREFIX - 1] = 0x0;

		visit_1905_TLV_structure(memory_structure->tlv_cstruct_list[i], callback, write_function, new_prefix, bb);
		i++;
		if (bb)
			blobmsg_close_array(bb, table);
	}
	if (bb)
		blobmsg_close_array(bb, array);

	return;
}
void visit_1905_TLV_structure(INT8U *memory_structure, PRINT_CALLBACK callback, PRINT_WRITER write_function, const char *prefix, struct blob_buf *bb)
{
	void *array = NULL, *table = NULL, *subarray = NULL, *subtable = NULL;
	// Buffer size to store a prefix string that will be used to show each
	// element of a structure on screen
	//
#define MAX_PREFIX  100

	if (memory_structure == NULL) {
		return;
	}

	if (bb)
		blobmsg_add_string(bb, "tlv_type", convert_1905_TLV_type_to_string(*memory_structure));
	// The first byte of any of the valid structures is always the "tlv_type"
	// field.
	//
	switch (*memory_structure) {
	case TLV_TYPE_END_OF_MESSAGE:
		{
			// There is nothing to visit. This TLV is always empty
			//
			return;
		}

	case TLV_TYPE_VENDOR_SPECIFIC:
		{
			struct vendorSpecificTLV *p;

			p = (struct vendorSpecificTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->vendorOUI), "vendorOUI", "0x%02x", p->vendorOUI);
			callback(bb, write_function, prefix, sizeof(p->m_nr), "m_nr", "%d", &p->m_nr);
			callback(bb, write_function, prefix, p->m_nr, "m", "0x%02x", p->m);

			return;
		}

	case TLV_TYPE_AL_MAC_ADDRESS_TYPE:
		{
			struct alMacAddressTypeTLV *p;

			p = (struct alMacAddressTypeTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->al_mac_address), "al_mac_address", "0x%02x", p->al_mac_address);

			return;
		}

	case TLV_TYPE_MAC_ADDRESS_TYPE:
		{
			struct macAddressTypeTLV *p;

			p = (struct macAddressTypeTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->mac_address), "mac_address", "0x%02x", p->mac_address);

			return;
		}

	case TLV_TYPE_DEVICE_INFORMATION_TYPE:
		{
			struct deviceInformationTypeTLV *p;
			INT8U i;

			p = (struct deviceInformationTypeTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->al_mac_address), "al_mac_address", "0x%02x", p->al_mac_address);
			callback(bb, write_function, prefix, sizeof(p->local_interfaces_nr), "local_interfaces_nr", "%d", &p->local_interfaces_nr);

			if (bb)
				array = blobmsg_open_array(bb, "local_interfaces");

			for (i = 0; i < p->local_interfaces_nr; i++) {
				char new_prefix[MAX_PREFIX];

				if (bb) {
					table = blobmsg_open_table(bb, NULL);
				}

				snprintf(new_prefix, MAX_PREFIX - 1, "%slocal_interfaces[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(p->local_interfaces[i].mac_address), "mac_address", "0x%02x",
					 p->local_interfaces[i].mac_address);
				callback(bb, write_function, new_prefix, sizeof(p->local_interfaces[i].media_type), "media_type", "0x%04x",
					 &p->local_interfaces[i].media_type);
				callback(bb, write_function, new_prefix, sizeof(p->local_interfaces[i].media_specific_data_size), "media_specific_data_size",
					 "%d", &p->local_interfaces[i].media_specific_data_size);

				if ((p->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11B_2_4_GHZ) ||
				    (p->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11G_2_4_GHZ) ||
				    (p->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11A_5_GHZ) ||
				    (p->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11N_2_4_GHZ) ||
				    (p->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11N_5_GHZ) ||
				    (p->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AC_5_GHZ) ||
				    (p->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AD_60_GHZ) ||
				    (p->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_802_11AF_GHZ)
				    ) {
					callback(bb, write_function, new_prefix,
						 sizeof(p->local_interfaces[i].media_specific_data.ieee80211.network_membership), "network_membership",
						 "0x%02x", p->local_interfaces[i].media_specific_data.ieee80211.network_membership);
					callback(bb, write_function, new_prefix, sizeof(p->local_interfaces[i].media_specific_data.ieee80211.role), "role",
						 "%d", &p->local_interfaces[i].media_specific_data.ieee80211.role);
					callback(bb, write_function, new_prefix, sizeof(p->local_interfaces[i].media_specific_data.ieee80211.ap_channel_band),
						 "ap_channel_band", "%d", &p->local_interfaces[i].media_specific_data.ieee80211.ap_channel_band);
					callback(bb, write_function, new_prefix,
						 sizeof(p->local_interfaces[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_1),
						 "ap_channel_center_frequency_index_1", "%d",
						 &p->local_interfaces[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_1);
					callback(bb, write_function, new_prefix,
						 sizeof(p->local_interfaces[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_2),
						 "ap_channel_center_frequency_index_2", "%d",
						 &p->local_interfaces[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_2);
				} else if ((p->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_1901_WAVELET)
					   || (p->local_interfaces[i].media_type == MEDIA_TYPE_IEEE_1901_FFT)
				    ) {
					callback(bb, write_function, new_prefix, sizeof(p->local_interfaces[i].media_specific_data.ieee1901.network_identifier),
						 "network_identifier", "0x%02x", p->local_interfaces[i].media_specific_data.ieee1901.network_identifier);
				}

				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES:
		{
			struct deviceBridgingCapabilityTLV *p;
			INT8U i, j;

			p = (struct deviceBridgingCapabilityTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->bridging_tuples_nr), "bridging_tuples_nr", "%d", &p->bridging_tuples_nr);
			if (bb)
				array = blobmsg_open_array(bb, "bridging_tuples");
			for (i = 0; i < p->bridging_tuples_nr; i++) {
				if (bb)
					table = blobmsg_open_table(bb, NULL);
				char new_prefix[MAX_PREFIX];

				snprintf(new_prefix, MAX_PREFIX - 1, "%sbridging_tuples[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(p->bridging_tuples[i].bridging_tuple_macs_nr), "bridging_tuple_macs_nr", "%d",
					 &p->bridging_tuples[i].bridging_tuple_macs_nr);

				if (bb)
					subarray = blobmsg_open_array(bb, "macs");
				for (j = 0; j < p->bridging_tuples[i].bridging_tuple_macs_nr; j++) {
					if (bb)
						subtable = blobmsg_open_table(bb, NULL);
					snprintf(new_prefix, MAX_PREFIX - 1, "%sbridging_tuples[%d]->bridging_tuple_macs[%d]->", prefix, i, j);
					new_prefix[MAX_PREFIX - 1] = 0x0;

					callback(bb, write_function, new_prefix, sizeof(p->bridging_tuples[i].bridging_tuple_macs[j].mac_address),
						 "mac_address", "0x%02x", p->bridging_tuples[i].bridging_tuple_macs[j].mac_address);
					if (subtable)
						blobmsg_close_table(bb, subtable);
				}
				if (subarray)
					blobmsg_close_array(bb, subarray);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST:
		{
			struct non1905NeighborDeviceListTLV *p;
			INT8U i;

			p = (struct non1905NeighborDeviceListTLV *)memory_structure;

			if (p->non_1905_neighbors_nr > 0 && NULL == p->non_1905_neighbors) {
				// Malformed structure
				return;
			}

			callback(bb, write_function, prefix, sizeof(p->local_mac_address), "local_mac_address", "0x%02x", p->local_mac_address);
			callback(bb, write_function, prefix, sizeof(p->non_1905_neighbors_nr), "non_1905_neighbors_nr", "%d", &p->non_1905_neighbors_nr);
			if (bb)
				array = blobmsg_open_array(bb, "neighbors");
			for (i = 0; i < p->non_1905_neighbors_nr; i++) {
				if (bb)
					table = blobmsg_open_table(bb, NULL);
				char new_prefix[MAX_PREFIX];

				snprintf(new_prefix, MAX_PREFIX - 1, "%snon_1905_neighbors[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(p->non_1905_neighbors[i].mac_address), "mac_address", "0x%02x",
					 p->non_1905_neighbors[i].mac_address);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case TLV_TYPE_NEIGHBOR_DEVICE_LIST:
		{
			struct neighborDeviceListTLV *p;
			INT8U i;

			p = (struct neighborDeviceListTLV *)memory_structure;

			if (p->neighbors_nr > 0 && NULL == p->neighbors) {
				// Malformed structure
				return;
			}

			callback(bb, write_function, prefix, sizeof(p->local_mac_address), "local_mac_address", "0x%02x", p->local_mac_address);
			callback(bb, write_function, prefix, sizeof(p->neighbors_nr), "neighbors_nr", "%d", &p->neighbors_nr);
			if (bb)
				array = blobmsg_open_array(bb, "neighbors");

			for (i = 0; i < p->neighbors_nr; i++) {
				char new_prefix[MAX_PREFIX];

				snprintf(new_prefix, MAX_PREFIX - 1, "%sneighbors[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				if (bb) {
					table = blobmsg_open_table(bb, NULL);
				}
				callback(bb, write_function, new_prefix, sizeof(p->neighbors[i].mac_address), "mac_address", "0x%02x",
					 p->neighbors[i].mac_address);
				callback(bb, write_function, new_prefix, sizeof(p->neighbors[i].bridge_flag), "bridge_flag", "%d",
					 &p->neighbors[i].bridge_flag);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case TLV_TYPE_LINK_METRIC_QUERY:
		{
			struct linkMetricQueryTLV *p;

			p = (struct linkMetricQueryTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->destination), "destination", "%d", &p->destination);
			callback(bb, write_function, prefix, sizeof(p->specific_neighbor), "specific_neighbor", "0x%02x", p->specific_neighbor);
			callback(bb, write_function, prefix, sizeof(p->link_metrics_type), "link_metrics_type", "%d", &p->link_metrics_type);

			return;
		}

	case TLV_TYPE_TRANSMITTER_LINK_METRIC:
		{
			struct transmitterLinkMetricTLV *p;
			INT8U i;

			p = (struct transmitterLinkMetricTLV *)memory_structure;

			if (p->transmitter_link_metrics == NULL) {
				// Malformed structure
				return;
			}

			callback(bb, write_function, prefix, sizeof(p->local_al_address), "local_al_address", "0x%02x", p->local_al_address);
			callback(bb, write_function, prefix, sizeof(p->neighbor_al_address), "neighbor_al_address", "0x%02x", p->neighbor_al_address);
			callback(bb, write_function, prefix, sizeof(p->transmitter_link_metrics_nr), "transmitter_link_metrics_nr", "%d",
				 &p->transmitter_link_metrics_nr);
			if (bb)
				array = blobmsg_open_array(bb, "tx_link_metrics");
			for (i = 0; i < p->transmitter_link_metrics_nr; i++) {
				if (bb)
					table = blobmsg_open_table(bb, NULL);
				char new_prefix[MAX_PREFIX];

				snprintf(new_prefix, MAX_PREFIX - 1, "%stransmitter_link_metrics[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(p->transmitter_link_metrics[i].local_interface_address),
					 "local_interface_address", "0x%02x", p->transmitter_link_metrics[i].local_interface_address);
				callback(bb, write_function, new_prefix, sizeof(p->transmitter_link_metrics[i].neighbor_interface_address),
					 "neighbor_interface_address", "0x%02x", p->transmitter_link_metrics[i].neighbor_interface_address);
				callback(bb, write_function, new_prefix, sizeof(p->transmitter_link_metrics[i].intf_type), "intf_type", "0x%04x",
					 &p->transmitter_link_metrics[i].intf_type);
				callback(bb, write_function, new_prefix, sizeof(p->transmitter_link_metrics[i].bridge_flag), "bridge_flag", "%d",
					 &p->transmitter_link_metrics[i].bridge_flag);
				callback(bb, write_function, new_prefix, sizeof(p->transmitter_link_metrics[i].packet_errors), "packet_errors", "%d",
					 &p->transmitter_link_metrics[i].packet_errors);
				callback(bb, write_function, new_prefix, sizeof(p->transmitter_link_metrics[i].transmitted_packets), "transmitted_packets",
					 "%d", &p->transmitter_link_metrics[i].transmitted_packets);
				callback(bb, write_function, new_prefix, sizeof(p->transmitter_link_metrics[i].mac_throughput_capacity),
					 "mac_throughput_capacity", "%d", &p->transmitter_link_metrics[i].mac_throughput_capacity);
				callback(bb, write_function, new_prefix, sizeof(p->transmitter_link_metrics[i].link_availability), "link_availability", "%d",
					 &p->transmitter_link_metrics[i].link_availability);
				callback(bb, write_function, new_prefix, sizeof(p->transmitter_link_metrics[i].phy_rate), "phy_rate", "%d",
					 &p->transmitter_link_metrics[i].phy_rate);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case TLV_TYPE_RECEIVER_LINK_METRIC:
		{
			struct receiverLinkMetricTLV *p;
			INT8U i;

			p = (struct receiverLinkMetricTLV *)memory_structure;

			if (p->receiver_link_metrics == NULL) {
				// Malformed structure
				return;
			}

			callback(bb, write_function, prefix, sizeof(p->local_al_address), "local_al_address", "0x%02x", p->local_al_address);
			callback(bb, write_function, prefix, sizeof(p->neighbor_al_address), "neighbor_al_address", "0x%02x", p->neighbor_al_address);
			callback(bb, write_function, prefix, sizeof(p->receiver_link_metrics_nr), "receiver_link_metrics_nr", "%d",
				 &p->receiver_link_metrics_nr);
			if (bb)
				array = blobmsg_open_array(bb, "rx_link_metrics");
			for (i = 0; i < p->receiver_link_metrics_nr; i++) {
				if (bb)
					table = blobmsg_open_table(bb, NULL);
				char new_prefix[MAX_PREFIX];

				snprintf(new_prefix, MAX_PREFIX - 1, "%sreceiver_link_metrics[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(p->receiver_link_metrics[i].local_interface_address), "local_interface_address",
					 "0x%02x", p->receiver_link_metrics[i].local_interface_address);
				callback(bb, write_function, new_prefix, sizeof(p->receiver_link_metrics[i].neighbor_interface_address),
					 "neighbor_interface_address", "0x%02x", p->receiver_link_metrics[i].neighbor_interface_address);
				callback(bb, write_function, new_prefix, sizeof(p->receiver_link_metrics[i].intf_type), "intf_type", "0x%04x",
					 &p->receiver_link_metrics[i].intf_type);
				callback(bb, write_function, new_prefix, sizeof(p->receiver_link_metrics[i].packet_errors), "packet_errors", "%d",
					 &p->receiver_link_metrics[i].packet_errors);
				callback(bb, write_function, new_prefix, sizeof(p->receiver_link_metrics[i].packets_received), "packets_received", "%d",
					 &p->receiver_link_metrics[i].packets_received);
				callback(bb, write_function, new_prefix, sizeof(p->receiver_link_metrics[i].rssi), "rssi", "%d",
					 &p->receiver_link_metrics[i].rssi);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case TLV_TYPE_LINK_METRIC_RESULT_CODE:
		{
			struct linkMetricResultCodeTLV *p;

			p = (struct linkMetricResultCodeTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->result_code), "result_code", "%d", &p->result_code);

			return;
		}

	case TLV_TYPE_SEARCHED_ROLE:
		{
			struct searchedRoleTLV *p;

			p = (struct searchedRoleTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->role), "role", "%d", &p->role);

			return;
		}

	case TLV_TYPE_AUTOCONFIG_FREQ_BAND:
		{
			struct autoconfigFreqBandTLV *p;

			p = (struct autoconfigFreqBandTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->freq_band), "freq_band", "%d", &p->freq_band);

			return;
		}

	case TLV_TYPE_SUPPORTED_ROLE:
		{
			struct supportedRoleTLV *p;

			p = (struct supportedRoleTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->role), "role", "%d", &p->role);

			return;
		}

	case TLV_TYPE_SUPPORTED_FREQ_BAND:
		{
			struct supportedFreqBandTLV *p;

			p = (struct supportedFreqBandTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->freq_band), "freq_band", "%d", &p->freq_band);

			return;
		}

	case TLV_TYPE_WSC:
		{
			struct wscTLV *p;

			p = (struct wscTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->wsc_frame_size), "wsc_frame_size", "%d", &p->wsc_frame_size);
			callback(bb, write_function, prefix, p->wsc_frame_size, "wsc_frame", "0x%02x", p->wsc_frame);

			return;
		}

	case TLV_TYPE_PUSH_BUTTON_EVENT_NOTIFICATION:
		{
			struct pushButtonEventNotificationTLV *p;
			INT8U i;

			p = (struct pushButtonEventNotificationTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->media_types_nr), "media_types_nr", "0x%02x", &p->media_types_nr);
			if (bb)
				array = blobmsg_open_array(bb, "media_types");
			for (i = 0; i < p->media_types_nr; i++) {
				if (bb)
					table = blobmsg_open_table(bb, NULL);
				char new_prefix[MAX_PREFIX];

				snprintf(new_prefix, MAX_PREFIX - 1, "%smedia_types[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(p->media_types[i].media_type), "media_type", "0x%04x",
					 &p->media_types[i].media_type);
				callback(bb, write_function, new_prefix, sizeof(p->media_types[i].media_specific_data_size), "media_specific_data_size", "%d",
					 &p->media_types[i].media_specific_data_size);

				if ((p->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11B_2_4_GHZ) ||
				    (p->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11G_2_4_GHZ) ||
				    (p->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11A_5_GHZ) ||
				    (p->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11N_2_4_GHZ) ||
				    (p->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11N_5_GHZ) ||
				    (p->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11AC_5_GHZ) ||
				    (p->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11AD_60_GHZ) ||
				    (p->media_types[i].media_type == MEDIA_TYPE_IEEE_802_11AF_GHZ)
				    ) {
					callback(bb, write_function, new_prefix, sizeof(p->media_types[i].media_specific_data.ieee80211.network_membership),
						 "network_membership", "0x%02x", p->media_types[i].media_specific_data.ieee80211.network_membership);
					callback(bb, write_function, new_prefix, sizeof(p->media_types[i].media_specific_data.ieee80211.role), "role", "%d",
						 &p->media_types[i].media_specific_data.ieee80211.role);
					callback(bb, write_function, new_prefix, sizeof(p->media_types[i].media_specific_data.ieee80211.ap_channel_band),
						 "ap_channel_band", "%d", &p->media_types[i].media_specific_data.ieee80211.ap_channel_band);
					callback(bb, write_function, new_prefix,
						 sizeof(p->media_types[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_1),
						 "ap_channel_center_frequency_index_1", "%d",
						 &p->media_types[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_1);
					callback(bb, write_function, new_prefix,
						 sizeof(p->media_types[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_2),
						 "ap_channel_center_frequency_index_2", "%d",
						 &p->media_types[i].media_specific_data.ieee80211.ap_channel_center_frequency_index_2);
				} else if ((p->media_types[i].media_type == MEDIA_TYPE_IEEE_1901_WAVELET)
					   || (p->media_types[i].media_type == MEDIA_TYPE_IEEE_1901_FFT)
				    ) {
					callback(bb, write_function, new_prefix, sizeof(p->media_types[i].media_specific_data.ieee1901.network_identifier),
						 "network_identifier", "0x%02x", p->media_types[i].media_specific_data.ieee1901.network_identifier);
				}
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case TLV_TYPE_PUSH_BUTTON_JOIN_NOTIFICATION:
		{
			struct pushButtonJoinNotificationTLV *p;

			p = (struct pushButtonJoinNotificationTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->al_mac_address), "al_mac_address", "0x%02x", p->al_mac_address);
			callback(bb, write_function, prefix, sizeof(p->message_identifier), "message_identifier", "%d", &p->message_identifier);
			callback(bb, write_function, prefix, sizeof(p->mac_address), "mac_address", "0x%02x", p->mac_address);
			callback(bb, write_function, prefix, sizeof(p->new_mac_address), "new_mac_address", "0x%02x", p->new_mac_address);
			return;
		}

	case TLV_TYPE_GENERIC_PHY_DEVICE_INFORMATION:
		{
			struct genericPhyDeviceInformationTypeTLV *p;
			INT8U i;

			p = (struct genericPhyDeviceInformationTypeTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->al_mac_address), "al_mac_address", "0x%02x", &p->al_mac_address);
			callback(bb, write_function, prefix, sizeof(p->local_interfaces_nr), "local_interfaces_nr", "%d", &p->local_interfaces_nr);
			if (bb)
				array = blobmsg_open_array(bb, "local_interfaces");
			for (i = 0; i < p->local_interfaces_nr; i++) {
				if (bb)
					table = blobmsg_open_table(bb, NULL);
				char new_prefix[MAX_PREFIX];

				snprintf(new_prefix, MAX_PREFIX - 1, "%slocal_interfaces[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(p->local_interfaces[i].local_interface_address), "local_interface_address",
					 "0x%02x", p->local_interfaces[i].local_interface_address);
				callback(bb, write_function, new_prefix, sizeof(p->local_interfaces[i].generic_phy_common_data.oui), "oui", "0x%02x",
					 p->local_interfaces[i].generic_phy_common_data.oui);
				callback(bb, write_function, new_prefix, sizeof(p->local_interfaces[i].generic_phy_common_data.variant_index), "variant_index",
					 "%d", &p->local_interfaces[i].generic_phy_common_data.variant_index);
				callback(bb, write_function, new_prefix, sizeof(p->local_interfaces[i].variant_name), "variant_name", "%s",
					 p->local_interfaces[i].variant_name);
				callback(bb, write_function, new_prefix, sizeof(p->local_interfaces[i].generic_phy_description_xml_url_len),
					 "generic_phy_description_xml_url_len", "%d", &p->local_interfaces[i].generic_phy_description_xml_url_len);
				callback(bb, write_function, new_prefix, sizeof(p->local_interfaces[i].generic_phy_common_data.media_specific_bytes_nr),
					 "media_specific_bytes_nr", "%d", &p->local_interfaces[i].generic_phy_common_data.media_specific_bytes_nr);
				callback(bb, write_function, new_prefix, p->local_interfaces[i].generic_phy_description_xml_url_len,
					 "generic_phy_description_xml_url", "%s", p->local_interfaces[i].generic_phy_description_xml_url);
				callback(bb, write_function, new_prefix, p->local_interfaces[i].generic_phy_common_data.media_specific_bytes_nr,
					 "media_specific_bytes", "0x%02x", p->local_interfaces[i].generic_phy_common_data.media_specific_bytes);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case TLV_TYPE_DEVICE_IDENTIFICATION:
		{
			struct deviceIdentificationTypeTLV *p;

			p = (struct deviceIdentificationTypeTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->friendly_name), "friendly_name", "%s", p->friendly_name);
			callback(bb, write_function, prefix, sizeof(p->manufacturer_name), "manufacturer_name", "%s", p->manufacturer_name);
			callback(bb, write_function, prefix, sizeof(p->manufacturer_model), "manufacturer_model", "%s", p->manufacturer_model);
			return;
		}

	case TLV_TYPE_CONTROL_URL:
		{
			struct controlUrlTypeTLV *p;

			p = (struct controlUrlTypeTLV *)memory_structure;

			callback(bb, write_function, prefix, strlen(p->url) + 1, "url", "%s", p->url);

			return;
		}

	case TLV_TYPE_IPV4:
		{
			struct ipv4TypeTLV *p;
			INT8U i, j;

			p = (struct ipv4TypeTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->ipv4_interfaces_nr), "ipv4_interfaces_nr", "%d", &p->ipv4_interfaces_nr);
			if (bb)
				array = blobmsg_open_array(bb, "ipv4_interfaces");
			for (i = 0; i < p->ipv4_interfaces_nr; i++) {
				if (bb)
					table = blobmsg_open_table(bb, NULL);
				char new_prefix[MAX_PREFIX];

				snprintf(new_prefix, MAX_PREFIX - 1, "%sipv4_interfaces[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(p->ipv4_interfaces[i].mac_address), "mac_address", "0x%02x",
					 p->ipv4_interfaces[i].mac_address);
				callback(bb, write_function, new_prefix, sizeof(p->ipv4_interfaces[i].ipv4_nr), "ipv4_nr", "%d",
					 &p->ipv4_interfaces[i].ipv4_nr);

				if (bb)
					subarray = blobmsg_open_array(bb, "ipv4");
				for (j = 0; j < p->ipv4_interfaces[i].ipv4_nr; j++) {
					if (bb)
						subtable = blobmsg_open_table(bb, NULL);
					snprintf(new_prefix, MAX_PREFIX - 1, "%sipv4_interfaces[%d]->ipv4[%d]->", prefix, i, j);
					new_prefix[MAX_PREFIX - 1] = 0x0;

					callback(bb, write_function, new_prefix, sizeof(p->ipv4_interfaces[i].ipv4[j].type), "type", "%d",
						 &p->ipv4_interfaces[i].ipv4[j].type);
					callback(bb, write_function, new_prefix, sizeof(p->ipv4_interfaces[i].ipv4[j].ipv4_address), "ipv4_address", "%ipv4",
						 p->ipv4_interfaces[i].ipv4[j].ipv4_address);
					callback(bb, write_function, new_prefix, sizeof(p->ipv4_interfaces[i].ipv4[j].ipv4_dhcp_server), "ipv4_dhcp_server",
						 "%ipv4", p->ipv4_interfaces[i].ipv4[j].ipv4_dhcp_server);
					if (subtable)
						blobmsg_close_table(bb, subtable);
				}
				if (subarray)
					blobmsg_close_array(bb, subarray);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case TLV_TYPE_IPV6:
		{
			struct ipv6TypeTLV *p;
			INT8U i, j;

			p = (struct ipv6TypeTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->ipv6_interfaces_nr), "ipv6_interfaces_nr", "%d", &p->ipv6_interfaces_nr);
			if (bb)
				array = blobmsg_open_array(bb, "ipv4_interfaces");
			for (i = 0; i < p->ipv6_interfaces_nr; i++) {
				if (bb)
					table = blobmsg_open_table(bb, NULL);
				char new_prefix[MAX_PREFIX];

				snprintf(new_prefix, MAX_PREFIX - 1, "%sipv6_interfaces[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(p->ipv6_interfaces[i].mac_address), "mac_address", "0x%02x",
					 p->ipv6_interfaces[i].mac_address);
				callback(bb, write_function, new_prefix, sizeof(p->ipv6_interfaces[i].ipv6_nr), "ipv6_nr", "%d",
					 &p->ipv6_interfaces[i].ipv6_nr);

				if (bb)
					subarray = blobmsg_open_array(bb, "ipv4");
				for (j = 0; j < p->ipv6_interfaces[i].ipv6_nr; j++) {
					if (bb)
						subtable = blobmsg_open_table(bb, NULL);
					snprintf(new_prefix, MAX_PREFIX - 1, "%sipv6_interfaces[%d]->ipv6[%d]->", prefix, i, j);
					new_prefix[MAX_PREFIX - 1] = 0x0;

					callback(bb, write_function, new_prefix, sizeof(p->ipv6_interfaces[i].ipv6[j].type), "type", "%d",
						 &p->ipv6_interfaces[i].ipv6[j].type);
					callback(bb, write_function, new_prefix, sizeof(p->ipv6_interfaces[i].ipv6[j].ipv6_address), "ipv6_address", "0x%02x",
						 p->ipv6_interfaces[i].ipv6[j].ipv6_address);
					callback(bb, write_function, new_prefix, sizeof(p->ipv6_interfaces[i].ipv6[j].ipv6_address_origin),
						 "ipv6_address_origin", "0x%02x", p->ipv6_interfaces[i].ipv6[j].ipv6_address_origin);
					if (subtable)
						blobmsg_close_table(bb, subtable);
				}
				if (subarray)
					blobmsg_close_array(bb, subarray);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case TLV_TYPE_GENERIC_PHY_EVENT_NOTIFICATION:
		{
			struct pushButtonGenericPhyEventNotificationTLV *pev;
			INT8U i;

			pev = (struct pushButtonGenericPhyEventNotificationTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(pev->local_interfaces_nr), "local_interfaces_nr", "%d", &pev->local_interfaces_nr);
			if (bb)
				array = blobmsg_open_array(bb, "ipv4_interfaces");
			for (i = 0; i < pev->local_interfaces_nr; i++) {
				if (bb)
					table = blobmsg_open_table(bb, NULL);
				char new_prefix[MAX_PREFIX];

				snprintf(new_prefix, MAX_PREFIX - 1, "%slocal_interfaces[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(pev->local_interfaces[i].oui), "oui", "0x%02x", pev->local_interfaces[i].oui);
				callback(bb, write_function, new_prefix, sizeof(pev->local_interfaces[i].variant_index), "variant_index", "%d",
					 &pev->local_interfaces[i].variant_index);
				callback(bb, write_function, new_prefix, sizeof(pev->local_interfaces[i].media_specific_bytes_nr), "media_specific_bytes_nr",
					 "%d", &pev->local_interfaces[i].media_specific_bytes_nr);
				callback(bb, write_function, new_prefix, pev->local_interfaces[i].media_specific_bytes_nr, "media_specific_bytes", "0x%02x",
					 pev->local_interfaces[i].media_specific_bytes);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case TLV_TYPE_1905_PROFILE_VERSION:
		{
			struct x1905ProfileVersionTLV *p;

			p = (struct x1905ProfileVersionTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->profile), "profile", "%d", &p->profile);

			return;
		}

	case TLV_TYPE_POWER_OFF_INTERFACE:
		{
			struct powerOffInterfaceTLV *p;
			INT8U i;

			p = (struct powerOffInterfaceTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->power_off_interfaces_nr), "power_off_interfaces_nr", "%d", &p->power_off_interfaces_nr);
			if (bb)
				array = blobmsg_open_array(bb, "ipv4_interfaces");
			for (i = 0; i < p->power_off_interfaces_nr; i++) {
				if (bb)
					table = blobmsg_open_table(bb, NULL);
				char new_prefix[MAX_PREFIX];

				snprintf(new_prefix, MAX_PREFIX - 1, "%spower_off_interfaces[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(p->power_off_interfaces[i].interface_address), "interface_address", "0x%02x",
					 p->power_off_interfaces[i].interface_address);
				callback(bb, write_function, new_prefix, sizeof(p->power_off_interfaces[i].media_type), "media_type", "0x%04x",
					 &p->power_off_interfaces[i].media_type);
				callback(bb, write_function, new_prefix, sizeof(p->power_off_interfaces[i].generic_phy_common_data.oui), "oui", "0x%02x",
					 p->power_off_interfaces[i].generic_phy_common_data.oui);
				callback(bb, write_function, new_prefix, sizeof(p->power_off_interfaces[i].generic_phy_common_data.variant_index),
					 "variant_index", "%d", &p->power_off_interfaces[i].generic_phy_common_data.variant_index);
				callback(bb, write_function, new_prefix, sizeof(p->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes_nr),
					 "media_specific_bytes_nr", "%d", &p->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes_nr);
				callback(bb, write_function, new_prefix, p->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes_nr,
					 "media_specific_bytes", "0x%02x", p->power_off_interfaces[i].generic_phy_common_data.media_specific_bytes);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case TLV_TYPE_INTERFACE_POWER_CHANGE_INFORMATION:
		{
			struct interfacePowerChangeInformationTLV *p_ch;
			INT8U i;

			p_ch = (struct interfacePowerChangeInformationTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p_ch->power_change_interfaces_nr), "power_change_interfaces_nr", "%d",
				 &p_ch->power_change_interfaces_nr);
			if (bb)
				array = blobmsg_open_array(bb, "ipv4_interfaces");
			for (i = 0; i < p_ch->power_change_interfaces_nr; i++) {
				if (bb)
					table = blobmsg_open_table(bb, NULL);
				char new_prefix[MAX_PREFIX];

				snprintf(new_prefix, MAX_PREFIX - 1, "%spower_change_interfaces[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(p_ch->power_change_interfaces[i].interface_address), "interface_address", "0x%02x",
					 p_ch->power_change_interfaces[i].interface_address);
				callback(bb, write_function, new_prefix, sizeof(p_ch->power_change_interfaces[i].requested_power_state), "requested_power_state",
					 "0x%02x", &p_ch->power_change_interfaces[i].requested_power_state);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case TLV_TYPE_INTERFACE_POWER_CHANGE_STATUS:
		{
			struct interfacePowerChangeStatusTLV *p;
			INT8U i;

			p = (struct interfacePowerChangeStatusTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->power_change_interfaces_nr), "power_change_interfaces_nr", "%d",
				 &p->power_change_interfaces_nr);
			if (bb)
				array = blobmsg_open_array(bb, "ipv4_interfaces");
			for (i = 0; i < p->power_change_interfaces_nr; i++) {
				if (bb)
					table = blobmsg_open_table(bb, NULL);
				char new_prefix[MAX_PREFIX];

				snprintf(new_prefix, MAX_PREFIX - 1, "%spower_change_interfaces[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(p->power_change_interfaces[i].interface_address), "interface_address", "0x%02x",
					 p->power_change_interfaces[i].interface_address);
				callback(bb, write_function, new_prefix, sizeof(p->power_change_interfaces[i].result), "result", "%d",
					 &p->power_change_interfaces[i].result);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	case TLV_TYPE_L2_NEIGHBOR_DEVICE:
		{
			struct l2NeighborDeviceTLV *p;
			INT8U i, j, k;
			void *subsubarray = NULL, *subsubtable = NULL;

			p = (struct l2NeighborDeviceTLV *)memory_structure;

			callback(bb, write_function, prefix, sizeof(p->local_interfaces_nr), "local_interfaces_nr", "%d", &p->local_interfaces_nr);
			if (bb)
				array = blobmsg_open_array(bb, "ipv4_interfaces");
			for (i = 0; i < p->local_interfaces_nr; i++) {
				if (bb)
					table = blobmsg_open_table(bb, NULL);
				char new_prefix[MAX_PREFIX];

				snprintf(new_prefix, MAX_PREFIX - 1, "%slocal_interfaces[%d]->", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;

				callback(bb, write_function, new_prefix, sizeof(p->local_interfaces[i].local_mac_address), "local_mac_address", "0x%02x",
					 p->local_interfaces[i].local_mac_address);
				callback(bb, write_function, new_prefix, sizeof(p->local_interfaces[i].l2_neighbors_nr), "l2_neighbors_nr", "%d",
					 &p->local_interfaces[i].l2_neighbors_nr);

				if (bb)
					subarray = blobmsg_open_array(bb, "ipv4");
				for (j = 0; j < p->local_interfaces[i].l2_neighbors_nr; j++) {
					if (bb)
						subtable = blobmsg_open_table(bb, NULL);
					snprintf(new_prefix, MAX_PREFIX - 1, "%slocal_interfaces[%d]->l2_neighbors[%d]->", prefix, i, j);
					new_prefix[MAX_PREFIX - 1] = 0x0;

					callback(bb, write_function, new_prefix, sizeof(p->local_interfaces[i].l2_neighbors[j].l2_neighbor_mac_address),
						 "l2_neighbor_mac_address", "0x%02x", p->local_interfaces[i].l2_neighbors[j].l2_neighbor_mac_address);
					callback(bb, write_function, new_prefix, sizeof(p->local_interfaces[i].l2_neighbors[j].behind_mac_addresses_nr),
						 "behind_mac_addresses_nr", "%d", &p->local_interfaces[i].l2_neighbors[j].behind_mac_addresses_nr);

					if (bb)
						subsubarray = blobmsg_open_array(bb, "behind_macs");
					for (k = 0; k < p->local_interfaces[i].l2_neighbors[j].behind_mac_addresses_nr; k++) {
						if (bb)
							subsubtable = blobmsg_open_table(bb, NULL);
						snprintf(new_prefix, MAX_PREFIX - 1, "%slocal_interfaces[%d]->l2_neighbors[%d]->behind_mac_addresses[%d]",
							 prefix, i, j, k);
						new_prefix[MAX_PREFIX - 1] = 0x0;

						callback(bb, write_function, new_prefix, 6, "behind_mac_addresses", "0x%02x",
							 p->local_interfaces[i].l2_neighbors[j].behind_mac_addresses[k]);
						if (subsubtable)
							blobmsg_close_table(bb, subsubtable);
					}
					if (subsubarray)
						blobmsg_close_array(bb, subsubarray);
					if (subtable)
						blobmsg_close_table(bb, subtable);
				}
				if (subarray)
					blobmsg_close_array(bb, subarray);
				if (table)
					blobmsg_close_table(bb, table);
			}
			if (array)
				blobmsg_close_array(bb, array);

			return;
		}

	default:
		{
			// Ignore
			//
			return;
		}
	}

	// This code cannot be reached
	//
	return;
}
