/*
 * utils.c: Common utility functions
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "utils.h"
#include "common.h"
#include "platform_crypto.h"

#include <stdlib.h>		// free(), malloc(), ...
#include <string.h>		// memcpy(), memcmp(), ...
#include <stdio.h>		// printf(), ...
#include <stdarg.h>		// va_list
#include <sys/time.h>		// gettimeofday()
#include <syslog.h>		// vsyslog()
#include <stdbool.h>

#include <libubus.h>
#include <libubox/blobmsg_json.h>
#include <easy/easy.h>

// The following variable is used to set which "print_std_DEBUG_*()"
// functions should be ignored:
//
//   0 => Only print ERROR messages
//   1 => Print ERROR and WARNING messages
//   2 => Print ERROR, WARNING and INFO messages
//   3 => Print ERROR, WARNING, INFO and DETAIL messages
static uint8_t gs_verbosity_level;

// We will use this variable to save the instant when "init_time()" was
// called. This way we sill be able to get relative timestamps later when
// someone calls "get_time()"
static struct timeval tv_begin;

void print_callback_blob(struct blob_buf *bb,
			 PRINT_WRITER write_function __attribute__ ((unused)),
			 const char *prefix __attribute__ ((unused)),
			 INT8U size,
			 const char *name,
			 const char *fmt,
			 void *data)
{
#define BUFFER_SIZE (256)
	char buffer[BUFFER_SIZE] = { 0x00 };

	if (!bb) {
		INFO("Blob is not initialized!!!");
		return;
	}

	if (0 != memcmp(fmt, "%s", 3)) {
		if (size * 2 > BUFFER_SIZE) {
			INFO("Buffer is smaller than data !!!");
			return;
		}
	}
	if (0 == memcmp(fmt, "%s", 3)) {
		blobmsg_add_string(bb, name, (char *)data);
	} else if (0 == memcmp(fmt, "%d", 3)) {
		switch (size) {
		case 1:
			{
				blobmsg_add_u32(bb, name, *(INT8U *) data);
				break;
			}
		case 2:
			{
				blobmsg_add_u32(bb, name, *(INT16U *) data);
				break;
			}
		case 4:
			{
				blobmsg_add_u32(bb, name, *(INT32U *) data);
				break;
			}
		case 6:
			{
				hwaddr_ntoa((INT8U *) data, buffer);
				blobmsg_add_string(bb, name, buffer);
				break;
			}
		default:
			{
				btostr((INT8U *) data, size, buffer);
				blobmsg_add_string(bb, name, buffer);
			}
		}
	} else if (0 == memcmp(fmt, "0x%02x", 7)) {
		if (size == 6) {
			hwaddr_ntoa((INT8U *) data, buffer);
			blobmsg_add_string(bb, name, buffer);
		} else {
			btostr((INT8U *) data, size, buffer);
			blobmsg_add_string(bb, name, buffer);
		}
	} else if (0 == memcmp(fmt, "0x%04x", 7)) {
		blobmsg_add_u32(bb, name, *(INT16U *) data);
	} else {
		btostr((INT8U *) data, size, buffer);
		blobmsg_add_string(bb, name, buffer);
	}
}

void print_callback(struct blob_buf *bb, PRINT_WRITER write_function, const char *prefix, INT8U size, const char *name, const char *fmt, void *p)
{

	if (bb) {
		INFO("Wrong callback function");
		return;
	}
	if (0 == memcmp(fmt, "%s", 3)) {
		// Strings are printed with triple quotes surrounding them
		//
		write_function("%s%s: \"\"\"%s\"\"\"\n", prefix, name, p);
		return;
	} else if (0 == memcmp(fmt, "%ipv4", 6)) {
		// This is needed because "size == 4" on IPv4 addresses, but we don't
		// want them to be treated as 4 bytes integers, so we change the
		// "fmt" to "%d" and do *not* returns (so that the IPv4 ends up being
		// printed as a sequence of bytes.
		//
		fmt = "%d";
	} else {
#define FMT_LINE_SIZE 20
		char fmt_line[FMT_LINE_SIZE];

		fmt_line[0] = 0x0;

		PLATFORM_SNPRINTF(fmt_line, FMT_LINE_SIZE - 1, "%%s%%s: %s\n", fmt);
		fmt_line[FMT_LINE_SIZE - 1] = 0x0;

		if (size == 1) {
			write_function(fmt_line, prefix, name, *(INT8U *) p);

			return;
		} else if (size == 2) {
			write_function(fmt_line, prefix, name, *(INT16U *) p);

			return;
		} else if (size == 4) {
			write_function(fmt_line, prefix, name, *(INT32U *) p);

			return;
		}
	}

	// If we get here, it's either an IPv4 address or a sequence of bytes
	//
	{
#define AUX1_SIZE 200		// Store a whole output line
#define AUX2_SIZE  20		// Store a fmt conversion

		INT16U i, j;

		char aux1[AUX1_SIZE];
		char aux2[AUX2_SIZE];

		INT16U space_left = AUX1_SIZE - 1;

		aux1[0] = 0x00;
		strncat(aux1, "%s%s: ", AUX1_SIZE);

		for (i = 0; i < size; i++) {
			// Write one element to aux2
			//
			PLATFORM_SNPRINTF(aux2, AUX2_SIZE - 1, fmt, *((INT8U *) p + i));

			// Obtain its length
			//
			for (j = 0; j < AUX2_SIZE; j++) {
				if (aux2[j] == 0x00) {
					break;
				}
			}

			// 'j' contains the number of chars in "aux2"
			// Check if there is enough space left in "aux1"
			//
			//   NOTE: The "+2" is because we are going to append to "aux1"
			//         the contents of "aux2" plus a ", " string (which is
			//         three chars long)
			//         The second "+2" is because of the final "\n"
			//
			if (j + 2 + 2 > space_left) {
				// No more space left
				//
				aux1[AUX1_SIZE - 6] = '.';
				aux1[AUX1_SIZE - 5] = '.';
				aux1[AUX1_SIZE - 4] = '.';
				aux1[AUX1_SIZE - 3] = '.';
				aux1[AUX1_SIZE - 2] = '.';
				aux1[AUX1_SIZE - 1] = 0x00;
				break;
			}
			// Append string to "aux1"
			//
			strncat(aux1, aux2, j);
			strncat(aux1, ", ", 2);
			space_left -= (j + 2);
		}

		write_function(aux1, prefix, name);
	}

	return;
}

void PLATFORM_SNPRINTF(char *dest, INT32U n, const char *format, ...)
{
	va_list arglist;

	va_start(arglist, format);
	vsnprintf(dest, n, format, arglist);
	va_end(arglist);

	return;
}

void print_std(const char *format, ...)
{
	va_list arglist;

	if (gs_verbosity_level == 0) {
		return;
	}

	va_start(arglist, format);
	vsyslog(LOG_DEBUG|LOG_CONS, format, arglist);
	va_end(arglist);

	return;
}

void set_print_level(int level)
{
	gs_verbosity_level = level;
}

void print_error(const char *format, ...)
{
	va_list arglist;

	if (gs_verbosity_level == 0) {
		return;
	}

	va_start(arglist, format);
	vsyslog(LOG_ERR|LOG_CONS|LOG_PERROR, format, arglist);
	va_end(arglist);

	return;
}

void print_warning(const char *format, ...)
{
	va_list arglist;

	if (gs_verbosity_level < 2) {
		return;
	}

	va_start(arglist, format);
	vsyslog(LOG_WARNING|LOG_CONS|LOG_PERROR, format, arglist);
	va_end(arglist);
	return;
}

void print_info(const char *format, ...)
{
	va_list arglist;

	if (gs_verbosity_level < 3) {
		return;
	}

	va_start(arglist, format);
	vsyslog(LOG_INFO|LOG_CONS, format, arglist);
	va_end(arglist);

	return;
}

// suppress output
void print_null(__attribute__ ((unused))
		const char *format, ...)
{
	return;
}

void print_debug(const char *format, ...)
{
	va_list arglist;

	if (gs_verbosity_level < 4) {
		return;
	}

	va_start(arglist, format);
	vsyslog(LOG_DEBUG|LOG_CONS, format, arglist);
	va_end(arglist);

	return;
}

INT32U get_time(void)
{
	struct timeval tv_end;
	INT32U diff;

	gettimeofday(&tv_end, NULL);

	diff = (tv_end.tv_usec - tv_begin.tv_usec) / 1000 + (tv_end.tv_sec - tv_begin.tv_sec) * 1000;

	return diff;
}

INT8U init_time(void)
{
	// Call "_timeval_print()" for the first time so that the initialization
	// time is saved for future reference.
	gettimeofday(&tv_begin, NULL);

	return 1;
}

INT16U getNextMid(void)
{
	static INT16U mid;
	static INT8U first_time = 1;

	if (first_time == 1) {
		// Start with a random MID. The standard is not clear about this, but
		// I think a random number is better than simply choosing zero, to
		// avoid start up problems (ex: one node boots and after a short time
		// it is reset and starts making use of the same MIDs all over again,
		// which will probably be ignored by other nodes, thinking they have
		// already processed these messages in the past)
		//
		DEBUG("Generating random MID for first time");
		first_time = 0;
		PLATFORM_GET_RANDOM_BYTES((INT8U *) &mid, sizeof(INT16U));
	} else {
		mid++;
	}

	return mid;
}
