/*
 * utils.h: Common utility functions
 *
 * Author: Vivek Dutta <vivek.dutta@iopsys.eu>
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef _UTILS_H_
#define _UTILS_H_
// This is an auxiliary function which is used when calling the "visit_*()"
// family of functions so that the contents of the memory structures can be
// printed on screen for debugging/logging purposes.
//
// The documentation for any of the "visit_*()" function explains what this
// functions should do and look like.
//
#include "common_defs.h"
#include "common.h"

#include <stdarg.h>		// va_list
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <libubox/blobmsg.h>

  // NOTE: This is part of the C standard, thus *all* platforms should have it
  // available... and that's why this include can exist in this "platform
  // independent" file

////////////////////////////////////////////////////////////////////////////////
// Hardware stuff
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// Typical libc stuff
////////////////////////////////////////////////////////////////////////////////

#define BOOL_STRCMP(x, y) (strncmp(x, y, strlen(x))?false:true)
#define BOOL_STRCCMP(x, y) (strncasecmp(x, y, strlen(x))?false:true)

#define DEBUG(fmt, args...) \
do { \
	print_debug(fmt, ##args); \
} while (0)

#define INFO(fmt, args...) \
do { \
	print_info(fmt, ##args); \
} while (0)

#define PRINT(fmt, args...) \
do { \
	print_std(fmt, ##args); \
} while (0)

#define ERROR(fmt, args...) \
do { \
	print_error("[%s:%d] " fmt, __func__, __LINE__, ##args); \
} while (0)

#define WARNING(fmt, args...) \
do { \
	print_warning("[%s:%d] " fmt, __func__, __LINE__, ##args); \
} while (0)

typedef void (*PRINT_WRITER) (const char *fmt, ...);
typedef void (*PRINT_CALLBACK) (struct blob_buf *bb,
				PRINT_WRITER writer,
				const char *prefix,
				INT8U size,
				const char *name,
				const char *fmt,
				void *p);

// Output to a string (see 'man 3 snprintf' on any Linux box)
//
void PLATFORM_SNPRINTF(char *dest, INT32U n, const char *format, ...);

// Output the provided format string (see 'man 3 printf' on any Linux box)
//
void print_std(const char *format, ...);
void print_null(const char *format, ...);

// Same as 'PLATFORM_PRINTF', but the message will only be processed if the
// platform has the pertaining debug level enabled
//
void print_error(const char *format, ...);
void print_warning(const char *format, ...);
void print_info(const char *format, ...);
void print_debug(const char *format, ...);

// Return the number of milliseconds ellapsed since the program started
//
INT32U get_time(void);

// Used to set the verbosity of the previous functions:
//
//   0 => Only print ERROR messages
//   1 => Print ERROR and WARNING messages
//   2 => Print ERROR, WARNING and INFO messages
//   3 => Print ERROR, WARNING, INFO and DETAIL messages
//
void set_print_level(int level);

////////////////////////////////////////////////////////////////////////////////
// Misc stuff
////////////////////////////////////////////////////////////////////////////////

// [PLATFORM PORTING NOTE]
//   Depending on what other platform headers you have included up until this
//   point, 'NULL' might or might not be defined. If so, define it here
//
#ifndef NULL
#define NULL (0x0)
#endif

INT8U init_time(void);

void print_callback(struct blob_buf *bb,
		    PRINT_WRITER write_function,
		    const char *prefix,
		    INT8U size,
		    const char *name,
		    const char *fmt,
		    void *p);

void print_callback_blob(struct blob_buf *bb,
			 PRINT_WRITER write_function,
			 const char *prefix,
			 INT8U size,
			 const char *name,
			 const char *fmt,
			 void *data);

INT16U getNextMid(void);
#endif
