/*
 * utils_uci.c: Common uci utility functions
 *
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "utils.h"
#include "utils_uci.h"
#include "al_ubus.h"
#include "platform_interfaces.h"
#include <uci.h>

#define UCI_DB_PATH "/lib/db/config"

bool get_boolean_val(const char *value)
{
	if (!value)
		return false;

	if (0 == strncasecmp(value, "true", 4) ||
	    '1' == value[0] ||
	    0 == strncasecmp(value, "on", 2) ||
	    0 == strncasecmp(value, "yes", 3) ||
	    0 == strncasecmp(value, "enabled", 7))
		return true;

	return false;
}

bool uci_get_option_db(const char *pkg_name, const char *sec_type, const char *option, char *val, size_t val_len)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;

	if (pkg_name == NULL || sec_type == NULL ||
	    option == NULL || val == NULL || val_len == 0)
		return false;

	memset(val, 0, val_len);

	ctx = uci_alloc_context();
	if (!ctx)
		return false;

	uci_set_confdir(ctx, UCI_DB_PATH);
	if (uci_load(ctx, pkg_name, &pkg)) {
		uci_free_context(ctx);
		return false;
	}

	struct uci_section *s = uci_lookup_section(ctx, pkg, sec_type);
	struct uci_option *o = uci_lookup_option(ctx, s, option);
	switch(o->type) {
	case UCI_TYPE_STRING:
		snprintf(val, val_len, "%s", o->v.string);
		break;
	case UCI_TYPE_LIST:
		uci_foreach_element(&o->v.list, e) {
			strncat(val, e->name, val_len);
			strncat(val, ",", val_len);
		}
		val[strlen(val) - 1] = '\0';
		break;
	}

	uci_unload(ctx, pkg);
	uci_free_context(ctx);

	return true;
}

bool uci_get_option(const char *pkg_name, const char *sec_type, const char *option, char *val, size_t val_len)
{
	struct uci_context *get_ctx;
	struct uci_package *pkg;
	struct uci_element *e;

	if (pkg_name == NULL || sec_type == NULL ||
	    option == NULL || val == NULL || val_len == 0)
		return false;

	memset(val, 0, val_len);

	get_ctx = uci_alloc_context();
	if (!get_ctx)
		return false;

	if (uci_load(get_ctx, pkg_name, &pkg)) {
		uci_free_context(get_ctx);
		return false;
	}

	struct uci_section *s = uci_lookup_section(get_ctx, pkg, sec_type);
	struct uci_option *o = uci_lookup_option(get_ctx, s, option);
	switch(o->type) {
	case UCI_TYPE_STRING:
		snprintf(val, val_len, "%s", o->v.string);
		break;
	case UCI_TYPE_LIST:
		uci_foreach_element(&o->v.list, e) {
			strncat(val, e->name, val_len);
			strncat(val, ",", val_len);
		}
		val[strlen(val) - 1] = '\0';
		break;
	}

	uci_unload(get_ctx, pkg);
	uci_free_context(get_ctx);

	return true;
}

bool uci_set_option(const char *pkg_name, const char *section, char *option, char *value)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;

	if (pkg_name == NULL || section == NULL ||
	    option == NULL || value == NULL)
		return false;

	ctx = uci_alloc_context();
	if (!ctx)
		return false;

	if (uci_load(ctx, pkg_name, &pkg)) {
		uci_free_context(ctx);
		return false;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, section)) {
			struct uci_option *opt = uci_lookup_option(ctx, s, option);

			if (!opt || opt->type != UCI_TYPE_STRING)
				break;

			struct uci_ptr ptr = {0};

			ptr.value = value;
			ptr.package = pkg_name;
			ptr.section = s->e.name;
			ptr.option = option;
			ptr.target = UCI_TYPE_OPTION;
			if (uci_lookup_ptr(ctx, &ptr, NULL, false) != UCI_OK || !UCI_LOOKUP_COMPLETE) {
				printf("exit from here");
				break;
			}
			if (uci_set(ctx, &ptr) == UCI_OK) {
				uci_save(ctx, ptr.p);
				uci_commit(ctx, &ptr.p, false);
				printf("uci set done ###\n");
				break;
			}
		}
	}
	uci_unload(ctx, pkg);
	uci_free_context(ctx);
	return true;
}

void validate_add_ifname(const char *ifname, const char *media)
{
	if (ifname == NULL || media == NULL)
		return;

	WARNING("ifname(%s), media(%s)", ifname, media);
	if (strncasecmp(media, "bridge", 6) == 0) {
		add_interfaces_from_bridge(ifname);
	} else {
		addInterface(ifname);
	}
}

bool uci_configure_al_iface()
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;
	const char *ifname, *media;
	bool enabled;

	ctx = uci_alloc_context();
	if (!ctx)
		return false;

	if (uci_load(ctx, "ieee1905", &pkg)) {
		uci_free_context(ctx);
		return false;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, "al-iface")) {
			enabled = get_boolean_val(uci_lookup_option_string(ctx, s, "enabled"));

			ifname = uci_lookup_option_string(ctx, s, "ifname");
			media = uci_lookup_option_string(ctx, s, "media");
			if (enabled) {
				validate_add_ifname(ifname, media);
			} else {
				WARNING("interface (%s) is not enabled", ifname);
			}

		}
	}
	uci_unload(ctx, pkg);
	uci_free_context(ctx);
	return true;
}

void uci_free_list_of_1905_interfaces(char **x, INT8U nr)
{
	uint8_t i;

	for (i = 0; i < nr; ++i) {
		FREE(x[i]);
	}

	FREE(x);
	return;
}

char **uci_get_list_of_1905_interfaces(uint8_t *nr)
{
	struct uci_context *uci_ctx;
	struct uci_package *pkg;
	struct uci_element *e;
	const char *ifname;
	bool enabled;
	char **list = NULL;
	uint8_t count;

	*nr = 0;
	uci_ctx = uci_alloc_context();
	if (!uci_ctx)
		return NULL;

	if (uci_load(uci_ctx, "ieee1905", &pkg)) {
		uci_free_context(uci_ctx);
		return NULL;
	}

	count = 0;
	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, "al-iface")) {
			enabled = get_boolean_val(uci_lookup_option_string(uci_ctx, s, "enabled"));

			ifname = uci_lookup_option_string(uci_ctx, s, "ifname");
			if (enabled) {
				if (count)
					list = (char **) realloc(list, sizeof(char *));
				else
					list = (char **) malloc(sizeof(char *));

				list[count] = strdup(ifname);
				count++;
			}

		}
	}
	uci_unload(uci_ctx, pkg);
	uci_free_context(uci_ctx);
	*nr = count;

	return list;
}

const char *get_ap_radio_name(char *ap_name)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;
	const char *val = NULL, *radio_name = NULL;

	if (ap_name == NULL)
		return NULL;

	ctx = uci_alloc_context();
	if (!ctx)
		return NULL;

	if (uci_load(ctx, "wireless", &pkg)) {
		uci_free_context(ctx);
		return NULL;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, UCI_WLAN_IFACE)) {
			val = uci_lookup_option_string(ctx, s, "ifname");
			if (strcmp(val, ap_name) == 0) {
				radio_name = uci_lookup_option_string(ctx, s, "device");
				break;
			}
		}
	}

	uci_unload(ctx, pkg);
	uci_free_context(ctx);
	return radio_name ? radio_name : NULL;
}

bool uci_set_wireless_interface_option(char *package_name, char *ifname, char *option, char *value)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;

	if (package_name == NULL || NULL == ifname || NULL == option || NULL == value)
		return false;

	ctx = uci_alloc_context();
	if (!ctx)
		return false;

	if (uci_load(ctx, package_name, &pkg)) {
		uci_free_context(ctx);
		return false;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, UCI_WLAN_IFACE)) {
			struct uci_option *opt = uci_lookup_option(ctx, s, "ifname");

			if (!opt || opt->type != UCI_TYPE_STRING)
				continue;
			if (strcmp(opt->v.string, ifname) == 0) {
				struct uci_ptr ptr = {0};

				ptr.value = value;
				ptr.package = package_name;
				ptr.section = s->e.name;
				ptr.option = option;
				ptr.target = UCI_TYPE_OPTION;
				if (uci_lookup_ptr(ctx, &ptr, NULL, false) != UCI_OK || !UCI_LOOKUP_COMPLETE) {
					break;
				}
				if (uci_set(ctx, &ptr) == UCI_OK) {
					uci_save(ctx, ptr.p);
				}
				break;
			}
		}
	}
	uci_commit(ctx, &pkg, false);
	uci_unload(ctx, pkg);
	uci_free_context(ctx);
	return false;
}

bool uci_del_wireless_interface_sec(char *package_name, char *ifname)
{
	struct uci_context *del_ctx;
	struct uci_package *pkg;
	struct uci_element *e;

	if (package_name == NULL || ifname == NULL)
		return false;

	del_ctx = uci_alloc_context();
	if (!del_ctx)
		return false;

	if (uci_load(del_ctx, package_name, &pkg)) {
		uci_free_context(del_ctx);
		return false;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, UCI_WLAN_IFACE)) {
			struct uci_option *opt = uci_lookup_option(del_ctx, s, "ifname");

			if (!opt || opt->type != UCI_TYPE_STRING)
				continue;

			if (strcmp(opt->v.string, ifname) == 0) {
				struct uci_ptr ptr = {0};

				ptr.package = package_name;
				ptr.section = s->e.name;
				ptr.target = UCI_TYPE_SECTION;
				if (uci_lookup_ptr(del_ctx, &ptr, NULL, false) != UCI_OK || !UCI_LOOKUP_COMPLETE) {
					break;
				}
				if (uci_delete(del_ctx, &ptr) == UCI_OK) {
					uci_save(del_ctx, ptr.p);
				}
				break;
			}
		}
	}
	uci_commit(del_ctx, &pkg, false);
	uci_unload(del_ctx, pkg);
	uci_free_context(del_ctx);
	return false;
}
bool uci_get_wireless_interface_option(char *package_name, char *ifname, char *option, char *value, size_t vlen)
{
	bool ret = false;
	struct uci_context *uci_ctx;
	struct uci_package *pkg;
	struct uci_element *e;

	if (NULL == ifname || NULL == option ||
	    value == NULL || vlen == 0 ||
	    package_name == NULL)
		return false;

	uci_ctx = uci_alloc_context();
	if (!uci_ctx)
		return false;

	if (uci_load(uci_ctx, package_name, &pkg)) {
		uci_free_context(uci_ctx);
		return false;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, UCI_WLAN_IFACE)) {
			struct uci_option *opt = uci_lookup_option(uci_ctx, s, "ifname");

			if (!opt || opt->type != UCI_TYPE_STRING)
				continue;
			if (strcmp(opt->v.string, ifname) == 0) {
				opt = uci_lookup_option(uci_ctx, s, option);
				if (opt && opt->type == UCI_TYPE_STRING) {
					strncpy(value, opt->v.string, vlen);
					ret = true;
					break;
				}
			}
		}
	}
	uci_unload(uci_ctx, pkg);
	uci_free_context(uci_ctx);

	return ret;
}

bool uci_check_wifi_iface(char *package_name, char *ifname)
{
	bool ret;
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;

	if (package_name == NULL || ifname == NULL)
		return false;

	ctx = uci_alloc_context();
	if (!ctx)
		return false;

	if (uci_load(ctx, package_name, &pkg)) {
		uci_free_context(ctx);
		return false;
	}

	ret = false;
	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, UCI_WLAN_IFACE)) {
			struct uci_option *opt = uci_lookup_option(ctx, s, "ifname");

			if (!opt || opt->type != UCI_TYPE_STRING)
				continue;
			if (strcmp(opt->v.string, ifname) == 0) {
				ret = true;
				break;
			}
		}
	}
	uci_unload(ctx, pkg);
	uci_free_context(ctx);

	return ret;
}

bool uci_add_new_sec(char *package_name, char *sec_name)
{
	bool ret = false;
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_section *s;

	if (package_name == NULL || sec_name == NULL)
		return false;

	ctx = uci_alloc_context();
	if (!ctx)
		return false;

	if (uci_load(ctx, package_name, &pkg)) {
		uci_free_context(ctx);
		return false;
	}

	if (uci_add_section(ctx, pkg, sec_name, &s) == UCI_OK) {
		if (uci_save(ctx, pkg) == UCI_OK) {
			ret = true;
		}
	}

	uci_unload(ctx, pkg);
	uci_free_context(ctx);

	return ret;
}

bool uci_add_wireless_iface_sec(char *package_name, char *interface_name)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_section *s;

	if (interface_name == NULL ||
	    package_name == NULL)
		return false;

	ctx = uci_alloc_context();
	if (!ctx)
		return false;

	if (uci_load(ctx, package_name, &pkg)) {
		uci_free_context(ctx);
		return false;
	}

	if (uci_add_section(ctx, pkg, UCI_WLAN_IFACE, &s) == UCI_OK) {
		struct uci_ptr ptr = {0};

		if (uci_save(ctx, pkg) != UCI_OK) {
			return false;
		}

		ptr.value = interface_name;
		ptr.package = package_name;
		ptr.section = s->e.name;
		ptr.option = "ifname";
		ptr.target = UCI_TYPE_OPTION;
		uci_lookup_ptr(ctx, &ptr, NULL, false);

		uci_set(ctx, &ptr);
		uci_save(ctx, ptr.p);
		uci_commit(ctx, &pkg, false);
	}

	uci_unload(ctx, pkg);
	uci_free_context(ctx);

	return true;
}
