/*
 * al_wsc.c: IEEE-1905 AL wsc handling
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "utils.h"

#include "al_wsc.h"
#include "al_datamodel.h"

#include "platform_crypto.h"
#include "platform_interfaces.h"
#include "platform_interfaces_ubus.h"

////////////////////////////////////////////////////////////////////////////////
// Private functions and data
////////////////////////////////////////////////////////////////////////////////

// "defines" used to fill the M1 and M2 message fields
//
#define ATTR_VERSION           (0x104a)
#define ATTR_MSG_TYPE          (0x1022)
#define WPS_M1                 (0x04)
#define WPS_M2                 (0x05)
#define ATTR_UUID_E            (0x1047)
#define ATTR_UUID_R            (0x1048)
#define ATTR_MAC_ADDR          (0x1020)
#define ATTR_ENROLLEE_NONCE    (0x101a)
#define ATTR_REGISTRAR_NONCE   (0x1039)
#define ATTR_PUBLIC_KEY        (0x1032)
#define ATTR_AUTH_TYPE_FLAGS   (0x1004)
#define WPS_AUTH_OPEN          (0x0001)
#define WPS_AUTH_WPAPSK        (0x0002)
#define WPS_AUTH_SHARED        (0x0004)	/* deprecated */
#define WPS_AUTH_WPA           (0x0008)
#define WPS_AUTH_WPA2          (0x0010)
#define WPS_AUTH_WPA2PSK       (0x0020)
#define ATTR_ENCR_TYPE_FLAGS   (0x1010)
#define WPS_ENCR_NONE          (0x0001)
#define WPS_ENCR_WEP           (0x0002)	/* deprecated */
#define WPS_ENCR_TKIP          (0x0004)
#define WPS_ENCR_AES           (0x0008)
#define ATTR_CONN_TYPE_FLAGS   (0x100d)
#define WPS_CONN_ESS           (0x01)
#define WPS_CONN_IBSS          (0x02)
#define ATTR_CONFIG_METHODS    (0x1008)
#define WPS_CONFIG_VIRT_PUSHBUTTON (0x0280)
#define WPS_CONFIG_PHY_PUSHBUTTON  (0x0480)
#define ATTR_WPS_STATE         (0x1044)
#define WPS_STATE_NOT_CONFIGURED (1)
#define WPS_STATE_CONFIGURED     (2)
#define ATTR_MANUFACTURER      (0x1021)
#define ATTR_MODEL_NAME        (0x1023)
#define ATTR_MODEL_NUMBER      (0x1024)
#define ATTR_SERIAL_NUMBER     (0x1042)
#define ATTR_PRIMARY_DEV_TYPE  (0x1054)
#define WPS_DEV_COMPUTER                           (1)
#define WPS_DEV_COMPUTER_PC                       (1)
#define WPS_DEV_COMPUTER_SERVER                   (2)
#define WPS_DEV_COMPUTER_MEDIA_CENTER             (3)
#define WPS_DEV_COMPUTER_ULTRA_MOBILE             (4)
#define WPS_DEV_COMPUTER_NOTEBOOK                 (5)
#define WPS_DEV_COMPUTER_DESKTOP                  (6)
#define WPS_DEV_COMPUTER_MID                      (7)
#define WPS_DEV_COMPUTER_NETBOOK                  (8)
#define WPS_DEV_COMPUTER_TABLET                   (9)
#define WPS_DEV_INPUT                              (2)
#define WPS_DEV_INPUT_KEYBOARD                    (1)
#define WPS_DEV_INPUT_MOUSE                       (2)
#define WPS_DEV_INPUT_JOYSTICK                    (3)
#define WPS_DEV_INPUT_TRACKBALL                   (4)
#define WPS_DEV_INPUT_GAMING                      (5)
#define WPS_DEV_INPUT_REMOTE                      (6)
#define WPS_DEV_INPUT_TOUCHSCREEN                 (7)
#define WPS_DEV_INPUT_BIOMETRIC_READER            (8)
#define WPS_DEV_INPUT_BARCODE_READER              (9)
#define WPS_DEV_PRINTER                            (3)
#define WPS_DEV_PRINTER_PRINTER                   (1)
#define WPS_DEV_PRINTER_SCANNER                   (2)
#define WPS_DEV_PRINTER_FAX                       (3)
#define WPS_DEV_PRINTER_COPIER                    (4)
#define WPS_DEV_PRINTER_ALL_IN_ONE                (5)
#define WPS_DEV_CAMERA                             (4)
#define WPS_DEV_CAMERA_DIGITAL_STILL_CAMERA       (1)
#define WPS_DEV_CAMERA_VIDEO                      (2)
#define WPS_DEV_CAMERA_WEB                        (3)
#define WPS_DEV_CAMERA_SECURITY                   (4)
#define WPS_DEV_STORAGE                            (5)
#define WPS_DEV_STORAGE_NAS                       (1)
#define WPS_DEV_NETWORK_INFRA                      (6)
#define WPS_DEV_NETWORK_INFRA_AP                  (1)
#define WPS_DEV_NETWORK_INFRA_ROUTER              (2)
#define WPS_DEV_NETWORK_INFRA_SWITCH              (3)
#define WPS_DEV_NETWORK_INFRA_GATEWAY             (4)
#define WPS_DEV_NETWORK_INFRA_BRIDGE              (5)
#define WPS_DEV_DISPLAY                            (7)
#define WPS_DEV_DISPLAY_TV                        (1)
#define WPS_DEV_DISPLAY_PICTURE_FRAME             (2)
#define WPS_DEV_DISPLAY_PROJECTOR                 (3)
#define WPS_DEV_DISPLAY_MONITOR                   (4)
#define WPS_DEV_MULTIMEDIA                         (8)
#define WPS_DEV_MULTIMEDIA_DAR                    (1)
#define WPS_DEV_MULTIMEDIA_PVR                    (2)
#define WPS_DEV_MULTIMEDIA_MCX                    (3)
#define WPS_DEV_MULTIMEDIA_SET_TOP_BOX            (4)
#define WPS_DEV_MULTIMEDIA_MEDIA_SERVER           (5)
#define WPS_DEV_MULTIMEDIA_PORTABLE_VIDEO_PLAYER  (6)
#define WPS_DEV_GAMING                             (9)
#define WPS_DEV_GAMING_XBOX                       (1)
#define WPS_DEV_GAMING_XBOX360                    (2)
#define WPS_DEV_GAMING_PLAYSTATION                (3)
#define WPS_DEV_GAMING_GAME_CONSOLE               (4)
#define WPS_DEV_GAMING_PORTABLE_DEVICE            (5)
#define WPS_DEV_PHONE                             (10)
#define WPS_DEV_PHONE_WINDOWS_MOBILE              (1)
#define WPS_DEV_PHONE_SINGLE_MODE                 (2)
#define WPS_DEV_PHONE_DUAL_MODE                   (3)
#define WPS_DEV_PHONE_SP_SINGLE_MODE              (4)
#define WPS_DEV_PHONE_SP_DUAL_MODE                (5)
#define WPS_DEV_AUDIO                             (11)
#define WPS_DEV_AUDIO_TUNER_RECV                  (1)
#define WPS_DEV_AUDIO_SPEAKERS                    (2)
#define WPS_DEV_AUDIO_PMP                         (3)
#define WPS_DEV_AUDIO_HEADSET                     (4)
#define WPS_DEV_AUDIO_HEADPHONES                  (5)
#define WPS_DEV_AUDIO_MICROPHONE                  (6)
#define WPS_DEV_AUDIO_HOME_THEATRE                (7)
#define ATTR_DEV_NAME          (0x1011)
#define ATTR_RF_BANDS          (0x103c)
#define WPS_RF_24GHZ           (0x01)
#define WPS_RF_50GHZ           (0x02)
#define WPS_RF_60GHZ           (0x04)
#define ATTR_ASSOC_STATE       (0x1002)
#define WPS_ASSOC_NOT_ASSOC     (0)
#define WPS_ASSOC_CONN_SUCCESS  (1)
#define ATTR_DEV_PASSWORD_ID   (0x1012)
#define DEV_PW_PUSHBUTTON      (0x0004)
#define ATTR_CONFIG_ERROR      (0x1009)
#define WPS_CFG_NO_ERROR       (0)
#define ATTR_OS_VERSION        (0x102d)
#define ATTR_VENDOR_EXTENSION  (0x1049)
#define WPS_VENDOR_ID_WFA_1    (0x00)
#define WPS_VENDOR_ID_WFA_2    (0x37)
#define WPS_VENDOR_ID_WFA_3    (0x2A)
#define WFA_ELEM_VERSION2      (0x00)
#define WPS_VERSION            (0x20)
#define ATTR_SSID              (0x1045)
#define ATTR_AUTH_TYPE         (0x1003)
#define ATTR_ENCR_TYPE         (0x100f)
#define ATTR_NETWORK_KEY       (0x1027)
#define ATTR_KEY_WRAP_AUTH     (0x101e)
#define ATTR_ENCR_SETTINGS     (0x1018)
#define ATTR_AUTHENTICATOR     (0x1005)

// Keys sizes
#define WPS_AUTHKEY_LEN    32
#define WPS_KEYWRAPKEY_LEN 16
#define WPS_EMSK_LEN       32

struct wscKey {
	INT8U *key;
	INT32U key_len;
	INT8U mac[6];
};

// Global variable to save the latest M1 message created
struct wsc_data {
	char *ifname;
	INT8U *m1;
	INT16U m1_size;
	struct wscKey *key;
};


static struct wsc_data *gs_wsc_data;
static INT8U gs_wsc_data_nr;

static void get_auth_types(uint16_t auth_mode, uint16_t *auth_types)
{
	if (IEEE80211_AUTH_MODE_OPEN & auth_mode) {
		*auth_types |= WPS_AUTH_OPEN;
	}
	if (IEEE80211_AUTH_MODE_WPA & auth_mode) {
		*auth_types |= WPS_AUTH_WPA;
	}
	if (IEEE80211_AUTH_MODE_WPAPSK & auth_mode) {
		*auth_types |= WPS_AUTH_WPAPSK;
	}
	if (IEEE80211_AUTH_MODE_WPA2 & auth_mode) {
		*auth_types |= WPS_AUTH_WPA2;
	}
	if (IEEE80211_AUTH_MODE_WPA2PSK & auth_mode) {
		*auth_types |= WPS_AUTH_WPA2PSK;
	}
}

// This is the key derivation function used in the WPS standard to obtain a
// final hash that is later used for encryption.
//
// The output is stored in the memory buffer pointed by 'res', which must be
// "SHA256_MAC_LEN" bytes long (ie. 're_len' must always be "SHA256_MAC_LEN",
// even if it is an input argument)
//
void _wps_key_derivation_function(INT8U *key, INT8U *label_prefix, INT32U label_prefix_len, char *label, INT8U *res, INT32U res_len)
{
	INT8U i_buf[4];
	INT8U key_bits[4];

	INT8U *addr[4];
	INT32U len[4];

	INT32U i, iter;

	INT8U hash[SHA256_MAC_LEN];
	INT8U *opos;

	INT32U left;

	INT8U *p;
	INT32U aux;

	aux = res_len * 8;
	p = key_bits;

	_I4B(&aux, &p);

	addr[0] = i_buf;
	addr[1] = label_prefix;
	addr[2] = (INT8U *) label;
	addr[3] = key_bits;
	len[0] = sizeof(i_buf);
	len[1] = label_prefix_len;
	len[2] = strlen(label);
	len[3] = sizeof(key_bits);

	iter = (res_len + SHA256_MAC_LEN - 1) / SHA256_MAC_LEN;
	opos = res;
	left = res_len;

	for (i = 1; i <= iter; i++) {
		p = i_buf;
		_I4B(&i, &p);

		PLATFORM_HMAC_SHA256(key, SHA256_MAC_LEN, 4, addr, len, hash);

		if (i < iter) {
			memcpy(opos, hash, SHA256_MAC_LEN);
			opos += SHA256_MAC_LEN;
			left -= SHA256_MAC_LEN;
		} else {
			memcpy(opos, hash, left);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// Public functions
////////////////////////////////////////////////////////////////////////////////
bool wsc_init(void)
{
	char **ifs_names = NULL;
	INT8U ifs_nr;
	INT8U i;

	ifs_names = PLATFORM_GET_LIST_OF_1905_INTERFACES(&ifs_nr);
	if (ifs_names == NULL || ifs_nr == 0) {
		ERROR("No interfaces detected");
		return false;
	}

	for (i = 0; i < ifs_nr; i++) {
		INT8U *m1 = NULL;
		INT16U m1_size = 0;
		void *key = NULL;
		uint8_t intf_type = get_interface_type(ifs_names[i]);

		if (intf_type != INTF_TYPE_WIFI)
			continue;

		if (gs_wsc_data_nr == 0)
			CALLOC(gs_wsc_data, gs_wsc_data_nr + 1);
		else
			REALLOC(gs_wsc_data, gs_wsc_data_nr + 1);

		if (wscBuildM1(ifs_names[i], &m1, &m1_size, &key)) {
			gs_wsc_data[gs_wsc_data_nr].ifname = strdup(ifs_names[i]);
			gs_wsc_data[gs_wsc_data_nr].m1 = m1;
			gs_wsc_data[gs_wsc_data_nr].m1_size = m1_size;
			gs_wsc_data[gs_wsc_data_nr].key = (struct wscKey *)key;
			gs_wsc_data_nr++;
		}
	}
	PLATFORM_FREE_LIST_OF_1905_INTERFACES(ifs_names, ifs_nr);

	return true;
}

void wsc_clean(void)
{
	INT8U index = 0;

	for (index = 0; index < gs_wsc_data_nr; index++) {
		FREE(gs_wsc_data[index].m1);
		FREE(gs_wsc_data[index].key);
	}
	gs_wsc_data_nr = 0;
	FREE(gs_wsc_data);
	gs_wsc_data = NULL;
}

bool get_wsc_m1_key(char *interface_name, INT8U **m1, INT16U *m1_size, void **key)
{
	INT8U index = 0;

	if (NULL == interface_name || NULL == m1 || NULL == m1_size || NULL == key) {
		WARNING("Invalid arguments!!");
		return false;
	}
	INFO("Have (%d) WCS M1 data in db", gs_wsc_data_nr);
	for (index = 0; index < gs_wsc_data_nr; index++) {
		if (strcmp(gs_wsc_data[index].ifname, interface_name) == 0) {
			*m1 = gs_wsc_data[index].m1;
			*m1_size = gs_wsc_data[index].m1_size;
			*key = gs_wsc_data[index].key;
			return true;
		}
	}
	return false;
}
//////////////////////////////////////// Enrollee functions ////////////////////
INT8U wscBuildM1(char *interface_name, INT8U **m1, INT16U *m1_size, void **key)
{
	INT8U *buffer;

	struct interfaceInfo *x;

	struct wscKey *private_key;

	INT8U *p;

	INT8U aux8;
	INT16U aux16;
	INT32U aux32;

	if (NULL == interface_name || NULL == m1 || NULL == m1_size || NULL == key) {
		WARNING("Invalid arguments !");
		return 0;
	}

	// First try to get the details from pre-build data
	if (get_wsc_m1_key(interface_name, m1, m1_size, key)) {
		INFO("WSC M1 entry present in DB returning from here");
		return 1;
	}

	x = PLATFORM_GET_1905_INTERFACE_INFO(interface_name);
	if (x == NULL) {
		WARNING("Could not retrieve info of interface %s", interface_name);
		return 0;
	}

	CALLOC(buffer, 1000);
	p = buffer;

	// VERSION
	{
		aux16 = ATTR_VERSION;
		_I2B(&aux16, &p);
		aux16 = 1;
		_I2B(&aux16, &p);
		aux8 = 0x10;
		_I1B(&aux8, &p);
	}

	// MESSAGE TYPE
	{
		aux16 = ATTR_MSG_TYPE;
		_I2B(&aux16, &p);
		aux16 = 1;
		_I2B(&aux16, &p);
		aux8 = WPS_M1;
		_I1B(&aux8, &p);
	}

	// UUID
	{
		aux16 = ATTR_UUID_E;
		_I2B(&aux16, &p);
		aux16 = 16;
		_I2B(&aux16, &p);
		_InB(x->uuid, &p, 16);
	}

	// MAC ADDRESS
	{
		aux16 = ATTR_MAC_ADDR;
		_I2B(&aux16, &p);
		aux16 = 6;
		_I2B(&aux16, &p);
		_InB(x->mac_address, &p, 6);
	}

	// ENROLLEE NONCE
	{
		INT8U enrollee_nonce[16];

		PLATFORM_GET_RANDOM_BYTES(enrollee_nonce, 16);

		aux16 = ATTR_ENROLLEE_NONCE;
		_I2B(&aux16, &p);
		aux16 = 16;
		_I2B(&aux16, &p);
		_InB(enrollee_nonce, &p, 16);
	}

	// PUBLIC KEY
	{
		INT8U *priv, *pub;
		INT16U priv_len = 0, pub_len = 0;

		PLATFORM_GENERATE_DH_KEY_PAIR(&priv, &priv_len, &pub, &pub_len);
		// TODO: ZERO PAD the pub key (doesn't seem to be really needed though)
		DEBUG("M1 plen|%d| publen|%d|", priv_len, pub_len);

		aux16 = ATTR_PUBLIC_KEY;
		_I2B(&aux16, &p);
		aux16 = pub_len;
		_I2B(&aux16, &p);
		_InB(pub, &p, pub_len);
		// The private key is one of the output arguments
		CALLOC(private_key, 1);
		CALLOC(private_key->key, priv_len);
		private_key->key_len = priv_len;
		if (priv_len > 0)
			memcpy(private_key->key, priv, priv_len);

		memcpy(private_key->mac, x->mac_address, 6);
		FREE(pub);
		FREE(priv);
	}

	// AUTHENTICATION TYPES
	{
		INT16U auth_types;

		auth_types = 0;

		if (x->interface_type_data.ieee80211.authentication_mode == 0) {
			ERROR("Unsupported authentication mode not sending M1 (%s)!!", interface_name);
			PLATFORM_FREE_1905_INTERFACE_INFO(x);
			return 0;
		}

		get_auth_types(x->interface_type_data.ieee80211.authentication_mode, &auth_types);

		aux16 = ATTR_AUTH_TYPE_FLAGS;
		_I2B(&aux16, &p);
		aux16 = 2;
		_I2B(&aux16, &p);
		_I2B(&auth_types, &p);
	}

	// ENCRYPTION TYPES
	{
		INT16U encryption_types;

		encryption_types = 0;

		if (IEEE80211_ENCRYPTION_MODE_NONE & x->interface_type_data.ieee80211.encryption_mode) {
			encryption_types |= WPS_ENCR_NONE;
		}
		if (IEEE80211_ENCRYPTION_MODE_TKIP & x->interface_type_data.ieee80211.encryption_mode) {
			encryption_types |= WPS_ENCR_TKIP;
		}
		if (IEEE80211_ENCRYPTION_MODE_AES & x->interface_type_data.ieee80211.encryption_mode) {
			encryption_types |= WPS_ENCR_AES;
		}

		aux16 = ATTR_ENCR_TYPE_FLAGS;
		_I2B(&aux16, &p);
		aux16 = 2;
		_I2B(&aux16, &p);
		_I2B(&encryption_types, &p);
	}

	// CONNECTION TYPES
	{
		// Two possible types: ESS or IBSS. In the 1905 context, enrollees will
		// always want to acts as "ESS" to create an "extended" network where
		// all APs share the same credentials as the registrar.

		aux16 = ATTR_CONN_TYPE_FLAGS;
		_I2B(&aux16, &p);
		aux16 = 1;
		_I2B(&aux16, &p);
		aux8 = WPS_CONN_ESS;
		_I1B(&aux8, &p);
	}

	// CONFIGURATION METHODS
	{
		// In the 1905 context, the configuration methods the AP is willing to
		// offer will always be these two

		aux16 = ATTR_CONFIG_METHODS;
		_I2B(&aux16, &p);
		aux16 = 2;
		_I2B(&aux16, &p);
		aux16 = WPS_CONFIG_PHY_PUSHBUTTON | WPS_CONFIG_VIRT_PUSHBUTTON;
		_I2B(&aux16, &p);
	}

	// WPS STATE
	{
		aux16 = ATTR_WPS_STATE;
		_I2B(&aux16, &p);
		aux16 = 1;
		_I2B(&aux16, &p);
		aux8 = WPS_STATE_NOT_CONFIGURED;
		_I1B(&aux8, &p);
	}

	// MANUFACTURER
	{
		aux16 = ATTR_MANUFACTURER;
		_I2B(&aux16, &p);
		aux16 = strlen(x->manufacturer_name);
		_I2B(&aux16, &p);
		_InB(x->manufacturer_name, &p, strlen(x->manufacturer_name));
	}

	// MODEL NAME
	{

		aux16 = ATTR_MODEL_NAME;
		_I2B(&aux16, &p);
		aux16 = strlen(x->model_name);
		_I2B(&aux16, &p);
		_InB(x->model_name, &p, strlen(x->model_name));
	}

	// MODEL NUMBER
	{
		aux16 = ATTR_MODEL_NUMBER;
		_I2B(&aux16, &p);
		aux16 = strlen(x->model_number);
		_I2B(&aux16, &p);
		_InB(x->model_number, &p, strlen(x->model_number));
	}

	// SERIAL NUMBER
	{
		aux16 = ATTR_SERIAL_NUMBER;
		_I2B(&aux16, &p);
		aux16 = strlen(x->serial_number);
		_I2B(&aux16, &p);
		_InB(x->serial_number, &p, strlen(x->serial_number));
	}

	// PRIMARY DEVICE TYPE
	{
		// In the 1905 context, they node sending a M1 message will always be
		// (at least) a "network router"

		INT8U oui[4] = { 0x00, 0x50, 0xf2, 0x00 };	// Fixed value from the
		// WSC spec

		aux16 = ATTR_PRIMARY_DEV_TYPE;
		_I2B(&aux16, &p);
		aux16 = 8;
		_I2B(&aux16, &p);
		aux16 = WPS_DEV_NETWORK_INFRA;
		_I2B(&aux16, &p);
		_InB(oui, &p, 4);
		aux16 = WPS_DEV_NETWORK_INFRA_ROUTER;
		_I2B(&aux16, &p);
	}

	// DEVICE NAME
	{
		aux16 = ATTR_DEV_NAME;
		_I2B(&aux16, &p);
		aux16 = strlen(x->device_name);
		_I2B(&aux16, &p);
		_InB(x->device_name, &p, strlen(x->device_name));
	}

	// RF BANDS
	{
		// TODO
		// We should here list all supported freq bands (2.4, 5.0, 60 GHz)...
		// however each interfaces is already "pre-configured" to one specific
		// freq band... thus we will always report back a single value (instead
		// of and OR'ed list).
		// This should probably be improved in the future.
		INT8U rf_bands;

		rf_bands = 0;

		if (MEDIA_TYPE_IEEE_802_11B_2_4_GHZ == x->interface_type ||
		    MEDIA_TYPE_IEEE_802_11G_2_4_GHZ == x->interface_type || MEDIA_TYPE_IEEE_802_11N_2_4_GHZ == x->interface_type) {
			rf_bands = WPS_RF_24GHZ;
		} else if (MEDIA_TYPE_IEEE_802_11A_5_GHZ == x->interface_type ||
			   MEDIA_TYPE_IEEE_802_11N_5_GHZ == x->interface_type || MEDIA_TYPE_IEEE_802_11AC_5_GHZ == x->interface_type) {
			rf_bands = WPS_RF_50GHZ;
		} else if (x->interface_type == MEDIA_TYPE_IEEE_802_11AD_60_GHZ) {
			rf_bands = WPS_RF_60GHZ;
		}

		aux16 = ATTR_RF_BANDS;
		_I2B(&aux16, &p);
		aux16 = 1;
		_I2B(&aux16, &p);
		_I1B(&rf_bands, &p);
	}

	// ASSOCIATION STATE
	{
		aux16 = ATTR_ASSOC_STATE;
		_I2B(&aux16, &p);
		aux16 = 2;
		_I2B(&aux16, &p);
		aux16 = WPS_ASSOC_NOT_ASSOC;
		_I2B(&aux16, &p);
	}

	// DEVICE PASSWORD ID
	{
		aux16 = ATTR_DEV_PASSWORD_ID;
		_I2B(&aux16, &p);
		aux16 = 2;
		_I2B(&aux16, &p);
		aux16 = DEV_PW_PUSHBUTTON;
		_I2B(&aux16, &p);
	}

	// CONFIG ERROR
	{
		aux16 = ATTR_CONFIG_ERROR;
		_I2B(&aux16, &p);
		aux16 = 2;
		_I2B(&aux16, &p);
		aux16 = WPS_CFG_NO_ERROR;
		_I2B(&aux16, &p);
	}

	// OS VERSION
	{
		// TODO: Fill with actual properties from the interface

		INT32U os_version = 0x00000001;

		aux16 = ATTR_OS_VERSION;
		_I2B(&aux16, &p);
		aux16 = 4;
		_I2B(&aux16, &p);
		aux32 = 0x80000000 | os_version;
		_I4B(&aux32, &p);
	}

	// VENDOR EXTENSIONS
	{
		aux16 = ATTR_VENDOR_EXTENSION;
		_I2B(&aux16, &p);
		aux16 = 6;
		_I2B(&aux16, &p);
		aux8 = WPS_VENDOR_ID_WFA_1;
		_I1B(&aux8, &p);
		aux8 = WPS_VENDOR_ID_WFA_2;
		_I1B(&aux8, &p);
		aux8 = WPS_VENDOR_ID_WFA_3;
		_I1B(&aux8, &p);
		aux8 = WFA_ELEM_VERSION2;
		_I1B(&aux8, &p);
		aux8 = 1;
		_I1B(&aux8, &p);
		aux8 = WPS_VERSION;
		_I1B(&aux8, &p);
	}

	PLATFORM_FREE_1905_INTERFACE_INFO(x);

	*m1 = buffer;
	*m1_size = abs(p - buffer);
	*key = private_key;

	return 1;
}


INT8U wscProcessM2(void *key, INT8U *m1, INT16U m1_size, INT8U *m2, INT16U m2_size)
{
	INT8U *p;
	struct wscKey *k;

	// "Useful data" we want to extract from M2
	INT8U ssid[64];
	INT8U ssid_present;
	INT8U bssid[6];
	INT8U bssid_present;
	INT16U auth_type;
	INT8U auth_type_present;
	INT16U encryption_type;
	INT8U encryption_type_present;
	INT8U network_key[64];
	INT8U network_key_present;

	// Keys we need to compute to authenticate and decrypt M2
	INT8U authkey[WPS_AUTHKEY_LEN];
	INT8U keywrapkey[WPS_KEYWRAPKEY_LEN];
	INT8U emsk[WPS_EMSK_LEN];

	if (key == NULL || m1 == NULL || m1_size == 0) {
		ERROR("Missing m1 or key");
		return 0;
	}

	if (m2 == NULL || m2_size == 0) {
		ERROR("Missing m2");
		return 0;
	}
	// "Intermediary data" we also need to extract from M2 to obtain the keys
	// that will let us decrypt the "useful data" from M2
	//
	INT8U *m2_nonce = NULL;
	INT8U m2_nonce_present;
	INT8U *m2_pubkey = NULL;
	INT8U m2_pubkey_present;
	INT16U m2_pubkey_len = 0;
	INT8U *m2_encrypted_settings = NULL;
	INT8U m2_encrypted_settings_present;
	INT16U m2_encrypted_settings_len = 0;
	INT8U *m2_authenticator = NULL;
	INT8U m2_authenticator_present;

	// "Intermediary data" we also need to extract from M1 to obtain the keys
	// that will let us decrypt the "useful data" from M2
	//
	INT8U *m1_nonce = NULL;
	INT8U m1_nonce_present;
	/*INT8U  *m1_pubkey; */ INT8U m1_pubkey_present;
	INT16U m1_pubkey_len;

	// "Intermediary data" contained in the "key" argument also needed to obtain
	// the keys that will let us decrypt the "useful data" from M2
	//
	INT8U *m1_privkey;
	INT16U m1_privkey_len;
	INT8U *m1_mac;

	k = (struct wscKey *)key;
	m1_privkey = k->key;
	m1_privkey_len = k->key_len;
	m1_mac = k->mac;

	// Extract "intermediary data" from M2
	m2_nonce_present = 0;
	m2_pubkey_present = 0;
	m2_encrypted_settings_present = 0;
	m2_authenticator_present = 0;
	p = m2;
	while (abs(p - m2) < m2_size) {
		INT16U attr_type;
		INT16U attr_len;

		_E2B(&p, &attr_type);
		_E2B(&p, &attr_len);

		if (attr_type == ATTR_REGISTRAR_NONCE) {
			if (attr_len != 16) {
				WARNING("Incorrect length (%d) for ATTR_REGISTRAR_NONCE", attr_len);
				return 0;
			}
			m2_nonce = p;

			p += attr_len;
			m2_nonce_present = 1;
		} else if (attr_type == ATTR_PUBLIC_KEY) {
			m2_pubkey_len = attr_len;
			m2_pubkey = p;

			p += attr_len;
			m2_pubkey_present = 1;
		} else if (attr_type == ATTR_ENCR_SETTINGS) {
			m2_encrypted_settings_len = attr_len;
			m2_encrypted_settings = p;

			p += attr_len;
			m2_encrypted_settings_present = 1;
		} else if (attr_type == ATTR_AUTHENTICATOR) {
			if (attr_len != 8) {
				WARNING("Incorrect length (%d) for ATTR_AUTHENTICATOR", attr_len);
				return 0;
			}
			m2_authenticator = p;

			p += attr_len;
			m2_authenticator_present = 1;
		} else {
			p += attr_len;
		}
	}
	if (0 == m2_nonce_present || 0 == m2_pubkey_present || 0 == m2_encrypted_settings_present || 0 == m2_authenticator_present) {
		WARNING("Missing attributes in the received M2 message");
		return 0;
	}
	// Extract "intermediary data" from M1
	m1_nonce_present = 0;
	m1_pubkey_present = 0;
	p = m1;
	while (abs(p - m1) < m1_size) {
		INT16U attr_type;
		INT16U attr_len;

		_E2B(&p, &attr_type);
		_E2B(&p, &attr_len);

		if (attr_type == ATTR_ENROLLEE_NONCE) {
			if (attr_len != 16) {
				WARNING("Incorrect length (%d) for ATTR_REGISTRAR_NONCE", attr_len);
				return 0;
			}
			m1_nonce = p;

			p += attr_len;
			m1_nonce_present = 1;
		} else if (attr_type == ATTR_PUBLIC_KEY) {
			m1_pubkey_len = attr_len;
			/*m1_pubkey     = p; */

			p += m1_pubkey_len;
			m1_pubkey_present = 1;
		} else {
			p += attr_len;
		}
	}
	if (0 == m1_nonce_present || 0 == m1_pubkey_present) {
		WARNING("Missing attributes in the received M1 message");
		return 0;
	}
	// With all the information we have just extracted from M1 and M2, obtain
	// the authentication/encryption keys.
	{
		INT8U *shared_secret;
		INT16U shared_secret_len;

		INT8U *addr[3];
		INT32U len[3];

		INT8U dhkey[SHA256_MAC_LEN];
		INT8U kdk[SHA256_MAC_LEN];

		INT8U keys[WPS_AUTHKEY_LEN + WPS_KEYWRAPKEY_LEN + WPS_EMSK_LEN];

		// With the enrolle public key (which we received in M1) and our private
		// key (which we generated above), obtain the Diffie Hellman shared
		// secret (when receiving M2, the enrollee will be able to obtain this
		// same shared secret using its private key and ou public key -contained
		// in M2-)
		//
		PLATFORM_COMPUTE_DH_SHARED_SECRET(&shared_secret, &shared_secret_len, m2_pubkey, m2_pubkey_len, m1_privkey, m1_privkey_len);
		// TODO: ZERO PAD the shared_secret (doesn't seem to be really needed
		// though)

		// Next, obtain the SHA-256 digest of this shared secret. We will call
		// this "dhkey"
		//
		addr[0] = shared_secret;
		len[0] = shared_secret_len;

		PLATFORM_SHA256(1, addr, len, dhkey);

		// Next, concatenate three things (the enrolle nonce contained in M1,
		// the enrolle MAC address, and the nonce we just generated before, and
		// calculate its HMAC (hash message authentication code) using "dhkey"
		// as the secret key.
		//
		addr[0] = m1_nonce;
		addr[1] = m1_mac;
		addr[2] = m2_nonce;
		len[0] = 16;
		len[1] = 6;
		len[2] = 16;

		PLATFORM_HMAC_SHA256(dhkey, SHA256_MAC_LEN, 3, addr, len, kdk);

		// Finally, take "kdk" and using a function provided in the "Wi-Fi
		// simple configuration" standard, obtain THREE KEYS that we will use
		// later ("authkey", "keywrapkey" and "emsk")
		//
		_wps_key_derivation_function(kdk, NULL, 0, "Wi-Fi Easy and Secure Key Derivation", keys, sizeof(keys));

		memcpy(authkey, keys, WPS_AUTHKEY_LEN);
		memcpy(keywrapkey, keys + WPS_AUTHKEY_LEN, WPS_KEYWRAPKEY_LEN);
		memcpy(emsk, keys + WPS_AUTHKEY_LEN + WPS_KEYWRAPKEY_LEN, WPS_EMSK_LEN);

		DEBUG("WPS keys: ");
		DEBUG(" Registrar pubkey(%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", m2_pubkey_len, m2_pubkey[0], m2_pubkey[1],
		      m2_pubkey[2], m2_pubkey[m2_pubkey_len - 3], m2_pubkey[m2_pubkey_len - 2], m2_pubkey[m2_pubkey_len - 1]);
		DEBUG(" Enrollee privkey(%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", m1_privkey_len, m1_privkey[0], m1_privkey[1],
		      m1_privkey[2], m1_privkey[m1_privkey_len - 3], m1_privkey[m1_privkey_len - 2], m1_privkey[m1_privkey_len - 1]);
		DEBUG(" Shared secret(%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", shared_secret_len, shared_secret[0],
		      shared_secret[1], shared_secret[2], shared_secret[shared_secret_len - 3], shared_secret[shared_secret_len - 2],
		      shared_secret[shared_secret_len - 1]);
		DEBUG(" DH key( 32 bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", dhkey[0], dhkey[1], dhkey[2], dhkey[29],
		      dhkey[30], dhkey[31]);
		DEBUG(" Enrollee nonce( 16 bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", m1_nonce[0], m1_nonce[1], m1_nonce[2],
		      m1_nonce[13], m1_nonce[14], m1_nonce[15]);
		DEBUG(" Registrar nonce( 16 bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", m2_nonce[0], m2_nonce[1], m2_nonce[2],
		      m2_nonce[13], m2_nonce[14], m2_nonce[15]);
		DEBUG(" KDK( 32 bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", kdk[0], kdk[1], kdk[2], kdk[29], kdk[30],
		      kdk[31]);
		DEBUG(" authkey(%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", authkey[0], authkey[1], authkey[2],
		      authkey[WPS_AUTHKEY_LEN - 3], authkey[WPS_AUTHKEY_LEN - 2], authkey[WPS_AUTHKEY_LEN - 1]);
		DEBUG(" keywrapkey(%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", keywrapkey[0], keywrapkey[1], keywrapkey[2],
		      keywrapkey[WPS_KEYWRAPKEY_LEN - 3], keywrapkey[WPS_KEYWRAPKEY_LEN - 2], keywrapkey[WPS_KEYWRAPKEY_LEN - 1]);
		DEBUG(" emsk(%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", emsk[0], emsk[1], emsk[2], emsk[WPS_EMSK_LEN - 3],
		      emsk[WPS_EMSK_LEN - 2], emsk[WPS_EMSK_LEN - 1]);
		FREE(shared_secret);
	}

	// With the just computed key, check the message authentication
	{
		// Concatenate M1 and M2 (excluding the last 12 bytes, where the
		// authenticator attribute is contained) and calculate the HMAC, then
		// check it against the actual authenticator attribute value.
		//
		INT8U hash[SHA256_MAC_LEN];

		INT8U *addr[2];
		INT32U len[2];

		addr[0] = m1;
		addr[1] = m2;
		len[0] = m1_size;
		len[1] = m2_size - 12;

		PLATFORM_HMAC_SHA256(authkey, WPS_AUTHKEY_LEN, 2, addr, len, hash);

		if (memcmp(m2_authenticator, hash, 8) != 0) {
			WARNING("Message M2 authentication failed");
			return 0;
		}
	}

	// With the just computed keys, decrypt the message and check the keywrap
	{
		INT8U *plain;
		INT32U plain_len;

		INT8U m2_keywrap_present;

		plain = m2_encrypted_settings + AES_BLOCK_SIZE;
		plain_len = m2_encrypted_settings_len - AES_BLOCK_SIZE;

		DEBUG("AP settings before decryption (%d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x",
		      plain_len, plain[0], plain[1], plain[2],
		      plain[plain_len - 3], plain[plain_len - 2],
		      plain[plain_len - 1]);

		DEBUG("IV (%d bytes)                           : 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x",
		      AES_BLOCK_SIZE, m2_encrypted_settings[0],
		      m2_encrypted_settings[1], m2_encrypted_settings[2],
		      m2_encrypted_settings[AES_BLOCK_SIZE - 3],
		      m2_encrypted_settings[AES_BLOCK_SIZE - 2],
		      m2_encrypted_settings[AES_BLOCK_SIZE - 1]);

		PLATFORM_AES_DECRYPT(keywrapkey, m2_encrypted_settings, plain, plain_len);
		DEBUG("AP settings after decryption (%d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x",
		      plain_len, plain[0], plain[1], plain[2],
		      plain[plain_len - 3], plain[plain_len - 2], plain[plain_len - 1]);

		// Remove padding
		plain_len -= plain[plain_len - 1];

		// Parse contents of decrypted settings
		ssid_present = 0;
		bssid_present = 0;
		auth_type_present = 0;
		encryption_type_present = 0;
		network_key_present = 0;
		m2_keywrap_present = 0;
		p = plain;
		while ((INT32U) (abs(p - plain)) < plain_len) {
			INT16U attr_type;
			INT16U attr_len;

			_E2B(&p, &attr_type);
			_E2B(&p, &attr_len);

			if (attr_type == ATTR_SSID) {
				_EnB(&p, ssid, attr_len);
				ssid[attr_len] = 0x00;
				ssid_present = 1;
			} else if (attr_type == ATTR_AUTH_TYPE) {
				_E2B(&p, &auth_type);
				auth_type_present = 1;
			} else if (attr_type == ATTR_ENCR_TYPE) {
				_E2B(&p, &encryption_type);
				encryption_type_present = 1;
			} else if (attr_type == ATTR_NETWORK_KEY) {
				_EnB(&p, network_key, attr_len);
				network_key[attr_len] = 0x00;
				network_key_present = 1;
			} else if (attr_type == ATTR_MAC_ADDR) {
				_EnB(&p, bssid, attr_len);
				bssid_present = 1;
			} else if (attr_type == ATTR_KEY_WRAP_AUTH) {
				// This attribute is always the last one contained in the plain
				// text, thus 4 bytes *before* where "p" is pointing right now
				// is the end of the plain text blob whose HMAC we are going to
				// compute to check the keywrap.
				//
				INT8U *end_of_hmac;
				INT8U hash[SHA256_MAC_LEN];

				INT8U *addr[1];
				INT32U len[1];

				end_of_hmac = p - 4;

				addr[0] = plain;
				len[0] = abs(end_of_hmac - plain);

				PLATFORM_HMAC_SHA256(authkey, WPS_AUTHKEY_LEN, 1, addr, len, hash);

				if (memcmp(p, hash, 8) != 0) {
					WARNING("Message M2 keywrap failed", attr_len);
					return 0;
				}

				p += attr_len;
				m2_keywrap_present = 1;
			} else {
				p += attr_len;
			}
		}
		if ( (ssid_present &
		     bssid_present &
		     auth_type_present &
		     encryption_type_present &
		     network_key_present &
		     m2_keywrap_present) == 0) {
			WARNING("Missing attributes in the configuration settings received in the M2 message");
			return 0;
		}
	}

	// Apply the security settings so that this AP clones the registrar
	// configuration
	char *m1_ifname = DMmacToInterfaceName(m1_mac);

	if (PLATFORM_CONFIGURE_80211_AP(m1_ifname, ssid, bssid, auth_type, encryption_type,
					network_key)) {
		return platform_update_bssid(m1_ifname);
	}

	return 0;
}

//////////////////////////////////////// Registrar functions ///////////////////
INT8U wscBuildM2(INT8U *m1, INT16U m1_size, INT8U **m2, INT16U *m2_size)
{
	INT8U *buffer;

	struct interfaceInfo *x;

	INT8U *p;

	INT8U aux8;
	INT16U var16;
	INT32U aux32;

	INT8U *m1_mac_address = NULL;
	INT8U m1_mac_address_present;
	INT8U *m1_nonce = NULL;
	INT8U m1_nonce_present;
	INT8U *m1_pubkey = NULL;
	INT8U m1_pubkey_present;
	INT16U m1_pubkey_len = 0;
	INT8U m1_rf_band = WPS_RF_24GHZ;

	INT8U *local_privkey;
	INT16U local_privkey_len;

	INT8U authkey[WPS_AUTHKEY_LEN];
	INT8U keywrapkey[WPS_KEYWRAPKEY_LEN];
	INT8U emsk[WPS_EMSK_LEN];

	INT8U registrar_nonce[16];

	char *registrar_interface_name = NULL;

	INT16U encryption_types;
	INT16U auth_types;

	// We first need to extract the following parameters contained in "M1":
	//
	//   - Mac address
	//   - Nounce
	//   - Public key
	//
	m1_mac_address_present = 0;
	m1_nonce_present = 0;
	m1_pubkey_present = 0;
	p = m1;
	while (abs(p - m1) < m1_size) {
		INT16U attr_type;
		INT16U attr_len;

		_E2B(&p, &attr_type);
		_E2B(&p, &attr_len);

		if (attr_type == ATTR_MAC_ADDR) {
			if (attr_len != 6) {
				WARNING("Incorrect length (%d) for ATTR_MAC_ADDR", attr_len);
				return 0;
			}
			m1_mac_address = p;

			p += attr_len;
			m1_mac_address_present = 1;
		} else if (attr_type == ATTR_ENROLLEE_NONCE) {
			if (attr_len != 16) {
				WARNING("Incorrect length (%d) for ATTR_ENROLLEE_NONCE", attr_len);
				return 0;
			}
			m1_nonce = p;

			p += attr_len;
			m1_nonce_present = 1;
		} else if (attr_type == ATTR_PUBLIC_KEY) {
			m1_pubkey_len = attr_len;
			m1_pubkey = p;

			p += attr_len;
			m1_pubkey_present = 1;
		} else if (attr_type == ATTR_RF_BANDS) {
			_E1B(&p, &m1_rf_band);
		} else {
			p += attr_len;
		}
	}
	// If this node is processing an M1 message, it must mean one of our
	// interfaces is the network registrar.
	switch (m1_rf_band) {
	case WPS_RF_24GHZ:
		registrar_interface_name = DMmacToInterfaceName(DMregistrarMacGet(IEEE80211_FREQUENCY_BAND_2_4_GHZ));
		break;
	case WPS_RF_50GHZ:
		registrar_interface_name = DMmacToInterfaceName(DMregistrarMacGet(IEEE80211_FREQUENCY_BAND_5_GHZ));
		break;
	case WPS_RF_60GHZ:
		registrar_interface_name = DMmacToInterfaceName(DMregistrarMacGet(IEEE80211_FREQUENCY_BAND_60_GHZ));
		break;
	}
	INFO("Building M2 from |%s|", registrar_interface_name);
	if (registrar_interface_name == NULL) {
		WARNING("None of this nodes' interfaces matches the registrar MAC address. Ignoring M1 message.");
		return 0;
	}

	if (0 == m1_mac_address_present || 0 == m1_nonce_present || 0 == m1_pubkey_present) {
		WARNING("Imcomplete M1 message received");
		return 0;
	}

	// Now we can build "M2"
	x = PLATFORM_GET_1905_INTERFACE_INFO(registrar_interface_name);
	if (x == NULL) {
		WARNING("Could not retrieve info of interface %s", registrar_interface_name);
		return 0;
	}

	CALLOC(buffer, 1000);
	p = buffer;

	// VERSION
	{
		var16 = ATTR_VERSION;
		_I2B(&var16, &p);
		var16 = 1;
		_I2B(&var16, &p);
		aux8 = 0x10;
		_I1B(&aux8, &p);
	}

	// MESSAGE TYPE
	{
		var16 = ATTR_MSG_TYPE;
		_I2B(&var16, &p);
		var16 = 1;
		_I2B(&var16, &p);
		aux8 = WPS_M2;
		_I1B(&aux8, &p);
	}

	// ENROLLEE NONCE
	{
		var16 = ATTR_ENROLLEE_NONCE;
		_I2B(&var16, &p);
		var16 = 16;
		_I2B(&var16, &p);
		_InB(m1_nonce, &p, 16);
	}

	// REGISTRAR NONCE
	{
		PLATFORM_GET_RANDOM_BYTES(registrar_nonce, 16);

		var16 = ATTR_REGISTRAR_NONCE;
		_I2B(&var16, &p);
		var16 = 16;
		_I2B(&var16, &p);
		_InB(registrar_nonce, &p, 16);
	}

	// UUID
	{
		var16 = ATTR_UUID_R;
		_I2B(&var16, &p);
		var16 = 16;
		_I2B(&var16, &p);
		_InB(x->uuid, &p, 16);
	}

	// PUBLIC KEY
	{
		INT8U *priv, *pub;
		INT16U priv_len = 0, pub_len = 0;

		PLATFORM_GENERATE_DH_KEY_PAIR(&priv, &priv_len, &pub, &pub_len);
		DEBUG("plen|%d|, publen|%d|", priv_len, pub_len);
		// TODO: ZERO PAD the pub key (doesn't seem to be really needed though)

		var16 = ATTR_PUBLIC_KEY;
		_I2B(&var16, &p);
		var16 = pub_len;
		_I2B(&var16, &p);
		_InB(pub, &p, pub_len);

		// We will use it later... save it.
		local_privkey = priv;
		local_privkey_len = priv_len;
		FREE(pub);
	}

	// Key derivation (no bytes are written to the output buffer in the next
	// block of code, we just obtain three cryptographic keys that are needed
	// later
	{
		INT8U *shared_secret;
		INT16U shared_secret_len;

		INT8U *addr[3];
		INT32U len[3];

		INT8U dhkey[SHA256_MAC_LEN];
		INT8U kdk[SHA256_MAC_LEN];

		INT8U keys[WPS_AUTHKEY_LEN + WPS_KEYWRAPKEY_LEN + WPS_EMSK_LEN];

		// With the enrolle public key (which we received in M1) and our private
		// key (which we generated above), obtain the Diffie Hellman shared
		// secret (when receiving M2, the enrollee will be able to obtain this
		// same shared secret using its private key and our public key
		// -contained in M2-)
		//
		PLATFORM_COMPUTE_DH_SHARED_SECRET(&shared_secret, &shared_secret_len, m1_pubkey, m1_pubkey_len, local_privkey, local_privkey_len);
		// TODO: ZERO PAD the shared_secret (doesn't seem to be really needed
		// though)

		// Next, obtain the SHA-256 digest of this shared secret. We will call
		// this "dhkey"
		//
		addr[0] = shared_secret;
		len[0] = shared_secret_len;

		PLATFORM_SHA256(1, addr, len, dhkey);

		// Next, concatenate three things (the enrollee nonce contained in M1,
		// the enrolle MAC address -also contained in M1-, and the nonce we just
		// generated before and calculate its HMAC (hash message authentication
		// code) using "dhkey" as the secret key.
		//
		addr[0] = m1_nonce;
		addr[1] = m1_mac_address;
		addr[2] = registrar_nonce;
		len[0] = 16;
		len[1] = 6;
		len[2] = 16;

		PLATFORM_HMAC_SHA256(dhkey, SHA256_MAC_LEN, 3, addr, len, kdk);

		// Finally, take "kdk" and using a function provided in the "Wi-Fi
		// simple configuration" standard, obtain THREE KEYS that we will use
		// later ("authkey", "keywrapkey" and "emsk")
		//
		_wps_key_derivation_function(kdk, NULL, 0, "Wi-Fi Easy and Secure Key Derivation", keys, sizeof(keys));

		memcpy(authkey, keys, WPS_AUTHKEY_LEN);
		memcpy(keywrapkey, keys + WPS_AUTHKEY_LEN, WPS_KEYWRAPKEY_LEN);
		memcpy(emsk, keys + WPS_AUTHKEY_LEN + WPS_KEYWRAPKEY_LEN, WPS_EMSK_LEN);

		DEBUG("WPS keys: ");
		DEBUG("  Enrollee pubkey   (%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", m1_pubkey_len, m1_pubkey[0], m1_pubkey[1],
		      m1_pubkey[2], m1_pubkey[m1_pubkey_len - 3], m1_pubkey[m1_pubkey_len - 2], m1_pubkey[m1_pubkey_len - 1]);
		DEBUG("  registrar privkey (%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", local_privkey_len, local_privkey[0],
		      local_privkey[1], local_privkey[2], local_privkey[local_privkey_len - 3], local_privkey[local_privkey_len - 2],
		      local_privkey[local_privkey_len - 1]);
		DEBUG("  shared secret     (%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", shared_secret_len, shared_secret[0],
		      shared_secret[1], shared_secret[2], shared_secret[shared_secret_len - 3], shared_secret[shared_secret_len - 2],
		      shared_secret[shared_secret_len - 1]);
		DEBUG("  dh key            ( 32 bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", dhkey[0], dhkey[1], dhkey[2], dhkey[29],
		      dhkey[30], dhkey[31]);
		DEBUG("  enrollee nonce    ( 16 bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", m1_nonce[0], m1_nonce[1], m1_nonce[2],
		      m1_nonce[13], m1_nonce[14], m1_nonce[15]);
		DEBUG("  registrar nonce   ( 16 bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", registrar_nonce[0], registrar_nonce[1],
		      registrar_nonce[2], registrar_nonce[13], registrar_nonce[14], registrar_nonce[15]);
		DEBUG("  kdk               ( 32 bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", kdk[0], kdk[1], kdk[2], kdk[29], kdk[30],
		      kdk[31]);
		DEBUG("  auth_key           (%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", authkey[0], authkey[1], authkey[2],
		      authkey[WPS_AUTHKEY_LEN - 3], authkey[WPS_AUTHKEY_LEN - 2], authkey[WPS_AUTHKEY_LEN - 1]);
		DEBUG("  keywrap_key        (%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", keywrapkey[0], keywrapkey[1], keywrapkey[2],
		      keywrapkey[WPS_KEYWRAPKEY_LEN - 3], keywrapkey[WPS_KEYWRAPKEY_LEN - 2], keywrapkey[WPS_KEYWRAPKEY_LEN - 1]);
		DEBUG("  emsk              (%3d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", emsk[0], emsk[1], emsk[2], emsk[WPS_EMSK_LEN - 3],
		      emsk[WPS_EMSK_LEN - 2], emsk[WPS_EMSK_LEN - 1]);
		FREE(shared_secret);
	}

	// AUTHENTICATION TYPES
	{
		auth_types = 0;

		if (x->interface_type_data.ieee80211.authentication_mode == 0) {
			ERROR("Unsupported authentication mode not sending M2 !!");
			PLATFORM_FREE_1905_INTERFACE_INFO(x);
			return 0;
		}

		get_auth_types(x->interface_type_data.ieee80211.authentication_mode, &auth_types);

		var16 = ATTR_AUTH_TYPE_FLAGS;
		_I2B(&var16, &p);
		var16 = 2;
		_I2B(&var16, &p);
		_I2B(&auth_types, &p);
	}

	// ENCRYPTION TYPES
	{
		encryption_types = 0;

		if (x->interface_type_data.ieee80211.encryption_mode == 0) {
			ERROR("Unsupported encryption mode !!");
			PLATFORM_FREE_1905_INTERFACE_INFO(x);
			return 0;
		}

		if (IEEE80211_ENCRYPTION_MODE_NONE & x->interface_type_data.ieee80211.encryption_mode) {
			encryption_types |= WPS_ENCR_NONE;
		}
		if (IEEE80211_ENCRYPTION_MODE_TKIP & x->interface_type_data.ieee80211.encryption_mode) {
			encryption_types |= WPS_ENCR_TKIP;
		}
		if (IEEE80211_ENCRYPTION_MODE_AES & x->interface_type_data.ieee80211.encryption_mode) {
			encryption_types |= WPS_ENCR_AES;
		}
		var16 = ATTR_ENCR_TYPE_FLAGS;
		_I2B(&var16, &p);
		var16 = 2;
		_I2B(&var16, &p);
		_I2B(&encryption_types, &p);
	}

	// CONNECTION TYPES
	{
		// Two possible types: ESS or IBSS. In the 1905 context, registrars
		// will always be "ESS", meaning they are willing to have their
		// credentials cloned by other APs in order to end up with a network
		// which is "roaming-friendly" ("ESS": "extended service set")

		var16 = ATTR_CONN_TYPE_FLAGS;
		_I2B(&var16, &p);
		var16 = 1;
		_I2B(&var16, &p);
		aux8 = WPS_CONN_ESS;
		_I1B(&aux8, &p);
	}

	// CONFIGURATION METHODS
	{
		// In the 1905 context, the configuration methods the AP is willing to
		// offer will always be these two

		var16 = ATTR_CONFIG_METHODS;
		_I2B(&var16, &p);
		var16 = 2;
		_I2B(&var16, &p);
		var16 = WPS_CONFIG_PHY_PUSHBUTTON | WPS_CONFIG_VIRT_PUSHBUTTON;
		_I2B(&var16, &p);
	}

	// MANUFACTURER
	{
		var16 = ATTR_MANUFACTURER;
		_I2B(&var16, &p);
		var16 = strlen(x->manufacturer_name);
		_I2B(&var16, &p);
		_InB(x->manufacturer_name, &p, strlen(x->manufacturer_name));
	}

	// MODEL NAME
	{
		var16 = ATTR_MODEL_NAME;
		_I2B(&var16, &p);
		var16 = strlen(x->model_name);
		_I2B(&var16, &p);
		_InB(x->model_name, &p, strlen(x->model_name));
	}

	// MODEL NUMBER
	{
		var16 = ATTR_MODEL_NUMBER;
		_I2B(&var16, &p);
		var16 = strlen(x->model_number);
		_I2B(&var16, &p);
		_InB(x->model_number, &p, strlen(x->model_number));
	}

	// SERIAL NUMBER
	{
		var16 = ATTR_SERIAL_NUMBER;
		_I2B(&var16, &p);
		var16 = strlen(x->serial_number);
		_I2B(&var16, &p);
		_InB(x->serial_number, &p, strlen(x->serial_number));
	}

	// PRIMARY DEVICE TYPE
	{
		// In the 1905 context, they node sending a M2 message will always be
		// (at least) a "network router"

		INT8U oui[4] = { 0x00, 0x50, 0xf2, 0x00 };	// Fixed value from the
		// WSC spec

		var16 = ATTR_PRIMARY_DEV_TYPE;
		_I2B(&var16, &p);
		var16 = 8;
		_I2B(&var16, &p);
		var16 = WPS_DEV_NETWORK_INFRA;
		_I2B(&var16, &p);
		_InB(oui, &p, 4);
		var16 = WPS_DEV_NETWORK_INFRA_ROUTER;
		_I2B(&var16, &p);
	}

	// DEVICE NAME
	{
		var16 = ATTR_DEV_NAME;
		_I2B(&var16, &p);
		var16 = strlen(x->device_name);
		_I2B(&var16, &p);
		_InB(x->device_name, &p, strlen(x->device_name));
	}

	// RF BANDS
	{
		INT8U rf_bands;

		rf_bands = 0;

		if (MEDIA_TYPE_IEEE_802_11B_2_4_GHZ == x->interface_type ||
		    MEDIA_TYPE_IEEE_802_11G_2_4_GHZ == x->interface_type || MEDIA_TYPE_IEEE_802_11N_2_4_GHZ == x->interface_type) {
			rf_bands = WPS_RF_24GHZ;
		} else if (MEDIA_TYPE_IEEE_802_11A_5_GHZ == x->interface_type ||
			   MEDIA_TYPE_IEEE_802_11N_5_GHZ == x->interface_type || MEDIA_TYPE_IEEE_802_11AC_5_GHZ == x->interface_type) {
			rf_bands = WPS_RF_50GHZ;
		} else if (x->interface_type == MEDIA_TYPE_IEEE_802_11AD_60_GHZ) {
			rf_bands = WPS_RF_60GHZ;
		}

		var16 = ATTR_RF_BANDS;
		_I2B(&var16, &p);
		var16 = 1;
		_I2B(&var16, &p);
		_I1B(&rf_bands, &p);
	}

	// ASSOCIATION STATE
	{
		var16 = ATTR_ASSOC_STATE;
		_I2B(&var16, &p);
		var16 = 2;
		_I2B(&var16, &p);
		var16 = WPS_ASSOC_CONN_SUCCESS;
		_I2B(&var16, &p);
	}

	// CONFIG ERROR
	{
		var16 = ATTR_CONFIG_ERROR;
		_I2B(&var16, &p);
		var16 = 2;
		_I2B(&var16, &p);
		var16 = WPS_CFG_NO_ERROR;
		_I2B(&var16, &p);
	}

	// DEVICE PASSWORD ID
	{
		var16 = ATTR_DEV_PASSWORD_ID;
		_I2B(&var16, &p);
		var16 = 2;
		_I2B(&var16, &p);
		var16 = DEV_PW_PUSHBUTTON;
		_I2B(&var16, &p);
	}

	// OS VERSION
	{
		// TODO: Fill with actual properties from the interface

		INT32U os_version = 0x00000001;

		var16 = ATTR_OS_VERSION;
		_I2B(&var16, &p);
		var16 = 4;
		_I2B(&var16, &p);
		aux32 = 0x80000000 | os_version;
		_I4B(&aux32, &p);
	}

	// VENDOR EXTENSIONS
	{
		var16 = ATTR_VENDOR_EXTENSION;
		_I2B(&var16, &p);
		var16 = 6;
		_I2B(&var16, &p);
		aux8 = WPS_VENDOR_ID_WFA_1;
		_I1B(&aux8, &p);
		aux8 = WPS_VENDOR_ID_WFA_2;
		_I1B(&aux8, &p);
		aux8 = WPS_VENDOR_ID_WFA_3;
		_I1B(&aux8, &p);
		aux8 = WFA_ELEM_VERSION2;
		_I1B(&aux8, &p);
		aux8 = 1;
		_I1B(&aux8, &p);
		aux8 = WPS_VERSION;
		_I1B(&aux8, &p);
	}

	// ENCRYPTED SETTINGS
	{
		// This is what we are going to do next:
		//
		//   1. Fill a tmp buffer ("aux") ith credential attributes (SSID,
		//      network key, etc...)
		//
		//   2. Add an HMAC to this tmp buffer (so that the enrollee, when
		//      receiving this buffer in M2, can be sure no one has tampered
		//      with these attributes)
		//
		//   3. Encryp the message + HMAC with AES (so that no one else
		//      snooping can have a look at these attributes)

		INT8U plain[200];
		INT8U hash[SHA256_MAC_LEN];
		INT8U *iv_start;
		INT8U *data_start;
		INT8U *r;
		INT8U pad_elements_nr;

		INT8U *addr[1];
		INT32U len[1];

		char *ssid;
		char *network_key;

		ssid = x->interface_type_data.ieee80211.ssid;
		network_key = x->interface_type_data.ieee80211.network_key;

		r = plain;

		// SSID
		var16 = ATTR_SSID;
		_I2B(&var16, &r);
		var16 = strlen(ssid);
		_I2B(&var16, &r);
		_InB(ssid, &r, strlen(ssid));

		// AUTH TYPE
		var16 = ATTR_AUTH_TYPE;
		_I2B(&var16, &r);
		var16 = 2;
		_I2B(&var16, &r);
		var16 = auth_types;
		_I2B(&var16, &r);

		// ENCRYPTION TYPE
		var16 = ATTR_ENCR_TYPE;
		_I2B(&var16, &r);
		var16 = 2;
		_I2B(&var16, &r);
		var16 = encryption_types;
		_I2B(&var16, &r);

		// NETWORK KEY
		var16 = ATTR_NETWORK_KEY;
		_I2B(&var16, &r);
		var16 = strlen(network_key);
		_I2B(&var16, &r);
		_InB(network_key, &r, strlen(network_key));

		// MAC ADDR
		var16 = ATTR_MAC_ADDR;
		_I2B(&var16, &r);
		var16 = 6;
		_I2B(&var16, &r);
		_InB(x->mac_address, &r, 6);

		INFO("AP configuration settings that we are going to send:");
		INFO("  - SSID            : %s", ssid);
		INFO("  - BSSID           : %02x:%02x:%02x:%02x:%02x:%02x",
		     x->mac_address[0], x->mac_address[1],
		     x->mac_address[2], x->mac_address[3],
		      x->mac_address[4], x->mac_address[5]);
		INFO("  - AUTH_TYPE       : 0x%04x", auth_types);
		INFO("  - ENCRYPTION_TYPE : 0x%04x", encryption_types);
		INFO("  - NETWORK_KEY     : %s", network_key);

		// Obtain the HMAC of the whole plain buffer using "authkey" as the
		// secret key.
		//
		addr[0] = plain;
		len[0] = abs(r - plain);
		PLATFORM_HMAC_SHA256(authkey, WPS_AUTHKEY_LEN, 1, addr, len, hash);

		// ...and add it to the same plain buffer (well, only the first 8 bytes
		// of the hash)
		//
		var16 = ATTR_KEY_WRAP_AUTH;
		_I2B(&var16, &r);
		var16 = 8;
		_I2B(&var16, &r);
		_InB(hash, &r, 8);

		// Finally, encrypt everything with AES and add the resulting blob to
		// the M2 buffer, as an "ATTR_ENCR_SETTINGS" attribute
		//
		//// Pad the length of the message to encrypt to a multiple of
		//// AES_BLOCK_SIZE. The new padded bytes must have their value equal to
		//// the amount of bytes padded (PKCS#5 v2.0 pad)
		////
		pad_elements_nr = AES_BLOCK_SIZE - (abs(r - plain) % AES_BLOCK_SIZE);
		for (aux8 = 0; aux8 < pad_elements_nr; aux8++) {
			_I1B(&pad_elements_nr, &r);
		}
		//// Add the attribute header ("type" and "lenght") to the M2 buffer,
		//// followed by the IV and the data to encrypt.
		////
		var16 = ATTR_ENCR_SETTINGS;
		_I2B(&var16, &p);
		var16 = AES_BLOCK_SIZE + abs(r - plain);
		_I2B(&var16, &p);
		iv_start = p;
		PLATFORM_GET_RANDOM_BYTES(p, AES_BLOCK_SIZE);
		p += AES_BLOCK_SIZE;
		data_start = p;
		_InB(plain, &p, abs(r - plain));
		//// Encrypt the data IN-PLACE. Note that the "ATTR_ENCR_SETTINGS"
		//// attribute containes both the IV and the encrypted data.
		////
		INT32U setting_len = abs(r - plain);

		DEBUG("AP settings before encryption (%d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", setting_len, data_start[0], data_start[1],
		      data_start[2], data_start[setting_len - 3], data_start[setting_len - 2], data_start[setting_len - 1]);
		DEBUG("IV (%d bytes)                           : 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", AES_BLOCK_SIZE, iv_start[0], iv_start[1],
		      iv_start[2], iv_start[AES_BLOCK_SIZE - 3], iv_start[AES_BLOCK_SIZE - 2], iv_start[AES_BLOCK_SIZE - 1]);
		PLATFORM_AES_ENCRYPT(keywrapkey, iv_start, data_start, setting_len);
		DEBUG("AP settings after  encryption (%d bytes): 0x%02x, 0x%02x, 0x%02x, ..., 0x%02x, 0x%02x, 0x%02x", setting_len, data_start[0], data_start[1],
		      data_start[2], data_start[setting_len - 3], data_start[setting_len - 2], data_start[setting_len - 1]);
	}

	// AUTHENTICATOR
	{
		// This one is easy: concatenate M1 and M2 (everything in the M2 buffer
		// up to this point) and calculate the HMAC, then append it to M2 as
		// a new (and final!) attribute
		//
		INT8U hash[SHA256_MAC_LEN];

		INT8U *addr[2];
		INT32U len[2];

		addr[0] = m1;
		addr[1] = buffer;
		len[0] = m1_size;
		len[1] = abs(p - buffer);

		PLATFORM_HMAC_SHA256(authkey, WPS_AUTHKEY_LEN, 2, addr, len, hash);

		var16 = ATTR_AUTHENTICATOR;
		_I2B(&var16, &p);
		var16 = 8;
		_I2B(&var16, &p);
		_InB(hash, &p, 8);
	}

	PLATFORM_FREE_1905_INTERFACE_INFO(x);

	*m2 = buffer;
	*m2_size = abs(p - buffer);

	FREE(local_privkey);
	return 1;
}

INT8U wscFreeM2(INT8U *m, INT16U m_size)
{
	if (0 == m_size || NULL == m) {
		return 1;
	}

	FREE(m);
	return 1;
}

//
//////////////////////////////////////// Common functions //////////////////////
//
INT8U wscGetType(INT8U *m, INT16U m_size)
{
	INT8U *p;

	p = m;
	while (abs(p - m) < m_size) {
		INT16U attr_type;
		INT16U attr_len;
		INT8U msg_type;

		_E2B(&p, &attr_type);
		_E2B(&p, &attr_len);

		if (attr_type == ATTR_MSG_TYPE) {
			if (attr_len != 1) {
				WARNING("Incorrect length (%d) for ATTR_MSG_TYPE", attr_len);
				return WSC_TYPE_UNKNOWN;
			}
			_E1B(&p, &msg_type);

			if (msg_type == WPS_M1) {
				return WSC_TYPE_M1;
			} else if (msg_type == WPS_M2) {
				return WSC_TYPE_M2;
			} else {
				return WSC_TYPE_UNKNOWN;
			}
		} else {
			p += attr_len;
		}
	}

	return WSC_TYPE_UNKNOWN;
}
