/*
 * lldp_payload.c: IEEE-1905 lldp payload handling
 *
 * Copyright (c) 2017, Broadband Forum
 * Copyright (C) 2019, IOPSYS Software Solutions AB
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Subject to the terms and conditions of this license, each copyright
 * holder and contributor hereby grants to those receiving rights under
 * this license a perpetual, worldwide, non-exclusive, no-charge,
 * royalty-free, irrevocable (except for failure to satisfy the
 * conditions of this license) patent license to make, have made, use,
 * offer to sell, sell, import, and otherwise transfer this software,
 * where such license applies only to those patent claims, already
 * acquired or hereafter acquired, licensable by such copyright holder or
 * contributor that are necessarily infringed by:
 * (a) their Contribution(s) (the licensed copyrights of copyright holders
 *     and non-copyrightable additions of contributors, in source or binary
 *     form) alone; or
 *
 * (b) combination of their Contribution(s) with the work of authorship to
 *     which such Contribution(s) was added by such copyright holder or
 *     contributor, if, at the time the Contribution is added, such addition
 *     causes such combination to be necessarily infringed. The patent
 *     license shall not apply to any other combinations which include the
 *     Contribution.
 *
 * Except as expressly stated above, no rights or licenses from any
 * copyright holder or contributor is granted under this license, whether
 * expressly, by implication, estoppel or otherwise.
 *
 * DISCLAIMER
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include "utils.h"
#include "lldp_payload.h"
#include "lldp_tlvs.h"

////////////////////////////////////////////////////////////////////////////////
// Actual API functions
////////////////////////////////////////////////////////////////////////////////

struct PAYLOAD *parse_lldp_PAYLOAD_from_packet(INT8U *packet_stream)
{
	struct PAYLOAD *ret;

	INT8U *p;
	INT8U i, j;

	if (packet_stream == NULL) {
		return NULL;
	}

	CALLOC(ret, 1);
	for (i = 0; i < MAX_LLDP_TLVS; i++) {
		ret->tlv_cstruct_list[i] = NULL;
	}

	p = packet_stream;
	i = 0;

	while (1) {
		INT8U *tlv;

		INT8U byte1, byte2;
		INT16U len;

		tlv = parse_lldp_TLV_from_packet(p);

		if (NULL == tlv || MAX_LLDP_TLVS == i) {
			// Parsing error or too many TLVs
			//
			for (j = 0; j < i; j++) {
				free_lldp_TLV_structure(ret->tlv_cstruct_list[j]);
			}
			FREE(ret);
			return NULL;
		}
		// The first byte of the TLV structure always contains the TLV type.
		// We need to check if we have reach the "end of LLPDPDU" TLV (ie. the
		// last one)
		//
		if (*tlv == TLV_TYPE_END_OF_LLDPPDU) {
			free_lldp_TLV_structure(tlv);
			break;
		} else {
			ret->tlv_cstruct_list[i] = tlv;
		}

		// All LLDP TLVs start with the same two bytes:
		//
		//   |byte #1 |byte #2 |
		//   |--------|--------|
		//   |TTTTTTTL|LLLLLLLL|
		//   |--------|--------|
		//    <-----><-------->
		//    7 bits   9 bits
		//    (type)   (lenght)
		//
		// We are interested in the length to find out how much we should
		// "advance" the stream pointer each time

		_E1B(&p, &byte1);
		_E1B(&p, &byte2);

		len = ((byte1 & 0x1) << 8) + byte2;

		p += len;
		i++;
	}

	// Before returning, we must make sure that this packet contained all the
	// needed TLVs (ie. "chassis ID", "port ID" and "time to live")
	//
	{
		INT8U chassis_id = 0;
		INT8U port_id = 0;
		INT8U time_to_live = 0;

		for (j = 0; j < i; j++) {
			if (TLV_TYPE_CHASSIS_ID == *(ret->tlv_cstruct_list[j])) {
				chassis_id++;
			} else if (TLV_TYPE_PORT_ID == *(ret->tlv_cstruct_list[j])) {
				port_id++;
			} else if (TLV_TYPE_TIME_TO_LIVE == *(ret->tlv_cstruct_list[j])) {
				time_to_live++;
			}
		}

		if (1 != chassis_id || 1 != port_id || 1 != time_to_live) {
			// There are too many (or too few) TLVs of one of the required types
			//
			for (j = 0; j < i; j++) {
				free_lldp_TLV_structure(ret->tlv_cstruct_list[j]);
			}
			FREE(ret);
			return NULL;
		}
	}

	return ret;
}

INT8U *forge_lldp_PAYLOAD_from_structure(struct PAYLOAD *memory_structure, INT16U *len)
{
	INT8U i;

	struct chassisIdTLV *x = NULL;
	struct portIdTLV *y = NULL;
	struct timeToLiveTypeTLV *z = NULL;

	INT8U *stream;
	INT16U stream_len;

	INT8U *buffer;
	INT16U total_len;

	struct endOfLldppduTLV end_of_lldppdu_tlv = {.tlv_type = TLV_TYPE_END_OF_LLDPPDU };

	// First of all, make sure that the provided PAYLOAD structure contains one
	// (and only one) of the required TLVs (ie. "chassis ID", "port ID" and
	// "time to live") and nothing else.
	//
	{
		INT8U chassis_id = 0;
		INT8U port_id = 0;
		INT8U time_to_live = 0;

		for (i = 0; i < MAX_LLDP_TLVS; i++) {
			if (memory_structure->tlv_cstruct_list[i] == NULL) {
				// No more TLVs
				//
				break;
			}

			if (TLV_TYPE_CHASSIS_ID == *(memory_structure->tlv_cstruct_list[i])) {
				chassis_id++;
				x = (struct chassisIdTLV *)memory_structure->tlv_cstruct_list[i];
			} else if (TLV_TYPE_PORT_ID == *(memory_structure->tlv_cstruct_list[i])) {
				port_id++;
				y = (struct portIdTLV *)memory_structure->tlv_cstruct_list[i];
			} else if (TLV_TYPE_TIME_TO_LIVE == *(memory_structure->tlv_cstruct_list[i])) {
				time_to_live++;
				z = (struct timeToLiveTypeTLV *)memory_structure->tlv_cstruct_list[i];
			} else {
				// Unexpected TLV!
				//
				return NULL;
			}
		}

		if (1 != chassis_id || 1 != port_id || 1 != time_to_live) {
			// There are too many (or too few) TLVs of one of the required types
			//
			return NULL;
		}
	}

	// Prepare the output buffer
	CALLOC(buffer, MAX_NETWORK_SEGMENT_SIZE);

	// Next, from each of the just filled structures, obtain its packet
	// representation (ie. bit stream layout) and fill 'buffer' with them in
	// order
	//
	total_len = 0;

	stream = forge_lldp_TLV_from_structure((INT8U *) x, &stream_len);
	if (stream == NULL) {
		// Could not forge the packet. Error?
		//
		WARNING("forge_lldp_TLV_from_structure(\"chassis ID\") failed!");
		FREE(buffer);
		return NULL;
	}
	memcpy(buffer + total_len, stream, stream_len);
	FREE(stream);
	total_len += stream_len;

	stream = forge_lldp_TLV_from_structure((INT8U *) y, &stream_len);
	if (stream == NULL) {
		// Could not forge the packet. Error?
		//
		WARNING("forge_lldp_TLV_from_structure(\"port ID\") failed!");
		FREE(buffer);
		return NULL;
	}
	memcpy(buffer + total_len, stream, stream_len);
	FREE(stream);
	total_len += stream_len;

	stream = forge_lldp_TLV_from_structure((INT8U *) z, &stream_len);
	if (stream == NULL) {
		// Could not forge the packet. Error?
		//
		WARNING("forge_lldp_TLV_from_structure(\"time to live\") failed!");
		FREE(buffer);
		return NULL;
	}
	memcpy(buffer + total_len, stream, stream_len);
	FREE(stream);
	total_len += stream_len;

	stream = forge_lldp_TLV_from_structure((INT8U *) &end_of_lldppdu_tlv, &stream_len);
	if (stream == NULL) {
		// Could not forge the packet. Error?
		//
		WARNING("forge_lldp_TLV_from_structure() failed!");
		FREE(buffer);
		return NULL;
	}
	memcpy(buffer + total_len, stream, stream_len);
	FREE(stream);
	total_len += stream_len;

	*len = total_len;

	return buffer;
}

void free_lldp_PAYLOAD_structure(struct PAYLOAD *memory_structure)
{
	INT8U i;

	i = 0;
	while (memory_structure->tlv_cstruct_list[i]) {
		free_lldp_TLV_structure(memory_structure->tlv_cstruct_list[i]);
		i++;
	}

	FREE(memory_structure);

	return;
}

INT8U compare_lldp_PAYLOAD_structures(struct PAYLOAD *memory_structure_1, struct PAYLOAD *memory_structure_2)
{
	INT8U i;

	if (NULL == memory_structure_1 || NULL == memory_structure_2) {
		return 1;
	}

	i = 0;
	while (1) {
		if (NULL == memory_structure_1->tlv_cstruct_list[i] && NULL == memory_structure_2->tlv_cstruct_list[i]) {
			// No more TLVs to compare! Return '0' (structures are equal)
			//
			return 0;
		}

		if (compare_lldp_TLV_structures(memory_structure_1->tlv_cstruct_list[i], memory_structure_2->tlv_cstruct_list[i]) != 0) {
			// TLVs are not the same
			//
			return 1;
		}

		i++;
	}

	// This point should never be reached
	//
	return 1;
}

void visit_lldp_PAYLOAD_structure(struct PAYLOAD *memory_structure,
				  PRINT_CALLBACK callback, PRINT_WRITER write_function, const char *prefix, struct blob_buf *bb)
{
	// Buffer size to store a prefix string that will be used to show each
	// element of a structure on screen
	//
#define MAX_PREFIX  100

	INT8U i;

	if (memory_structure == NULL) {
		return;
	}

	i = 0;
	while (memory_structure->tlv_cstruct_list[i] != NULL) {
		// In order to make it easier for the callback() function to present
		// useful information, append the type of the TLV to the prefix
		//
		char new_prefix[MAX_PREFIX];

		switch (*(memory_structure->tlv_cstruct_list[i])) {
		case TLV_TYPE_END_OF_LLDPPDU:
			{
				PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "%sTLV(END_OF_LLDPPDU)", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;
				break;
			}

		case TLV_TYPE_CHASSIS_ID:
			{
				PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "%sTLV(CHASSIS_ID)", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;
				break;
			}

		case TLV_TYPE_PORT_ID:
			{
				PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "%sTLV(PORT_ID)", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;
				break;
			}

		case TLV_TYPE_TIME_TO_LIVE:
			{
				PLATFORM_SNPRINTF(new_prefix, MAX_PREFIX - 1, "%sTLV(TIME_TO_LIVE)", prefix, i);
				new_prefix[MAX_PREFIX - 1] = 0x0;
				break;
			}

		default:
			{
				// Unknown TLV. Ignore.
				break;
			}
		}

		visit_lldp_TLV_structure(memory_structure->tlv_cstruct_list[i], callback, write_function, new_prefix, bb);
		i++;
	}

	return;
}
